
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/debugfs.h>
#include <linux/ltt-tracer.h>
#include <linux/notifier.h>
#include <linux/jiffies.h>
#include <linux/marker.h>
#include <linux/oom.h>


struct marker_probe_array {
	struct rcu_head rcu;
	struct marker_probe_closure c[0];
};

/**
 * __mark_empty_function - Empty probe callback
 * @mdata: marker data
 * @probe_private: probe private data
 * @call_private: call site private data
 * @fmt: format string
 * @...: variable argument list
 *
 * Empty callback provided as a probe to the markers. By providing this to a
 * disabled marker, we make sure the  execution flow is always valid even
 * though the function pointer change and the marker enabling are two distinct
 * operations that modifies the execution flow of preemptible code.
 */
notrace void __mark_empty_function_brocade(const struct marker *mdata,
	void *probe_private, void *call_private, const char *fmt, va_list *args)
{
}
EXPORT_SYMBOL(__mark_empty_function_brocade);

/*
 * marker_probe_cb Callback that prepares the variable argument list for probes.
 * @mdata: pointer of type struct marker
 * @call_private: caller site private data
 * @...:  Variable argument list.
 *
 * Since we do not use "typical" pointer based RCU in the 1 argument case, we
 * need to put a full smp_rmb() in this branch. This is why we do not use
 * rcu_dereference() for the pointer read.
 */
notrace void marker_probe_cb_brocade(const struct marker *mdata,
		void *call_private, ...)
{
	va_list args;
	char ptype;

	/*
	 * rcu_read_lock_sched does two things : disabling preemption to make
	 * sure the teardown of the callbacks can be done correctly when they
	 * are in modules and they insure RCU read coherency.
	 */
	rcu_read_lock_sched_notrace();
	ptype = mdata->ptype;
	if (likely(!ptype)) {
		marker_probe_func *func;
		/* Must read the ptype before ptr. They are not data dependant,
		 * so we put an explicit smp_rmb() here. */
		smp_rmb();
		func = mdata->single.func;
		/* Must read the ptr before private data. They are not data
		 * dependant, so we put an explicit smp_rmb() here. */
		smp_rmb();
		va_start(args, call_private);
		func(mdata, mdata->single.probe_private, call_private,
			mdata->format, &args);
		va_end(args);
	} else {
		struct marker_probe_array *multi;
		int i;
		/*
		 * Read mdata->ptype before mdata->multi.
		 */
		smp_rmb();
		multi = mdata->multi;
		/*
		 * multi points to an array, therefore accessing the array
		 * depends on reading multi. However, even in this case,
		 * we must insure that the pointer is read _before_ the array
		 * data. Same as rcu_dereference, but we need a full smp_rmb()
		 * in the fast path, so put the explicit barrier here.
		 */
		smp_read_barrier_depends();
		for (i = 0; multi->c[i].func; i++) {
			va_start(args, call_private);
			multi->c[i].func(mdata, multi->c[i].probe_private,
				call_private, mdata->format, &args);
			va_end(args);
		}
	}
	rcu_read_unlock_sched_notrace();
}
EXPORT_SYMBOL(marker_probe_cb_brocade);

/*
 * Export oom callbacks
 */
#include <linux/oom.h>
int
register_oom_notifier_brocade(struct notifier_block *nb)
{
	return (register_oom_notifier(nb));
}
EXPORT_SYMBOL(register_oom_notifier_brocade);

int
unregister_oom_notifier_brocade(struct notifier_block *nb)
{
	return (unregister_oom_notifier(nb));

}
EXPORT_SYMBOL(unregister_oom_notifier_brocade);
