// Add definitions of some utility functions to be used for active/standby
// failover code.


#include <asm/silkworm_40x0_misc.h>
#include <asm/prom.h>
#include <asm/io.h>
#include <asm/fsl_hcalls.h>
#include <linux/fsl_hypervisor.h>
#include <linux/delay.h>



bool silkworm_is_ha_config(void){
    struct device_node * root = of_find_node_by_path("/");
    int i=0;        

    i = of_device_is_compatible(root,"ha-failover-config-hv");
    of_node_put(root);

    return i ? true:false;
}
bool silkworm_is_plat_wmem_config(void){
    struct device_node * root = of_find_node_by_path("/");
    int i=0;        
  
    if(has_fsl_hypervisor()) {
        i = of_device_is_compatible(root,"ha-plat-wmem-config-hv");
    } else {
        i = of_device_is_compatible(root,"ha-plat-wmem-config");
    }
    of_node_put(root);

    return i ? true:false;
}
// return STATUS_NA if active/standby status is not relevant, e.g. on MM or not
// running with HV.
// return STATUS_ACTIVE  for ACTIVE
// return STATUS_STANDBY for STANDBY
// returns the information from the device tree held with HV.
silkworm_partition_status silkworm_get_partition_status_by_handle(int handle){
	if(!has_fsl_hypervisor())
		return STATUS_NA;

	else if(!silkworm_is_ha_config())
		return STATUS_UNKNOWN;

	else{
		const char * path = "/chosen";
		const char * prop = "status";
		char val[32];
		uint32_t len = sizeof(val);
		int r;
		
		if((r = fh_partition_get_dtprop(handle,          // handle = -1 implies self
			virt_to_phys((void*)path),
			virt_to_phys((void*)prop),
			virt_to_phys(val),
			&len)) == 0){
				if(strcmp(val,"active") == 0)
					return STATUS_ACTIVE;
				else if(strcmp(val,"standby") == 0)
					return STATUS_STANDBY;
			}
	}
	return  STATUS_UNKNOWN;
}


// return STATUS_NA if active/standby status is not relevant, e.g. on MM or not
// running with HV.
// return STATUS_ACTIVE  for ACTIVE
// return STATUS_STANDBY for STANDBY
// returns the information from the device tree held with HV.
silkworm_partition_status silkworm_get_partition_status(void){
	// handle = -1 implies self
	return(silkworm_get_partition_status_by_handle(-1));
}



// return the partition id of self - either 0 or 1.
// should be used by code that need this infromation for any purpose.
// -1 on error e.g. use in case w/o HV.
// partition id is read from /chosen/label which is of the following format
// p<id>-linux.
int silkworm_get_partition_id(void){

    struct device_node * n = of_find_node_by_path("/");
    struct property * p;
    int l;
    int pid = -1;

    if(!has_fsl_hypervisor()){
        return pid;
    }

    if(!n){
        return pid;
    }

    p = of_find_property(n,"label",&l);
    if(!p){
        return pid;
    }

    sscanf(p->value,"p%d-linux",&pid);
    return pid;
}


static struct property sp;
static char name[32];
static char value[32];

// set the current status of partition as active/standby
// status = STATUS_ACTIVE or STATUS_STANDBY
// who = 0 implies self, who = 1 implies the other partition as applicable.
// The function sets the status of the partition in the device tree with HV.
// in case of self, it also sets the status in the GOS local device tree.
// should be called after a partition has gone active from standby status for
// self and the other partition before rebooting it.
int silkworm_set_partition_status(silkworm_partition_status status, int who){
    char val[32];
    const char * path = "/chosen";
    const char * prop = "status";

    sp.name = name;
    sp.value=value;

    if(status == STATUS_ACTIVE)
        strcpy(val,"active");
    else if (status == STATUS_STANDBY)
        strcpy(val,"standby");
    else
        return -1;        

    if(!has_fsl_hypervisor())
        return -1;
    else if(!silkworm_is_ha_config())
        return -1;
    else if(who == 0){
        int r;
   		if((r = fh_partition_set_dtprop(-1,          // handle = -1 implies self
						    virt_to_phys((void*)path),
						    virt_to_phys((void*)prop),
                            virt_to_phys(val),
						    strlen(val)+1) == 0)){
            // set the local device tree status
            int l;
            struct device_node * n = of_find_node_by_path("/chosen");
            struct property * p = of_find_property(n,"status",&l);
            strcpy(sp.value,val);
            strcpy(sp.name,p->name);
            sp.length = strlen(val) + 1;
            sp._flags = p->_flags;
            sp.unique_id = p->unique_id;
            prom_update_property(n,&sp,p);
            of_node_put(n);
            return 0;
        }
    }else if(who == 1){
        // find the partition handle from device tree, then modofy it
        struct device_node * n = of_find_compatible_node(0,0,"fsl,hv-partition-handle");
        struct property * p;
        int l,handle,r;
        if(!n)
            return -1;
        p = of_find_property(n,"reg",&l);
        if(!p)
            return -1;
        handle = *((int*)p->value);
        if((r = fh_partition_set_dtprop(handle, 
						    virt_to_phys((void*)path),
						    virt_to_phys((void*)prop),
                            virt_to_phys(val),
						    strlen(val)+1) == 0))
            return 0;
    }
    return -1;
}


// claims the localbus,cpld,norflash
// retuns success /faliure
int silkworm_claim_localbus(){


    struct device_node *node=NULL;
    int ret=0;

   if(!silkworm_is_ha_config())
        return -1;
   
	 node=of_find_compatible_node(0,0,"fsl,elbc");
    if(node==NULL){
                printk(KERN_ERR "\n   not able to find the device with local bus\n");
                return -1;
        }
    ret=fsl_hv_claim(node);
    if (ret) {
			pr_err("Failed to claim localbus device!\n");
			of_node_put(node);
			return -1;
	}
    of_node_put(node);

    node=of_find_node_by_type(NULL, "cpld"); 
    if(node==NULL){
                printk(KERN_ERR "\n  not able to find the device with local buscpld \n");
		of_node_put(node);
                return -1;
    }

    ret=fsl_hv_claim(node);
    if (ret) {
			pr_err("Failed to claim cpld device!\n");
			of_node_put(node);
			return -1;
		}
    of_node_put(node);
    return 0;

}


// read device tree and get the the virtual irq for IRQ 0
// returns irq number

int silkworm_lc_get_cpldirq(){
    struct device_node *node=NULL;
    int err_irq=0;

	
    if(!silkworm_is_ha_config())
        return -1;
	
    node=of_find_node_by_type(NULL, "cpld"); 
    if(node==NULL){
                printk(KERN_ERR "\n   not able to find the device with local buscpld \n");
		of_node_put(node);
                return -1;
        }

    err_irq = of_irq_to_resource(node, 0, NULL);
    if (err_irq == NO_IRQ) {
		printk(KERN_ERR "\n   of_irq_to_resource not able to find the device with local buscpld \n");
		pr_info("Can't get %s property '%s'\n", node->full_name,
			"interrupts");
		of_node_put(node);
		return -ENODEV;
	}

    of_node_put(node);

    return err_irq;
}

#if defined(CONFIG_SILKWORM_LC) || defined(CONFIG_SILKWORM_MM) || defined(CONFIG_SILKWORM_PB) || defined(CONFIG_SILKWORM_EMBEDDED)

/*
 * silkworm_machine_check()
 *
 * Need to allow drivers an opportunity to deal with errors that cause
 * machine checks to occur. To do that, we must not initiate a kernal
 * panic on the first MC that occurs.
 *
 * This function will count the number of MCs that occur within a
 * specfied window of time. Until that threshold is exceeded, this function
 * will advance the NIP and return "recoverable" status.
 *
 * Once the threshold is exceeded, then return "unrecoverable" status to
 * allow a kernel panic to occur.
 *
 */
#define PCI_ERR_TIME_WINDOW (15 * HZ) /* 15 secs */
// #define PCI_ERR_TIME_WINDOW (60 * HZ) /* 60 secs */
// #define PCI_ERR_THRESHOLD 50
#define PCI_ERR_THRESHOLD 75
// #define PCI_ERR_THRESHOLD 500
// #define PCI_ERR_THRESHOLD 100

static	int	mc_count = 0;
static int mc_threshold = PCI_ERR_THRESHOLD;
static unsigned long base_jiff = 0;	/* Time in jiffies since we started counting */
static int mc_enable = 0;

int
silkworm_machine_check(struct pt_regs *regs, unsigned long reason)
{
	u32 core;
	int recoverable = 0;
	unsigned long curr_jiff = jiffies;
	static u32 stack_dumped = 0;

	if (mc_enable == 0){
		recoverable = 0;
		goto Out2;
	}

	if ((reason & MCSR_LD)||(reason & MCSR_LDG)) {

		/*
		 * Show who caused the MC on the 1st MC only to prevent
		 * flooding the console in the likely event of multiple errors
		 */
		core = mfspr(SPRN_PIR) & 0x3fff;
		if (!stack_dumped) {
			int orig_loglevel;

			orig_loglevel = console_loglevel;
			console_loglevel = 8;
			printk("%s: CORE %d\n", __FUNCTION__,
				   core);
			show_regs(regs);
			console_loglevel = orig_loglevel;
			stack_dumped = 1;
		}
		

		mc_count++;
		if (time_after_eq(base_jiff + PCI_ERR_TIME_WINDOW, curr_jiff)) {

			/*
			 * if specified time window has not been exceeded, then
			 * see if we have exceeded the event threshold.
			 */

			if (mc_count >= mc_threshold) {

				/*
				 * We've had too many MCs within the defined time
				 * window. Whatever driver is supposed to be dealing
				 * with the problem isn't doing so, so consider this
				 * event to be un-recoverable.
				 */
				printk(KERN_ERR "%s: Excessive machine checks %d threshold %d\n",
					   __FUNCTION__, mc_count, mc_threshold);

				recoverable = 0;

				/*
				 * Start a new interval in case we are called again. This
				 * will ensure we keep returning 0 (unrecoverable).
				 */
				base_jiff = jiffies; 
				goto Out;
			}
		} else {

				/*
				 * The errors did not happen within the defined
				 * time window, so we start over.
				 */
				base_jiff = jiffies;
				mc_count = 1;
		}

		core = mfspr(SPRN_PIR) & 0x3fff;
#if 0
		printk(KERN_ERR "%s %d: Advancing NIP for core %d\n",
			   __FUNCTION__, __LINE__, core);
#endif
		/* Let error handlers take care of whatever caused this */
		regs->nip += 4;
		recoverable = 1;
	}

 Out:
	mdelay(20);

 Out2:
	return(recoverable);
}

void
silkworm_machine_check_control(int enable, int threshold)
{
	if (enable) {
		mc_enable = 1;
		if (threshold > 0) {
			mc_threshold = threshold;
		}
		base_jiff = jiffies;

	} else {
		mc_enable = 0;
	}

}

#endif
// read device tree and get the the virtual irq for IRQ 11
// returns irq number

int silkworm_lc_get_phyirq() {
	struct device_node *node = NULL;
	int err_irq = 0;
	
	if(!silkworm_is_ha_config())
		return -1;
	
	node=of_find_node_by_type(NULL, "cpld"); 
	if(node == NULL) {
		printk(KERN_ERR
			"\n   not able to find the device with local buscpld\n");
		of_node_put(node);
		return -1;
	}

	// NOTE: IRQ 11 is the second entry in the IRQ list!
	// IRQs 1 through 10 are not available.
	err_irq = of_irq_to_resource(node, 1, NULL);

	if (err_irq == NO_IRQ) {
		printk(KERN_ERR
			"\n   of_irq_to_resource not able to find the device "
			"with local buscpld \n");
		pr_info("Can't get %s property '%s'\n", node->full_name, "interrupts");
		of_node_put(node);
		return -ENODEV;
	}

	of_node_put(node);

	return err_irq;
}


EXPORT_SYMBOL(silkworm_claim_localbus);
EXPORT_SYMBOL(silkworm_is_ha_config);
EXPORT_SYMBOL(silkworm_get_partition_status);
EXPORT_SYMBOL(silkworm_get_partition_id);
EXPORT_SYMBOL(silkworm_set_partition_status);
EXPORT_SYMBOL(silkworm_lc_get_cpldirq);
EXPORT_SYMBOL(silkworm_lc_get_phyirq);
EXPORT_SYMBOL(silkworm_machine_check_control);
EXPORT_SYMBOL(silkworm_is_plat_wmem_config);
