/*
 * Embedded platform kernel support.
 */

#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/phy.h>

#include <asm/system.h>
#include <asm/time.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <mm/mmu_decl.h>
#include <asm/prom.h>
#include <asm/udbg.h>
#include <asm/mpic.h>

#include <linux/of_platform.h>
#include <sysdev/fsl_soc.h>
#include <sysdev/fsl_pci.h>
#include <linux/root_dev.h>

#include <brocade/platform.h>
#include <brocade/silkworm.h>
#include <brocade/silkworm_features.h>
#include <brocade/silkworm_cpld.h>
#include <brocade/cpld.h>
#include <brocade/silkworm_bhpc.h>
#include "../85xx/corenet_ds.h"

#ifdef CONFIG_SILKWORM
#ifdef CONFIG_SILKWORM_8BIT_CPLD
extern int cpld_init(void);
#else
extern int cpld_init32(void);
#endif
#endif
#ifdef CONFIG_BOOKE_WDT
extern void ppc_booke_sw_wdt_heartbeat(void);
#endif

#ifndef PCI_VENDOR_ID_TUNDRA
#define PCI_VENDOR_ID_TUNDRA 0x10e3
#endif

#define NUM_OF_SLOTS_PER_NTB	(4)
#define SLOT_ADDRESS_SPACE		(128*1024*1024) /* 128 MB*/
/* NTB claims a part of the address space to itself*/
#define NTB_ADDRESS_SPACE		(128*1024*1024) /* 128 MB*/
#define NUM_OF_NTBS				(3)
#define TOTAL_NTB_ADDRESS_SPACE \
		((NUM_OF_SLOTS_PER_NTB*SLOT_ADDRESS_SPACE)+	\
		(NTB_ADDRESS_SPACE))
#define PCIE_SWITCH_ADDRESS_SPACE_BEGIN	0x70000000
#define PCIE_SWITCH_ADDRESS_SPACE	(128*1024*1024)
#define FIRST_PCIE_SWITCH_SLOT_NO	4

static int irq_table[MM_MAX_IRQS];

static void of_cpld_init(void)
{
	struct device_node *np;
	int index, ret;

	/* NOS uses External IRQs 0-11 */
	for_each_node_by_type(np, "cpld") {
		for (index = 0; index < MM_MAX_IRQS; index++) {
			ret = irq_of_parse_and_map(np, index);
			pr_debug("%s: Init cpld IRQ %d ret %d\n", __FUNCTION__, index, ret);
			irq_table[index] = ret;

		}
	}
}
/*
 * Provide a way for NOS to get the virtual interrupt numbers
 * for the external IRQs.
 */
int silkworm_irq_table(int *table, int bytes)
{
	int size;

	size = (bytes > sizeof(irq_table))?sizeof(irq_table):bytes;
	memcpy(table, irq_table, size);

	return (0);
}
EXPORT_SYMBOL(silkworm_irq_table);

void __init mpc85xx_embedded_pic_init(void)
{
	struct mpic *mpic;
	struct resource r;
	struct device_node *np = NULL;

	np = of_find_node_by_type(np, "open-pic");

	if (np == NULL) {
		printk(KERN_ERR "Could not find open-pic node\n");
		return;
	}

	if (of_address_to_resource(np, 0, &r)) {
		printk(KERN_ERR "Failed to map mpic register space\n");
		of_node_put(np);
		return;
	}

	mpic = mpic_alloc(np, r.start, MPIC_PRIMARY | MPIC_ENABLE_COREINT |
			  MPIC_BIG_ENDIAN | MPIC_BROKEN_FRR_NIRQS |
			  MPIC_SINGLE_DEST_CPU,
			0, 256, " OpenPIC  ");
	BUG_ON(mpic == NULL);

	mpic_init(mpic);

	of_cpld_init();
}

#ifdef CONFIG_PCI
static int primary_phb_addr;
#endif

/*
 * Setup the architecture
 */
#ifdef CONFIG_SMP
void __init mpc85xx_smp_init(void);
#endif
static void __init mpc85xx_embedded_setup_arch(void)
{
#ifdef CONFIG_PCI
	struct device_node *np;
#endif

	if (ppc_md.progress)
		ppc_md.progress("mpc85xx_embedded_setup_arch()", 0);

#ifdef CONFIG_SMP
	mpc85xx_smp_init();
#endif

#ifdef CONFIG_PCI
	for_each_node_by_type(np, "pci") {
		if (of_device_is_compatible(np, "fsl,p4080-pcie") ||
			of_device_is_compatible(np, "fsl,t1040-pcie")) {
			struct resource rsrc;
			of_address_to_resource(np, 0, &rsrc);
			if ((rsrc.start & 0xfffff) == primary_phb_addr)
				fsl_add_bridge(np, 1);
			else
				fsl_add_bridge(np, 0);
		}
	}
#endif

#ifdef CONFIG_ROOT_NFS
	ROOT_DEV = Root_NFS;
#else
	ROOT_DEV = Root_SDA1;
#endif

#ifdef CONFIG_SILKWORM
#ifdef CONFIG_SILKWORM_8BIT_CPLD
	cpld_init();
#else
	cpld_init32();
#endif
#endif

	printk("Brocade Silkworm T10xx port (C) 2014 Brocade Communications Systems, Inc.\n");
}

static void show_cpuinfo(struct seq_file *m)
{
	uint svr_reg = mfspr(SPRN_SVR);
	if (svr_reg & 0x00080000) {
		seq_printf(m, "chipset\t\t: T10xx (svr = %x, sec engine enabled)\n", svr_reg);
	} else {
		seq_printf(m, "chipset\t\t: T10xx (svr = %x)\n", svr_reg);
	}
	seq_printf(m, "Vendor\t\t: Freescale Semiconductor\n");
}


static const struct of_device_id of_device_ids[] __devinitconst = {
	{
		.compatible = "brocade,ds1742"
	},
	{
		.compatible = "brocade,silkworm-gen-nvram"
	},
	{}
};

static int __init declare_of_embedded_platform_devices(void)
{
	int err;

	err = of_platform_bus_probe(NULL, of_device_ids, NULL);
	if (err)
		return err;

	return 0;
}
machine_device_initcall(silkworm_embedded, declare_of_embedded_platform_devices);

/*
 * Called very early, device-tree isn't unflattened
 */
static int __init t104x_embedded_probe(void)
{
        unsigned long root = of_get_flat_dt_root();

        if (of_flat_dt_is_compatible(root, "brocade,SILKWORM_EMBEDDED")) {
#ifdef CONFIG_PCI
		/* xxx - galak */
		primary_phb_addr = 0x8000;
#endif
#ifdef CONFIG_SILKWORM
		ppc_md.feature_call = ppc44x_do_feature_call;
#endif
		return 1;
	} else {
		return 0;
	}
}

define_machine(silkworm_embedded) {
        .name                   = "SILKWORM_EMBEDDED",
        .probe                  = t104x_embedded_probe,
        .setup_arch             = mpc85xx_embedded_setup_arch,
        .init_IRQ               = mpc85xx_embedded_pic_init,
#ifdef CONFIG_PCI
        .pcibios_fixup_bus      = fsl_pcibios_fixup_bus,
#endif
/* coreint doesn't play nice with lazy EE, use legacy mpic for now */
#ifdef CONFIG_PPC64
        .get_irq                = mpic_get_irq,
#else
        .get_irq                = mpic_get_coreint_irq,
#endif
        .restart                = fsl_rstcr_restart,
        .calibrate_decr         = generic_calibrate_decr,
        .progress               = udbg_progress,
        .power_off				= cpld_power_off,
#ifdef CONFIG_PPC64
        .power_save             = book3e_idle,
#else
        .power_save             = e500_idle,
#endif
#ifdef CONFIG_BOOKE_WDT
        .heartbeat              = ppc_booke_sw_wdt_heartbeat,
#endif
        .init_early             = corenet_ds_init_early,
        .show_cpuinfo           = show_cpuinfo,
};

machine_arch_initcall(silkworm_embedded, corenet_ds_publish_devices);

#ifdef CONFIG_SWIOTLB
machine_arch_initcall(silkworm_embedded, swiotlb_setup_bus_notifier);
#endif

#ifdef CONFIG_SILKWORM_ENABLE_GEN1_HACK
// force CPU PCIe root complex to gen1, all devices direct connect PCIe to
// CPU will have gen1. Need more change if want to set PCIe switch
/*
 * Clear any pending PCIe errors in the PCIe controller so
 * they don't cause unwanted interrupts later.
 */
static void clear_root_pcie_errors(struct pci_dev *dev)
{
	struct device_node *np;
	struct resource res;
	volatile void __iomem *pex_base_addr = NULL;
	volatile void __iomem *regaddr;
	u32 errdetect1 = 0;

	if (dev) {

		np = pci_device_to_OF_node(dev);
		if (np) {
			pr_debug("%s: Dev %s Node %s\n", __FUNCTION__,
				   pci_name(dev), np->full_name);
			np = of_get_parent(np);
			if (np) {
				pr_debug("%s: Dev %s Parent Node %s\n", __FUNCTION__,
				   pci_name(dev), np->full_name);
			} else {
				pr_err( "%s: Dev %s - Cannot find parent node\n",
					   __FUNCTION__, pci_name(dev));
			}
		}

		of_address_to_resource(np, 0, &res);

		pex_base_addr = ioremap(res.start,
								(unsigned long)(res.end - res.start));

		if (pex_base_addr) {
			regaddr = pex_base_addr + 0xe00;	/* ERR_DR */
			errdetect1 = readl_be(regaddr);
			pr_debug("%s: ERR_DR 0x%8.8x\n", __FUNCTION__, errdetect1);
			writel_be(errdetect1, regaddr);

			regaddr = pex_base_addr + 0x020;  // PME_DR
			errdetect1 = readl_be(regaddr);
			pr_debug("%s: PME_DR 0x%8.8x\n", __FUNCTION__, errdetect1);
			writel_be(errdetect1, regaddr);
			iounmap(pex_base_addr);
		} else {
			pr_err("%s: Could not map PCIe controller\n",
				   __FUNCTION__);
		}
	} else {
		pr_err("%s: Could not find PCIe controller\n", __FUNCTION__);
	}

}
/************************************************************
 * PCI fixup routines to take care of special cases
 *
 *
 ************************************************************
 */
#define LINK_RETRAIN_TIMEOUT (HZ / 10)
static int
set_link_speed(struct pci_dev *dev, int speed)
{
	u32 regval32 = 0;
	u16 regval16 = 0;
	unsigned long start_jiffies;
	int rc = 0, i = 0, pos;

	pos = pci_find_capability(dev, PCI_CAP_ID_EXP);

	if (pos) {

		/* Set the desired target speed in bits 3:0 */
		pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL2, &regval32);
		regval32 &= ~(0xF);
		regval32 |= speed;
		pci_write_config_dword(dev, pos + PCI_EXP_LNKCTL2, regval32);

		/*
		 * If link training is already in progress, then
		 * need to wait for it to complete before triggering
		 * it again.
		 */
		pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
		if (regval16 & PCI_EXP_LNKSTA_LT) {
			start_jiffies = jiffies;
			for (;;) {
				pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
				if (!(regval16 & PCI_EXP_LNKSTA_LT)) {
					break;
				}
				if (time_after(jiffies,
							   start_jiffies + LINK_RETRAIN_TIMEOUT)) {
					break;
				}
				msleep(1);
			}
		}

		/* Do a link re-train to change the speed */
		pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
		regval16 |= PCI_EXP_LNKCTL_RL;
		pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

		/* Now wait for link training to complete */
		pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
		if (regval16 & PCI_EXP_LNKSTA_LT) {
			start_jiffies = jiffies;
			for (;;) {
				pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
				if (!(regval16 & PCI_EXP_LNKSTA_LT)) {
					break;
				}
				i++;
				if (time_after(jiffies,
							   start_jiffies + LINK_RETRAIN_TIMEOUT)) {
					break;
				}
				msleep(1);
			}
		}

		/* See if link re-training completed */
		pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
		pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL2, &regval32);
		if (regval16 & PCI_EXP_LNKSTA_LT) {
			u32 linkctl = 0, slotctl = 0;

			pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL, &linkctl);
			pci_read_config_dword(dev, pos + PCI_EXP_SLTCTL, &slotctl);
			printk(KERN_ERR "%s %d: Dev %s Link stuck in training "
				   "linkctl 0x%8.8x linkctl2 0x%8.8x slotctl 0x%8.8x\n",
				   __FUNCTION__, i, pci_name(dev), linkctl, regval32, slotctl);
			rc = -1;
		}
	}
	return (rc);
}

static void __devinit hotplug_fixup_root(struct pci_dev *dev)
{
	/*
	 * Hot-reset sometimes causes error bits to be set in PCIe
	 * controller. These will cause problems with other things
	 * (such as BMAN) that share the same interrupt as the PCIe
	 * errors, so they must be cleared here.
	 */
	clear_root_pcie_errors(dev);
	set_link_speed(dev, 1);
}

/* when freescale cpu, call hotplug_fixup_root	*/
DECLARE_PCI_FIXUP_EARLY(PCI_VENDOR_ID_FREESCALE, PCI_ANY_ID, hotplug_fixup_root);
#endif
