

/*
 * Copyright (C) 2012 Brocade Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <linux/types.h>
#include <linux/version.h>
#include <linux/interrupt.h> 
#include <linux/pci.h>
#include <linux/delay.h>

#include <linux/of_platform.h>
#define SA_SHIRQ IRQF_SHARED
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/i8259.h>
#include <asm/irq.h>
#ifdef  CONFIG_WATCHDOG
extern int kick_watchdog(void);
#endif
#include <brocade/silkworm_pcie.h>

static int
get_link_status(struct pci_dev *dev, int pos)
{
	u16 link_status = 0, slot_status = 0;
	int link_down = -1;

	if (dev) {

		/* Clear all the w1c status */
		pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &link_status);
		pci_write_config_word(dev, pos + PCI_EXP_LNKSTA, link_status);
		pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &link_status);

		if (link_status & PCI_EXP_LNKSTA_DLLLA) {
#if 0
			printk(KERN_ERR "%s: UP dev %s. link status 0x%4.4x\n",
				   __FUNCTION__, pci_name(dev), link_status);
#endif
			link_down = 0;
		} else {

			if (slot_status & PCI_EXP_SLTSTA_PDS) {

				/* If we expect the link to be up but is isn't, issue a warning */
				pci_read_config_word(dev, pos + PCI_EXP_SLTSTA, &slot_status);
				printk(KERN_ERR "%s: DOWN dev %s. link status  0x%4.4x "
					   "slot status 0x%4.4x\n", __FUNCTION__, pci_name(dev),
					   link_status, slot_status);
			}
			link_down = 1;
		}
	}

	return (link_down);
}

static void
set_link_speed(struct pci_dev *dev, int pos, int speed,
						   int original_link_speed)
{
	u32 regval32 = 0;
	u16 regval16 = 0;

	if (dev) {

		pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL2, &regval32);
		regval32 &= ~(0xF);

		switch (speed) {

		case 1:

			/* regval32 Already set to Gen 1 speed */
			break;

		case 2:

			/* Set link speed to gen 2 via link control2 0x98[3:0] */

			regval32 |= 0x02;

			break;

		default:

			/* Leave speed as it was */
			regval32 |= original_link_speed;
		}

		pci_write_config_dword(dev, pos + PCI_EXP_LNKCTL2, regval32);

		/* Set link retrain via link control 0x78[5] */

		pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
		regval16 |= PCI_EXP_LNKCTL_RL;
		pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);
		msleep(5);

	}
}

/*
 * Description:
 *  This function is used to determine the type of port represent by
 *  the specified pci device object
 *
 *
 * Inputs:
 *   dev - The pci_dev object whose port number is to be determined.
 *
 *
 * Outputs:
 *   N/A
 *
 * Returns:
 *  -1 - Port type could not be read
 *  PCI_EXP_TYPE_DOWNSTREAM - Downstream port
 *  PCI_EXP_TYPE_UPSTREAM   - Upstream port
 *  PCI_EXP_TYPE_ROOT_PORT  - Root complex
 *  PCI_EXP_TYPE_ENDPOINT   - Endpoint device
 *
 */

int
pex_get_port_type(struct pci_dev *dev)
{
	int pos, port_type = -1;
	u16 regval16;

	if (dev) {
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP); 

		/* This is the PCIE capabilities register */
		if (pos) {
			pci_read_config_word(dev, pos + PCI_EXP_FLAGS, &regval16);
			port_type = ((regval16 & PCI_EXP_FLAGS_TYPE) >> 4);
		}
	}
	return (port_type);

}

/*
 *
 * Performs reset of a PLX PCIe switch port using device-specific
 * registers as per:
 *   errata 29 for pex8632
 *
 * Inputs:
 *   dev - pci device object of the port to be reset
 *   speed - Optionally set the link speed:
 *           2 = Gen 2 speed (5.0G)
 *           1 = Gen 1 speed (2.5G)
 *           Any other value = leave speed as-is
 * 
 * This function may be compatible with other switch types. Check
 * device data sheets to see.
 * 
 *
 * Returns:
 * -1 = Error, link status is unknown
 *  1 = link is down
 *  0 = link is up
 */
static u32 pex8632_port_reset_map[] = {
	0x00010000,	/* Ports 0, 4, 8  */
	0x00020000,	/* Ports 1, 5, 9  */
	0x00040000,	/* Ports 2, 6, 19 */
	0x00080000,	/* Ports 3, 7, 11 */
};

int
pex8632_errata_port_reset(struct pci_dev *dev, int speed)
{
	struct pci_dev *dis_dev = NULL;
	int bitmap, pos, port, link_down = -1, disable_offset = 0x234;
	int domain = 0, idx, port_type, devfn;
	u16 regval16;
	u32 regval32, original_link_speed = 0;

	if (!dev) {
		printk(KERN_ERR "%s: NULL device\n", __FUNCTION__);
		goto Exit;
	}

	pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
	if (!pos) {
		printk(KERN_ERR "%s: Cannot find PCI_CAP_ID_EXP pos = %d dev %s\n",
			   __FUNCTION__, pos, pci_name(dev));
		goto Exit;
	}

	port_type = pex_get_port_type(dev);
	if (port_type != PCI_EXP_TYPE_DOWNSTREAM) {
		printk(KERN_ERR "%s: Invalid port type %d. S/B %d dev %s\n",
			   __FUNCTION__, port_type, PCI_EXP_TYPE_DOWNSTREAM,
			   pci_name(dev));
		goto Exit;
	}

	port = PCI_SLOT(dev->devfn);

	/* Find the device that has the disable register for this port */
	if ((port >= 0)&&(port <= 3)) {
		devfn = PCI_DEVFN(0, 0);
		dis_dev = pci_get_domain_bus_and_slot(domain, dev->bus->primary, devfn);
	} else if ((port >= 4)&&(port <= 7)) {
		devfn = PCI_DEVFN(4, 0);
		dis_dev = pci_get_domain_bus_and_slot(domain, dev->bus->number, devfn);
	} else if ((port >= 8)&&(port <= 11)) {
		devfn = PCI_DEVFN(8, 0);
		dis_dev = pci_get_domain_bus_and_slot(domain, dev->bus->number, devfn);
	} else {
		printk(KERN_ERR "%s: Invalid port number %d for dev %s\n",
			   __FUNCTION__, port, pci_name(dev));
		goto Exit;
	}

	if (!dis_dev) {
		printk(KERN_ERR "%s: Could not find dis_dev for port %d domain %d bus %d "
			   "devfn %d for dev %s\n", __FUNCTION__, port, domain,
			   dev->bus->number, devfn, pci_name(dev));
		goto Exit;
	}

	idx = port % 4;
	bitmap = pex8632_port_reset_map[idx];


/* STEP 1: BRING THE LINK DOWN */

	/*	Set link disable via link control 0x78[4] */

	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
	regval16 |= PCI_EXP_LNKCTL_LD;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

	/*	Set link retrain via link control 0x78[5] */

	regval16 |= PCI_EXP_LNKCTL_RL;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

/* STEP 2: DISABLE THE PORT */

	/* Set link speed to gen 1 via link control2 0x98[3:0] */

	pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL2, &regval32);
	original_link_speed = regval32 & 0xF; /* save for later use */
	regval32 &= ~(0xF);
	pci_write_config_dword(dev, pos + PCI_EXP_LNKCTL2, regval32);

	udelay(10);

	/* Set port disable via 0x234[19:16] */

	pci_read_config_dword(dis_dev, disable_offset, &regval32);
	regval32 |= bitmap;
	pci_write_config_dword(dis_dev, disable_offset, regval32);


/* STEP 3: ENABLE THE PORT */

	/* 
	 * Enable DLLA. This is not part of the errata, but the link
	 * does not come up if we don't do this.
	 */

	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
	regval16 &= ~PCI_EXP_LNKCTL_LD;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

	/* Enable port via 0x234[19:16] */

	pci_read_config_dword(dis_dev, disable_offset, &regval32);
	regval32 &= ~bitmap;
	pci_write_config_dword(dis_dev, disable_offset, regval32);
	msleep(5);

	/* If the link isn't up, try disable/enable again */

	pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
	if (!(regval16 & PCI_EXP_LNKSTA_DLLLA)) {

#if 0
		printk(KERN_INFO "%s: %s Link not up 0x%4.4x. Retrying...\n",
			   __FUNCTION__, pci_name(dev), regval16);
#endif
		pci_read_config_dword(dis_dev, disable_offset, &regval32);
		regval32 |= bitmap;
		pci_write_config_dword(dis_dev, disable_offset, regval32);
		mdelay(1);

		pci_read_config_dword(dis_dev, disable_offset, &regval32);
		regval32 &= ~bitmap;
		pci_write_config_dword(dis_dev, disable_offset, regval32);
		msleep(5);
	}

/* STEP 4: OPTIONALLY SHIFT TO GEN 2 SPEED */

	set_link_speed(dev, pos, speed, original_link_speed);


	link_down = get_link_status(dev, pos);


	pci_dev_put(dis_dev);

Exit:

	return (link_down);
}
EXPORT_SYMBOL(pex8632_errata_port_reset);

/*
 *
 * Performs reset of a PLX PCIe switch port using device-specific
 * registers as per:
 *   errata 17 for pex8614
 *   errata 18 for pex8608 and 8606
 *   errata 
 *
 * if "speed" parameter is set to 2 (Gen 2 speed), then also performs:
 *  errata 12 for pex8614
 *  errata 13 for pex8606
 *
 * Inputs:
 *   dev - pci device object of the port to be reset
 *   speed - Optionally set the link speed:
 *           2 = Gen 2 speed (5.0G)
 *           1 = Gen 1 speed (2.5G)
 *           Any other value = leave speed as-is
 * 
 * This function may be compatible with other switch types. Check
 * device data sheets to see.
 * 
 *
 * Returns:
 * -1 = Error, link status is unknown
 *  1 = link is down
 *  0 = link is up
 */

static u32 pex8614_port_reset_map[] = {
	0x01,	/* Port 0  */
	0x01,	/* Port 1  */
	0x02,	/* Port 2  */
	0x00,	/* Port 3 Not implemented */
	0x04,	/* Port 4  */
	0x04,	/* Port 5  */
	0x08,	/* Port 6  */
	0x08,	/* Port 7  */
	0x10,	/* Port 8  */
	0x10,	/* Port 9  */
	0x20,	/* Port 10 */
	0x00,	/* Port 11 Not implemented */
	0x40,	/* Port 12 */
	0x00,	/* Port 13 Not implemented */
	0x80	/* Port 14 */
};
static int table_size = (sizeof(pex8614_port_reset_map)/sizeof(uint));

int
pex8614_errata_port_reset(struct pci_dev *dev, int speed)
{

	int pos, port, disable_offset, quiet_offset, link_down = -1;
	u32 regval32 = 0, port_bitmap, original_link_speed;
	u16 regval16 = 0;
	struct pci_dev *pdev = NULL;

	if (!dev) {
		printk(KERN_ERR "%s: NULL device\n", __FUNCTION__);
		goto Exit;
	}

	pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
	if (!pos) {
		printk(KERN_ERR "%s: Cannot find PCI_CAP_ID_EXP pos = %d dev %s\n",
			   __FUNCTION__, pos, pci_name(dev));
		goto Exit;
	}

	port = PCI_SLOT(dev->devfn);
	pdev = dev->bus->self;

	if ((port >= 0)&&(port < table_size)) {
		if (port & 0x01) {
			disable_offset = 0x234; /* For odd port */
			quiet_offset = 0x22C;
		} else {
			disable_offset = 0x230; /* For even port */
			quiet_offset = 0x228;
		}

		port_bitmap = pex8614_port_reset_map[port];

	} else {
		printk(KERN_ERR "%s: Invalid port number %d for dev %s\n",
			   __FUNCTION__, port, pci_name(dev));
		goto Exit;
	}


/* STEP 1: BRING THE LINK DOWN */

	/*	Set link disable via link control 0x78[4] */

	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
	regval16 |= PCI_EXP_LNKCTL_LD;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

	/*	Set link retrain via link control 0x78[5] */

	regval16 |= PCI_EXP_LNKCTL_RL;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);


/* STEP 2: DISABLE THE PORT */

	/* 	Set detect.quiet wait via 0x228[11] or 0x22C[11] */

	pci_read_config_dword(dev, quiet_offset, &regval32);
	regval32 |= 0x00000800;
	pci_write_config_dword(dev, quiet_offset, regval32);

	/* Set link speed to gen 1 via link control2 0x98[3:0] */

	pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL2, &regval32);
	original_link_speed = regval32 & 0xF; /* save for later use */
	regval32 &= ~(0xF);
	pci_write_config_dword(dev, pos + PCI_EXP_LNKCTL2, regval32);

	/* Set link retrain via 0x78[5] */

	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
	regval16 |= PCI_EXP_LNKCTL_RL;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

	/* Wait 10us */

	udelay(10);

	/* Set port disable via 0x230[7:0] or 0x234[7:0] */

	pci_read_config_dword(pdev, disable_offset, &regval32);
	regval32 |= port_bitmap;
	pci_write_config_dword(pdev, disable_offset, regval32);
	mdelay(1);


/* STEP 3: ENABLE THE PORT */


	/* Set detect.quiet wait via 0x228[11] or 0x22C[11] */

	pci_read_config_dword(dev, quiet_offset, &regval32);
	regval32 |= 0x00000800;
	pci_write_config_dword(dev, quiet_offset, regval32);

	/* 
	 * Enable DLLA. This is not part of the errata, but the link
	 * does not come up if we don't do this.
	 */

	pci_read_config_word(dev, pos + PCI_EXP_LNKCTL, &regval16);
	regval16 &= ~PCI_EXP_LNKCTL_LD;
	pci_write_config_word(dev, pos + PCI_EXP_LNKCTL, regval16);

	/* Enable port via 0x230[7:0] or 0x234[7:0] */

	pci_read_config_dword(pdev, disable_offset, &regval32);
	regval32 &= ~port_bitmap;
	pci_write_config_dword(pdev, disable_offset, regval32);
	msleep(5);

	/* If link not up, then disable/enable port via 0x230[7:0] or 0x234[7:0] */

	pci_read_config_word(dev, pos + PCI_EXP_LNKSTA, &regval16);
	if (!(regval16 & PCI_EXP_LNKSTA_DLLLA)) {

#if 0
		printk(KERN_ERR "%s: %s Link not up 0x%4.4x. Retrying...\n",
			   __FUNCTION__, pci_name(dev), regval16);
#endif
		pci_read_config_dword(pdev, disable_offset, &regval32);
		regval32 |= port_bitmap;
		pci_write_config_dword(pdev, disable_offset, regval32);
		mdelay(1);

		pci_read_config_dword(pdev, disable_offset, &regval32);
		regval32 &= ~port_bitmap;
		pci_write_config_dword(pdev, disable_offset, regval32);
		msleep(5);
	}

/* STEP 4: OPTIONALLY SHIFT TO GEN 2 SPEED FOR ERRATA 12 */

	set_link_speed(dev, pos, speed, original_link_speed);


	link_down = get_link_status(dev, pos);


 Exit:

	return (link_down);

}
EXPORT_SYMBOL(pex8614_errata_port_reset);


/* This could also be used for the pex8608 */
static void __devinit pex8606_errata(struct pci_dev *dev)
{
	int port_type;

	/* Verify it's a DSP */

	if (dev) {
		port_type = pex_get_port_type(dev);

		if (port_type == PCI_EXP_TYPE_DOWNSTREAM) {

			pex8614_errata_port_reset(dev, 0);
		}
	}
}

#if 0

/* Not yet ready to enable this */

static void __devinit pex8632_errata(struct pci_dev *dev)
{
	int port_type;

	if (dev) {
		port_type = pex_get_port_type(dev);

		/* Filter for DSP only */

		if (port_type == PCI_EXP_TYPE_DOWNSTREAM) {

			pex8632_errata_port_reset(dev, 0);
		}
	}
}
DECLARE_PCI_FIXUP_EARLY(PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_8632, pex8632_errata);
#endif

DECLARE_PCI_FIXUP_EARLY(PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_8606, pex8606_errata);

/*
 * Parse the memory and bus ranges for the specified device from
 * the device tree.
 */
int
find_pci_of_resources(struct pci_dev *dev, u32 *memory, u32 *busses)
{
	struct device_node *node = NULL;
	const int *bus_range, *ranges;
	int rc = 0, len;

	if (dev) {
		node = pci_device_to_OF_node(dev);
		if (node) {

			bus_range = of_get_property(node, "bus-range", &len);
			if (bus_range) {
				*busses = ((bus_range[1] << 16) & 0xFF0000) |
					((bus_range[0] << 8) & 0xFF00) |
					((dev->bus->number) & 0xFF);

				pr_debug("%s %d: Dev %s node [%s] "
					   "Busses %d,%d-%d  0x%8.8x\n", __FUNCTION__, __LINE__,
					   pci_name(dev), node->full_name, dev->bus->number,
					   bus_range[0], bus_range[1], *busses);

			} else {
				printk(KERN_ERR "%s %d: Dev %s node [%s] No Busses\n",
					   __FUNCTION__, __LINE__, pci_name(dev), node->full_name);
			}

			ranges = of_get_property(node, "ranges", &len);
			if (ranges == NULL) {
				printk(KERN_ERR "%s %d: Can't get ranges for PCI-PCI "
					   "bridge %s\n",
					   __FUNCTION__, __LINE__, node->full_name);
			} else {
				unsigned int flags;
				u64 start = 0, size, end = 0;

				flags = ranges[0];
				size = of_read_number(&ranges[6], 2);
				if (size != 0) {
					if (flags & 0x02000000) {
						start = of_read_number(&ranges[1], 2);
						end = start + size - 1;
						*memory = ((start >> 16) & 0xFFFF) | (end & 0xFFFF0000);
						pr_debug("%s %d: MEM start 0x%llx end 0x%llx "
							   "flags 0x%x Ranges len %d Limit/base 0x%x\n",
							   __FUNCTION__, __LINE__, start, end, flags,
							   len, *memory);
					} else {
						printk(KERN_ERR "%s %d: Flags are 0\n",
							   __FUNCTION__, __LINE__);
					}
				} else {
					printk(KERN_ERR "%s %d: Ranges Size is 0\n",
						   __FUNCTION__, __LINE__);
				}
			}

		} else {
			pr_debug("%s %d: Could not find node for dev %s\n",
				   __FUNCTION__, __LINE__, pci_name(dev));
		}
	}

	return (rc);
}
/*
 * Get the memory and bus range resources from device tree and
 * program the device accordingly. If they are not in the device
 * tree, then just leave the device as-is.
 */
void
hotplug_fixup_resources(struct pci_dev *dev)
{
	u32 prev_busses = 0, prev_memory = 0;
	u32 memory = 0, busses = 0;

	find_pci_of_resources(dev, &memory, &busses);

	if (busses) {
        pci_read_config_dword(dev, PCI_PRIMARY_BUS, &prev_busses);
        pci_write_config_dword(dev, PCI_PRIMARY_BUS, busses);
        pr_debug("%s %d: Fix up busses from 0x%8.8x to 0x%8.8x for dev %s\n",
               __FUNCTION__, __LINE__, prev_busses,  busses, pci_name(dev));
    }

    if (memory) {
        pci_read_config_dword(dev, PCI_MEMORY_BASE, &prev_memory);
        pci_write_config_dword(dev, PCI_MEMORY_BASE, memory);
        pr_debug("%s %d: Fix up memory from 0x%8.8x to 0x%8.8x for dev %s\n",
               __FUNCTION__, __LINE__, prev_memory, memory, pci_name(dev));
    }

}


