/*
 *    Copyright (c) 1996-2007 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Initialization and Helper routines for SIN(System Inventory)
 */

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <brocade/sin.h>
#include <brocade/silkworm_sin.h>


sw_bd	silkworm = {0, 0};
EXPORT_SYMBOL(silkworm);

/* Preprocessor Defines */
#define SIN_ENTRIES_MAX         32   
                                     
/* Type Definitions */
struct sin_table {
	unsigned int	 st_flags;
	sin_rec_t	 st_rec;
};

/*
 * SIN Table flags
 */
#define	ST_ASSIGNED		0x00000001


/* Function Macros */
#define	SPTOINDEX(sp)		(*(sp))


/* Global Variables */
static struct sin_table sin_table[SIN_ENTRIES_MAX];
static sin_stats_t sin_table_stats;


int __init
sin_init(void)
{
	int i;

	/* Initialize the statistics */

	sin_table_stats.ss_version = SIN_VERSION;
	sin_table_stats.ss_size = sizeof (sin_rec_t);
	sin_table_stats.ss_used = 0;
	sin_table_stats.ss_total = SIN_ENTRIES_MAX;

	/* Walk the table and initialize some fields in each record. */

	for (i = 0; i < sin_table_stats.ss_total; i++) {
		sin_table[i].st_flags &= ~ST_ASSIGNED;
	}

	return (0);
}

int
sin_rewind(int *statep)
{
	if (statep == NULL)
		return (-1);

	*statep = 0;

	return (0);
}

sin_rec_t *
sin_get(sin_rec_t *srp, int *statep)
{
	int i, n;

	/* Check for bad input parameters */

	if (srp == NULL || statep == NULL)
		return (NULL);

	/*
	 * If there are no entries used, don't waste time
	 * searching.
	 */

	if (sin_table_stats.ss_used == 0)
		return (NULL);

	/* Parameters were good and we've got records; start searching... */

	n = SPTOINDEX(statep);

	for (i = n; i < sin_table_stats.ss_total; i++) {
		if (sin_table[i].st_flags & ST_ASSIGNED)
			break;
	}

	/*
	 * Sanity check, the statistics might have lied and there really
	 * were no records.
	 */

	if (i >= sin_table_stats.ss_total)
		return (NULL);

	/* At this point, we found a record, copy it to the caller. */

	memcpy(srp, &sin_table[i].st_rec, sizeof (sin_rec_t));

	/* Update the index state */

	SPTOINDEX(statep) = i + 1;

	return (srp);
}

//static int
//sin_cmp(const sin_rec_t *srp1, const sin_rec_t *srp2)
//{
//	return (memcmp(srp1, srp2, sizeof (sin_rec_t)));
//}

#ifdef	NOTYET
static sin_rec_t *
sin_search(sin_rec_t *srp, int *statep)
{
	int i, n, match = 0;
	struct sin_table *cstp = NULL;

	/* Check for bad input parameters */

	if (srp == NULL)
		return (NULL);

	/*
	 * If there are no entries used, don't waste time
	 * searching.
	 */

	if (sin_table_stats.ss_used == 0)
		return (NULL);

	/*
	 * The state pointer is optional. Figure out if it was passed and
	 * if it was, if it is valid. Otherwise, start at the very begin-
	 * ning, a very good place to start.
	 */

	n = ((statep != NULL) ? SPTOINDEX(statep) : 0);

	/*
	 * Sanity check, the caller has already reached the end and needs
	 * to do rewind before searching again.
	 */

	if (n >= sin_table_stats.ss_total) {
		return (NULL);
	}

	/* Walk the database, searching as directed. */

	for (i = n; i < sin_table_stats.ss_total; i++) {
		match = 0;
		cstp = &sin_table[i];

		if (!(cstp->st_flags & ST_ASSIGNED))
			continue;

		match = (sin_cmp(srp, &cstp->st_rec) == 0);

		if (match) break;
	}

	/*
	 * If a state pointer was passed, update it to reflect the location
	 * of where the search left off.
	 */

	if (statep != NULL) {
		SPTOINDEX(statep) = i + 1;
	}

	if (match && cstp != NULL) {
		memcpy(srp, &cstp->st_rec, sizeof (sin_rec_t));
	}

	return (srp);
}
#endif

int
sin_add(sin_rec_t *srp)
{
	int i;
	struct sin_table *cstp = NULL;

	/* Check for up-front errors */

	if (srp == NULL)
		return (-1);

	/*
	 * If the table is full, don't bother searching for an empty
	 * slot
	 */

	if (sin_table_stats.ss_used == sin_table_stats.ss_total)
		return (-1);

	/* Find an empty database entry and add the new license. */

	for (i = 0; i < SIN_ENTRIES_MAX; i++) {
		cstp = &sin_table[i];

		if (!(cstp->st_flags & ST_ASSIGNED)) {
			cstp->st_flags |= ST_ASSIGNED;
			break;
		}
	}

	/* Update the statistics */

	sin_table_stats.ss_used++;

	/*
	 * The record ID is the table index plus one. This is a completely
	 * arbitrary assignment as might as well be some other unique
	 * number.
	 */

	srp->sr_id = i + 1;

	/* Copy the record into the table from the caller */

	memcpy(&cstp->st_rec, srp, sizeof (sin_rec_t));

	return (0);
}

#ifdef	NOTYET
static int
sin_remove(sin_rec_t *srp)
{
	int i;
	struct sin_table *cstp = NULL;

	/* Check for up-front errors */

	if (srp == NULL) {
		return (-1);
	}

	/*
	 * If there are no entries in the table, then we can't very
	 * well try to remove anything.
	 */

	if (sin_table_stats.ss_used == 0)
		return (-1);

	/* Find the key in the database */

	for (i = 0; i < sin_table_stats.ss_total; i++) {
		cstp = &sin_table[i];

		if (sin_cmp(srp, &cstp->st_rec) == 0)
			break;
	}

	/* Remove it by clearing the "assigned" flag */

	cstp->st_flags &= ~ST_ASSIGNED;

	/* Update the statistics */

	sin_table_stats.ss_used--;

	return (0);
}
#endif

/*
 * int sin_stats()
 *
 * Description:
 *   This routine returns system inventory statistics.
 *
 * Input(s):
 *   *ssp - A pointer to a system inventory statistics structure to copy
 *          data to.
 *
 * Output(s):
 *   *ssp - A pointer to a system inventory statistics structure updated with
 *          current statistics.
 *
 * Returns:
 *   0 if OK, otherwise -1 on error.
 *
 */
int
sin_stats(sin_stats_t *ssp)
{
	if (ssp == NULL)
		return (-1);

	memcpy(ssp, &sin_table_stats, sizeof (sin_stats_t));

	return (0);
}


/*
 * int read_silkworm_bdinfo()
 *
 * Description:
 *   Reads the sin record for silkworm board and fills in the struct.
 *
 * Input(s):
 *
 *
 * Output(s):
 *
 *
 * Returns:
 *   0 if OK, otherwise -1 on error
 *
 */
static int __init
read_silkworm_bdinfo(sw_bd* silkworm)
{
    /* The silkworm board sin record */
    static sin_rec_t sr_silkwormbd = {0};
    int state = 0;

    
    /* If the sin struct for silkworm_bd has not been updated, do so */
    if(!sr_silkwormbd.sr_class) {

	/* Reset the system inventory search state */
	sin_rewind(&state);

	/* Find the silkworm sin_record */
	if( sin_find_sr(SIN_CLASS_BOARD, SIN_CLASS_BOARD_PROCESSOR, 
  	      &sr_silkwormbd, &state))    {
	    return -1;
	}
    }

    silkworm->board = ((u_int *)(sr_silkwormbd.sr_data))[0];
    silkworm->board_rev = ((u_int *)(sr_silkwormbd.sr_data))[1];

    printk("read_silkworm_bdinfo: silkworm->board = %u, silkworm->board_rev ="
			" %u\n", silkworm->board, silkworm->board_rev);

    return 0;
}


/*
 * int sin_tbl_init()
 *
 * Description:
 *   Initializes the sin(system inventory) records
 *
 * Input(s):
 *
 *
 * Output(s):
 *
 *
 * Returns:
 *   0 if OK, otherwise -1 on error
 *
 */
int __init 
sin_tbl_init(sin_rec_t* sintable, unsigned long tblsize)
{
    unsigned long sin_total = 0;
    int i, retval=0;

    sin_init();

    sin_total = tblsize / sizeof (sin_rec_t);

    /* %%%AC When do we remove these, sin_remove? Also, do we check max */
    for(i=0; i < sin_total; i++) {
	/* %%%AC Check retval ? */
	retval = sin_add(sintable + i);
    }

    /* Read and cache the silkworm board information */
    read_silkworm_bdinfo(&silkworm);

    return retval;
}


/*
 * sin_rec_t* sin_find_sr()
 *
 * Description:
 *   Find the sin record matching the class and subclass. 
 *   Search starts at the current pointer \ state location.
 *
 * Input(s):
 *
 *
 * Output(s):
 *
 *
 * Returns:
 *   0 if OK, otherwise -1 on error
 *
 */
int 
sin_find_sr(int sr_class, int sr_subclass, sin_rec_t* sinrec, int* state)
{
    sin_rec_t sr, *srp;
	
    /*
     * Walk the system inventory records and if a match is found
     * return the record.
     */

     while ((srp = sin_get(&sr, state)) != NULL) {
	 if ((srp->sr_class == sr_class) && 
	              (srp->sr_subclass == sr_subclass)) {
	    memcpy(sinrec, srp, sizeof (sin_rec_t));
	    return 0;
	}
     }

     return -1;
}

/*
 * int ioctl_sinstats(void** arg)
 *
 * Description:
 *   ioctl implementation for sinstats
 *
 * Input(s):
 *  arg: User memory where sinstats are copied.	
 *
 * Output(s):
 *
 * Returns:
 *   0 if OK, otherwise -1 on error
 *
 */
int ioctl_sinstats(void** arg)
{
    sin_stats_t sstats ={0};

    if(sin_stats(&sstats))
	return -ENOENT;

    return (copy_to_user(*arg, &sstats, sizeof (sstats))
	? -EFAULT:0);
}

/*
 * int ioctl_sintbl(void** arg)
 *
 * Description:
 *   ioctl implementation for sintbl, this is all the sin records.
 *
 * Input(s):
 *  arg: User memory where sintbl is copied.	
 *
 * Output(s):
 *
 * Returns:
 *   0 if OK, otherwise -1 on error
 *
 */
int ioctl_sintbl(void** arg)
{
    sin_stats_t sstats ={0};
    sin_rec_t *srp, *tempsrp = NULL;
    int state, nsr;
    int retval, blksize;


    /* Get stats to decide buffer size, etc */
    if(sin_stats(&sstats))
	return -ENOENT;

    /* Verify if the user space is enough to write all records? */
    blksize = sstats.ss_size * sstats.ss_used;
    if (!access_ok (VERIFY_WRITE, *arg, blksize))
	return -EINVAL;

    /* Allocate for the buffer */
    if( !(srp = (sin_rec_t*)kmalloc(blksize, GFP_KERNEL)) )
	return -ENOMEM;

     /* Reset the system inventory search state */
     sin_rewind(&state);

     tempsrp = srp;
     nsr=0;

     /* Read and store each sin record */
     while ( ((tempsrp = sin_get(tempsrp, &state)) != NULL) 
	  && (nsr < sstats.ss_used)) {
	  tempsrp++;
	  nsr++;
      } /* while */

     /* Copy for returning to user space */
     retval =  (__copy_to_user(*arg, srp, blksize) ? -EFAULT : 0);

     kfree(srp);
     return retval;
}

/*
 * int sin_get_swbd(sw_bd* info)
 *
 * Description:
 *   Retrieve the board info structure.
 *
 * Input(s):
 *  info: Buffer where board info is copied to.	
 *
 * Output(s):
 *  info: Contain the Silkworm board info. 
 *
 * Returns:
 *   0 if OK, otherwise -1 on error.
 *
 */
int 
sin_get_swbd(sw_bd* bdinfo)
{
	if (bdinfo == NULL)
		return (-1);

	memcpy(bdinfo, &silkworm, sizeof (sw_bd));
	return 0;
}

/*
 * void sin_copy_raw_atboot(voi* data)
 *
 * Description:
 *   Copies the data from the Bootloader.
 *   It verifies if the version and other element are compatible.
 *   The data contains Statistics followed by sin-table holding records.
 *   Used on platforms that require u-boot bootloader.
 *
 * Input(s):
 *  data: Buffer where the sin info is passed by bootloader
 *
 * Output(s):
 *  sin-stats: Copies the statistics from bootloader data.
 *  sin-table: Copies the sin-table from bootloader data.
 *
 * Returns:
 *   None
 *
 */
void
sin_copy_raw_atboot(void *data)
{
	sin_stats_t *sstat =  (sin_stats_t *)(data);
	struct sin_table *sin_rec = (struct sin_table *) ( (char *)data + sizeof(sin_stats_t));

	if (sstat->ss_version != SIN_VERSION) {
		printk(KERN_ERR "sin_init: Incompatible SIN version, skipping\n");
		return;
	}

	if (sstat->ss_size != sizeof (sin_rec_t)) {
		printk(KERN_ERR "sin_init: Incompatible SIN record size, skipping\n");
		return;
	}

	if (sstat->ss_total > SIN_ENTRIES_MAX) {
		sstat->ss_total =  SIN_ENTRIES_MAX;
		printk(KERN_ERR "sin_init: Truncating SIN record entries\n");
	}

	memcpy( &sin_table_stats, sstat, sizeof(sin_stats_t));

	memcpy( sin_table, sin_rec, sizeof(struct sin_table) * sstat->ss_total);

	{
		/*
		 * Temporary work-around for Europa TTV. Walk through
		 * the sin records and find the one that defines the SWBD.
		 * If it's SWBD88 (Europa), then change it to SWBD98 (EUROPA_TTV)
		 *
		 * Just changing swilkworm.board does not work for FOS, because
		 * the sin command (used by FOS) gets it's data from the
		 * sin records rather than from swilkworm.board
		 *
		 */
		int i;
		sin_rec_t *ptr;

		for (i = 0; i < sin_table_stats.ss_total; i++) {
			// sin_table[i].st_flags &= ~ST_ASSIGNED;
			ptr = &sin_table[i].st_rec;

			if ((ptr->sr_class == SIN_CLASS_BOARD)&&
				(ptr->sr_subclass == SIN_CLASS_BOARD_PROCESSOR)) {
				if (((u_int  *)ptr->sr_data)[0] == 88) {
					((u_int  *)ptr->sr_data)[0] = 98;
					printk("%s: Override SWBD88 with SWBD%d\n",
						   __FUNCTION__, ((u_int  *)ptr->sr_data)[0]);
				}
			}
		}
	}

	/* Read and cache the silkworm board information */
	read_silkworm_bdinfo(&silkworm);
}

EXPORT_SYMBOL(sin_rewind);
EXPORT_SYMBOL(sin_get);
EXPORT_SYMBOL(sin_stats);
EXPORT_SYMBOL(sin_tbl_init);
EXPORT_SYMBOL(ioctl_sinstats);	/* todo - cleanup */
EXPORT_SYMBOL(ioctl_sintbl);	/* todo - cleanup */
EXPORT_SYMBOL(sin_get_swbd);
EXPORT_SYMBOL(sin_find_sr);

