/*
 * Freescale Hypervisor Platform
 *
 * This is the platform file for corenet based SoC setup under the
 * Freescale Hypervisor.
 *
 * Maintained by Kumar Gala (see MAINTAINERS for contact information)
 * Copyright 2008-2011 Freescale Semiconductor Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <asm/ehv_pic.h>
#include <asm/fsl_hcalls.h>
#include <linux/of_platform.h>

void __init fsl_hv_pic_init(void)
{
	struct device_node *np;
	int coreint_flag = 1;

	np = of_find_compatible_node(NULL, NULL, "epapr,hv-pic");

	if (np == NULL) {
		printk(KERN_ERR "Could not find ehv_pic node\n");
		of_node_put(np);
		return;
	}

	if (!of_find_property(np, "has-external-proxy", NULL))
		coreint_flag = 0;

	ehv_pic_init(np, coreint_flag);

	of_node_put(np);
}

/**
 * has_fsl_hypervisor - return TRUE if we're running under FSL hypervisor
 *
 * This function checks to see if we're running under the Freescale
 * hypervisor, and returns zero if we're not, or non-zero if we are.
 *
 * First, it checks if MSR[GS]==1, which means we're running under some
 * hypervisor.  Then it checks if there is a hypervisor node in the device
 * tree.  Currently, that means there needs to be a node in the root called
 * "hypervisor" and which has a property named "fsl,hv-version".
 */
int has_fsl_hypervisor(void)
{
	struct device_node *node;
	int ret;

	if (!(mfmsr() & MSR_GS))
		return 0;

	node = of_find_node_by_path("/hypervisor");
	if (!node)
		return 0;

	ret = of_find_property(node, "fsl,hv-version", NULL) != NULL;

	of_node_put(node);

	return ret;
}
EXPORT_SYMBOL(has_fsl_hypervisor);

#include <asm/silkworm_40x0_misc.h>
void fsl_hv_restart(char *cmd)
{
	printk(KERN_INFO "hv restart\n");
#if (defined(CONFIG_SILKWORM_LC) || defined(CONFIG_SILKWORM_PB)) 
    if(silkworm_is_ha_config() && (silkworm_get_partition_status() == STATUS_ACTIVE))
        fh_partition_stop(-1);
    else
        fh_partition_restart(-1);
#else
    fh_partition_restart(-1);
#endif
}
EXPORT_SYMBOL(fsl_hv_restart);

#ifdef CONFIG_SILKWORM
extern void Disable_MACs(void);
/*
 * fh_platform_reset_self - reboot the current partition 
 * Returns an epapr error code if reboot failed.  Does not return if it succeeds.
 */
unsigned int fh_platform_reset_self(void) 
{
	Disable_MACs();

#ifdef CONFIG_SILKWORM_LC
	if(silkworm_is_ha_config() && (silkworm_get_partition_status() == STATUS_ACTIVE)) {
		return(fh_partition_stop(-1));
	}
	else
		return(fh_partition_restart(-1));
#else
	return(fh_partition_restart(-1));
#endif //CONFIG_SILKWORM_LC
};
EXPORT_SYMBOL(fh_platform_reset_self);


/*
 * fh_platform_reset_peer - reboot the xth peer partition 
 *            where x is the peer partition number starting from 1 based on the ordinal appearance in the dt 
 * With the current two partitions configuration, only fh_platform_reset_peer(1) makes sense.
 *
 * Returns an epapr error code if reboot failed.  Returns 0 if it succeeds.
 */

unsigned int fh_platform_reset_peer(unsigned int partition) 
{       
	unsigned int i;
	struct device_node *peer_node;
	const u32 *door_bell;

	if (partition == 0) {
		printk(KERN_ERR "%s: illegal arguemnt, cannot use fh_platform_reset_peer(0)\n", __func__);
		return(EV_ENOENT);
	}

	peer_node = NULL;                
	for(i=0;i<partition;i++) {
		peer_node = of_find_compatible_node(peer_node, NULL, "fsl,hv-shutdown-doorbell");
		if (!peer_node) {
			printk(KERN_ERR "%s: in loop number %d can't find of_find_compatible_node of hv-doorbell-send-handle\n", __func__, i);
			return(EV_ENOENT);
		}                       
	}

	door_bell = of_get_property(peer_node, "reg", NULL);
	if (!door_bell) {
		printk(KERN_ERR "%s: can't find reg\n", __func__);
		return(EV_ENOENT);
	}

	printk(KERN_ERR "%s: Sending door bell %d\n", __func__, *door_bell);                         
	return (ev_doorbell_send(*door_bell));
	
};
EXPORT_SYMBOL(fh_platform_reset_peer);


void fsl_hv_set_guest_panic(void)
{	
	brcd_hvctl(GUEST_PANIC_STATE, NULL, 0);
}
EXPORT_SYMBOL(fsl_hv_set_guest_panic);


void fsl_hv_clear_guest_panic(void)
{	
	brcd_hvctl(CLEAR_GUEST_IN_PANIC, NULL, 0);
}
EXPORT_SYMBOL(fsl_hv_clear_guest_panic);


void fsl_hv_restart_unc(void)
{
	printk(KERN_INFO "hv restart unconditionally\n");
	fh_partition_restart_unc(-1);
}
EXPORT_SYMBOL(fsl_hv_restart_unc);
#endif //CONFIG_SILKWORM


void fsl_hv_halt(void)
{
	printk(KERN_INFO "hv exit\n");
	fh_partition_stop(-1);
}
EXPORT_SYMBOL(fsl_hv_halt);

/**
 * fsl_hv_claim -- claim a device
 */
int fsl_hv_claim(struct device_node *node)
{
	int ret;
	const u32 *handle = of_get_property(node, "fsl,hv-device-handle", &ret);
	if (!handle || (ret != 4))
		return -ENODEV;
	ret = fh_claim_device(*handle);
	if (ret)
		return -EIO;
	return 0;
}
EXPORT_SYMBOL(fsl_hv_claim);

/**
 * fsl_hv_partition_stop_dma -- perform deferred dma-stop on (stopped) partition
 */
int fsl_hv_partition_stop_dma(struct device_node *node)
{
	int ret;
	const u32 *handle = of_get_property(node, "reg", &ret);
	if (!handle || (ret != 4))
		return -ENODEV;
	ret = fh_partition_stop_dma(*handle);
	if (ret) {
		pr_err("FOO: ret = %d\n", ret);
		return -EIO;
	}
	return 0;
}
EXPORT_SYMBOL(fsl_hv_partition_stop_dma);

/**
 * fsl_hv_partition_reboot -- perform a reboot on a partition
 */
int fsl_hv_partition_reboot(struct device_node *node)
{
	int ret;
	const u32 *handle = of_get_property(node, "reg", &ret);
	if (!handle || (ret != 4))
		return -ENODEV;
	ret = fh_partition_start(*handle, 0, 1);
	if (ret)
		return -EIO;
	return 0;
}
EXPORT_SYMBOL(fsl_hv_partition_reboot);

/**
 * fsl_hv_is_first_boot -- determines if the current partition has been rebooted
 *
 * Returns zero if the current partition has been rebooted
 */
int fsl_hv_is_first_boot(void)
{
	/* TODO: confirm that this node is always called "hypervisor" */
	int ret;
	struct device_node *node = of_find_node_by_name(NULL, "hypervisor");
	if (!node || !of_get_property(node, "fsl,hv-reason-stopped", &ret))
		return 1;
	return 0;
}
EXPORT_SYMBOL(fsl_hv_is_first_boot);

