/*
 * Freescale Hypervisor DMA
 *
 * Issues hcall to enable a device's DMA once it requests a map.
 *
 * This is important when handling a partition reset, since the whole
 * system does not reset, and thus old DMA activity could corrupt the new
 * partition instance.  The driver needs a chance to stop old DMA activity
 * (typically by resetting the device) before the IOMMU gives the device
 * access to memory.
 *
 * Copyright 2010-2011 Freescale Semiconductor Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */ 

#include <linux/dma-mapping.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/spinlock.h>

#include <asm/fsl_hcalls.h>

#include <sysdev/fsl_soc.h>

/*
 * Maintain one set of wrapper ops per real ops.  Besides avoiding the
 * maintenance hassle of having a hardcoded set of ops here that calls
 * through to the real ops dynamically, this allows any ops that are NULL in
 * the real ops struct (e.g.  set_dma_mask) to be NULL in the wrapper struct
 * as well, so the caller can act appropriately.
 */
struct fsl_hv_dma_wrapper {
	struct list_head list;
	struct dma_map_ops *real_ops;
	struct dma_map_ops *wrap_ops;
};

static LIST_HEAD(fsl_hv_dma_wrappers);
static DEFINE_SPINLOCK(fsl_hv_dma_lock);

int fsl_hv_enable_liodn(int liodn_index)
{
	int len, i, hvret;
	const u32 *handles;
	int found = 0;
	int ret = 0;
	struct device_node *of_node = NULL; /* TODO Chintan: this is gonna die!! */

	handles = of_get_property(of_node, "fsl,hv-dma-handle", &len);
	if (!handles || len & 3) {
		pr_warning("%s: warning: no hv dma handle in %s, "
			   "cannot enable\n", __func__, of_node->full_name);
		return -EINVAL;
	}

	for (i = 0; i < len / 4; i++) {
		if (liodn_index >= 0 && liodn_index != i)
			continue;

		hvret = fh_dma_enable(handles[i]);
		if (hvret != 0) {
			pr_err("%s: fh_dma_enable returned %d on "
			       "handle %u in %s\n", __func__, hvret,
			       handles[i], of_node->full_name);
			ret = -EIO;
		} else {
			pr_debug("%s: enabled dma handle %u on %s\n",
				 __func__, handles[i], of_node->full_name);
		}

		found = 1;
	}

	if (!found) {
		pr_err("%s: no liodn index %d in %s, cannot enable hv dma\n",
		       __func__, liodn_index, of_node->full_name);
		return -EINVAL;
	}

	return ret;
}

static void fsl_hv_dev_enable_dma(struct device *dev)
{
	struct dev_archdata *ad = &dev->archdata;

	if (ad->of_node) {
		fsl_hv_enable_liodn(-1);
	} else {
		dev_warn(dev, "%s: warning: no of_node, cannot enable hv dma",
			 __func__);
	}

	set_dma_ops(dev, ad->prev_dma_ops);
}

static void *fsl_hv_dma_alloc_coherent(struct device *dev, size_t size,
				       dma_addr_t *dma_handle, gfp_t gfp)
{
	struct dev_archdata *ad = &dev->archdata;
	fsl_hv_dev_enable_dma(dev);
	return ad->prev_dma_ops->alloc_coherent(dev, size, dma_handle, gfp);
}

static dma_addr_t fsl_hv_dma_map_page(struct device *dev, struct page *page,
				      unsigned long offset, size_t size,
				      enum dma_data_direction dir,
				      struct dma_attrs *attrs)
{
	struct dev_archdata *ad = &dev->archdata;
	fsl_hv_dev_enable_dma(dev);
	return ad->prev_dma_ops->map_page(dev, page, offset, size, dir, attrs);
}

static int fsl_hv_dma_map_sg(struct device *dev, struct scatterlist *sg,
			     int nents, enum dma_data_direction dir,
			     struct dma_attrs *attrs)
{
	struct dev_archdata *ad = &dev->archdata;
	fsl_hv_dev_enable_dma(dev);
	return ad->prev_dma_ops->map_sg(dev, sg, nents, dir, attrs);
}

static int fsl_hv_dma_bus_notify(struct notifier_block *nb,
				 unsigned long action, void *data)
{
	struct device *dev = data;
	struct dev_archdata *ad = &dev->archdata;
	struct fsl_hv_dma_wrapper *wrap;
	struct list_head *node;

	if (action != BUS_NOTIFY_ADD_DEVICE)
		return NOTIFY_DONE;

	ad->prev_dma_ops = get_dma_ops(dev);
	if (!ad->prev_dma_ops)
		return NOTIFY_DONE;

	spin_lock(&fsl_hv_dma_lock);

	list_for_each(node, &fsl_hv_dma_wrappers) {
		wrap = list_entry(node, struct fsl_hv_dma_wrapper, list);

		if (wrap->real_ops == ad->prev_dma_ops)
			break;
	}

	if (node == &fsl_hv_dma_wrappers) {
		wrap = kmalloc(sizeof(struct fsl_hv_dma_wrapper), GFP_KERNEL);
		if (!wrap) {
			spin_unlock(&fsl_hv_dma_lock);
			pr_err("%s: out of memory\n", __func__);
			return NOTIFY_BAD;
		}

		wrap->wrap_ops = kmalloc(sizeof(struct dma_map_ops),
					 GFP_KERNEL);
		if (!wrap->wrap_ops) {
			spin_unlock(&fsl_hv_dma_lock);
			pr_err("%s: out of memory\n", __func__);
			kfree(wrap);
			return NOTIFY_BAD;
		}

		wrap->real_ops = ad->prev_dma_ops;
		memcpy(wrap->wrap_ops, wrap->real_ops,
		       sizeof(struct dma_map_ops));

		wrap->wrap_ops->alloc_coherent = fsl_hv_dma_alloc_coherent;
		wrap->wrap_ops->map_page = fsl_hv_dma_map_page;
		wrap->wrap_ops->map_sg = fsl_hv_dma_map_sg;

		list_add(&wrap->list, &fsl_hv_dma_wrappers);
	}

	spin_unlock(&fsl_hv_dma_lock);

	set_dma_ops(dev, wrap->wrap_ops);
	return NOTIFY_DONE;
}

static struct notifier_block fsl_hv_dma_plat_bus_notifier = {
	.notifier_call = fsl_hv_dma_bus_notify,
	.priority = -1, /* Run after the real dma ops */
};

static struct notifier_block fsl_hv_dma_of_bus_notifier = {
	.notifier_call = fsl_hv_dma_bus_notify,
	.priority = -1,
};

int __init fsl_hv_dma_setup_bus_notifier(void)
{
//	pamu_enable_liodn = fsl_hv_enable_liodn;

	bus_register_notifier(&platform_bus_type,
			      &fsl_hv_dma_plat_bus_notifier);
	bus_register_notifier(&of_platform_bus_type,
			      &fsl_hv_dma_of_bus_notifier);

	return 0;
}

