/*
 * Support for indirect PCI bridges.
 *
 * Copyright (C) 1998 Gabriel Paubert.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/prom.h>
#include <asm/pci-bridge.h>
#include <asm/machdep.h>

#if defined(CONFIG_SILKWORM_PCIE_FAILOVER_HACK)
#include <asm/silkworm_40x0_misc.h>
#endif
#ifdef CONFIG_FSL_HYPERVISOR
#include <linux/fsl_hypervisor.h>
#endif

#ifdef CONFIG_SILKWORM_TRI_FPGA
extern int silkworm_tri_domain;
extern int silkworm_tri_offset;
#endif

static int
indirect_read_config(struct pci_bus *bus, unsigned int devfn, int offset,
		     int len, u32 *val)
{
	struct pci_controller *hose = pci_bus_to_host(bus);
	volatile void __iomem *cfg_data;
	u8 cfg_type = 0;
	u32 bus_no, reg;

#if defined(CONFIG_SILKWORM_PCIE_FAILOVER_HACK) 
	extern struct silkworm_pcie_shared_mem *spsm;
	unsigned long flags = 0;
#endif


	if (hose->indirect_type & PPC_INDIRECT_TYPE_NO_PCIE_LINK) {
		if (bus->number != hose->first_busno)
			return PCIBIOS_DEVICE_NOT_FOUND;
		if (devfn != 0)
			return PCIBIOS_DEVICE_NOT_FOUND;
	}

	if (ppc_md.pci_exclude_device)
		if (ppc_md.pci_exclude_device(hose, bus->number, devfn))
			return PCIBIOS_DEVICE_NOT_FOUND;

	if (hose->indirect_type & PPC_INDIRECT_TYPE_SET_CFG_TYPE)
		if (bus->number != hose->first_busno)
			cfg_type = 1;

	bus_no = (bus->number == hose->first_busno) ?
			hose->self_busno : bus->number;

#ifdef CONFIG_SILKWORM_TRI_FPGA
	/*
	 * A Brocade-custom FPGA is being used to interface the three NTBs to the
	 * CPU to reduce the latency of a PCIe switch and PCIe-PCI bridge. Due
	 * to constraints of this FPGA, the config space for the NTBs is accessible
	 * only thru device 0's config address space starting at offset 0x100.
	 * Device 0 is an endpoint, not a bridge, so it has to act as a proxy for
	 * the NTBs as far as config address space is concerned.
	 * 
	 * So check for NTB access (Bus 1, dev 1, 2, or 3) and add the appropriate
	 * offset.
	 */

	if ((hose->global_number == silkworm_tri_domain)&&
		(bus_no == (hose->first_busno + 1))&&(PCI_SLOT(devfn) > 0)) {
		pr_err("%s: Domain %d bus %d first bus %d "
			   "dev %d reg 0x%x with OFFset 0x%x\n",
			   __FUNCTION__, hose->global_number, bus_no,
			   hose->first_busno,
			   PCI_SLOT(devfn), offset, offset +  silkworm_tri_offset);
		offset += silkworm_tri_offset;
	}

#endif

	if (hose->indirect_type & PPC_INDIRECT_TYPE_EXT_REG)
		reg = ((offset & 0xf00) << 16) | (offset & 0xfc);
	else
		reg = offset & 0xfc;

#ifdef CONFIG_FSL_HYPERVISOR
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK
{
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK_DEBUG
    extern bool active_pcie_owner;
    if(!active_pcie_owner){
        printk(KERN_ERR "**crs - (B,D,F,R,L) = (%d,%d,%d,0x%x,%d)\n",bus_no,(devfn & 0xf8) >> 3,devfn & 0x7,offset,len);
		*val = 0xffffffff;
		return PCIBIOS_SUCCESSFUL;
    }
#endif
	if (has_fsl_hypervisor())
	{
	/* attempt to grab pci lock */
		local_irq_save(flags);
	arch_spin_lock((arch_spinlock_t *)&spsm->pci_lock);
	}

}
#endif
#endif
	
	if (hose->indirect_type & PPC_INDIRECT_TYPE_BIG_ENDIAN)
		out_be32(hose->cfg_addr, (0x80000000 | (bus_no << 16) |
			 (devfn << 8) | reg | cfg_type));
	else
		out_le32(hose->cfg_addr, (0x80000000 | (bus_no << 16) |
			 (devfn << 8) | reg | cfg_type));

	/*
	 * Note: the caller has already checked that offset is
	 * suitably aligned and that len is 1, 2 or 4.
	 */
	cfg_data = hose->cfg_data + (offset & 3);
	switch (len) {
	case 1:
		*val = in_8(cfg_data);
		break;
	case 2:
		*val = in_le16(cfg_data);
		break;
	default:
		*val = in_le32(cfg_data);
		break;
	}
#ifdef CONFIG_FSL_HYPERVISOR
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK
	if (has_fsl_hypervisor())
	{
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK_DEBUG
    printk(KERN_INFO "r -> (%d,%d,%d,0x%x,%d,%x)\n",bus_no,(devfn & 0xf8) >> 3,devfn & 0x7,offset,len,*val);
#endif
	/* clear pci lock */
	arch_spin_unlock((arch_spinlock_t *)&spsm->pci_lock);
	local_irq_restore(flags);
	}
#endif
#endif

	return PCIBIOS_SUCCESSFUL;
}

static int
indirect_write_config(struct pci_bus *bus, unsigned int devfn, int offset,
		      int len, u32 val)
{
	struct pci_controller *hose = pci_bus_to_host(bus);
	volatile void __iomem *cfg_data;
	u8 cfg_type = 0;
	u32 bus_no, reg;

	if (hose->indirect_type & PPC_INDIRECT_TYPE_NO_PCIE_LINK) {
		if (bus->number != hose->first_busno)
			return PCIBIOS_DEVICE_NOT_FOUND;
		if (devfn != 0)
			return PCIBIOS_DEVICE_NOT_FOUND;
	}

	if (ppc_md.pci_exclude_device)
		if (ppc_md.pci_exclude_device(hose, bus->number, devfn))
			return PCIBIOS_DEVICE_NOT_FOUND;

	if (hose->indirect_type & PPC_INDIRECT_TYPE_SET_CFG_TYPE)
		if (bus->number != hose->first_busno)
			cfg_type = 1;

	bus_no = (bus->number == hose->first_busno) ?
			hose->self_busno : bus->number;

#ifdef CONFIG_SILKWORM_TRI_FPGA
	/*
	 * A Brocade-custom FPGA is being used to interface the three NTBs to the
	 * CPU to reduce the latency of a PCIe switch and PCIe-PCI bridge. Due
	 * to constraints of this FPGA, the config space for the NTBs is accessible
	 * only thru device 0's config address space starting at offset 0x800.
	 * Device 0 is an endpoint, not a bridge, so it has to act as a proxy for
	 * the NTBs as far as config address space is concerned.
	 * 
	 * So check for NTB access (Bus 1, dev 1, 2, or 3) and add the appropriate
	 * offset.
	 */
	if ((hose->global_number == silkworm_tri_domain)&&
		(bus_no == (hose->first_busno + 1))&&(devfn & 0x18)) {
		pr_err("%s: Domain %d bus %d first bus %d "
				 "dev %d reg 0x%x with offset 0x%x\n",
				 __FUNCTION__, hose->global_number, bus_no,
				 hose->first_busno,
				 PCI_SLOT(devfn), offset, offset +  silkworm_tri_offset);
		offset += silkworm_tri_offset;
	}
#endif

	if (hose->indirect_type & PPC_INDIRECT_TYPE_EXT_REG)
		reg = ((offset & 0xf00) << 16) | (offset & 0xfc);
	else
		reg = offset & 0xfc;


#if defined(CONFIG_SILKWORM_PCIE_FAILOVER_HACK)
{
    extern bool active_pcie_owner;
    if(!active_pcie_owner){
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK_DEBUG
        printk(KERN_INFO "**cws (B,D,F,R,L,V) = (%d,%d,%d,0x%x,%d,0x%x)\n",bus_no,(devfn & 0xf8) >> 3,devfn & 0x7,offset,len,val);
#endif
        return PCIBIOS_SUCCESSFUL;
    }
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK_DEBUG
    else
        printk(KERN_INFO "**cwa (B,D,F,R,L,V) = (%d,%d,%d,0x%x,%d,0x%x)\n",bus_no,(devfn & 0xf8) >> 3,devfn & 0x7,offset,len,val);
#endif    
}
#endif


	if (hose->indirect_type & PPC_INDIRECT_TYPE_BIG_ENDIAN)
		out_be32(hose->cfg_addr, (0x80000000 | (bus_no << 16) |
			 (devfn << 8) | reg | cfg_type));
	else
		out_le32(hose->cfg_addr, (0x80000000 | (bus_no << 16) |
			 (devfn << 8) | reg | cfg_type));

	/* surpress setting of PCI_PRIMARY_BUS */
	if (hose->indirect_type & PPC_INDIRECT_TYPE_SURPRESS_PRIMARY_BUS)
		if ((offset == PCI_PRIMARY_BUS) &&
			(bus->number == hose->first_busno))
		val &= 0xffffff00;

	/* Workaround for PCI_28 Errata in 440EPx/GRx */
	if ((hose->indirect_type & PPC_INDIRECT_TYPE_BROKEN_MRM) &&
			offset == PCI_CACHE_LINE_SIZE) {
		val = 0;
	}

	/*
	 * Note: the caller has already checked that offset is
	 * suitably aligned and that len is 1, 2 or 4.
	 */
	cfg_data = hose->cfg_data + (offset & 3);
	switch (len) {
	case 1:
		out_8(cfg_data, val);
		break;
	case 2:
		out_le16(cfg_data, val);
		break;
	default:
		out_le32(cfg_data, val);
		break;
	}
#ifdef CONFIG_SILKWORM_PCIE_FAILOVER_HACK_DEBUG
    printk(KERN_INFO "w ->(%d,%d,%d,0x%x,%d,%x)\n",bus_no,(devfn & 0xf8) >> 3,devfn & 0x7,offset,len,val);
#endif
	return PCIBIOS_SUCCESSFUL;
}

static struct pci_ops indirect_pci_ops =
{
	.read = indirect_read_config,
	.write = indirect_write_config,
};

void __init
setup_indirect_pci(struct pci_controller* hose,
		   resource_size_t cfg_addr,
		   resource_size_t cfg_data, u32 flags)
{
	resource_size_t base = cfg_addr & PAGE_MASK;
	void __iomem *mbase;

	mbase = ioremap(base, PAGE_SIZE);
	hose->cfg_addr = mbase + (cfg_addr & ~PAGE_MASK);
	if ((cfg_data & PAGE_MASK) != base)
		mbase = ioremap(cfg_data & PAGE_MASK, PAGE_SIZE);
	hose->cfg_data = mbase + (cfg_data & ~PAGE_MASK);
	hose->ops = &indirect_pci_ops;
	hose->indirect_type = flags;
}
