/* ePAPR hypervisor byte channel device driver
 *
 * Copyright 2009-2010 Freescale Semiconductor, Inc.
 *
 * Author: Timur Tabi <timur@freescale.com>
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 *
 * This driver support three distinct interfaces, all of which are related to
 * ePAPR hypervisor byte channels.
 *
 * 1) An early-console (udbg) driver.  This provides early console output
 * through a byte channel.  The byte channel handle must be specified in a
 * Kconfig option.
 *
 * 2) A normal console driver.  Output is sent to the byte channel designated
 * for stdout in the device tree.  The console driver is for handling kernel
 * printk calls.
 *
 * 3) A tty driver, which is used to handle user-space input and output.  The
 * byte channel used for the console is designated as the default tty.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/sysrq.h>
#include <asm/epapr_hcalls.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/of_device.h>
#include <linux/cdev.h>
#include <linux/console.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <asm/udbg.h>
#include <linux/serial.h>
#include <linux/delay.h>

/* Per-byte channel private data */
struct ehv_bc_data {
	struct device *dev;
	struct tty_port port;
	struct tty_struct *ttys;
	uint32_t handle;
	unsigned int rx_irq;
	unsigned int tx_irq;
};

#ifdef CONFIG_MAGIC_SYSRQ
static unsigned long sysrq_initiated = 0;
static unsigned long sysrq_requested = 0;

#define MAGIC_SYSRQ_MAX_TIMEOUT 5*HZ
#define MAGIC_SYSRQ_IGNORE_CHAR 0 // NULL
#define MAGIC_SYSRQ_INITIATED_CHAR '\x0f' // ^O
#define MAGIC_SYSRQ_REQUESTED_CHAR '\x0e' // ^N
#endif

/* Array of byte channel objects */
static struct ehv_bc_data *bcs;

/* Byte channel handle for stdout (and stdin), taken from device tree */
static unsigned int stdout_bc;

/* Virtual IRQ for the byte channel handle for stdin, taken from device tree */
static unsigned int stdout_irq;

static struct serial_cir_buf_t *ehv_serial_log_buf = NULL;
static int *sl_next_free = NULL;

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
static struct serial_cir_buf_t *ehv_serial_log_buf_ttys2 = NULL;
static struct serial_cir_buf_t *ehv_serial_log_buf_ttys3 = NULL;
static int *sl_next_free_ttys2 = NULL;
static int *sl_next_free_ttys3 = NULL;

static uint32_t bc_handle_ttyS2 = 0;
static uint32_t bc_handle_ttyS3 = 0;

struct tty_struct ttys2;
struct tty_struct ttys3;

struct ehv_bc_data *bc_ttys2 = NULL;
struct ehv_bc_data *bc_ttys3 = NULL;

static uint32_t restrict_access_to_ttys2_ttys3 = 0;
#endif

/**************************** SUPPORT FUNCTIONS ****************************/

/*
 * return TRUE if we're running under FSL hypervisor
 *
 * This function checks to see if we're running under the Freescale
 * hypervisor, and returns zero if we're not, or non-zero if we are.
 *
 * First, it checks if MSR[GS]==1, which means we're running under some
 * hypervisor.  Then it checks if there is a hypervisor node in the device
 * tree.  Currently, that means there needs to be a node in the root called
 * "hypervisor" and which has a property named "fsl,hv-version".
 */

static inline void serial_log_save_char(unsigned char c)
{
	if (!ehv_serial_log_buf) return;

	if (!c) return;

	if (*sl_next_free >= SERIAL_LOG_FULL_LEN) {
		*sl_next_free = SERIAL_LOG_WRAP_LEN;
		ehv_serial_log_buf->wraped_once++;
	}
	ehv_serial_log_buf->data[(*sl_next_free)++] = c;
}

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
static inline void serial_log_save_char_rx(uint32_t handle, unsigned char c)
{
	static struct serial_cir_buf_t *ehv_serial_log_buf_rx = NULL;
	static int *sl_next_free_rx = NULL;

	if (bc_handle_ttyS2 == handle) {
		ehv_serial_log_buf_rx = ehv_serial_log_buf_ttys2;
		sl_next_free_rx = sl_next_free_ttys2;
	}else if (bc_handle_ttyS3 == handle ) {
		ehv_serial_log_buf_rx = ehv_serial_log_buf_ttys3;
		sl_next_free_rx = sl_next_free_ttys3;
	}

	if (!ehv_serial_log_buf_rx) return;

	if (!c) return;

	if (*sl_next_free_rx >= SERIAL_LOG_FULL_LEN) {
		*sl_next_free_rx = SERIAL_LOG_WRAP_LEN;
		ehv_serial_log_buf_rx->wraped_once++;
	}
	ehv_serial_log_buf_rx->data[(*sl_next_free_rx)++] = c;
}
#endif

static int has_fsl_hypervisor(void)
{
	struct device_node *node;
	int ret;

	if (!(mfmsr() & MSR_GS))
		return 0;

	node = of_find_node_by_path("/hypervisor");
	if (!node)
		return 0;

	ret = of_find_property(node, "fsl,hv-version", NULL) != NULL;

	of_node_put(node);

	return ret;
}

/*
 * find the byte channel handle to use for the console
 *
 * The byte channel to be used for the console is specified via a "stdout"
 * property in the /chosen node.
 *
 * For compatible with legacy device trees, we also look for a "stdout" alias.
 */
static void find_console_handle(void)
{
	struct device_node *np, *np2;
	const char *sprop = NULL;
	const uint32_t *iprop;

	np = of_find_node_by_path("/chosen");
	if (np)
		sprop = of_get_property(np, "stdout-path", NULL);

	if (!np || !sprop) {
		of_node_put(np);
		np = of_find_node_by_name(NULL, "aliases");
		if (np)
			sprop = of_get_property(np, "stdout", NULL);
	}

	if (!sprop) {
		of_node_put(np);
		return;
	}

	/* We don't care what the aliased node is actually called.  We only
	 * care if it's compatible with "epapr,hv-byte-channel", because that
	 * indicates that it's a byte channel node.  We use a temporary
	 * variable, 'np2', because we can't release 'np' until we're done with
	 * 'sprop'.
	 */
	np2 = of_find_node_by_path(sprop);
	of_node_put(np);
	np = np2;
	if (!np) {
		pr_warning("ehv-bc: stdout node '%s' does not exist\n", sprop);
		return;
	}

	/* Is it a byte channel? */
	if (!of_device_is_compatible(np, "epapr,hv-byte-channel"))
		goto exit;

	stdout_irq = irq_of_parse_and_map(np, 0);
	if (stdout_irq == NO_IRQ) {
		pr_err("ehv-bc: no 'interrupts' property in %s node\n", sprop);
		goto exit;
	}

	iprop = of_get_property(np, "reg", NULL);
	if (!iprop) {
		pr_err("ehv-bc: no 'reg' property in %s node\n", np->name);
		goto exit;
	}
	stdout_bc = be32_to_cpu(*iprop);

exit:
	of_node_put(np);
}

/*************************** EARLY CONSOLE DRIVER ***************************/

#ifdef CONFIG_PPC_EARLY_DEBUG_EHV_BC

/*
 * send a byte to a byte channel, wait if necessary
 *
 * This function sends a byte to a byte channel, and it waits and
 * retries if the byte channel is full.  It returns if the character
 * has been sent, or if some error has occurred.
 *
 */
static void byte_channel_spin_send(const char data)
{
	int ret, count;

	do {
		count = 1;
		ret = ev_byte_channel_send(CONFIG_PPC_EARLY_DEBUG_EHV_BC_HANDLE,
					   &count, &data);
	} while (ret == EV_EAGAIN);

	// Append the character to log buffer
	serial_log_save_char(data);
}

/*
 * The udbg subsystem calls this function to display a single character.
 * We convert CR to a CR/LF.
 */
static void ehv_bc_udbg_putc(char c)
{
	if (c == '\n')
		byte_channel_spin_send('\r');

	byte_channel_spin_send(c);
}

/*
 * early console initialization
 *
 * PowerPC kernels support an early printk console, also known as udbg.
 * This function must be called via the ppc_md.init_early function pointer.
 * At this point, the device tree has been unflattened, so we can obtain the
 * byte channel handle for stdout.
 *
 * We only support displaying of characters (putc).  We do not support
 * keyboard input.
 */
void __init udbg_init_ehv_bc(void)
{
	unsigned int rx_count, tx_count;
	unsigned int ret;

	/* Check if we're running as a guest of a hypervisor */
	if (!(mfmsr() & MSR_GS))
		return;

	/* Verify the byte channel handle */
	ret = ev_byte_channel_poll(CONFIG_PPC_EARLY_DEBUG_EHV_BC_HANDLE,
				   &rx_count, &tx_count);
	if (ret)
		return;

	udbg_putc = ehv_bc_udbg_putc;
	register_early_udbg_console();

	udbg_printf("ehv-bc: early console using byte channel handle %u\n",
		    CONFIG_PPC_EARLY_DEBUG_EHV_BC_HANDLE);
}

#endif

/****************************** CONSOLE DRIVER ******************************/

static struct tty_driver *ehv_bc_driver;

/*
 * Byte channel console sending worker function.
 *
 * For consoles, if the output buffer is full, we should just spin until it
 * clears.
 */

#ifdef CONFIG_SILKWORM     
#define ehv_bc_console_byte_channel_send_timeout 		25
#define ehv_bc_console_byte_channel_send_MAX_TIMEOUT 		(5*ehv_bc_console_byte_channel_send_timeout)

#ifdef CONFIG_WATCHDOG
extern void kick_watchdog(void);
#endif
#endif

static int ehv_bc_console_byte_channel_send(unsigned int handle, const char *s,
			     unsigned int count)
{
	unsigned int len;
	int ret = 0;
	int temp=0, count1=count;
        char *ctemp;
        ctemp=s;
#ifdef CONFIG_SILKWORM     
	int time;
	
	time = 0;	
#ifdef CONFIG_WATCHDOG
	kick_watchdog();
#endif
	while (count) {
		len = min_t(unsigned int, count, EV_BYTE_CHANNEL_MAX_BYTES);
		ret = ev_byte_channel_send(handle, &len, s);

		if (ret != EV_EAGAIN) {
			count -= len;
			s += len;	
		} else {			 
#ifdef CONFIG_WATCHDOG
			kick_watchdog();
#endif
			mdelay(ehv_bc_console_byte_channel_send_timeout);							
			time += ehv_bc_console_byte_channel_send_timeout;

			if (time>ehv_bc_console_byte_channel_send_MAX_TIMEOUT) {
				break;
			}
		}
	}
#else //CONFIG_SILKWORM
	while (count) {
		len = min_t(unsigned int, count, EV_BYTE_CHANNEL_MAX_BYTES);
		do {
			ret = ev_byte_channel_send(handle, &len, s);
		} while (ret == EV_EAGAIN);
		count -= len;
		s += len;
	}
#endif //CONFIG_SILKWORM

	for(temp=0; temp < count1; temp++) {
                serial_log_save_char(ctemp[temp]);
        }

	return 0;
}

/*
 * write a string to the console
 *
 * This function gets called to write a string from the kernel, typically from
 * a printk().  This function spins until all data is written.
 *
 * We copy the data to a temporary buffer because we need to insert a \r in
 * front of every \n.  It's more efficient to copy the data to the buffer than
 * it is to make multiple hcalls for each character or each newline.
 */
static void ehv_bc_console_write(struct console *co, const char *s,
				 unsigned int count)
{
	unsigned int handle = (unsigned int)co->data;
	char s2[EV_BYTE_CHANNEL_MAX_BYTES];
	unsigned int i, j = 0;
	char c;

	for (i = 0; i < count; i++) {
		c = *s++;

		if (c == '\n')
			s2[j++] = '\r';

		s2[j++] = c;
		if (j >= (EV_BYTE_CHANNEL_MAX_BYTES - 1)) {
			if (ehv_bc_console_byte_channel_send(handle, s2, j))
				return;
			j = 0;
		}
	}

	if (j)
		ehv_bc_console_byte_channel_send(handle, s2, j);
}

/*
 * When /dev/console is opened, the kernel iterates the console list looking
 * for one with ->device and then calls that method. On success, it expects
 * the passed-in int* to contain the minor number to use.
 */
static struct tty_driver *ehv_bc_console_device(struct console *co, int *index)
{
	*index = co->index;

	return ehv_bc_driver;
}

static struct console ehv_bc_console = {
	.name		= "ttyEHV",
	.write		= ehv_bc_console_write,
	.device		= ehv_bc_console_device,
	.flags		= CON_PRINTBUFFER | CON_ENABLED,
};

/*
 * Console initialization
 *
 * This is the first function that is called after the device tree is
 * available, so here is where we determine the byte channel handle and IRQ for
 * stdout/stdin, even though that information is used by the tty and character
 * drivers.
 */
static int __init ehv_bc_console_init(void)
{
	find_console_handle();
	if (!stdout_bc) {
		pr_debug("ehv-bc: stdout is not a byte channel\n");
		return -ENODEV;
	}

#ifdef CONFIG_PPC_EARLY_DEBUG_EHV_BC
	/* Print a friendly warning if the user chose the wrong byte channel
	 * handle for udbg.
	 */
	if (stdout_bc != CONFIG_PPC_EARLY_DEBUG_EHV_BC_HANDLE)
		pr_warning("ehv-bc: udbg handle %u is not the stdout handle\n",
			   CONFIG_PPC_EARLY_DEBUG_EHV_BC_HANDLE);
#endif

	ehv_bc_console.data = (void *)stdout_bc;

	/* add_preferred_console() must be called before register_console(),
	   otherwise it won't work.  However, we don't want to enumerate all the
	   byte channels here, either, since we only care about one. */

	add_preferred_console(ehv_bc_console.name, ehv_bc_console.index, NULL);
	register_console(&ehv_bc_console);

	pr_info("ehv-bc: registered console driver for byte channel %u\n",
		stdout_bc);

	return 0;
}
console_initcall(ehv_bc_console_init);

/******************************** TTY DRIVER ********************************/

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
/*
	ISR registered to receive dpconsole data with out an application 
	opening a deivce and activating a port.
*/
static irqreturn_t ehv_bc_tty_rx_isr_ttys2_ttys3(int irq, void *data)
{
	struct ehv_bc_data *bc = data;
	unsigned int rx_count, tx_count, len;
	int count;
	char buffer[EV_BYTE_CHANNEL_MAX_BYTES];
	int loop=0;

#ifdef CONFIG_MAGIC_SYSRQ
	unsigned long ind, sysrq_timeout;
#endif

	/* Find out how much data needs to be read, and then ask the TTY layer
	 * if it can handle that much.  We want to ensure that every byte we
	 * read from the byte channel will be accepted by the TTY layer.
	 */
	ev_byte_channel_poll(bc->handle, &rx_count, &tx_count);
	count = rx_count;

	/* 'count' is the maximum amount of data the TTY layer can accept at
	 * this time.  However, during testing, I was never able to get 'count'
	 * to be less than 'rx_count'.  I'm not sure whether I'm calling it
	 * correctly.
	 */

	while (count > 0) {
		len = min_t(unsigned int, count, sizeof(buffer));

		/* Read some data from the byte channel.  This function will
		 * never return more than EV_BYTE_CHANNEL_MAX_BYTES bytes.
		 */
		ev_byte_channel_receive(bc->handle, &len, buffer);

		/* 'len' is now the amount of data that's been received. 'len'
		 * can't be zero, and most likely it's equal to one.
		 */

#ifdef CONFIG_MAGIC_SYSRQ
		for (ind = 0; ind < len; ind++) {
			if (sysrq_on()) {
				if (buffer[ind] == '\x0f') { /* ^O */
					buffer[ind] = 0;
					if (!sysrq_requested) {
						sysrq_requested = jiffies;
						continue; /* don't print sysrq key */
					}
					sysrq_requested = 0;
				} else if (sysrq_requested) {
					sysrq_timeout = sysrq_requested + HZ*5;
					sysrq_requested = 0;
					if (time_before(jiffies, sysrq_timeout)) {
					//	handle_sysrq(buffer[ind], ttys);
						buffer[ind] = 0;
						continue;
					}
				}
			}
		}
#endif

	 	for(loop=0; loop < len; loop++) {
               serial_log_save_char_rx(bc->handle, buffer[loop]);
        }

		count -= len;
	}

	return IRQ_HANDLED;
}

int redirect_ttys2_ttys3()
{
    int ret;

    restrict_access_to_ttys2_ttys3 = 1;

    ret = request_irq(bc_ttys2->rx_irq, ehv_bc_tty_rx_isr_ttys2_ttys3, 0, "ehv-bc", bc_ttys2);
    if (ret < 0)
    {
        dev_err(bc_ttys2->dev, "could not request rx irq %u (ret=%i)\n", bc_ttys2->rx_irq, ret);
        return ret;
    }

    ret = request_irq(bc_ttys3->rx_irq, ehv_bc_tty_rx_isr_ttys2_ttys3, 0, "ehv-bc", bc_ttys3);
    if (ret < 0)
    {
        dev_err(bc_ttys3->dev, "could not request rx irq %u (ret=%i)\n", bc_ttys3->rx_irq, ret);
        return ret;
    }
}

EXPORT_SYMBOL(redirect_ttys2_ttys3);
#endif

/*
 * byte channel receive interupt handler
 *
 * This ISR is called whenever data is available on a byte channel.
 */
static irqreturn_t ehv_bc_tty_rx_isr(int irq, void *data)
{
	struct ehv_bc_data *bc = data;
	struct tty_struct *ttys = bc->ttys;
	unsigned int rx_count, tx_count, len;
	int count;
	char buffer[EV_BYTE_CHANNEL_MAX_BYTES];
	int ret;

#ifdef CONFIG_MAGIC_SYSRQ
	unsigned long ind, sysrq_timeout;
#endif

	/* Find out how much data needs to be read, and then ask the TTY layer
	 * if it can handle that much.  We want to ensure that every byte we
	 * read from the byte channel will be accepted by the TTY layer.
	 */
	ev_byte_channel_poll(bc->handle, &rx_count, &tx_count);
	count = tty_buffer_request_room(ttys, rx_count);

	/* 'count' is the maximum amount of data the TTY layer can accept at
	 * this time.  However, during testing, I was never able to get 'count'
	 * to be less than 'rx_count'.  I'm not sure whether I'm calling it
	 * correctly.
	 */

	while (count > 0) {
		len = min_t(unsigned int, count, sizeof(buffer));

		/* Read some data from the byte channel.  This function will
		 * never return more than EV_BYTE_CHANNEL_MAX_BYTES bytes.
		 */
		ev_byte_channel_receive(bc->handle, &len, buffer);

		/* 'len' is now the amount of data that's been received. 'len'
		 * can't be zero, and most likely it's equal to one.
		 */

#ifdef CONFIG_MAGIC_SYSRQ
		for (ind = 0; ind < len; ind++) {
			if (sysrq_on()) {
				if (buffer[ind] == MAGIC_SYSRQ_INITIATED_CHAR) {
					buffer[ind] = MAGIC_SYSRQ_IGNORE_CHAR;
					if (!sysrq_initiated) {
						sysrq_requested = 0;
						sysrq_initiated = jiffies;
						continue;
					}
				} else if (buffer[ind] == MAGIC_SYSRQ_REQUESTED_CHAR) {
					buffer[ind] = MAGIC_SYSRQ_IGNORE_CHAR;
					if (sysrq_initiated && !sysrq_requested) {
						sysrq_timeout = sysrq_initiated + MAGIC_SYSRQ_MAX_TIMEOUT;
						if (time_before(jiffies, sysrq_timeout)) {
							sysrq_requested = 1;
							continue;
						}
					}
				} else if (sysrq_initiated && sysrq_requested) {
					sysrq_timeout = sysrq_initiated + MAGIC_SYSRQ_MAX_TIMEOUT;
					if (time_before(jiffies, sysrq_timeout)) {
						sysrq_initiated = sysrq_requested = 0;
						handle_sysrq(buffer[ind], ttys);
						buffer[ind] = MAGIC_SYSRQ_IGNORE_CHAR;
						continue;
					}
				}
				sysrq_initiated = sysrq_requested = 0;
			}
		}
#endif


		/* Pass the received data to the tty layer.  Note that this
		 * function calls tty_buffer_request_room(), so I'm not sure if
		 * we should have also called tty_buffer_request_room().
		 */
		ret = tty_insert_flip_string(ttys, buffer, len);

		/* 'ret' is the number of bytes that the TTY layer accepted.
		 * If it's not equal to 'len', then it means the buffer is
		 * full, which should never happen.  If it does happen, we can
		 * exit gracefully, but we drop the last 'len - ret' characters
		 * that we read from the byte channel.
		 */
		if (ret != len)
			break;

		count -= len;
	}

	/* Tell the tty layer that we're done. */
	tty_flip_buffer_push(ttys);

	return IRQ_HANDLED;
}

/*
 * byte channel transmit interupt handler
 *
 * This ISR is called whenever space becomes available for transmitting
 * characters on a byte channel.
 */
static irqreturn_t ehv_bc_tty_tx_isr(int irq, void *data)
{
	struct ehv_bc_data *bc = data;

	// disable_irq_nosync(bc->tx_irq);
	tty_wakeup(bc->ttys);

	return IRQ_HANDLED;
}

static int ehv_bc_tty_write(struct tty_struct *ttys, const unsigned char *s,
			    int count)
{
	struct ehv_bc_data *bc = ttys->driver_data;
	unsigned int ret;
	unsigned int len;
	int written = 0;
	int temp=0, count1=count;
	char *ctemp;
	ctemp=s;
	while (count) {
		len = min_t(unsigned, count, EV_BYTE_CHANNEL_MAX_BYTES);
		ret = ev_byte_channel_send(bc->handle, &len, s);
		if (ret) {
			if (ret == EV_EAGAIN){
                    /* If we can't write all the data at once,
                     * then enable the TX IRQ.  We'll get an
                     * interrupt when there's more room in the
                     * output buffer.
                     */
                    // enable_irq(bc->tx_irq);
            }
			break;
		}
		written += len;
		count -= len;
		s += len;
	}

	for(temp=0; temp < count1; temp++) {
		serial_log_save_char(ctemp[temp]);
	}

	return written;
}

/* This function can be called multiple times for a given tty_struct, which is
 * why we initialize bc->ttys in ehv_bc_tty_port_activate() instead.
 */
static int ehv_bc_tty_open(struct tty_struct *ttys, struct file *filp)
{
	struct ehv_bc_data *bc = &bcs[ttys->index];

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
	if (restrict_access_to_ttys2_ttys3 &&
        (bc_handle_ttyS2 == bc->handle || bc_handle_ttyS3 == bc->handle))
	{
		return -EACCES;
	}
#endif

	return tty_port_open(&bc->port, ttys, filp);
}

static void ehv_bc_tty_close(struct tty_struct *ttys, struct file *filp)
{
#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
	struct ehv_bc_data *bc=NULL;

	if (ttys)
		bc = ttys->driver_data;

	if (bc == NULL)
		return -EACCES;

	if (restrict_access_to_ttys2_ttys3 &&
        (bc_handle_ttyS2 == bc->handle || bc_handle_ttyS3 == bc->handle))
	{
		return -EACCES;
	}
#else
	struct ehv_bc_data *bc = ttys->driver_data;
#endif
	tty_port_close(&bc->port, ttys, filp);
}

/*
 * return the amount of space in the output buffer
 *
 * This is actually a contract between the driver and the tty layer outlining
 * how much write room the driver can guarantee will be sent OR BUFFERED.  This
 * driver MUST honor the return value.
 */
static int ehv_bc_tty_write_room(struct tty_struct *ttys)
{
	struct ehv_bc_data *bc = ttys->driver_data;
	unsigned int rx_count, tx_count;
	unsigned int ret;

	/* Returns the amount of free space in the TX buffer */
	ret = ev_byte_channel_poll(bc->handle, &rx_count, &tx_count);
	if (ret)
		/* Failure can occur only if the byte channel handle is wrong */
		return -EINVAL;

	return tx_count;
}

/*
 * Stop sending data to the tty layer
 *
 * This function is called when the tty layer's input buffers are getting full,
 * so the driver should stop sending it data.  The easiest way to do this is to
 * disable the RX IRQ, which will prevent ehv_bc_tty_rx_isr() from being
 * called.
 *
 * The hypervisor will continue to queue up any incoming data.  If there is any
 * data in the queue when the RX interrupt is enabled, we'll immediately get an
 * RX interrupt.
 */
static void ehv_bc_tty_throttle(struct tty_struct *ttys)
{
	struct ehv_bc_data *bc = ttys->driver_data;

	disable_irq(bc->rx_irq);
}

/*
 * Resume sending data to the tty layer
 *
 * This function is called after previously calling ehv_bc_tty_throttle().  The
 * tty layer's input buffers now have more room, so the driver can resume
 * sending it data.
 */
static void ehv_bc_tty_unthrottle(struct tty_struct *ttys)
{
	struct ehv_bc_data *bc = ttys->driver_data;

	/* If there is any data in the queue when the RX interrupt is enabled,
	 * we'll immediately get an RX interrupt.
	 */
	enable_irq(bc->rx_irq);
}

/*
 * TTY driver operations
 *
 * If we could ask the hypervisor how much data is still in the TX buffer, or
 * at least how big the TX buffers are, then we could implement the
 * .wait_until_sent and .chars_in_buffer functions.
 */
static const struct tty_operations ehv_bc_ops = {
	.open		= ehv_bc_tty_open,
	.close		= ehv_bc_tty_close,
	.write		= ehv_bc_tty_write,
	.write_room	= ehv_bc_tty_write_room,
	.throttle	= ehv_bc_tty_throttle,
	.unthrottle	= ehv_bc_tty_unthrottle,
};

/*
 * initialize the TTY port
 *
 * This function will only be called once, no matter how many times
 * ehv_bc_tty_open() is called.  That's why we register the ISR here, and also
 * why we initialize tty_struct-related variables here.
 */
static int ehv_bc_tty_port_activate(struct tty_port *port,
				    struct tty_struct *ttys)
{
	struct ehv_bc_data *bc = container_of(port, struct ehv_bc_data, port);
	int ret;

	bc->ttys = ttys;
	ttys->driver_data = bc;

	ret = request_irq(bc->rx_irq, ehv_bc_tty_rx_isr, 0, "ehv-bc", bc);
	if (ret < 0) {
		dev_err(bc->dev, "could not request rx irq %u (ret=%i)\n",
		       bc->rx_irq, ret);
		return ret;
	}

	ret = request_irq(bc->tx_irq, ehv_bc_tty_tx_isr, 0, "ehv-bc", bc);
	if (ret < 0) {
		dev_err(bc->dev, "could not request tx irq %u (ret=%i)\n",
		       bc->tx_irq, ret);
		free_irq(bc->rx_irq, bc);
		return ret;
	}

	/* The TX IRQ is enabled only when we can't write all the data to the
	 * byte channel at once, so by default it's disabled.
	 */
	// disable_irq(bc->tx_irq);

	return 0;
}

static void ehv_bc_tty_port_shutdown(struct tty_port *port)
{
	struct ehv_bc_data *bc = container_of(port, struct ehv_bc_data, port);

	free_irq(bc->tx_irq, bc);
	free_irq(bc->rx_irq, bc);
	bc->ttys = NULL;
}

static const struct tty_port_operations ehv_bc_tty_port_ops = {
	.activate = ehv_bc_tty_port_activate,
	.shutdown = ehv_bc_tty_port_shutdown,
};

static int __devinit ehv_bc_tty_of_probe(struct of_device *of_dev,
					 const struct of_device_id *match)
{
	struct device_node *np = of_dev->node;
	struct ehv_bc_data *bc;
	const uint32_t *reg;
	unsigned int handle;
	int ret;
	static unsigned int index = 1;
	unsigned int i;

	reg = of_get_property(np, "reg", NULL);
	if (!reg) {
		dev_err(&of_dev->dev, "no 'reg' property in %s node\n",
			np->name);
		return -ENODEV;
	}

	/* We already told the console layer that the index for the console
	 * device is zero, so we need to make sure that we use that index when
	 * we probe the console byte channel node.
	 */
	handle = be32_to_cpu(*reg);
	i = (handle == stdout_bc) ? 0 : index++;
	bc = &bcs[i];

	bc->handle = handle;

	bc->rx_irq = irq_of_parse_and_map(np, 0);
	bc->tx_irq = irq_of_parse_and_map(np, 1);
	if ((bc->rx_irq == NO_IRQ) || (bc->tx_irq == NO_IRQ)) {
		dev_err(&of_dev->dev, "no 'interrupts' property in %s node\n",
			np->name);
		ret = -ENODEV;
		goto error;
	}

	bc->dev = tty_register_device(ehv_bc_driver, bc - bcs, &of_dev->dev);
	if (IS_ERR(bc->dev)) {
		ret = PTR_ERR(bc->dev);
		dev_err(&of_dev->dev, "could not register tty (ret=%i)\n", ret);

		goto error;
	}

	tty_port_init(&bc->port);
	bc->port.ops = &ehv_bc_tty_port_ops;

	dev_set_drvdata(&of_dev->dev, bc);

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
	if (index == 2) {
        bc_ttys2 = bc;
		bc_handle_ttyS2 = bc->handle;
		bc->ttys = &ttys2;
	}else if (index == 3) {
        bc_ttys3 = bc;
		bc_handle_ttyS3 = bc->handle;
		bc->ttys = &ttys3;
	}
#endif
		
	dev_info(&of_dev->dev, "registered /dev/%s%u for byte channel %u\n",
		ehv_bc_driver->name, index, bc->handle);

	return 0;

error:
	irq_dispose_mapping(bc->tx_irq);
	irq_dispose_mapping(bc->rx_irq);

	return ret;
}

static int ehv_bc_tty_of_remove(struct of_device *of_dev)
{
	struct ehv_bc_data *bc = dev_get_drvdata(&of_dev->dev);

	tty_unregister_device(ehv_bc_driver, bc - bcs);

	irq_dispose_mapping(bc->tx_irq);
	irq_dispose_mapping(bc->rx_irq);

	return 0;
}

static const struct of_device_id ehv_bc_tty_of_ids[] = {
	{ .compatible = "epapr,hv-byte-channel" },
	{}
};

MODULE_DEVICE_TABLE(of, ehv_bc_tty_of_ids);

static struct of_platform_driver ehv_bc_tty_of_driver = {
	.name		= "ehv-bc",
	.match_table	= ehv_bc_tty_of_ids,
	.probe		= ehv_bc_tty_of_probe,
	.remove		= ehv_bc_tty_of_remove,
};


static void ehv_serial_log_init(void)
{
	ehv_serial_log_buf = (struct serial_cir_buf_t *) kzalloc(sizeof(
			struct serial_cir_buf_t), GFP_KERNEL);

	if (NULL == ehv_serial_log_buf) {
		printk(KERN_ERR "[%s]:LINE %d memory allocation failed \n",
			__FUNCTION__, __LINE__);
		return;
	}

	ehv_serial_log_buf->magic = SERIAL_LOG_MAGIC;
	ehv_serial_log_buf->total_size = SERIAL_LOG_FULL_LEN;
	ehv_serial_log_buf->wrap_point = SERIAL_LOG_WRAP_LEN;
	ehv_serial_log_buf->next_free = 0;
	ehv_serial_log_buf->wraped_once = 0;
	sl_next_free = &(ehv_serial_log_buf->next_free);

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
	ehv_serial_log_buf_ttys2 = (struct serial_cir_buf_t *) kzalloc(sizeof(
			struct serial_cir_buf_t), GFP_KERNEL);

	if (NULL == ehv_serial_log_buf_ttys2) {
		printk(KERN_ERR "[%s]:LINE %d memory allocation failed \n",
			__FUNCTION__, __LINE__);
		return;
	}

	ehv_serial_log_buf_ttys3 = (struct serial_cir_buf_t *) kzalloc(sizeof(
			struct serial_cir_buf_t), GFP_KERNEL);

	if (NULL == ehv_serial_log_buf_ttys3) {
		printk(KERN_ERR "[%s]:LINE %d memory allocation failed \n",
			__FUNCTION__, __LINE__);
		return;
	}

	ehv_serial_log_buf_ttys2->magic = SERIAL_LOG_MAGIC;
	ehv_serial_log_buf_ttys2->total_size = SERIAL_LOG_FULL_LEN;
	ehv_serial_log_buf_ttys2->wrap_point = SERIAL_LOG_WRAP_LEN;
	ehv_serial_log_buf_ttys2->next_free = 0;
	ehv_serial_log_buf_ttys2->wraped_once = 0;
	sl_next_free_ttys2 = &(ehv_serial_log_buf_ttys2->next_free);

	ehv_serial_log_buf_ttys3->magic = SERIAL_LOG_MAGIC;
	ehv_serial_log_buf_ttys3->total_size = SERIAL_LOG_FULL_LEN;
	ehv_serial_log_buf_ttys3->wrap_point = SERIAL_LOG_WRAP_LEN;
	ehv_serial_log_buf_ttys3->next_free = 0;
	ehv_serial_log_buf_ttys3->wraped_once = 0;
	sl_next_free_ttys3 = &(ehv_serial_log_buf_ttys3->next_free);
#endif
}

char *ehv_serial_log_buf_get(void)
{
	return((char *) ehv_serial_log_buf);
}

EXPORT_SYMBOL(ehv_serial_log_buf_get);

#if defined CONFIG_VDX3720_REDIRECT_DPCONSOLES
char *ehv_serial_log_buf_ttys2_get(void)
{
	return((char *) ehv_serial_log_buf_ttys2);
}

EXPORT_SYMBOL(ehv_serial_log_buf_ttys2_get);

char *ehv_serial_log_buf_ttys3_get(void)
{
	return((char *) ehv_serial_log_buf_ttys3);
}
EXPORT_SYMBOL(ehv_serial_log_buf_ttys3_get);
#endif

/**
 * ehv_bc_init - ePAPR hypervisor byte channel driver initialization
 *
 * This function is called when this module is loaded.
 */
static int __init ehv_bc_init(void)
{
	struct device_node *np;
	unsigned int count = 0; /* Number of elements in bcs[] */

	int ret;

	pr_info("ePAPR hypervisor byte channel console driver\n");

	if (!has_fsl_hypervisor()) {
		pr_info("ehv-bc: no hypervisor found\n");
		return -ENODEV;
	}

	/* Count the number of byte channels */
	for_each_compatible_node(np, NULL, "epapr,hv-byte-channel")
		count++;

	if (!count)
		return -ENODEV;

	/* The array index of an element in bcs[] is the same as the tty index
	 * for that element.  If you know the address of an element in the
	 * array, then you can use pointer math (e.g. "bc - bcs") to get its tty
	 * index.
	 */
	bcs = kzalloc(count * sizeof(struct ehv_bc_data), GFP_KERNEL);
	if (!bcs)
		return -ENOMEM;

	ehv_bc_driver = alloc_tty_driver(count);
	if (!ehv_bc_driver) {
		ret = -ENOMEM;
		goto error;
	}

	ehv_bc_driver->owner = THIS_MODULE;
	ehv_bc_driver->driver_name = "ehv-bc";
	ehv_bc_driver->name = ehv_bc_console.name;
	ehv_bc_driver->type = TTY_DRIVER_TYPE_CONSOLE;
	ehv_bc_driver->subtype = SYSTEM_TYPE_CONSOLE;
	ehv_bc_driver->init_termios = tty_std_termios;
	ehv_bc_driver->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_DYNAMIC_DEV;
	tty_set_operations(ehv_bc_driver, &ehv_bc_ops);

	ret = tty_register_driver(ehv_bc_driver);
	if (ret) {
		pr_err("ehv-bc: could not register tty driver (ret=%i)\n", ret);
		goto error;
	}

	ret = of_register_platform_driver(&ehv_bc_tty_of_driver);
	if (ret) {
		pr_err("ehv-bc: could not register platform driver (ret=%i)\n",
		       ret);
		goto error;
	}

	// Create a buffer for each byte channel to store console log
	ehv_serial_log_init();

	return 0;

error:
	if (ehv_bc_driver) {
		tty_unregister_driver(ehv_bc_driver);
		put_tty_driver(ehv_bc_driver);
	}

	kfree(bcs);

	return ret;
}


/**
 * ehv_bc_exit - ePAPR hypervisor byte channel driver termination
 *
 * This function is called when this driver is unloaded.
 */
static void __exit ehv_bc_exit(void)
{
	tty_unregister_driver(ehv_bc_driver);
	put_tty_driver(ehv_bc_driver);
	kfree(bcs);
}

module_init(ehv_bc_init);
module_exit(ehv_bc_exit);

MODULE_AUTHOR("Timur Tabi <timur@freescale.com>");
MODULE_DESCRIPTION("ePAPR hypervisor byte channel console driver");
MODULE_LICENSE("GPL");
