/*
 * Copyright (C) 2003 Brocade Communications
 *
 * driver to emulate 36bit mmap2 operation for CPLD
 *   - allows mapping upto one page of physical addresses
 *   - written for supporting 36bit mapping for PPC 440.
 *
 *     Based on driver from ldd2 samples
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 */
#include <linux/module.h>
#include <linux/kernel.h>   /* printk() */
#include <linux/init.h>
#include <linux/slab.h>	    /* kmalloc() */
#include <linux/fs.h>	    /* everything... */
#include <linux/errno.h>    /* error codes */
#include <linux/types.h>    /* size_t */

#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/tlb.h>

static int cpld_major = 0;
module_param(cpld_major, int, 0);
MODULE_AUTHOR("Kalpesh Jasapara");

/*
 * Forwards for our methods.
 */
static int cpld_open (struct inode *inode, struct file *filp);
static int cpld_close(struct inode *inode, struct file *filp);
static int cpld_mmap(struct file *filp, struct vm_area_struct *vma);
static int cpld_map_page(unsigned long va, phys_addr_t pa, pgprot_t prot);

/*
 * Global defines
 */
#define MAX_SIMPLE_DEV 1

/*
 * various sub-devices.
 */

/* Device 0 uses remap_page_range */
struct file_operations cpld_mmap_ops = {
	.open		= cpld_open,
	.release	= cpld_close,
	.mmap		= cpld_mmap,
	.owner		= THIS_MODULE
};

/*
 * Open the device; all we have to do here is to up the usage count and
 * set the right fops.
 */
static int 
cpld_open (struct inode *inode, struct file *filp)
{
unsigned int dev = MINOR(inode->i_rdev);

	if (dev >= MAX_SIMPLE_DEV) 
		return -ENODEV;
	filp->f_op = &cpld_mmap_ops;
	return 0;
}

/*
 * Closing is even cpldr.
 */
static int 
cpld_close(struct inode *inode, struct file *filp)
{
	return 0;
}

/*
 * Common VMA ops.
 */

static void 
cpld_vma_open(struct vm_area_struct *vma)
{ }

static void 
cpld_vma_close(struct vm_area_struct *vma)
{ }


/*
 * The mmap_page version of mmap.
 */

static struct vm_operations_struct cpld_mmap_vm_ops = {
	open:  cpld_vma_open,
	close: cpld_vma_close,
};

/*
 * Architectures vary in how they handle caching for addresses 
 * outside of main memory.
 */
static inline int noncached_address(unsigned long addr)
{
	return addr >= __pa(high_memory);
}

static int 
cpld_mmap(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long long offset = 
		(((unsigned long long)vma->vm_pgoff) << PAGE_SHIFT); 
	if (offset >= __pa(high_memory) || (filp->f_flags & O_SYNC))
		vma->vm_flags |= VM_SPECIAL;
	vma->vm_flags |= VM_RESERVED;

	if (noncached_address(offset) || (filp->f_flags & O_SYNC))
		vma->vm_page_prot =
			__pgprot(pgprot_val(vma->vm_page_prot) | _PAGE_NO_CACHE | _PAGE_GUARDED);

	/* cpld_mmap - supports only one page in size */ 
	if ( vma->vm_end - vma->vm_start > PAGE_SIZE)
		return  -EINVAL;

	if ( cpld_map_page(vma->vm_start, offset, vma->vm_page_prot) )
		return -EAGAIN;

	/* vm_ops prevents module unloading when mappings exist */ 
	vma->vm_ops = &cpld_mmap_vm_ops;

	/* Initialize the vma - Increment mod use count */
	cpld_vma_open(vma);

	return 0;
}

/*
 * Map a single page of VA to the specified physical address 
 *  -- currently the page size is only the error checking that we
 *  this is an optimized version compared to a full fledged 64 bit
 *  support changes in /dev/mem
 */
static int 
cpld_map_page(unsigned long va, phys_addr_t pa, pgprot_t prot)
{
	pmd_t *pd;
	pte_t *pt;
	struct mm_struct *mm = current->mm;
	spinlock_t *ptl;
	int err = -ENOMEM;

	/* Use upper 10 bits of VA to index the first level map */
	pd = pmd_offset(pud_offset(pgd_offset(mm, va), va), va);
	/* Use middle 10 bits of VA to index the second-level map */
	pt = pte_alloc_map_lock(mm, pd, va, &ptl);
	if (pt != NULL) {
		err = 0;
		set_pte_at(mm, va, pt,  pte_mkspecial(pfn_pte((pa >> PAGE_SHIFT), prot))); 
	}
	pte_unmap_unlock(pt, ptl);
	return err;
}

/*
 * Module housekeeping.
 */
static int __init 
cpld_init(void)
{
int result;


	result = register_chrdev(cpld_major, "mem_page", &cpld_mmap_ops);
	if (result < 0)
	{
		printk(KERN_WARNING "cpld: unable to get major %d\n", 
			cpld_major);
		return result;
	}
	if (cpld_major == 0)
		cpld_major = result;
	return 0;
}


static void __exit
cpld_cleanup(void)
{
	unregister_chrdev(cpld_major, "mem_page");
}


module_init(cpld_init);
module_exit(cpld_cleanup);
