/*
 * An rtc driver for the Dallas DS1742
 *
 * Copyright (C) 2006 Atsushi Nemoto <anemo@mba.ocn.ne.jp>
 * Copyright (C) 2011 Brocade Communications <mlei@brocade.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Copyright (C) 2006 Torsten Ertbjerg Rasmussen <tr@newtec.dk>
 *  - nvram size determined from resource
 *  - this ds1742 driver now supports ds1743.
 */

#include <linux/bcd.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <linux/rtc.h>
#include <linux/io.h>
#include <linux/miscdevice.h>
#include <linux/of.h>
#include <linux/of_platform.h>

#define DRV_VERSION "0.4"

#define RTC_SIZE		8

#define RTC_CONTROL		0
#define RTC_CENTURY		0
#define RTC_SECONDS		1
#define RTC_MINUTES		2
#define RTC_HOURS		3
#define RTC_DAY			4
#define RTC_DATE		5
#define RTC_MONTH		6
#define RTC_YEAR		7

#define RTC_CENTURY_MASK	0x3f
#define RTC_SECONDS_MASK	0x7f
#define RTC_DAY_MASK		0x07

/* Bits in the Control/Century register */
#define RTC_WRITE		0x80
#define RTC_READ		0x40

/* Bits in the Seconds register */
#define RTC_STOP		0x80

/* Bits in the Day register */
#define RTC_BATT_FLAG		0x80

struct rtc_plat_data {
	struct rtc_device *rtc;
	void __iomem *ioaddr_rtc;
	size_t size;
	resource_size_t baseaddr;
	unsigned long last_jiffies;
};

static int ds1742_rtc_set_time(struct device *dev, struct rtc_time *tm)
{
	struct rtc_plat_data *pdata = dev_get_drvdata(dev);
	void __iomem *ioaddr = pdata->ioaddr_rtc;
	u8 century;

	century = bin2bcd((tm->tm_year + 1900) / 100);

	writeb(RTC_WRITE, ioaddr + RTC_CONTROL);

	writeb(bin2bcd(tm->tm_year % 100), ioaddr + RTC_YEAR);
	writeb(bin2bcd(tm->tm_mon + 1), ioaddr + RTC_MONTH);
	writeb(bin2bcd(tm->tm_wday) & RTC_DAY_MASK, ioaddr + RTC_DAY);
	writeb(bin2bcd(tm->tm_mday), ioaddr + RTC_DATE);
	writeb(bin2bcd(tm->tm_hour), ioaddr + RTC_HOURS);
	writeb(bin2bcd(tm->tm_min), ioaddr + RTC_MINUTES);
	writeb(bin2bcd(tm->tm_sec) & RTC_SECONDS_MASK, ioaddr + RTC_SECONDS);

	/* RTC_CENTURY and RTC_CONTROL share same register */
	writeb(RTC_WRITE | (century & RTC_CENTURY_MASK), ioaddr + RTC_CENTURY);
	writeb(century & RTC_CENTURY_MASK, ioaddr + RTC_CONTROL);
	return 0;
}

static int ds1742_rtc_read_time(struct device *dev, struct rtc_time *tm)
{
	struct rtc_plat_data *pdata = dev_get_drvdata(dev);
	void __iomem *ioaddr = pdata->ioaddr_rtc;
	unsigned int year, month, day, hour, minute, second, week;
	unsigned int century;

	/* give enough time to update RTC in case of continuous read */
	if (pdata->last_jiffies == jiffies)
		msleep(1);
	pdata->last_jiffies = jiffies;
	writeb(RTC_READ, ioaddr + RTC_CONTROL);
	second = readb(ioaddr + RTC_SECONDS) & RTC_SECONDS_MASK;
	minute = readb(ioaddr + RTC_MINUTES);
	hour = readb(ioaddr + RTC_HOURS);
	day = readb(ioaddr + RTC_DATE);
	week = readb(ioaddr + RTC_DAY) & RTC_DAY_MASK;
	month = readb(ioaddr + RTC_MONTH);
	year = readb(ioaddr + RTC_YEAR);
	century = readb(ioaddr + RTC_CENTURY) & RTC_CENTURY_MASK;
	writeb(0, ioaddr + RTC_CONTROL);
	tm->tm_sec = bcd2bin(second);
	tm->tm_min = bcd2bin(minute);
	tm->tm_hour = bcd2bin(hour);
	tm->tm_mday = bcd2bin(day);
	tm->tm_wday = bcd2bin(week);
	tm->tm_mon = bcd2bin(month) - 1;
	/* year is 1900 + tm->tm_year */
	tm->tm_year = bcd2bin(year) + bcd2bin(century) * 100 - 1900;

	if (rtc_valid_tm(tm) < 0) {
		dev_err(dev, "retrieved date/time is not valid.\n");
		rtc_time_to_tm(0, tm);
	}
	return 0;
}

static const struct rtc_class_ops ds1742_rtc_ops = {
	.read_time	= ds1742_rtc_read_time,
	.set_time	= ds1742_rtc_set_time,
};

static struct rtc_device *ds1742rtc;

static int ds1742_rtc_open(struct inode *inode, struct file *file)
{
	int err;
	const struct rtc_class_ops *ops = ds1742rtc->ops;

	if (test_and_set_bit_lock(RTC_DEV_BUSY, &ds1742rtc->flags))
		return -EBUSY;

	file->private_data = ds1742rtc;

	err = ops->open ? ops->open(ds1742rtc->dev.parent) : 0;
	/* something has gone wrong */
	clear_bit_unlock(RTC_DEV_BUSY, &ds1742rtc->flags);
	return err;
}

static ssize_t
ds1742_rtc_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	unsigned long data;
	ssize_t ret;

	data = 0;
	if (sizeof(int) != sizeof(long) &&
	    count == sizeof(unsigned int))
		ret = put_user(data, (unsigned int __user *)buf) ?:
			sizeof(unsigned int);
	else
		ret = put_user(data, (unsigned long __user *)buf) ?:
			sizeof(unsigned long);
	return ret;
}

static long ds1742_rtc_ioctl(struct file *file,
		unsigned int cmd, unsigned long arg)
{
	int err = 0;
	struct rtc_device *rtc = file->private_data;
	const struct rtc_class_ops *ops = rtc->ops;
	struct rtc_time tm;
	void __user *uarg = (void __user *) arg;

	err = mutex_lock_interruptible(&rtc->ops_lock);
	if (err)
		return err;

	/* check that the calling task has appropriate permissions
	 * for certain ioctls. doing this check here is useful
	 * to avoid duplicate code in each driver.
	 */
	switch (cmd) {
	case RTC_SET_TIME:
		if (!capable(CAP_SYS_TIME))
			err = -EACCES;
		break;
	}

	if (err)
		goto done;

	/* try the driver's ioctl interface */
	if (ops->ioctl) {
		err = ops->ioctl(rtc->dev.parent, cmd, arg);
		if (err != -ENOIOCTLCMD) {
			mutex_unlock(&rtc->ops_lock);
			return err;
		}
	}

	/* if the driver does not provide the ioctl interface
	 * or if that particular ioctl was not implemented
	 * (-ENOIOCTLCMD), we will try to emulate here.
	 *
	 * Drivers *SHOULD NOT* provide ioctl implementations
	 * for these requests.  Instead, provide methods to
	 * support the following code, so that the RTC's main
	 * features are accessible without using ioctls.
	 *
	 * RTC and alarm times will be in UTC, by preference,
	 * but dual-booting with MS-Windows implies RTCs must
	 * use the local wall clock time.
	 */

	switch (cmd) {
	case RTC_RD_TIME:
		mutex_unlock(&rtc->ops_lock);

		err = rtc_read_time(rtc, &tm);
		if (err < 0)
			return err;

		if (copy_to_user(uarg, &tm, sizeof(tm)))
			err = -EFAULT;
		return err;

	case RTC_SET_TIME:
		mutex_unlock(&rtc->ops_lock);

		if (copy_from_user(&tm, uarg, sizeof(tm)))
			return -EFAULT;

		return rtc_set_time(rtc, &tm);

	case RTC_UIE_ON:
		mutex_unlock(&rtc->ops_lock);
		return (0);

	case RTC_UIE_OFF:
		mutex_unlock(&rtc->ops_lock);
		return (0);

	default:
		err = -EINVAL;
		break;
	}

done:
	mutex_unlock(&rtc->ops_lock);
	return err;
}

static int ds1742_rtc_release(struct inode *inode, struct file *file)
{
	struct rtc_device *rtc = file->private_data;

	/* We shut down the repeating IRQs that userspace enabled,
	 * since nothing is listening to them.
	 *  - Update (UIE) ... currently only managed through ioctls
	 *  - Periodic (PIE) ... also used through rtc_*() interface calls
	 *
	 * Leave the alarm alone; it may be set to trigger a system wakeup
	 * later, or be used by kernel code, and is a one-shot event anyway.
	 */
	if (rtc->ops->release)
		rtc->ops->release(rtc->dev.parent);

	clear_bit_unlock(RTC_DEV_BUSY, &rtc->flags);
	return 0;
}

static const struct file_operations ds1742_rtc_fops = {
	.owner		= THIS_MODULE,
	.read		= ds1742_rtc_read,
	.unlocked_ioctl	= ds1742_rtc_ioctl,
	.open		= ds1742_rtc_open,
	.release	= ds1742_rtc_release,
};

static struct miscdevice ds1742_rtc_dev =
{
	RTC_MINOR,
	"rtc",
	&ds1742_rtc_fops
};

static int __devinit ds1742_rtc_probe(struct of_device *dev,
					const struct of_device_id *match)
{
	struct device_node *dp = dev->node;
	struct resource res;
	struct rtc_plat_data *pdata = NULL;
	unsigned int cen, sec;
	void __iomem *ioaddr = NULL;
	int ret = 0;
	const char* cp;

	if (of_address_to_resource(dp, 0, &res)) {
		dev_err(&dev->dev, "Can't get IO address from device tree\n");
		return -ENODEV;
	}

	pdata = kzalloc(sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return -ENOMEM;

	dev_set_drvdata(&dev->dev, pdata);

	pdata->size = res.end - res.start + 1;
	pdata->baseaddr = res.start;
	printk(KERN_INFO "ds1742_rtc_probe 0x%llx\n", pdata->baseaddr);

	ioaddr = ioremap(pdata->baseaddr, pdata->size);
	if (!ioaddr) {
		ret = -ENOMEM;
		goto out;
	}
	pdata->ioaddr_rtc = ioaddr;

	cp = of_get_property(dp, "fsl,hv-claimable", NULL);
	if (!cp || strncmp(cp, "active", 6) == 0) {

		printk("rtc active\n");

		/* turn RTC on if it was not on */
		sec = readb(ioaddr + RTC_SECONDS);
		if (sec & RTC_STOP) {
			sec &= RTC_SECONDS_MASK;
			cen = readb(ioaddr + RTC_CENTURY) & RTC_CENTURY_MASK;
			writeb(RTC_WRITE, ioaddr + RTC_CONTROL);
			writeb(sec, ioaddr + RTC_SECONDS);
			writeb(cen & RTC_CENTURY_MASK, ioaddr + RTC_CONTROL);
		}
		if (!(readb(ioaddr + RTC_DAY) & RTC_BATT_FLAG))
			dev_warn(&dev->dev, "voltage-low detected.\n");
	}

	ds1742rtc = kzalloc(sizeof(struct rtc_device), GFP_KERNEL);
	if (ds1742rtc == NULL) {
		ret = -ENOMEM;
		goto out;
	}

	ds1742rtc->ops = &ds1742_rtc_ops;
	ds1742rtc->owner = THIS_MODULE;
	ds1742rtc->dev.parent = &dev->dev;

	mutex_init(&ds1742rtc->ops_lock);

	strlcpy(ds1742rtc->name, dev_name(&dev->dev), RTC_DEVICE_NAME_SIZE);
	dev_set_name(&ds1742rtc->dev, "rtc0");

	pdata->rtc = ds1742rtc;
	pdata->last_jiffies = jiffies;

//	ds1742nv[0].base = pdata->ioaddr_nvram + 0x0800;
//	ds1742nv[1].base = pdata->ioaddr_nvram + 0x0008;

	ret = misc_register(&ds1742_rtc_dev);
	if (ret) {
		printk(KERN_ERR "rtc: can't misc_register on minor=%d\n",
		    RTC_MINOR);
		goto out;
	}

	printk(KERN_INFO "Brocade NVRAM/RTC driver v" DRV_VERSION "\n");
	return 0;
 out:
	if (pdata->ioaddr_rtc)
		iounmap(pdata->ioaddr_rtc);
	if (pdata) kfree(pdata);
	if (ds1742rtc) kfree(ds1742rtc);
	return ret;
}

static int ds1742_rtc_remove(struct of_device *dev)
{
	struct rtc_plat_data *pdata = dev_get_drvdata(&dev->dev);
	if (!pdata)
		return 0;
	dev_set_drvdata(&dev->dev, NULL);

	misc_deregister(&ds1742_rtc_dev);

	iounmap(pdata->ioaddr_rtc);
	kfree(pdata);
	kfree(ds1742rtc);
	return 0;
}

static struct of_device_id ds1742_rtc_match[] =
{
	{
		.type = "rtc",
		.compatible = "brocade,ds1742",
	},
	{},
};
static struct of_platform_driver ds1742_rtc_driver = {
	.name	= "ds1742-rtc",
	.match_table = ds1742_rtc_match,
	.probe		= ds1742_rtc_probe,
	.remove		= ds1742_rtc_remove,
};

/*
 * NVRAM code section
 */
extern spinlock_t rtc_lock;
static DEFINE_SPINLOCK(nvram_state_lock);
#define NVRAM_WRITE     1 /* opened for writing (exclusive) */
#define NVRAM_EXCL      2 /* opened with O_EXCL */

#define NVRAM_BYTES     (0x7f0)

static char contents[NVRAM_BYTES];

struct nvram_dev {
	int id;
	resource_size_t paddr;	
    void __iomem *base;
    size_t size;
	int open_cnt;
	int open_mode;
};

static int num_nvs = 0;
static struct nvram_dev ds1742nv[8] = {
	{ 0, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 1, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 2, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 3, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 4, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 5, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 6, 0, NULL, NVRAM_BYTES, 0, 0 },
	{ 7, 0, NULL, NVRAM_BYTES, 0, 0 },
};

static int ds1742_nvram_open(struct inode *inode, struct file *file)
{
	int minor = iminor(inode) - NVRAM0_MINOR;
	struct nvram_dev *nv = &ds1742nv[minor];

	if (nv->base == NULL) return -ENODEV;

	lock_kernel();
	spin_lock(&nvram_state_lock);

	if ((nv->open_cnt && (file->f_flags & O_EXCL)) ||
	    (nv->open_mode & NVRAM_EXCL) ||
	    ((file->f_mode & FMODE_WRITE) && (nv->open_mode & NVRAM_WRITE))) {
		spin_unlock(&nvram_state_lock);
		unlock_kernel();
		return -EBUSY;
	}

	if (file->f_flags & O_EXCL)
		nv->open_mode |= NVRAM_EXCL;
	if (file->f_mode & FMODE_WRITE)
		nv->open_mode |= NVRAM_WRITE;
	nv->open_cnt++;

	file->private_data = nv;

	spin_unlock(&nvram_state_lock);
	unlock_kernel();

	return 0;
}

static int ds1742_nvram_release(struct inode *inode, struct file *file)
{
	struct nvram_dev *nv = file->private_data;

	spin_lock(&nvram_state_lock);

	nv->open_cnt--;

	if (nv->open_cnt == 0) file->private_data = NULL;

	/* if only one instance is open, clear the EXCL bit */
	if (nv->open_mode & NVRAM_EXCL)
		nv->open_mode &= ~NVRAM_EXCL;
	if (file->f_mode & FMODE_WRITE)
		nv->open_mode &= ~NVRAM_WRITE;

	spin_unlock(&nvram_state_lock);

	return 0;
}

static loff_t ds1742_nvram_llseek(struct file *file, loff_t offset, int origin)
{
	lock_kernel();
	switch (origin) {
	case 0:
		/* nothing to do */
		break;
	case 1:
		offset += file->f_pos;
		break;
	case 2:
		offset += NVRAM_BYTES;
		break;
	}
	unlock_kernel();
	return (offset >= 0) ? (file->f_pos = offset) : -EINVAL;
}

static ssize_t ds1742_nvram_read(struct file *file, char __user *buf,
						size_t count, loff_t *ppos)
{
	struct nvram_dev *nv = file->private_data;
	void __iomem *ioaddr = nv->base;
	unsigned int pos = *ppos;
	char *tmp;

	spin_lock_irq(&rtc_lock);

	for (tmp = contents; count-- > 0 && pos < nv->size; ++pos, ++tmp)
		*tmp = readb(ioaddr + pos);

	spin_unlock_irq(&rtc_lock);

	if (copy_to_user(buf, contents, tmp - contents))
		return -EFAULT;

	*ppos = pos;

	return tmp - contents;
}

static ssize_t ds1742_nvram_write(struct file *file, const char __user *buf,
						size_t count, loff_t *ppos)
{
	struct nvram_dev *nv = file->private_data;
	void __iomem *ioaddr = nv->base;
	unsigned int pos = *ppos;
	char *tmp;
	int len;

	len = (nv->size - pos) < count ? (nv->size - pos) : count;
	if (copy_from_user(contents, buf, len))
		return -EFAULT;

	spin_lock_irq(&rtc_lock);

	for (tmp = contents; count-- > 0 && pos < nv->size; ++pos, ++tmp)
		writeb(*tmp, ioaddr + pos);

	spin_unlock_irq(&rtc_lock);

	*ppos = pos;

	return tmp - contents;
}

static const struct file_operations ds1742_nvram_fops = {
	.owner		= THIS_MODULE,
	.llseek		= ds1742_nvram_llseek,
	.read		= ds1742_nvram_read,
	.write		= ds1742_nvram_write,
	.open		= ds1742_nvram_open,
	.release	= ds1742_nvram_release,
};

static struct miscdevice ds1742_nvram_dev[8] = {
	{ NVRAM0_MINOR, "nvram0", &ds1742_nvram_fops },
	{ NVRAM1_MINOR, "nvram1", &ds1742_nvram_fops },
	{ NVRAM2_MINOR, "nvram2", &ds1742_nvram_fops },
	{ NVRAM3_MINOR, "nvram3", &ds1742_nvram_fops },
	{ NVRAM4_MINOR, "nvram4", &ds1742_nvram_fops },
	{ NVRAM5_MINOR, "nvram5", &ds1742_nvram_fops },
	{ NVRAM6_MINOR, "nvram6", &ds1742_nvram_fops },
	{ NVRAM7_MINOR, "nvram7", &ds1742_nvram_fops },
};

static int __devinit ds1742_nv_probe(struct of_device *dev,
					const struct of_device_id *match)
{
	struct device_node *dp = dev->node;
	struct resource res;
	size_t size;
	void __iomem *ioaddr = NULL;
	int ret = 0;

	if (of_address_to_resource(dp, 0, &res)) {
		dev_err(&dev->dev, "Can't get IO address from device tree\n");
		return -ENODEV;
	}

	size = res.end - res.start + 1;
	if (!request_mem_region(res.start, size, dev_name(&dev->dev))) {
		return -EBUSY;
	}

	printk(KERN_INFO "ds1742_nv_probe %d->0x%llx 0x%x\n", num_nvs, res.start, size);

	ioaddr = ioremap(res.start, size);
	if (!ioaddr) {
		ret = -ENOMEM;
		goto out;
	}

	ret = misc_register(&ds1742_nvram_dev[num_nvs]);
	if (ret) {
		printk(KERN_ERR "nvram: can't misc_register on minor=%d\n",
		    NVRAM0_MINOR);
		goto out;
	}

	dev_set_drvdata(&dev->dev, (void *)num_nvs);

	ds1742nv[num_nvs].paddr = res.start;
	ds1742nv[num_nvs].base = ioaddr;
	ds1742nv[num_nvs++].size = size;
	printk(KERN_INFO "NVRAM%d created\n", num_nvs-1);
	return 0;
 out:
	if (ioaddr)
		iounmap(ioaddr);
	release_mem_region(res.start, size);
	return ret;
}

static int ds1742_nv_remove(struct of_device *dev)
{
	int id = (int)dev_get_drvdata(&dev->dev);

	dev_set_drvdata(&dev->dev, NULL);
	misc_deregister(&ds1742_nvram_dev[id]);

	iounmap(ds1742nv[id].base);
	release_mem_region(ds1742nv[id].paddr, ds1742nv[id].size);
	ds1742nv[id].base = NULL;
	return 0;
}

static struct of_device_id ds1742_nv_match[] =
{
	{
		.type = "nvram",
		.compatible = "brocade,ds1742",
	},
	{},
};
static struct of_platform_driver ds1742_nv_driver = {
	.name	= "ds1742-nvram",
	.match_table = ds1742_nv_match,
	.probe		= ds1742_nv_probe,
	.remove		= ds1742_nv_remove,
};

static __init int ds1742_of_init(void)
{
	of_register_platform_driver(&ds1742_rtc_driver);
	return of_register_platform_driver(&ds1742_nv_driver);
}

static __exit void ds1742_of_exit(void)
{
	of_unregister_platform_driver(&ds1742_rtc_driver);
	of_unregister_platform_driver(&ds1742_nv_driver);
}

module_init(ds1742_of_init);
module_exit(ds1742_of_exit);

MODULE_DESCRIPTION("Brocade DS1742 RTC driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:rtc-ds1742");
MODULE_ALIAS_MISCDEV(RTC_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM0_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM1_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM2_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM3_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM4_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM5_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM6_MINOR);
MODULE_ALIAS_MISCDEV(NVRAM7_MINOR);
