/*
 *    Copyright (c) 1996-2013 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      Intel/DEC 21554 non-transparent PCI-to-PCI bridge (aka Drawbridge)
 *      device driver.
 *
 **		major overhaul to match 2.6.30 - Ming Lei
 */
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <asm/processor.h>

#include <brocade/cpld.h>
#include <brocade/swbd.h>
#include <brocade/silkworm_sin.h>	/* for sin_get_swbd() */

#include "../include/db.h"

/* Preprocessor Defines */

/* Function Macros */

#ifdef DEBUG
#define DBGPRINT(x...) printk(x)
#else
#define DBGPRINT(x...)
#endif

static int  debugAccessors = 0;


/* PCI Configuration Access Command Types */

#define	PCI_TYPE0_CONFIG		0x0
#define	PCI_TYPE1_CONFIG		0x1

#define	PCI_TYPE0_CONFIG_MASK		0x000007FF
#define	PCI_TYPE1_CONFIG_MASK		0x0FFFFFFF

#define	SLOT_SHIFT			11

#define	DB_CONFIG(bus, devfn, reg) \
	 ((((bus)   & 0xFF) << 16) | \
	 (((devfn) & 0xFF) <<  8) | \
	 ((reg)   & 0xFC))

#define	CFGTYPE0(bus, devfn, reg) \
	    ((DB_CONFIG(bus, devfn, reg) & PCI_TYPE0_CONFIG_MASK) | \
	     PCI_TYPE0_CONFIG | (1 << (PCI_SLOT(devfn) + SLOT_SHIFT)))

#define	CFGTYPE1(bus, devfn, reg) \
	    ((DB_CONFIG(bus, devfn, reg) & PCI_TYPE1_CONFIG_MASK) | \
	     PCI_TYPE1_CONFIG)

#define DRAWBRIDGE_CONFIGURED(ptr) \
	((ptr)&& \
	((ptr)->usable))

#define DRAWBRIDGE_NOT_CONFIGURED(dp) (!DRAWBRIDGE_CONFIGURED(dp))

/* Type definitions */

enum {
    DB_NORECONFIG = 0,
    DB_RECONFIG = 1
};

/* These devices are used in both Saturn and Neptune */

/* PCI memory BAR, 32 bit, non-prefetchable */
#define DB_MEM32_NONPREFETCH	0x0

db_known_dev_t common_known_devs[] = {
	/* Freescale 8548 */
	{
		vendor: 0x1057,
		device: 0x0012,
		bar_sz: {0x100000, 0x400000, 0x0, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, DB_MEM32_NONPREFETCH,
					  0x0, 0x0, 0x0, 0x0}
	},
	{
		vendor: 0x1057,
		device: 0x0013,
		bar_sz: {0x100000, 0x2000000, 0x800000, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, DB_MEM32_NONPREFETCH,
					  0x0, 0x0, 0x0, 0x0}
	},
	/* Condor */
	{
		vendor: 0x1657,
		device: 0x000a,
		bar_sz: {0x1000000, 0x0, 0x100, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, PCI_BASE_ADDRESS_SPACE_IO,
					0x0, 0x0, 0x0}
	},
	/* Anzio */
	{
		vendor: 0x1657,
		device: 0x000f,
		bar_sz: {0x1000000, 0x0, 0x0, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, 0x0, 0x0, 0x0, 0x0}
	},
	/* Broadcom 1250 */
	{
		vendor: 0x166d,
		device: 0x0001,
		bar_sz: {0x1000000, 0x0, 0x1000, 0x1000, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, DB_MEM32_NONPREFETCH,
					DB_MEM32_NONPREFETCH, 0x0, 0x0}
	},
	/* Condor2 */
	{
		vendor: 0x1657,
		device: 0x0010,
		bar_sz: {0x1000000, 0x0, 0x100, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, PCI_BASE_ADDRESS_SPACE_IO,
					0x0, 0x0, 0x0}
	},
	/* BCM */
	{
		vendor: 0x14e4,
		device: 0xb312,
		bar_sz: {0x1000000, 0x0, 0x0, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, 0, 0x0, 0x0, 0x0}
	},
#if 1
	/* PCIe switch */
	{
		vendor: 0x10B5,
		device: 0x8604,
		bar_sz: {0x10000, 0x20000, 0x10000, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, DB_MEM32_NONPREFETCH,
				   0x0, 0x0, 0x0, 0x0}
	},
	/* PCIe switch */
	{
		vendor: 0x10B5,
		device: 0x8606,
		bar_sz: {0x10000, 0x20000, 0x10000, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, DB_MEM32_NONPREFETCH,
				   0x0, 0x0, 0x0, 0x0}
	},
#endif
	/* Condor3 */
	{
		vendor: 0x1657,
		device: 0x001d,
		bar_sz: {0x1000000, 0x0, 0x100, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, PCI_BASE_ADDRESS_SPACE_IO,
				0x0, 0x0, 0x0}
	},

	/* Dummy Anvil */
	{
		vendor: 0x1657,
		device: 0x001a,
		bar_sz: {0x100000, 0x0, 0x0, 0x0, 0x0, 0x0},
		bar_attr: {DB_MEM32_NONPREFETCH, 0x0, 0x0, 0x0, 0x0, 0x0}
	},

	/* Last entry marker */
	{
		vendor: 0x0,
		device: 0x0,
		bar_sz: {0x0, 0x0, 0x0, 0x0, 0x0, 0x0},
		bar_attr: {0x0, 0x0, 0x0, 0x0, 0x0, 0x0}
	}
};


char db_name[] = "Drawbridge 2155x";

int db_count_other = 0, db_count_21554 = 0, db_fixup_needed = 0;



/*
 * Intel acquired the Digital PCI-to-PCI bridge intellectual property. As
 * such the PCI vendor ID may* change from DEC to Intel at some near or
 * distant date in the future.
 */
static struct pci_device_id db_id_table[] = {
	 { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_21554,
	   PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	 { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_DEC_21554,
	   PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	 { PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_21555,
	   PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	 { 0, }
};

/* Function Prototypes */

extern void kick_watchdog(void);

/*
 * Device instantiation
 */

static int	db_probe(pci_dev_t *dp, const struct pci_device_id *id);

/*
 * PCI configuration access functions for devices behind the bridge.
 */
static int	 db_read_byte(
				struct pci_bus *dp, int devfn, int reg, uint8_t *value);
static int	 db_read_word(
				struct pci_bus *dp, int devfn, int reg, uint16_t *value);
static int	 db_read_dword(
				struct pci_bus *dp, int devfn, int reg, uint32_t *value);
static int	 db_write_byte(
				struct pci_bus *dp, int devfn, int reg, uint8_t value);
static int	 db_write_word(
				struct pci_bus *dp, int devfn, int reg, uint16_t value);
static int	 db_write_dword(
				struct pci_bus *dp, int devfn, int reg, uint32_t value);

int	db_init(struct pci_dev *dp, const struct pci_device_id *id);
struct	pci_bus * db_bus_exists(struct pci_bus *bus, int nr);
static	int db_scan_bus(pci_dev_t *dp, struct pci_ops *ops,
				void *sysdata, unsigned int *maxbus);
static void db_read_irq(struct pci_dev *dp);
static int db_scan_child_bus(
				struct pci_bus *bus, unsigned *maxbus, int reconfig,
				db_usr_if_t *usr);

static int db_scan_slot(
				struct pci_bus *bus, int devfn,
				struct pci_dev **newdev,
				int reconfig, db_usr_if_t *usr);
static int db_scan_bridge(
				struct pci_bus *bus, struct pci_dev *dp,
				unsigned int *maxbus, int reconfig, int is_blade_root_bus,
				int root_bus_number, db_usr_if_t *usr);
static struct pci_bus * db_add_new_bus(struct pci_bus *parent,
				       struct pci_dev *dp, int busnr);
int db_device_exists(struct pci_dev *new_dp, unsigned char bus);
static void db_set_master(struct pci_dev *dp);

static unsigned int db_calc_resource_flags(unsigned int flags);
static int db_scan_blade_cmn(struct pci_dev *drawbridge, struct pci_dev **dev,
						int devid, int reconfig, db_usr_if_t *usr);

static int db_primary_reset_uninit(void);

static int db_alloc_small_bar(struct pci_dev *dp, db_t *dbp, int rel_blade,
						struct resource *res, u32 sz, u32 *base);

static struct pci_driver db_driver = {
	 name:		db_name,
	 id_table:	db_id_table,
	 probe:		db_probe,
	 remove:	db_remove_drawbridge,
	 suspend:	NULL,
	 resume:	NULL
};

extern db_platform_cfg_t *platform_cfg_tbl[];
extern int db_cfg_tbl_sz;


static int
db_read_config(struct pci_bus *bus, unsigned int devfn, int where,
			   int size, u32 *val)
{
	int err = -1;
	u8 v1;
	u16 v2;
	u32 v4;

	switch (size) {
	case 1:
		err = db_read_byte(bus, devfn, where, &v1);
		*val = v1;
		break;

	case 2:
		err = db_read_word(bus, devfn, where, &v2);
		*val = v2;
		break;

	case 4:
		err = db_read_dword(bus, devfn, where, &v4);
		*val = v4;
		break;
	}

	return (err);
}

static int
db_write_config(struct pci_bus *bus, unsigned int devfn, int where,
				int size, u32 val)
{
	int err = -1;

	switch (size) {
	case 1:
		err = db_write_byte(bus, devfn, where, val);
		break;

	case 2:
		err = db_write_word(bus, devfn, where, val);
		break;

	case 4:
		err = db_write_dword(bus, devfn, where, val);
		break;
	}

	return (err);
}

static struct pci_ops db_ops = {
	 db_read_config,
	 db_write_config
};

/* Pointer to current platform configuration instance.  */
db_platform_cfg_t *cur_cfg_ptr = NULL;

/*
 * db_pci_size()
 *
 * Description:
 *
 *  Computes the amount of address space from the
 *  contents of a BAR.
 *
 * Input(s):
 *   base: The contents of the BAR
 *   mask: Mask indicating which bits of the BAR
 *	   to use in determining the size
 *
 * Output(s):
 *
 * Returns:
 *   The amount of address space requested by the BAR
 *
 */

static u32 db_pci_size(u32 base, unsigned long mask)
{
	u32 size = mask & base;		/* Find the significant bits */

	/* Get the lowest of them to find the decode size */

	size = size & ~(size-1);

	return (size-1);		/* extent = size - 1 */
}

static db_find_t	db_find_db;

void
db_register_find_db(db_find_t func)
{
	db_find_db = func;
}

/*
 *  struct pci_dev *db_get_device()
 *
 * Description:
 *
 *  Get the data structure for a specified PCI drawbridge
 *  on a specified bus. The global PCI device list is
 *  searched for the data structure.
 *
 * Input(s):
 *  devnum - The PCI device or PCI slot number
 *  bus    - The number of the bus connected to the
 *           primary side of the specified PCI device.
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  NULL - No such PCI device in the global PCI device list
 *  NON-NULL - A pointer to the desired data structure.
 *
 */
struct pci_dev *
db_get_device(int devnum, int bus)
{
	struct pci_dev *dp;
	uint32_t  devfn;

	if (db_find_db) {
		db_find_db(bus, devnum, &dp);
		if (dp != NULL)
			return (dp);
	}
	devfn = PCI_DEVFN(devnum, 0); /* slot, func */
	dp = pci_get_domain_bus_and_slot(cur_cfg_ptr->pci_domain, bus, devfn);
	if (dp) {
		pci_dev_put(dp); /* Keep reference count sane */
	}
	return (dp);
}

/*
 * Description
 *
 *  Map in the CSR space of a drawbridge to make it
 *  accessable.
 *
 *  Input(s):
 *   **csr  - Handle to return the the mapped memory address.
 *            All CSR registers are accessed as an offset from
 *            this address
 *   base_addr - The physical memory start address of the
 *               CSR space.
 *
 *  Output(s):
 *   **csr  - Contains the base address of the virtual through
 *            Through wich the CSRs can be accessed.
 *
 *  Return(s):
 *   0 for success, < 0 for error
 */

/*
 * The printk routine won't print a 36-bit number
 * even if we use 'll' in the format specifier.
 * So break it up into 2 numbers for display
 * purposes.
 */

#define first_wd(x)	(((long long)x)>> (sizeof (int) * 8))
#define second_wd(x)	((x) & (int)(~0))

/*
 *  Description:
 *
 *   Get a pointer to the CSR space of a specified
 *   non-transparent bridge. The CSR space is assumed
 *   to already be mapped.
 *
 * Input(s):
 *   bridge_idx - Indicates which local drawbridge (1 or 2) to use
 *   local  - True = Asking for local CSR space for this
 *            bridge. False = Asking for the CSR  space of the
 *            remote (on the other CP) drawbridge connected to the
 *            secondary side of this bridge.
 *
 *  NOTE: Remote CSR is no longer supported. To avoid an interface
 *        change, the local CSR space is always returned.
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  A pointer to the 1st location of the desired CSR space,
 *  or NULL in the event of an error.
 *
 */
db_csr_t*
db_get_csrbase(int bridge_idx, int local)
{
   struct pci_dev *dp;
   db_t   *dbp;
   db_csr_t  *csr;
   int db_devnum, primary_bus;

   db_devnum = db_idx_to_devnum(bridge_idx);
   if (db_devnum < 0) {
	   printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
			  __FUNCTION__, bridge_idx);
	   return (NULL);
   }
   primary_bus = cur_cfg_ptr->primary_busses[bridge_idx];

   dp = db_get_device(db_devnum, primary_bus);
   if (!dp) {
	   printk(KERN_ERR "\nSpecified PCI device not found in PCI list");
	   return (NULL);
   }

    dbp = (db_t *)pci_get_drvdata(dp);
    if (DRAWBRIDGE_NOT_CONFIGURED(dbp)) {
	 printk(KERN_ERR "\nSpecified PCI device has not been configured");
	 return (NULL);
    }

    csr = dbp->db_local_csr;

    return (csr);

}
/*
 * int db_set_cache_line_size()
 *
 * Description:
 *	Set up cache line size for both primary and secondary CLS register.
 *
 * Input(s):
 *    bridge - Index of the drawbridge (1..n)
 *    size   - The cache size going to be set. Eligible value is:
 *		       4, 8, 16, 32.
 *
 *  Output(s):
 *   None
 *
 *  Returns:
 *   0: success, < 0: fail
 */

int
db_set_cache_line_size(int bridge_idx, u8 size)
{
	struct pci_dev *dp;
	db_t *dbp;
	u8	cls = 0;
	int db_devnum, primary_bus;


	db_devnum = db_idx_to_devnum(bridge_idx);
	if (db_devnum < 0) {
		printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
			   __FUNCTION__, bridge_idx);
		return (-1);
	}
	primary_bus = cur_cfg_ptr->primary_busses[bridge_idx];

	dp = db_get_device(db_devnum, primary_bus);
	if (!dp) {
		printk(KERN_ERR "%s: Bridge %d not found in PCI device list\n",
			   __FUNCTION__, bridge_idx);
		return (-1);
	}

	dbp = (db_t *)pci_get_drvdata(dp);
	if (DRAWBRIDGE_NOT_CONFIGURED(dbp))
		return (-1);

	pci_read_config_byte(dp, PCI_CACHE_LINE_SIZE, &cls);
	DBGPRINT("local cache line size = %d\n", cls);
	cls = size;
	pci_write_config_byte(dp, PCI_CACHE_LINE_SIZE, cls);
	pci_write_config_byte(dp, DB_PCI_SCD_CACHE_LINE_SIZE, cls);

	dbp->cfg_save.cache_line = cls;
	dbp->cfg_save.secondary_cache_line = cls;

	return (0);
}
/*
 *  unsigned int db_get_irq()
 *
 * Description:
 *
 *  Get the value of the PCI IRQ register for a drawbridge device.
 *
 * Input(s):
 *  bridge  - Specifies the index of the desired PCI drawbridge (1..n).
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  0 - If the IRQ register cannot be read.
 *  NON-0 - The IRQ register contents.
 *
 */
uint32_t
db_get_irq(int bridge_idx)
{
  struct pci_dev *db;
  db_t *dbp;
  int db_devnum, primary_bus;

  db_devnum = db_idx_to_devnum(bridge_idx);
  if (db_devnum < 0) {
	  printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
			 __FUNCTION__, bridge_idx);
	  return (0);
  }
  primary_bus = cur_cfg_ptr->primary_busses[bridge_idx];

  db = db_get_device(bridge_idx, primary_bus);

  if (!db) {
    printk(KERN_ERR "Bridge %d not found in PCI device list\n", bridge_idx);
    return (0);
  }

  dbp = (db_t *)pci_get_drvdata(db);
  if (DRAWBRIDGE_NOT_CONFIGURED(dbp)) {
    printk(KERN_ERR "Bridge %d has not been configured\n", bridge_idx);
    return (0);
  }

  return ((unsigned int)db->irq);

}
/*
 * int db_get_forwarding_window()
 *
 * Description:
 *   Get the beginning and end address of ASIC PCI address space
 *   provided by a specified drawbridge. All accesses to this space
 *   will be forwarded by the drawbridge to the switch blade ASICs.
 *
 * Input(s):
 *   bridge_idx - Index of the specified drawbridge, 1..n.
 *   rsrc   - Pointer to a resource structure to be filled
 *            in with the start and end addresses of the window
 *
 * Output(s):
 *   The start and end addresses of the forwarding window via
 *   the *rsrc pointer
 *
 * Returns:
 *   0 - Success
 *   -EFAULT - Fail, Invalid input parameters
 *   -ENODEV - Fail, specified bridge is not valid
 *
 */
int
db_get_forwarding_window(int bridge_idx, struct resource *rsrc)
{
  struct pci_dev *db;
  int db_devnum, primary_bus;

  if (!rsrc) {
	  printk(KERN_ERR "%s: invalid input parameters\n", __FUNCTION__);
	  return (-EFAULT);
  }

  db_devnum = db_idx_to_devnum(bridge_idx);
   if (db_devnum < 0) {
	   printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
			  __FUNCTION__, bridge_idx);
	   return (-EFAULT);
   }
  primary_bus = cur_cfg_ptr->primary_busses[bridge_idx];

  db = db_get_device(bridge_idx, primary_bus);
  if (!db) {
	  printk(KERN_ERR "%s DB: Bridge %d not found in PCI device list\n",
			 __FUNCTION__, bridge_idx);
	  return (-ENODEV);
  }

  rsrc->start = (int)db->resource[WINDOW_START_IDX].start;
  rsrc->end  = (int)db->resource[WINDOW_STOP_IDX].end;

  DBGPRINT("DB: Forwarding window for bridge %d: 0x%8.8llx to 0x%8.8llx\n",
	   bridge_idx, rsrc->start, rsrc->end);

  return (0);

}
/*
 * int db_get_drawbridge_status()
 *
 * Description:
 *   This routine is used to return the saved error status
 *   from the last PCI config transaction. This data comes
 *   from the PCI_STATUS (0x06) and DB_PCI_SSTATUS (0x4x)
 *   registers of the target drawbridge. Currently only
 *   SERR and PERR are tracked.
 *
 * Input(s):
 *   *bridge           - Target drawbridge.
 *   *primary_status   - pointer used to pass back primary
 *                       side status
 *   *secondary_status - Pointer used to pass back secondary
 *                       side status.
 *
 * Output(s):
 *   Primary and secondary bridge SERR and PERR status
 *
 * Returns:
 *   0 - Success
 *  -EFAULT - Fail, invalid input parameters
 *  -ENODEV - specified drawbridge does not exist, or is
 *            not configured.
 *
 */
int
db_get_drawbridge_status(struct pci_dev *bridge, int *primary_status,
			 int *secondary_status)
{
	db_t *dbp;



	if ((!bridge)||(!primary_status)||(!secondary_status))
		return (-EFAULT);

	*primary_status = 0;
	*secondary_status = 0;

	dbp = (db_t *)pci_get_drvdata(bridge);
	if (DRAWBRIDGE_CONFIGURED(dbp)) {
		*primary_status = dbp->primary_status;
		*secondary_status = dbp->secondary_status;
	} else {
		printk(KERN_ERR "%s: Bridge %d has not been configured\n", __FUNCTION__,
		       PCI_FUNC(bridge->devfn));
		return (-ENODEV);
	}

	return (0);
}
/*
 * int db_clear_drawbridge_status()
 *
 * Description:
 *   This routine is used to clear the primary and
 *   secondary status registers of the parent drawbridge
 *   of a specified pci device. status is cleared by
 *   reading the status registers and writing the contents
 *   back to the registers. any asserted error bits will
 *   be cleared.
 *
 *
 * Input(s):
 *   *dev - PCI device (not the drawbridge itself).
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   0 - Success
 *  -EFAULT - Fail, invalid input parameters
 *  -ENODEV - Parent drawbridge does not exist, or is
 *            not configured.
 *
 */
static int
db_clear_drawbridge_status(db_t *dbp)
{
	uint16_t  regval = 0;
    struct pci_dev *dbdp;

	if (DRAWBRIDGE_CONFIGURED(dbp)) {

		/* get data structure for drawbridge */

		dbdp = dbp->db_dev;

		/* Read and reset status registers */

		pci_read_config_word(dbdp, PCI_STATUS, &regval);
		pci_write_config_word(dbdp, PCI_STATUS, regval);
		pci_read_config_word(dbdp, DB_PCI_SSTATUS, &regval);
		pci_write_config_word(dbdp, DB_PCI_SSTATUS, regval);
	} else {
	    return (-ENODEV);
	}

	return (0);
}
/*
 *
 * void db_dump_csr ()
 *
 * Description:
 *
 *  This function collects draw bridge information for panic dump
 *
 * Input(s):
 *  dev   - pointer to the drawbridge whose configuration
 *         and status registers are to be dumped.
 *  buf - pointer to a buffer to hold the drawbridge
 *         configuration and status register contents.
 *
 * Output(s):
 *  The contents of the drawbridge configuration and status
 *  registers.
 *
 * Return(s):
 *  N/A
 *
 */
void
db_dump_csr(struct pci_dev *dev, db_csr_t *buf)
{
	db_t *dbp;

	/* Check Drawbridge device */
	if ((!dev)||(!buf))
		return;

	/* Check if Drawbridges are initialized */
	if (!(dbp = pci_get_drvdata(dev)))
		return;

	/* If Drawbridge's csr registers are mapped */
	if (dbp->db_local_csr) {
	    memcpy(buf, dbp->db_local_csr, sizeof (db_csr_t));
	}
}

/*
 *
 * void db_save_bridge_config()
 *
 * Description:
 *
 *  Used to save the configuration of a drawbridge
 *  prior to resetting it.
 *
 * Input(s):
 *  dp   - pointer to the drawbridge whose configuration
 *         is to be saved.
 *  save - pointer to a buffer to hold the drawbridge
 *         configuration.
 *
 * Output(s):
 *  The contents of the drawbridge configuration registers.
 *
 * Return(s):
 *  N/A
 *
 */
void
db_save_bridge_config(struct pci_dev *dp, db_cfg_save_t *save)
{

    DBGPRINT("Saving drawbridge registers\n");

    /* Primary base address */
    pci_read_config_dword(dp, PCI_BASE_ADDRESS_0, &save->pbase_address0);
    pci_read_config_dword(dp, PCI_BASE_ADDRESS_1, &save->pbase_address1);
    pci_read_config_dword(dp, PCI_BASE_ADDRESS_2, &save->pbase_address2);


    DBGPRINT("db_save_bridge_config: Save BAR2: 0x%8.8x\n",
	     save->pbase_address2);

    pci_read_config_dword(dp, PCI_BASE_ADDRESS_3, &save->pbase_address3);
    pci_read_config_dword(dp, PCI_BASE_ADDRESS_4, &save->pbase_address4);
    pci_read_config_dword(dp, PCI_BASE_ADDRESS_5, &save->pbase_address5);

    /* Secondary base address */
    pci_read_config_dword(dp, DB_PCI_SBASE_CSR_MEM, &save->sbase_csr_mem);
    pci_read_config_dword(dp, DB_PCI_SBASE_CSR_IO, &save->sbase_csr_io);
    pci_read_config_dword(dp, DB_PCI_SBASE_ADDRESS_0, &save->sbase_address0);
    pci_read_config_dword(dp, DB_PCI_SBASE_ADDRESS_1, &save->sbase_address1);
    pci_read_config_dword(dp, DB_PCI_SBASE_ADDRESS_2, &save->sbase_address2);

    /* Downstream TLBs */
    pci_read_config_dword(dp, DB_PCI_DSTR_XBAR_0, &save->ds_tb0);
    pci_read_config_dword(dp, DB_PCI_DSTR_XBAR_1, &save->ds_tb1);
    pci_read_config_dword(dp, DB_PCI_DSTR_XBAR_2, &save->ds_tb2);
    pci_read_config_dword(dp, DB_PCI_DSTR_XBAR_3, &save->ds_tb3);

    /* Upstream TLBs */
    pci_read_config_dword(dp, DB_PCI_USTR_XBAR_0, &save->us_tb0);
    pci_read_config_dword(dp, DB_PCI_USTR_XBAR_1, &save->us_tb1);

    /* Reset Control */
    pci_read_config_dword(dp, DB_PCI_RESET_CTL, &save->reset_ctrl);

    /* Command registers */
    pci_read_config_word(dp, PCI_COMMAND, &save->primary_command);
    pci_read_config_word(dp, DB_PCI_SCOMMAND, &save->secondary_command);

    /* Status registers */
    pci_read_config_word(dp, PCI_STATUS, &save->primary_status);
    pci_read_config_word(dp, DB_PCI_SSTATUS, &save->secondary_status);

    /* Chip Control registers */
    pci_read_config_word(dp, DB_PCI_CCR0, &save->chip_ctrl0);
    pci_read_config_word(dp, DB_PCI_CCR1, &save->chip_ctrl1);

    /* Configuration Control and Status */
    pci_read_config_word(dp,  DB_PCI_CFG_CSR, &save->config_ctrl);

    /* Latency registers */
    pci_read_config_byte(dp, PCI_LATENCY_TIMER, &save->primary_latency);
    pci_read_config_byte(dp, DB_PCI_SLATENCY_TIMER, &save->secondary_latency);

    pci_read_config_byte(dp, PCI_INTERRUPT_LINE, &save->irq);

    pci_read_config_byte(dp, PCI_CACHE_LINE_SIZE, &save->cache_line);
    pci_read_config_byte(dp, DB_PCI_SCD_CACHE_LINE_SIZE,
			 &save->secondary_cache_line);


}
/*
 *
 * void db_restore_bridge_config()
 *
 * Description:
 *
 *  Used to restore the configuration of a drawbridge
 *  after resetting it.
 *
 * Input(s):
 *  dp   - pointer to the drawbridge whose configuration
 *         is to be saved.
 *  save - pointer to a buffer holding the drawbridge
 *         configuration.
 *
 * Output(s):
 *  N/A
 *
 * Return(s):
 *  N/A
 *
 */
void
db_restore_bridge_config(struct pci_dev *dp, db_cfg_save_t *save)
{
  uint32_t data;

  /* First, disable bridge so the subsequent register writes */
  /* won't cause a downstream transaction to occur */

  DBGPRINT("Restoring drawbridge registers\n");

  data = save->primary_command & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY |
				    PCI_COMMAND_MASTER | PCI_COMMAND_SERR | PCI_COMMAND_PARITY);

  pci_write_config_word(dp, PCI_COMMAND, data);

  data = save->secondary_command  & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY |
				       PCI_COMMAND_MASTER | PCI_COMMAND_SERR |
					   PCI_COMMAND_PARITY);

  pci_write_config_word(dp, DB_PCI_SCOMMAND, data);

    /* Primary base address */
    pci_write_config_dword(dp, PCI_BASE_ADDRESS_0, save->pbase_address0);
    pci_write_config_dword(dp, PCI_BASE_ADDRESS_1, save->pbase_address1);

    DBGPRINT("db_restore_bridge_config: Restore BAR2: 0x%8.8x\n",
	     save->pbase_address2);

    pci_write_config_dword(dp, PCI_BASE_ADDRESS_2, save->pbase_address2);
    pci_write_config_dword(dp, PCI_BASE_ADDRESS_3, save->pbase_address3);
    pci_write_config_dword(dp, PCI_BASE_ADDRESS_4, save->pbase_address4);
    pci_write_config_dword(dp, PCI_BASE_ADDRESS_5, save->pbase_address5);

    /* Secondary base address */
    pci_write_config_dword(dp, DB_PCI_SBASE_CSR_MEM, save->sbase_csr_mem);
    pci_write_config_dword(dp, DB_PCI_SBASE_CSR_IO, save->sbase_csr_io);
    pci_write_config_dword(dp, DB_PCI_SBASE_ADDRESS_0, save->sbase_address0);
    pci_write_config_dword(dp, DB_PCI_SBASE_ADDRESS_1, save->sbase_address1);
    pci_write_config_dword(dp, DB_PCI_SBASE_ADDRESS_2, save->sbase_address2);

    /* Downstream TLBs */
    pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_0, save->ds_tb0);
    pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_1, save->ds_tb1);
    pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_2, save->ds_tb2);
    pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_3, save->ds_tb3);

    /* Upstream TLBs */
    pci_write_config_dword(dp, DB_PCI_USTR_XBAR_0, save->us_tb0);
    pci_write_config_dword(dp, DB_PCI_USTR_XBAR_1, save->us_tb1);

    /* Reset Control */
    pci_write_config_dword(dp, DB_PCI_RESET_CTL, save->reset_ctrl);


    /* Status registers */
    pci_write_config_word(dp, PCI_STATUS, save->primary_status);
    pci_write_config_word(dp, DB_PCI_SSTATUS, save->secondary_status);

    /* Chip Control registers */
    pci_write_config_word(dp, DB_PCI_CCR0, save->chip_ctrl0);
    pci_write_config_word(dp, DB_PCI_CCR1, save->chip_ctrl1);

    /* Configuration Control and Status */
    pci_write_config_word(dp,  DB_PCI_CFG_CSR, save->config_ctrl);

    /* Latency registers */
    pci_write_config_byte(dp, PCI_LATENCY_TIMER, save->primary_latency);
    pci_write_config_byte(dp, DB_PCI_SLATENCY_TIMER, save->secondary_latency);

    pci_write_config_byte(dp, PCI_INTERRUPT_LINE, save->irq);

    pci_write_config_byte(dp, PCI_CACHE_LINE_SIZE, save->cache_line);
    pci_write_config_byte(dp, DB_PCI_SCD_CACHE_LINE_SIZE,
			  save->secondary_cache_line);

    /* Command registers */
    pci_write_config_word(dp, PCI_COMMAND, save->primary_command);
    pci_write_config_word(dp, DB_PCI_SCOMMAND, save->secondary_command);
}

/*
 * int db_timeout_restore_drawbridge()
 *
 * Description:
 *   Restore the PCI configuration of a specified
 *   drawbridge after a timeout reset.
 *
 *   WARNING: This function is called in interrupt context
 *            and must not do anything that could block.
 *
 *            Interrupts may be disabled, so this function
 *            must also ensure the HW watchdog timer does not expire.
 *
 * Input(s):
 *  dp   - pointer to the drawbridge whose configuration
 *         is to be saved.
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *  0 - Success
 *  <0 - Error, invalid bridge or no register cache.
 *
 */
int
db_timeout_restore_drawbridge(struct pci_dev *db)
{

	kick_watchdog();

	if ((db)&&(pci_get_drvdata(db))) {
	  db_t *dbp = (db_t *)pci_get_drvdata(db);
	  db_restore_bridge_config(db, &dbp->cfg_save);
	} else {
		kick_watchdog();
		printk(KERN_ERR "DB ERROR:  Cannot restore bridge - no "
		       "register cache\n");
		return (-EFAULT);
	}

	kick_watchdog();

	return (0);
}
/*
 *
 * int db_pci_primary_reset()
 *
 * Description:
 *
 *  Call the platform specific routine to perform a primary reset of all
 *  drawbridges.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  N/A
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
int
db_pci_primary_reset(void)
{
	int i, db_devnum, primary_bus;
	struct pci_dev *db_dev;
	db_t *dbp;
	int error = 0;

	if (cur_cfg_ptr != NULL && cur_cfg_ptr->reset_method != NULL) {
		error = cur_cfg_ptr->reset_method();
	} else {
		DBGPRINT("db_pci_primary_reset: reset function is NULL!\n");
		return (-1);
	}

	/*
	 * Restore the bridge contents if we have previously
	 * saved them. If we haven't saved them, then
	 * just leave quietly, and expect the caller to
	 * ensure the bridges get reprogrammed.
	 */
	for (i = 1; i <= cur_cfg_ptr->num_db; i++) {

		db_devnum = db_idx_to_devnum(i);
		if (db_devnum < 0) {
			printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
				   __FUNCTION__, i);
			continue;
		}
		primary_bus = cur_cfg_ptr->primary_busses[i];

		db_dev = db_get_device(db_devnum, primary_bus);
		if ((db_dev)&&(pci_get_drvdata(db_dev))) {
			dbp = (db_t *)pci_get_drvdata(db_dev);
			db_restore_bridge_config(db_dev, &dbp->cfg_save);
		}
	}

	return (error);
}

/*
 * int db_read_bases()
 *
 * Description:
 *   Read a specified number of base address registers (BARs) from
 *   a specified PCI device, allocate the requested resources (address
 *   space) and record the resources assigned.
 *
 *   If the device is already configured, don't alter it, just
 *   record the resources assigned. The device is already configured
 *   if the BAR is non-zero and not all 1's.
 *
 * Input(s):
 *   dp - Pointer to data structure defining the PCI device.
 *   howmany - Indicates how many BARs to read.
 *   usr - Optional user interface structure that specifies the
 *         users callback methods and token; if NULL, the default,
 *         client-independent methods must be used;
 *
 *  NOTE: This method should be used only on systems where the
 *        PCI devices should not be reconfigured unless a PCI
 *        reset of the entire blade has been done prior to calling
 *
 * Output(s):
 *   None
 *
 * Returns:
 *    < 0 - PCI Scan failed
 *      0 - PCI Scan pass so far.
 *
 */
int
db_read_bases(struct pci_dev *dp, unsigned int howmany, db_usr_if_t *usr)
{
	unsigned int bar;
	int i;
	int error = 0;
	u32 sz = 0;
	u32 base = 0;						/* base address */
	struct resource *res = NULL;
	db_t *dbp = NULL;
	db_known_dev_t *known_devp;
	db_usr_cmd_get_dev_attr_t	usr_data;

    dbp = (db_t *)dev_get_drvdata(&dp->dev);
	if (!dbp) {
		printk(KERN_ERR "%s:  NULL dp sysdata pointer\n", __FUNCTION__);
		return (-EINVAL);
	}

	if (usr != NULL) {
		int	rc;

		if (usr->ctrl_func == NULL) {
			printk(KERN_ERR "%s:  NULL usr callback pointer\n", __FUNCTION__);
			return (-EINVAL);
		}

  		memset(&usr_data, 0, sizeof (db_usr_cmd_get_dev_attr_t));
		usr_data.bus		= dp->bus->number;
		usr_data.devfn		= dp->devfn;
		usr_data.num_bars	= DB_MAX_BARS;
		usr_data.vendor		= dp->vendor;
		usr_data.device		= dp->device;

		rc = (usr->ctrl_func)(usr->data, DB_USR_CMD_GET_DEV_ATTR, &usr_data);
		if (rc != 0) {
			printk(KERN_ERR "%s: usr get attr bus %d devfn %d vdr %0x "
				   "devid %0x rc %d\n",
				__FUNCTION__, usr_data.bus, usr_data.devfn,
				usr_data.vendor, usr_data.device, rc);
			return (-EINVAL);
		}
		known_devp = (db_known_dev_t *)&usr_data.vendor;

	} else {

		if (cur_cfg_ptr->known_devs == NULL) {
			printk(KERN_ERR "%s: current platform %d has no known devices\n",
				__FUNCTION__, cur_cfg_ptr->swbd);
			return (-EINVAL);
		}

		/* Look up the device in the list of known ones */
		i = 0;
		known_devp = NULL;
		while (1) {
			if (cur_cfg_ptr->known_devs[i].vendor == 0) {	/* end of list */
				break;
			} else if (dp->vendor == cur_cfg_ptr->known_devs[i].vendor &&
				dp->device == cur_cfg_ptr->known_devs[i].device) {
				known_devp = &cur_cfg_ptr->known_devs[i];
				break;
			} else {
				i++;
			}
		}

		if (known_devp == NULL) {
			printk(KERN_ERR
				   "%s: No known device matching vend=0x%x dev=0x%x "
				   "bus=%d slot=%d\n", __FUNCTION__, dp->vendor, dp->device,
				   dp->bus->number,	PCI_SLOT(dp->devfn));
			return (-EINVAL);
		}
	}

	/*
	 * The balance of this code should read base registers only. Don't do
	 * anything that would disable, re-configure, or otherwise stop the
	 * normal operation of the device.
	 */

	for (i = 0; i < howmany; i++) {
		res = &dp->resource[i];
    	res->name = "not supported";
		bar = PCI_BASE_ADDRESS_0 + (i * 4);

		/* Determine memory base address */
		if ((error = pci_read_config_dword(dp, bar, &base))) {
			printk(KERN_ERR
				"%s: config read failed bus=%d slot=%d off=0x%x, ret=%d",
				__FUNCTION__, dp->bus->number, PCI_SLOT(dp->devfn), bar, error);
			return (error);
		}

		if (base == 0 || base == 0xffffffff) {	/* unconfigured BAR */
			continue;
		}

		/* Set BAR attributes (prefetch, type, etc.) */
    	res->flags |= known_devp->bar_attr[i];
		base &= PCI_BASE_ADDRESS_MEM_MASK;

		/* Check if this active BAR is recognized */
		if ((sz = known_devp->bar_sz[i]) == 0) {
			printk(KERN_ERR
				"%s: WARNING unknown size for bar 0x%x bus=%d slot=%d\n",
				__FUNCTION__, bar, dp->bus->number, PCI_SLOT(dp->devfn));
			continue;
		}

		DBGPRINT("  BAR %d (0x%x) Contents: 0x%8.8x  Size: 0x%8.8x\n",
				 i, bar, base, sz);

    	if ((base & PCI_BASE_ADDRESS_SPACE) != PCI_BASE_ADDRESS_SPACE_MEMORY) {
			/* We only deal with memory BARs */
			continue;
		}


		res->start = base;
		res->end = res->start + (unsigned long) sz - 1;

		dbp->lowest_mem_addr = res->end + 1;

		DBGPRINT("  Defining device memory window from: 0x%8.8llx "
				 "to: 0x%8.8llx LMA 0x%x\n", res->start, res->end,
				 dbp->lowest_mem_addr);
		DBGPRINT("  Memory size requested: 0x%8.8x\n", sz);
		DBGPRINT("  Device: %d  Bus: %d\n", PCI_SLOT(dp->devfn),
				dp->bus->number);
	} /* for bars */

	return (error);
}


/*
 *
 * int db_primary_reset_uninit()
 *
 * Description:
 *
 *  Encapsulation of the processor-specific un-initialization in
 *  support of the DB primary bus interface reset.  The reset
 *  control function should only *control* the interface, not
 *  initialize it: including mapping/unmapping resources.
 *
 *  A subsequent designer/developer may wish to generalize this
 *  to support all platform/processor-specific un-initialization
 *
 * Input(s):
 *  None.
 *
 * Output(s):
 *  global gpio_base member of DB control structure is invalidated
 *  with NULL value
 *
 * Returns:
 *  0 - Success
 *
 */
static int
db_primary_reset_uninit(void)
{
  /* Unmap the GPIO control register space. */
  if ((cur_cfg_ptr != NULL) && (cur_cfg_ptr->gpio_base != NULL)) {
	iounmap(cur_cfg_ptr->gpio_base);
	cur_cfg_ptr->gpio_base = NULL;
  }
	return (0);
}

/*
 * int db_alloc_small_bar()
 *
 * Description:
 *   Called to allocate PCI address space on systems that use
 *   the "dynamic" address allocation method.  Allocates small
 *   BAR regions from a reserved region high in the blade allocation
 *   range to reduce fragmentation due to alignment requirements.
 *   e.g. if there's a 16 MB BAR, followed by a 4KB BAR, which is again
 *   followed by a 16 MB BAR, there will be a 16MB-4KB fragment due
 *   to the requirement that the last 16 MB bar be aligned to a 16 MB
 *   starting address.
 *
 *   NOTE: This function is called from CPU-specific code.
 *
 * Input(s):
 *  dp -  Pointer to the device structure of the PCI device
 *        being configured.
 *  dbp - DB pointer.
 *  rel_blade -	Relative blade number (0 index)
 *  res - Pointer the resource table for the particular BAR
 *        of the PCI device being configured
 *  sz  - Size of address space being requested for this BAR
 *  base - Pointer used to pass back the base address of the
 *         physical PCI address space assigned to this BAR.
 *
 * Output(s):
 *  Base address of the physical PCI address space assigned to
 *  the target BAR of the target PCI device.
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
static int
db_alloc_small_bar(
	struct pci_dev *dp,
	db_t *dbp,
	int rel_blade,
	struct resource *res,
	u32 sz,
	u32 *base)
{
	u32 offset;
	u32 remainder;
	u32 align_amt;


	if (dp == NULL || dbp == NULL || res == NULL || base == NULL) {
		printk(KERN_ERR "%s: NULL ptr dp=0x%x dbp=0x%x res=0x%x base=0x%x\n",
			__FUNCTION__, (unsigned)dp, (unsigned)dbp,
			(unsigned)res, (unsigned)base);
		return (-EINVAL);
	}

	/*
	 * Align the offset with the BAR size being allocated.
	 */
	offset = dbp->small_bar_alloc_off[rel_blade];
	remainder = offset % (sz + 1);

    DBGPRINT("%s: sz=0x%x off=0x%x remainder=0x%x\n", __FUNCTION__,
				sz, offset, remainder);

	if (remainder != 0) { 		/* Need to align offset */
		align_amt = (sz + 1) - remainder;
		offset += align_amt;
    	DBGPRINT("%s: align_amt=0x%x new off=0x%x\n", __FUNCTION__,
			align_amt, offset);
	}

	/* Validate that there's enough space left to alloc this BAR */
	if ((offset + sz + 1) > cur_cfg_ptr->blade_alloc_sz) {
		printk(KERN_ERR "%s: ERROR - PCI device %d class 0x%x on bus %d "
			"exceeds memory limit for slot small BAR alloc area (0x%x). "
			"cur alloc offset=0x%x\n", __FUNCTION__,
			PCI_SLOT(dp->devfn), dp->class, dp->bus->number,
			DB_SMALL_BAR_ALLOC_SZ, offset);
		res->start = res->end = *base = 0;
		return (-EINVAL);
	}

	dbp->small_bar_alloc_off[rel_blade] = offset + (sz + 1);

	/*
	 * The BAR start address is the DB base addr + offset of blade
	 * within the DB + aligned small BAR allocation offset.
	 */
    res->start =
    	(dbp->window_start + (rel_blade * cur_cfg_ptr->blade_alloc_sz) +
		 offset) & PCI_BASE_ADDRESS_MEM_MASK;

    res->end = res->start + sz;

	DBGPRINT("%s: start=0x%08x end=0x%08x next offset=0x%x\n",
			__FUNCTION__, (unsigned)res->start, (unsigned)res->end,
			(unsigned)dbp->small_bar_alloc_off[rel_blade]);

	*base =  res->start;

	return (0);
}


/*
 * int db_dynamic_address_allocate()
 *
 * Description:
 *
 *   Used to allocate PCI address space on systems that use
 *   the "dynamic" address allocation method.
 *
 *   It's dynamic in that each blade slot is pre-allocated
 *   a specific range of PCI address space. The PCI devices
 *   in that slot can be assigned however much space they require
 *   within that range. The assignment is done on a first-come-first-serve
 *   basis.
 *
 *   NOTE: This function is called from CPU-specific code.
 *
 * Input(s):
 *  dp -  Pointer to the device structure of the PCI device
 *        being configured.
 *  res - Pointer the resource table for the particular BAR
 *        of the PCI device being configured
 *  sz  - Size of address space being requested for this BAR
 *  base - Pointer used to pass back the base address of the
 *         physical PCI address space assigned to this BAR.
 *
 * Output(s):
 *  Base address of the physical PCI address space assigned to
 *  the target BAR of the target PCI device.
 *
 * Returns:
 *  0 - Success
 *  Non-0 - Failed.
 *
 */
int
db_dynamic_address_allocate(struct pci_dev *dp,
				struct resource *res,
				u32 sz, u32 *base)
{
    u32 l;
    db_t *dbp = NULL;
    int error = 0;
    unsigned int relative_blade;
    u32 remainder;
	u32 align_amt = 0;
	u32 max_alloc_sz;

    dbp = (db_t *)dev_get_drvdata(&dp->dev);
	if (!dbp) {
		printk(KERN_ERR "\n%s:  Null dp sysdata pointer", __FUNCTION__);
		return (-EINVAL);
	}

    if (!base) {
	printk(KERN_ERR "\n%s: Null base pointer", __FUNCTION__);
	return (-EINVAL);
    }

    /*
     * This is the relative blade number behind the drawbridge,
     * which can be 0 to MAX_SLOTS_PER_DB. Don't confuse this
     * with the physical chassis slots.
     */

    relative_blade = dbp->blade_slot;

	/*
	 * If the requested size is small, allocate it from the top of the
	 * blade range to minimize fragmentation.
	 */
	if ((sz + 1) <= DB_SMALL_BAR_ALLOC_THRESH) {
		return (db_alloc_small_bar(dp, dbp, relative_blade, res, sz, base));
	}

	/* Reserve portion of blade addr range for small BAR allocations. */
	max_alloc_sz = cur_cfg_ptr->blade_alloc_sz - DB_SMALL_BAR_ALLOC_SZ;

    /*
     * next_slot_offset always contains the next available address
     * for the blade slot being scanned.
     */
	DBGPRINT(" window_start 0x%8.8X relblade %d blade_alloc_sz 0x%8.8x\n",
			 dbp->window_start, relative_blade, cur_cfg_ptr->blade_alloc_sz);

    l = dbp->window_start + (relative_blade * cur_cfg_ptr->blade_alloc_sz) +
	dbp->next_slot_offset[relative_blade];

	DBGPRINT(" l = 0x%8.8X Next slot offset=%#08X\n", l,
			 dbp->next_slot_offset[relative_blade]);

    res->start = l & PCI_BASE_ADDRESS_MEM_MASK;

    /* Align start address with requested region size. */
    remainder = res->start % (sz + 1);
    DBGPRINT(" sz=0x%08X res->start=0x%08llX remainder=0x%0X\n",
	    sz, res->start, remainder);
    if (remainder != 0) {
	  align_amt = (sz + 1) - remainder;
	  res->start += align_amt;
	  dbp->next_slot_offset[relative_blade] += align_amt;
	  DBGPRINT(" Adjusted start=0x%08llx\n", res->start);
    }

	/*
	 * Make sure that what we are about to allocate does not
	 * exceed the region available to this blade.
	 */

    if (((dbp->next_slot_offset[relative_blade] + sz + 1) > max_alloc_sz) &&
		(dp->class != 0x68000)) {
		printk(KERN_ERR "DB: WARNING - PCI device %d class 0x%x on bus %d "
			   "exceeds memory limit for slot.\n", PCI_SLOT(dp->devfn),
			   dp->class, dp->bus->number);
		printk(KERN_ERR "              Requested size: 0x%8.8x  Slot max: "
			   "0x%8.8x\n", sz + 1, max_alloc_sz);

		res->start = res->end = l = 0;
        return (-EINVAL);
	}


	/* Allocate what device asks for. Note: sz is region size - 1 */
    res->end = res->start + sz;
    dbp->next_slot_offset[relative_blade] += sz + 1;
    DBGPRINT(" Dynamic end: Next slot offset=%#08X\n",
	     dbp->next_slot_offset[relative_blade]);

    *base = l;

	/*
	 * The following is used for multiple bridges downstream.
	 * But only do this if the allocation just performed did not
	 * come from the "small bars" area
	 */

	dbp->lowest_mem_addr = res->end + 1;

    return (error);
}


/*
 *
 * int db_pci_chip_reset()
 *
 * Description:
 *
 *  Issue a chip reset to a specified drawbridge. The
 *  bridge configuration is saved, then restored after
 *  the reset.
 *
 * Input(s):
 *  dp - Pointer to data structure for the drawbridge
 *
 * Output(s):
 *  None
 *
 * Return(s):
 *  0   - Success
 *  < 0 - Fail
 *
 */
int
db_pci_chip_reset(struct pci_dev *dp)
{
  uint32_t data = 0;
  db_cfg_save_t save;

  DBGPRINT("DrawBridge Chip Reset\n");

  memset(&save, 0, sizeof (save));

  if (!dp) {
    printk(KERN_ERR "\nDB ERROR: no pci device specified for reset\n");
    return (-EFAULT);
  }

  db_save_bridge_config(dp, &save);

  /* Force a chip reset. Clears all config registers, */
  /* and re-reads strapped pins. */

  pci_read_config_dword(dp, DB_PCI_RESET_CTL, &data);
  data |= 2;

  DBGPRINT("DB Chip Reset via CTL Register write data: 0x%x\n", data);

  pci_write_config_dword(dp, DB_PCI_RESET_CTL, data);

  /*
   * Don't allow anything else to run here
   * because the drawbridge is in an unpredictable
   * state until this reset completes.
   * But don't let the delay to exceed 1ms.
   *
   */
  udelay(1000);

  pci_read_config_dword(dp, DB_PCI_RESET_CTL, &data);

  DBGPRINT("DB Chip Reset via CTL read Register: 0x%x  %s\n",
	  data, ((data & 2)?"NOT DONE":"DONE"));

  db_restore_bridge_config(dp, &save);


  return (0);
}
/*
 *
 * int db_secondary_reset_control()
 *
 * Description:
 *
 *  Control state of the PCI Secondary-side reset for a
 *  specified drawbridge. The bridge configuration is not affected.
 *  The reset is not propagated to any downstream devices.
 *
 *  Don't try to print anything in here, because this may be
 *  called with interrupts disabled.
 *
 *
 * Input(s):
 *  bridge_idx - Index of the desired Drawbridge (1..n)
 *  assert - True = assert the reset, False = de-assert the reset
 *
 * Output(s):
 *  None
 *
 * Return(s):
 *  0   = Success
 *  -EFAULT = Fail, specified bridge does not exist.
 *
 */
int
db_secondary_reset_control(int bridge_idx, int assert)
{
  uint32_t data = 0;
  struct pci_dev *dp = NULL;
  int db_devnum, primary_bus;

  db_devnum = db_idx_to_devnum(bridge_idx);
  if (db_devnum < 0) {
	  printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
			  __FUNCTION__, bridge_idx);
	  return (-EFAULT);
  }

  primary_bus = cur_cfg_ptr->primary_busses[bridge_idx];

  dp = db_get_device(db_devnum, primary_bus);
  if (!dp) {
	  printk(KERN_ERR "%s: db_get_device failed\n", __FUNCTION__);
    return (-EFAULT);
  }

  pci_read_config_dword(dp, DB_PCI_RESET_CTL, &data);

  if (assert)
    data |= 1;
  else
    data &= ~1;

  pci_write_config_dword(dp, DB_PCI_RESET_CTL, data);

  return (0);
}
/*
 *
 * int db_bridge_secondary_reset()
 *
 * Description:
 *
 *  Issue a PCI Secondary-side reset to a specified bridge.
 *  The bridge configuration is not affected. The reset is
 *  propagated to ALL downstream devices.
 *
 *  NOTE: Do not use this function to do secondary-side
 *        reset of drawbridge devices. For drawbridges,
 *        use db_drawbridge_secondary_reset().
 *
 * Input(s):
 *  *dp - Pointer to the PCI device structure corresponding to the
 *        target device. It should be a transparent bridge.
 *
 * Output(s):
 *  None
 *
 * Return(s):
 *  N/A
 *
 */
void
db_bridge_secondary_reset(struct pci_dev *dp)
{
	uint16_t control = 0;

	if (dp) {
		pci_read_config_word(dp, PCI_BRIDGE_CONTROL, &control);
		control |= PCI_BRIDGE_CTL_BUS_RESET;
		pci_write_config_word(dp, PCI_BRIDGE_CONTROL, control);
		control &= ~PCI_BRIDGE_CTL_BUS_RESET;
		pci_write_config_word(dp, PCI_BRIDGE_CONTROL, control);

		/*
		 * Ensure the bus register on the primary
		 * side is cleared. Other code uses that
		 * register to determine if downstream
		 * devices need to be configured.
		 */

		pci_write_config_dword(dp, PCI_PRIMARY_BUS, 0);
	}

}
/*
 *
 * int db_drawbridge_secondary_reset()
 *
 * Description:
 *
 *  Issue a PCI Secondary-side reset to a specified drawbridge.
 *  The bridge configuration is not affected. The reset is
 *  not propagated to any downstream devices.
 *
 * Input(s):
 *  bridge - Identify which DrawBridge 1, 2, or 3
 *
 * Output(s):
 *  None
 *
 * Return(s):
 *  0   - Success
 *  < 0 - Fail
 *
 */
int
db_drawbridge_secondary_reset(int bridge)
{
  int status;
  ulong flags;

  DBGPRINT("DrawBridge (%d) Secondary Side Reset\n", bridge);

  SYS_INT_SAVE();
  status = db_secondary_reset_control(bridge, 1);

  if (status == 0) {

    /*
     * Don't allow anything else to run here
     * because the drawbridge is in an unpredictable
     * state until this reset completes.
     * But don't let the delay to exceed 1ms.
     *
     */
    udelay(1000);

    status = db_secondary_reset_control(bridge, 0);

  }

  SYS_INT_RESTORE();

  return (status);

}
/*
 * int db_reset_drawbridge()
 *
 * Description:
 *
 *  Issue the appropriate reset for drawbridge devices.
 *  This function is a wrapper for the actual reset routine
 *  so we can easily change how resets are done without changing
 *  the function interface.
 *
 *  All drawbridge devices will be reset.
 *
 * Inputs(s):
 *  None
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  0   - Success
 *  < 0 - Fail
 *
 */
int
db_reset_drawbridge(void)
{
  int status = 0;

  DBGPRINT("DB: 21554 bridges: %d  other bridges: %d\n",
	   db_count_21554, db_count_other);

  if ((cur_cfg_ptr->db_fixup_method)&&(db_fixup_needed)) {
      status = cur_cfg_ptr->db_fixup_method();
      db_fixup_needed = (status)?1:0;
  } else {
      status = db_pci_primary_reset();
  }


  DBGPRINT("DB: 21554 bridges: %d  other bridges: %d\n",
	   db_count_21554, db_count_other);

  return (status);

}
/*
 * static int db_prescan_init()
 *
 * Description:
 *   This routine performs initialization of the Drawbridge PCI-to-PCI
 *   non-transparent bridge device that is needed prior to scanning for
 *   devices downstream of the bridge.
 *
 * Input(s):
 *   *dp  - A pointer to the PCI device structure corresponding to the device
 *          being instantiated.
 *   *id  - A pointer to the PCI device ID structure entry which matched for
 *          this device.
 *   *cmd - A pointer to return the original contents of the PCI
 *          command register
 *
 * Output(s):
 *   *cmd - Original contents of the PCI command register
 *
 *
 * Returns:
 *   PCI_BIOS_SUCCESSFUL if OK, otherwise < 0 on error.
 *
 */
static int
db_prescan_init(struct pci_dev *dp, const struct pci_device_id *id,
		 uint16_t *cmd)
{
	uint32_t data;
	db_t *dbp;
	struct pci_bus *child = NULL;
	struct list_head *list;
	int i;

	/*
	 * Find the structure for PCI bus 0, which
	 * should be the first (and only??) element
	 * directly attached to pci_root_buses by
	 * the Linux kernel
	 */
	list_for_each(list, &pci_root_buses) {

	  child = pci_bus_b(list);
	  break;

	}

	/*
	 * Now get the subordinate bus number from
	 * the PCI bus 0 structure. This is the highest
	 * PCI bus number found so far. We have to start
	 * our numbering after that one. This will account
	 * for any bridges on PCI bus 0 that Linux has
	 * scanned before FabOS loads.
	 *
	 * If there is no PCI bus 0 structure, then we
	 * have no PCI subsystem, so don't even try to
	 * touch any PCI hardware.
	 *
	 */
	if (child) {

	  dp->bus->secondary = child->subordinate;

	} else {
	  printk(KERN_ERR "DB ERROR: PCI bus 0 does not exist. PCI "
		 "subsystem unusable\n");
	  goto done;
	}

	dbp = (db_t *)pci_get_drvdata(dp);
	if (!dbp) {
		printk(KERN_ERR "%s: DrawBridge data is missing\n", db_name);
		goto done;
	}



	DBGPRINT("\nCSR Space start: 0x%8.8llx  CSR Space len: 0x%8.8llx",
		 pci_resource_start(dp, 0), pci_resource_len(dp, 0));

	dbp->db_dev = dp;

	/*
	 * This is the bus attached to the drawbridge secondary
	 * side. Config transactions to this bus must use type 0
	 * format. Config transactions to buses downstream of this
	 * bus must use type 1 format.
	 */

	i = db_devnum_to_idx(PCI_SLOT(dp->devfn));
	if (i < 0) {
		printk(KERN_CRIT "%s: Invalid bridge index %d for bridge %d\n",
			   __FUNCTION__, i, PCI_SLOT(dp->devfn));
		return (-1);
	}

	dbp->root_bus_number = cur_cfg_ptr->root_busses[i];
	dbp->primary_bus_number = cur_cfg_ptr->primary_busses[i];

	/* Disable the device before we start altering its registers */

	pci_read_config_word(dp, PCI_COMMAND, cmd);
	data = *cmd & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY |
			PCI_COMMAND_MASTER | PCI_COMMAND_SERR | PCI_COMMAND_PARITY);
	pci_write_config_word(dp, PCI_COMMAND, data);
	pci_write_config_word(dp, DB_PCI_SCOMMAND, data);

	/* TBD: CMI fix - Set the cache line size (in units of dwords) */
	db_set_cache_line_size(i, 8);

	/*
	 * Enable to perform downstream configuration transactions
	 * when the Downstream Configuration Data register is
	 * accessed.
	 */

	pci_read_config_dword(dp, DB_PCI_CFG_OWN, &data);
	data |= 0x02020000;
	pci_write_config_dword(dp, DB_PCI_CFG_OWN, data);


	/*
	 * Enable secondary clock output, then issue secondary
	 * reset. This prevents the chip from hanging when it tries to
	 * do secondary-side transactions.
	 */

	pci_write_config_word(dp, DB_PCI_CCR0, 0);
	pci_write_config_dword(dp, DB_PCI_RESET_CTL, DB_RESET_CTL_SEC);
	pci_write_config_dword(dp, DB_PCI_RESET_CTL, 0);

	/* This is Mem BAR1, offset 0x18 */
	pci_read_config_dword(dp, PCI_BASE_ADDRESS_2, &data);
	data &= (~(0xf));

	DBGPRINT("\ndb_prescan_init: BAR2 = 0x%8.8x\n", data);

	if (data != 0) {

	  dbp->window_start = dbp->lowest_mem_addr = data;
	  dbp->window_stop = dbp->window_start + 0x28000000;

	  DBGPRINT("\nForwarding Window From: 0x%8.8x  To:  0x%8.8x  "
		   "Root Bus %d  IDSel %d \n", dbp->window_start,
		   dbp->window_stop, dbp->root_bus_number, PCI_SLOT(dp->devfn));

	  /* Program the downstream Translation Base Addresses */

	  /* Mem BAR1, offset 0x18 to xlated BAR1, offset 0x98 */
	  pci_read_config_dword(dp, PCI_BASE_ADDRESS_2, &data);
	  data &= (~(0xf));
	  pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_1, data); // For BAR1

	  /* Mem BAR2, offset 0x1C to xlated BAR2, offset 0x9C */
	  pci_read_config_dword(dp, PCI_BASE_ADDRESS_3, &data);
	  data &= (~(0xf));
	  pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_2, data);

	  /* Mem BAR3, offset 0x20 to xlated BAR3, offset 0xA0 */
	  pci_read_config_dword(dp, PCI_BASE_ADDRESS_4, &data);
	  data &= (~(0xf));
	  pci_write_config_dword(dp, DB_PCI_DSTR_XBAR_3, data);

	  /*
	   * Map the bridge CSR into PCI address space.
	   *   This is the "local" CSR space
	   *
	   */

	  pci_read_config_dword(dp, PCI_BASE_ADDRESS_0, &data);
	  data &= (~(0xf));

	  /*
	   * Clear the array of next (chassis) slot offsets to allocate. Each
	   * element holds the next offset from the bridge base address to
	   * allocate to the next request or PCI memory.
	   *
	   * Also init the offsets used for tracking small BAR allocations.
	   */
	  for (i = 0; i < MAX_SLOTS_PER_DB; i++) {
		  dbp->next_slot_offset[i] = 0x00000000;
		  dbp->small_bar_alloc_off[i] =
			  cur_cfg_ptr->blade_alloc_sz - DB_SMALL_BAR_ALLOC_SZ;
	  }

	  /* Set the latency timer values for the primary and secondary sides. */
	  pci_write_config_byte(dp, PCI_LATENCY_TIMER, cur_cfg_ptr->ntb_lt_primary);
	  pci_write_config_byte(dp, DB_PCI_SLATENCY_TIMER,
			cur_cfg_ptr->ntb_lt_secondary);

	} else {
	  printk(KERN_ERR "Forwarding window not defined\n");
	  return (-EADDRNOTAVAIL);
	}

	if (pci_enable_device(dp))
		goto done_local_iounmap;

	db_set_master(dp);

	dbp->usable = 1;

	/*
	 * Need a bus object for the secondary side of the drawbridge.
	 * Linux does not know to create this bus, so we have to
	 * do it as part of the drawbridge init.
	 */
	if (!dp->subordinate) {
		int error = 0;

		if ((error = db_add_drawbridge_subordinate(dp))) {
			printk(KERN_CRIT "%s: Cannot add supordinate bus for device %s\n",
				   __FUNCTION__, pci_name(dp));
			return (error);
		}
	}
	return (0);

 done_local_iounmap:
	DBGPRINT("\nERROR - at done_local_iounmap");
	release_mem_region(data, sizeof (db_csr_t));

 done:

	return (-ENODEV);
}

/*
 * static void db_postscan_init()
 *
 * Description:
 *   This routine performs the non-transparent bridge initialization
 *   needed after all the downstream devices have been scanned
 *
 * Input(s):
 *   *dp - A pointer to the PCI device structure corresponding to the device
 *         being instantiated.
 *   cmd   The previously read contents of the pci command register
 *
 * Output(s):
 *   None
 *
 *
 * Returns:
 *   None
 *
 */
static void db_postscan_init(struct pci_dev *dp, uint16_t cmd)
{
	uint16_t regval = 0;
	db_t *dbp;

  /* Finish setting up the non-transparent bridge */

  db_read_irq(dp);

  /* Enable forwarding of SERR# from downstream */

  pci_read_config_word(dp, DB_PCI_CCR0, &regval);
  regval |= DB_CCR0_SERR_FWD;
  pci_write_config_word(dp, DB_PCI_CCR0, regval);

  /* Enable SERR#, parity error, I/O Space, Memory Space, and Master */
  /* for both primary and secondary */

  cmd |= (PCI_COMMAND_IO | PCI_COMMAND_MEMORY |
	  PCI_COMMAND_MASTER | PCI_COMMAND_SERR | PCI_COMMAND_PARITY);

  pci_write_config_word(dp, PCI_COMMAND, cmd);

  /*
   * Check if memory access should be disabled on the secondary side
   * initially.  This is to prevent PCI masters on the blades from
   * transferring data to the CP on failover.
   */
  if ((cur_cfg_ptr != NULL) && (!cur_cfg_ptr->enable_secondary_mem))
	cmd &= ~PCI_COMMAND_MEMORY;

  pci_write_config_word(dp, DB_PCI_SCOMMAND, cmd);

  /* Clear SERR# if it came on */
  pci_read_config_word(dp, PCI_STATUS, &cmd);
  pci_write_config_word(dp, PCI_STATUS, cmd);

  pci_read_config_word(dp, DB_PCI_SSTATUS, &cmd);
  pci_write_config_word(dp, DB_PCI_SSTATUS, cmd);

  dbp = (db_t *)pci_get_drvdata(dp);

  /*
   * Save the final drawbridge config register state
   */
  if (DRAWBRIDGE_CONFIGURED(dbp))
	  db_save_bridge_config(dp, &dbp->cfg_save);
  else
	  printk(KERN_ERR "DB ERROR: Cannot cache Drawbridge configuration "
		 "registers\n");

}

/*
 * int db_scan_drawbridge()
 *
 * Description:
 *
 *   This is the entry point for the system module to use. The CP must
 *   be unfenced, and the PCI clock enabled. The master/slave handshake
 *   between the two CPs (if both installed) must already be done or
 *   else the results will be unpredictable.
 *
 * Input(s):
 *   *dp - A pointer to the PCI device structure corresponding to the device
 *         being instantiated.
 *   recursive - If true, configure the bridge and do a configuration scan
 *               of all devices behind the bridge. If false, just configure
 *               the bridge and return.
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   0 if OK, otherwise < 0 on error.
 *
 */
int
db_scan_drawbridge(struct pci_dev *dp, int recursive)
{
  uint16_t cmd = 0;
  uint32_t max = 0;
  int error = 0, needInit, bridge_idx, root_bus;
  db_t *dbp = NULL;


  bridge_idx = db_devnum_to_idx(PCI_SLOT(dp->devfn));
  if (bridge_idx < 0) {
	  printk(KERN_CRIT "%s: Invalid bridge index %d for bridge %d\n",
			 __FUNCTION__, bridge_idx, PCI_SLOT(dp->devfn));
	  return (-1);
  }
  root_bus = cur_cfg_ptr->root_busses[bridge_idx];

  DBGPRINT("\n\n=> Scan device %s at Bus %d, Device %d, Function %d "
		 "bridge idx %d root bus %d\n\n",
		 db_name, dp->bus->number, PCI_SLOT(dp->devfn),
		 PCI_FUNC(dp->devfn), bridge_idx, root_bus);

  DBGPRINT("\tParent Primary Bus: %d   Secondary Bus: %d  Subordinate "
		 "Bus: %d\n", dp->bus->primary, dp->bus->secondary,
		 dp->bus->subordinate);

  dbp = (db_t *)pci_get_drvdata(dp);

  if (DRAWBRIDGE_CONFIGURED(dbp)) {
    DBGPRINT("DrawBridge has already been configured\n");
    needInit = 0;
  } else {
    DBGPRINT("New DrawBridge found - configuration needed\n");
    needInit = 1;
  }

  if (needInit)
    error = db_prescan_init(dp, NULL, &cmd);

  if (error != 0) {

    return (error);

  } else {

    dbp = (db_t *)pci_get_drvdata(dp);

    if (DRAWBRIDGE_NOT_CONFIGURED(dbp)) {
      printk(KERN_ERR "\n\nDB ERROR: Incomplete configuration. Secondary bus "
	     "assigned, but no private data\n\n");
      debugAccessors = 0;
      return (-EFAULT);
    }

    if (recursive) {
      error = db_scan_bus(dp, &db_ops, dbp, &max);
    } else {
      max = dp->bus->secondary;
    }

    if (needInit) {

      dp->bus->subordinate = max;

      db_postscan_init(dp, cmd);
    }

  }


  DBGPRINT("\n\n\tAFTER Found %s at Bus %d, Device %d, Function %d.\n",
      db_name,
      dp->bus->number, PCI_SLOT(dp->devfn), PCI_FUNC(dp->devfn));

  DBGPRINT("\tParent Primary Bus: %d  Secondary Bus: %d  Subordinate Bus: %d\n",
      dp->bus->primary, dp->bus->secondary, dp->bus->subordinate);

  if (dp->subordinate) {
    DBGPRINT("\tSubordinate: Primary Bus: %d   Secondary Bus: %d  ",
	dp->subordinate->primary, dp->subordinate->secondary);
    DBGPRINT("Subordinate Bus: %d\n\n", dp->subordinate->subordinate);
  }

  return (error);

}

/*
 *  int db_disable_bridge()
 *
 *   Description:
 *	This function is used to disable a transparent PCI bridge.
 *	This needs to be done if PCI scan of a blade fails. If the
 *	bridge is left enabled (especially the bus registers) then
 *	the bridge may interfere with transactions intended for
 *	other blades and make them look bad.
 *
 *     So far, this problem has only been seen to happen if the
 *     bridge subordinate register is left with 0xFF.
 *
 *  Input(s):
 *     dp - Pointer to bridge object
 *
 *  Output(s):
 *     None
 *
 *  Returns:
 *     0 - No errors
 *     NonZero - error occured while accessing a config register.
 *
 *               The primary_status and secondary_status
 *               members of the db_t struct (a member of dp)
 *               can be examined to determine if SERR or PAERR
 *               occured on the last register access.
 *
 */
int
db_disable_bridge(struct pci_dev *dp)
{
	unsigned int buses = 0;
	int error, accum_error = 0;

	/*
	 * We keep going even if the register accesses generate
	 * errors because not clearing these registers
	 * can make an entire logical switch appear faulty.
	 */

	if ((error = pci_read_config_dword(dp, PCI_PRIMARY_BUS,
					   &buses) != 0)) {
		accum_error = error;
	}

	/*
	 * Clear the bus registers so this bridge
	 * doesn't try to forward any config transactions
	 */

	buses &= DB_BRIDGE_BUS_MASK;
	if ((error = pci_write_config_dword(dp, PCI_PRIMARY_BUS, buses))) {
		accum_error = error;
	}

	/*
	 * Turn off memory, I/O and master enables so normal
	 * transactions do not get forwarded.
	 */

	if ((error = pci_write_config_word(dp, PCI_COMMAND, 0x0000))) {
		accum_error = error;
	}

	return (accum_error);
}

/*
 * int db_do_scan_blade()
 *
 * Description:
 *   Scan a slot and try to recover if the scan fails.
 *
 *   If the scan fails, perfrom a recovery procedure on the
 *   slot and rescan up to SCAN_RETRY_MAX times. Must take care
 *   to release any objects created by the failed scan before
 *   rescanning. Otherwise, memory leaks or duplicate entries
 *   could occur.
 *
 * Input(s):
 *   *drawbridge - Pointer to PCI device data structure for the drawbridge
 *                 that the target slot is behind.
 *   *temp       - Pointer to a PCI device structure template for the
 *                 blade that is to be scanned.
 *   **dphandle  - Handle (pointer to pointer) to use in passing back
 *                 the PCI device tree that's found (if any).
 *   reconfig    - Flags whether blade devices should be reconfigured.  On an HA
 *                 failover certain blades, e.g. Scimitar, should not be.
 *   usr         - Optional user interface structure that specifies the
 *                 users callback methods and token; if NULL, the default,
 *                 client-independent methods must be used;
 *
 * Output(s):
 *   *dphandle   - List of PCI devices found at this slot. The first
 *                 node should be the bridge device
 *		 - If there is no device to be found, it will reset pointer
 *		   to NULL.
 *
 */
static int
db_do_scan_blade(struct pci_dev *drawbridge, struct pci_dev *temp,
		 struct pci_dev **dphandle, int reconfig, db_usr_if_t *usr)
{
	db_t *dbp = NULL;
	struct pci_dev *dp;
	int error = 0;
	struct pci_bus *bus = NULL;
	unsigned int max;

	/* dphandle need to be initialized */
	*dphandle = NULL;

	/* coverity[deref_ptr] */
	bus = temp->bus;
	/* coverity[deref_ptr] */
	max = bus->secondary;

	if ((temp)&&(bus)) {
		dbp = (db_t *)dev_get_drvdata(&bus->dev);
	} else {
		DBGPRINT("\nMissing system  data. Cannot scan/configure "
		       "this Device: %d", PCI_SLOT(temp->devfn));
		return (DB_DRAWBRIDGE_UNINITIALIZED);
	}

	pr_debug("DB: Scanning bus %04x:%02x\n", pci_domain_nr(bus), bus->number);

	/*
	 * Ensure the status fields are
	 * cleared before 1st register
	 * access so we can detect new SERRs.
	 */
	dbp->primary_status = 0;
	dbp->secondary_status = 0;

	/*
	 * Relative slot number needs to be assigned before scan device. The slot
	 * number is relative to the drawbridge it is behind.
	 */
        dbp->blade_slot = PCI_SLOT(temp->devfn) - DB_FIRST_BLADE_SLOT;
        DBGPRINT("\n ****%s slot number:%d devfn:%d\n",
                __FUNCTION__, dbp->blade_slot, temp->devfn);

	/*
	 * Preserve the original temp device
	 * in case we have to print debug messages.
	 * Calls downstream of db_scan_slot() will
	 * alter the contents of the struct.
	 */
	error = db_scan_slot(bus, temp->devfn, &dp, reconfig, usr);

	if ((dp)&&(!error)) {

		/*
		 * Found something.
		 */

		*dphandle = dp;  /* make sure it gets back to the caller */

		if ((dp->class >> 8) == PCI_CLASS_BRIDGE_PCI) {
			/*
			 * db_scan_slot found a bridge at devid,
			 * so we are OK to proceed
			 */
			error = db_scan_bridge(bus, dp, &max, reconfig, 1, 0, usr);
			if (error) {
				/* Found nothing behind the bridge */
				DBGPRINT(KERN_WARNING "DB WARNING: PCI "
				       "scan found an isolated bridge "
				       "device. Bus %d  DevID %d\n",
				       temp->bus->number,
				       PCI_SLOT(temp->devfn));
				db_disable_bridge(dp);
				error = DB_ISOLATED_BRIDGE;

			} else {
				/* Found something behind the bridge */
				if (max > bus->subordinate)
					bus->subordinate = max;

				if (max > drawbridge->bus->subordinate)
					drawbridge->bus->subordinate =	max;

				error = 0;
			}

		} else {

			/*
			 * We found something, but it was not a bridge.
			 * This is invalid, so we need to reset the
			 * bridge via blade fence/unfence and try again.
			 */
			DBGPRINT("DB WARNING: Requested to "
			       "scan a non-bridge device\n");
			DBGPRINT("              Vendor: 0x%x"
			       " Device:  0x%x Class 0x%x\n",
			       dp->vendor, dp->device,
			       (dp->class >> 8));

			error = DB_NONBRIDGE_DEVICE;
		}
	} else {
		/*
		 * If nothing was found in this slot, or
		 * we got an error scanning the blade bridge,
		 * then consider it an empty slot and let
		 * the caller do error recovery.
		 *
		 */
		error = DB_EMPTY_BLADE_SLOT;
	}

	pr_debug("DB: Bus scan for %04x:%02x returning with max=%02x, error=%d\n",
		pci_domain_nr(bus), bus->number, max, error);
	return (error);
}


int
db_add_drawbridge_subordinate(struct pci_dev *drawbridge)
{
  unsigned int max;
  db_t *dbp = NULL;
  struct pci_bus *parent, *child;

  if (!drawbridge) {
	return (DB_INVALID_PARAMETER);
  }

  dbp = (db_t *)pci_get_drvdata(drawbridge);
  if (!drawbridge->subordinate) {

    DBGPRINT("%s: Drawbridge has "
			 "no subordinate bus\n", __FUNCTION__);

    /*
	 * Max (busses scanned so far) behind this drawbridge will just
	 * be the DB root_bus_number value.
     */
    parent = drawbridge->bus;
    max = dbp->root_bus_number;

    DBGPRINT("%s: Parent Bus %d  "
		   "Parent Secondary %d  Parent Sub %d max bus %d\n", __FUNCTION__,
		   parent->number, parent->secondary, parent->subordinate, max);

    /*
     * We need to create a subordinate bus (bus
     * this bridge bridges to) for
     * the drawbridge. But first check to see if
     * a bus by this number already exists. If
     * it does, then we are out of sync with
     * the PCI subsystem, and do not dare
     * touch any PCI hardware.
     *
     * The new buss structure is attached to
     * drawbridge->subordinate.
     */
    child = db_bus_exists(parent, max);
    if (child) {

      DBGPRINT("DB ERROR: The requested Drawbridge subordinate bus (%d) "
	     "is already in use\n", max);
      return (DB_BUS_UNAVAILABLE);

    } else {

      DBGPRINT("%s: Adding new bus %d\n", __FUNCTION__, max);
      child = db_add_new_bus(parent, drawbridge, max);

      if (child) {
		dev_set_drvdata(&child->dev, dbp);
		child->ops = &db_ops;
		child->subordinate = child->number;
      } else {
		return (DB_BUS_CREATE_FAIL);
      }
    }
  } else {
    DBGPRINT("%s: Drawbridge already had "
			 "subordinate bus\n", __FUNCTION__);
  }
  return (0);
}

/*
 * int db_scan_blade_cmn()
 *
 * Description:
 *  Scan a single specified slot behind a specified DrawBridge
 *  This function is called from two publicly exported wrapper
 *  functions to perform the slot scan with and without device
 *  configuration:
 *  	db_scan_blade()
 *  	db_scan_blade_noreconfig()
 *
 *  This routine should be called when a new blade is discovered
 *  in the switch or chassis. It should not be called for blades
 *  that have already been scanned.
 *
 * Input(s):
 *  drawbridge - Pointer to data structure of the desired parent
 *               DrawBridge.
 *  **dev      - Handle (pointer to pointer) to use in passing back
 *               the PCI device tree that's found (if any).
 *	           - If there is any error, pointer (*dev) will be reset to NULL.
 *  devid      - The PCI device id of the desired slot. Must be
 *               a PCI bridge.
 *  reconfig   - Flags whether blade devices should be reconfigured.  On an HA
 *               failover certain blades, e.g. Scimitar, should not be.
 *   usr       - Optional user interface structure that specifies the
 *               users callback methods and token; if NULL, the default,
 *               client-independent methods must be used;
 *
 * Output(s):
 *  If the specified device exists and is a PCI bridge, the
 *  bridge and children devices will be added to the global
 *  PCI device list. This devices found on this slot will
 *  also be passed back to the caller via **dev.
 *
 * Returns:
 *   0 - Success
 *   Any other value - Fail. Value is a code indicating what went wrong.
 *		       See linux/include/linux/db.h for definitions.
 *
 */
static int
db_scan_blade_cmn(struct pci_dev *drawbridge, struct pci_dev **dev,
			int devid, int reconfig, db_usr_if_t *usr)
{
	int error = 0;
	struct pci_dev *dev0 = NULL, *dp;
	struct pci_bus *bus;
	db_t *dbp = NULL;
	u32 save0xcc = 0, regval32 = 0;

	if ((!drawbridge)||(!dev)) {
		error = DB_INVALID_PARAMETER;
		goto out;
	}

	/* Need to initialize *dev first */
	*dev = NULL;

	dbp = (db_t *)pci_get_drvdata(drawbridge);
	if (DRAWBRIDGE_NOT_CONFIGURED(dbp)) {
		error = DB_DRAWBRIDGE_UNINITIALIZED;
		goto out;
	}

	if ((devid < 0)||(devid > 31)) {
		error = DB_INVALID_DEVICE_ID;
		goto out;
	}

	/*
	 * Drawbridge subordinate bus is now created in db_prescan_init().
	 * If the bus does not exist by now, then something has failed
	 * earlier and we cannot go any further.
	 */
	if (!drawbridge->subordinate) {
		return (DB_BUS_UNAVAILABLE);
	}

	bus = drawbridge->subordinate;

	dev_dbg(&drawbridge->dev,
		"DB: Scanning PCI DevID %d of bus %d (secondary = %d) behind "
		"DrawBridge on blade %d\n", devid, bus->number, bus->secondary,
		dbp->blade_slot);

	/* Create a device template */
	if ((dev0 = kmalloc(sizeof (struct pci_dev), GFP_KERNEL)) == NULL) {
		printk(KERN_ERR "%s: cannot allocate data\n", __FUNCTION__);
		return (-ENOMEM);
	}
	memset(dev0, 0, sizeof (struct pci_dev));
	INIT_LIST_HEAD(&dev0->bus_list);
	dev0->bus = bus;
	dev0->devfn = PCI_DEVFN(devid, 0); /* slot, func */

	if (pci_get_slot(bus, dev0->devfn)) {
	/*
     * db_scan_blade() should only be called to scan a new blade. If
     * it's already in the PCI list, then it was not properly
     * removed. Don't proceed or we will get duplicate objects
     * in the PCI list and cause lots of confusion.
     */
		printk(KERN_ERR "DB: the device behind Drawbridge %d is "
			"already in PCI device list.\nRe-scan not permitted\n",
			PCI_SLOT(drawbridge->devfn));
		error = DB_DEVICE_EXISTS;
		goto out;
	} else {
		/*
		 * Defect #44823
		 * Clear any error status that may be left over
		 * from previous PCI accesses. Otherwise, the error
		 * checks in the downstream code could get fooled into
		 * thinking a new error occured.
		 */
		error = db_clear_drawbridge_status(dbp);
		if (error) {
			printk(KERN_ERR "DB: Clear drawbridge status failed\n");
			goto out;
		}

		/*
		 * Disable NTB error reporting while the blade is
		 * being scanned. Otherwise, we get lots of PCI
		 * errors that we don't care about.
		 */
		pci_read_config_dword(drawbridge, 0xcc, &save0xcc);
		regval32 = 0;
		pci_write_config_dword(drawbridge, 0xcc, regval32);

		error = db_do_scan_blade(drawbridge, dev0, &dp, reconfig, usr);
		pci_write_config_dword(drawbridge, 0xcc, save0xcc);
		*dev = dp;
		if (!error) {
			pci_bus_add_devices(drawbridge->subordinate);
		} else {
			goto out;
		}
	}


	bus = db_bus_exists(dp->bus, dp->subordinate->number);
	DBGPRINT("%s: input %p %s %s ", __FUNCTION__,
		   dp->bus, dev_name(&(dp->bus)->dev),
		   dev_name(&(dp->subordinate)->dev));
	if (bus) {
		DBGPRINT("get %s ", dev_name(&bus->dev));
		DBGPRINT("input %d\n", dp->subordinate->number);
	}

out:
	if (error) {
		printk(KERN_ERR "%s: error=%d\n", __FUNCTION__, error);
	}
	if (dev0) {
		kfree(dev0);
	}

	if ((*dev)&&(dbp)) {

		dbp->blade_bridge_refcount[dbp->blade_slot] = pci_dev_refcount(*dev);
	}
	return (error);
}

/*
 * int db_scan_blade()
 *
 * Description:
 *  Wrapper routine that invokes the internal scan blade routine.  This
 *  routine assumes that the blade devices should be reconfigured.
 *
 * Input(s):
 *  drawbridge - Pointer to data structure of the desired parent
 *               DrawBridge.
 *  **dev      - Handle (pointer to pointer) to use in passing back
 *               the PCI device tree that's found (if any).
 *	       - If there is any error, pointer (*dev) will be reset to NULL.
 *  devid      - The PCI device id of the desired slot. Must be
 *               a PCI bridge.
 *
 * Output(s):
 *  **dev is potentially set by the internal blade scan routine.
 *
 * Returns:
 *   Passes on the return value from the internal blade scan routine.
 *
 */
int
db_scan_blade(struct pci_dev *drawbridge, struct pci_dev **dev, int devid)
{
    /* Scan blade and reconfigure devices */
    return (db_scan_blade_cmn(
				drawbridge, dev, devid, DB_RECONFIG, (db_usr_if_t *)NULL));
}

/*
 * int db_scan_blade_noreconfig()
 *
 * Description:
 *  Wrapper routine that invokes the internal scan blade routine.  This
 *  routine assumes that the blade devices should not be reconfigured.
 *
 * Input(s):
 *  drawbridge - Pointer to data structure of the desired parent
 *               DrawBridge.
 *  **dev      - Handle (pointer to pointer) to use in passing back
 *               the PCI device tree that's found (if any).
 *	       - If there is any error, pointer (*dev) will be reset to NULL.
 *  devid      - The PCI device id of the desired slot. Must be
 *               a PCI bridge.
 *   usr       - Optional user interface structure that specifies the
 *               users callback methods and token; if NULL, the default,
 *               client-independent methods must be used;
 *
 * Output(s):
 *  **dev is potentially set by the internal blade scan routine.
 *
 * Returns:
 *   Passes on the return value from the internal blade scan routine.
 *
 */
int
db_scan_blade_noreconfig(struct pci_dev *drawbridge, struct pci_dev **dev,
						 int devid, db_usr_if_t *usr)
{
    /* Scan blade and without reconfiguring devices */
    return (db_scan_blade_cmn(drawbridge, dev, devid, DB_NORECONFIG, usr));
}

/*
 * static int db_probe()
 *
 * Description:
 *   This routine instantiates a discovered instance of the Drawbridge PCI-
 *   to-PCI non-transparent bridge device.
 *   This probe routine returns error when it does not understand the
 *   the device. SInce it has asked for PCI_ANY_ID, it is important to
 *   that it ignore devices it does not know so that other registered
 *   drivers can probe that.
 *
 * Input(s):
 *   *dp - A pointer to the PCI device structure corresponding to the device
 *         being instantiated.
 *   *id - A pointer to the PCI device ID structure entry which matched for
 *         this device.
 *
 * Output(s):
 *   *dp - A pointer to the PCI device structure with device-specific data
 *         attached to it at successful completion of this routine.
 *
 * Returns:
 *   0 if OK, otherwise < 0 on error.
 *
 */
static int
db_probe(struct pci_dev *dp, const struct pci_device_id *id)
{
  db_t *dbp = NULL;
  int i;

  if (cur_cfg_ptr->db_fixup_method) {

      db_fixup_needed = 1;
      if (id->device == PCI_DEVICE_ID_DEC_21554) {
	  db_count_21554++;
	  if (cur_cfg_ptr != NULL && db_count_21554 == cur_cfg_ptr->num_db)
	      db_fixup_needed = 0;
      } else if (id->device == PCI_DEVICE_ID_INTEL_21555) {
	  db_count_other++;
      } else {

		/* Should we return some erro number ? */
		  return (-1);
      }

  } else {
      db_fixup_needed = 0;
  }

  if (dp->class != 0x68000)
      return (-1);

  if ((dbp = kmalloc(sizeof (db_t), GFP_KERNEL)) == NULL) {
	  printk(KERN_ERR "%s: cannot allocate data\n", db_name);
	  return (-ENOMEM);
  }

  dbp->usable = 0;
  dbp->cur_bus_index = 1;

  for (i = 0; i < MAX_SLOTS_PER_DB; i++) {
	  dbp->blade_bridge_refcount[i] = 0;
  }
  pci_set_drvdata(dp, dbp);

  db_save_bridge_config(dp, &dbp->cfg_save);

  DBGPRINT("\nDB_PROBE: Registers cached\n");

  DBGPRINT("\n\n******** Found %s at Bus 0x%x, Slot 0x%x, "
	   "Function 0x%x.*********\n",
	   db_name, dp->bus->number, PCI_SLOT(dp->devfn), PCI_FUNC(dp->devfn));

  DBGPRINT("*************** Primary Bus: %d   Secondary Bus: %d  ",
	 dp->bus->primary, dp->bus->secondary);

  DBGPRINT("Subordinate Bus: %d\n", dp->bus->subordinate);


  DBGPRINT("*************** Pointer dp: 0x%8.8x   db->bus: 0x%8.8x \n",
	 (int)dp, (int)dp->bus);


  return (dp->bus->subordinate);
}

/*
 * void db_remove_child_devices()
 *
 *  Description:
 *   Remove an instance of a PCI device from the
 *   global PCI device list.
 *
 *   WARNING: IF THIS FUNCTION IS CALLED WITH
 *            DRAWBRIDGE, THE CALLER MUST FIRST RELEASE
 *            THE driver_data OR MEMORY LEAK WILL OCCUR
 *
 *  Input(s):
 *    dp - Pointer to PCI device structure to be
 *         removed from the device list
 */
void
db_remove_child_devices(struct pci_dev *dp)
{
	pci_remove_bus_device(dp);
}

/*
 * void db_remove_drawbridge()
 *
 * Description:
 *   This routine removes all downstream PCI bus and device instances
 *   attached to a specified draw bridge. The draw bridge instance
 *   itself is not removed, but it's device-specific data is.
 *
 *   Removing the drawbridge instance without removing the downstream
 *   device instances should not be allowed to happen anyway. This routines
 *   removes all child devices before releasing the device-specific data.
 */
void __devexit
db_remove_drawbridge(struct pci_dev *dp)
{
	printk(KERN_ERR "%s: Cannot support drawbridge removal at this time. "
		   "A reboot is required\n", __FUNCTION__);
}
/*
 *  refcount_fixup()
 *
 *    Adjust the reference count of a specified device to what it
 *    was when the PCI scan was completed. This is to prevent a
 *    memory leak or kernel panic if a user of the device did
 *    something to alter the reference count.
 *
 * Inputs:
 *   pdev - pci_dev object of the target device
 *   expected_refcount - what the refernce count should be.
 *
 */
static void
refcount_fixup(struct pci_dev *pdev, int expected_refcount)
{
	int actual_refcount, delta, i;

	if (pdev) {
		actual_refcount = pci_dev_refcount(pdev);

		if (expected_refcount > actual_refcount) {
			delta = expected_refcount - actual_refcount;
			pr_debug("%s: Increment actual refcount by %d\n",
				   __FUNCTION__, delta);
			for (i = 0; i < delta; i++) {
				pci_dev_get(pdev);
			}
		} else {
			delta = actual_refcount - expected_refcount;
			pr_debug("%s: Decrement actual refcount by %d\n",
				   __FUNCTION__, delta);

			for (i = 0; i < delta; i++) {
				pci_dev_put(pdev);
			}
		}

		pr_debug("%s: After fixup actual refcount %d expected refcount %d\n",
			   __FUNCTION__, pci_dev_refcount(pdev), expected_refcount);

	} else {
		printk(KERN_ERR "%s: No pci_dev device. Cannot fixup refcount\n",
			   __FUNCTION__);
	}

}
/*
 * int db_remove_tree()
 *
 * Description:
 *
 *  Remove all child busses and devices from the global PCI
 *  list for a specified device. The bridge itself is also
 *  removed. If the device is not a bridge, then only that
 *  device is removed.
 *
 *  If the device is a bridge, sets the subordinate of the parent
 *  bus value to the  parent secondary bus value.
 *
 * WARNING: Never kfree the dp data, as the caller is
 *          expecting to do this.
 *
 * Input(s):
 *   dp - Pointer to the specified PCI device
 *
 * Output(s):
 *   None
 *
 * Returns:
 *  0 - Success
 *  <0 - The specified PCI device is not in the PCI
 *       global device list.
 *
 */

int
db_remove_tree(struct pci_dev *dp)
{
	db_t *dbp = NULL;
	int relative_slot;

	if (!dp) {
		printk(KERN_ERR "%s: No parent bridge specified\n", __FUNCTION__);
		return (-ENODEV);
	}

	/*
	 * Reset the next address offset of the slot being removed.
	 */
    dbp = (db_t *)dev_get_drvdata(&dp->dev);
	if (dbp) {
		uint32_t refcount;

		dev_dbg(&dp->dev, "db_remove_tree: blade_slot=%d\n", dbp->blade_slot);

		/* Compute relative slot number (0 .. #blades per DB - 1) */
		relative_slot = PCI_SLOT(dp->devfn) - DB_FIRST_BLADE_SLOT;
		dbp->next_slot_offset[relative_slot] = 0;
		dbp->small_bar_alloc_off[relative_slot] =
			cur_cfg_ptr->blade_alloc_sz - DB_SMALL_BAR_ALLOC_SZ;
		DBGPRINT("db_remove_tree: cleared next slot offset for slot %d\n",
				 relative_slot);

		/*
		 * If the current reference count is not the same as it was when
		 * the scan completed, then we have to restore the original reference
		 * count before cleaning up the objects. Otherwise, either a memory
		 * leak or memory corruption will occur
		 */
		refcount = pci_dev_refcount(dp);
		if (refcount != dbp->blade_bridge_refcount[relative_slot]) {

			printk(KERN_ERR "%s: Dev %s relblade %d REFCOUNT MISMATCH:  "
				   "IS %d  S/B %d\n", __FUNCTION__, pci_name(dp),
				   relative_slot, refcount,
				   dbp->blade_bridge_refcount[relative_slot]);

			refcount_fixup(dp, dbp->blade_bridge_refcount[dbp->blade_slot]);
		}

		pci_remove_bus_device(dp);

	} else {
		printk(KERN_ERR "%s: Null dp sysdata pointer.\n", __FUNCTION__);
	}

	return (0);
}

#define PCI_byte_BAD 0
#define PCI_word_BAD (pos & 1)
#define PCI_dword_BAD (pos & 3)

#define DB_PCI_OP_READ(size, type, len) \
int db_pci_read_config_##size \
	(struct pci_dev *dev, int pos, type *value) \
{                                   \
    int res;                            \
    u32 data = 0;                           \
    if (PCI_##size##_BAD) return PCIBIOS_BAD_REGISTER_NUMBER;   \
    res = dev->bus->ops->read(dev->bus, dev->devfn, pos, len, &data);      \
    *value = (type)data;                        \
    return (res);								\
}

#define DB_PCI_OP_WRITE(size, type, len) \
int db_pci_write_config_##size \
	(struct pci_dev *dev, int pos, type value)		\
{                                   \
    if (PCI_##size##_BAD) return PCIBIOS_BAD_REGISTER_NUMBER;   \
    return (dev->bus->ops->write(dev->bus, dev->devfn, pos, len, value)); \
}

DB_PCI_OP_READ(byte, u8, 1)
DB_PCI_OP_READ(word, u16, 2)
DB_PCI_OP_READ(dword, u32, 4)
DB_PCI_OP_WRITE(byte, u8, 1)
DB_PCI_OP_WRITE(word, u16, 2)
DB_PCI_OP_WRITE(dword, u32, 4)

/*
 * int db_read_xxxxsize()
 *
 * Description:
 *
 *   Routine for reading PCI device configuration registers.
 *   Implemented as a macro, this routine will be instantiated
 *   inline for each of the types of access (byet, word, dword).
 *
 * Input(s):
 *   dp - Pointer to data structure for the PCI bus where the device resides.
 *	 devfn - device function number on which the configuration register is read.
 *   reg - The configuration register being read.
 *   value - Pointer to storage for the read result
 *
 * Output(s):
 *   The contents of the specified register in the parameter
 *   value.
 *
 * Returns:
 *   0 - If the operation succeeded.
 *   <0 - If there was an error.
 *
 */
#define	DB_CONFIG_READ(size, type) \
static int								\
db_read_##size(struct pci_bus *dp, int devfn, int reg, type *value)			\
{									\
	int config, bus = dp->number;			        \
	int error = 0;							\
	u16 status = 0;                                                 \
	db_t *dbp = NULL;                                               \
	pci_dev_t *dbdp = NULL;                                         \
																\
    dbp = (db_t *)dev_get_drvdata(&dp->dev);						\
	if (!dbp) {														\
		printk(KERN_ERR "\ndb_read_%s: Null dp sysdata pointer", #type);  \
			return (-EINVAL);                                             \
	}                                                               \
                                                                        \
	if (dbp->db_dev) {                                              \
		dbdp = dbp->db_dev;					\
	} else {                                                        \
		printk(KERN_ERR "\ndb_read_%s:  Null db_dev pointer", #type);     \
		return (-EINVAL);                                             \
	}                                                               \
                                                                        \
	*value = 0;							\
                                                                        \
	if (debugAccessors)						\
		DBGPRINT("db_read_%s: Bus %d, Slot %d, Function %d, "		\
			"Reg %d. Value @ %p = ", #type, bus,       		\
			PCI_SLOT(devfn), PCI_FUNC(devfn), reg, value);	\
                                                                        \
	/*								\
	 * Use the configuration access semaphore method as documented	\
	 * on page 6-3 of the "21554 PCI-to-PCI Bridge for Embedded	\
	 * Applications User's Manual"...				\
	 */								\
	db_pci_read_config_word(dbdp, DB_PCI_CFG_CSR, &status);     	\
									\
	db_pci_write_config_dword(dbdp, DB_PCI_CFG_OWN, 0x00020000); 	\
									\
	/* 								\
	 * This is supposed to work as a semaphore to prevent concurrent \
	 * access to the configuration registers. But doing this		\
	 * caused frequent hangs during the PCI scan even though only	\
	 * one process was doing config accesses. Leave this disabled 	\
	 * until we have time to pursue it. Higher-level	code is already \
	 * synchronizing access to the PCI config scan.			\
	 */								\
									\
/*                                                                      \
	do {								\
		db_pci_read_config_word(dbdp, DB_PCI_CFG_CSR, &status);   	\
	} while (status & DB_CFG_DSTR_OS);				\
*/                                                                      \
	/* ...we have bus ownership, issue the transaction... */       	\
									\
	if ((bus == 0) || (bus == dbp->root_bus_number))	\
		config = CFGTYPE0(bus, devfn, reg);			\
	else								\
		config = CFGTYPE1(bus, devfn, reg);	       		\
	db_pci_write_config_dword(dbdp, DB_PCI_DSTR_CFG_AR, config);	\
	db_pci_read_config_##size(dbdp, DB_PCI_DSTR_CFG_DR | (reg & 0x03), \
				 value);				\
									\
	/* Read, save and reset status registers */			\
									\
	db_pci_read_config_word(dbdp, PCI_STATUS,				\
				&dbp->primary_status);			\
	db_pci_write_config_word(dbdp, PCI_STATUS,				\
				dbp->primary_status);			\
									\
	db_pci_read_config_word(dbdp, DB_PCI_SSTATUS,			\
				&dbp->secondary_status);		\
	db_pci_write_config_word(dbdp, DB_PCI_SSTATUS,			\
				dbp->secondary_status);			\
									\
	/* Check if there were any errors we care about */		\
									\
	if (dbp->primary_status &= DB_FATAL_ERROR_MASK) {		\
		error = -EINVAL;					\
	}								\
									\
	if (dbp->secondary_status &= DB_FATAL_ERROR_MASK) {		\
		error = -EINVAL;					\
	}								\
									\
	return (error);							\
}

/*
 * int db_write_xxxxsize()
 *
 * Description:
 *
 *   Routine for reading PCI configuration registers.
 *   Implemented as a macro, this routine will be instantiated
 *   inline for each of the types of access (byet, word, dword).
 *
 * Input(s):
 *   dp - Pointer to a data structure for the PCI device whose
 *        configuration register is being accessed.
 *   reg - The configuration register being read.
 *   value - Pointer to storage for the read result
 *
 * Output(s):
 *   The contents of the specified register in the parameter
 *   value.
 *
 * Returns:
 *   0 - If the operation succeeded.
 *   <0 - If there was an error.
 *
 */
#define	DB_CONFIG_WRITE(size, type) \
static int								\
db_write_##size(struct pci_bus *dp, int devfn, int reg, type value)			\
{									\
	int config, bus = dp->number;		\
	int error = 0;							\
	u16 status = 0;			                                \
                                                                        \
	db_t *dbp = NULL;                                               \
	pci_dev_t *dbdp = NULL;                                         \
                                                                        \
    dbp = (db_t *)dev_get_drvdata(&dp->dev);						\
	if (!dbp) {														\
		printk(KERN_ERR "\ndb_write_%s:  Null dp sysdata pointer", #type);\
			return (-EINVAL);                                             \
	}                                                               \
                                                                        \
	if (dbp->db_dev) {                                              \
		dbdp = dbp->db_dev;					\
	} else {                                                        \
		printk(KERN_ERR "\ndb_write_%s:  Null db_dev pointer", #type);    \
		return (-EINVAL);                                             \
	}                                                               \
											\
	if (debugAccessors)						\
		DBGPRINT("db_write_%s: Bus %d, Slot %d, Function %d, "	\
		"Reg %d. Value = %x\n", #type, bus,       		\
	       PCI_SLOT(devfn), PCI_FUNC(devfn), reg, value);	\
									\
	/*								\
	 * Use the configuration access semaphore method as documented	\
	 * on page 6-3 of the "21554 PCI-to-PCI Bridge for Embedded	\
	 * Applications User's Manual"...				\
	 */								\
									\
	db_pci_write_config_dword(dbdp, DB_PCI_CFG_OWN, 0x00020000); 	\
	db_pci_read_config_word(dbdp, DB_PCI_CFG_CSR, &status);     	\
                                                                        \
	/* Doing this seems to cause intermittant hangs. */             \
	/* Leaving it out until we can find out why */                  \
/*			                                                \
	do {								\
		db_pci_read_config_word(dbdp, DB_PCI_CFG_CSR, &status);   	\
	} while (config & DB_CFG_DSTR_OS);				\
*/                                                                      \
									\
	/* ...we have bus ownership, issue the transaction... */       	\
									\
	if ((bus == 0) || (bus == dbp->root_bus_number))	\
		config = CFGTYPE0(bus, devfn, reg);			\
	 else								\
		config = CFGTYPE1(bus, devfn, reg);	       		\
                                                                        \
	db_pci_write_config_dword(dbdp, DB_PCI_DSTR_CFG_AR, config);	\
	db_pci_write_config_##size(dbdp, DB_PCI_DSTR_CFG_DR | (reg & 0x03),\
 				value); 				\
									\
	/* Read, save and reset status registers */			\
									\
	db_pci_read_config_word(dbdp, PCI_STATUS,				\
				&dbp->primary_status);			\
	db_pci_write_config_word(dbdp, PCI_STATUS,				\
				dbp->primary_status);			\
									\
	db_pci_read_config_word(dbdp, DB_PCI_SSTATUS,			\
				&dbp->secondary_status);		\
	db_pci_write_config_word(dbdp, DB_PCI_SSTATUS,			\
				dbp->secondary_status);			\
									\
	/* Check if there were any errors we care about */		\
									\
	if (dbp->primary_status &= DB_FATAL_ERROR_MASK) {		\
		error = -EINVAL;					\
	}								\
									\
	if (dbp->secondary_status &= DB_FATAL_ERROR_MASK) {		\
		error = -EINVAL;					\
	}								\
                                                                        \
	return (error);							\
}

DB_CONFIG_READ(byte, byte)
DB_CONFIG_READ(word, word)
DB_CONFIG_READ(dword, dword)
DB_CONFIG_WRITE(byte, byte)
DB_CONFIG_WRITE(word, word)
DB_CONFIG_WRITE(dword, dword)

/*
 * void db_set_master()
 *
 * Description:
 *   Set the PCI bus master bit in the PCI command
 *   register for a specified PCI device
 *
 * Input(s):
 *   dp - Pointer to data structure defining the
 *        the desired PCI device
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   N/A
 *
 */
static void
db_set_master(struct pci_dev *dp)
{
  u16 cmd = 0;

  pci_read_config_word(dp, PCI_COMMAND, &cmd);
  if (! (cmd & PCI_COMMAND_MASTER)) {
    DBGPRINT("PCI: Enabling bus mastering for device %s\n", pci_name(dp));
    cmd |= PCI_COMMAND_MASTER;
    pci_write_config_word(dp, PCI_COMMAND, cmd);
  }
  pcibios_set_master(dp);
}

/*
 * unsigned int db_calc_resource_flags()
 *
 * Description:
 *   Translate the low bits of the PCI base
 *   to the resource type
 *
 * Input(s):
 *   flags - Contents of the read from a base address register.
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   IORESOURCE_IO - If the BAR wants I/O address space
 *   IORESOURCE_MEM - If the BAR wants memory address space
 *
 */
static unsigned int db_calc_resource_flags(unsigned int flags)
{
  if (flags & PCI_BASE_ADDRESS_SPACE_IO)
    return (IORESOURCE_IO);

  if (flags & PCI_BASE_ADDRESS_MEM_PREFETCH)
    return (IORESOURCE_MEM | IORESOURCE_PREFETCH);

  return (IORESOURCE_MEM);
}

/*
 * int db_configure_bases()
 *
 * Description:
 *   Read a specified number of base address registers (BARs) from
 *   a specified PCI device, allocate the requested resources (address
 *   space) and record the resources assigned.
 *
 *   If the device is already configured, don't alter it, just
 *   record the resources assigned. The device is already configured
 *   if the BAR is non-zero and not all 1's.
 *
 *   NOTE: This routine is _always_ going to disable the transparent
 *         blade bridge and reconfigure all PCI devices on the blade.
 *	   This will block all PCI access to the switch ASICs until
 *         the PCI scan of the blade has been completed.
 *
 * Input(s):
 *   dp - Pointer to data structure defining the PCI device.
 *   howmany - Indicates how many BARs to read.
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   < 0 = PCI scan fail
 *     0 = PCI scan pass
 *
 */
int
db_configure_bases(struct pci_dev *dp, unsigned int howmany)
{
  unsigned int slot, pos, reg, next;
  int error = 0;
  u32 l = 0, sz = 0, tmp;
  u16 cmd = 0;
  struct resource *res = NULL;
  db_t *dbp = NULL;

	dbp = (db_t *)dev_get_drvdata(&dp->dev);
	if (!dbp) {
		printk(KERN_ERR "\n%s:  Null dp sysdata pointer", __FUNCTION__);
		return (-EINVAL);
	}

  DBGPRINT("RECONFIGURING BASE REG for Device %d on bus %d blade_slot %d\n",
	 PCI_SLOT(dp->devfn), dp->bus->number, dbp->blade_slot);

  /* Disable IO and memory while we fiddle */
  if ((error = pci_read_config_word(dp, PCI_COMMAND, &cmd)))
    return (error);

  tmp = cmd & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY);
  pci_write_config_word(dp, PCI_COMMAND, tmp);

  for (pos = 0; pos < howmany; pos = next) {
    next = pos+1;
    res = &dp->resource[pos];
    res->name = "not supported";
    reg = PCI_BASE_ADDRESS_0 + (pos << 2);

    /* Determine memory base address */

    if ((error = pci_read_config_dword(dp, reg, &l)))
      return (error);

    pci_write_config_dword(dp, reg, ~0);
    if ((error =  pci_read_config_dword(dp, reg, &sz))) {
	    printk(KERN_ERR "%s: bus %d dev %d - failed to "
		    "read BAR %d (0x%x) after writing all 1's\n",
		__FUNCTION__, dp->bus->number, PCI_SLOT(dp->devfn), pos, reg);
	    return (error);
    }

    DBGPRINT(" BAR%d requested size=%#08X\n", pos, sz);

    if (!sz || sz == 0xffffffff) {
      pci_write_config_dword(dp, reg, l);
      continue;
    }

	if (l == 0xffffffff)
		l = 0;

    slot = PCI_SLOT(dp->devfn);


	if ((l & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_MEMORY) {

		sz = db_pci_size(sz, PCI_BASE_ADDRESS_MEM_MASK);

		/*
		 * Compute the start addr based on the current allocation method: fixed
		 * or dynamic.
		 */

		error = cur_cfg_ptr->alloc_method(dp, res, sz, &l);

		if (error)
			return (error);

		pci_write_config_byte(dp, PCI_LATENCY_TIMER, 0x80);

		DBGPRINT("  Defining device memory window from: 0x%8.8llx "
				 "to: 0x%8.8llx New LMA 0x%8.8x\n", res->start, res->end,
				 dbp->lowest_mem_addr);
		DBGPRINT("  Memory size requested: 0x%8.8x, Max Allowed: 0x%8.8x  "
				 "Actual Allotted: 0x%8.8llx\n", sz, DB_SLOT_ALLOTMENT - 1,
				 res->end - res->start);
		DBGPRINT("  Device: %d  Bus: %d\n", PCI_SLOT(dp->devfn),
				 dp->bus->number);

    } else { /* It's I/O space, not memory space */

		res->start = l & PCI_BASE_ADDRESS_IO_MASK;
		sz = ~(sz & PCI_BASE_ADDRESS_IO_MASK) & 0xffff;
		res->end = res->start + (unsigned long) sz;
    }

      /* Write the starting address to the device's BAR */
      pci_write_config_dword(dp, reg, res->start);


    DBGPRINT("  Window Start: 0x%8.8x  slot: %d  RelativeBlade: %d  "
	     "Base: 0x%8.8x\n", dbp->window_start, slot, dbp->blade_slot, l);


    /*
     * If bit 2 of BAR is a 1, it's either reserved or
     * is requesting 64-bit address space. Linux
     * doesn't support this as yet.
     *
     */
    res->flags |= (l & 0xf) | db_calc_resource_flags(l);
    if ((l & (PCI_BASE_ADDRESS_SPACE | PCI_BASE_ADDRESS_MEM_TYPE_MASK))
	== (PCI_BASE_ADDRESS_SPACE_MEMORY | PCI_BASE_ADDRESS_MEM_TYPE_64)) {
      printk(KERN_ERR "PCI 64-bit address requested: Reg: 0x%x  Data: 0x%x\n",
	     reg, l);
      pci_read_config_dword(dp, reg+4, &l);
      next++;
#if BITS_PER_LONG == 64
      res->start |= ((unsigned long) l) << 32;
      res->end = res->start + sz;
      pci_write_config_dword(dp, reg+4, ~0);
      pci_read_config_dword(dp, reg+4, &tmp);
      pci_write_config_dword(dp, reg+4, l);
      if (l)
	res->end = res->start + (((unsigned long) ~l) << 32);
#else
      if (l) {
	printk(KERN_ERR "PCI: Unable to handle 64-bit address for device %s\n",
               pci_name(dp));
	printk(KERN_ERR "PCI: Reg: 0x%x  Data: 0x%x\n", reg+4, l);
	res->start = 0;
	res->flags = 0;
	continue;
      }
#endif
    }
  }

  /* Original code restored the command register to it's original */
  /* state. That works only if the the device was configured by   */
  /* the boot prom. Since ours doesn't, we have to enable IO and  */
  /* memory here. */

  cmd |= (PCI_COMMAND_IO | PCI_COMMAND_MEMORY);
  pci_write_config_word(dp, PCI_COMMAND, cmd);

  return (error);

}

/*
 * struct pci_bus * db_alloc_bus()
 *
 * Description:
 *   Allocate and initialize a new PCI bus data structure.
 *   The caller is responsible for freeing the memory when
 *   no longer needed.
 *
 * Input(s):
 *   None
 *
 * Output(s):
 *   A pointer to the new data structure
 *
 */
static struct pci_bus * db_alloc_bus(void)
{
	struct pci_bus *b = NULL;

	b = kzalloc(sizeof (*b), GFP_KERNEL);
	if (b) {
		INIT_LIST_HEAD(&b->node);
		INIT_LIST_HEAD(&b->children);
		INIT_LIST_HEAD(&b->devices);
		INIT_LIST_HEAD(&b->slots);
		INIT_LIST_HEAD(&b->resources);
		b->max_bus_speed = PCI_SPEED_UNKNOWN;
		b->cur_bus_speed = PCI_SPEED_UNKNOWN;
	}
	return (b);
}

static struct pci_bus *db_alloc_child_bus(struct pci_bus *parent,
						struct pci_dev *bridge, int busnr)
{
	struct pci_bus *child;

	/*
	 * Allocate a new bus, and inherit stuff from the parent..
	 */
	child = db_alloc_bus();
	if (!child) {
		printk(KERN_ERR "\nCould not allocate new bus number: %d", busnr);
		return (NULL);
	}

	child->parent = parent;
	child->ops = parent->ops;
	child->sysdata = parent->sysdata;
	child->bus_flags = parent->bus_flags;

	dev_set_drvdata(&child->dev, dev_get_drvdata(&parent->dev));

	/*
	 * initialize some portions of the bus device, but don't register it
	 * now as the parent is not properly set up yet.  This device will get
	 * registered later in pci_bus_add_devices()
	 */
	child->dev.class = parent->dev.class; // inplace for pcibus_class
	dev_set_name(&child->dev, "%04x:%02x", pci_domain_nr(child), busnr);

	/* XXX BRCD comments see if still valid statements */
	  /*
	   * Set up the primary, secondary and subordinate
	   * bus numbers. These will get programmed into
	   * the bridge secondary and subordinate bus
	   * registers as well.
	   *
	   *  primary     = bus on the primary side of dp
	   *  secondary   = the new bus; will be on secondary side of dp
	   *  subordinate = 0xff so that the subsequent scan
	   *                behind this bridge will find all
	   *                lower-level busses. This value
	   *                gets changed to the actual subordinate
	   *                value later.
	   */

	/*
	 * Set up the primary, secondary and subordinate
	 * bus numbers.
	 */
	child->number = child->secondary = busnr;
	child->primary = parent->secondary;
	child->subordinate = 0xff;

	if (!bridge)
		return (child);

	child->self = bridge;
	child->bridge = get_device(&bridge->dev);

#ifdef XXX
	/* Set up default resource pointers and names.. */
	for (i = 0; i < PCI_BRIDGE_RESOURCE_NUM; i++) {
		child->resource[i] = &bridge->resource[PCI_BRIDGE_RESOURCES+i];
		child->resource[i]->name = child->name;
	}
#endif
	bridge->subordinate = child;

	return (child);
}

/*
 * pci_bus * db_add_new_bus()
 *
 * Description:
 *   Allocate a new bus instance and add it the the parent
 *   bus's list of children buses. Attach the bridge instance
 *   to the new bus (child->self).
 *
 * Input(s):
 *  parent  - Parent bus of the new bus to be created.
 *  dp      - Bridge device whose secondary side connects to
 *            the new bus being created.
 *  busnr   - Number to assign to the new bus.
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   NULL - Error, no bus instance created.
 *   NON-NULL - Pointer to the new bus instance.
 *
 */
static struct pci_bus * db_add_new_bus(struct pci_bus *parent,
				       struct pci_dev *dp, int busnr)
{
	struct pci_bus *child;

	child = db_alloc_child_bus(parent, dp, busnr);
	if (child) {
#ifdef XXX
		down_write(&pci_bus_sem);
#endif
		list_add_tail(&child->node, &parent->children);
#ifdef XXX
		up_write(&pci_bus_sem);
#endif

	dev_dbg(&child->dev, "NEW CHILD BUS created\n");
	  DBGPRINT("\nNEW CHILD BUS: number: %d Primary: %d  Secondary: %d  "
		   "Sub: %x", child->number, child->primary, child->secondary,
		   child->subordinate);
	  DBGPRINT("\n  Parent Bus Name: %s Number: %d  Primary: %d  "
		   "Secondary: %d  Sub: %x\n",
		   parent->name, parent->number, parent->primary,
		   parent->secondary, parent->subordinate);
	}

	return (child);
}
static unsigned int
compute_bus_method(struct pci_bus *bus, struct pci_dev *dp,
				unsigned int *maxbus, int is_blade_root_bus,
					   int root_bus_number, db_t *dbp)
{

	unsigned int max = 0;

  /* If this is the first bridge on the blade, we initialize as per slot */
  if (likely(is_blade_root_bus)) {

	DBGPRINT("\n ****%s slot number:%d devfn:%d\n",
            __FUNCTION__, dbp->blade_slot, dp->devfn);

    /*
     * Compute max bus number to scan:
     * dbp->blade_slot = relative slot number range 0..max_slots_per_db - 1
     * cur_cfg_ptr->busses_per_slot = num PCI busses allocated to each slot
     * dbp->root_bus_number is bus on the secondary side of the db
     *
     * Example:
     *
     *          Chassis Slot 1
     *          blade_slot=0, busses_per_slot=10, root_bus_number=2
     *          max = ((0 + 1) * 10) + 2 = 12
     *
     *          Chassis Slot 2
     *          blade_slot=1, busses_per_slot=1, root_bus_number=2
     *          max = ((1 + 1) * 10) + 2 = 22
     *
     *          Chassis Slot 7
     *          blade_slot=0, busses_per_slot=1, root_bus_number=0x42
     *          max = ((0 + 1) * 1) + 0x42 = 76
     */
	max = (((dbp->blade_slot + 1) * cur_cfg_ptr->busses_per_slot) +
		   dbp->root_bus_number + 1);

    dbp->cur_bus_index = 1;

	printk(KERN_CRIT "%s: blade_slot %d busses_per_slot %d  "
		   "root_bus_number %d 0x%x max %d\n", __FUNCTION__, dbp->blade_slot,
		   cur_cfg_ptr->busses_per_slot,
		   dbp->root_bus_number, dbp->root_bus_number, max);

  } else {
	/*
	 * ... Otherwise we compute based on the usual enumeration instead of slot
	 */
    max = root_bus_number + dbp->cur_bus_index;
    ++dbp->cur_bus_index;
  }

  return (max);
}

/*
 * int db_scan_bridge()
 *
 * Description:
 *   Scan for PCI devices on the secondary side of a
 *   specified PCI bridge device. Configure the bridge
 *   to forward transactions to the downstream devices.
 *
 * Input(s):
 *   bus  - The bus to which the primary of the bridge is attached
 *   *dp  - Pointer to struct for the bridge device.
 *   *maxbus - Highest bus number so far.
 *   reconfig - Flags whether blade devices should be reconfigured.  On an HA
 *  			failover certain blades, e.g. Scimitar, should not be.
 *   usr  - Optional user interface structure that specifies the
 *          users callback methods and token; if NULL, the default,
 *          client-independent methods must be used;
 *   root_bus_number - The backplane bus number
 *   is_blade_root_bus - The bridge device is the blade bridge
 *
 * Output(s):
 *   *maxbus - Updated to indicate the highest bus number found
 *             downstream of this bridge.
 *
 * Returns:
 *   < 0 = Scan failed
 *     0 = Scan passed
 *
 */
static int db_scan_bridge(struct pci_bus *bus, struct pci_dev *dp,
				unsigned int *maxbus, int reconfig, int is_blade_root_bus,
				int root_bus_number, db_usr_if_t *usr)
{
	unsigned int buses = 0, base, mem_base, mem_limit, max;
	unsigned short cr = 0;
	int error = 0;
	struct pci_bus *child;
	db_t *dbp = NULL;
	u16 bctl;
	int broken = 0;

    dbp = (db_t *)dev_get_drvdata(&dp->dev);
	if (!dbp) {
		printk(KERN_ERR
			"\nMissing private data. Cannot scan/configure this bridge");
		printk(KERN_ERR "\nBridge Vendor:  %x  Device:  %x  Slot: %d  "
			   "Name: %s\n", dp->vendor, dp->device, PCI_SLOT(dp->devfn),
			   "not supported");
		return (-EINVAL);
	}

	if (cur_cfg_ptr->compute_bus_method) {
		max = cur_cfg_ptr->compute_bus_method(bus, dp, maxbus,
											  is_blade_root_bus,
											  root_bus_number, dbp);
	} else {
		max = compute_bus_method(bus, dp, maxbus,
								 is_blade_root_bus,
								 root_bus_number, dbp);
	}

	/*
	 * Create a child bus and insert it into the tree of buses.
	 */

	/*
	 * Read the existing primary/secondary/subordinate bus
	 * number configuration to determine if the PCI bridge
	 * has already been configured by the system.  If so,
	 * do not modify the configuration, merely note it.
	 */
	pci_read_config_dword(dp, PCI_PRIMARY_BUS, &buses);

	dev_dbg(&dp->dev, "scanning behind bridge SwitchSlot: %d  Device: %d "
		"max: %d, config %06x\n",
		dbp->blade_slot + 1, PCI_SLOT(dp->devfn), max,
		buses & 0xffffff);

	/* Check if setup is sensible at all */
	if ((buses & 0xff) != bus->number || ((buses >> 8) & 0xff) <= bus->number) {
        dev_dbg(&dp->dev, "DB: bus configuration invalid\n");
        broken = 1;
    }

	pci_read_config_word(dp, PCI_BRIDGE_CONTROL, &bctl);

	if ((buses & 0xffff00) && !broken) {
		unsigned int cmax, busnr;

		/*
		 * Bus already configured by firmware, process it in the first
		 * pass and just note the configuration.
		 */
		pr_debug("db_scan_bridge: Bridge at slot %d is already configured "
			"(0x%8.8x)\n", dbp->blade_slot + 1, buses);

	    busnr = (buses >> 8) & 0xFF;	/* extract secondary bus number */

		/*
		 * However, we continue to descend down the hierarchy and
		 * scan remaining child buses.
		 */
		child = pci_find_bus(pci_domain_nr(bus), busnr);
		if (!child) {
			child = db_add_new_bus(bus, dp, busnr);
			if (!child)
				return (-EINVAL);
			child->primary = buses & 0xFF;
			child->subordinate = (buses >> 16) & 0xFF;
			child->bridge_ctl = bctl;
		}

		DBGPRINT("\n  Bridge Buss registers already configured\n");
		DBGPRINT("    PCI bridge %s blade_slot: %d  Device: %d Bus: %d\n",
			pci_name(dp), dbp->blade_slot, PCI_SLOT(dp->devfn), busnr);
		DBGPRINT("    Number: %d  Primary: %d  Secondary: %d  "
				 "Subordinate: %d\n", busnr, child->primary, child->secondary,
				 child->subordinate);

    	error = db_scan_child_bus(child, &cmax, reconfig, usr);
		if (error)
			return (error);
		if (cmax > max)
			max = cmax;
		if (child->subordinate > max)
			max = child->subordinate;
	} else {

    /*
     * Bus registers are 0, so assume the bridge is not configured. Is OK to
     * disable it
     *
     * Clear all status bits and turn off memory, I/O and master enables.
     */

	dev_dbg(&dp->dev, "DB: db_scan_bridge: Bridge at slot %d "
			"is NOT configured (0x%8.8x)\n",
			dbp->blade_slot + 1, buses);

	pci_read_config_word(dp, PCI_COMMAND, &cr);
	pci_write_config_word(dp, PCI_COMMAND, 0x0000);
	/* Clear errors */
	pci_write_config_word(dp, PCI_STATUS, 0xffff);

    if (!(child = pci_find_bus(pci_domain_nr(bus), max))) {
		child = db_add_new_bus(bus, dp, max);
		if (!child)
			return (-EINVAL);
		pr_debug("DB: Bus %d not in list, create a new one",
				child->number);
	}
#ifdef DEBUG
	else
		dev_dbg(&child->dev, "bus exists\n");
#endif

	/*
	 * Configure the bus numbers for this bridge:
	 */
    DBGPRINT("\n  Bridge Bus registers are being configured\n");
    buses &= DB_BRIDGE_BUS_MASK;
    buses = (buses & 0xff000000)
		| ((unsigned int)(child->primary)		<<  0)
		| ((unsigned int)(child->secondary)		<<  8)
		| ((unsigned int)(child->subordinate)	<< 16);

    pci_write_config_dword(dp, PCI_PRIMARY_BUS, buses);

	/* We have to capture the ranges for all the bridges downstream */
    if (likely(is_blade_root_bus)) {

	  /* Initialize the lowest mem start for the first bridge */
      dbp->lowest_mem_addr = dbp->window_start +
	    (cur_cfg_ptr->blade_alloc_sz * dbp->blade_slot);
     }

	/* The start of next bridge is end of previous bridge */
    mem_base = dbp->lowest_mem_addr;
	DBGPRINT(" New LMA 0x%8.8x New mem_base 0x%8.8x\n",
			 dbp->lowest_mem_addr, mem_base);

    /* TBD: CMI fix - Set the cache line size (in units of dwords) */
    pci_write_config_byte(dp, PCI_CACHE_LINE_SIZE, 8);
	/* Configure the primary and secondary side latency timers */
	pci_write_config_byte(dp, PCI_LATENCY_TIMER, cur_cfg_ptr->tb_lt_primary);
	pci_write_config_byte(dp, PCI_SEC_LATENCY_TIMER,
		cur_cfg_ptr->tb_lt_secondary);

	child->bridge_ctl = bctl;
	/* Now we can scan all subordinate buses... */
    error = db_scan_child_bus(child, &max, reconfig, usr);
    /*
     * If scan failed, we must not leave the bridge subordinate
     * bus set to 0xFF. Otherwise it interferes with scan of
     * other blades.
     */
    if (error) {
	    pci_read_config_dword(dp, PCI_PRIMARY_BUS, &buses);
	    buses &= DB_BRIDGE_BUS_MASK;
	    pci_write_config_dword(dp, PCI_PRIMARY_BUS, buses);
	    return (error);
    }

	/*
	 * Re-read the primary/sec/sub bus reg because the high byte
	 * contains the latency timer value that we set above.
	 */
	pci_read_config_dword(dp, PCI_PRIMARY_BUS,  &buses);
    buses = (buses & 0xff00ffff)
      | ((unsigned int)(child->subordinate) << 16);
    pci_write_config_dword(dp, PCI_PRIMARY_BUS, buses);

    /*
     * Set the subordinate bus number to its real value.
	 */
	child->subordinate = max;
	pci_write_config_byte(dp, PCI_SUBORDINATE_BUS, max);


	/* We need to find if the bridge has any device under it */
	/* Such case can happen for PCIe switch if the port is not connected */

	mem_limit = dbp->lowest_mem_addr - 1;

	/* Check if no memory was allocated for this bridge */
	/* XXX: We do not deal with IO space */
	if (mem_base != mem_limit) {
    	base = (mem_base>>16 & 0xfff0);
    	base |= (mem_limit & 0xFFF00000);
    	pci_write_config_dword(dp, PCI_MEMORY_BASE, base);

    	DBGPRINT("\n For Bus %d :: Calculated Downstream memory: from 0x%8.8x "
				 "to 0x%8.8x\n", child->number, mem_base, mem_limit);
    	DBGPRINT("\nDownstream Bridge window base reg:  0x%8.8x\n", base);

    	/* Enable Bus Master, memory, and I/O decode on secondary bus */
    	/*    cr |= 0x07; */
      	cr |= (PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER |\
			   PCI_COMMAND_SERR | PCI_COMMAND_PARITY);
      	pci_write_config_word(dp, PCI_COMMAND, cr);
	} else {
		/* Disable the bridge as it has no device under it */
		pci_write_config_word(dp, PCI_COMMAND, 0x0000);
		pci_write_config_word(dp, PCI_STATUS, 0xffff);
	}
  }

	pci_read_config_dword(dp, PCI_MEMORY_BASE,  &base);
	pci_read_config_dword(dp, PCI_PRIMARY_BUS,  &buses);

	DBGPRINT("\nBridge window: from 0x%8.8x  to 0x%8.8x\n",
      (base << 16), (((base & 0xffff0000) | 0xFFFFF)));
	DBGPRINT("Primary Bus:  %2.2d  Secondary Bus %2.2d  ",
      (buses & 0xff), ((buses >> 8) & 0xff));
	DBGPRINT(" Subordinate Bus: %2.2d\n",
      ((buses >> 16) & 0xff));

	*maxbus = max;
	return (error);
}
/*
 * void db_read_irq()
 *
 * Description:
 *   Read interrupt line and base address registers.
 *   The architecture-dependent code can tweak these, of course.
 *
 * Input(s):
 *   dp - Pointer to a PCI device data structure
 *
 * Output(s):
 *   Contents of the IRQ register, via the dp structure.
 *
 * Returns:
 *   < 0 = IRQ Read failed
 *     0 = IRQ Read passed
 */
static void db_read_irq(struct pci_dev *dp)
{
	unsigned char irq = 0;

	pci_read_config_byte(dp, PCI_INTERRUPT_PIN, &irq);
	dp->pin = irq;
	if (irq)
		pci_read_config_byte(dp, PCI_INTERRUPT_LINE, &irq);

	if ((irq == 0)||(irq == 0xFF))
	  irq = cur_cfg_ptr->irq_num;

	dp->irq = irq;
}

/* copy from pcibios_setup_bus_devices */
static void db_setup_dma(struct pci_dev *dp)
{
#ifdef CONFIG_PPC32
	struct dev_archdata *sd = &dp->dev.archdata;

	/* Hook up default DMA ops */
	sd->dma_ops = &dma_direct_ops;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 34)
	set_dma_offset(&dp->dev, PCI_DRAM_OFFSET);
#else
	sd->dma_data = (void *)PCI_DRAM_OFFSET;
#endif
#else
	BUG(); // not supported yet
#endif
}

/*
 * int db_setup_device()
 *
 * Description:
 *   Configure a PCI device, bridge or non-bridge.
 *
 * Input(s):
 *   dp - Pointer to data structure for the PCI device
 *
 *   reconfig - Flags whether blade devices should be reconfigured.  On an HA
 *  		failover certain blades, e.g. Scimitar, should not be.
 *   usr  - Optional user interface structure that specifies the
 *          users callback methods and token; if NULL, the default,
 *          client-independent methods must be used;
 *
 * Output(s):
 *   Resource allocation information for the PCI device
 *   via the dp structure.
 *
 * Returns:
 *   0  - Success
 *   <0 - Error, device could not be configured and the
 *        data structure contents are undefined.
 *
 */
int
db_setup_device(struct pci_dev *dp, int reconfig, db_usr_if_t *usr)
{
	u32 class;
	u8 hdr_type;
	struct pci_slot *slot;
	int error = 0;

	if (pci_read_config_byte(dp, PCI_HEADER_TYPE, &hdr_type))
		return (-EIO);

	dev_set_drvdata(&dp->dev, dev_get_drvdata(&dp->bus->dev));
	dp->sysdata = dp->bus->sysdata;
	dp->dev.parent = dp->bus->bridge;
	dp->dev.bus = &pci_bus_type;
	dp->hdr_type = hdr_type & 0x7f;
	dp->multifunction = !!(hdr_type & 0x80);
	dp->error_state = pci_channel_io_normal;

	list_for_each_entry(slot, &dp->bus->slots, list)
		if (PCI_SLOT(dp->devfn) == slot->number)
			dp->slot = slot;

	/*
	 * Assume 32-bit PCI; let 64-bit PCI cards (which are far rarer)
	 * set this higher, assuming the system even supports it.
	 */
	dp->dma_mask = 0xffffffff;

	dev_set_name(&dp->dev, "%04x:%02x:%02x.%d", pci_domain_nr(dp->bus),
			 dp->bus->number, PCI_SLOT(dp->devfn),
			 PCI_FUNC(dp->devfn));

	pci_read_config_dword(dp, PCI_CLASS_REVISION, &class);
	dp->revision = class & 0xff;
	class >>= 8;		/* upper 3 bytes */
	dp->class = class;
	class >>= 8;

	dev_dbg(&dp->dev, "found %d:%02x (S=%d) [%04x:%04x] class %06x "
			"header type %02x\n",
			dp->bus->number, dp->devfn, PCI_SLOT(dp->devfn),
			dp->vendor, dp->device, class, dp->hdr_type);

	switch (dp->hdr_type) {		    /* header type */
	case PCI_HEADER_TYPE_NORMAL:		    /* standard header */
		if (class == PCI_CLASS_BRIDGE_PCI)
			goto bad;

		db_setup_dma(dp);
		db_read_irq(dp);
#ifdef XXX
		pci_read_bases(dp, 6, PCI_ROM_ADDRESS);
#endif
		pci_read_config_word(dp, PCI_SUBSYSTEM_VENDOR_ID,
							 &dp->subsystem_vendor);
		pci_read_config_word(dp, PCI_SUBSYSTEM_ID, &dp->subsystem_device);

		/*
		 * If device reconfiguration is specified we call the configure
		 * bases method.  This will disable the device memory and IO
		 * and perform the discovery scan which involves writing 1's to the
		 * BARs and reading back the region size followed by assigning
		 * base addresses.  If reconfig is not to be performed we simply
		 * read in the existing BAR values and determine the size by
		 * inference or hard coded values, depending on the current platform.
		 */
		if (reconfig) {
			if (cur_cfg_ptr->config_bases_method) {
		    if ((error = cur_cfg_ptr->config_bases_method(dp, 6)))
			return (error);
			} else {
				return (-EFAULT);
			}
		} else {
			if (cur_cfg_ptr->read_bases_method) {
				if ((error = cur_cfg_ptr->read_bases_method(dp, 6, usr)))
					return (error);
			} else {
				return (-EFAULT);
			}
		}
		break;

	case PCI_HEADER_TYPE_BRIDGE:		    /* bridge header */
		if (class != PCI_CLASS_BRIDGE_PCI)
			goto bad;

#ifdef XXX
		/*
		 * The PCI-to-PCI bridge spec requires that subtractive
		 * decoding (i.e. transparent) bridge must have programming
		 * interface code of 0x01.
		 */
		pci_read_irq(dp);
		dp->transparent = ((dp->class & 0xff) == 1);
		pci_read_bases(dp, 2, PCI_ROM_ADDRESS1);
#endif

		/* This device need a patch to forward the memory cycles */
		if (dp->vendor == 0x10b5 && ((dp->device == 0x8111) ||
									 (dp->device == 0x8112))) {
			DBGPRINT("Found the PLX PCI-PCIe bridge\n");
			error = pci_write_config_byte(dp, 0x44, 0x0);
			if (error)
				DBGPRINT("Error patching the PLX\n");
		}

		if (cur_cfg_ptr->config_bases_method) {
			if ((error = cur_cfg_ptr->config_bases_method(dp, 2)))
			return (error);
		} else {
			return (-EFAULT);
		}
		break;

	default:				    /* unknown header */
		dev_err(&dp->dev, "unknown header type %02x, "
			"ignoring device\n", dp->hdr_type);
		return (-EIO);

	bad:
		dev_err(&dp->dev, "ignoring class %02x (doesn't match header "
			"type %02x)\n", class, dp->hdr_type);
		dp->class = PCI_CLASS_NOT_DEFINED;
	}

	/* We found a fine healthy device, go go go... */
	return (0);
}

/*
 * int db_scan_device()
 *
 * Description:
 *   Look for a specified PCI device. If found, read the config
 *   data for the device, sanity-check it, and  create and fill
 *   in a device data structure.
 *
 * Input(s):
 *   temp - Pointer to a partially filled data structure
 *          for the PCI device.
 *
 *   newdev - Handle to return device, if found
 *
 *   reconfig - Flags whether blade devices should be reconfigured.  On an HA
 *  		failover certain blades, e.g. Scimitar, should not be.
 *   usr  - Optional user interface structure that specifies the
 *          users callback methods and token; if NULL, the default,
 *          client-independent methods must be used;
 *
 * Output(s):
 *   newdev - A new data structure containing the device resource
 *            allocation. If no device found, or an error is detected,
 *            this will not be modified.
 *
 *   If Device was found, a pointer-to-pointer is returned in newdev
 *   to a data structure with resource allocation information
 *   for the device. This is NOT the same as the temp
 *   pointer passed into the routine, and the caller is
 *   responsible for freeing the memory when it is no
 *   longer needed.
 *
 * Returns:
 *   0       - No error occured, and device was found
 *   -ENODEV - No error occured, but no device was found
 *   Any other value - Error occured
 *
 */
static int db_scan_device(struct pci_bus *bus, int devfn,
			  struct pci_dev **newdev,
			  int reconfig, db_usr_if_t *usr)
{
	struct pci_dev *dp;
	u32 l;
	int status = 0;

	if (!newdev) {
		printk(KERN_ERR "DB %s: Invalid input parameters\n", __FUNCTION__);
		return (-EINVAL);
	}

	*newdev = NULL;

	if ((status = pci_bus_read_config_dword(bus, devfn, PCI_VENDOR_ID, &l))) {
		printk(KERN_ERR "%s: failed to read PCI_VENDOR_ID dev %d bus %d, "
			"ret=%d\n", __FUNCTION__, PCI_SLOT(devfn),
			bus->number, status);
		return (status);
	}

	/* some broken boards return 0 or ~0 if a slot is empty: */
	if (l == 0xffffffff || l == 0x00000000 ||
		l == 0x0000ffff || l == 0xffff0000) {
		return (-ENODEV);
	}

	dp = alloc_pci_dev();
	if (!dp) {
		printk(KERN_ERR "%s: alloc_pci_dev(%d) failed\n",
			   __FUNCTION__, sizeof (*dp));
		return (-ENOMEM);
	}

	dp->bus = bus;
	dp->devfn = devfn;
	dp->vendor = l & 0xffff;
	dp->device = (l >> 16) & 0xffff;

	status = db_setup_device(dp, reconfig, usr);
	if (status < 0) {
		DBGPRINT("db_scan_device Discarding device node\n");
		kfree(dp);
	} else {
		*newdev = dp;
	}

  	return (status);
}
static int db_scan_single_device(struct pci_bus *bus, int devfn,
								 struct pci_dev **newdev,
								 int reconfig, db_usr_if_t *usr)
{
	struct pci_dev *dp;
	int error;

	dp = pci_get_slot(bus, devfn);
	if (dp) {
		*newdev = dp;
		dev_dbg(&dp->dev, "%s: device exists\n", __FUNCTION__);
		return (0);
	}

	error = db_scan_device(bus, devfn, newdev, reconfig, usr);
	if (error)
		return (error);

	pci_device_add(*newdev, bus);
	return (0);
}

/*
 * int db_device_exists()
 *
 * Description:
 *
 *   Determine if a specified PCI device already exists
 *   in the global PCI device list.
 *
 *   A device has been found by the scan, and we need to
 *   know if we already know about it. If the device is
 *   already in the global PCI list, then it has already
 *   been scanned and we should just ignore it this time.
 *
 *   Basic assumption is that if a device was physically
 *   removed, it will have been removed from the global
 *   PCI list prior to calling db_scan_drawbridge(). Also
 *   assume that the bus and device number combination is
 *   unique, although we do check a few other things.
 *
 * Input(s):
 *   new_dp - Pointer to data structure defining the
 *            PCI device that was just found.
 *   bus    - Number of the PCI bus attached to the device
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   0 - The device is already in the global PCI list
 *   Non 0 - The device is not in the global PCI list
 *
 */
// XXX delete!
int
db_device_exists(struct pci_dev *new_dp, unsigned char bus)
{
  int exists = 0;
  struct pci_dev *dp = NULL;

  if (!new_dp) {
    printk(KERN_ERR "\ndb_device_exists: Invalid Device pointer\n");
    return (0);
  }

  dp =  pci_get_domain_bus_and_slot(cur_cfg_ptr->pci_domain,
									bus, new_dp->devfn);
  if (dp) {
	  pci_dev_put(dp); /* Keep reference count sane */
	  if ((dp->vendor == new_dp->vendor)&&(dp->device == new_dp->device)&&
		  (dp->hdr_type == new_dp->hdr_type)&&(dp->class == new_dp->class)) {
		  exists = 1;
	  }

  }

  return (exists);

}
/*
 * int db_scan_slot()
 *
 * Description:
 *   Check for the existance of 1 or more PCI functions
 *   at a specified PCI slot. If a function exists, scan
 *   the device to configure and/or record its resource
 *   usage. Check if the device is already in the global
 *   PCI device list, and add it if it is not there.
 *
 * Input(s):
 *   temp - Data structure defining the PCI device slot
 *          to be scanned.
 *
 *   newdev - Handle to return device, if found
 *
 *   reconfig - Flags whether blade devices should be reconfigured.  On an HA
 *  		failover certain blades, e.g. Scimitar, should not be.
 *   usr  - Optional user interface structure that specifies the
 *          users callback methods and token; if NULL, the default,
 *          client-independent methods must be used;
 *
 * Output(s):
 *   newdev - A new data structure containing the device resource
 *            allocation. The new PCI device data structure is also
 *            added to the global PCI device list.
 *
 *            If an error is returned, then newdev is not modified.
 *
 * Returns:
 *   0    - No errors, and 1 or more new PCI functions were found.
 *   <0   - An error occured, and no valid PCI function was found
 *
 *          Can check contents of primary_status
 *          and secondary_status members of the db_t struct
 *          to see if SERR or parity caused the error.
 *
 */
static int db_scan_slot(struct pci_bus *bus, int devfn,
			struct pci_dev **newdev,
			int reconfig, db_usr_if_t *usr)
{
	struct pci_dev *dp;
	struct pci_dev *first_dev = NULL;
	int fn, error = 0;

	if (!newdev)
		return (-EINVAL);

	*newdev = NULL;

	error = db_scan_single_device(bus, devfn, &dp, reconfig, usr);
	if (error) {
		if (error == -ENODEV)
			error = 0;
		else {
			printk(KERN_ERR "%s: db_scan_single_device() returned %d "
				"possible HW problem dev %d bus %d",
				__FUNCTION__, error,
				PCI_SLOT(devfn), bus->number);
			return (error);
		}
	}
	first_dev = dp;

	if (dp && dp->multifunction) {
		for (fn = 1; fn < 8; fn++) {
			error = db_scan_single_device(bus, devfn + fn, &dp, reconfig, usr);
			if (error) {
			/*
			 * The -ENODEV case is not an error
			 * so don't propagate it out of this
			 * function. Otherwise, large amounts
			 * of other code has to change also
			 */
				if (error == -ENODEV) {
					error = 0;
					continue;
				}

				printk(KERN_ERR "%s: db_scan_single_device() returned %d "
					"possible HW problem dev %d bus %d",
					__FUNCTION__, error,
					PCI_SLOT(devfn), bus->number);

				/* Possible hardware error, don't continue */
				return (error);
			}

			if (!dp) { /* Probable software error, don't continue */
				printk(KERN_ERR "\n%s: dp ptr is NULL", __FUNCTION__);
				return (-EINVAL);
			}

		/*
		 * See if the device is already in the global PCI
		 * device chain. If it is, discard this node.
		 */
			if (db_device_exists(dp, bus->number)) {
				printk(KERN_ERR "This device already exists, db_device_exists "
					   "got called\n");
				BUG();
                kfree(dp);
				dp = NULL;
				continue;
			}

			DBGPRINT("  AddToList:  Bus %2d, device %3d, function %2d:\n",
					 dp->bus->number, PCI_SLOT(dp->devfn),
					 PCI_FUNC(dp->devfn));

			dp->multifunction = 1;
		}
	}

	if (first_dev) {
		*newdev = first_dev;
	}

	return (error);
}
/*
 * int db_do_scan_bus()
 *
 * Description:
 *   Scan a specified PCI bus to find any devices
 *   attached to the bus.
 *
 * Input(s):
 *   bus - Data structure defining the PCI bus.
 *   *maxbus - Max PCI bus found so far
 *   reconfig - Flags whether blade devices should be reconfigured.  On an HA
 *  			failover certain blades, e.g. Scimitar, should not be.
 *   usr  - Optional user interface structure that specifies the
 *          users callback methods and token; if NULL, the default,
 *          client-independent methods must be used;
 *
 * Output(s):
 *   *maxbus - Updated to contain the max PCI bus found
 *             downstream of this bus.
 *
 *   Any new PCI devices or busses found are added to
 *   the global PCI device and bus list.
 *
 * Returns:
 *   < 0  = Scan fail
 *     0  = Scan Passed
 *
 */
static int db_scan_child_bus(struct pci_bus *bus, unsigned int *maxbus,
							int reconfig, db_usr_if_t *usr)
{
	unsigned int devfn, max = bus->secondary;
	struct pci_dev *dp;
	int error = 0;

	pr_debug("DB: Scanning bus %04x:%02x (secondary = %d)\n",
		 pci_domain_nr(bus), bus->number, bus->secondary);

	/*
	 * Go find them, Rover!
	 *
	 * It may be OK if db_scan_slot() doesn't find
	 * anything - might be just an empty bus. But
	 * a non-zero error code means some kind of
	 * error occured while scanning. In that case,
	 * bail out now because something
	 * is wrong with this bus or its bridge.
	 *
	 */
	for (devfn = 0; devfn < 0x100; devfn += 8) {
		error = db_scan_slot(bus, devfn, &dp, reconfig, usr);
		if (error) {
			pr_err("DB: db_scan_child_bus: Call to db_scan_slot "
				 "failed %d  DevId: 0x%4.4x Bus: 0x%4.4x\n",
				 error, PCI_SLOT(devfn), bus->number);
			return (error);
		}
	}

	/*
	 * After performing arch-dependent fixup of the bus, look behind
	 * all PCI-to-PCI bridges on this bus.
	 */
	list_for_each_entry(dp, &bus->devices, bus_list) {
		switch (dp->class >> 8) {
		case PCI_CLASS_BRIDGE_PCI:
			pr_debug("DB: Scanning bridge 0x%x:0x%x at bus %d\n", dp->vendor,
					 dp->device, bus->number);
			error = db_scan_bridge(
						bus, dp, &max, DB_RECONFIG, 0, bus->number, usr);
			if (error)
			  return (error);
			break;
		}
	}

	/*
	 * We've scanned the bus and so we know all about what's on
	 * the other side of any bridges that may be on this bus plus
	 * any devices.
	 *
	 * Return how far we've got finding sub-buses.
	 */
	pr_debug("PCI: Bus scan for %04x:%02x returning with max=%02x\n",
		pci_domain_nr(bus), bus->number, max);

	/* Don't update the maxbus if anything failed */
	if (!error)
	  *maxbus = max;

	return (error);
}

/*
 * struct pci_bus *db_bus_exists()
 *
 * Description:
 *   Search a specified PCI bus list to see if it
 *   contains a specified bus.
 *
 * Input(s):
 *   bus - Parent bus containing the PCI bus list.
 *   nr  - Number of the PCI bus to search for.
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   NULL - The specified bus is not in the list
 *   NON-NULL - The bus is in the list; a pointer to
 *              the bus data structure is returned.
 *
 */
struct pci_bus *
db_bus_exists(struct pci_bus *bus, int nr)
{
	return (pci_find_bus(pci_domain_nr(bus), nr));
}

/*
 * unsigned int db_scan_bus()
 *
 * Description:
 *   Scan a "top-level" PCI bus to find and configure
 *   all downstream devices and busses.
 *
 *   This routine is only called when a DrawBridge is found.
 *   All downstream busses are scanned by db_do_scan_bus()
 *   Without coming through here.
 *
 * Input(s):
 *   db - Pointer to a structure defining the parent
 *        PCI bridge.
 *   ops - Structure containing function pointers to
 *         the PCI operations such as read/write
 *   sysdata - Private instance data
 *
 * Output(s):
 *   The global PCI device and bus lists are updated
 *   to contain all new devices and busses found
 *   downstream of this bridge.
 *
 * Returns:
 *   < 0 = PCI scan fail
 *     0 = PCI scan pass
 */
static int
db_scan_bus(struct pci_dev *db,
		       struct pci_ops *ops,
		       void *sysdata,
		       unsigned int *maxbus)
{
	/*
	 * parent is the bus instance on the primary
	 * side of the bridge.
	 */

	struct pci_bus *parent = db->bus, *child = NULL;
	int error = 0;
	unsigned int max;
	db_t *dbp = (db_t *)sysdata;

	if (!dbp) {
		return (-1);
	}

	max = dbp->root_bus_number;

	printk("%s: Parent Bus %d  Parent Secondary %d  "
		   "Parent Sub %d\n", __FUNCTION__, parent->number,
		   parent->secondary, parent->subordinate);

	if (!(child = db_bus_exists(parent, max))) {

		DBGPRINT("db_scan_bus2: Adding new bus %d\n", max);
		child = db_add_new_bus(parent, db, max);
	}

	if (child) {
		dev_set_drvdata(&child->dev, sysdata);
	  child->ops = ops;

	  error = db_scan_child_bus(child, &max, DB_RECONFIG, (db_usr_if_t *)NULL);
		if (!error)
			pci_bus_add_devices(child);
	}

	if (error)
	  printk(KERN_ERR "\nFail to scan behind Drawbridge...\n");
	else {
		/* coverity[var_deref_op] */
	  child->subordinate = max;
	  *maxbus = max;
	}

	return (error);
}

/*
 * uint32_t db_get_dma_membase()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 *
 */
uint32_t
db_get_dma_membase(int bridge, int virtual)
{

    return (-ENOSYS);

}
/*
 *
 * uint32_t db_get_dma_pcibase()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
uint32_t
db_get_dma_pcibase(int bridge)
{
    return (-ENOSYS);
}

/*
 * int db_get_dma_pcibase()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
int
db_set_doorbell(u_int door, int bridge, int local, int state)
{
    return (-ENOSYS);
}
/*
 * uint32_t db_get_dma_memsize()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
uint32_t
db_get_dma_memsize(int bridge)
{
    return (-ENOSYS);
}

/*
 * int db_mailbox_get()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
int
db_mailbox_get(int bridge, int local, int mailbox, int *data)
{
    return (-ENOSYS);
}

/*
 * int db_mailbox_send()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
int
db_mailbox_send(int bridge, int local, int mailbox, int data)
{
    return (-ENOSYS);
}

/*
 * int db_get_doorbell()
 *
 * Description:
 *
 *  Obsolete function. Interface is retained until
 *  all calls from other modules are removed.
 *
 * Input(s):
 *  N/A
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  -ENOSYS
 */
int
db_get_doorbell(int bridge, int local, u16 *irq)
{
    return (-ENOSYS);
}

/*
 * int db_secondary_mem_enable()
 *
 * Description:
 *
 *  Enable or disable PCI memory transactions on the drawbridges'
 *  secondary side interfaces.  Performs this operation on all known
 *  drawbridges for the current platform.  Note that we need to disable
 *  master capability prior to making a change to the memory setting.
 *
 * Input(s):
 *  enable	= 1 for enable, 0 for disable
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  0 on success
 *  < 0 on failure
 */
int
db_secondary_mem_enable(int enable)
{
	uint16_t cmd;
	int i, ret, db_devnum, primary_bus;
	struct pci_dev *dp;


	if (cur_cfg_ptr == NULL) {
		printk(KERN_ERR "%s: DB module not initialized!\n", __FUNCTION__);
		return (-1);
	}

	for (i = 1; i <= cur_cfg_ptr->num_db; i++) {

		db_devnum = db_idx_to_devnum(i);
		if (db_devnum < 0) {
			printk(KERN_ERR "\n%s: Invalid bridge index %d\n",
				   __FUNCTION__, i);
			continue;
		}

		primary_bus = cur_cfg_ptr->primary_busses[i];

		dp = db_get_device(db_devnum, primary_bus);
		if (dp == NULL) {
			printk(KERN_ERR "%s: DB %d not found!\n", __FUNCTION__, i);
			continue;
		}

		/*
		 * Read in the secondary side command register and
		 * set memory bit according to whether we're enabling or disabling.
		 */
		ret = pci_read_config_word(dp, DB_PCI_SCOMMAND, &cmd);
		if (ret) {
			printk(KERN_ERR "%s: failed to read cmd reg, db=%d ret=%d\n",
				__FUNCTION__, i, ret);
			return (ret);
		}

		/* Disable master capability if enabled */
		if (cmd & PCI_COMMAND_MASTER) {
			cmd &= ~PCI_COMMAND_MASTER;

			ret = pci_write_config_word(dp, DB_PCI_SCOMMAND, cmd);

			if (ret) {
				printk(KERN_ERR
					"%s: failed to disable mastership, db=%d ret=%d\n",
					__FUNCTION__, i, ret);
				return (ret);
			}

			cmd |= PCI_COMMAND_MASTER;
		}

		if (enable)
			cmd |= PCI_COMMAND_MEMORY;
		else
			cmd &= ~PCI_COMMAND_MEMORY;

		ret = pci_write_config_word(dp, DB_PCI_SCOMMAND, cmd);
		if (ret) {
			printk(KERN_ERR
				"%s: failed to write updated mem setting, db=%d ret=%d\n",
				__FUNCTION__, i, ret);
			return (ret);
		}

	} /* for DB's */

	return (0);
}

/*
 * int db_is_secondary_mem_enabled()
 *
 * Description:
 *
 *  Called to determine if DB secondary side memory accesses are enabled.
 *  Since we enable/disable all DB's as a group, this function simply checks
 *  the secondary side command register for the first DB to make this
 *  determination.
 *
 * Input(s):
 * 	dp: passed in pci device.
 *
 * Output(s):
 *  None
 *
 * Returns:
 *  0 memory is not enabled
 *  > 0 memory is enabled
 */
int
db_is_secondary_mem_enabled(struct pci_dev *dp)
{
	uint16_t cmd;
	int ret;


	if (cur_cfg_ptr == NULL) {
		printk(KERN_ERR "%s: DB module not initialized!\n", __FUNCTION__);
		return (0);
	}

	if (dp == NULL) {
		printk(KERN_ERR "%s: DB 1 not found!\n", __FUNCTION__);
		return (0);
	}

	/* Read in the secondary side command register and see if mem bit is on */
	ret = pci_read_config_word(dp, DB_PCI_SCOMMAND, &cmd);
	if (ret) {
		printk(KERN_ERR "%s: failed to read cmd reg, db=1 ret=%d\n",
				__FUNCTION__, ret);
		return (ret);
	}

	return ((cmd & PCI_COMMAND_MEMORY) != 0);
}

/*
 * int db_init_module()
 *
 * Description:
 *   Module initializer. Called by the Linux kernel
 *   when the module is loaded. Registers db_driver as
 *   the pci_driver data structure containing routines for
 *   the kernel to call to probe for and remove PCI
 *   devices.
 *
 * Input(s):
 *   None
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   0 - Success
 *   Non 0 - Failed
 *
 */
int __init
db_init_module(void)
{
	int status = 0;
	int ret;
	int i;
	sw_bd board_info;
	struct list_head *list;
	struct pci_bus *child = NULL;

	/* Read the current platform SWBD number */
	if ((ret = sin_get_swbd(&board_info)) != 0) {
		printk(KERN_ERR "Failed to read board ID, ret=%d\n", ret);
		return (-ENODEV);
	}

	DBGPRINT("db_init_module: swbd=%d\n", board_info.board);

	/* Configure platform specific parameters based on platform ID. */
	cur_cfg_ptr = NULL;
	for (i = 0; i < db_cfg_tbl_sz; i++) {
		if (platform_cfg_tbl[i]->swbd == board_info.board) {
			cur_cfg_ptr = platform_cfg_tbl[i];
			break;
		}
	}

	if (cur_cfg_ptr == NULL) {
		printk(KERN_ERR "Unsupported platform %d\n", board_info.board);
		return (-ENODEV);
	}

	/*
	 * DEFECT000045112
	 * Initialize DB reset interface on CPU.  This is done to
	 * avoid blocking in the reset function which may be invoked
	 * from a bottom-half.
	 *
	 * Free any resources during module unload.
	 */
    db_primary_reset_init();

	DBGPRINT("db_init_module: Calling pci_register_driver\n");

	status |= pci_register_driver(&db_driver);

	/*
	 * Save the current subordinate bus number of the root
	 * bus so it can be restored when db-module is unloaded.
	 * The root bus (bus 0) should be the first (and only??)
	 * bus directly attached to pci_root_buses by the Linux
	 * kernel
	 */

	list_for_each(list, &pci_root_buses) {

	    child = pci_bus_b(list);
	    break;

	}

	if (child) {

	    DBGPRINT("db_init_module: Saving Root subordinate bus (0x%x)\n",
		     child->subordinate);
	    cur_cfg_ptr->cp_subordinate_bus =  child->subordinate;

	}

	DBGPRINT("db_init_module: Return from pci_module_init\n");

	return (status);
}

/*
 * void db_exit_module()
 *
 * Description:
 *   Called when the module is removed.
 *   Registers db_driver as the pci_driver data
 *   structure containing routines for the kernel
 *   to call to remove PCI devices.
 *
 * Input(s):
 *   None
 *
 * Output(s):
 *   None
 *
 * Returns:
 *   N/A
 *
 */
void __exit
db_exit_module(void)
{
    struct list_head *list;
    struct pci_bus *child = NULL;

    DBGPRINT("db_exit_module: Entered\n");

    pci_unregister_driver(&db_driver);

	/*
	 * DEFECT000045112
	 * uninitialize primary reset interface
	 */

	db_primary_reset_uninit();

    /*
     * Find the structure for PCI bus 0, which
     * should be the first (and only??) element
     * directly attached to pci_root_buses by
     * the Linux kernel
     */
    list_for_each(list, &pci_root_buses) {

	child = pci_bus_b(list);
	break;

    }

    /*
     * If the root bus is found, restore the subordinate bus
     * number to whatever it was when this module was loaded.
     * This is being done to enable db-module to be unloaded
     * and re-loaded without a reboot in between.
     *
     * WARNING: This is very platform specific. The assumption is
     *          that any PCI busses found after db-module loaded
     *          were found by db-module itself.
     *
     */

    if (child) {

	DBGPRINT("db_exit_module: Restore Root subordinate bus "
		 "from 0x%x to 0x%x\n", child->subordinate,
		 cur_cfg_ptr->cp_subordinate_bus);
	child->subordinate = cur_cfg_ptr->cp_subordinate_bus;

    }

    DBGPRINT("db_exit_module: Exiting\n");

}

/*
 * Linux initcall declarations
 */
module_init(db_init_module)
module_exit(db_exit_module)

MODULE_LICENSE("GPL");

EXPORT_SYMBOL(db_reset_drawbridge);
EXPORT_SYMBOL(db_scan_drawbridge);
EXPORT_SYMBOL(db_remove_tree);
EXPORT_SYMBOL(db_get_csrbase);
EXPORT_SYMBOL(db_get_irq);
EXPORT_SYMBOL(db_get_forwarding_window);
EXPORT_SYMBOL(db_pci_primary_reset);
EXPORT_SYMBOL(db_drawbridge_secondary_reset);
EXPORT_SYMBOL(db_bridge_secondary_reset);
EXPORT_SYMBOL(db_pci_chip_reset);
EXPORT_SYMBOL(db_remove_child_devices);
EXPORT_SYMBOL(db_scan_blade);
EXPORT_SYMBOL(db_scan_blade_noreconfig);
EXPORT_SYMBOL(db_bus_exists);
EXPORT_SYMBOL(db_set_cache_line_size);
EXPORT_SYMBOL(db_secondary_reset_control);
EXPORT_SYMBOL(db_secondary_mem_enable);
EXPORT_SYMBOL(db_is_secondary_mem_enabled);
EXPORT_SYMBOL(db_get_drawbridge_status);
EXPORT_SYMBOL(db_timeout_restore_drawbridge);
EXPORT_SYMBOL(db_add_drawbridge_subordinate);
EXPORT_SYMBOL(db_dump_csr);
EXPORT_SYMBOL(db_register_find_db);

EXPORT_SYMBOL(db_get_dma_membase);
EXPORT_SYMBOL(db_get_dma_memsize);
EXPORT_SYMBOL(db_get_dma_pcibase);
EXPORT_SYMBOL(db_set_doorbell);
EXPORT_SYMBOL(db_get_doorbell);
EXPORT_SYMBOL(db_mailbox_get);
EXPORT_SYMBOL(db_mailbox_send);
