#ident "@(#) $Id$"

/*
 *
 *    Copyright (c) 1996-2013 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *	Header for Intel/DEC 21554 non-transparent PCI-to-PCI bridge
 *	(aka Drawbridge) device driver.
 *
 */

#ifndef __DB_H_PRIVATE__
#define	__DB_H_PRIVATE__

#ifdef __cplusplus
extern "C" {
#endif

#include <linux/db.h>

#if !defined(PCI_DEVICE_ID_DEC_21554)
#define	PCI_DEVICE_ID_DEC_21554	0x0046
#endif


	/*
	 * This is temporary until Linux is moved
	 * into ClearCase. No one can see updates to
	 * the Linux include files without a great
	 * deal of pain
	 */
#define GPIO0_OR	0xef600700
#define GPIO0_TCR	0xef600704
#define GPIO0_ODR	0xef600718
#define GPIO0_IR	0xef60071c

	/*
	 * These are defined in linux/include/asm-ppc/processor.h,
	 * but are only seen if compiling for PPC.
	 * So if we aren't compiling for a PPC, define them here
	 * so people can continue to compile for Intel platforms
	 *
	 */
#ifndef CONFIG_PPC
#define DCRN_CHCR0	0x0B1		/* Chip Control Register 1  */
#define GPIO0_OR	0xef600700
#define GPIO0_TCR	0xef600704
#define GPIO0_ODR	0xef600718
#define GPIO0_IR	0xef60071c
#endif

#ifndef FABOSBUILD
#define CP_PCIPS_IRQ3		28	/* PCI/PS Changes */
#endif

/* interrupt control macros */
#define SYS_INT_SAVE()			local_irq_save(flags);
#define SYS_INT_RESTORE()		local_irq_restore(flags)

//#define	DB_COUNT_MAX		2 /* Max number of drawbridges */
#define DB_BRIDGE_BUS_MASK	0xFF000000
 
#define WINDOW_START_IDX	2 /* Forwarding window start index */
#define WINDOW_STOP_IDX		4 /* Forwarding window stop index  */

#define MAX_SLOTS_PER_DB	4 /* max num chassis slots per drawbridge */

#define DB_MAX_BARS			6 /* max num BAR registers per device */

/*
 * For saving DrawBridge config registers
 * when a save-reset-restore needs to
 * be done.
 */
typedef struct db_cfg_save_s {

	uint32_t pbase_address0;  /* Primary BARs */
	uint32_t pbase_address1;
	uint32_t pbase_address2;
	uint32_t pbase_address3;
	uint32_t pbase_address4;
	uint32_t pbase_address5;

	uint32_t sbase_csr_mem;  /* Secondary BARs */
	uint32_t sbase_csr_io;
	uint32_t sbase_address0;
	uint32_t sbase_address1;
	uint32_t sbase_address2;

	uint32_t ds_tb0; /* Downstream translated base registers */
	uint32_t ds_tb1;
	uint32_t ds_tb2;
	uint32_t ds_tb3;

	uint32_t us_tb0; /* Upstream translated base registers */
	uint32_t us_tb1;

	uint32_t reset_ctrl;

	uint16_t primary_command;
	uint16_t secondary_command;

	uint16_t primary_status;
	uint16_t secondary_status;

	uint16_t chip_ctrl0;
	uint16_t chip_ctrl1;

	uint16_t config_ctrl;

	uint8_t  primary_latency;
	uint8_t  secondary_latency;

	uint8_t	irq;
	uint8_t	cache_line;
	uint8_t	secondary_cache_line;

} db_cfg_save_t;

/*
 * Various things needed by the drawbridge
 * scan routines.
 *
 * One of these is created for and 'owned by' each drawbridge.
 * The struct is attached to the driver_data member of the drawbridge
 * pci_dev structure.
 *
 * Each PCI device found behind a drawbridge links to this
 * structure via the sysdata member of each device's pci_dev
 * structure.
 *
 */

typedef struct db_s {

	/* Points to the owning drawbridge instance */


	void		*db_dev;

	/* Points to local bridge CSR registers */


	db_csr_t	*db_local_csr;

	/* Points to remote bridge (on other CP) CSR registers */


	db_csr_t	*db_remote_csr;

	/*
	 * Virtual address and size of the PCI downstream
	 * forwarding window used to write into an area of
	 * the other CP's DRAM
	 */

	uint32_t	*reserved_mem_pci_vbase;
	uint32_t	reserved_memsize;

	void		*db_data;
	dword		window_start;
	dword		window_stop;
	dword		lowest_mem_addr;
	dword		root_bus_number;
	dword		primary_bus_number;
	int			blade_slot;


	/* To enumerate multiple bridges on the BP */
	int			cur_bus_index;
	dword		cur_mem_addr;

	/* drawbridge status registers */

	uint16_t	primary_status;
	uint16_t	secondary_status;

	db_cfg_save_t	cfg_save;

	uint8_t		usable;
	uint8_t		remap_memory;

	/*
	 * When allocating just the amount of PCI memory the dev asks for, 
	 * we need to keep track of the next available offset.
	 */
	uint32_t	next_slot_offset[MAX_SLOTS_PER_DB];

	/*
	 * When allocating mem a dev asks for, we reserve the end of a
	 * blade's allocation range for small BAR requests so as not to
	 * fragment memory too much due to alignment adjustments.
	 */
	uint32_t	small_bar_alloc_off[MAX_SLOTS_PER_DB];

	uint32_t	blade_bridge_refcount[MAX_SLOTS_PER_DB];

} db_t;

typedef struct pci_dev pci_dev_t;

/* Define a known device's (as ID'd by vendor and device IDs) BAR sizes */
typedef struct db_known_dev_s {
	unsigned short vendor;
	unsigned short device;
	unsigned long bar_sz[DB_MAX_BARS];
	unsigned long bar_attr[DB_MAX_BARS];
} db_known_dev_t;

/*
 * Specifies whether (non-bridge) device memory allocation should be done
 * on a fixed region basis where the same amount of memory is allocated
 * regardless of what the devices requests or dynamically determined by
 * respecting what the device requests.
 */
typedef enum {
	DB_ALLOC_METHOD_FIXED = 1,	/* use same amount for all devices */
	DB_ALLOC_METHOD_DYNAMIC		/* use what the device asks for */
} db_alloc_t;

/*
 * Holds platform specific configuration parameters and methods.  One instance
 * is instantiated for each supported platform.
 */

typedef int (*db_alloc_fn_t)(struct pci_dev *dp, struct resource *res,
				       u32 sz, u32 *base);
typedef int (*db_fixup_fn_t)(void);
typedef int (*db_config_bases_fn_t)(struct pci_dev *dp, unsigned int howmany);
typedef int (*db_read_bases_fn_t)(
				struct pci_dev *dp, unsigned int howmany, db_usr_if_t *usr);
typedef int (*db_setup_dma_fn_t)(void);
typedef int (*db_reset_fn_t)(void);
typedef int (*db_primary_reset_init_fn_t)(void);
typedef int (*db_primary_reset_uninit_fn_t)(void);
typedef unsigned int (*db_compute_bus_number_fn_t) (struct pci_bus *bus,
						  struct pci_dev *dp,
						  unsigned int *maxbus, int is_blade_root_bus,
						  int root_bus_number, db_t *dbp);

typedef struct db_platform_cfg_s {
	unsigned int	swbd;		/* SWBD platform ID */
	int				num_db;		/* number of drawbridges */
	int				busses_per_slot; /* num buses to alloc per chass slot */

	int		irq_num;	/* internal IRQ num to pgm in DBs */

    /*
     * Retains the subordinate PCI bus as it was defined
     * prior to doing any scanning behind the drawbridges.
     * It needs to be saved so that it can be restored
     * if all blades are hot-removed. Otherwise, when blades
     * are put back in the system, they will be assigned the
     * wrong bus numbers. This value is obtained from Linux
     * at db-module init time.
     */
	int		cp_subordinate_bus;

    /*
     * Where in the CPU memory map do PCI physical addresses start
     * This is a 36-bit value for 440 processors, and the exact value
     * depends on how the boot monitor has set up the TLB mapping.
     * For 405 processors, this value is 0.
     */

	phys_addr_t	physical_pci_base;

    /*
     * The offset of the beginning of PCI address space assigned to
     * the drawbridges. This is a 32 bit value. For 405 processors
     * this value is always 0. For 440 processors, it is the first
     * 32-bit physical base address assigned by the boot monitor
     * to a drawbridge (DB 1).
     */

	unsigned long 	physical_db_base;

    /*
     * Holds the amount of address space allowed per device
     * When using the fixed allocation method
     */

	uint32_t 	device_alloc_sz;

    /*
     * Holds the amount of address space allowed per blade slot.
     * Used for both fixed and dynamic allocation methods
     */

	uint32_t	blade_alloc_sz;	/* Slot allocation size in bytes */

    /*
     * Needed on systems where the drawbridge reset could
     * occur when there is no secondary side clock.
     */
	uint32_t 	fwd_win_offset;		/* forwarding win offset from base */
	uint32_t 	const  *base_addr;	/* ptr to array of base addrs */
	uint32_t	*gpio_base;			/* virt mem base addr of gpio registers */

    /* Methods */

	db_fixup_fn_t		db_fixup_method; /* Need to check for reset problems */

	db_config_bases_fn_t	config_bases_method; /* For read/config of BARs */
    db_read_bases_fn_t		read_bases_method; /* For reading of cur BARs */

    /*
     * Used on systems where the ASICs can be DMA masters
     */
	db_setup_dma_fn_t	setup_dma_method; /* For CPU<->ASIC DMA setup */

	db_reset_fn_t		reset_method;	/* ptr to func to reset all DBs */

	db_alloc_fn_t		alloc_method; /* device allocation method */

	/* Platform-specific method to calculate blade bus numbers */
	db_compute_bus_number_fn_t	compute_bus_method;
	/*
	 * Latency timer values for the non-transparent (drawbridge) and
	 * transparent bridges' primary and secondary sides.
	 */
	uint8_t ntb_lt_primary;		/* non-transparent bridge primary bus */
	uint8_t ntb_lt_secondary;	/* non-transparent bridge secondary bus */
	uint8_t tb_lt_primary;		/* transparent bridge primary bus */
	uint8_t tb_lt_secondary;	/* transparent bridge secondary bus */

	/* Flags whether DB secondary side memory should be initially enabled */
	int enable_secondary_mem;

	/* PCI domain number. This is the PCI/PCIe controller instance */
	int pci_domain;

	/*
	 * Table containing the db secondary-side bus numbers. These form the
	 * root bus from which all downstream bus numbers are derived for
	 * busses downstream of the db.
	 */
	int const *root_busses;
	int const *primary_busses;

	/*
	 * For the case where device BARs should only be read and not configured,
	 * we need to know the sizes beforehand.  The following points to the
	 * list of known devices to be consulted.
	 */
	db_known_dev_t *known_devs;
} db_platform_cfg_t;

struct pci_dev *db_get_device(int devnum, int bus);
void db_save_bridge_config(struct pci_dev *dp, db_cfg_save_t *save);
void db_remove_drawbridge(pci_dev_t *dp);

int db_primary_reset_init(void);
int db_configure_bases(struct pci_dev *dp, unsigned int howmany);
int db_read_bases(struct pci_dev *dp, unsigned int howmany, db_usr_if_t *usr);
int db_dynamic_address_allocate(struct pci_dev *dp,
								struct resource *res,
								u32 sz, u32 *base);
void db_restore_bridge_config(struct pci_dev *dp, db_cfg_save_t *save);
#ifdef __cplusplus
}
#endif

#endif /* __DB_H_PRIVATE__ */
