/*
 * (C) Copyright 2003-2004
 * Humboldt Solutions Ltd, adrian@humboldt.co.uk.

 * This is a combined i2c adapter and algorithm driver for the
 * MPC107/Tsi107 PowerPC northbridge and processors that include
 * the same I2C unit (8240, 8245, 85xx).
 *
 * Release 0.8
 *
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/of_platform.h>
#include <linux/of_i2c.h>
#include <linux/slab.h>

#include <linux/io.h>
#include <linux/fsl_devices.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
// #include <linux/sysctl.h>
#include <asm/uaccess.h>

#ifdef CONFIG_SILKWORM
#include <brocade/cpld.h>
#include <linux/time.h>
#include "i2c-stats.h"
#endif

#include <asm/mpc52xx.h>
#include <sysdev/fsl_soc.h>

#define DRV_NAME "mpc-i2c"

#define MPC_I2C_CLOCK_LEGACY   0
#define MPC_I2C_CLOCK_PRESERVE (~0U)

#define MPC_I2C_FDR   0x04
#define MPC_I2C_CR    0x08
#define MPC_I2C_SR    0x0c
#define MPC_I2C_DR    0x10
#define MPC_I2C_DFSRR 0x14

#define CCR_MEN  0x80
#define CCR_MIEN 0x40
#define CCR_MSTA 0x20
#define CCR_MTX  0x10
#define CCR_TXAK 0x08
#define CCR_RSTA 0x04

#define CSR_MCF  0x80
#define CSR_MAAS 0x40
#define CSR_MBB  0x20
#define CSR_MAL  0x10
#define CSR_SRW  0x04
#define CSR_MIF  0x02
#define CSR_RXAK 0x01

#ifdef CONFIG_SILKWORM
#define IIC_TIMEOUT		(HZ/4)		/* 250 msec */
#else
#define IIC_TIMEOUT		(HZ)
#endif
#define IIC_RETRY		2

#ifdef CONFIG_SILKWORM
int core_i2c_debug = 0;
#endif

// int i2cTurboCnfg = 1;
int i2cTurboFlag = 0;

#if 0
/*
 * Comment this out for now since these parameters
 * i2cTurboCnfg, i2cTurboFlag are not used at user space
 */
static unsigned one_time_on = 0;
static unsigned one_time_off = 0;
static ctl_table i2c_tab[]={
	{ 18, "i2cTurboFlag", &i2cTurboFlag,
		sizeof(int), 0644, NULL, &proc_dointvec},
	{ 19, "i2cTurboCnfg", &i2cTurboCnfg,
		sizeof(int), 0644, NULL, &proc_dointvec},
	{ 20, "i2c_debug", &core_i2c_debug, sizeof(int), 0644, NULL,
&proc_dointvec},
	{0}
};

static ctl_table i2c_root_tab[]={
	{CTL_DEBUG, "debug", NULL, 0, 0555, i2c_tab},
	{0}
};
#endif

/* Types to handle at interrupt context */
enum { OTHER, BASIC_READ, BASIC_WRITE };


struct mpc_i2c {
	struct device *dev;
	void __iomem *base;
	u32 interrupt;
	wait_queue_head_t queue;
	struct i2c_adapter adap;
	int irq;
	u8 *data;
	u8 cmd;		/* OTHER, BASIC_READ, BASIC_WRITE */
	u16 count;	/* bytes to read */
};

struct mpc_i2c_divider {
	u16 divider;
	u16 fdr;	/* including dfsrr */
};

struct mpc_i2c_data {
	void (*setup)(struct device_node *node, struct mpc_i2c *i2c,
		      u32 clock, u32 prescaler);
	u32 prescaler;
};

static inline void writeccr(struct mpc_i2c *i2c, u32 x)
{
	writeb(x, i2c->base + MPC_I2C_CR);
}

/* process a byte read from the interrupt service routine */
static void byteread(struct mpc_i2c *i2c)
{
	int l=i2c->count;
	if (l <= 1) {	/* should only happen with l == 1, but lets be defensive */
		i2c->cmd = OTHER;
		/* switch off master bit, causes stop to be sent */
		// writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_TXAK);
		/* Do not generate stop on last byte */
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX);
	}
	if (l == 2) {
		/* switch on txak bit, do not send ack bit after getting next byte */
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
	}
	*i2c->data++ = readb(i2c->base + MPC_I2C_DR);
	i2c->count = --l;
	if (l <= 0) {		/* should only happen with l == 0 */
		wake_up(&i2c->queue);
		return;
	}
}

#if 0
static void bytewrite(struct mpc_i2c *i2c)
{
	u32 l=i2c->count--;
	u32 x=i2c->interrupt;
	if ((x & CSR_MCF) || !(x & CSR_MAL) || (l == 0)) {
		wake_up(&i2c->queue);
		return;
	}
	writeb(*i2c->data++, i2c->base + MPC_I2C_DR);
}
#endif



static irqreturn_t mpc_i2c_isr(int irq, void *dev_id)
{
	struct mpc_i2c *i2c = dev_id;
	u32 x;
	if (readb(i2c->base + MPC_I2C_SR) & CSR_MIF) {
		/*
		 * Read again to allow register to stabilise
		 * Tracy has a theory that there is a race condition here
		 * during access of the i2c->interrupt value
		 */

		i2c->interrupt = x = readb(i2c->base + MPC_I2C_SR);
		writeb(0, i2c->base + MPC_I2C_SR);
		switch (i2c->cmd) {
		default: BUG();
		case OTHER: wake_up(&i2c->queue); break;
		case BASIC_READ:
			if ((x & (CSR_MCF|CSR_MAL)) != CSR_MCF)
				wake_up(&i2c->queue);
			else 
				byteread(i2c);
			break;
		/* case BASIC_WRITE: bytewrite(i2c); break; */
		}

	}
	return IRQ_HANDLED;
}

/* Sometimes 9th clock pulse isn't generated, and slave doesn't release
 * the bus, because it wants to send ACK.
 * Following sequence of enabling/disabling and sending start/stop generates
 * the pulse, so it's all OK.
 */
static void mpc_i2c_fixup(struct mpc_i2c *i2c)
{
	printk("%s: called\n", __FUNCTION__);
	writeccr(i2c, 0);							// reset/disable
	udelay(30);
	writeccr(i2c, CCR_MEN);						// enable
	udelay(30);
	writeccr(i2c, CCR_MSTA | CCR_MTX);			// disable, rdy for
	udelay(30);									// masterSTART/xmit mode
	writeccr(i2c, CCR_MSTA | CCR_MTX | CCR_MEN); // enable mstST/xmit mode
	udelay(30);
	writeccr(i2c, CCR_MEN);						// slaveSTOP/rcv mode
	udelay(30);
}

/* 
 * According the FSL, this sequence will force any devices
 * off the bus that were left waiting to complete a transfer.
 * We will try this with the Mercury Line Card SFPs
 */
void mpc_bus_flush(struct mpc_i2c *i2c)
{
	unsigned int sr;
	static int once = 0;
	if (!once++) {
		printk("\n%s: new version(8) w/ fixup\n", __FUNCTION__);
	}

	/* Disable and reset the controller I2CCR[MEN] = 0 */
	writeccr(i2c, 0);
	udelay(30);

	/* Change controller to master mode I2CCR[MSTA] = 1 */
	writeccr(i2c, CCR_MSTA); /* 0x20 */
	udelay(30);

	/* Re-enable controller I2CCR[MEN] = 1 */
	writeccr(i2c, CCR_MSTA | CCR_MEN); /* 0xA0 */
	udelay(30);

	/* Read I2CDR. Should Generate 9 clocks and a STOP, w/o a start, MSTA */
	(void)readb(i2c->base + MPC_I2C_DR);

	/* Put in slave mode I2CCR[MSTA] = 0, this generates a STOP */
	writeccr(i2c, CCR_MEN); /* 0x80 */
	udelay(30);

	sr = readb(i2c->base + MPC_I2C_SR);
	if ((sr & (CSR_MBB | CSR_MAL | CSR_RXAK)) ==
											(CSR_MBB | CSR_MAL | CSR_RXAK)) {
		int	i;

		printk("\n%s: (called %d) SR: %#x\n", __FUNCTION__, once, sr);
		/* sometimes one or more fixups in a row finishes the recov */
		for (i = 0; i < 5; i++) {
			mpc_i2c_fixup(i2c);
			sr = readb(i2c->base + MPC_I2C_SR);
			printk("%s: SR after%d: %#x\n", __FUNCTION__, i+1, sr);
			if (!(sr & CSR_MBB)) {
				break;
			}
		}
	}
}

/*
 * wait for I2C event
 * After this routine returns i2c->interrupt is 0
 */
static int i2c_wait(struct mpc_i2c *i2c, unsigned timeout, int writing)
{
	unsigned long orig_jiffies = jiffies;
	u32 x, sr;
	int result = 0;

#ifdef CONFIG_SILKWORM
	struct timeval	perf_begin;
	do_gettimeofday(&perf_begin);
#endif

	if (i2c->irq == NO_IRQ) {
		while (!(readb(i2c->base + MPC_I2C_SR) & CSR_MIF)) {
		/* We don't expect this code to be used as we expect to have an irq */
			schedule();
			if (time_after(jiffies, orig_jiffies + timeout)) {
				dev_dbg(i2c->dev, "timeout\n");
				// printk("%s %d: I2C: timeout\n", __FUNCTION__, __LINE__);
				writeccr(i2c, 0);
				result = -EIO;
				break;
			}
		}
		x = readb(i2c->base + MPC_I2C_SR);
		writeb(0, i2c->base + MPC_I2C_SR);
	} else {
		/*
		 * Interrupt mode. Timeout must be well under 1 sec.
		 * Is there is a race here with the i2c->interrupt variable
		 * wait_event_interruptible_timeout is a macro, which includes a loop
		 * that tests the condition
		 */

		result = wait_event_timeout(i2c->queue,
			(i2c->interrupt & CSR_MIF), timeout);

		if (unlikely(!(i2c->interrupt & CSR_MIF))) {
			dev_dbg(i2c->dev, "wait timeout\n");
			printk(KERN_ERR "%s %d: I2C: wait timeout\n", __FUNCTION__, __LINE__);
			writeccr(i2c, 0);
			result = -ETIMEDOUT;
		}

		x = i2c->interrupt;
		i2c->interrupt = 0;
	}

	if (result < 0) {
		goto end;
	}

	if (!(x & CSR_MCF)) {
		dev_dbg(i2c->dev, "unfinished\n");
		// printk("%s %d: I2C: unfinished\n", __FUNCTION__, __LINE__);
		result = -EIO;
		goto end;
	}

	if (x & CSR_MAL) {
		dev_dbg(i2c->dev, "MAL\n");
		// printk("%s %d: I2C: MAL, SR: %#x\n", __FUNCTION__, __LINE__, x);
		/* let's try the flush/fixup here */
		mpc_bus_flush(i2c);
		sr = readb(i2c->base + MPC_I2C_SR);

		if (sr & CSR_MAL) {
			result = -EIO;
			goto end;
		}
	}

	if (writing && (x & CSR_RXAK)) {
		dev_dbg(i2c->dev, "No RXAK\n");
		// printk("%s %d: I2C: No RXAK\n", __FUNCTION__, __LINE__);
		/* generate stop */
		writeccr(i2c, CCR_MEN);
		result = -EIO;
		goto end;
	}

 end:
#ifdef CONFIG_SILKWORM
	perf_stats_xfer_end(&perf_stats->wait_times, &perf_begin);
#endif
	return (result);
}

#ifdef CONFIG_SILKWORM
static unsigned char  mpc_i2c_get_set_freq(struct mpc_i2c *i2c,
										   unsigned char new_data)
{
	unsigned char old_data;
		old_data = readb(i2c->base + MPC_I2C_FDR);
		writeb(new_data, i2c->base + MPC_I2C_FDR);
		return (old_data);
}
#endif

#if defined(CONFIG_PPC_MPC52xx) || defined(CONFIG_PPC_MPC512x)
static const struct mpc_i2c_divider mpc_i2c_dividers_52xx[] __devinitconst = {
	{20, 0x20}, {22, 0x21}, {24, 0x22}, {26, 0x23},
	{28, 0x24}, {30, 0x01}, {32, 0x25}, {34, 0x02},
	{36, 0x26}, {40, 0x27}, {44, 0x04}, {48, 0x28},
	{52, 0x63}, {56, 0x29}, {60, 0x41}, {64, 0x2a},
	{68, 0x07}, {72, 0x2b}, {80, 0x2c}, {88, 0x09},
	{96, 0x2d}, {104, 0x0a}, {112, 0x2e}, {120, 0x81},
	{128, 0x2f}, {136, 0x47}, {144, 0x0c}, {160, 0x30},
	{176, 0x49}, {192, 0x31}, {208, 0x4a}, {224, 0x32},
	{240, 0x0f}, {256, 0x33}, {272, 0x87}, {288, 0x10},
	{320, 0x34}, {352, 0x89}, {384, 0x35}, {416, 0x8a},
	{448, 0x36}, {480, 0x13}, {512, 0x37}, {576, 0x14},
	{640, 0x38}, {768, 0x39}, {896, 0x3a}, {960, 0x17},
	{1024, 0x3b}, {1152, 0x18}, {1280, 0x3c}, {1536, 0x3d},
	{1792, 0x3e}, {1920, 0x1b}, {2048, 0x3f}, {2304, 0x1c},
	{2560, 0x1d}, {3072, 0x1e}, {3584, 0x7e}, {3840, 0x1f},
	{4096, 0x7f}, {4608, 0x5c}, {5120, 0x5d}, {6144, 0x5e},
	{7168, 0xbe}, {7680, 0x5f}, {8192, 0xbf}, {9216, 0x9c},
	{10240, 0x9d}, {12288, 0x9e}, {15360, 0x9f}
};

static int __devinit mpc_i2c_get_fdr_52xx(struct device_node *node, u32 clock,
					  int prescaler)
{
	const struct mpc_i2c_divider *div = NULL;
	unsigned int pvr = mfspr(SPRN_PVR);
	u32 divider;
	int i;

	if (clock == MPC_I2C_CLOCK_LEGACY)
		return -EINVAL;

	/* Determine divider value */
	divider = mpc5xxx_get_bus_frequency(node) / clock;

	/*
	 * We want to choose an FDR/DFSR that generates an I2C bus speed that
	 * is equal to or lower than the requested speed.
	 */
	for (i = 0; i < ARRAY_SIZE(mpc_i2c_dividers_52xx); i++) {
		div = &mpc_i2c_dividers_52xx[i];
		/* Old MPC5200 rev A CPUs do not support the high bits */
		if (div->fdr & 0xc0 && pvr == 0x80822011)
			continue;
		if (div->divider >= divider)
			break;
	}

	return div ? (int)div->fdr : -EINVAL;
}

static void __devinit mpc_i2c_setup_52xx(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
	int ret, fdr;

	if (clock == MPC_I2C_CLOCK_PRESERVE) {
		dev_dbg(i2c->dev, "using fdr %d\n",
			readb(i2c->base + MPC_I2C_FDR));
		return;
	}

	ret = mpc_i2c_get_fdr_52xx(node, clock, prescaler);
	fdr = (ret >= 0) ? ret : 0x3f; /* backward compatibility */

	writeb(fdr & 0xff, i2c->base + MPC_I2C_FDR);

	if (ret >= 0)
		dev_info(i2c->dev, "clock %d Hz (fdr=%d)\n", clock, fdr);
}
#else /* !(CONFIG_PPC_MPC52xx || CONFIG_PPC_MPC512x) */
static void __devinit mpc_i2c_setup_52xx(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
}
#endif /* CONFIG_PPC_MPC52xx || CONFIG_PPC_MPC512x */

#ifdef CONFIG_PPC_MPC512x
static void __devinit mpc_i2c_setup_512x(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
	struct device_node *node_ctrl;
	void __iomem *ctrl;
	const u32 *pval;
	u32 idx;

	/* Enable I2C interrupts for mpc5121 */
	node_ctrl = of_find_compatible_node(NULL, NULL,
					    "fsl,mpc5121-i2c-ctrl");
	if (node_ctrl) {
		ctrl = of_iomap(node_ctrl, 0);
		if (ctrl) {
			/* Interrupt enable bits for i2c-0/1/2: bit 24/26/28 */
			pval = of_get_property(node, "reg", NULL);
			idx = (*pval & 0xff) / 0x20;
			setbits32(ctrl, 1 << (24 + idx * 2));
			iounmap(ctrl);
		}
		of_node_put(node_ctrl);
	}

	/* The clock setup for the 52xx works also fine for the 512x */
	mpc_i2c_setup_52xx(node, i2c, clock, prescaler);
}
#else /* CONFIG_PPC_MPC512x */
static void __devinit mpc_i2c_setup_512x(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
}
#endif /* CONFIG_PPC_MPC512x */

#ifdef CONFIG_FSL_SOC
static const struct mpc_i2c_divider mpc_i2c_dividers_8xxx[] __devinitconst = {
	{160, 0x0120}, {192, 0x0121}, {224, 0x0122}, {256, 0x0123},
	{288, 0x0100}, {320, 0x0101}, {352, 0x0601}, {384, 0x0102},
	{416, 0x0602}, {448, 0x0126}, {480, 0x0103}, {512, 0x0127},
	{544, 0x0b03}, {576, 0x0104}, {608, 0x1603}, {640, 0x0105},
	{672, 0x2003}, {704, 0x0b05}, {736, 0x2b03}, {768, 0x0106},
	{800, 0x3603}, {832, 0x0b06}, {896, 0x012a}, {960, 0x0107},
	{1024, 0x012b}, {1088, 0x1607}, {1152, 0x0108}, {1216, 0x2b07},
	{1280, 0x0109}, {1408, 0x1609}, {1536, 0x010a}, {1664, 0x160a},
	{1792, 0x012e}, {1920, 0x010b}, {2048, 0x012f}, {2176, 0x2b0b},
	{2304, 0x010c}, {2560, 0x010d}, {2816, 0x2b0d}, {3072, 0x010e},
	{3328, 0x2b0e}, {3584, 0x0132}, {3840, 0x010f}, {4096, 0x0133},
	{4608, 0x0110}, {5120, 0x0111}, {6144, 0x0112}, {7168, 0x0136},
	{7680, 0x0113}, {8192, 0x0137}, {9216, 0x0114}, {10240, 0x0115},
	{12288, 0x0116}, {14336, 0x013a}, {15360, 0x0117}, {16384, 0x013b},
	{18432, 0x0118}, {20480, 0x0119}, {24576, 0x011a}, {28672, 0x013e},
	{30720, 0x011b}, {32768, 0x013f}, {36864, 0x011c}, {40960, 0x011d},
	{49152, 0x011e}, {61440, 0x011f}
};

static u32 __devinit mpc_i2c_get_sec_cfg_8xxx(void)
{
	struct device_node *node = NULL;
	u32 __iomem *reg;
	u32 val = 0;

	node = of_find_node_by_name(NULL, "global-utilities");
	if (node) {
		const u32 *prop = of_get_property(node, "reg", NULL);
		if (prop) {
			/*
			 * Map and check POR Device Status Register 2
			 * (PORDEVSR2) at 0xE0014
			 */
			reg = ioremap(get_immrbase() + *prop + 0x14, 0x4);
			if (!reg)
				printk(KERN_ERR
				       "Error: couldn't map PORDEVSR2\n");
			else
				val = in_be32(reg) & 0x00000080; /* sec-cfg */
			iounmap(reg);
		}
	}
	if (node)
		of_node_put(node);

	return val;
}

static int __devinit mpc_i2c_get_fdr_8xxx(struct device_node *node, u32 clock,
					  u32 prescaler)
{
	const struct mpc_i2c_divider *div = NULL;
	u32 divider;
	int i;

	if (clock == MPC_I2C_CLOCK_LEGACY)
		return -EINVAL;

	/* Determine proper divider value */
	if (of_device_is_compatible(node, "fsl,mpc8544-i2c"))
		prescaler = mpc_i2c_get_sec_cfg_8xxx() ? 3 : 2;
	if (!prescaler)
		prescaler = 1;

	divider = fsl_get_sys_freq() / clock / prescaler;

	pr_debug("I2C: src_clock=%d clock=%d divider=%d\n",
		 fsl_get_sys_freq(), clock, divider);

	/*
	 * We want to choose an FDR/DFSR that generates an I2C bus speed that
	 * is equal to or lower than the requested speed.
	 */
	for (i = 0; i < ARRAY_SIZE(mpc_i2c_dividers_8xxx); i++) {
		div = &mpc_i2c_dividers_8xxx[i];
		if (div->divider >= divider)
			break;
	}

	return div ? (int)div->fdr : -EINVAL;
}

static void __devinit mpc_i2c_setup_8xxx(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
	int ret, fdr;

	if (clock == MPC_I2C_CLOCK_PRESERVE) {
		dev_dbg(i2c->dev, "using dfsrr %d, fdr %d\n",
			readb(i2c->base + MPC_I2C_DFSRR),
			readb(i2c->base + MPC_I2C_FDR));
		return;
	}

	ret = mpc_i2c_get_fdr_8xxx(node, clock, prescaler);
	fdr = (ret >= 0) ? ret : 0x1031; /* backward compatibility */

	writeb(fdr & 0xff, i2c->base + MPC_I2C_FDR);
	writeb((fdr >> 8) & 0xff, i2c->base + MPC_I2C_DFSRR);

	if (ret >= 0)
		dev_info(i2c->dev, "clock %d Hz (dfsrr=%d fdr=%d)\n",
			 clock, fdr >> 8, fdr & 0xff);
}

#else /* !CONFIG_FSL_SOC */
static void __devinit mpc_i2c_setup_8xxx(struct device_node *node,
					 struct mpc_i2c *i2c,
					 u32 clock, u32 prescaler)
{
}
#endif /* CONFIG_FSL_SOC */

static void mpc_i2c_start(struct mpc_i2c *i2c)
{
#ifdef CONFIG_SILKWORM
	/* Reset the device in case it thinks the bus is busy */
	writeccr(i2c, 0);
	udelay(20);
#endif
	/* Clear arbitration */
	writeb(0, i2c->base + MPC_I2C_SR);
	/* Start with MEN */
	writeccr(i2c, CCR_MEN);
	i2c->cmd = OTHER;	/* reset the interrupt processing */
	i2c->interrupt = 0;
}

static void mpc_i2c_stop(struct mpc_i2c *i2c)
{
#ifdef CONFIG_SILKWORM
/*
 * This code is manually asserting STOP command on the bus in part because
 * the controller provides no way to do it automatically after writes.
 * The code MUST ensure correct timing or else the STOP may not be
 * recognized by other devices. This is required by MSA spec for SFP
 * devices.
 */
	udelay(20); /* Ensure STOP is not asserted too soon */
#endif
	writeccr(i2c, CCR_MEN);
#ifdef CONFIG_SILKWORM
	udelay(20); /* Ensure STOP remains asserted long enough */
#endif
}

#ifdef CONFIG_SILKWORM
/*
 *
 */
static void mpc_access_debug(struct mpc_i2c *i2c, int target,
		    const u8 * data, int length, int restart, int read)
{
	int i;
	unsigned char cpld_data;

	printk("-----------------------------\n");
	printk("\n%s(): %s target 0x%x i2c->base 0x%x"
		   "length %d restart 0x%x",
		   __FUNCTION__, ((read)?"READ":"WRITE"),
		   target, (int)i2c->base, length, restart);

	for (i = 0; i < length; i++) {

		if (!(i % 0x10)) {
			printk("\n0x%4.4x: ", i);
		}
		printk("%02x ", data[i]);

	}
	printk("\n");
	i = cpld_read_register(MET_I2C_SELECT, &cpld_data);
	printk("\n%s():cpld_data: 0x%x  ret: %d\n",
		   __FUNCTION__, cpld_data, i);
	printk("-----------------------------\n");
}
#endif

static int mpc_write(struct mpc_i2c *i2c, int target,
		     const u8 *data, int length, int restart)
{
	int i, result = 0;
	unsigned timeout = i2c->adap.timeout;
	u32 flags = restart ? CCR_RSTA : 0;

#ifdef CONFIG_SILKWORM
	struct timeval	perf_begin;
	perf_stats_xfer_begin(&perf_stats->write_xfers,
						  &perf_stats->write_bytes,
						  &perf_begin, length);
#endif

	/* Start as master */
	writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX | flags);
	/* Write target byte */
	writeb((target << 1), i2c->base + MPC_I2C_DR);

	if (i2c_wait(i2c, timeout, 1) < 0) {
		//		printk("%s %d: I2C: timeout1 Target-0x%x\n", __FUNCTION__, __LINE__, target);
		result = -1;
		goto end;
	}

#ifdef CONFIG_SILKWORM
	if (core_i2c_debug > 2) {
		mpc_access_debug(i2c, target, data, length, restart, 0);
	}
#endif

	for (i = 0; i < length; i++) {
		/* Write data byte */
		writeb(data[i], i2c->base + MPC_I2C_DR);

		if (i2c_wait(i2c, timeout, 1) < 0) {
			//printk("\n%s %d: I2C: timeout2\n", __FUNCTION__, __LINE__);
			result = -1;
			goto end;
		}
	}

 end:
#ifdef CONFIG_SILKWORM
	perf_stats_xfer_end(&perf_stats->write_xfers, &perf_begin);
#endif
	return (result);
}

static int mpc_read(struct mpc_i2c *i2c, int target,
		    u8 *data, int length, int restart)
{
	unsigned timeout = i2c->adap.timeout;
	int i, result = length;
	u32 flags = restart ? CCR_RSTA : 0;

#ifdef CONFIG_SILKWORM
	struct timeval	perf_begin;
	perf_stats_xfer_begin(&perf_stats->read_xfers,
						  &perf_stats->read_bytes,
						  &perf_begin, length);
#endif

	i2c->interrupt = 0;
	/* Switch to read - restart */
	writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX | flags);
	/* Write target address byte - this time with the read flag set */
	writeb((target << 1) | 1, i2c->base + MPC_I2C_DR);

	if (i2c_wait(i2c, timeout, 1) < 0) {
		result = -1;
		goto end;
	}

#ifdef CONFIG_SILKWORM
	/*
	 * This delay is added to comply with SFP MSA spec that requires
	 * a minimum of 20us between end of an address cycle and beginning of
	 * the data cycle. There is no way to tell if an SFP is being
	 * accessed, so we have to always do this.
	 */
	udelay(20);
#endif

	if (length == 0) goto maybedebug;

	if (length == 1)
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
	else
		writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA);

 #if 0
	/*
	 * Add debug/Trace code here to print to console one-time
	 * whenever the i2cTurboCnfg is going from:
	 *  0 to non-zero (1 = I2C_TURBO_CNFG_ENABLE_SFP,
	 *				   2 = I2C_TURBO_CNFG_ENABLE_ALL mode)
	 *  non-zero to 0 (0 = I2C_TURBO_CNFG_DISABLED) 
	 *  This can be ifdef'd out in the GA version.
	 */
	if (i2cTurboCnfg) {
		if (!one_time_on) {
			printk("i2cTurboCnfg on=%d\n", i2cTurboCnfg);
		}
		one_time_on++;
		one_time_off = 0;
	} else {
		if (!one_time_off) {
			printk("i2cTurboCnfg off\n");
		}
		one_time_off++;
		one_time_on = 0;
	}
#endif

	if ((i2c->irq) && (i2cTurboFlag)) {
		/* Interrupts are enabled, do this the fast way */
		i2c->data = data;
		i2c->count = length;
		i2c->cmd = BASIC_READ;
		i2c->interrupt = 0;
		/* Dummy read */
		readb(i2c->base + MPC_I2C_DR);
		if (i2c_wait(i2c, timeout * length, 0) < 0) {
			result = -1;
			goto end;
		}
	} else {
		/* Dummy read */
		readb(i2c->base + MPC_I2C_DR);


		for (i = 0; i < length; i++) {
			if (i2c_wait(i2c, timeout, 0) < 0) {
				result = -1;
				goto end;
			}

			/* Generate txack on next to last byte */
			if (i == length - 2)
				writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_TXAK);
			/* Do not generate stop on last byte */
			if (i == length - 1)
				writeccr(i2c, CCR_MIEN | CCR_MEN | CCR_MSTA | CCR_MTX);
			data[i] = readb(i2c->base + MPC_I2C_DR);
		}
	}
maybedebug:

#ifdef CONFIG_SILKWORM
	if (core_i2c_debug > 2) {
		mpc_access_debug(i2c, target, data, length, restart, 1);
	}
#endif

 end:
#ifdef CONFIG_SILKWORM
	perf_stats_xfer_end(&perf_stats->read_xfers, &perf_begin);
#endif
	return (result);
}

static int mpc_xfer(struct i2c_adapter *adap, struct i2c_msg *msgs, int num)
{
	struct i2c_msg *pmsg;
	int i;
	int ret = 0;
	unsigned long orig_jiffies = jiffies;
	struct mpc_i2c *i2c = i2c_get_adapdata(adap);

#ifdef CONFIG_SILKWORM
	int freq_change = 0;
	unsigned char old_freq = 0;

	for (i = 0; ret >= 0 && i < num; i++) {
		pmsg = &msgs[i];
		if (pmsg->flags & I2C_M_SET_MIN_FREQ ){
			/* setup the clock speed 
			* set I2C bus 1 clock to 20KHz to access power supply
			* eeprom. For MPC8548 CPU:
			*   clock = 1/2 of CCB clock (400MHz) divided by
			*           frequency divider
			*   clock = ((400MHz / 2) / 10240) ~= 20KHz
			*   frequency divider of 10240 = frequency divider
			*/

			old_freq = mpc_i2c_get_set_freq(i2c, 0x38);
			freq_change = 1;
			break;
		}
	}
#endif

	mpc_i2c_start(i2c);

	/* Allow bus up to 1s to become not busy */
	while (readb(i2c->base + MPC_I2C_SR) & CSR_MBB) {
		if (signal_pending(current)) {
			dev_dbg(i2c->dev, "Interrupted\n");
			printk(KERN_ERR "\n%s %d: I2C: Interrupted\n", __FUNCTION__, __LINE__);
			writeccr(i2c, 0);
			return -EINTR;
		}
		if (time_after(jiffies, orig_jiffies + IIC_TIMEOUT)) {
			dev_dbg(i2c->dev, "timeout\n");
			if (readb(i2c->base + MPC_I2C_SR) ==
			    (CSR_MCF | CSR_MBB | CSR_RXAK))
				mpc_i2c_fixup(i2c);
			printk(KERN_ERR "\n%s %d: I2C: timeout\n", __FUNCTION__, __LINE__);
			return -EIO;
		}
		schedule();
	}

	for (i = 0; ret >= 0 && i < num; i++) {
		pmsg = &msgs[i];
		dev_dbg(i2c->dev,
			"Doing %s %d bytes to 0x%02x - %d of %d messages\n",
			pmsg->flags & I2C_M_RD ? "read" : "write",
			pmsg->len, pmsg->addr, i + 1, num);
		if (pmsg->flags & I2C_M_RD)
			ret =
			    mpc_read(i2c, pmsg->addr, pmsg->buf, pmsg->len, i);
		else
			ret =
			    mpc_write(i2c, pmsg->addr, pmsg->buf, pmsg->len, i);
	}
	mpc_i2c_stop(i2c);

#ifdef CONFIG_SILKWORM
	if (freq_change) {
		/* reset the clock speed */
		mpc_i2c_get_set_freq(i2c, old_freq);
		freq_change = 0;
	}
#endif
	return (ret < 0) ? ret : num;
}

static u32 mpc_functionality(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static /* const */ struct i2c_algorithm mpc_algo = {
	.master_xfer = mpc_xfer,
	.functionality = mpc_functionality,
	.fixup_i2c		= NULL,        // Set in board initialization
};

static struct i2c_adapter mpc_ops = {
	.owner = THIS_MODULE,
	.name = "MPC adapter",
	.algo = &mpc_algo,
	.class = I2C_CLASS_HWMON,
	.timeout = IIC_TIMEOUT,
	.retries = IIC_RETRY /* Doesn't seem to be used - yet */
};

static int __devinit fsl_i2c_probe(struct of_device *op,
				   const struct of_device_id *match)
{
	struct mpc_i2c *i2c;
	const u32 *prop;
	u32 clock = MPC_I2C_CLOCK_LEGACY;
	int result = 0;
	int plen;

	i2c = kzalloc(sizeof(*i2c), GFP_KERNEL);
	if (!i2c)
		return -ENOMEM;

	i2c->dev = &op->dev; /* for debug and error output */

	init_waitqueue_head(&i2c->queue);

	i2c->base = of_iomap(op->node, 0);
	if (!i2c->base) {
		dev_err(i2c->dev, "failed to map controller\n");
		result = -ENOMEM;
		goto fail_map;
	}

	i2c->cmd = OTHER;	/* Not strictly rqd as OTHER=0 & we just set it to 0 */
	i2c->irq = irq_of_parse_and_map(op->node, 0);
	if (i2c->irq != NO_IRQ) { /* i2c->irq = NO_IRQ implies polling */
		result = request_irq(i2c->irq, mpc_i2c_isr,
				     IRQF_SHARED, "i2c-mpc", i2c);
		if (result < 0) {
			dev_err(i2c->dev, "failed to attach interrupt\n");
			goto fail_request;
		}
	}

	if (of_get_property(op->node, "fsl,preserve-clocking", NULL)) {
		clock = MPC_I2C_CLOCK_PRESERVE;
	} else {
		prop = of_get_property(op->node, "clock-frequency", &plen);
		if (prop && plen == sizeof(u32))
			clock = *prop;
	}

	if (match->data) {
		struct mpc_i2c_data *data = match->data;
		data->setup(op->node, i2c, clock, data->prescaler);
	} else {
		/* Backwards compatibility */
		if (of_get_property(op->node, "dfsrr", NULL))
			mpc_i2c_setup_8xxx(op->node, i2c, clock, 0);
	}

	dev_set_drvdata(&op->dev, i2c);

	i2c->adap = mpc_ops;
	i2c_set_adapdata(&i2c->adap, i2c);
	i2c->adap.dev.parent = &op->dev;

	result = i2c_add_adapter(&i2c->adap);
	if (result < 0) {
		dev_err(i2c->dev, "failed to add adapter\n");
		goto fail_add;
	}
	of_register_i2c_devices(&i2c->adap, op->node);

	return result;

 fail_add:
	dev_set_drvdata(&op->dev, NULL);
	free_irq(i2c->irq, i2c);
 fail_request:
	irq_dispose_mapping(i2c->irq);
	iounmap(i2c->base);
 fail_map:
	kfree(i2c);
	return result;
};

static int __devexit fsl_i2c_remove(struct of_device *op)
{
	struct mpc_i2c *i2c = dev_get_drvdata(&op->dev);

	i2c_del_adapter(&i2c->adap);
	dev_set_drvdata(&op->dev, NULL);

	if (i2c->irq != NO_IRQ)
		free_irq(i2c->irq, i2c);

	irq_dispose_mapping(i2c->irq);
	iounmap(i2c->base);
	kfree(i2c);
	return 0;
};

static struct mpc_i2c_data mpc_i2c_data_512x __devinitdata = {
	.setup = mpc_i2c_setup_512x,
};

static struct mpc_i2c_data mpc_i2c_data_52xx __devinitdata = {
	.setup = mpc_i2c_setup_52xx,
};

static struct mpc_i2c_data mpc_i2c_data_8313 __devinitdata = {
	.setup = mpc_i2c_setup_8xxx,
};

static struct mpc_i2c_data mpc_i2c_data_8543 __devinitdata = {
	.setup = mpc_i2c_setup_8xxx,
	.prescaler = 2,
};

static struct mpc_i2c_data mpc_i2c_data_8544 __devinitdata = {
	.setup = mpc_i2c_setup_8xxx,
	.prescaler = 3,
};

static const struct of_device_id mpc_i2c_of_match[] = {
	{.compatible = "mpc5200-i2c", .data = &mpc_i2c_data_52xx, },
	{.compatible = "fsl,mpc5200b-i2c", .data = &mpc_i2c_data_52xx, },
	{.compatible = "fsl,mpc5200-i2c", .data = &mpc_i2c_data_52xx, },
	{.compatible = "fsl,mpc5121-i2c", .data = &mpc_i2c_data_512x, },
	{.compatible = "fsl,mpc8313-i2c", .data = &mpc_i2c_data_8313, },
	{.compatible = "fsl,mpc8543-i2c", .data = &mpc_i2c_data_8543, },
	{.compatible = "fsl,mpc8544-i2c", .data = &mpc_i2c_data_8544, },
	/* Backward compatibility */
	{.compatible = "fsl-i2c", },
	{},
};
MODULE_DEVICE_TABLE(of, mpc_i2c_of_match);

/* Structure for a device driver */
static struct of_platform_driver mpc_i2c_driver = {
	.match_table	= mpc_i2c_of_match,
	.probe		= fsl_i2c_probe,
	.remove		= __devexit_p(fsl_i2c_remove),
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= DRV_NAME,
	},
};

#ifdef CONFIG_SILKWORM
static int read_bus_i2c_debug(char *buf, char **start, off_t offset, int len, int *eof,
                 void *private)
{
	int nr = 0;

	// MOD_INC_USE_COUNT;
	nr += sprintf(buf+nr, "%d\n", core_i2c_debug);
	// MOD_DEC_USE_COUNT;

	return nr;
}

static int write_bus_i2c_debug(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];

	// MOD_INC_USE_COUNT;

	if(count > PAGE_SIZE)
		count = PAGE_SIZE;

	if(copy_from_user(ptr, buffer, count)) {
		// MOD_DEC_USE_COUNT;
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return -EFAULT;
	}

	if(*ptr) {
		if (*ptr >= '0' && *ptr <= '9') {
			core_i2c_debug = *ptr - '0';
			printk(KERN_INFO "%s :Setting "
					   "core_i2c_debug to %d \n", __FUNCTION__,
				   core_i2c_debug);
		} else {
			printk(KERN_ERR "Invalid debug level.\n");
		}
	}

	// MOD_DEC_USE_COUNT;
	return count;
}
/*
 *
 */
static void iic_debug_create_proc(void)
{

	struct proc_dir_entry *proc_dir_i2c;
	struct proc_dir_entry *proc_debug_i2c;

	proc_dir_i2c = proc_mkdir("driver/iic", NULL);

	if (proc_dir_i2c == NULL) {

		printk(KERN_ERR "\n\n%s: Could not create directory driver/iic/\n",
			   __FUNCTION__);
		return;

	}
	proc_debug_i2c = create_proc_entry("i2c_debug", 0, proc_dir_i2c);

	if (!proc_debug_i2c) {
		printk(KERN_ERR "%s: Could not create /driver/iic/i2c_debug",
			   __FUNCTION__);
		//		i2cproc_cleanup();
		return;
 	}
	proc_debug_i2c->read_proc = &read_bus_i2c_debug;
	proc_debug_i2c->write_proc = &write_bus_i2c_debug;

#if 0
	// much simpler to use sysctl these days
	register_sysctl_table(i2c_root_tab);
#endif

	if (!perf_stats) {
		perf_stats_init();
	}
}
#endif

static int __init fsl_i2c_init(void)
{
	int rv;

#ifdef CONFIG_SILKWORM
	printk(KERN_DEBUG "----------->%s(): %s Initialization entered",
		   __FUNCTION__, __FILE__);
	iic_debug_create_proc();
#endif
	rv = of_register_platform_driver(&mpc_i2c_driver);
	if (rv)
		printk(KERN_ERR DRV_NAME
		       " of_register_platform_driver failed (%i)\n", rv);
	return rv;
}

static void __exit fsl_i2c_exit(void)
{
	of_unregister_platform_driver(&mpc_i2c_driver);
}

module_init(fsl_i2c_init);
module_exit(fsl_i2c_exit);
// EXPORT_SYMBOL(i2cTurboCnfg);
EXPORT_SYMBOL(i2cTurboFlag);

MODULE_AUTHOR("Adrian Cox <adrian@humboldt.co.uk>");
MODULE_DESCRIPTION("I2C-Bus adapter for MPC107 bridge and "
		   "MPC824x/83xx/85xx/86xx/512x/52xx processors");
MODULE_LICENSE("GPL");
