/*
 *
 *    Copyright (c) 2007 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *		Provide support for performance statistics to be read via proc file system
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <asm/uaccess.h>
#include "i2c-stats.h"

i2c_perf_stats_t i2c_perf, *perf_stats = NULL;

static char *
next_token(char *cmd)
{
    char *ptr;

    ptr = cmd;
    while ((*ptr != ' ') && (*ptr != '\0')) {
        ptr++;
    }
    while (*ptr == ' ') {
         (ptr++);
    }

    return (ptr);
}

static char *
SkipWhiteSpace(char *buff)
{

  while ((buff)&&((*buff == 0x20)||(*buff == 0x9)))
    buff++;

  return (buff);

}

static void
clear_debug_stats(perf_debug_stats_t *ptr)
{
	ptr->min = 0;
	ptr->max = 0;
	ptr->sum = 0;
	ptr->count = 0;

}

void
perf_stats_clear(i2c_perf_stats_t *ptr)
{

	clear_debug_stats(&(ptr->read_xfers));
	clear_debug_stats(&(ptr->read_bytes));
	clear_debug_stats(&(ptr->write_xfers));
	clear_debug_stats(&(ptr->write_bytes));
	clear_debug_stats(&(ptr->wait_times));
	clear_debug_stats(&(ptr->clients));
	clear_debug_stats(&(ptr->error_times));
	clear_debug_stats(&(ptr->arb_times));
	clear_debug_stats(&(ptr->arb_loops));
	clear_debug_stats(&(ptr->intr_times));

}

/*
 *
 */
int
perf_stats_show(char *buf, char **start, off_t pos, int count, int *eof,
	  void *data)
{

	u_int len = 0, limit;
	i2c_perf_stats_t *perf_stats;

	if (data == NULL) {
		printk("\n%s: No statistics data\n", __FUNCTION__);
		return (-1);
	}

	perf_stats = (i2c_perf_stats_t*) data;

	*eof = 1;

	/*
	 * Guarantee we always have at least 1 line of space
	 * in the buffer. If the format of the text changes
	 * such that one line is > 90 characters, then this
	 * constant needs to change to match.
	 */
	limit = count - 90;

	len = sprintf(buf + len, "\n\t\tI2C performance statistics\n\n");



	len += sprintf(buf + len,
		       "\tTotal \t\t    Time(us)\n");

	len += sprintf(buf + len,
		       "\tCount\t    Min\t    Max\t\tTotal");

	len += sprintf(buf + len,
				   "\n\t----------  ------------------------------\n");



	len += sprintf(buf + len,
				   "Reads \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->read_xfers.count, perf_stats->read_xfers.min,
				   perf_stats->read_xfers.max,
				   perf_stats->read_xfers.sum);

	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "Writes \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->write_xfers.count, perf_stats->write_xfers.min,
				   perf_stats->write_xfers.max,
				   perf_stats->write_xfers.sum);

	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "Waits \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->wait_times.count,
				   perf_stats->wait_times.min,
				   perf_stats->wait_times.max,
				   perf_stats->wait_times.sum);

	if (len >= limit)
		return (len);

	/* Not yet implemented */
	/*
	len += sprintf(buf + len,
				   "IRQs \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->intr_times.count,
				   perf_stats->intr_times.min,
				   perf_stats->intr_times.max,
				   perf_stats->intr_times.sum);


	if (len >= limit)
		return (len);
	*/

	len += sprintf(buf + len,
				   "Arbs \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->arb_times.count,
				   perf_stats->arb_times.min,
				   perf_stats->arb_times.max,
				   perf_stats->arb_times.sum);

	if (len >= limit)
		return (len);

	/* Not yet implemented */
	/*
	len += sprintf(buf + len,
				   "Errors \t%10.10u  %6.6lu  %10.10lu  %10.10lu\n",
				   perf_stats->error_times.count,
				   perf_stats->error_times.min,
				   perf_stats->error_times.max,
				   perf_stats->error_times.sum);

	if (len >= limit)
		return (len);
	*/

	len += sprintf(buf + len,
		       "\n\t\tMisc. Counts\n");

	len += sprintf(buf + len,
		       "\tMin\tMax\t    Total");

	len += sprintf(buf + len,
				   "\n\t-------------------------------------\n");

	len += sprintf(buf + len,
				   "Reads \t%6.6lu  %10.10lu  %10.10lu Bytes\n",
				   perf_stats->read_bytes.min, perf_stats->read_bytes.max, 
				   perf_stats->read_bytes.sum);


	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "Writes \t%6.6lu  %10.10lu  %10.10lu Bytes\n",
				   perf_stats->write_bytes.min, perf_stats->write_bytes.max, 
				   perf_stats->write_bytes.sum);


	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "Clients\t%6.6lu  %10.10lu  %10.10lu Waiting\n",
				   perf_stats->clients.min,
				   perf_stats->clients.max,
				   perf_stats->clients.sum);

	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "Arbs \t%6.6lu  %10.10lu  %10.10lu Loops\n",
				   perf_stats->arb_loops.min,
				   perf_stats->arb_loops.max,
				   perf_stats->arb_loops.sum);

	if (len >= limit)
		return (len);

	len += sprintf(buf + len,
				   "\nArbitrations lost: %u Current Clients: %u\n\n",
				   perf_stats->arb_loops.count,
				   perf_stats->clients.count);

	//	printk ("\n%s: Exiting. Count %d len %d\n", __FUNCTION__, count, len);

	return (len);
}
/*
 *
 * Allow user to exercise some limited control of
 * the stats:
 *  1. Specify number of bytes to transfer at a time.
 *     Should not exceed 4 with ppc4xx.
 *  2. Specify software buffer size, up to 128 bytes
 *  3. Reset stats counters
 *
 */
int perf_stats_write(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];
	char *cptr;
	// int len = 0;

#if 0
	if (count >= PAGE_SIZE - 1)
		len = PAGE_SIZE - 1;
	else
		len = count;
#endif

	if(count > PAGE_SIZE)
		count = PAGE_SIZE;

	if(copy_from_user(ptr, buffer, count)) {
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return -EFAULT;
	}

	if(*ptr) {

		//	ptr[len] = '\0';
		ptr[count] = '\0';

		/* skip to the first word */
		cptr = SkipWhiteSpace(ptr);

		while (*cptr) {
			printk("::%c::\n", *cptr);
			if (*cptr == '-') {
				// int i;

				cptr++;
				switch (*cptr) {
#if 0
				case 'f': /* FIFO size */
					cptr = next_token(cptr);
					i = (int) simple_strtol(cptr, NULL, 10);
					printk("\n%s: -f options %s [%d]\n", __FUNCTION__,
						   cptr, i);

					if ((i >= 1)&&(i <= 128)) {
						core_cntl_tct = i;
					}
					break;
#endif

#if 0
				case 'b': /* Buffer size */
					cptr = next_token(cptr);
					i = (int) simple_strtol(cptr, NULL, 10);
					printk("\n%s: -b options %s [%d]\n", __FUNCTION__, cptr, i);
					if ((i >= 0)&&(i <= 128)) {
						i2c_buff_size = i;
					}
					break;
#endif
				case 'r': /* Reset stats */
					perf_stats_clear(&i2c_perf);
					break;

				default:
					printk("\n%s: Unknown options %s\n", __FUNCTION__, cptr);
					break;
				}
			} else {
				printk("\n%s: missing leading '-' character %s\n", __FUNCTION__,
					   cptr);
				// return (len);
				return (count);
			}

			cptr = next_token(cptr);
		} /* while */
	}

	return (count);
}

void
perf_stats_init(void)
{
	struct proc_dir_entry *stats_file;

	if (!perf_stats) {

		perf_stats = &i2c_perf;
		perf_stats_clear(perf_stats);

		stats_file = create_proc_read_entry("driver/iic/perf_stats", 0, NULL,
											&perf_stats_show, perf_stats);

		if (stats_file == NULL) {
			printk("\n\n%s: Could not create file driver/iic/perf_stats\n",
				   __FUNCTION__);
			return;
		}

		stats_file->write_proc = &perf_stats_write;
		//		stats_file->owner = THIS_MODULE;  Hari 10/08/2009

	}
}
void
perf_stats_xfer_begin(perf_debug_stats_t *xfers, perf_debug_stats_t *bytes,
					  struct timeval *perf_begin, int length)
{

	xfers->count++;
	PERF_STATS_UPDATE(length, bytes);
	do_gettimeofday(perf_begin);

}
void
perf_stats_xfer_end(perf_debug_stats_t *xfers, struct timeval *perf_begin)
{
	struct timeval perf_end;
	u_long diff;

	do_gettimeofday(&perf_end);
	PERF_TIME_DIFF_PTR(perf_begin, &perf_end, diff);
	PERF_STATS_UPDATE(diff, xfers);
}
