/*
 * Virtual I2C bus driver.
 * Simplifies access to complex multiplexed I2C bus topologies, by presenting
 * each multiplexed bus segment as a virtual I2C adapter.  Supports multi-level
 * mux'ing (mux behind a mux), as well as arbitration for exclusive bus access
 * for those systems which require it (some bladed chassis for example).
 *
 * Brian Kuschak <bkuschak@yahoo.com>
 * Kalpesh Jasapara <kjasapara@yahoo.com> 
 *
 * Adapted from i2c-adap-ibm_ocp.c 
 * Original file Copyright 2000-2002 MontaVista Software Inc.
 *
 */

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/i2c-virtual.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/io.h>
#if defined(CONFIG_I2C_IBM_IIC)
extern void iic_mode_control(struct i2c_adapter *adap, int fsmenable);
#endif


#define VIRT_TIMEOUT		(HZ/2)		/* 500msec */
#define VIRT_RETRIES		3		

/* exclusive access to the bus */
#define I2C_LOCK(adap) rt_mutex_lock(&adap->bus_lock)
#define I2C_UNLOCK(adap) rt_mutex_unlock(&adap->bus_lock)

#ifdef CORE_I2C_DEBUG
extern int core_i2c_debug;
#define DEB(x) if (core_i2c_debug>=1) x;
#define DEB2(x) if (core_i2c_debug>=2) x;
#else
#define DEB(x)
#define DEB2(x)
#endif /* CORE_I2C_DEBUG */

static int i2c_scan = 0;

/* First the I2C 'algorithm' driver code: */

/*
 * Description: acquire exclusive access (if needed), select the mux,
 * perform the transfer on the real i2c adapter, deselect mux and drop
 * exclusive access.
 */
static int
virt_i2c_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
	struct i2c_virt_priv 	*priv = i2c_get_adapdata(adap);
	struct i2c_adapter 	*real = priv->real_adap;
	struct i2c_adapter 	*phys = i2c_get_phys_adap(adap); 
	int ret;

 	 DEB2(printk(KERN_DEBUG "i2c-virtual.o: virt_i2c_xfer: %s with %d msgs.\n",
		            adap->name,num));

	/* Grab the lock for the _real_ adapter.  We already hold the lock for 
	 * the virtual adapter.  Then select the right mux port and perform 
	 * the transfer.
	 */
	I2C_LOCK(real);

	/*
	 * Transfer the write enable flag to the adapter private data so
	 * the mux select/deselect routines can get it. The adapter is shared
	 * among multiple devices, and this flag could get changed if it's
	 * set before getting here.
	 */

	priv->mux.write_enable = msgs[0].flags & I2C_M_WRITE_ENABLE;

	if ((ret = priv->mux.select(real, &priv->mux)) < 0) {
		I2C_UNLOCK(real);
		return ret;
	}

	if (phys && phys->algo->fixup_i2c) {
		phys->algo->fixup_i2c(phys);
	}

#if defined(CONFIG_I2C_IBM_IIC)
	/* mode control is only supported on IBM chipset (AMCC) and not by Freescale PowerQuicks */
	/* Set the 400Khz for this device if enabled */ 
	if (priv && (priv->i2c_cap & I2C_CAP_FSM_400KHZ) && phys) {
		iic_mode_control(phys, 1);
	}
#endif

        ret = real->algo->master_xfer(real, msgs, num);

#if defined(CONFIG_I2C_IBM_IIC)
	/* mode control is only supported on IBM chipset (AMCC) and not by Freescale PowerQuicks */
	/* reset the 400Khz if we have set it - next fixup will clear it */
	if (priv && (priv->i2c_cap & I2C_CAP_FSM_400KHZ) && phys) {
		iic_mode_control(phys, 0);
	}
#endif

	/* Do not care about return value of deselect */	
	priv->mux.deselect(real, &priv->mux);
	I2C_UNLOCK(real);

	return ret;
}


static u32
virt_i2c_func(struct i2c_adapter *adap)
{
	/* FIXME : we should probably return the functionality of the 
	 * real adapter...
	 */
	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_10BIT_ADDR |
	    I2C_FUNC_PROTOCOL_MANGLING;
}

static struct i2c_algorithm virt_i2c_algo = {
	.name = "Virtual I2C algorithm driver",
	.id = I2C_HW_VIRT,
	.master_xfer = virt_i2c_xfer,		/* master_xfer		*/
	.smbus_xfer = NULL,
	.functionality = virt_i2c_func		/* functionality        */
};

/*
 * Description: Register bus structure
 */
static int
i2c_virt_add_bus(struct i2c_adapter *adap)
{
	adap->id |= virt_i2c_algo.id;
	adap->algo = &virt_i2c_algo;
	adap->timeout = VIRT_TIMEOUT;
	adap->retries = VIRT_RETRIES;

	i2c_add_adapter(adap);

	/* scan the I2C bus for valid 7 bit addresses
	 * (ie things that ACK on 1byte read)
	 * TODO: check for 10-bit mode and never run as a slave.
	 */

	if (i2c_scan) {

		int found = 0;
		int i;
		DEB(printk(KERN_INFO "i2c%d: scanning bus. Found ",
		       i2c_adapter_id(adap)));
		for (i = 0; i < 0x7f; i++) {
			int ret;
			struct i2c_msg msg;
			char data[1];
			msg.addr = i;
			msg.buf = data;
			msg.len = 1;
			msg.flags = I2C_M_RD;
			if ((ret = virt_i2c_xfer(adap, &msg, 1)) == 1) {
				printk("0x%02x ", i);
				found++;
			}
		}
		
		if (found)
			DEB(printk("\n"));
		else
			DEB(printk("Nothing\n"));

	}
	return 0;
}

static int
i2c_virt_del_bus(struct i2c_adapter *adap)
{
	int ret;
	if ((ret = i2c_del_adapter(adap)) < 0)
		return ret;
	return 0;
}

/* The 'adapter' driver code: */

static int
i2c_virt_reg(struct i2c_client *client)
{
	return 0;
}
static int
i2c_virt_unreg(struct i2c_client *client)
{
	return 0;
}

/*
 * Called to create a 'virtual' i2c bus which represents a multiplexed bus
 * segment.  Mux_addr and mux_val are passed to the select and deselect
 * callback functions to perform hardware-specific mux control.
 */
struct i2c_adapter *register_virtual_i2c_adap(struct i2c_adapter *real_adap, 
					unsigned long mux_addr, 
					unsigned long mux_val, 
					void *select_cb,
					void *deselect_cb,
					int id)
{
	struct i2c_adapter *adap;
	struct i2c_virt_priv *priv;
	struct i2c_adapter *phys, *temp;

	if (!(adap = kmalloc(sizeof (struct i2c_adapter) + sizeof(struct i2c_virt_priv), 
			GFP_KERNEL))) {
		printk("register_virtual_i2c_adap: Failed allocation\n");
		return NULL;
	}

	memset(adap, 0, sizeof(struct i2c_adapter) + sizeof(struct i2c_virt_priv));

	snprintf(adap->name, sizeof(adap->name)-1,
		"Virtual I2C (i2c%d, mux %02lx:%02lx)",
		i2c_adapter_id(real_adap), mux_addr, mux_val);
	adap->name[sizeof(adap->name)-1] = 0;

	priv = (void*) (adap+1); /* point to first byte of i2c_virt_priv */
	i2c_set_adapdata(adap, priv);
	adap->id = I2C_HW_VIRT;
	adap->algo = NULL;
	//adap->client_register = i2c_virt_reg;
	//adap->client_unregister = i2c_virt_unreg;

	priv->real_adap = real_adap;
	phys = real_adap;
	temp = real_adap;
	while ((temp = i2c_get_phys_adap(phys)) != NULL) {
		phys = temp;
	}
	priv->phys_adap = phys;
	priv->mux.addr = mux_addr;
	priv->mux.value = mux_val;
	priv->mux.select = select_cb;
	priv->mux.deselect = deselect_cb;
	priv->mux.id = id;

	/* knj -
	 * deselect the Mux for this virtual adapter - fixes failover issues
	 *
	 * BOBM - Remove this functionality. The deselect function assumes it
	 * has exclusive access to the global I2C bus. But this CP has not done
	 * a bus arbitration at this point, so this reset causes bus collisions with
	 * the OCP and intermittant I2C errors. No one seems to remember what
	 * failover issues were being fixed by this code. In any case, the proper
	 * way to take care of failover issues is to deal with them in the
	 * FOS failover code.
	 */

	if (i2c_virt_add_bus(adap) < 0) {
 	 	printk(KERN_ERR "i2c-virtual.o: register_virtual_i2c_adap: "
			" i2c_virt_add_bus failed for adapter %s id %d\n",
		            adap->name, id);
		return NULL;
	}

	DEB(printk("i2c%d: Virtual I2C bus "
		"(Physical bus i2c%d, multiplexor %02lx port %ld)\n",
		i2c_adapter_id(adap), i2c_adapter_id(real_adap), 
		mux_addr, mux_val));

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	return adap;
}

int unregister_virtual_i2c_adap(struct i2c_adapter *adap)
{
	int ret;
	if((ret = i2c_virt_del_bus(adap)) != 0)  {
 	 	printk(KERN_ERR "i2c-virtual.o: unregister_virtual_i2c_adap: "
			" i2c_virt_del_bus failed for adapter %s\n",
		            adap->name);
		return ret;
	}

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

EXPORT_SYMBOL(register_virtual_i2c_adap);
EXPORT_SYMBOL(unregister_virtual_i2c_adap);

MODULE_AUTHOR("B Kuschak <www.yahoo.com>, K Jasapara <www.yahoo.com>");
MODULE_DESCRIPTION("Virual I2C driver for multiplexed I2C busses");
MODULE_LICENSE("GPL");


