/*
 *
 *    Copyright (c) 1996-2014 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *	To support multiple levels of multiplexed I2C busses, we have created
 *	a new I2C adapter type (I2C_HW_VIRT), which is a 'virtual' adapter.
 *	The virtual adapters are registered here, based on the topology of the
 *	platform.  The virtual bus represents the bus segment when a given mux
 *	port is enabled.
 *
 * 	To avoid polluting the platform-independent linux I2C driver code,
 * 	we put all
 *	the Brocade-specifics for I2C multiplexing into this file.  We iterate
 *	through each real I2C adapter in the system, and register a callback
 *	for each one.
 *	The virtual adapter driver will call this callback immediately before
 *	accessing the device, allowing us to do platform-specific multiplexor
 *	control from within the linux I2C driver framework.  For those bladed
 *	systems which also require exclusive access to a shared I2C bus,
 *	we arbitrate for
 *	the bus first.
 *
 *    Ulysses -		CPLD multiplexor.  CPLD arbitration logic.
 *    Terminator - 	No Mux, no arbitration.
 *    Meteor - 		CPLD multiplexors, CPLD arbitration logic
 *    Saturn - 		CPLD multiplexors, CPLD arbitration logic
 *    Pulsar - 		I2C bus multiplexors, no arbitration
 *    Elara  -      I2C Bus multiplexors, no arbitration
 */

//#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/i2c-virtual.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <asm/delay.h>
//#include <asm/ocp.h>
#include <asm/io.h>
#include <asm/time.h>
#include <brocade/platform.h>
#include <brocade/swbd.h>
#include <brocade/cpld.h>
#include "i2c-stats.h"

extern sw_bd silkworm;
extern i2c_perf_stats_t *perf_stats;

/* Forwared reference */
struct mpc_i2c;
extern void mpc_bus_flush(struct mpc_i2c *i2c);
int p3041_fixup_locked_bus(struct i2c_adapter *adap);

#ifdef CORE_I2C_DEBUG
extern int core_i2c_debug;
#define DEB(x) if (core_i2c_debug>=1) x;
#define DEB2(x) if (core_i2c_debug>=2) x;
#else
#define DEB(x)
#define DEB2(x)
#endif /* CORE_I2C_DEBUG */

#ifndef elemof
#define elemof(a)	(sizeof(a) / sizeof((a)[0]))
#endif

/* Exclusive bus ownership shouldn't last longer than this */
#define	ARBITRATION_TIMEOUT	(1*HZ)

#define FIXUP_I2C(adap) \
	{ \
		struct i2c_adapter *phys = i2c_get_phys_adap(adap); \
		if (phys && phys->algo->fixup_i2c) { \
			phys->algo->fixup_i2c(phys); \
		} \
	}

#define MUX_METSAT_PS	16
/*
 * Define table of values to be written to
 * CPLD register to connect CP I2C to FRUs via
 * an analog mux (not controlled via I2C).
 * Each mux output represents 1 virtual bus.
 * The virtual bus numbers always start at 2
 * because physical busses 0, 1 get virtual bus
 * numbers 0, 1, respectively.
 *
 * This table does not need to be in any particular order. The
 * order simply determines the virtual bus number assignment
 * for each FRU. The virtual bus number determined here must
 * be used in the i2c_client_table_t device tables in platform module
 */

/* meteor_mux_map same for SWBD21 & SWBD42 */
static
uint8_t	meteor_mux_map[] = {
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, /* slots 1-10; virt busses 2-11 */
	11, 12, /* Core slot 0 and core 1; virt busses 12-13 */
	MUX_METSAT_PS, /* Power supplies; virt bus 14 */
	17, 18, /* WWN0  and WWN1; virt busses 15,16 */
	19, 20, 21, /* blower 1, 2, 3; virt busses 17-19 */
	};

/*
 * sprint_mux_map same for Sprint and descendents. Cannot
 * just use the meteor/saturn map because the power supplies
 * and fans are different - They follow the Pulsar model where
 * the power supplies both have the same slave address,
 * and the fans have no I2C devices. There is a single fan
 * controller on the main board instead of one on each fan
 * assembly.
 *
 */
static
uint8_t	sprint_mux_map[] = {
	1, 5,		/* Slot 1, 5; virt bus 2, 3 */
	16, 20,		/* Power supplies 1,2; virt busses 4,5 */
	17, 18,		/* WWN0, WWN1; virt busses 6, 7 */
	19,			/* Fan controller; virt bus 8 */
};
/*
*Spear mux table need modification
*to 4  power supply *
*/
static
uint8_t	spear_mux_map[] = {
	1, 5,		/* Slot 1, 5; virt bus 2, 3 */
	16, 20,		/* Power supplies 1,2; virt busses 4,5 */
	17, 18,		/* WWN0, WWN1; virt busses 6, 7 */
	19,		/* Fan controller; virt bus 8 */
	24, 11,         /* Power supply 3,2 virt bus 9,10 */

};
#define MARATHONJR_VIRTUAL_BUS 2
#define SPEAR_VIRTUAL_BUS 2
/* neptune_mux_map same for SWBD62 */
static
uint8_t	neptune_mux_map[] = {
	0x01, 0x02, 0x03, 0x04, /* Slots 1-4 port blades; virt busses 2-5 */
    0x0B,                   /* Slot 5 Core blade 0; virt bus 6 */
    0x05, 0x06,             /* Slots 6-7 CP0,1; Virt busses 7,8 */
    0x0C,                   /* Slot 8 Core blade 1; virt bus 9 */
    0x07, 0x08, 0x09, 0x0A, /* Slots 9-12 port blades; virt busses 10-13 */
	0x11, 0x12,             /* WWN0  and WWN1; virt busses 14,15 */
	0x13, 0x14, 0x15,       /* blower 1, 2, 3; virt busses 16-18 */
    0x16, 0x17, 0x18, 0x19  /* PS 1, 2, 3; virt busses 19-22 */
};

/* pluto_mux_map for SWBD77 */
static
uint8_t	pluto_mux_map[] = {
	0x01, 0x02,		/* Slots 1-2 port blades; virt busses 2-3 */
    0x0B,           /* Slot 3 Core blade 0; virt bus 4 */
    0x05, 0x06,     /* Slots 4-5 CP0,1; Virt busses 5,6 */
    0x0C,           /* Slot 6 Core blade 1; virt bus 7 */
    0x07, 0x08, 	/* Slots 7-8 port blades; virt busses 8-9 */
	0x11, 0x12,		/* WWN0  and WWN1; virt busses 10,11 */
	0x13, 0x14,		/* blower 1, 2; virt busses 12-13 */
    0x16, 0x17		/* PS 1, 2; virt busses 14-15 */
};

static spinlock_t i2c_spinlock = SPIN_LOCK_UNLOCKED;

/* pointer to the list of i2c_bus_mappings */
static LIST_HEAD(i2c_map_list);
struct i2c_bus_mapping {
	struct i2c_adapter	*real;
	struct i2c_adapter	*virt;
	unsigned long		mux_addr;
	unsigned long		mux_val;

	/* private data */
	struct list_head	list;
};


static struct i2c_adapter *add_virt_bus(struct i2c_adapter *real,
                                        unsigned long mux_addr,
                                        unsigned long mux_val,
                                        void *select_cb,
                                        void *deselect_cb,
										int id);

static int remove_virt_bus(struct i2c_adapter *virt);
static int cpjrplus_fixup_locked_bus(struct i2c_adapter *adap);


/*
 * int elara_recover_locked_iic1_bus()
 *
 * Description:
 *  Ensures that there are no "stuck" devices on the selected i2c bus.
 *  Stuck devices are those that are in the middle of processing a
 *  read or write transaction.  As such, their internal state machines
 *  may not allow them to respond to new transactions.
 *
 * Elara CPLD has I2C Bit Banging Flush register, which can be used
 * to get IICBus1 to get out of its stuck state and then clears the
 * state machine from its intermediate state.
 *
 * Returns:
 *    0 success
 *   -1 flush done but error detected
 *   -2 flush not done
 *
 */
static int
// elara_recover_locked_iic1_bus(void) Hari Narayanan 09/28/2009
elara_recover_locked_iic1_bus(struct i2c_adapter *adap)
{
	u_char status;
	int error = 0;
	u_char ctrl;

	/* clear any residual status and then flush it */
	cpld_write_register(ELARA_CPLD_IIC_FLUSH_CTRL, 0);

	/* Enable the I2C Bus Flush */
	cpld_write_register(ELARA_CPLD_IIC_FLUSH_CTRL,
								ELARA_CPLD_BUS_FLUSH_ENABLE);

	udelay(200);

	/* Read the status after bus flush */
	cpld_read_register(ELARA_CPLD_IIC_FLUSH_STATUS, &status);

	if (status & ELARA_CPLD_BUS_FLUSH_DONE) {

		if (status & ELARA_CPLD_BUS_FLUSH_ERROR) {
			cpld_read_register(ELARA_CPLD_IIC_FLUSH_CTRL, &ctrl);
			/* this will be enabled after testing  */
			printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
			 =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
			error = -1;
		}

	} else {
		cpld_read_register(ELARA_CPLD_IIC_FLUSH_CTRL, &ctrl);
		/* this will be enabled after testing  */
		printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
		 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
		error = -2;
	}

	/* reset the status register by clearing the FLUSH_ENABLE */
	cpld_write_register(ELARA_CPLD_IIC_FLUSH_CTRL, 0);

	return (error);
}

/* @kevin */
/*
 * int elara2_recover_locked_iic1_bus()
 *
 * Description:
 *  Ensures that there are no "stuck" devices on the selected i2c bus.
 *  Stuck devices are those that are in the middle of processing a
 *  read or write transaction.  As such, their internal state machines
 *  may not allow them to respond to new transactions.
 *
 * elara2 CPLD has I2C Bit Banging Flush register, which can be used
 * to get IICBus1 to get out of its stuck state and then clears the
 * state machine from its intermediate state.
 *
 * Returns:
 *    0 success
 *   -1 flush done but error detected
 *   -2 flush not done
 *
 */
static int elara2_recover_locked_iic1_bus(void)
{
    u_char status;
    int error = 0;
    int retrycnt = 3; /* @kevin : for flushing retry */
    u_char ctrl;

    do {
        /* clear any residual status and then flush it */
        cpld_write_register(ELARA2_CPLD_IIC_FLUSH_CTRL, 0);

        /* Enable the I2C Bus Flush */
        cpld_write_register(ELARA2_CPLD_IIC_FLUSH_CTRL,
                            ELARA2_CPLD_BUS_FLUSH_ENABLE);

        udelay(200);

        /* Read the status after bus flush */
        cpld_read_register(ELARA2_CPLD_IIC_FLUSH_STATUS, &status);

        error = 0; /* @kevin */
        if (status & ELARA2_CPLD_BUS_FLUSH_DONE) {

            if (status & ELARA2_CPLD_BUS_FLUSH_ERROR) {
                cpld_read_register(ELARA2_CPLD_IIC_FLUSH_CTRL, &ctrl);
            //  printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush "
            //      "status regx =  %#x ctrl reg =  %#x try %d \n",
            //       __FUNCTION__, __LINE__, status, ctrl, retrycnt);
                error = -1;
            }

        } else {
            cpld_read_register(ELARA2_CPLD_IIC_FLUSH_CTRL, &ctrl);
            //printk(KERN_DEBUG "%s %d:  Flush not done status reg = %#x ctrl "
            //  "reg =  %#x try %d \n", __FUNCTION__, __LINE__, status, ctrl,
            //  retrycnt);
            error = -2;
        }
    } while( error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
    cpld_write_register(ELARA2_CPLD_IIC_FLUSH_CTRL, 0);

    return (error);
}

/* @kevin */
/*
 * int callisto_recover_locked_iic1_bus()
 *
 * Description:
 *  Ensures that there are no "stuck" devices on the selected i2c bus.
 *  Stuck devices are those that are in the middle of processing a
 *  read or write transaction.  As such, their internal state machines
 *  may not allow them to respond to new transactions.
 *
 * callisto CPLD has I2C Bit Banging Flush register, which can be used
 * to get IICBus1 to get out of its stuck state and then clears the
 * state machine from its intermediate state.
 *
 * Returns:
 *    0 success
 *   -1 flush done but error detected
 *   -2 flush not done
 *
 */
static int callisto_recover_locked_iic1_bus(void)
{
    u_char status;
    int error = 0;
    int retrycnt = 3; /* @kevin : for flushing retry */
    u_char ctrl;

    do {
        /* clear any residual status and then flush it */
        cpld_write_register(CALLISTO_CPLD_IIC_FLUSH_CTRL, 0);

        /* Enable the I2C Bus Flush */
        cpld_write_register(CALLISTO_CPLD_IIC_FLUSH_CTRL,
                            CALLISTO_CPLD_BUS_FLUSH_ENABLE);

        udelay(200);

        /* Read the status after bus flush */
        cpld_read_register(CALLISTO_CPLD_IIC_FLUSH_STATUS, &status);

        error = 0; /* @kevin */
        if (status & CALLISTO_CPLD_BUS_FLUSH_DONE) {

            if (status & CALLISTO_CPLD_BUS_FLUSH_ERROR) {
                cpld_read_register(CALLISTO_CPLD_IIC_FLUSH_CTRL, &ctrl);
                //printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush "
                //  "status regx =  %#x ctrl reg =  %#x try %d \n",
                //   __FUNCTION__, __LINE__, status, ctrl, retrycnt);
                error = -1;
            }

        } else {
            cpld_read_register(CALLISTO_CPLD_IIC_FLUSH_CTRL, &ctrl);
            //printk(KERN_DEBUG "%s %d:  Flush not done status reg = %#x ctrl "
            //  "reg =  %#x try %d \n", __FUNCTION__, __LINE__, status, ctrl,
            //  retrycnt);
            error = -2;
        }
    } while( error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
    cpld_write_register(CALLISTO_CPLD_IIC_FLUSH_CTRL, 0);

    return (error);
}

/* @kevin */
/*
 * int carme_recover_locked_iic1_bus()
 *
 * Description:
 *  Ensures that there are no "stuck" devices on the selected i2c bus.
 *  Stuck devices are those that are in the middle of processing a
 *  read or write transaction.  As such, their internal state machines
 *  may not allow them to respond to new transactions.
 *
 * callisto CPLD has I2C Bit Banging Flush register, which can be used
 * to get IICBus1 to get out of its stuck state and then clears the
 * state machine from its intermediate state.
 *
 * Returns:
 *    0 success
 *   -1 flush done but error detected
 *   -2 flush not done
 *
 */
static int carme_recover_locked_iic1_bus(void)
{
    u_char status;
    int error = 0;
    int retrycnt = 3; /* @kevin : for flushing retry */
    u_char ctrl, ctrl_byte;

    do {
        /* clear any residual status and then flush it */
        cpld_write_register(CARME_CPLD_IIC_FLUSH_CTRL, 0);

        /* Enable the I2C Bus Flush */
        cpld_write_register(CARME_CPLD_IIC_FLUSH_CTRL,
                            CARME_CPLD_BUS_FLUSH_ENABLE);

        udelay(200);

        /* Read the status after bus flush */
        cpld_read_register(CARME_CPLD_IIC_FLUSH_STATUS, &status);

        error = 0;
        if (status & CARME_CPLD_BUS_FLUSH_DONE) {

            if (status & CARME_CPLD_BUS_FLUSH_ERROR) {
                cpld_read_register(CARME_CPLD_IIC_FLUSH_CTRL, &ctrl);
                /* this will be enabled after testing  */
                //printk("%s %d: Flush done but Error in bus flush status regx \
                // =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
                error = -1;
            }

        } else {
            cpld_read_register(CARME_CPLD_IIC_FLUSH_CTRL, &ctrl);
            /* this will be enabled after testing  */
            //printk("%s %d:  Flush not done status reg = %#x ctrl reg =  %#x\n",
            //__FUNCTION__, __LINE__, status, ctrl);
            error = -2;
        }
    } while( error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
    cpld_write_register(CARME_CPLD_IIC_FLUSH_CTRL, 0);

    return (error);
}
/* Bit banging I2C bus flush operation supported by CPLD */
static int
starlifter_recover_locked_iic1_bus(void)
{
	u_int status;
	int error = 0, ret;
    	int retrycnt = 3; /* @kevin : for flushing retry */
	u_int ctrl_reg;
	u_int ctrl;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register32(STARLIFTER_CPLD_IIC_CTRL, &ctrl_reg);

	cpld_write_register32(STARLIFTER_CPLD_IIC_CTRL,
			(ctrl_reg & ~STARLIFTER_CPLD_IIC_MUX_SEL));
    do {
	/* clear any residual status and then flush it */
	cpld_write_register32(STARLIFTER_CPLD_IIC_FLUSH_CTRL, 0);

	/* Enable the I2C Bus Flush */
	cpld_write_register32(STARLIFTER_CPLD_IIC_FLUSH_CTRL,
								STARLIFTER_CPLD_BUS_FLUSH_ENABLE);

	udelay(200);

	/* Read the status after bus flush */
	ret = cpld_read_register32(STARLIFTER_CPLD_IIC_FLUSH_STATUS, &status);
	error = 0;
	if (status & STARLIFTER_CPLD_BUS_FLUSH_DONE) {

		if (status & STARLIFTER_CPLD_BUS_FLUSH_ERROR) {
			cpld_read_register32(STARLIFTER_CPLD_IIC_FLUSH_CTRL, &ctrl);
			/* this will be enabled after testing  */
			/* printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
			 =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);*/
			error = -1;
		}
		/* printk(KERN_DEBUG "Flush success, rc=%d, st=%x,ctrl=%x\n",retrycnt, status, ctrl); */
	} else {
		cpld_read_register32(STARLIFTER_CPLD_IIC_FLUSH_CTRL, &ctrl);
		/* this will be enabled after testing  */
		/* printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
		 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl); */
		error = -2;
	}
    } while( error & retrycnt--);

	/* reset the status register by clearing the FLUSH_ENABLE */
	cpld_write_register32(STARLIFTER_CPLD_IIC_FLUSH_CTRL, 0);
	/* Write back original control reg */
	cpld_write_register32(STARLIFTER_CPLD_IIC_CTRL, ctrl_reg);
	/* Give some time to select CPU as I2C bus1 master */
	udelay(200);

	return (error);
}

static int
wedge_recover_locked_iic1_bus(void)
{
	uint8_t status;
	int error = 0, ret;
	int retrycnt = 3; /* @kevin : for flushing retry */
	uint8_t ctrl_reg, ctrl;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register(WEDGE_CPLD_IIC_CTRL, &ctrl_reg);

	cpld_write_register(WEDGE_CPLD_IIC_CTRL,
			(ctrl_reg & ~WEDGE_CPLD_IIC_MUX_SEL));
	do {
	/* clear any residual status and then flush it */
	cpld_write_register(WEDGE_CPLD_IIC_FLUSH_CTRL, 0);

	/* Enable the I2C Bus Flush */
	cpld_write_register(WEDGE_CPLD_IIC_FLUSH_CTRL,
								WEDGE_CPLD_BUS_FLUSH_ENABLE);
	udelay(200);

	/* Read the status after bus flush */
	ret = cpld_read_register(WEDGE_CPLD_IIC_FLUSH_STATUS, &status);
	error = 0;
	if (status & WEDGE_CPLD_BUS_FLUSH_DONE) {

		if (status & WEDGE_CPLD_BUS_FLUSH_ERROR) {
			cpld_read_register(WEDGE_CPLD_IIC_FLUSH_CTRL, &ctrl);
			/* this will be enabled after testing  */
			/* printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
			 =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);*/
			error = -1;
		}
		/* printk(KERN_DEBUG "Flush success, rc=%d, st=%x,ctrl=%x\n",retrycnt, status, ctrl); */
	} else {
		cpld_read_register(WEDGE_CPLD_IIC_FLUSH_CTRL, &ctrl);
		/* this will be enabled after testing  */
		/* printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
		 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl); */
		error = -2;
	}
	} while( error & retrycnt--);

	/* reset the status register by clearing the FLUSH_ENABLE */
	cpld_write_register(WEDGE_CPLD_IIC_FLUSH_CTRL, 0);
	/* Write back original control reg */
	cpld_write_register(WEDGE_CPLD_IIC_CTRL, ctrl_reg);
	/* Give some time to select CPU as I2C bus1 master */
	udelay(200);

	return (error);
}


static int
skybolt_recover_locked_iic1_bus(void)
{
	uint8_t status;
	int error = 0, ret;
    int retrycnt = 3; /* @kevin : for flushing retry */
	uint8_t ctrl_reg, ctrl;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register(SKYBOLT_CPLD_IIC_CTRL, &ctrl_reg);

    cpld_write_register(SKYBOLT_CPLD_IIC_CTRL,
			(ctrl_reg & ~SKYBOLT_CPLD_IIC_MUX_SEL));
    do {
	/* clear any residual status and then flush it */
	cpld_write_register(SKYBOLT_CPLD_IIC_FLUSH_CTRL, 0);

	/* Enable the I2C Bus Flush */
	cpld_write_register(SKYBOLT_CPLD_IIC_FLUSH_CTRL,
								SKYBOLT_CPLD_BUS_FLUSH_ENABLE);
	udelay(200);

	/* Read the status after bus flush */
	ret = cpld_read_register(SKYBOLT_CPLD_IIC_FLUSH_STATUS, &status);
	error = 0;
	if (status & SKYBOLT_CPLD_BUS_FLUSH_DONE) {

		if (status & SKYBOLT_CPLD_BUS_FLUSH_ERROR) {
			cpld_read_register(SKYBOLT_CPLD_IIC_FLUSH_CTRL, &ctrl);
			/* this will be enabled after testing  */
			/* printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
			 =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);*/
			error = -1;
		}
		/* printk(KERN_DEBUG "Flush success, rc=%d, st=%x,ctrl=%x\n",retrycnt, status, ctrl); */
	} else {
		cpld_read_register(SKYBOLT_CPLD_IIC_FLUSH_CTRL, &ctrl);
		/* this will be enabled after testing  */
		/* printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
		 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl); */
		error = -2;
	}
    } while( error & retrycnt--);

	/* reset the status register by clearing the FLUSH_ENABLE */
	cpld_write_register(SKYBOLT_CPLD_IIC_FLUSH_CTRL, 0);
	/* Write back original control reg */
	cpld_write_register(SKYBOLT_CPLD_IIC_CTRL, ctrl_reg);
	/* Give some time to select CPU as I2C bus1 master */
	udelay(200);

	return (error);
}

static int
castorT_fixup_locked_iic1_bus(struct i2c_adapter *adap)
{
    u_int status;
    int error = 0;
	int retrycnt = 3;
	u_int ctrl_reg;
	u_int ctrl;
	u_int model_id;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register32(CASTOR_CPLD_IIC_CTRL, &ctrl_reg);

	cpld_write_register32(CASTOR_CPLD_IIC_CTRL,
			(ctrl_reg & ~CASTOR_CPLD_IIC_MUX_SEL));

	do {

	    /* clear any residual status and then flush it */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

	    /* Enable the I2C Bus Flush */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL,
                                CASTOR_CPLD_BUS_FLUSH_ENABLE);

	    udelay(200);

    	/* Read the status after bus flush */
	    cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_STATUS, &status);

    	if (status & CASTOR_CPLD_BUS_FLUSH_DONE) {

        	if (status & CASTOR_CPLD_BUS_FLUSH_ERROR) {
            	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	            /* this will be enabled after testing  */
    	       //  printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
        	     =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
            	error = -1;
        	}

    	} else {
        	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	        /* this will be enabled after testing  */
    	  //  printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
        	 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
	        error = -2;
    	}
	} while(error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
   	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

    /* Write back original control reg */
    cpld_write_register32(CASTOR_CPLD_IIC_CTRL, ctrl_reg);
    /* Give some time to select CPU as I2C bus1 master */
    udelay(200);

    /*
     * Fix for Defect 519263
     * the EEPROM was not sending the ACK to the access, So
     * we flush the i2c master for evry boot
     */
    if(error = p3041_fixup_locked_bus(adap)){
    	printk(KERN_ERR "%s : i2c Flush didn't completed : %d\n", __FUNCTION__, error);
	error = -3;
    }

    return (error);
}

static int
castor_fixup_locked_iic1_bus(struct i2c_adapter *adap)
{
    u_int status;
    int error = 0;
	int retrycnt = 3;
	u_int ctrl_reg;
	u_int ctrl;
	u_int model_id;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register32(CASTOR_CPLD_IIC_CTRL, &ctrl_reg);

	cpld_write_register32(CASTOR_CPLD_IIC_CTRL,
			(ctrl_reg & ~CASTOR_CPLD_IIC_MUX_SEL));

	do {

	    /* clear any residual status and then flush it */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

	    /* Enable the I2C Bus Flush */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL,
                                CASTOR_CPLD_BUS_FLUSH_ENABLE);

	    udelay(200);

    	/* Read the status after bus flush */
	    cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_STATUS, &status);

    	if (status & CASTOR_CPLD_BUS_FLUSH_DONE) {

        	if (status & CASTOR_CPLD_BUS_FLUSH_ERROR) {
            	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	            /* this will be enabled after testing  */
    	        // printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
        	     =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
            	error = -1;
        	}

    	} else {
        	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	        /* this will be enabled after testing  */
    	    // printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
        	 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
	        error = -2;
    	}
	} while(error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
   	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

    /* Write back original control reg */
    cpld_write_register32(CASTOR_CPLD_IIC_CTRL, ctrl_reg);
    /* Give some time to select CPU as I2C bus1 master */
    udelay(200);

    /*
     * Fix for Defect 519263
     * the EEPROM was not sending the ACK to the access, So
     * we flush the i2c master for evry boot
     */
    if(error = p3041_fixup_locked_bus(adap)){
	printk(KERN_ERR "%s : i2c Flush didn't completed : %d\n", __FUNCTION__, error);
	error = -3;
		
    }
    return (error);
}


static int
blackbird_fixup_locked_iic1_bus(struct i2c_adapter *adap)
{
    u_int status;
    int error = 0;
	int retrycnt = 3;
	u_int ctrl_reg;
	u_int ctrl;
	u_int model_id;

	/* I2C MUX ENABLE bit is set by default.
	 * Clear I2C_MUX_SEL bit:4 to select sysFPGA as
	 * Bus Master.
	 */
	cpld_read_register32(CASTOR_CPLD_IIC_CTRL, &ctrl_reg);

	cpld_write_register32(CASTOR_CPLD_IIC_CTRL,
			(ctrl_reg & ~CASTOR_CPLD_IIC_MUX_SEL));

	do {

	    /* clear any residual status and then flush it */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

	    /* Enable the I2C Bus Flush */
    	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL,
                                CASTOR_CPLD_BUS_FLUSH_ENABLE);

	    udelay(200);

    	/* Read the status after bus flush */
	    cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_STATUS, &status);

    	if (status & CASTOR_CPLD_BUS_FLUSH_DONE) {

        	if (status & CASTOR_CPLD_BUS_FLUSH_ERROR) {
            	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	            /* this will be enabled after testing  */
    	        // printk(KERN_DEBUG "%s %d: Flush done but Error in bus flush status regx \
        	     =  %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
            	error = -1;
        	}

    	} else {
        	cpld_read_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, &ctrl);
	        /* this will be enabled after testing  */
    	    // printk(KERN_DEBUG "%s %d:  Flush not done status reg = \
        	 %#x ctrl reg =  %#x\n",  __FUNCTION__, __LINE__, status, ctrl);
	        error = -2;
    	}
	} while(error & retrycnt--);

    /* reset the status register by clearing the FLUSH_ENABLE */
   	cpld_write_register32(CASTOR_CPLD_IIC_FLUSH_CTRL, 0);

    /* Write back original control reg */
    cpld_write_register32(CASTOR_CPLD_IIC_CTRL, ctrl_reg);
    /* Give some time to select CPU as I2C bus1 master */
    udelay(200);

    /*
     * Fix for Defect 519263
     * the EEPROM was not sending the ACK to the access, So
     * we flush the i2c master for evry boot
     */
    if(error = p3041_fixup_locked_bus(adap)){
	printk(KERN_ERR "%s : i2c Flush didn't completed : %d\n", __FUNCTION__, error);
	error = -3;		
    }

    return (error);
}



/*
 * Meteor uses multiplexors controlled by the CPLD
 */
static int meteor_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;
	int ret = 0;

	cpld_read_register(MET_I2C_SELECT, &data);

	/* Ignore the SEL_MASK value as set */
	data = (char) (data & ~MET_CP_I2C_SEL_MASK);

	/* Maintain the other bits as is */
	data |= (char) ((mux->value & MET_CP_I2C_SEL_MASK)
		| MET_CP_I2C_WRITE_MASK);

	if((ret = cpld_write_register(MET_I2C_SELECT, data)) < 0) {
		printk("%s: mux selection failed %d id %d\n",
			__FUNCTION__, data, mux->id);
		return (ret);
	}

	DEB2(printk("Meteor I2C mux selected [muxval: 0x%02x]! id %d\n",
				data, mux->id));

	return ret;
}

static int meteor_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;
	cpld_read_register(MET_I2C_SELECT, &data);

	if (mux->value == MUX_METSAT_PS) {
		/* Add a jiffies delay for WP hold time */
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(2);
	}


	/* Last position is reserved as 'disconnected' */
	data = (data | MET_CP_I2C_SEL_MASK);

	if(cpld_write_register(MET_I2C_SELECT, data) < 0) {
		printk("%s: mux deselection failed %d id %d\n",
			__FUNCTION__, data, mux->id);
		/* We do not care to return an error */
	}

	DEB2(printk("Meteor I2C mux deselected - muxval 0x%02x! id %d\n",
				data, mux->id));

	return 0;
}

/*
 * Zentron uses multiplexors controlled by the CPLD
 * Use this method with the 8548 internal I2C controller
 */
static int neptune8548_select_mux(struct i2c_adapter *adap,
								  struct i2c_mux_ctrl *mux)
{
	char data;
	int ret = 0;

	cpld_read_register(MET_I2C_SELECT, &data);

	/* Ignore the SEL_MASK value as set */
	data = (char) (data & ~MET_CP_I2C_SEL_MASK);

	/* Maintain the other bits as is */
	data |= (char) ((mux->value & MET_CP_I2C_SEL_MASK));

	if (mux->write_enable) {
		data |= MET_CP_I2C_WRITE_MASK;
	}
	if((ret = cpld_write_register(MET_I2C_SELECT, data)) < 0) {
		printk("%s: mux selection failed %d id %d\n",
			__FUNCTION__, data, mux->id);
		goto error_exit;
	}

	DEB2(printk("Meteor I2C mux selected [muxval: 0x%02x]! id %d\n",
				data, mux->id));

 error_exit:

	return ret;
}


/*
 * Zentron uses multiplexors controlled by the CPLD
 * Use this method with the external PCA 9665 controller
 */
static int neptune_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;
	int ret = 0;

	/* The PCA controller is mostly #23 */
	DEB2(printk("PCA 9665 IIC  Select Mux called for 0x%x \n", mux->value));
	/* First select the Philips IIC */
	cpld_write_register(0x42, 0x1);

	/* Enable the Philips I2C interrupt */
	cpld_read_register(0x6F, &data);
	data &= ~(0x2);
	cpld_write_register(0x6F, data);

	cpld_read_register(MET_I2C_SELECT, &data);

	/* Ignore the SEL_MASK value as set */
	data = (char) (data & ~MET_CP_I2C_SEL_MASK);

	/* Maintain the other bits as is */
	data |= (char) ((mux->value & MET_CP_I2C_SEL_MASK));

	if (mux->write_enable) {
		data |= MET_CP_I2C_WRITE_MASK;
	}

	if((ret = cpld_write_register(MET_I2C_SELECT, data)) < 0) {
		printk("%s: mux selection failed %d\n",
			__FUNCTION__, data);
		return (ret);
	}

	return 0;
}

static int neptune_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;

	cpld_read_register(MET_I2C_SELECT, &data);

	if (mux->write_enable) {
		/* Add a jiffies delay for WP hold time */
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(2);
	}

	/*
	 * Last position is reserved as 'disconnected'.
	 * Ensure write enable is turned off
	 */
	data = ((data | MET_CP_I2C_SEL_MASK)&(~MET_CP_I2C_WRITE_MASK));

	if(cpld_write_register(MET_I2C_SELECT, data) < 0) {
		printk(KERN_ERR "%s: mux deselection failed %d id %d\n",
			__FUNCTION__, data, mux->id);
		/* We do not care to return an error */
	}
	/* Select the CPU IIC controller */

	cpld_write_register(0x42, 0x0);
	DEB2(printk("Meteor I2C mux deselected - muxval 0x%02x! id %d\n",
				data, mux->id));

	return 0;
}

/* Blazer2 VPD is shared by CPU and backplane, there is 3 I2C switches to
 * isolate the CPU and backplane I2C bus, and they are controlled by CPLD.
 * I treated it as virtual I2C bus.
 */
static int blazer2_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;
	int ret=0,retry;

	cpld_read_register(BLAZER2_BACKPLANE_SUPPORT_REG,&data);
	if(data & MET_CP_I2C_WR_ACTIVE_MASK)
	{
		/*I2C bus has been granted by CPLD.*/
		return 0;
	}

	retry=0;
	do{
		data |= MET_CP_I2C_WRITE_MASK;
		if((ret = cpld_write_register(BLAZER2_BACKPLANE_SUPPORT_REG, data)))
		{
			printk("%s: mux selection failed %d id %d\n",
				__FUNCTION__, data, mux->id);
			return ret;
		}
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
		cpld_read_register(BLAZER2_BACKPLANE_SUPPORT_REG,&data);
		if(data & MET_CP_I2C_WR_ACTIVE_MASK) break;
		retry++;

	}while(retry<3);

	if(retry>=3)
	{
		 printk("%s: mux selection failed %d id %d\n",
				 __FUNCTION__, data, mux->id);
		 return -1;
	}
	return 0;

}
static int blazer2_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	char data;

	cpld_read_register(BLAZER2_BACKPLANE_SUPPORT_REG,&data);
	data &= ~MET_CP_I2C_WRITE_MASK;
	cpld_write_register(BLAZER2_BACKPLANE_SUPPORT_REG,data);
	return 0;
}


/* Pulsar uses PCA9544 and PCA9548 I2c-controlled bus switches.  We try to
 * share code for these two different parts.
 */
static int pcf954x_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux,
				unsigned long value)
{
	struct i2c_msg msg;
	int i2c_switch_addr = mux->addr;
	char buf[1];
	int ret = 0;

	/* send I2C commands to the bus switch(es) to select the right path */
	if (adap->algo->master_xfer) {
		msg.addr	= i2c_switch_addr;
		msg.flags	= 0;
		msg.len		= 1;
		buf[0] 		= value;
		msg.buf 	= buf;

		if((ret = adap->algo->master_xfer(adap,&msg,1)) < 0) {
			printk("%s: Failed to select the I2C mux "
					"(addr=%02hx, val=%02hx, err=%d id=%d)!\n",
					__FUNCTION__, i2c_switch_addr, (int)value, ret,
				mux->id);
		}
		else {
			DEB2(printk("Pulsar I2C mux selected "
				"[i2c addr: %hX, port: %d id: %d]!\n",
				i2c_switch_addr, value, mux->id));
		}
	}
	else {
		printk(KERN_ERR "pcf954x_select_mux: no master_xfer?? id%d\n",
			   mux->id);
	}

	return ret;
}

static int pcf954x_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux,
				unsigned long value)
{
	struct i2c_msg msg;
	int i2c_switch_addr = mux->addr;
	char buf[1];
	int ret = 0;

	/* send I2C commands to the bus switch(es) to select the right path */
	if (adap->algo->master_xfer) {
		msg.addr	= i2c_switch_addr;
		msg.flags	= 0;
		msg.len		= 1;
		buf[0] 		= 0;	/* select nothing */
		msg.buf 	= buf;

		if((ret = adap->algo->master_xfer(adap,&msg,1)) < 0) {
			printk("Failed to deselect the I2C multiplexor "
					"(addr=%02hx, val=%02hx, err=%d id=%d)!\n",
					i2c_switch_addr, 0, ret, mux->id);

		}
		else {
			DEB2(printk("Pulsar I2C mux deselected "
				"[i2c addr: %hX id: %d]!\n",
				i2c_switch_addr, mux->id));
		}
	}
	else {
		printk(KERN_ERR "pcf954x_deselect_mux: no master_xfer?? id %d\n",
			   mux->id);
	}

	/*
	 * The error will be printk'ed - but we do not report error in case of
	 * deselection failure
	 */
	return (0);
}

/*
 * The 9544 and 9548 use different register definitions, hence these two
 * functions
 */
static int pcf9544_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	int i2c_mux_sel     = mux->value;
	/* FIXME - use #defines */
	return(pcf954x_select_mux(adap, mux, ((i2c_mux_sel & 0x3) + 4)));
}

static int pcf9548_select_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	int i2c_mux_sel     = mux->value;
	/* FIXME - use #defines */
	return(pcf954x_select_mux(adap, mux, 1<<(i2c_mux_sel & 0x7)));
}

/* Disabling the mux is the same for both 9544 and 9548 */
static int pcf9544_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	return(pcf954x_deselect_mux(adap, mux, 0));
}

static int pcf9548_deselect_mux(struct i2c_adapter *adap, struct i2c_mux_ctrl *mux)
{
	return(pcf954x_deselect_mux(adap, mux, 0));
}

/*
 * Here we have the arbitration functions for bladed systems, where multiple CPs
 * access a shared I2C bus.Arbitrate and acquire exclusive access to the bus.
 * This code is called from the i2c-core driver _without_ the I2C_LOCK held.
 *
 * Add back the Ulysses arbitration here once Ulysses supports virtual i2c.
 * Meteor/Saturn use the old method of mucking with the CPLD bits.
 */

/*
 * Deleted question and discussion about why arb. is needed....
 * Arbitration is required for all multiple CP systems - PERIOD.
 *	-even if I2C is multi-master capable
 *		i2c controller cannot monitor beyond the mux when segment
 *		is not connected
 *	- insertion and removal of CP's cannot monitor the i2c bus
 */

/*
 * static int meteor_acquire_bus()
 *
 * Description:
 *   arbitrate with the other CP for access to the I2C address mux in the CPLD.
 *   This implements a semaphore-like function between dual  CPs.
 *
 *   On single processor systems, arbitration is not
 *   needed,
 *
 *   Arbitration uses the I2C active bit in the
 *   CPLD. The CP that wins the arbitration leaves
 *   this bit asserted until it has finished its
 *   sequence of I2C transactions.
 *
 *   In the event both CPs want access at the same
 *   time, the Active CP wins. If there is no active
 *   CP, then CP0 (in slot 5) wins. The algorithm is
 *   biased to let the Active CP win.
 *
 *   The function will continue arbitrating for access up
 *   to 2 seconds. If timeout occurrs, a negative value
 *   is returned. Timeout is not unusual, so the caller
 *   should retry the call.
 *
 *   After winning the arbitration, the non-master
 *   CP will make one last check to see if there is
 *   a conflict. If there is conflict, treat it as
 *   losing the arbitration.
 *
 * Input(s):
 *   i2c_adapter, virtual/real adapter structure
 *
 * Output(s):
 *   N/A
 *
 * Return(s):
 *   0 = The caller wins the arbitration
 *   I2C_CP_ERR = Cannot try to arbitrate due to CP being fenced
 *   I2C_TIMEOUT_ERR = Could not win arbitration within the
 *                     2 second limit.
 *  -1 = Cannot try to arbitrate due lack of internal structure
 *       initialization. Generally means I2C access is being
 *       attempted before kernel initialization is completed.
 *
 */
static int
meteor_acquire_bus(struct i2c_adapter *adap)
{

	int OKToAccess = 0;
	int cp_fenced = 0, waitCntr = 0;
	ulong timeOut, flags;
	uint8_t cp_misc, cpld_hk_fence, cpld_i2c, cp_misc_err;
	int ret_value = 0;
	struct timeval	perf_begin;
	uint loops = 0;

	DEB2(printk("meteor acquire bus 0x%08x\n", adap));

	/*
	 * increment arb counter, client counter.
	 * Acquire start time
	 */
	perf_stats->arb_times.count++;
	perf_stats->clients.count++;
	do_gettimeofday(&perf_begin);

	/*
	 * Cannot arbitrate if our housekeeping signals
	 * are fenced. They get fenced under anyone of
	 * these conditions:
	 *  1. CP error out is asserted
	 *  2. HK self-fence is asserted
	 *  3. HK fence-in is asserted (From the OCP).
	 *
	 * If any of these conditions are true, we must
	 * stay off the I2C bus.
	 */
	cpld_read_register(MET_CP_MISC_WRITE, &cp_misc);
	cpld_read_register(MET_CP_HK_FENCE, &cpld_hk_fence);
	cpld_read_register(MET_CP_MISC_ERR, &cp_misc_err);
	if (cp_misc & MET_CP_SECTION_ERR_OUT) {
		cp_fenced = I2C_CP_ERR;
	} else if (cpld_hk_fence & (MET_HK_SELF_FENCE | MET_HK_FORCE_FENCE_IN)) {
		cp_fenced = I2C_FENCE_ERR;
	}

	/* Reset Write-active bit */
	DEB2(printk("meteor_acquire_bus: Clear write-active bit \n"));
	spin_lock_irqsave(&i2c_spinlock, flags);
	cpld_read_register(MET_I2C_SELECT, &cpld_i2c);
	cpld_i2c &= ~(CP_I2C_WR_ACTIVE_MASK);
	cpld_write_register(MET_I2C_SELECT, cpld_i2c);
	spin_unlock_irqrestore(&i2c_spinlock, flags);

	if (cp_fenced != 0) {
		DEB2(printk("\nmeteor_acquire_bus: "
			"CP_MISC_WRITE 0x%2x, "
			"CP_HK_FENCE 0x%2x, "
			"CP_MISC_ERR 0x%2x - "
			"can't arbitrate for i2c bus",
			cp_misc, cpld_hk_fence, cp_misc_err));
		if (current) {
			DEB(printk("\n\tprocess %d (%s)\n",
				current->pid, current->comm));
	  	}

		ret_value = cp_fenced;
		goto END;

	}

	/* time out if can't get access in 1s */
	timeOut = jiffies + HZ;
	loops = 1;
	do {

		spin_lock_irqsave(&i2c_spinlock, flags);
		cpld_read_register(MET_I2C_SELECT, &cpld_i2c);

		/* Check if i2c is being used */
		if (!(cpld_i2c & CP_I2C_RD_ACTIVE_MASK)) {

			/* It's not in use, so notify that I want to access i2c bus */
			cpld_i2c |= CP_I2C_WR_ACTIVE_MASK;
			cpld_write_register(MET_I2C_SELECT, cpld_i2c);
			spin_unlock_irqrestore(&i2c_spinlock, flags);

			DEB2(printk("meteor_acquire_bus: Notify access\n"));

			/*
			 * Waiting to ensure the CPLD bit has time to
			 * propagate from one CP to the other.
			 */
			udelay(50);

			/*
			 * Since I'm not the I2C master, I have to look
			 * out for and handle race conditions. Start by
			 * checking again if i2c is being used (in case two CPs want
			 * to access the i2c bus at the exact same time).
			 * If so, the other CP wins by default since it is
			 * the I2C master. We drop our request, sleep a bit,
			 * and try again.
			 */
			spin_lock_irqsave(&i2c_spinlock, flags);
			cpld_read_register(MET_I2C_SELECT, &cpld_i2c);
			spin_unlock_irqrestore(&i2c_spinlock, flags);

			if (!(cpld_i2c & CP_I2C_RD_ACTIVE_MASK)) {
				OKToAccess = 1;
				DEB2(printk("meteor_acquire_bus: OK to access\n"));
				break;
			} else {
				/*
				 * Here if 2 cp wanted access at the exact same time
				 * CP0 wins, check cp ID differently in diff platform.
				 */
				spin_lock_irqsave(&i2c_spinlock, flags);
				cpld_read_register(MISC_READ_OFFSET, &cp_misc);
				spin_unlock_irqrestore(&i2c_spinlock, flags);
				if (cp_misc & MET_CP_CP_SLOT_ID) { /* CP1 looses */
					OKToAccess = 0;

					/* cpld_delay(25); */
					current->state = TASK_INTERRUPTIBLE;
					schedule_timeout((HZ * 25) / 1000L);

					DEB(printk("meteor_acquire_bus: "
					"lost arbitration - trying again\n"));
				} else {
					OKToAccess = 1;
					DEB2("meteor_acquire_bus: OK to access\n");
					break;
				}
			} /* checking bus the second time */
		} else {
			spin_unlock_irqrestore(&i2c_spinlock, flags);
			/* I2C is already in use, so Re-schedule this process */
			DEB(printk("meteor_acquire_bus: I2C is busy (0x%2x), "
			  "Re-schedule process\n", cpld_i2c));

			/* cpld_delay(25); */
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout((HZ * 25) / 1000L);

			/* ck if time to insert FAIRNESS logic */
			waitCntr += 1;
			if (waitCntr == 2) {
				spin_lock_irqsave(&i2c_spinlock, flags);
				cpld_read_register(MET_I2C_SELECT, &cpld_i2c);
				cpld_i2c |= CP_I2C_WR_ACTIVE_MASK;
				cpld_write_register(MET_I2C_SELECT, cpld_i2c);
				spin_unlock_irqrestore(&i2c_spinlock, flags);
			}
		} /* checking bus the first time */

		loops++;
	} while ((!OKToAccess) && (jiffies < timeOut));

	/*
	 * -----------------------------------------------------
	 * Check to see if we timed out while waiting for access.
	 * Caller should be able to handle possible timeouts.
	 * -----------------------------------------------------
	 */
	if (!OKToAccess) {
		spin_lock_irqsave(&i2c_spinlock, flags);
		cpld_read_register(MET_I2C_SELECT, &cpld_i2c);
		cpld_i2c &= ~(CP_I2C_WR_ACTIVE_MASK);
		cpld_write_register(MET_I2C_SELECT, cpld_i2c);
		spin_unlock_irqrestore(&i2c_spinlock, flags);
		DEB(printk("meteor_acquire_bus: timed out (0x%x)waiting "
			"for I2C access!\n", cpld_i2c));

		/*
		 * increment arb timeout counter
		 * We are using the otherwise unused arb_loops
		 * counter for this purpose.
		 */
		perf_stats->arb_loops.count++;
		ret_value = I2C_TIMEOUT_ERR;
	}

END:

	/* Acquire stop time and calculate elapsed time, min, max, sum */
	perf_stats_xfer_end(&perf_stats->arb_times, &perf_begin);

	/* Calculate min, max, and total times the arb routine has looped */
	PERF_STATS_UPDATE(loops, &perf_stats->arb_loops);

	/* How may clients in this routine */
	PERF_STATS_UPDATE(perf_stats->clients.count, &perf_stats->clients);

	/* Client exiting */
	perf_stats->clients.count--;

	return (ret_value);

}

/*
 * int meteor_release_bus()
 *
 * Description:
 *   Release ownership of the I2C address mux
 *
 *   Do not call this routine arbitrarily. Only
 *   Call it after a call to cpld_arbitrate_i2c_mux()
 *   has been made. Otherwise a warning message will
 *   be displayed.
 *
 * Input(s):
 *   i2c_adapter - virtual/real adapter
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *    0 = success
 *   -1 = Failure. Could not turn off the
 *        the I2C active bit in the CPLD.
 *
 */
static int meteor_release_bus(struct i2c_adapter *adap)
{
	uint8_t data;
	unsigned long flags;

	DEB2(printk("meteor release bus 0x%08x\n", adap));

	/* Release I2C ownership - Clearing Write-active and Write-enable bit */
	spin_lock_irqsave(&i2c_spinlock, flags);
	cpld_read_register(MET_I2C_SELECT, &data);
	data &= ~(CP_I2C_WR_ACTIVE_MASK);
	cpld_write_register(MET_I2C_SELECT, data);
	spin_unlock_irqrestore(&i2c_spinlock, flags);

	return 0;
}

/*-----------------------------------------------------------------------------
 *
 * This function uses the CPLD to drive clocks so that any device
 * that has not completed a bus transaction will finish and release the bus.
 *  1. Set the 3-1 mux to the CPLD
 *  2. Start the clock burst
 *  3. Wait for the clock burst to finsh
 *  4. Check error status
 *  5. Restore the 3-1 mux to its previous setting
 *
 */
int
iic_fixup_locked_bus(struct i2c_adapter *adap,
                            u_char       rsel,  // Select register
                            u_char       rctl,  // Control register
                            u_char       rstt)  // Status register
{
    u_char source, status;
    int error = 0;

    cpld_read_register(rsel,&source);
    cpld_write_register(rsel, CPLD_FPGA_SOURCE);

    /* clear any residual status and then flush it */
    cpld_write_register(rctl, 0);
    cpld_write_register(rctl, CPLD_BUS_FLUSH_ENABLE);

    udelay(200);

    cpld_read_register(rstt, &status);

    if (status & CPLD_BUS_FLUSH_DONE) {

        if (status & CPLD_BUS_FLUSH_ERROR) {

            u_char ctrl;

            cpld_read_register(rctl, &ctrl);
            printk("%s %d: ****** ERROR IN BUS FLUSH 0x%x ctrl 0x%x\n",
                   __FUNCTION__, __LINE__, status, ctrl);
            error = -1;

        }
    } else {

        u_char ctrl;

        cpld_read_register(rctl, &ctrl);
        printk("%s %d: $$$$$$$ BUS FLUSH NOT COMPLETED status 0x%x ctrl 0x%x\n",               __FUNCTION__, __LINE__, status, ctrl);

        error = -2;
    }

    cpld_write_register(rsel, source);

	return (error);
}

//-----------------------------------------------------------------------------
static int
zentron_fixup_locked_bus(struct i2c_adapter *adap)
{
    return (iic_fixup_locked_bus(adap, CPLD_IIC_SOURCE_SELECT,
                                 CPLD_IIC_FLUSH_CTRL, CPLD_IIC_FLUSH_STATUS));
}

//-----------------------------------------------------------------------------
static int
cpjrplus_fixup_locked_bus(struct i2c_adapter *adap)
{
    return (iic_fixup_locked_bus(adap, CPLD_CPJRPLUS_IIC_SOURCE_SELECT,
                        CPLD_CPJRPLUS_IIC_FLUSH_CTRL, CPLD_CPJRPLUS_IIC_FLUSH_STATUS));
}
//-----------------------------------------------------------------------------
static int
thor_fixup_locked_bus(struct i2c_adapter *adap)
{
    return (iic_fixup_locked_bus(adap, THOR_CPLD_IIC_SOURCE_SELECT,
                        THOR_CPLD_IIC_FLUSH_CTRL, THOR_CPLD_IIC_FLUSH_STATUS));
}
//-----------------------------------------------------------------------------
static int
mpc85xx_fixup_locked_bus(struct i2c_adapter *adap)
{
	int rc = 0;
#ifdef CONFIG_I2C_MPC
	struct mpc_i2c *i2c = i2c_get_adapdata(adap);

	if (i2c) {
		mpc_bus_flush(i2c);
	} else {
		printk(KERN_ERR "%s: No mpc_i2c struct. Cannot do I2C bus flush\n",
			   __FUNCTION__);
		rc = -1;
	}
#endif
	return (rc);
}
int
p3041_fixup_locked_bus(struct i2c_adapter *adap)
{
        int rc = 0;
        struct mpc_i2c *i2c = i2c_get_adapdata(adap);

        if (i2c) {
                mpc_bus_flush(i2c);
        } else {
                printk(KERN_ERR "%s: No mpc_i2c struct. Cannot do I2C bus flush\n",
                           __FUNCTION__);
                rc = -1;
        }
        return (rc);
}

/*
 * This function is used to add virtual busses behind a pcf9548 I2C switch
 */
static void
add_pcf9548_virt_bus(struct i2c_adapter *adap, int start_addr, int num_switches, int idx)
{
	int sw, i;

	for (sw = 0; sw < num_switches; sw++) {

		for(i=0; i<=7; i++) {
				add_virt_bus(adap, start_addr + sw, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, idx);
		}
	}
}

//-----------------------------------------------------------------------------
static void
add_virt_bus_castorT(struct i2c_adapter *adap)
{
    int ret;
    struct i2c_adapter *virt[0];
    int i, sfpidx, total_sw = 1;

    adap->algo->fixup_i2c = &castorT_fixup_locked_iic1_bus;

    /*
     * Invoke unlock routine at startup in case any device (such as
     * FRU header/seeprom) was left hung up by reboot.  On castor this
     * will affect the virtual i2c driver's ability to configure the
     * i2c switch during init and lead to init failure.  And because
     * of that failure the unlock never has a chance to be called.
     */
    if ((ret = castorT_fixup_locked_iic1_bus(adap))) {
        printk(KERN_ERR "CastorT i2c unlock, ret=%d\n", ret);
    }


    /* The first level mux (PCA9548) has 8 outputs.
     * Each Mux Output is selected by making that
     * bit (0-7) to 1 in PCA9548 Control Register. Register
     * a virtual bus for each one. As each output is a
     * separate I2C Bus electrically.
     * They are all at the same level on the same board, so
     * use the same ID value. The ID gets displayed by Linux
     * error messages and is used to determine which FRU is
     * causing the problem.
     */

    /* mux val = 0, Power Monitor, DS1780, LM 75's */
    add_virt_bus(adap, (0xE4 >> 1), 0,
            &pcf9548_select_mux, &pcf9548_deselect_mux, 0);

    /* Add all the First level QSFP Mux */
    virt[0] = add_virt_bus(adap, (0xE4 >> 1), 3,
            &pcf9548_select_mux, &pcf9548_deselect_mux, 0);

    sfpidx = 0;
    if (virt[0]) {
    /* GE QSFP Ports 49 through 52 */
        for(i = 0; i <= 3; i++, sfpidx++) {
            add_virt_bus(virt[0], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
        }
        /* Adding the two PS hooked to this MUX */
        add_virt_bus(virt[0], (0xE0 >> 1), 6,
               &pcf9548_select_mux, &pcf9548_deselect_mux, 1);

        add_virt_bus(virt[0], (0xE0 >> 1), 7,
               &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

}
//-----------------------------------------------------------------------------
static void
add_virt_bus_castor(struct i2c_adapter *adap)
{
	int ret;
	struct i2c_adapter *virt[7];
	int i, sfpidx, total_sw = 7;

	adap->algo->fixup_i2c = &castor_fixup_locked_iic1_bus;

	/*
	 * Invoke unlock routine at startup in case any device (such as
	 * FRU header/seeprom) was left hung up by reboot.  On castor this
	 * will affect the virtual i2c driver's ability to configure the
	 * i2c switch during init and lead to init failure.  And because
	 * of that failure the unlock never has a chance to be called.
	 */
	if ((ret = castor_fixup_locked_iic1_bus(adap))) {
		printk(KERN_ERR "Castor i2c unlock, ret=%d\n", ret);
	}

	/* The first level mux (PCA9548) has 8 outputs.
	 * Each Mux Output is selected by making that
	 * bit (0-7) to 1 in PCA9548 Control Register. Register
	 * a virtual bus for each one. As each output is a
	 * separate I2C Bus electrically.
	 * They are all at the same level on the same board, so
	 * use the same ID value. The ID gets displayed by Linux
	 * error messages and is used to determine which FRU is
	 * causing the problem.
	 */

	/* mux val = 0, Power Monitor, DS1780, LM 75's */
	add_virt_bus(adap, (0xE4 >> 1), 0,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* Add all the First level SFP Mux's */
	for (i = 0; i < total_sw; i++) {
		virt[i] = add_virt_bus(adap, (0xE4 >> 1), i+1,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
	}


	/*
	 * Following are 2nd level virtual busses for the SFPs
	 * Muxes are added in the order of the SFP's, check HW spec.
	 * Muxes are all on the same board and on the save level,
	 * so use the same ID for all of them
	 */

	sfpidx = 0;
	if (virt[4]) {
	/* GE SFP Ports 1 through 8 */
		for(i = 0; i <= 7; i++, sfpidx++)
			add_virt_bus(virt[4], (0xE0 >> 1), i,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
	}

	sfpidx = 0;
    if (virt[3]) {
    /* GE SFP Ports 9 through 16 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[3], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[0]) {
    /* GE SFP Ports 17 through 24 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[0], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[1]) {
    /* GE SFP Ports 25 through 32 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[1], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[5]) {
    /* GE SFP Ports 33 through 40 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[5], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[6]) {
    /* GE SFP Ports 41 through 48 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[6], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[2]) {
    /* GE QSFP Ports 49 through 52 */
        for(i = 0; i <= 3; i++, sfpidx++) {
            add_virt_bus(virt[2], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		/* Adding the two PS hooked to this MUX */
        add_virt_bus(virt[2], (0xE0 >> 1), 6,
               &pcf9548_select_mux, &pcf9548_deselect_mux, 1);

        add_virt_bus(virt[2], (0xE0 >> 1), 7,
               &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }
}


//-----------------------------------------------------------------------------
static void
add_virt_bus_rigel1(struct i2c_adapter *adap)
{
	int ret;
	int i, sfpidx, total_sw = 7;
	struct i2c_adapter *virt[total_sw];

	adap->algo->fixup_i2c = &castor_fixup_locked_iic1_bus;

	/*
	 * Invoke unlock routine at startup in case any device (such as
	 * FRU header/seeprom) was left hung up by reboot.  On Rigel this
	 * will affect the virtual i2c driver's ability to configure the
	 * i2c switch during init and lead to init failure.  And because
	 * of that failure the unlock never has a chance to be called.
	 */
	if ((ret = castor_fixup_locked_iic1_bus(adap))) {
		printk(KERN_ERR "Rigel i2c unlock, ret=%d\n", ret);
	}

	/* The first level mux (PCA9548) has 8 outputs.
	 * Each Mux Output is selected by making that
	 * bit (0-7) to 1 in PCA9548 Control Register. Register
	 * a virtual bus for each one. As each output is a
	 * separate I2C Bus electrically.
	 * They are all at the same level on the same board, so
	 * use the same ID value. The ID gets displayed by Linux
	 * error messages and is used to determine which FRU is
	 * causing the problem.
	 */

	/* mux val = 0, Power Monitor, DS1780, LM 75's, VBus 4 (why not 3?) */
	add_virt_bus(adap, (0xE4 >> 1), 0,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* Add all the First level SFP Mux's, VBus 5 - 9, 10, 11 not used, yet */
	for (i = 0; i < total_sw; i++) {
		virt[i] = add_virt_bus(adap, (0xE4 >> 1), i+1,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
	}

	/* QSFP Ports [8:1] VBus 12 - 19 */
	sfpidx = 0;
	if (virt[0]) {
		for(i = 0; i <= 7; i++, sfpidx++) {
			add_virt_bus(virt[0], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* QSFP Ports [16:9] VBus 20 - 27 */
	sfpidx = 0;
	if (virt[1]) {
		for(i = 0; i <= 7; i++, sfpidx++) {
			add_virt_bus(virt[1], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* QSFP Ports [24:17] VBus 28 - 35 */
	sfpidx = 0;
	if (virt[2]) {
		for(i = 0; i <= 7; i++, sfpidx++) {
			add_virt_bus(virt[2], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* QSFP Ports [32:25] VBus 35 - 43 */
	sfpidx = 0;
	if (virt[3]) {
		for(i = 0; i <= 7; i++, sfpidx++) {
			add_virt_bus(virt[3], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* mixed switch w/ QSFPs and PSs and temp sensors */
	sfpidx = 0;
	if (virt[4]) {

		/* QSFP Ports [36:33] VBus 44 - 47 */
		for(i = 0; i <= 3; i++, sfpidx++) {
			add_virt_bus(virt[4], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

		/* PS0 VBus 48 */
		add_virt_bus(virt[4], (0xE0 >> 1), 4,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

		/* PS1 VBus 49 */
		add_virt_bus(virt[4], (0xE0 >> 1), 5,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

		/* COB_TEMP VBus 50 */
		add_virt_bus(virt[4], (0xE0 >> 1), 6,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

		/* CPU_TEMP VBus 51 */
		add_virt_bus(virt[4], (0xE0 >> 1), 7,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
	}
}

static void
add_virt_bus_rigel2(struct i2c_adapter *adap)
{
	int ret;
	int i, total_sw = 7;
	struct i2c_adapter *virt[total_sw];

	adap->algo->fixup_i2c = &castor_fixup_locked_iic1_bus;

	/*
	 * Invoke unlock routine at startup in case any device (such as
	 * FRU header/seeprom) was left hung up by reboot.  On Rigel this
	 * will affect the virtual i2c driver's ability to configure the
	 * i2c switch during init and lead to init failure.  And because
	 * of that failure the unlock never has a chance to be called.
	 */
	if ((ret = castor_fixup_locked_iic1_bus(adap))) {
		printk(KERN_ERR "Rigel i2c unlock, ret=%d\n", ret);
	}

	/* mux val = 0, Power Monitor, DS1780, LM 75's, VBus 4 (why not 3?) */
	add_virt_bus(adap, (0xEE >> 1), 0,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* Add all the First level SFP Mux's, VBus 5 - 9, 10, 11 */
	for (i = 0; i < total_sw; i++) {
		virt[i] = add_virt_bus(adap, (0xEE >> 1), i+1,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
	}

	/* QSFP Ports [8:1] [16:9] [24:17] */
	if (virt[0]) {
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[0], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[0], (0xE2 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[0], (0xE4 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* VMON[2:0] COB[2:0]_p1v0 */
	if (virt[1]) {
		for(i = 0; i <= 2; i++) {
			add_virt_bus(virt[1], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		for(i = 3; i <= 5; i++) {
			add_virt_bus(virt[1], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* COB[2:0]_TEMP PSU0 SFP[2:1]_p3v3 CPU_TEMP */
	if (virt[2]) {
		for(i = 0; i <= 2; i++) {
			add_virt_bus(virt[2], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		add_virt_bus(virt[2], (0xE0 >> 1), 3,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		for(i = 4; i <= 5; i++) {
			add_virt_bus(virt[2], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		/* port 6 of this mux not used */
		add_virt_bus(virt[2], (0xE0 >> 1), 7,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

	}

	/* DC */
	if (virt[5]) {
		/* VMON[2:0] COB[2:0] Temp Sensors */
		for(i = 0; i <= 6; i++) {
			add_virt_bus(virt[5], (0xE6 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

		/* QSFP Ports [8:1] [16:9] [24:17] */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[5], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[5], (0xE2 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[5], (0xE4 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

		/* COB[5:3]_TEMP PSU1 SFP[2:1]_p3v3 */
		for(i = 0; i <= 2; i++) {
			add_virt_bus(virt[5], (0xE8 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		// Port 3 unused
		add_virt_bus(virt[5], (0xE8 >> 1), 4,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		for(i = 5; i <= 6; i++) {
			add_virt_bus(virt[5], (0xE8 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		// Port 7 unused
	}

	/*
	 ********************************************************
	 *
	 *  Note that these PICs are out of the typical order
	 *  because they will be going in and out.
	 *
	 ********************************************************
	 */

	/* MB PIC */
	if (virt[4]) {
		/* QSFP PIC - TPS2481 LTC3880 Temp Sensors 5 unused */
		/* MTP PIC - TPS2481 Temp Sensors Mini-pods[7:0] */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[4], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

		/* QSFP PIC - QSFP Ports [8:1] */
		/* MTP PIC - unused */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[4], (0xE2 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		/* QSFP PIC - QSFP Ports [12:9] 4 unused */
		/* MTP PIC - unused */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[4], (0xE4 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

	/* DC PIC */

	if (virt[6]) {
		/* QSFP PIC - TPS2481 LTC3880 Temp Sensors 5 unused */
		/* MTP PIC - TPS2481 Temp Sensors Mini-pods[7:0] */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[6], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

		/* QSFP PIC - QSFP Ports [8:1] */
		/* MTP PIC - unused */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[6], (0xE2 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
		/* QSFP PIC - QSFP Ports [12:9] 4 unused */
		/* MTP PIC - unused */
		for(i = 0; i <= 7; i++) {
			add_virt_bus(virt[6], (0xE4 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}
	}

}

static void
add_virt_bus_wedge(struct i2c_adapter *adap)
{
		int ret, i;
		int mux_ports = 8;
		struct i2c_adapter *virt[8];

		adap->algo->fixup_i2c = &wedge_recover_locked_iic1_bus;

		/*
		 * Invoke unlock routine at startup in case any device (such as
		 * FRU header/seeprom) was left hung up by reboot.  On callisto this
		 * will affect the virtual i2c driver's ability to configure the
		 * i2c switch during init and lead to init failure.  And because
		 * of that failure the unlock never has a chance to be called.
		 */
		if ((ret = wedge_recover_locked_iic1_bus())) {
			printk(KERN_ERR "FATAL wedge i2c unlock failure, ret=%d\n",
					ret);
		}


		for (i = 0; i < mux_ports; i++) {
			virt[i] = add_virt_bus(adap, (0xE4 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/*
		 * Except for port 0 of the main mux, all the
		 * other ports have a mux.
		 */

		/* Thermal sensors and volt monitoring */
		// for (i = 0; i < mux_ports; i++) {
		//	add_virt_bus(virt[1], (0xE0 >> 1), i,
		//	&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		// }

		/* SFPs 16-23 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[1], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}


		/* SFPs 24-31 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[2], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/* QSFPs */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[3], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/* SFPs 8-15 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[4], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/* SFPs 0-7 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[5], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/* SFPs 32-39 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[6], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}

		/* SFPs 40-47 */
		for (i = 0; i < mux_ports; i++) {
			add_virt_bus(virt[7], (0xE0 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
		}
}
//-----------------------------------------------------------------------------
static void
add_virt_bus_blackbird(struct i2c_adapter *adap)
{
	int ret;
	struct i2c_adapter *virt[2];
	int i, sfpidx, total_sw = 2;

	adap->algo->fixup_i2c = &blackbird_fixup_locked_iic1_bus;

	/*
	 * Invoke unlock routine at startup in case any device (such as
	 * FRU header/seeprom) was left hung up by reboot.  On blackbird this
	 * will affect the virtual i2c driver's ability to configure the
	 * i2c switch during init and lead to init failure.  And because
	 * of that failure the unlock never has a chance to be called.
	 */
	if ((ret = blackbird_fixup_locked_iic1_bus(adap))) {
		printk(KERN_ERR "Blackbird i2c unlock, ret=%d\n", ret);
	}

	/* The first level mux (PCA9548) has 8 outputs.
	 * Each Mux Output is selected by making that
	 * bit (0-7) to 1 in PCA9548 Control Register. Register
	 * a virtual bus for each one. As each output is a
	 * separate I2C Bus electrically.
	 * They are all at the same level on the same board, so
	 * use the same ID value. The ID gets displayed by Linux
	 * error messages and is used to determine which FRU is
	 * causing the problem.
	 */

	/* Add all the First level SFP Mux's */
	for (i = 0; i < total_sw; i++) {
		virt[i] = add_virt_bus(adap, (0xE8 >> 1), i,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
	}

	/* mux val = 2, Temperature sensor, TMP411A */
	add_virt_bus(adap, (0xE8 >> 1), 2,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* mux val = 3, Voltage monitor, MAX16047*/
	add_virt_bus(adap, (0xE8 >> 1), 3,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* mux val = 4, WWN EEPROM */
	add_virt_bus(adap, (0xE8 >> 1), 4,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);


	/* mux val = 5, Voltage Margin, MAX4562 */
	add_virt_bus(adap, (0xE8 >> 1), 5,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* mux val = 6, Cobra temperature, TMP411A */
	add_virt_bus(adap, (0xE8 >> 1), 6,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

	/* mux val = 7, TEST-IO card */
	add_virt_bus(adap, (0xE8 >> 1), 7,
			&pcf9548_select_mux, &pcf9548_deselect_mux, 0);



	/*
	 * Following are 2nd level virtual busses for the SFPs
	 * Muxes are added in the order of the SFP's, check HW spec.
	 * Muxes are all on the same board and on the save level,
	 * so use the same ID for all of them
	 */

	sfpidx = 0;
    if (virt[0]) {
    /* GE SFP Ports 0 through 7 */
        for(i = 0; i <= 7; i++, sfpidx++)
            add_virt_bus(virt[0], (0xE0 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }

	sfpidx = 0;
    if (virt[1]) {
    /* GE SFP Ports 8 through 13 */
        for(i = 0; i <= 5; i++, sfpidx++)
            add_virt_bus(virt[1], (0xE2 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
    }


	sfpidx = 0;
    if (virt[1]) {
    /* GE QSFP Ports 0 through 1 */
        for(i = 6; i <= 7; i++, sfpidx++) {
            add_virt_bus(virt[1], (0xE2 >> 1), i,
                &pcf9548_select_mux, &pcf9548_deselect_mux, 1);
		}

    }
}

//-----------------------------------------------------------------------------
static void
add_virt_bus_lc(struct i2c_adapter *adap)
{
	int sfpidx;

	adap->algo->fixup_i2c = &mpc85xx_fixup_locked_bus;

	/*
	 * Muxes are all on the same board and on the save level,
	 * so use the same ID for all of them. The ID gets displayed
	 * by Linux error messages and is used to determine which
	 * FRU is causing the problem.
	 */

	/* SFP Ports 0 through 15 */
	sfpidx = 0;
	add_pcf9548_virt_bus(adap, 0x70, 2, sfpidx);

	if (silkworm.board == SWBD1004 ||
		silkworm.board == SWBD1008) {

		/* SFP Ports 16 through 47 */
		add_pcf9548_virt_bus(adap, 0x72, 4, sfpidx);

    } else if (silkworm.board == SWBD1010) {

        add_pcf9548_virt_bus(adap, 0x72, 3, sfpidx);

    }

}
/*
 * This fn is called once for each adapter. We use this to add our
 * platform-specific callback functions.
 */
static int attach_adapter(struct i2c_adapter *adap)
{
	int i, sfpidx, pca=0;
	struct i2c_adapter *virt;
	int index = i2c_adapter_id(adap);

	/* Register virtual busses based on our specific platform type */
	switch(silkworm.board) {
		case SWBD21:
		case SWBD42:
			/* Only IIC1 bus on Meteor has muxes */
			if(index != 1)
				break;

			/* Meteor has to arbitrate for the shared I2C bus */
			adap->algo->acquire_exclusive = &meteor_acquire_bus;
			adap->algo->release_exclusive = &meteor_release_bus;


			/*
			 * Use the mux_map table index for an ID. This gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */

			for(i=0; i<elemof(meteor_mux_map); i++)
				add_virt_bus(adap, 0, meteor_mux_map[i],
					&meteor_select_mux, &meteor_deselect_mux, i);

			printk("iic%d: Registered I2C multiplexor callback for "
				"SWBD%d on IIC bus %d.\n", index, silkworm.board, index);
			break;

		case SWBD62:
		case SWBD77:
		case SWBD141:
			/* Only IIC1 and PCA9665 bus has muxes */
			if (index < 1)
				break;

			/* We have to now rely on adapter name */
			if (!strcmp(adap->name, "PCA9665 adapter")) {
				pca = 1;
			}

			if (index == 1) {

				/*
				 * Directors have to arbitrate for the shared I2C bus. Neptune
				 * re-uses the Meteor arbitration methods
				 */
				adap->algo->acquire_exclusive = &meteor_acquire_bus;
				adap->algo->release_exclusive = &meteor_release_bus;

				/*
				 * platform-specific bus unlocking
				 */
				// adap->algo->fixup_i2c = &zentron_fixup_locked_bus;
				adap->algo->fixup_i2c = &mpc85xx_fixup_locked_bus;
				/*
				 * Use the mux_map table index for an ID. This gets displayed
				 * by Linux error messages and is used to determine which
				 * FRU is causing the problem.
				 *
				 */
				if (pca) {
					/* This is PCA9665 buffer mode adapter */
					for(i=0; i<elemof(neptune_mux_map); i++) {
						add_virt_bus(adap, 0, neptune_mux_map[i],
							 &neptune_select_mux, &neptune_deselect_mux, i);
					}
				} else if (silkworm.board == SWBD77) {

					/* This is CPU IIC1  adapter */
					for(i=0; i<elemof(pluto_mux_map); i++) {
						add_virt_bus(adap, 0, pluto_mux_map[i],
							 &neptune8548_select_mux, &neptune_deselect_mux, i);
					}
				} else {
					/* This is CPU IIC1  adapter */
					for(i=0; i<elemof(neptune_mux_map); i++) {
						add_virt_bus(adap, 0, neptune_mux_map[i],
							 &neptune8548_select_mux, &neptune_deselect_mux, i);
					}
				}
			} else {
				if (pca) {
					/* This is PCA9665 buffer mode adapter */
					adap->algo->acquire_exclusive = &meteor_acquire_bus;
					adap->algo->release_exclusive = &meteor_release_bus;
					for(i=0; i<elemof(neptune_mux_map); i++) {
						add_virt_bus(adap, 0, neptune_mux_map[i],
							 &neptune_select_mux, &neptune_deselect_mux, i);
					}
				} else
					break;
			}
			printk("iic%d: Registered I2C multiplexor callback for "
				"SWBD%d on IIC bus %d.\n", index, silkworm.board, index);

			break;

		case SWBD46:
			/* Only IIC1 bus on Sprint has muxes */
			if(index != 1)
				break;

			/*
			 * Use the mux_map table index for an ID. This gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */

			for(i=0; i<elemof(sprint_mux_map); i++)
				add_virt_bus(adap, 0, sprint_mux_map[i],
				 &meteor_select_mux, &meteor_deselect_mux, i);

			/*
			 * Get the virtual adapter for the bus connecting
			 * to MarathonJr board - that's where the mux's are attached.
			 */

			virt = i2c_get_adapter(MARATHONJR_VIRTUAL_BUS);

			/*
			 * ID 7 (0-6 assigned above) used for all four muxes
			 * since they are all at the same level and on the
			 * same board
			 */

			/* Create virtual busses 9-16. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x70, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 17-24. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x71, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 25-32. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x72, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 33-40. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x73, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			printk("iic%d: Registered I2C multiplexor callback for "
				"SWBD%d on IIC bus %d.\n", index, silkworm.board, index);
			break;

		case SWBD79:
			/* Only IIC1 bus on SPEAR has muxes */
			if(index != 1)
				break;
			/*
		         * platform-specific bus unlocking
		         */
			/*
			 * Use the mux_map table index for an ID. This gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			for(i=0; i<elemof(spear_mux_map); i++)
                               add_virt_bus(adap, 0, spear_mux_map[i],
		                             &meteor_select_mux, &meteor_deselect_mux, i);
			/*
                         * Get the virtual adapter for the bus connecting
                         * to MarathonJr board - that's where the mux's are attached.
                         */
			 virt = i2c_get_adapter(SPEAR_VIRTUAL_BUS);

			/*
			 ID 9 (0-8 assigned above) used for all four muxes
			* since they are all at the same level and on the
		        * same board
			*/
			/*TBD:SPEAR... Not sure about effect of presence Switch PCA9548 address EC on ID parameter
			in all add_virt_bus function below  */
			/* Create virtual busses 11-18. */
                        for(i=0; i<=7; i++) {
                               add_virt_bus(virt, 0x70, i,
		                                &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
	                   }
			 /* Create virtual busses 19-26. */
			for(i=0; i<=7; i++) {
			      add_virt_bus(virt, 0x71, i,
			                       &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
			   }
			/* Create virtual busses 27-34. */
	                for(i=0; i<=7; i++) {
	                      add_virt_bus(virt, 0x72, i,
		                              &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
		            }
			    /* Create virtual busses 35-42. */
		        for(i=0; i<=7; i++) {
		               add_virt_bus(virt, 0x73, i,
				       &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
	                    }
			    /* Create virtual busses 43-50. */
		        for(i=0; i<=7; i++) {
		               add_virt_bus(virt, 0x74, i,
				       &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
	                    }
			    /* Create virtual busses 51-58. */
		        for(i=0; i<=7; i++) {
		               add_virt_bus(virt, 0x76, i,
				       &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
	                    }

			printk("iic%d: Registered I2C multiplexor callback for "
			                       "SWBD%d on IIC bus %d.\n", index, silkworm.board, index);
		                break;

		case SWBD83:
		  {
			struct i2c_adapter *virt_parent;

			/* Only IIC1 bus on SPIKE has muxes */
			if(index != 1)
				break;

			/*
		         * platform-specific bus unlocking
		         */
			/*
			 * Use the mux_map table index for an ID. This gets displayed
		         * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			*/
			for(i=0; i<elemof(sprint_mux_map); i++)
                               add_virt_bus(adap, 0, sprint_mux_map[i],
		                             &meteor_select_mux, &meteor_deselect_mux, i);
			/*
                         * Get the virtual adapter for the bus connecting
                         * to BP  I2C devices  that's where the mux's are attached.
                         */
			 virt_parent = i2c_get_adapter(SPEAR_VIRTUAL_BUS);
			/*Parent Switch .Behind this switch SFP Switches are available*/
			 /*Bus Number-9*/
			 virt = add_virt_bus(virt_parent, 0x76, 0,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

			/*
			 ID 9 (0-8 assigned above) used for all four muxes
			* since they are all at the same level and on the
		        * same board
			*/
			/*I2C virt bus number 10-17*/
                        for(i=0; i<=7; i++) {
                               add_virt_bus(virt, 0x70, i,
		                                &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
	                   }

			 /* Create virtual busses 17-24. */
			for(i=0; i<=7; i++) {
			      add_virt_bus(virt, 0x71, i,
			                       &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
			   }
			/* Create virtual busses 25-31. */
	                for(i=0; i<=7; i++) {
	                      add_virt_bus(virt, 0x72, i,
		                              &pcf9548_select_mux, &pcf9548_deselect_mux, 9);
		            }

			/*Virt bus adapter for Sensor devices */
			/* Create virtual busses 32-38. */
		        for(i=0; i<=7; i++) {
		               add_virt_bus(virt_parent, 0x74, i,
				       &pcf9548_select_mux, &pcf9548_deselect_mux, 10);
	                    }

			printk("iic%d: Registered I2C multiplexor callback for "
			                       "SWBD%d on IIC bus %d.\n", index, silkworm.board, index);
		  }
		  break;
		case SWBD67:
			/* Only IIC1 bus on Mace has muxes */
			if(index != 1)
				break;

			/*
			 * platform-specific bus unlocking
			 */
			adap->algo->fixup_i2c = &cpjrplus_fixup_locked_bus;

			/*
			 * Use the mux_map table index for an ID. This gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			for(i=0; i<elemof(sprint_mux_map); i++)
				add_virt_bus(adap, 0, sprint_mux_map[i],
				 &meteor_select_mux, &neptune_deselect_mux, i);

			/*
			 * Get the virtual adapter for the bus connecting
			 * to MarathonJr board - that's where the mux's are attached.
			 */

			virt = i2c_get_adapter(MARATHONJR_VIRTUAL_BUS);

			/*
			 * ID 7 (0-6 assigned above) used for all four muxes
			 * since they are all at the same level and on the
			 * same board
			 */

			/* Create virtual busses 9-16. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x70, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 17-24. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x71, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 25-32. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x72, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 33-40. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x73, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			/* Create virtual busses 41-48. */
			for(i=0; i<=7; i++) {
				add_virt_bus(virt, 0x74, i,
				 &pcf9548_select_mux, &pcf9548_deselect_mux, 7);
			}

			printk("iic%d: Registered I2C multiplexor callback for "
				"SWBD%d on IIC bus %d.\n", index, silkworm.board, index);
			break;

  		 case SWBD23:
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;

			/* The first level mux has 4 outputs.
 			 * Register a virtual bus for each one. They are all
			 * at the same level and on the same board, so use
			 * the same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			virt = add_virt_bus(adap, 0x76, 0,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 1,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 2,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 3,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);

			/* There are 2nd level virtual buses for the SFPs */
			if(!virt)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/*
			* Muxes are all on the same board and on the save level,
			* so use the same ID for all of them
			*/
			/* SFP Ports 0 through 7 */
			sfpidx = 0;
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			/* SFP Ports 8 through 15 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			/* SFP Ports 16 through 23 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD32:	/* Pulsar-32 */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;

			/* The first level mux has 4 outputs.
	 		 * Register a virtual bus for each one. They are all at the
			 * same level and on the same board, so use the same ID value.
			 * The ID gets displayed by Linux error messages and is used
			 * to determine which FRU is causing the problem.
			 */
			virt = add_virt_bus(adap, 0x76, 0,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 1,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 2,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 3,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);

			/* There are 2nd level virtual buses for the SFPs */
			if(!virt)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/*
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */
			/* SFP Ports 0 through 7 */
			sfpidx = 0;
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			/* SFP Ports 8 through 15 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			/* SFP Ports 16 through 23 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			/* SFP Ports 24 through 31 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(virt, 0x73, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD58: /* Pulsar2 (32 port 440gp w/ Condors */
		case SWBD66: /* Spirit (40 port 440epx w/ C2s) */
		case SWBD71: /* Tomahawk (24 port 440EPX w/ geye2s) */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/*
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them. The ID gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			/* SFP Ports 0 through 7 */
			sfpidx = 0;
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(adap, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* SFP Ports 8 through 15 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(adap, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* SFP Ports 16 through 23 */
			for(i=0; i<=7; i++, sfpidx++)
				add_virt_bus(adap, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			if ((silkworm.board == SWBD58) ||
				(silkworm.board == SWBD66)) {
				/* SFP Ports 24 through 31 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(adap, 0x73, i,
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			}
			if (silkworm.board == SWBD66) {
				/* SFP Ports 32 through 39 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(adap, 0x74, i,
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD88: /* Europa BP. SFP access only */
		case SWBD98: /* Europa_ttv BP. SFP access only */
			{
			/* Only IIC1 bus on Pulsar has muxes */
			uint8_t regval1, regval2;

			if(index != 1)
				break;

			/* Bring these Muxes out of reset first via CPLD*/

			cpld_read_register(EUROPA_IIC_RESET_ENABLE_REG, &regval1);
			cpld_read_register(EUROPA_IIC_RESET_CNTRL_REG, &regval2);
			regval1 |= EUROPA_IIC_MUX_RESET;
			regval2 |= EUROPA_IIC_MUX_RESET;
			cpld_write_register(EUROPA_IIC_RESET_ENABLE_REG, regval1);
			cpld_write_register(EUROPA_IIC_RESET_CNTRL_REG, regval2);
			regval1 &= ~EUROPA_IIC_MUX_RESET;
			cpld_write_register(EUROPA_IIC_RESET_ENABLE_REG, regval1);

			/*
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them. The ID gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */

			add_pcf9548_virt_bus(adap, 0x70, 3, 0);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			}
			break;

		case SWBD70:	/* LittleBear 8G */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			for(i=0; i<8; i++)
				add_virt_bus(adap, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			for(i=0; i<2; i++, sfpidx++)
				add_virt_bus(adap, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD73: /* Blazer3 (Dirk) */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;
			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/* SFP Ports 0 through 5 */
			sfpidx = 0;
			for(i=0; i<=5; i++, sfpidx++)
				add_virt_bus(adap, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD87:	/* Kestrel2 */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 0)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/* SFP Ports 0 through 5 */
			for(i=0; i<6; i++)
				add_virt_bus(adap, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* Equalizer 0 through 7 */
			for(i=0; i<8; i++)
				add_virt_bus(adap, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* Equalizer 8 through 15 */
			for(i=0; i<8; i++)
				add_virt_bus(adap, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* Equalizer 16 through 19 */
			for(i=0; i<4; i++)
				add_virt_bus(adap, 0x73, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD72: /* Bulova (440EPX) */
		case SWBD75: /* Superhawk2 */
		case SWBD82: /* Utah */
		case SWBD86: /* Falcon2 */
			/* Only IIC1 bus on Pulsar has muxes */
			if(index != 1)
				break;

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_ENABLE_REG,
				PULSAR_IIC_MUX_RESET_ENABLE | PULSAR_IIC_RESET_ENABLE);
			cpld_write_register(PULSAR_INDIVIDUAL_RESET_CNTRL_REG, 0);

			/*
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them. The ID gets displayed
			 * by Linux error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			/* SFP Ports 0 through 7 */
			for(i=0; i<8; i++)
				add_virt_bus(adap, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD29:	/* Icecube */
			/* Only IIC1 bus on Icecube has muxes */
			if(index != 1)
				break;

			/* The single level Icecube mux has 4 outputs
			 * Register a virtual bus for each one. Don't really
			 * need an ID, but the API requires one.
			 */
			virt = add_virt_bus(adap, 0x76, 0,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 1,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 2,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 3,
				&pcf9544_select_mux, &pcf9544_deselect_mux, 0);

			/* Bring the Mux out of reset via CPLD
			 * No reset enable due to CPLD space constraints
			 */
			cpld_write_register(STEALTH_INDIVIDUAL_RESET_CNTRL_REG,
					    STEALTH_DIS_I2C_PWR_RESET);
			udelay(30000); /* Delay 30 msec */
			cpld_write_register(STEALTH_INDIVIDUAL_RESET_CNTRL_REG, 0);
			udelay(30000); /* Delay 30 msec */

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD34:	/* Stealth */
			break;

		case SWBD37: /* Blazer2 */
			if(index!=1) break;

			/*
			 * Don't really need an ID, but the API requires one.
			 */
			add_virt_bus(adap,0,0,&blazer2_select_mux,
					&blazer2_deselect_mux, 0);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD43:
		case SWBD45:
		case SWBD51:
		case SWBD61:
			break;

		case SWBD64:    // Thor
			if(index != 1) break;       // Only bus one
            // Platform specific IIC bus unlock
            adap->algo->fixup_i2c = &thor_fixup_locked_bus;
                        // Fall through
		case SWBD44:	/* Viking */
			{
			struct i2c_adapter *dc_virt;	/* daughter card */

			/* Only IIC1 bus on Viking has muxes */
			if(index != 1)      // Double checked for Thor
				break;

			/* The first level mux (PCA9548) has 8 outputs.
	 		* Register a virtual bus for each one. They are all
			 * at the same level on the same board, so use the
			 * same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which
			 * FRU is causing the problem.
			 */
			virt = add_virt_bus(adap, 0x76, 0,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			dc_virt = add_virt_bus(adap, 0x76, 1,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 2,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 3,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 4,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 5,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 6,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, 0x76, 7,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/* Bring these Muxes out of reset first via CPLD*/
			// TODO - CPLD bug - need to select i2c_reset instead of
			// i2c_sfp_reset - the bootrom should do this for us...
			cpld_write_register(VIKING_INDIVIDUAL_RESET_ENABLE_REG,
				VIKING_IIC_MUX_RESET_ENABLE | VIKING_IIC_RESET_ENABLE);
			cpld_write_register(VIKING_INDIVIDUAL_RESET_CNTRL_REG, 0);

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (virt) {
				/* SFP Ports 32 through 39 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* SFP Ports 40 through 47 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* SFP Ports 48 through 55 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* SFP Ports 56 through 63 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, 0x73, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				if (silkworm.board == SWBD64) {
					/* SFP Ports 64 through 71 */
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt, 0x74, i,
						&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

					/* SFP Ports 72 through 79 */
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt, 0x75, i,
						&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
					}
			}

			/*
			 * Following are daughter card virtual busses
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (dc_virt) {
				/* SFP Ports 01 through 07 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(dc_virt, 0x70, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

				/* SFP Ports 08 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(dc_virt, 0x71, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

				/* SFP Ports 16 through 23 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(dc_virt, 0x72, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

				/* SFP Ports 24 through 31 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(dc_virt, 0x73, i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 2);
			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			}
			break;

		case SWBD76:	/* Elara */
			{
			int ret;
			unsigned char rst_en_reg, rst_reg;

			/* IIC1 bus on Elara has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;

			adap->algo->fixup_i2c = &elara_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On elara this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = elara_recover_locked_iic1_bus(adap))) { /* Hari 09/28/2009 */
				printk(KERN_ERR "FATAL Elara i2c unlock failure, ret=%d\n",
						ret);
			}

			/* The first level mux (PCA9548) has 8 outputs. But
             * Elara Electrical Spec mentions use of only 5
             * MUX Outputs. Each Mux Output is selected by making that
             * bit (0-7) to 1 in PCA9548 Control Register. Register
             * a virtual bus for each one. As each output is a
			 * separate I2C Bus electrically.
             * They are all at the same level on the same board, so
			 * use the same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which FRU is
			 * causing the problem.
			 */
			add_virt_bus(adap, (0xE4 >> 1), 0,
                    /* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
                    /* mux out = 1, not assigned in spec */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
                    /* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 3,
                    /* mux val = 3 , PCIe Switch */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
                    /* mux val = 4, PS1 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 5,
                    /* mux val = 5, PS0 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 6,
                    /* mux val = 6, Not assigned */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 7,
                    /* mux val = 7, Not assigned */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
            /*
             * DS1780,LM75 will be on Bus 2
             * PCIe will be on Bus 5
             * PS1 on Bus 6, PS0 on Bus 7
             * WWN0 and WWN1 on Bus 1
             * FRU Header SEEPROM on Bus 1
             * GE SFP 0-7 bus 10-17
             * GE SFP 8-15 bus 18-25
             * GE SFP 16-23 bus 26-33
             * FC SFP 0-7 bus 34-41
             * FC SFP 8-15 bus 42-49
			 *
             * ALL SFP's access I2C address is 0xA0 >> 1, same as
             * PS0,PS1... Is it because of device has fixed address
             *
             *
             * mux_val = 0, this is passed to
             * pcf9548_(se/de)select, which is then written
             * to corresponding bit in the PCA9548 Control
             * register selecting the mux line as the I2C
             * Bus Segment downstream
             *
             */

			/*
             * Clear reset for all I2C Mux. Elara has I2C_SW_RST
             * connected to all PCA9548 Mux. Clear the reset bit
             * which will stop reset happening if any
             * Why is this required ??? TODO Check write
             */
			cpld_read_register(ELARA_I2C_CONTROL_PATH_RST_EN_REG,
						&rst_en_reg);
			cpld_write_register(ELARA_I2C_CONTROL_PATH_RST_EN_REG,
                       rst_en_reg | ELARA_I2C_MUX_RST);

			/* enable write to bit 4 in 0x4D */
			cpld_read_register(ELARA_I2C_CONTROL_PATH_RST_REG,
						&rst_reg);
			/* deassert I2C Mux rst if any happening */
			cpld_write_register(ELARA_I2C_CONTROL_PATH_RST_REG,
                      rst_reg & ~ELARA_I2C_MUX_RST);

			udelay(1000);  /* Will this be enough.. ?? */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (virt) {
				/* GE SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE2 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 16 through 23 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE6 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* FC SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE8 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

                /* FC SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xEA >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			}
			break;

		case SWBD95:    /* @kevin : Elara2E */
		case SWBD96:    /* @kevin : Elara2F */
			{
			int ret;

			/* IIC1 bus has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;
			unsigned char rst_en_reg, rst_reg;

			adap->algo->fixup_i2c = &elara2_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On elara2 this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = elara2_recover_locked_iic1_bus())) {
				printk(KERN_ERR "Elara2 i2c unlock, ret=%d\n",
						ret);
			}

			/* The first level mux (PCA9548) has 8 outputs. But
			 * elara2 Electrical Spec mentions use of only 3
			 * MUX Outputs. Each Mux Output is selected by making that
			 * bit (0-7) to 1 in PCA9548 Control Register. Register
			 * a virtual bus for each one. As each output is a
			 * separate I2C Bus electrically.
			 * They are all at the same level on the same board, so
			 * use the same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which FRU is
			 * causing the problem.
			 */
			add_virt_bus(adap, (0xE4 >> 1), 0,
				/* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
				/* mux val = 1, PS0 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
				/* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 3,
				/* mux val = 3, PS1 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
				/* mux val = 4, PS2 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 5,
				/* mux val = 5, PS3 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 6,
				/* mux val = 6, PS4 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 7,
				/* mux val = 7, PS5 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/*
			 * Clear reset for all I2C Mux. Elara2 has I2C_SW_RST
			 * connected to all PCA9548 Mux. Clear the reset bit
			 * which will stop reset happening if any
			 * Why is this required ??? TODO Check write
			 */
			cpld_read_register(ELARA2_I2C_CONTROL_PATH_RST_EN_REG,
						&rst_en_reg);
			cpld_write_register(ELARA2_I2C_CONTROL_PATH_RST_EN_REG,
						rst_en_reg | ELARA2_I2C_MUX_RST);

			/* enable write to bit 4 in 0x4D */
			cpld_read_register(ELARA2_I2C_CONTROL_PATH_RST_REG,
						&rst_reg);
			/* deassert I2C Mux rst if any happening */
			cpld_write_register(ELARA2_I2C_CONTROL_PATH_RST_REG,
						rst_reg & ~ELARA2_I2C_MUX_RST);

			udelay(1000);  /* Will this be enough.. ?? */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */
			if (virt) {
				/* GE SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE2 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 16 through 23 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE6 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* FC SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE8 >> 1), (i),
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;
			}

		case SWBD97:    /* @kevin : CallistoE */
		case SWBD107:   /* @kevin : CallistoF */
			{
			int ret;
			struct i2c_adapter *virt1;

			/* IIC1 bus has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;
			unsigned char rst_en_reg, rst_reg;

			adap->algo->fixup_i2c = &callisto_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On callisto this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = callisto_recover_locked_iic1_bus())) {
				printk(KERN_ERR "Callisto i2c unlock, ret=%d\n",
						ret);
			}

			/* The first level mux (PCA9548) has 8 outputs. But
			 * callisto Electrical Spec mentions use of only 5
			 * MUX Outputs. Each Mux Output is selected by making that
			 * bit (0-7) to 1 in PCA9548 Control Register. Register
			 * a virtual bus for each one. As each output is a
			 * separate I2C Bus electrically.
			 * They are all at the same level on the same board, so
			 * use the same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which FRU is
			 * causing the problem.
			 */
			add_virt_bus(adap, (0xE4 >> 1), 0,
				/* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
				/* mux out = 1,PCIe SWitch */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
				/* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt1 = add_virt_bus(adap, (0xE4 >> 1), 3,
				/* mux val = 3 , PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
				/* mux val = 4, PS0 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 5,
				/* mux val = 5, PS1 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 6,
				/* mux val = 6, PS2 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 7,
				/* mux val = 7, PS3 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			/*
			 *
			 * ALL SFP's access I2C address is 0xA0 >> 1, same as
			 * PS0,PS1... Is it because of device has fixed address
			 *
			 *
			 * mux_val = 0, this is passed to
			 * pcf9548_(se/de)select, which is then written
			 * to corresponding bit in the PCA9548 Control
			 * register selecting the mux line as the I2C
			 * Bus Segment downstream
			 *
			 */

			/*
			 * Clear reset for all I2C Mux. Callisto has I2C_SW_RST
			 * connected to all PCA9548 Mux. Clear the reset bit
			 * which will stop reset happening if any
			 * Why is this required ??? TODO Check write
			 */
			cpld_read_register(CALLISTO_I2C_CONTROL_PATH_RST_EN_REG,
						&rst_en_reg);
			cpld_write_register(CALLISTO_I2C_CONTROL_PATH_RST_EN_REG,
						rst_en_reg | CALLISTO_I2C_MUX_RST);

			/* enable write to bit 4 in 0x4D */
			cpld_read_register(CALLISTO_I2C_CONTROL_PATH_RST_REG,
						&rst_reg);
			/* deassert I2C Mux rst if any happening */
			cpld_write_register(CALLISTO_I2C_CONTROL_PATH_RST_REG,
						rst_reg & ~CALLISTO_I2C_MUX_RST);

			udelay(1000);  /* Will this be enough.. ?? */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (virt) {
				/* GE SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE2 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 16 through 23 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE6 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 24-31 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE8 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 32-39 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xEA >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}

			sfpidx =0;
			if (virt1) {
				/* GE SFP Ports 40 through 47 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt1, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 48 through 55 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt1, (0xE2 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 56 through 59 */
				for(i=0; i<=3; i++, sfpidx++)
					add_virt_bus(virt1, (0xE6 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* FC SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt1, (0xE8 >> 1), (i),
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* FC SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt1, (0xEA >> 1), (i),
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
				index, silkworm.board);
			break;
			}
		case SWBD112:
		case SWBD123:
			/* For this platform, handled in platform driver */
			break;
		case SWBD116:   /* @kevin : Carme */
			{
			int ret;

			/* IIC1 bus has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;

			adap->algo->fixup_i2c = &carme_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On elara2 this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = carme_recover_locked_iic1_bus())) {
				printk(KERN_ERR "FATAL Carme i2c unlock failure, ret=%d\n",
						ret);
			}

			/* The first level mux (PCA9548) has 8 outputs. But
			 * elara2 Electrical Spec mentions use of only 3
			 * MUX Outputs. Each Mux Output is selected by making that
			 * bit (0-7) to 1 in PCA9548 Control Register. Register
			 * a virtual bus for each one. As each output is a
			 * separate I2C Bus electrically.
			 * They are all at the same level on the same board, so
			 * use the same ID value. The ID gets displayed by Linux
			 * error messages and is used to determine which FRU is
			 * causing the problem.
			 */
			add_virt_bus(adap, (0xE4 >> 1), 0,
				/* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
				/* mux out = 1,PCIe SWitch */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
				/* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 3,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 5,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 6,
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 7,
				/* mux val = 7, RealTime PWR Monitor */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			udelay(1000);  /* Will this be enough.. ?? */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */
			if (virt) {
				/* GE SFP Ports 0 through 5 */
				for(i=0; i<=5; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}

			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;
			}
		case SWBD122:
			{
			int ret;
			struct i2c_adapter *dc_virt;

			/* IIC1 bus has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;
			unsigned int rst_en_reg, rst_reg;

			adap->algo->fixup_i2c = &callisto_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On callisto this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = callisto_recover_locked_iic1_bus())) {
				printk(KERN_ERR "Starlifter i2c unlock, ret=%d\n",
						ret);
			}

	    		/* The first level mux (PCA9548) has 8 outputs.
   		        * Each Mux Output is selected by making that
             		* bit (0-7) to 1 in PCA9548 Control Register. Register
    		        * a virtual bus for each one. As each output is a
             		* separate I2C Bus electrically.
			* They are all at the same level on the same board, so
		     	* use the same ID value. The ID gets displayed by Linux
			* error messages and is used to determine which FRU is
			* causing the problem.
			*/
			add_virt_bus(adap, (0xE4 >> 1), 0,
                    /* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
                    /* mux out = 1,LM75 on DC */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
                    /* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 3,
                    /* mux val = 3 , FRU-HDR DC  */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
                    /* mux val = 4, PS0 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, (0xE4 >> 1), 5,
                    /* mux val = 5, PS1 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 6,
                    /* mux val = 6, PS3 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 7,
                    /* mux val = 7, PS4 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
            		/*
			 *
             		 * ALL SFP's access I2C address is 0xA0 >> 1, same as
             		 * PS0,PS1... Is it because of device has fixed address
             		 *
             		 *
             		 * mux_val = 0, this is passed to
              		 * pcf9548_(se/de)select, which is then written
             		 * to corresponding bit in the PCA9548 Control
             		 * register selecting the mux line as the I2C
             		 * Bus Segment downstream
             		 *
             		 */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (virt) {
				/* GE SFP Ports 0 through 7 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

				/* GE SFP Ports 8 through 15 */
				for(i=0; i<=7; i++, sfpidx++)
					add_virt_bus(virt, (0xE2 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;
			}
		case SWBD127:
			{
			int ret;
			struct i2c_adapter *virt1, *virt2;

			/* IIC1 bus has muxes/I2C Switches */
			if(index != 1) /* index = i2c_adapter id for IIC1 */
				break;
			unsigned int rst_en_reg, rst_reg;

			adap->algo->fixup_i2c = &starlifter_recover_locked_iic1_bus;

			/*
			 * Invoke unlock routine at startup in case any device (such as
			 * FRU header/seeprom) was left hung up by reboot.  On callisto this
			 * will affect the virtual i2c driver's ability to configure the
			 * i2c switch during init and lead to init failure.  And because
			 * of that failure the unlock never has a chance to be called.
			 */
			if ((ret = starlifter_recover_locked_iic1_bus())) {
				printk(KERN_ERR "Starlifter i2c unlock, ret=%d\n",
						ret);
			}

	    		/* The first level mux (PCA9548) has 8 outputs.
   		        * Each Mux Output is selected by making that
             		* bit (0-7) to 1 in PCA9548 Control Register. Register
    		        * a virtual bus for each one. As each output is a
             		* separate I2C Bus electrically.
			* They are all at the same level on the same board, so
		     	* use the same ID value. The ID gets displayed by Linux
			* error messages and is used to determine which FRU is
			* causing the problem.
			*/
			add_virt_bus(adap, (0xE4 >> 1), 0,
                    /* mux val = 0, DS1780, LM 75's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 1,
                    /* mux out = 1,LM75 on DC */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt = add_virt_bus(adap, (0xE4 >> 1), 2,
                    /* mux val = 2, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 3,
                    /* mux val = 3 , FRU-HDR DC  */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			add_virt_bus(adap, (0xE4 >> 1), 4,
                    /* mux val = 4, PS0 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			add_virt_bus(adap, (0xE4 >> 1), 5,
                    /* mux val = 5, PS1 */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

			virt1 = add_virt_bus(adap, (0xE4 >> 1), 6,
                    /* mux val = 6, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);
			virt2 = add_virt_bus(adap, (0xE4 >> 1), 7,
                    /* mux val = 7, PCA9548 Switches to SFP's */
				&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

            		/*
			 *
             		 * ALL SFP's access I2C address is 0xA0 >> 1, same as
             		 * PS0,PS1... Is it because of device has fixed address
             		 *
             		 *
             		 * mux_val = 0, this is passed to
              		 * pcf9548_(se/de)select, which is then written
             		 * to corresponding bit in the PCA9548 Control
             		 * register selecting the mux line as the I2C
             		 * Bus Segment downstream
             		 *
             		 */

			sfpidx = 0;

			/*
			 * Following are 2nd level virtual busses for the SFPs
			 *
			 * Muxes are all on the same board and on the save level,
			 * so use the same ID for all of them
			 */

			if (virt) {
				/* GE SFP Ports 0 through 5 */
				for(i=0; i<=5; i++, sfpidx++)
					add_virt_bus(virt, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}
			sfpidx = 0;
			if (virt1) {
				/* GE SFP Ports 6 through 11 */
				for(i=0; i<=5; i++, sfpidx++)
					add_virt_bus(virt1, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}
			sfpidx = 0;
			if (virt2) {
			/* GE SFP Ports 12 through 17 */
				for(i=0; i<=5; i++, sfpidx++)
					add_virt_bus(virt2, (0xE0 >> 1), i,
					&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
			}
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;
			}
		case SWBD131:
			/* Only IIC1 bus has muxes */
			if(index != 1)
				break;

			add_virt_bus_castor(adap);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;
		case SWBD137:
		case SWBD151:
            /* Only IIC1 bus has muxes */
            if(index != 1)
                break;

            add_virt_bus_castorT(adap);
            printk("iic%d: Registered I2C mux callback for SWBD%d\n",
                    index, silkworm.board);
            break;

		/* Rigel-I */
		case SWBD153:
			/* Only IIC1 bus has muxes */
			if(index != 1)
			    break;
			add_virt_bus_rigel1(adap);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
				index, silkworm.board);
			break;
		case SWBD154:
			/* Only IIC1 bus has muxes */
			if(index != 1)
			    break;
			add_virt_bus_rigel2(adap);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
				index, silkworm.board);
			break;

 		case SWBD138:
            /* Only IIC1 bus has muxes */
            if(index != 1)
                break;

            add_virt_bus_blackbird(adap);
            printk("iic%d: Registered I2C mux callback for SWBD%d\n",
                    index, silkworm.board);
			break;

		case SWBD161:
		case SWBD162:
			{
				if (index != 1)
					break;

				add_virt_bus_wedge(adap);
				printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
				break;
			}
		case SWBD148:
			{
				int ret;
				struct i2c_adapter *virt_sw_ts, *virt_sw_e2, *virt_sw_ge_sfp1;
				struct i2c_adapter *virt_sw_ge_sfp2, *virt_sw_fc_sfp1,
					*virt_sw_fc_sfp2, *virt_sw_fc_sfp3;
				unsigned char rst_en_reg, rst_reg;

				/* IIC1 bus has muxes/I2C Switches */
				if(index != 1) /* index = i2c_adapter id for IIC1 */
					break;
				adap->algo->fixup_i2c = &skybolt_recover_locked_iic1_bus;

				/*
				 * Invoke unlock routine at startup in case any device (such as
				 * FRU header/seeprom) was left hung up by reboot.  On callisto this
				 * will affect the virtual i2c driver's ability to configure the
				 * i2c switch during init and lead to init failure.  And because
				 * of that failure the unlock never has a chance to be called.
				 */
				if ((ret = skybolt_recover_locked_iic1_bus())) {
					printk(KERN_ERR "FATAL skybolt i2c unlock failure, ret=%d\n",
							ret);
				}
				/* The first level mux (PCA9548) has 8 outputs.
				 * Each Mux Output is selected by making that
				 * bit (0-7) to 1 in PCA9548 Control Register. Register
				 * a virtual bus for each one. As each output is a
				 * separate I2C Bus electrically.
				 * They are all at the same level on the same board, so
				 * use the same ID value. The ID gets displayed by Linux
				 * error messages and is used to determine which FRU is
				 * causing the problem.
				 */
				add_virt_bus(adap, (0xE4 >> 1), 0,
						/* mux val = 0, MB Temp Sensors LM 75's IIC(4) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				virt_sw_ts = add_virt_bus(adap, (0xE4 >> 1), 1,
						/* mux val = 1, PCA9548 Switch to Thermal Sensor's IIC(5) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				virt_sw_e2 = add_virt_bus(adap, (0xE4 >> 1), 2,
						/* mux val = 2, PCA9548 second level i2c switch IIC(6) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				add_virt_bus(adap, (0xE4 >> 1), 3,
						/* mux val = 3 , FRU-HDR DC IIC(7) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				add_virt_bus(adap, (0xE4 >> 1), 4,
						/* mux val = 4, PS0 IIC(8) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				add_virt_bus(adap, (0xE4 >> 1), 5,
						/* mux val = 5, PS1 IIC(9) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				virt_sw_ge_sfp1 = add_virt_bus(adap, (0xE4 >> 1), 6,
						/* mux val = 6, PCA9548 Switch to GE SFP's IIC(10) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				virt_sw_ge_sfp2 = add_virt_bus(adap, (0xE4 >> 1), 7,
						/* mux val = 7, PCA9548 Switch to GE SFP's IIC(11) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 0);

				/* Registering the Temperature Sensors  */
				if(virt_sw_ts)
				{
					add_virt_bus(virt_sw_ts, (0xE0 >> 1), 0,
							/* mux val = 0, to  P 3041 Thermal Sensor IIC(12) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

					add_virt_bus(virt_sw_ts, (0xE0 >> 1), 1,
							/* mux val = 1, to  Blitzer_0 Thermal Sensor IIC(13) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

					add_virt_bus(virt_sw_ts, (0xE0 >> 1), 2,
							/* mux val = 2, to  Blitzer_1 Thermal Sensor IIC(14) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

					add_virt_bus(virt_sw_ts, (0xE0 >> 1), 3,
							/* mux val = 3, to CN6880_0 Thermal Sensor IIC(15) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 1);

					add_virt_bus(virt_sw_ts, (0xE0 >> 1), 4,
							/* mux val = 4, to CN6880_1 Thermal Sensor IIC(16) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 1);
				}

				if(virt_sw_e2)
				{
					/* second level switch 0xE2 */
					add_virt_bus(virt_sw_e2, (0xE2 >> 1), 0,
							/* mux val = 0, to ltc3880 IIC(17) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

					virt_sw_fc_sfp1 = add_virt_bus(virt_sw_e2, (0xE2 >> 1), 1,
							/*  mux val = 1, PCA9548 Switch to FC SFP's IIC(18) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

					virt_sw_fc_sfp2 = add_virt_bus(virt_sw_e2, (0xE2 >> 1), 2,
							/* mux val = 2, PCA9548 Switch to FC SFP's IIC(19) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

					virt_sw_fc_sfp3 = add_virt_bus(virt_sw_e2, (0xE2 >> 1), 3,
							/* mux val = 3, PCA9548 Switch to FC SFP's IIC(20) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);


					add_virt_bus(virt_sw_e2, (0xE2 >> 1), 4,
							/* mux val = 4, LTC3880 IIC(21) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

					add_virt_bus(virt_sw_e2, (0xE2 >> 1), 7,
							/* mux val = 7, MAX16047 IIC(22) */
							&pcf9548_select_mux, &pcf9548_deselect_mux, 2);
				}

				if(virt_sw_fc_sfp1)
				{
					/* FC SFP Ports 0 through 7 IIC(23) to IIC(29)*/
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt_sw_fc_sfp1, (0xE0 >> 1), i,
								/* mux val = 0-7, PCA9548 Switch to GE SFP's */
								&pcf9548_select_mux, &pcf9548_deselect_mux, 3);
				}

				if(virt_sw_fc_sfp2)
				{
					/* FC SFP Ports 8 through 15 IIC(30) to IIC(36)*/
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt_sw_fc_sfp2, (0xE0 >> 1), i,
								/* mux val = 0-7, PCA9548 Switch to GE SFP's */
								&pcf9548_select_mux, &pcf9548_deselect_mux, 3);
				}

				if(virt_sw_fc_sfp3)
				{
					/* FC SFP Ports 16 through 23 IIC(37) to IIC(43) */
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt_sw_fc_sfp3, (0xE0 >> 1), i,
								/* mux val = 7, PCA9548 Switch to GE SFP's */
								&pcf9548_select_mux, &pcf9548_deselect_mux, 3);
				}

				add_virt_bus(virt_sw_e2, (0xE2 >> 1),5,
						/* mux val = 5, to QSFP[1] IIC(44) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

				add_virt_bus(virt_sw_e2, (0xE2 >> 1), 6,
						/* mux val = 6, to QSFP[2] IIC(45) */
						&pcf9548_select_mux, &pcf9548_deselect_mux, 2);

				if (virt_sw_ge_sfp1) {
					/* GE SFP Ports 0 through 7 IIC(46) to IIC(52) */
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt_sw_ge_sfp1, (0xE0 >> 1), i,
								/* mux val = 0-7, PCA9548 Switch to GE SFP's */
								&pcf9548_select_mux, &pcf9548_deselect_mux, 4);
				}

				if (virt_sw_ge_sfp2) {
					/* GE SFP Ports 8 through 15 IIC(53) to IIC(59) */
					for(i=0; i<=7; i++, sfpidx++)
						add_virt_bus(virt_sw_ge_sfp2, (0xE0 >> 1), i,
								/* mux val = 0-7, PCA9548 Switch to GE SFP's */
								&pcf9548_select_mux, &pcf9548_deselect_mux, 4);
				}

			break;
			}

		case SWBD1004:	/* Mercury 48 port LC */
		case SWBD1005:	/* Mercury 16 port LC */
		case SWBD1006:	/* Mercury 72 port 1G LC*/
		case SWBD1007:	/* Mercury 48 port 1G LC*/
		case SWBD1008:	/* Mercury 36 port 10G LC*/
		case SWBD1010:  /* Mercury 27 port 40G LC*/

			/* Only IIC1 bus has muxes */
			if(index != 1)
				break;

			add_virt_bus_lc(adap);
			printk("iic%d: Registered I2C mux callback for SWBD%d\n",
					index, silkworm.board);
			break;

		case SWBD156:  /* Redbird */
		case SWBD157:  /* Graphte */
		case SWBD158:  /* Carbon */
		case SWBD167:  /* Shenlong */
			/* virtual busses created in platform driver */
			break;

		default:
			printk("%s: Platform type [%d],  "
				"Did not register I2C multiplexor callback.\n",
				adap->name, silkworm.board);
			break;

	}
	return 0;
}

/* Simple wrapper to export for later loaded modules */
int silkworm_attach_i2c_adapter(struct i2c_adapter *adap)
{
	return(attach_adapter(adap));
}

/*
 * Register a virtual bus and also store some information about this bus
 * that will be useful to FabOS clients.
 */
static struct i2c_adapter *add_virt_bus(struct i2c_adapter *real,
                                        unsigned long mux_addr,
                                        unsigned long mux_val,
                                        void *select_cb,
                                        void *deselect_cb,
										int id)

{
	struct i2c_bus_mapping *pmap;
	struct i2c_adapter *virt;

	if((virt = register_virtual_i2c_adap(real, mux_addr, mux_val,
					select_cb, deselect_cb, id)) == NULL)
		return (virt);

	/* Allocate and setup the pmap structure */
	if((pmap = kmalloc(sizeof(struct i2c_bus_mapping),
					GFP_KERNEL)) == NULL) {
		printk(KERN_ERR "%s: Failed to allocate bus mapping\n",
			__FUNCTION__);
		return (NULL);
	}

	pmap->real = real;
	pmap->mux_addr = mux_addr;
	pmap->mux_val = mux_val;
	pmap->virt = virt;

	list_add_tail(&pmap->list, &i2c_map_list);

	return virt;
}

/* Deregister a virtual bus. */
static int remove_virt_bus(struct i2c_adapter *virt)
{
        int ret = 0;
        struct list_head *l;
        struct i2c_bus_mapping *pmap;

        if (virt == NULL) {
		printk("remove_virt_bus: NULL virtual adapater\n");
                return -EINVAL;
	}

        if ((ret = unregister_virtual_i2c_adap(virt)) != 0) {
                return ret;
	}

        list_for_each(l, &i2c_map_list) {
                pmap = list_entry(l, struct i2c_bus_mapping, list);
                if(pmap && pmap->virt == virt) {
                        list_del(l);
                        kfree(pmap);
                        break;
                }
        }

        return ret;
}


/*
 * Called by FabOS (and other) clients to find out which i2c bus to use to
 * get to a specific bus segment.
 */
struct i2c_adapter *i2c_lookup_adapter(struct i2c_adapter *base, unsigned long mux_addr, unsigned long mux_val)
{
	struct list_head *l;
	struct i2c_bus_mapping *pmap;

	list_for_each(l, &i2c_map_list) {
		pmap = list_entry(l, struct i2c_bus_mapping, list);
		if(pmap->real == base && pmap->mux_addr == mux_addr &&
				pmap->mux_val == mux_val) {
			return pmap->virt;
		}
	}
	return NULL;	/* none found */
}

/* Shows the silkworm i2c bus mappings for the platform */
static int i2c_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data)
{
	int len;
	char *buf = page;
	struct list_head *l;
	struct i2c_bus_mapping *pmap;

        buf += sprintf(buf, "Silkworm i2c bus_mapping\n");

	list_for_each(l, &i2c_map_list) {
		pmap = list_entry(l, struct i2c_bus_mapping, list);

		if(pmap) {
        		buf += sprintf(buf, "base=iic%d  mux_addr=0%02hx, "
				"mux_val=%02hx(%d)  =>  iic%d\n",
				i2c_adapter_id(pmap->real), (int)pmap->mux_addr,
				(int)pmap->mux_val, (int)pmap->mux_val,
				i2c_adapter_id(pmap->virt));
		}
	}

        len = buf - page;
        if (len <= off+count) *eof = 1;
        *start = page + off;
        len -= off;
        if (len>count) len = count;
        if (len<0) len = 0;
        return len;

}

static int __init i2c_mux_init(void)
{
	int status  = 0, i;
	struct i2c_adapter *adap;



	/*
	 * Iterate through the _real_ adapters
	 * If it's not a virtual adapter, then it must
	 * be a _real_ adapter. Just filter out I2C_HW_VIRT so
	 * that no changes will be needed for new hardware
	 */
	for(i=0; i<I2C_ADAP_MAX; i++) {

		if((adap = i2c_get_adapter(i)) != NULL &&
		    adap->id != I2C_HW_VIRT) {
			attach_adapter(adap);
		}
	}

	create_proc_read_entry("driver/iic/silkworm_i2c_map", 0, NULL,
		&i2c_read_proc, NULL);

	if (!perf_stats) {
		perf_stats_init();
	}

	return (status);
}

static void __exit i2c_mux_exit(void)
{
}

module_init(i2c_mux_init);
module_exit(i2c_mux_exit);

EXPORT_SYMBOL(i2c_lookup_adapter);
EXPORT_SYMBOL(add_virt_bus);
EXPORT_SYMBOL(remove_virt_bus);
EXPORT_SYMBOL(pcf9544_select_mux);
EXPORT_SYMBOL(pcf9544_deselect_mux);
EXPORT_SYMBOL(pcf9548_select_mux);
EXPORT_SYMBOL(pcf9548_deselect_mux);
EXPORT_SYMBOL(silkworm_attach_i2c_adapter);

MODULE_DESCRIPTION("I2C driver to register multiplexor callbacks");

