/*
 * QorIQ 10G MDIO Controller
 *
 * Copyright 2012-2014 Freescale Semiconductor, Inc.
 *
 * Authors: Andy Fleming <afleming@freescale.com>
 *          Timur Tabi <timur@freescale.com>
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/phy.h>
#include <linux/mdio.h>
#include <linux/of_platform.h>
#include <linux/of_mdio.h>
#include <linux/mm.h>
#include <linux/io.h>

#include "xgmac_mdio.h"

#ifdef CONFIG_SILKWORM_XMDIO_BURST_WRITE
int XMDIO_DEBUG_LVL = 0;
#define MDIO_PRINT0 if (XMDIO_DEBUG_LVL>3) printk
#endif

/* Number of microseconds to wait for a register to respond */
#define TIMEOUT	1000

/* Taken from memac_mdio.c */
#define MDIO_STAT_ENC (1 << 6)
#define MDIO_STAT_HOLD_15_CLK (7 << 2)

/*
 * Wait untill the MDIO bus is free
 */
static int xgmac_wait_until_free(struct device *dev,
				 struct tgec_mdio_controller __iomem *regs)
{
	uint32_t status;

	/* Wait till the bus is free */
	status = spin_event_timeout(
		!((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY), TIMEOUT, 0);
	if (!status) {
		dev_err(dev, "timeout waiting for bus to be free\n");
		return -ETIMEDOUT;
	}

	return 0;
}

/*
 * Wait untill the MDIO command is completed
 */
static int xgmac_wait_until_completion(struct device *dev,
                                 struct tgec_mdio_controller __iomem *regs)
{
        uint32_t status;

        /* Wait till the command is completed */
        status = spin_event_timeout(
                ((in_be32(&regs->mdio_stat)) & MDIO_STAT_CMP), TIMEOUT, 0);
        if (!status) {
                dev_err(dev, "timeout waiting for command to complete\n");
                return -ETIMEDOUT;
        }

        return 0;
}

static int xgmac_clear_completion_stat(struct device *dev,
                                 struct tgec_mdio_controller __iomem *regs)
{
	uint32_t mdio_stat;
        uint32_t status;
	
	mdio_stat = in_be32(&regs->mdio_stat);
	mdio_stat |= MDIO_STAT_CMP;
	out_be32(&regs->mdio_stat, mdio_stat);

        status = spin_event_timeout(
                !((in_be32(&regs->mdio_stat)) & MDIO_STAT_CMP), TIMEOUT, 0);
        if (!status) {
                dev_err(dev, "timeout waiting for clearing completion\n");
                return -ETIMEDOUT;
        }

	return 0;
}

/*
 * Write value to the PHY for this device to the register at regnum,waiting
 * until the write is done before it returns.  All PHY configuration has to be
 * done through the TSEC1 MIIM regs.
 */
static int xgmac_mdio_write(struct mii_bus *bus, int phy_id, int devad, int regnum, u16 value)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	uint16_t dev_addr;
	u32 mdio_ctl, mdio_stat;
	int ret;

	mdio_stat = in_be32(&regs->mdio_stat);
	if (regnum & MII_ADDR_C45) {
		/* Clause 45 (ie 10G) */
		dev_addr = (regnum >> 16) & 0x1f;
		mdio_stat |= MDIO_STAT_ENC | MDIO_STAT_HOLD_15_CLK;
	} else {
		/* Clause 22 (ie 1G) */
		dev_addr = regnum & 0x1f;
		mdio_stat &= ~MDIO_STAT_ENC;
	}

	out_be32(&regs->mdio_stat, mdio_stat);

	ret = xgmac_wait_until_free(&bus->dev, regs);
	if (ret)
		return ret;

	/* Set the port and dev addr */
	mdio_ctl = MDIO_CTL_PORT_ADDR(phy_id) | MDIO_CTL_DEV_ADDR(dev_addr);
	out_be32(&regs->mdio_ctl, mdio_ctl);

	/* Set the register address */
	if (regnum & MII_ADDR_C45) {
		out_be32(&regs->mdio_addr, regnum & 0xffff);

		ret = xgmac_wait_until_free(&bus->dev, regs);
		if (ret)
			return ret;
		ret = xgmac_wait_until_completion(&bus->dev, regs);
		if (ret)
			return ret;
		__asm__ ("isync");

		ret = xgmac_clear_completion_stat(&bus->dev, regs);
		if (ret)
			return ret;
		__asm__ ("isync");
	}

	/* Write the value to the register */
	out_be32(&regs->mdio_data, MDIO_DATA(value));

	ret = xgmac_wait_until_completion(&bus->dev, regs);
	if (ret)
		return ret;
	__asm__ ("isync");

	ret = xgmac_clear_completion_stat(&bus->dev, regs);
	if (ret)
                return ret;
	__asm__ ("isync");

	if (in_be32(&regs->mdio_data) != MDIO_DATA(value))
		printk("### MDIO write error. ###\n");

	return 0;
}

/*
 * Reads from register regnum in the PHY for device dev, returning the value.
 * Clears miimcom first.  All PHY configuration has to be done through the
 * TSEC1 MIIM regs.
 */
static int xgmac_mdio_read(struct mii_bus *bus, int phy_id, int devad, int regnum)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	uint16_t dev_addr;
	uint32_t mdio_stat;
	uint32_t mdio_ctl;
	uint16_t value;
	int ret;

	mdio_stat = in_be32(&regs->mdio_stat);
	if (regnum & MII_ADDR_C45) {
		dev_addr = (regnum >> 16) & 0x1f;
		mdio_stat |= MDIO_STAT_ENC | MDIO_STAT_HOLD_15_CLK;
	} else {
		dev_addr = regnum & 0x1f;
		mdio_stat &= ~MDIO_STAT_ENC;
	}

	out_be32(&regs->mdio_stat, mdio_stat);

	ret = xgmac_wait_until_free(&bus->dev, regs);
	if (ret)
		return ret;

	/* Set the Port and Device Addrs */
	mdio_ctl = MDIO_CTL_PORT_ADDR(phy_id) | MDIO_CTL_DEV_ADDR(dev_addr);
	out_be32(&regs->mdio_ctl, mdio_ctl);

	/* Set the register address */
	if (regnum & MII_ADDR_C45) {
		out_be32(&regs->mdio_addr, regnum & 0xffff);

		ret = xgmac_wait_until_free(&bus->dev, regs);
		if (ret)
			return ret;
		ret = xgmac_wait_until_completion(&bus->dev, regs);
		if (ret)
			return ret;
		__asm__ ("isync");

		ret = xgmac_clear_completion_stat(&bus->dev, regs);
		if (ret)
			return ret;
		__asm__ ("isync"); 
	}

	/* Initiate the read */
	out_be32(&regs->mdio_ctl, mdio_ctl | MDIO_CTL_READ);

	ret = xgmac_wait_until_free(&bus->dev, regs);
	if (ret)
		return ret;

	ret = xgmac_wait_until_completion(&bus->dev, regs);
	if (ret)
		return ret;
	__asm__ ("isync");

	ret = xgmac_clear_completion_stat(&bus->dev, regs);
	if (ret)
		return ret;
	__asm__ ("isync");

	/* Return all Fs if nothing was there */
	if (in_be32(&regs->mdio_stat) & MDIO_STAT_RD_ER) {
		dev_err(&bus->dev, "MDIO read error\n");
		return 0xffff;
	}

	value = in_be32(&regs->mdio_data) & 0xffff;
	dev_dbg(&bus->dev, "read %04x\n", value);

	return value;
}

#ifdef CONFIG_SILKWORM_XMDIO_BURST_WRITE
int xgmac_mdio_addr(struct mii_bus *bus, int port_addr,
			int dev_addr, int regnum)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	u32 mdio_ctl, mdio_stat;

	/* Setup the MII Mgmt clock speed */
	mdio_stat = MDIO_STAT_CLKDIV(0x100);

	/* workaround for erratum XAUI3, only needs for Rev1 silicon */
	if ((mfspr(SPRN_SVR) & 0xff) == 0x10)
		mdio_stat |= 0xc;
	out_be32(&regs->mdio_stat, mdio_stat);

	/* Wait till the bus is free */
	while ((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY)
		cpu_relax();

	/* Set the port and dev addr */
	mdio_ctl = MDIO_CTL_PORT_ADDR(port_addr) | MDIO_CTL_DEV_ADDR(dev_addr);
	out_be32(&regs->mdio_ctl, mdio_ctl);

	/* Set the register address */
	out_be32(&regs->mdio_addr, regnum & 0xffff);

	/* Wait till the bus is free */
	while ((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY)
		cpu_relax();

	MDIO_PRINT0("MDIO ADDR  0x%x-%x %x\n", port_addr, dev_addr, regnum);

	return (0);
}

int xgmac_mdio_put(struct mii_bus *bus, int port_addr,
			int dev_addr, u16 value)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	u32 mdio_ctl, mdio_stat;

	/* Setup the MII Mgmt clock speed */
	mdio_stat = MDIO_STAT_CLKDIV(0x100);

	out_be32(&regs->mdio_stat, mdio_stat);

	/* Wait till the bus is free */
	while ((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY)
		cpu_relax();

	/* Set the port and dev addr */
	mdio_ctl = MDIO_CTL_PORT_ADDR(port_addr) | MDIO_CTL_DEV_ADDR(dev_addr);
	out_be32(&regs->mdio_ctl, mdio_ctl);

	/* Wait till the bus is free */
	while ((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY)
		cpu_relax();

	/* Write the value to the register */
	out_be32(&regs->mdio_data, MDIO_DATA(value));

	/* Wait till the MDIO write is complete */
	while ((in_be32(&regs->mdio_data)) & MDIO_DATA_BSY)
		cpu_relax();

	MDIO_PRINT0("MDIO PUT 0x%x-%x %x xmioData:0x%x\n", port_addr, dev_addr, value, MDIO_DATA(value));

	return (0);
}

int xgmac_mdio_get(struct mii_bus *bus, int port_addr,
			int dev_addr, int regnum)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	u32 mdio_ctl, mdio_stat, tmp;

	/* Setup the MII Mgmt clock speed */
	mdio_stat = MDIO_STAT_CLKDIV(0x100);

	out_be32(&regs->mdio_stat, mdio_stat);

	/* Wait till the bus is free */
	while ((in_be32(&regs->mdio_stat)) & MDIO_STAT_BSY)
		cpu_relax();

	/* Set the Port and Device Addrs */
	mdio_ctl =  MDIO_CTL_READ | MDIO_CTL_PORT_ADDR(port_addr) | MDIO_CTL_DEV_ADDR(dev_addr);
	out_be32(&regs->mdio_ctl, mdio_ctl);

	/* Wait till the MDIO write is complete */
	while ((in_be32(&regs->mdio_data)) & MDIO_DATA_BSY)
		cpu_relax();

	tmp = in_be32(&regs->mdio_data);

	MDIO_PRINT0("MDIO GET 0x%x-%x %x xmioData -> 0x%x\n", port_addr, dev_addr, regnum, tmp);

        /* if read error bit is set, return error */
        if (in_be32(&regs->mdio_stat) & MDIO_STAT_RD_ER)
                return (-EREMOTEIO);

	return (tmp & 0xffff);
}
#endif

/* Reset the MIIM registers, and wait for the bus to free */
static int xgmac_mdio_reset(struct mii_bus *bus)
{
	struct tgec_mdio_controller __iomem *regs = bus->priv;
	int ret;

	mutex_lock(&bus->mdio_lock);

	/* Setup the MII Mgmt clock speed */
	clrbits32(&regs->mdio_stat, MDIO_STAT_ENC);

	ret = xgmac_wait_until_free(&bus->dev, regs);

	mutex_unlock(&bus->mdio_lock);

	return ret;
}

static int xgmac_mdio_probe(struct of_device *pdev, const struct of_device_id *match)
{
	struct device_node *np = pdev->node;
	struct mii_bus *bus;
	struct resource res;
	int ret;

#ifndef CONFIG_SILKWORM_XMDIO_BURST_WRITE
	/*
	 * do not check xmdio device available on standby since there is
	 * no actual HW access in this probe routine.
	*/
	if (!of_device_is_available(np))
		return -ENODEV;
#endif

	ret = of_address_to_resource(np, 0, &res);
	if (ret) {
		dev_err(&pdev->dev, "could not obtain address\n");
		return ret;
	}

	bus = mdiobus_alloc_size(PHY_MAX_ADDR * sizeof(int));
	if (!bus)
		return -ENOMEM;

	bus->name = "Freescale XGMAC MDIO Bus";
	bus->read = xgmac_mdio_read;
	bus->write = xgmac_mdio_write;
	bus->reset = xgmac_mdio_reset;
#ifdef CONFIG_SILKWORM_XMDIO_BURST_WRITE
	bus->addr = xgmac_mdio_addr;
	bus->put = xgmac_mdio_put;
	bus->get = xgmac_mdio_get;
#endif
	bus->irq = bus->priv;
	bus->parent = &pdev->dev;
	snprintf(bus->id, MII_BUS_ID_SIZE, "%llx", (unsigned long long)res.start);

	/* Set the PHY base address */
	bus->priv = of_iomap(np, 0);
	if (!bus->priv) {
		ret = -ENOMEM;
		goto err_ioremap;
	}

	ret = of_mdiobus_register(bus, np);
	if (ret) {
		dev_err(&pdev->dev, "cannot register MDIO bus\n");
		goto err_registration;
	}

	dev_set_drvdata(&pdev->dev, bus);

	return 0;

err_registration:
	iounmap(bus->priv);

err_ioremap:
	mdiobus_free(bus);

	return ret;
}

static int xgmac_mdio_remove(struct of_device *pdev)
{
	struct mii_bus *bus = dev_get_drvdata(&pdev->dev);

	mdiobus_unregister(bus);
	iounmap(bus->priv);
	mdiobus_free(bus);

	return 0;
}

static struct of_device_id xgmac_mdio_match[] = {
	{
		.compatible = "fsl,fman-xmdio",
	},
	{
		.compatible = "fsl,fman-memac-mdio",
	},
	{},
};
MODULE_DEVICE_TABLE(of, xgmac_mdio_match);

static struct of_platform_driver xgmac_mdio_driver = {
	.name = "fsl-fman_xmdio",
	.match_table = xgmac_mdio_match,
	.probe = xgmac_mdio_probe,
	.remove = xgmac_mdio_remove,
};

static int __init xgmac_mdio_init(void)
{
	return of_register_platform_driver(&xgmac_mdio_driver);
}

static void __exit xgmac_mdio_exit(void)
{
	of_unregister_platform_driver(&xgmac_mdio_driver);
}

subsys_initcall_sync(xgmac_mdio_init);
module_exit(xgmac_mdio_exit);

MODULE_DESCRIPTION("Freescale QorIQ 10G MDIO Controller");
MODULE_LICENSE("GPL v2");
