/*
 * Brocade PCI Express PCI Hot Plug Driver
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/signal.h>
#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <linux/aer.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <brocade/silkworm_bhpc.h>
#include "../pci.h"
#include "pciehp.h"
#include <linux/pcieport_if.h>

#ifdef DEBUG
#define DBGPRINT(x...) printk(x)
#else
#define DBGPRINT(x...)
#endif

#define BHPC_PCI_DRV_NAME "brocade_hpc"


extern void silkworm_machine_check_control(int enable, int threshold);
extern int pex8614_errata_port_reset(struct pci_dev *dev, int speed);

/* Forward declarations */

static int show_dsp_link(struct pci_dev *pdev);
static int map_sfm_usp(uint slot, unsigned long __iomem **base);
static struct controller *get_slot_controller(uint8_t slotno);


/* Protos for bhpc_svc_ops */
static int bhpc_initialize(void);
static int bhpc_plat_remove(void);
static int bhpc_blade_insert(uint8_t slot, struct pci_dev **dev);
static int bhpc_blade_remove(uint8_t slot);
static int bhpc_restore_config(uint8_t slot);
static int bhpc_upstream_enable_get(uint8_t slot, uint8_t *state);
static int bhpc_upstream_enable_set(uint8_t slot, uint8_t state);
static int bhpc_command(uint8_t slot, int command);

/* Protos for pcie_port_service_driver */
static int bhpc_hpc_probe(struct pcie_device *dev);

/* Protos for hpc_ops */

static int bhpc_get_adapter_status (struct hotplug_slot *hp_slot, u8 *status);
static int bhpc_get_power_status(struct hotplug_slot *hp_slot, u8 *status);

/* Protos for AER */

static void bhpc_aer_remove(struct pci_dev *dev);
static int bhpc_aer_probe(struct pci_dev *dev, const struct pci_device_id *ent);
static pci_ers_result_t bhpc_error_detected(struct pci_dev *dev,
											enum pci_channel_state state);

/* Private data */

static int bhpc_initialized = 0;

static bhpc_slot_map_table slot_map_table;
static int interrupts_enabled = 0;

static bhpc_platform chassis, *chassis_ptr = &chassis;
static bhpc_ops *plat_ops;

static struct hotplug_slot_ops saved_hpc_ops;

static bhpc_svc_ops svc_ops = {

	.initialize = bhpc_initialize,
	.remove = bhpc_plat_remove,
	.blade_insert = bhpc_blade_insert,
	.blade_remove = bhpc_blade_remove,
	.upstream_enable_set = bhpc_upstream_enable_set,
	.upstream_enable_get = bhpc_upstream_enable_get,
	.restore_config = bhpc_restore_config,
	.command = bhpc_command,

};

static struct pcie_port_service_driver bhpc_service_ops = {

	.probe = bhpc_hpc_probe,
	// .remove = bhpc_hpc_remove,
};

static struct pcie_port_service_driver hpc_service_ops;

static struct hotplug_slot_ops bhpc_hpc_ops = {

	.get_power_status = bhpc_get_power_status,
	.get_adapter_status = bhpc_get_adapter_status,

};

static struct pci_error_handlers bhpc_aer_handler = {
	.error_detected = bhpc_error_detected,
};

static struct pci_device_id bhpc_aer_ids[] = {
	{ PCI_DEVICE(PCI_VENDOR_ID_BROCADE, PCI_ANY_ID) },
	{ 0, }
};

MODULE_DEVICE_TABLE(pci, bhpc_aer_ids);

static struct pci_driver bhpc_aer_driver = {
	.name		= BHPC_PCI_DRV_NAME,
	//	.owner		= THIS_MODULE,
	.id_table	= bhpc_aer_ids,  
	.probe		= bhpc_aer_probe,
	.remove		= bhpc_aer_remove,
	.err_handler = &bhpc_aer_handler

};
#define PSTATUS_ERROR_MASK (PCI_STATUS_SIG_TARGET_ABORT | PCI_STATUS_REC_TARGET_ABORT | \
						   PCI_STATUS_REC_MASTER_ABORT | PCI_STATUS_SIG_SYSTEM_ERROR | \
						   PCI_STATUS_DETECTED_PARITY)


extern int enable_fault;

/*
 *  save_device_state()
 *
 *    Save configuration registers of a hierarchy of devices.
 *
 * Inputs:
 *    dev - pci_dev object of the parent device of
 *          the hierarchy
 *
 */
static void
save_device_state(struct pci_dev *dev)
{
	int err;
	struct pci_bus *pbus;
	struct pci_dev *pdev, *tdev;

	pbus = dev->bus;
	pdev = pbus->self;


	pr_debug("%s: Device %s refcount %d Parent Dev %s refcount %d\n",
		   __FUNCTION__, pci_name(dev), pci_dev_refcount(dev),
		   pci_name(pdev), pci_dev_refcount(pdev));

	err = pci_save_state(dev);

	/*
	 * This call decrements the refcount of the initial dev and
	 * also modifies the dev pointer. Save the original pointer
	 * so we can restore the refcount later.
	 */
	tdev = dev;
	for_each_pci_dev(dev) {

		pr_debug("%s: Dev %s refcount %d\n",
			   __FUNCTION__, pci_name(dev), pci_dev_refcount(dev));
		err = pci_save_state(dev);
		if (err) {
			printk(KERN_ERR "%s: pci dev %s save state failed %d\n",
				   __FUNCTION__, pci_name(dev), err);
		}
	}

	/*
	 * Call pci_dev_get() to compensate for the refcount
	 * decrement that occured above.
	 */

	pci_dev_get(tdev);

	pr_debug("%s %d: Device %s refcount %d\n",
		   __FUNCTION__, __LINE__, pci_name(tdev), pci_dev_refcount(tdev));

}
#if 0
/*
 * restore_device_state()
 *
 *    Restore the config registers of a device hierarchy (for future use)
 *
 *  Inputs:
 *    dev -  pci_dev object of the parent device of
 *          the hierarchy
 *
 */
static void
restore_device_state(struct pci_dev *dev)
{
	int err;
	struct pci_dev *tdev;

	pr_debug("%s: Device %s\n", __FUNCTION__, pci_name(dev));
	err = pci_restore_state(dev);
	err |= pci_enable_device(dev);
	pci_set_master(dev);

	tdev = dev;
	for_each_pci_dev(dev) {

		pr_debug("%s DEV %s\n", __FUNCTION__, pci_name(dev));
		err = pci_restore_state(dev);
		err |= pci_enable_device(dev);
		pci_set_master(dev);
		if (err) {
			printk(KERN_ERR "%s: pci dev %s restore state failed %d\n",
				   __FUNCTION__, pci_name(dev), err);
		}
	}

	pci_dev_get(tdev);
}
#endif

#if 0
/*
 * bhpc_error_notify()
 *
 *   Notify whoever registered of a PCIe error interrupt
 *
 * Inputs:
 *    ddata - Contains all the data collected about the interrupt
 *
 */
void
bhpc_error_notify(bhpc_error_data *ddata)
{

	if (chassis_ptr->error_fn) {
		chassis_ptr->error_fn(ddata);
	} else {
		printk("%s: No error callback for slot %d\n",
			   __FUNCTION__, ddata->slot);
	}
}
#endif
/*
 * Enables the links of a downstream PCIe port
 */
static
int enable_links(struct pci_dev *dsp_dev)
{
	int pos, link_down = 0;
	u16 val;
	u32 sval;
	int retryCnt =10;

	pos = pci_find_capability(dsp_dev, PCI_CAP_ID_EXP);
	pci_read_config_word(dsp_dev, pos + PCI_EXP_LNKCTL, &val);

	/* Enable the links */
	val &= ~PCI_EXP_LNKCTL_LD;
	pci_write_config_word(dsp_dev, pos + PCI_EXP_LNKCTL, val);
	mdelay(5);
	

	/* Disable this as per PLX FAE */
#if 0
	/* Force link retrain. May not be needed */
	val |= PCI_EXP_LNKCTL_RL;
	pci_write_config_word(dsp_dev, pos + PCI_EXP_LNKCTL, val);
#endif

	while ( retryCnt ) {
		/*
		 * Delay is required for the link to come up after a write
		 * to enable and retrain the link
		 */
		mdelay(20); 

		pci_read_config_dword(dsp_dev, pos + PCI_EXP_LNKCTL, &sval);
		if ((sval >> 20) & (0x7)) {
			u32 slot_ctrl = 0;

			pci_read_config_dword(dsp_dev, pos + PCI_EXP_SLTCTL, &slot_ctrl);
			printk(KERN_ERR "%s: Dev %s dsp Link is up. link stat %08x (%08x) "
				"slt ctrl 0x%8.8x\n", __FUNCTION__, pci_name(dsp_dev),
				sval, ((sval >> 20) & (0x7)), slot_ctrl);
			return (0) ; 		/* Link is up with either x1, x2 or x4 width */
		}
		else {
			retryCnt--;
		}
	}
	if (retryCnt <= 0) {
		u32 slot_ctrl = 0;


		pci_read_config_dword(dsp_dev, pos + PCI_EXP_SLTCTL, &slot_ctrl);

		/*
		 * If a blade is there, then try to bring the link up.
		 * If no powered-on blade, then link down is normal.
		 */
		if (slot_ctrl & (PCI_EXP_SLTSTA_PDS << 16)) {
			// link_down = link_recovery_pex8614(dsp_dev, pos);
			link_down = pex8614_errata_port_reset(dsp_dev, 0);
		}

		if (link_down) {
			/* Unable to retrain the link with SFM PCIe Switch- Link is down */
			printk(KERN_ERR "%s: DeV %s dsp link is not up. link stat %08x (%08x) "
				"slt ctrl 0x%8.8x\n", __FUNCTION__, pci_name(dsp_dev),
				sval, ((sval >> 20) & (0x7)), slot_ctrl);
		}
	}

	return (link_down);
}
/*
 * Perform any cleanup needed on MM DSP. Usually done before unloading the
 * hotplug service driver to prevent an unexpected interrupt from occuring
 * while the data structures are being deleted.
 *
 * Currently, the only thing being done is:
 *  1) Clear slot status register of each DSP
 *
 */
static void
cleanup_dsps(void)
{
	int i, pos;
	u32 regval32 = 0;
	bhpc_slot_map *ptr;
	struct controller *ctrl = NULL;
	struct pci_dev *dev = NULL;

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if (ptr) {
			ctrl = (struct controller*)ptr->private;
			if ((ctrl)&&(ctrl->pcie->port)) {

				dev = ctrl->pcie->port;
				pos = pci_find_capability(ctrl->pcie->port, PCI_CAP_ID_EXP);
				if (pos) {
					pci_read_config_dword(ctrl->pcie->port,
										  pos + PCI_EXP_SLTCTL, &regval32);
					pci_write_config_dword(ctrl->pcie->port,
										   pos + PCI_EXP_SLTCTL, regval32);
				}
			}
		}
	}
}
/*
 * search the table to find the entry that
 * matches this port device number. Store the
 * controller object in that entry.
 */
static
int store_slot_controller(struct controller *ctrl, int domain, int devnum)
{
	int rc = -ENODEV, i;
	bhpc_slot_map *ptr;
	struct pci_dev *pdev = ctrl->pcie->port;

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if ((ptr)&&((ptr->domain == domain)&&(ptr->port == devnum))) {
			if (!ptr->private) {
				ctrl_dbg(ctrl, "Adding to table: domain %d port %d slot %d\n",
						 domain, devnum, ptr->slot);
				ptr->private = (void*) ctrl;
				if (enable_links(pdev) !=0) {
					//printk(" Store controller ====================slot %d\n", i ); 
                 	continue;
                }
			} else {
				ctrl_info(ctrl, "Discard controller\n");
			}
			rc = 0;
			break;
		}
			
	}

	return (rc);
}
/*
 * dump_slot_table_map()
 *
 *  Dump slot-port map table on request. Just dumps to stdout.
 *
 */
static int
dump_slot_table_map(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	int i;
	bhpc_slot_map *map_ptr;
	struct controller *ctrl;
	char ptr[PAGE_SIZE];
	int regdump = 0;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(ptr, buffer, count)) {
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return (-EFAULT);
	}

	if (!sscanf(ptr, "%d", &regdump)) {
		return (-EFAULT);
	};

	if (regdump > 0) {
		for (i = 0; i < chassis.slot_map->max_slots; i++) {
			map_ptr = &chassis.slot_map->slots[i];
			if (map_ptr) {
				ctrl = get_slot_controller(map_ptr->slot);
				if (ctrl) {
					printk(KERN_ERR "%s: Found slot %d in chassis slot table "
						   "entry %d Port is %d Dev %s\n", __FUNCTION__, i,
						   map_ptr->slot, map_ptr->port, pci_name(ctrl->pcie->port));
				} else {
					printk(KERN_ERR "%s: Found slot %d in chassis slot table. "
						   "Port is %d Dev empty\n", __FUNCTION__,
						   map_ptr->slot, map_ptr->port);
				}
			}
		}
	}
	return (count);
}
/*
 * now search the table to find the entry that
 * matches this slot number. Return the
 * controller object in that entry, if there is one.
 */
static
struct controller *get_slot_controller(uint8_t slotno)
{
	int i;
	bhpc_slot_map *ptr;
	struct controller *ctrl = NULL;

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if ((ptr)&&(ptr->slot == slotno)) {
			printk(KERN_DEBUG "%s: Found slot %d in chassis slot table. "
				   "Port is %d\n", __FUNCTION__, slotno, ptr->port);
			ctrl = (struct controller*)ptr->private;
			break;
		}
			
	}

	return (ctrl);
}

/*
 * Search the table to find the entry that
 * matches this port number. Return the
 * logical slot number for that entry, if there is one.
 */

int bhpc_get_slot_number(uint8_t portnum, uint8_t *slotnum)
{
	int i, rc = -1;
	bhpc_slot_map *ptr;

	if (!slotnum) {
		printk(KERN_ERR "%s:Invalid input pointer\n", __FUNCTION__);
		return (rc);
	}

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if ((ptr)&&(ptr->port == portnum)) {
			printk(KERN_DEBUG "%s: Found port %d in chassis slot table. "
				   "Slot is %d\n", __FUNCTION__, ptr->port, ptr->slot);
			*slotnum = ptr->slot;
			rc = 0;
			break;
		}		
	}

	return (rc);
}
/*
 * bhpc_save_reference_count()
 *
 *   Save the reference count of the MM DSP that connects
 *   to a blade in a specified logical slot number.
 *
 *  Inputs:
 *    slotno - Specified logical slot number
 *    refcount - Reference count to be stored.
 *
 */
static void
bhpc_save_reference_count(uint8_t slotno, uint32_t refcount)
{
	int i;
	bhpc_slot_map *ptr;

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if ((ptr)&&(ptr->slot == slotno)) {
			ptr->refcount = refcount;
			pr_debug("%s: Found slot %d in chassis slot table. "
				   "Port is %d refcount %d\n", __FUNCTION__,
				   slotno, ptr->port, ptr->refcount);
			break;
		}		
	}
}
/*
 *  bhpc_get_reference_count()
 *
 *    Retrieve the stored reference count of the MM DSP that
 *    connects to the blade in a specified logical slot number
 *
 * Inputs:
 *   slotno - Specified logical slot number
 *   refcount - Pointer to pass back the reference count
 *
 */
static void
bhpc_get_reference_count(uint8_t slotno, uint32_t *refcount)
{

	int i;
	bhpc_slot_map *ptr;

	for (i = 0; i < chassis.slot_map->max_slots; i++) {
		ptr = &chassis.slot_map->slots[i];
		if ((ptr)&&(ptr->slot == slotno)) {
			*refcount = ptr->refcount;
			pr_debug("%s: Found slot %d in chassis slot table. "
				   "Port is %d refcount %d\n", __FUNCTION__,
				   slotno, ptr->port, *refcount);
			break;
		}		
	}

}
struct pci_dev*
bhpc_get_blade_bridge(uint8_t slot)
{
	struct pci_dev *dev = NULL;
	struct list_head *tmp;
	struct controller *ctrl;

	ctrl = get_slot_controller(slot);

	if (ctrl) {
		tmp = ctrl->pcie->port->subordinate->devices.next;
		if (tmp) {

			dev = pci_dev_b(tmp);
			if (dev) {
				pr_debug("%s: tmp dev %p %s rct %d\n", __FUNCTION__,
					   tmp, pci_name(dev), pci_dev_refcount(dev));
			} else {
				printk(KERN_ERR "%s: No dev\n", __FUNCTION__);
			}
		} else {
			printk(KERN_ERR "%s: could not get devices.next\n", __FUNCTION__);
		}
	}
	return (dev);
}
/*
 * bhpc_error_detected()
 *
 * Description:
 *   This method is used to notify a device driver when a non-recoverable PCI
 *   error has originated from one of its devices. If the error is considered
 *   a fatal error (anything that causes I/O accesses to fail), the state will
 *   be pci_channel_io_frozen. If the error is non-fatal (I/O accesses are not
 *   failing), the state will be pci_channel_io_normal.
 *
 *   The actual error handling will be deferred to NOS since Brocade is using
 *   a proprietary hardware design. So this method will always return
 *   PCI_ERS_RESULT_CAN_RECOVER, which tells the Linux AER code to clear the
 *   switch port and RC port error status and take no further action.
 *
 * When Called:
 *   Whenever the Linux AER service driver has detected an error and traced to
 *   the device specified by the dev input parameter.
 *
 * Inputs:
 *   dev - pointer to a device associated with an error.
 *   state - Indicates the current state of the device. Valid values are:
 *           pci_channel_io_normal = The device can accept I/O
 *                                   accesses (but not DMA)
 *           pci_channel_io_frozen = I/O to the device is blocked
 *           pci_channel_io_perm_failure = Whatever is in the slot is
 *                                         not responding
 *
 * Outputs:
 *   N/A
 *
 *
 * Returns:
 *   PCI_ERS_RESULT_CAN_RECOVER - No further action is needed from Linux
 *
 */
static pci_ers_result_t
bhpc_error_detected(struct pci_dev *dev,
					enum pci_channel_state state)
{

	/*
	 * XXX Need to call platform module log_fn().
	 *     For now, just display the device identification
	 *
	 * XXX - How do we get more info on the nature of the error???
	 *       Linux has read the error info out of the various
	 *       registers, but how do we get it???
	 */
	printk(KERN_ERR "%s: Error reported for device %s\n",
		   __FUNCTION__, pci_name(dev));

	pci_cleanup_aer_uncorrect_error_status(dev);
	return(PCI_ERS_RESULT_CAN_RECOVER);
}
/*
 * bhpc_aer_remove()
 *
 * Description:
 *   This method is used to do any processing needed when a device is 
 *   removed from the system.
 *
 * When Called:
 *   When a board is powered off or removed from the system, or when the
 *   BHPC driver is being removed from memory.
 *
 * Inputs:
 *   dev - pointer to a device being removed from the system
 *
 *
 * Outputs:
 *   N/A
 *
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static void
bhpc_aer_remove(struct pci_dev *dev)
{

	DBGPRINT(KERN_DEBUG "%s: Entry dev %s\n", __FUNCTION__, pci_name(dev));

	pci_disable_pcie_error_reporting(dev);

}
/*
 * bhpc_aer_probe()
 *
 * Description:
 *
 * When Called:
 *   Whenever a device is found that matches the ID defined by the
 *   pci_device_id table that was registered with Linux in the
 *   bhpc_initialize() method
 *
 *
 * Inputs:
 *   dev - pointer to structure representing a PCIe device
 *  *ent - Entry in the pci_device_id table that matched this device.
 *
 *
 * Outputs:
 *   N/A
 *
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static int
bhpc_aer_probe(struct pci_dev *dev, const struct pci_device_id *ent)
{
	int rc = 0;

	DBGPRINT(KERN_DEBUG "%s: Entry dev %s\n", __FUNCTION__, psi_name(dev));

	pci_save_state(dev);
	rc = pci_enable_pcie_error_reporting(dev);
	if (rc) {
		printk(KERN_INFO "%s: Could not enable error reporting for dev %d "
			   "bus %d ID 0x%x\n", __FUNCTION__, PCI_SLOT(dev->devfn),
			   dev->bus->number, dev->device);
		/*
		 * Some devices do not support error reporting. This is
		 * unfortunate, but don't return an error for it or else
		 * we will not be able to release the device object later
		 * when the blade is removed.
		 */
		rc = 0;
	}

	return (rc);

}
/*
 * bhpc_get_adapter_status()
 *
 * Description:
 *   This callback method is called by the LHPC to determine if a blade
 *   is in the specified chassis slot. If platform module has registered
 *   a get_adapter_status() callback with BHPC, then the BHPC will translate
 *   the input parameters to a form that platform module can understand and
 *   forward the request to platform module
 *
 * When Called:
 *   When the LHPC is notified of a blade hot-plug in event,
 *
 *
 * Inputs:
 *   slot - Object representing the Linux view of a PCI/PCIe hotpluggable slot.
 *
 *
 * Outputs:
 *   status - Pass back the blade present status. True = slot contains a
 *            blade with a closed slider (if it has one). False = slot is
 *            empty or slider is open (if it has one).
 *
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static int
bhpc_get_adapter_status (struct hotplug_slot *hp_slot, u8 *status)
{
	uint8_t portnum, slotnum = 0;
	int rc = 0;
	struct slot *slot;

	if ((!hp_slot)||(!status)) {
		printk(KERN_ERR "%s: Invalid intput pointer\n",
			   __FUNCTION__);
		rc = -EINVAL;
		goto Exit;
	}

	slot = (struct slot*)hp_slot->private;
	if (!slot) {
		printk("%s: No slot object available\n", __FUNCTION__);
		rc = -EINVAL;
		goto Exit;
	}
	portnum = (slot->ctrl->slot_cap >> 19) & 0x1FFF;
	printk("%s port number %d name %s\n", __FUNCTION__,
		   portnum, pci_name(slot->ctrl->pcie->port));
	rc = bhpc_get_slot_number(portnum, &slotnum);
	if (rc) {
		printk(KERN_DEBUG "%s: Cannot find slot number for port number %d\n",
			   __FUNCTION__, portnum);
		rc = -ENODEV;
		goto Exit;
	}

	if (chassis_ptr->plat_ops->get_adapter_status) {
		rc = chassis_ptr->plat_ops->get_adapter_status(slotnum, status);
		if (rc) {
			printk(KERN_DEBUG "%s: Cannot get adapter status for slot %d "
				   "port %d\n", __FUNCTION__, slotnum, portnum);
		}
	}


 Exit:

	return (rc);

}
/*
 * bhpc_get_power_status()
 *
 * Description:
 *   This callback method is called by the LHPC to determine the state of
 *   the slot power. If platform module has registered a get_power_status()
 *   callback with BHPC, then the BHPC will translate the input parameters to
 *   a form that platform module can understand and forward the request to
 *   platform module
 *
 *
 * When Called:
 *   When the LHPC is notified of a blade hot-plug in event,
 *
 *
 * Inputs:
 *   slot - Object representing the Linux view of a PCI/PCIe slot.
 *
 *
 * Outputs:
 *   *status - Pass back the slot power status. True = slot is powered on.
 *				 False = slot is powered off.
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static int
bhpc_get_power_status(struct hotplug_slot *hp_slot, u8 *status)
{
	uint8_t portnum, slotnum = 0;
	int rc = 0;
	struct slot *slot;

	if ((!hp_slot)||(!status)) {
		printk(KERN_ERR "%s: Invalid intput pointer\n",
			   __FUNCTION__);
		rc = -EINVAL;
		goto Exit;
	}

	slot = (struct slot*)hp_slot->private;
	if (!slot) {
		printk("%s: No slot object available\n", __FUNCTION__);
		rc = -EINVAL;
		goto Exit;
	}

	portnum = (slot->ctrl->slot_cap >> 19) & 0x1FFF;
	printk("%s port number %d name %s\n", __FUNCTION__,
		   portnum, pci_name(slot->ctrl->pcie->port));
	rc = bhpc_get_slot_number(portnum, &slotnum);
	if (rc) {
		printk(KERN_DEBUG "%s: Cannot find slot number for port number %d\n",
			   __FUNCTION__, portnum);
		rc = -ENODEV;
		goto Exit;
	}

	if (plat_ops->get_power_status) {
		rc = plat_ops->get_power_status(slotnum, status);
		if (rc) {
			printk(KERN_DEBUG "%s: Cannot get power status for slot %d "
				   "port %d\n", __FUNCTION__, slotnum, portnum);
		}
	}

 Exit:

	return (rc);

}

/*
 * bhpc_upstream_enable_get()
 *
 * Description:
 *  This function is used by the BHPC to determine if  upstream
 *  (blade to MM) PCI memory accesses via the PCIe switch downstream
 *  ports for a specified logical slot have been enabled.
 *
 *
 * When Called:
 *   When blade drivers call the fabsys_is_pci_enable_upstream() API. 
 *
 *
 * Inputs:
 *   slot - Identifies the chassis logical slot.
 *
 *
 * Outputs:
 *   state - 1 = Primary-side bus mastering is enabled for the device.
 *           0 = Primary-side bus mastering is disabled for the device.
 *
 *
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static int
bhpc_upstream_enable_get(uint8_t slot, uint8_t *state)
{
	struct controller *ctrl;
	u16 cmd;
	int rc = 0;

	if (!bhpc_initialized) {
		printk(KERN_ERR "%s: Module not initialized\n", __FUNCTION__);
		rc = -1;
		goto Exit;
	}

	if (!state) {
		printk(KERN_ERR "%s: Invalid output pointer for slot %d\n",
			   __FUNCTION__, slot);
		rc = -EINVAL;
		goto Exit;
	}

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -ENODEV;
		goto Exit;
	}

	pci_read_config_word(ctrl->pcie->port, PCI_COMMAND, &cmd);

	if (cmd & PCI_COMMAND_MASTER) {
		*state = 1;
	} else {
		*state = 0;
	}

 Exit:

	return (rc);

}
/*
 * bhpc_upstream_enable_set()
 *
 * Description:
 *  This function is used by the BHPC to enable upstream (blade to MM)
 *  PCI memory accesses via the PCIe switch downstream ports for a specified
 *  logical slot.
 *
 *
 *
 * When Called:
 *   When blade drivers call the fabsys_pci_enable_upstream() API. 
 *
 *
 * Inputs:
 *   
 *   slot - Identifies the chassis logical slot.
 *   state - 1 = enable primary-side bus mastering for the device.
 *           0 = disable primary-side bus mastering for the device.
 *
 * Outputs:
 *   None
 *
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 *
 */
static int
bhpc_upstream_enable_set(uint8_t slot, uint8_t state)
{
	struct controller *ctrl;
	int rc = 0;

	if (!bhpc_initialized) {
		printk(KERN_ERR "%s: Module not initialized\n", __FUNCTION__);
		rc = -1;
		goto Exit;
	}

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -1;
		goto Exit;
	}

	if (state) {
		pci_set_master(ctrl->pcie->port);
	} else {
		pci_clear_master(ctrl->pcie->port);
	}

 Exit:

	return (rc);

}

/*
 *
 */
static int
bhpc_restore_config(uint8_t slot)
{
	int rc = 0;

	printk("%s: NOT YET IMPLEMENTED. Slot %d\n", __FUNCTION__, slot);

	return (rc);

}

static int
validate_dsp_link(struct pci_dev *dsp_dev)
{
	int pos, link_down = 0;
	u16 link_status = 0;
	struct pci_dev *pdev = NULL;
	u32 vid = 0;

	if (!dsp_dev) {
		printk(KERN_ERR "%s: INVALID PCI DEV\n", __FUNCTION__);
		return (-1);
	}

	pdev = dsp_dev->bus->self;
	pci_read_config_dword(pdev, PCI_VENDOR_ID, &vid);

	/*
	 * First, make sure the USP is working. If it isn't, nothing
	 * else is going to work
	 */
	if (vid == ~0) {
		printk(KERN_ERR "%s: PEX switch is not operational. Vendor ID 0x%8.8x "
			   "DSP %s\n", __FUNCTION__, vid, pci_name(dsp_dev));
		return (-1);
	}
	show_dsp_link(dsp_dev);

	pos = pci_find_capability(dsp_dev, PCI_CAP_ID_EXP); 
	if (pos) {

		/* Link Status/control */
		pci_read_config_word(dsp_dev, pos + PCI_EXP_LNKSTA, &link_status);
		if (!(link_status &  PCI_EXP_LNKSTA_DLLLA)) { // 1

			/* if link is down, try to recover it */
			link_down = pex8614_errata_port_reset(dsp_dev, 0);

		}
	}

	return (link_down);
}

/*
 * bhpc_get_port_type()
 *
 * Description:
 *  This function is used to determine the type of port represented by
 *  the specified struct slot object.
 *
 *
 * When Called:
 *   When a request to perform a PCI configuration scan of a slot is received
 *
 *
 * Inputs:
 *   p_slot - the slot object of the MM PCIe switch downstream port
 *            that connects to the target blade
 *
 *
 * Outputs:
 *   port_type - used to pass back the type of the port. Valid values are
 *               PCI_EXP_TYPE_ROOT_PORT, PCI_EXP_TYPE_UPSTREAM,
 *               PCI_EXP_TYPE_DOWNSTREAM, and 0. The latter indicates
 *               an endpoint, which is what the secondary side of an
 *               NTP will look like.
 *
 *
 * Returns:
 *   0 - Success - port_type output contains a valid value
 *  <0 - Failure - port_type output does not contain a valid value
 *
 */
static int
bhpc_get_port_type(struct slot *p_slot, int *port_type)
{
	unsigned int devfn = 0, device;
	int pos, type, link_down = 0;
	u16 reg, cmd;
	u32 l, vid;
	struct pci_bus *bus;
	struct pci_dev *dsp_dev;

	dsp_dev = p_slot->ctrl->pcie->port;
	bus = dsp_dev->subordinate;
	device = p_slot->ctrl->slot_cap >> 19;
	devfn = PCI_DEVFN(0, 0);

	DBGPRINT("%s: ENTRY for slot caps 0x%x device %d pci_dev %d on bus %d subord bus %d "
		   "devfn 0x%x\n", __FUNCTION__,  p_slot->ctrl->slot_cap, device,
		   PCI_SLOT(dsp_dev->devfn), dsp_dev->bus->number,
		   dsp_dev->subordinate->number, devfn);

	link_down = validate_dsp_link(dsp_dev);
	if (link_down) {
		printk(KERN_ERR "%s: %s Cannot read port type - DSP link is down\n",
			   __FUNCTION__, pci_name(dsp_dev));
		goto Exit;
	}

	pci_bus_read_config_dword(bus, devfn, PCI_VENDOR_ID, &vid);
	pci_bus_read_config_word(bus, devfn, PCI_COMMAND, &cmd);

	DBGPRINT("%s: Vendor ID 0x%x cmd 0x%x\n", __FUNCTION__, vid, cmd);

	pci_read_config_dword(dsp_dev, PCI_MEMORY_BASE, &l);
	DBGPRINT("%s %d: parent bridge %s Mem rsrc 0x%x 0x%x\n", __FUNCTION__,
		   __LINE__, pci_name(dsp_dev), l, l << 16);

	pos = pci_bus_find_capability(bus, devfn, PCI_CAP_ID_EXP); 
	if (!pos) {

		u32 regval1 = 0, regval2 = 0;

		pci_read_config_dword(dsp_dev, PCI_VENDOR_ID, &regval1);
		pci_read_config_dword(dsp_dev->bus->self, PCI_VENDOR_ID, &regval2);

		printk(KERN_ERR "%s: cannot read port type for dev %d bus %d "
			   "VendorID 0x%x dsp %s 0x%8.8x usp %s 0x%8.8x\n",
			   __FUNCTION__, PCI_SLOT(devfn), bus->number,
			   vid, pci_name(dsp_dev), regval1,
			   pci_name(dsp_dev->bus->self), regval2);

		return -ENODEV;
	}

	/* This is the PCIE capabilities register */
	pci_bus_read_config_word(bus, devfn, pos + PCI_EXP_FLAGS, &reg);
	type = ((reg & PCI_EXP_FLAGS_TYPE) >> 4);

	DBGPRINT("%s %d: pos 0x%x port type %d reg 0x%x\n",
		   __FUNCTION__, __LINE__, pos, type, reg);

	*port_type = type;

 Exit:

	return (link_down);

}
/*
 * Disable SERR reporting on the OMM. Need to do this at failover
 * to prevent SERRs from being reported to the failing-over MM as
 * the new Active MM begins taking control
 */
static int
disable_omm_serr(int slot)
{
	unsigned long __iomem *csr = NULL;
	u32 *addr = NULL, regval32a, regval32b;
	u32 *dbg_ctrl_address, dbg_ctrl_data;
	int port_num, port_offset = 0, rc = 0;

	rc = map_sfm_usp(slot, &csr);
	if (rc) {
		printk(KERN_ERR "%s: Cannot map SFM%d bridge BAR\n",
			   __FUNCTION__, slot);
		goto Exit;
	}

	dbg_ctrl_address = (u32 *)(csr + (PEX8606_DEBUG_CTRL/ sizeof(int)));
	dbg_ctrl_data = ioread32(dbg_ctrl_address);
	port_num = (dbg_ctrl_data >> PEX8606_NTP_SHIFT) & 0xF;
	dbg_ctrl_data &= ~PEX8606_PORT_CONFIG_MASK;

	switch(port_num) {

	case BHPC_SFM_MM0_PORT: /* NTP is currently port 0, USP is port 4 */

		port_offset = PEX8606_PORT4_OFFSET;
		break;

	case BHPC_SFM_MM1_PORT: /* NTP is currently port 4, USP is port 0 */

		port_offset = PEX8606_PORT0_OFFSET;
		break;

	default: /* NTP port number is invalid */
		printk(KERN_ERR "%s: NTP is incorrectly programmed to port %d (0x%x)\n",
			   __FUNCTION__, port_num, dbg_ctrl_data);
		goto Exit;
	}

	/* Block hot resets and seeprom reload from the OMM */
	addr = (u32 *)(csr + (PEX8606_DEBUG_CTRL/ sizeof(int)));
	regval32a = ioread32(addr);
	regval32b = regval32a | 0x003E0000;
	iowrite32(regval32b, addr);

#if 0
	regval32b = ioread32(addr);
	printk(KERN_ERR "%s: ############### Debug control word before 0x%8.8x "
		   "after 0x%8.8x\n", __FUNCTION__, regval32a, regval32b);
#endif

	/* Block SERR from going upstream to the OMM */
	addr = (u32 *)(csr + ((0x3C + port_offset)/ sizeof(int)));
	regval32a = ioread32(addr);
	regval32b = regval32a & 0xFFFCFFFF;
	iowrite32(regval32b, addr);

#if 0
	regval32b = ioread32(addr);
	printk(KERN_ERR "%s: port %d port offset 0x%x regval32a 0x%8.8x "
		   "regval32b 0x%8.8x\n", __FUNCTION__, port_num, port_offset,
		   regval32a, regval32b);
#endif

	/* Block SERR from being reported to the OMM */
	addr = (u32 *)(csr + ((0x4 + port_offset)/ sizeof(int)));
	regval32a = ioread32(addr);
	iowrite32(regval32a, addr);
	regval32b = regval32a & 0x7;
	iowrite32(regval32b, addr);

#if 0
	regval32b = ioread32(addr);
	printk(KERN_ERR "%s: port %d port offset 0x%x regval32a 0x%8.8x "
		   "regval32b 0x%8.8x\n", __FUNCTION__, port_num, port_offset,
		   regval32a, regval32b);
#endif

 Exit:
	if (csr) {
		iounmap(csr);
	}
	return (rc);
}

/*
 * Function for promoting PEX8713 PCIe switch NTP to USP 
 *
 * The PEX8713 provides for one Upstream Port (USP) and two Non-Transparent
 * ports (NTP0, NTP1).
 *
 * The default USP always connects to the backplane PCI bus via a PCI-PCIe bridge.
 * The two default NTPs connect to the two backplane PCIe links.
 *
 * When a CP becomes Active, it must detect the type of port it sees (USP or NTPx).
 * If the port is NTP0/NTP1, then it must promote that port to be the USP. All the
 * required registers are available in config space.
 *
 * Algorithm:
 * 1.  Block hot reset from causing reload of the seeprom. Has to be done
 *     via mmio.
 * 2.  Determine which port number we are connected to via PCIe Link Capability
 *     Register 0x74[31:24]
 * 3.  Read VS0 Failover (0xF00) config register
 * 4.  Determine which port is current USP via bits 4:0
 * 5.  Determine if we are currently NT0 (via bits 12:8) or NT1 (via bits 20:16).
 * 6.  Change bits 4:0 to be our port number. This will make us the USP.
 * 7.  Change NTx (x = 0 or 1 as determined in step 4) bits to be the port that
 *     was the USP in step 3.
 * 8.  Write the new value back to the VS0 Failover register
 * 9.  Restore the BAR0 and Command registers (they get changed as a result
 *     of the write to VS0 Failover register)
 * 10. Issue hot reset via USP bridge control register secondary bus
 *     reset bit 0x3c[22].
 *
 * Inputs:
 *  csr -  virtual base address of the PEX CSR space.
 *  bar0 - Current value programmed into BAR 0 of the port being accessed
 *  bus - The PCI bus connected to the port being accessed.
 *  devfn - Encoded device and function number for the port being accessed
 *  pdev - Parent device. This is the MM/CP PCIe switch DSP connected to the
 *         port being accessed.
 *  
 */

static int
promote_port_pex8713(unsigned long __iomem *csr, u32 bar0, struct pci_bus *bus,
					 unsigned int devfn, struct pci_dev *pdev)
{
	int rc = 0, pos, portnum, oldusp, ntx;
	u32 reg32 = 0;
	u16 reg16 = 0, cmd = 0;
	u32 *dbg_ctrl_address, dbg_ctrl_data;
	u32 vsfailover = 0;


	/* 1. Prevent hot reset from re-loading the seeprom */
	dbg_ctrl_address = (u32 *)(csr + (PEX8713_VS_DEBUG/ sizeof(int)));
	dbg_ctrl_data = ioread32(dbg_ctrl_address);
	dbg_ctrl_data |= PEX8713_DISABLE_SEEPROM_LOAD;
	iowrite32(dbg_ctrl_data, dbg_ctrl_address);

	pos = pci_bus_find_capability(bus, devfn, PCI_CAP_ID_EXP);

	/*2. Read Link Capability register and get the port number */

	pci_bus_read_config_dword(bus, devfn, pos + PCI_EXP_LNKCAP, &reg32);
	portnum = (reg32 >> 24) & 0xFF;

	/* 3. Read VS Failover register */

	pci_bus_read_config_dword(bus, devfn, PEX8713_VS0_FAILOVER, &vsfailover);
	printk(KERN_INFO "%s: vsfailover via config 0x%8.8x addr 0x%x\n",
		   __FUNCTION__, vsfailover, PEX8713_VS0_FAILOVER);

	/* 4. Determine which port is currently USP */

	oldusp = vsfailover & PEX8713_PORT_MASK;
	DBGPRINT(KERN_ERR "%d : oldusp : %8.8x vsfailover %8.8x \n",
		   __LINE__, oldusp, vsfailover);

	/* 5. Determine if we are NT0 or NT1 */

	ntx = (vsfailover >> PEX8713_NT0_SHIFT) & PEX8713_PORT_MASK;
	DBGPRINT("%d: NT0 (ntx) is port %8.8x\n", __LINE__, ntx);
	if (portnum == ntx) {
		ntx = 0;	/* We are NT0 */
	} else {
		ntx = (vsfailover >> PEX8713_NT1_SHIFT) & PEX8713_PORT_MASK;
		if (portnum == ntx) {
			ntx = 1;	/* We are NT1 */
		} else {
			printk(KERN_ERR "%s: Port %d is not an NTP. link cap. 0x%8.8x "
				   "vs failover 0x%8.8x.\n", __FUNCTION__,
				   portnum, reg32, vsfailover);

			rc = -1; /* We are not an NT. Something is wrong, cannot proceed */
			goto Out;
		}
	}
	DBGPRINT(KERN_ERR "%s: We are port %d, NT%d\n", __FUNCTION__, portnum, ntx);

	/* 6. Change VS Failover Upstream port bits to our port number */

	vsfailover = vsfailover & ~PEX8713_PORT_MASK;
	vsfailover |= portnum;


	/* 7. Change NTx bits to the value of the current USP */

	if (ntx == 0) {
		oldusp = (oldusp << PEX8713_NT0_SHIFT);
		vsfailover = (vsfailover & ~PEX8713_NT0_MASK) | oldusp;
	} else {
		oldusp = (oldusp << PEX8713_NT1_SHIFT);
		vsfailover = (vsfailover & ~PEX8713_NT1_MASK) | oldusp;
	}

	/* Save command register so it can be restored */
	pci_bus_read_config_word(bus, devfn, PCI_COMMAND, &cmd);

	/* 8. Write new value back to VS Failover register and we become the USP */

	printk(KERN_INFO "%s: Writing new vsfailover 0x%8.8x\n",
		   __FUNCTION__, vsfailover);

	pci_bus_write_config_dword(bus, devfn, PEX8713_VS0_FAILOVER, vsfailover);
	udelay(100);

	/* 9 Restore BAR0 and command registers */
	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, bar0);
	pci_bus_write_config_word(bus, devfn, PCI_COMMAND, cmd);

	/*
	 * We actually need a config read here. Not sure why, but without it,
	 * subsequent mmio will cause a machine check unless we delay for
	 * something like 100ms
	 */
	pci_bus_read_config_dword(bus, devfn, PEX8713_VS0_FAILOVER, &vsfailover);
	printk(KERN_INFO "%s: Read back config new vsfailover 0x%8.8x addr 0x%x\n",
		   __FUNCTION__, vsfailover, PEX8713_VS0_FAILOVER);

	/* 10. Issue hot reset */

	pci_bus_read_config_word(bus, devfn, PCI_BRIDGE_CONTROL, &reg16);
	reg16 |= (PCI_BRIDGE_CTL_BUS_RESET);
	pci_bus_write_config_word(bus, devfn, PCI_BRIDGE_CONTROL, reg16);

	udelay(100); /* propagation delay for the command */

	reg16 &= ~PCI_BRIDGE_CTL_BUS_RESET;
	pci_bus_write_config_word(bus, devfn, PCI_BRIDGE_CONTROL, reg16);

	msleep(20);/* Allow time for all downstream devices to complete */

 Out:

	return (rc);
}

/*
 * Function for promoting PEX8664 PCIe switch NTP to USP 
 *
 * Assumes only ports 0 and 1 will be used for NTP and USP. Also
 * assumes this port is currently an NTP.
 *
 * Algorithm:
 *  1. Set bit 9 of Debug Control register (offset 0x350 from BAR0
 *     address) to enable SW to change the USP/NTP port mapping.
 *     This register is accessible only via memory-mapped I/O.
 *  2. Read VS0 (offset 0x360) register bits 12,10:8 to determine the current
 *     NTP port number. This register is accessible only via memory-mapped I/O.
 *  3. Program VS0 register bits 4,2:0 for the new USP and
 *     bits 12,10:8 for the new NTP.
 *  4. Clear bit 9 of the Debug Control register
 *  5. Reset the bus hierarchy via the PCI command register. This is
 *     a config access.
 */
static int
promote_port_pex8664(unsigned long __iomem *csr, u32 bar0, struct pci_bus *bus,
					 unsigned int devfn, struct pci_dev *pdev)
{
	int rc = 0;
	/* XXX - Figure this out later */

	return (rc);
}

/*
 * Give the SFM PCIe switch USP a temporary base address
 * and ioremap it for the caller. Pass back the virtual
 * address, if successful
 *
 * Returns:
 *  0 - Successful
 *  Non-zero - could not map the address.
 *
 */
static int
map_sfm_usp(uint slot, unsigned long __iomem **base)
{
	struct controller *ctrl;
	struct slot *p_slot;
	struct pci_dev *dsp_dev = NULL;
	struct pci_bus *bus;
	unsigned int devfn, device;
	struct pci_controller *hose;
	unsigned long __iomem *csr = NULL;
	resource_size_t paddr = 0;
	int rc = 0;
	u32 regval, bar0;
	u16 cmd;

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -1;
		goto Exit2;
	}

	p_slot = ctrl->slot;
	if (!p_slot) {
		printk(KERN_ERR "%s: No slot object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -2;
		goto Exit2;
	}

	/*
	 * First, make sure the link to the SFM is up.
	 * If it isn't, do not proceed or else PCIe
	 * errors will occur
	 */

	dsp_dev = p_slot->ctrl->pcie->port;
	rc = validate_dsp_link(dsp_dev);
	if (rc) {
		printk(KERN_ERR "%s: Dev %s link is not up 0x%x\n", __FUNCTION__,
			   pci_name(dsp_dev), rc);
		rc = BHPC_DSP_LINK_DOWN;
		goto Exit2;
	}
	bus = dsp_dev->subordinate;
	device = p_slot->ctrl->slot_cap >> 19;
	devfn = PCI_DEVFN(0, 0);

	pci_read_config_dword(dsp_dev, PCI_MEMORY_BASE, &regval);
	bar0 = regval << 16;

	/*
	 * Write address to BAR0 of the blade's USP/NTP
	 */

	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, bar0);
	pci_bus_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &regval);

	/* Enable device memory decoders */

	pci_bus_read_config_word(bus, devfn, PCI_COMMAND, &cmd);
	cmd |= PCI_COMMAND_MEMORY;
	pci_bus_write_config_word(bus, devfn, PCI_COMMAND, cmd);

	/* Calculate resource from BAR0 value and hose pci offset */
	hose = pci_bus_to_host(bus);
	paddr = bar0 + hose->pci_mem_offset;

#if 0
	printk(KERN_ERR "%s %d: BAR0 0x%x dsp cmd 0x%x pci_offset 0x%llx paddr 0x%llx\n",
		   __FUNCTION__, __LINE__, regval, cmd, hose->pci_mem_offset, (u64)paddr);

#endif
	/* ioremap() the resource */

	csr = ioremap_nocache(paddr, PEX8606_TOTAL_CSR_SIZE);
	if (!csr) {
		printk(KERN_ERR "%s: Cannot ioremap blade bridge\n",
			   __FUNCTION__);

		rc = -3;
		goto Exit2;
	}

	*base = csr;

 Exit2:

	return (rc);

}
/*
 * Check the state of the SFM DSP links to the asics. If they are not
 * up, give them some time to come up. Only print if we had to wait, since
 * that shouldn't happen.
 */
int
show_sfm_dev_data(unsigned long __iomem *csr, struct pci_dev *pdev, int line)
{
	u32 port1_link = 0, port5_link = 0, port7_link = 0;
	u32 port1_slotctrl = 0, port5_slotctrl = 0, port7_slotctrl = 0;
	u32 *mem_address, not_ready = 1, count = 0;
	int rc = 0;

	not_ready = 1;
	do {

		mem_address = (u32 *)(csr + ((PEX8606_PORT1_OFFSET + 0x7C)/ sizeof(int)));
		port1_slotctrl = ioread32(mem_address);
		port1_slotctrl = (port1_slotctrl >> 19) & 0xFFFF;
		mem_address = (u32 *)(csr + ((PEX8606_PORT1_OFFSET + 0x78)/ sizeof(int)));
		port1_link = ioread32(mem_address);

		mem_address = (u32 *)(csr + ((PEX8606_PORT5_OFFSET + 0x7C)/ sizeof(int)));
		port5_slotctrl = ioread32(mem_address);
		port5_slotctrl = (port5_slotctrl >> 19) & 0xFFFF;
		mem_address = (u32 *)(csr + ((PEX8606_PORT5_OFFSET + 0x78)/ sizeof(int)));
		port5_link = ioread32(mem_address);

		mem_address = (u32 *)(csr + ((PEX8606_PORT7_OFFSET + 0x7C)/ sizeof(int)));
		port7_slotctrl = ioread32(mem_address);
		port7_slotctrl = (port7_slotctrl >> 19) & 0xFFFF;
		mem_address = (u32 *)(csr + ((PEX8606_PORT7_OFFSET + 0x78)/ sizeof(int)));
		port7_link = ioread32(mem_address);

		if ((port1_link & 0x20000000)||(port5_link & 0x20000000)||
			(port7_link & 0x20000000)){
			not_ready = 0; 
		} else {
			msleep(20);
			count++;
		}

	} while ((not_ready)&&(count < 10));

	if ((not_ready)||(count > 0)) {
		printk(KERN_ERR "%s %d: Waited %d times. Parent dev %s Prt%d 0x%8.8x "
			   "Prt%d 0x%8.8x Prt%d 0x%8.8x nrdy %d\n",
			   __FUNCTION__, line, count, pci_name(pdev),
		   port1_slotctrl, port1_link, port5_slotctrl, port5_link,
		   port7_slotctrl, port7_link, not_ready);

		rc = BHPC_DSP_LINK_DOWN;
	}

	return (rc);
}
/*
 * Check the state of all SFM links. If they are not up, give them some
 * time to come up. The DSP link registers are not accessible via
 * config space, so have to give the switch a temporary base address
 */
int
pciehp_show_sfm_data(uint8_t slot)
{

	struct controller *ctrl;
	struct slot *p_slot;
	struct pci_dev *dsp_dev = NULL;
	struct pci_bus *bus;
	unsigned int devfn, device;
	u32 regval, bar0;
	unsigned long __iomem *csr = NULL;
	u16 cmd;
	int rc = 0, pos, not_ready = 1, count = 0;
	struct pci_controller *hose;
	resource_size_t paddr = 0;

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);

		goto Exit2;
	}

	p_slot = ctrl->slot;
	if (!p_slot) {
		printk(KERN_ERR "%s: No slot object for slot %d\n", __FUNCTION__, slot);
		rc = -1;
		goto Exit2;
	}

	dsp_dev = p_slot->ctrl->pcie->port;
	bus = dsp_dev->subordinate;
	device = p_slot->ctrl->slot_cap >> 19;
	devfn = PCI_DEVFN(0, 0);

	/* 1st, make sure DSP link is up before proceeding */
	not_ready = 1;
	pos = pci_find_capability(dsp_dev, PCI_CAP_ID_EXP);
	do {

		pci_read_config_dword(dsp_dev, pos + PCI_EXP_LNKCTL, &regval);

		if (regval & 0x20000000) {
			not_ready = 0; 
		} else {
			msleep(20);
			count++;
		}

	} while ((not_ready)&&(count < 25)) ;

	if (not_ready) {
		rc = BHPC_DSP_LINK_DOWN;
		goto Exit2;
	}

	if (count > 0) {
		pci_read_config_dword(dsp_dev, pos + PCI_EXP_LNKCTL, &regval);
		printk(KERN_ERR "%s: Waited %d times for DSP %s link up 0x%8.8x nrdy %d\n",
			   __FUNCTION__, count, pci_name(dsp_dev), regval, not_ready);
	}

	/*
	 * Determine temporary BAR0 address from the window assigned to
	 * the MM's DSP.
	 */

	pci_read_config_dword(dsp_dev, PCI_MEMORY_BASE, &regval);
	bar0 = regval << 16;

	/*
	 * Write address to BAR0 of the blade's USP/NTP
	 */

	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, bar0);
	pci_bus_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &regval);

	/* Enable device memory decoders */

	pci_bus_read_config_word(bus, devfn, PCI_COMMAND, &cmd);
	cmd |= PCI_COMMAND_MEMORY;
	pci_bus_write_config_word(bus, devfn, PCI_COMMAND, cmd);

	/* Calculate resource from BAR0 value and hose pci offset */
	hose = pci_bus_to_host(bus);
	paddr = bar0 + hose->pci_mem_offset;

	/* ioremap() the resource */

	csr = ioremap_nocache(paddr, PEX8606_TOTAL_CSR_SIZE);
	if (!csr) {
		printk(KERN_ERR "%s: Cannot temporarily ioremap blade bridge\n",
			   __FUNCTION__);

		rc = -1;
		goto Exit2;
	}

	rc = show_sfm_dev_data(csr, dsp_dev, __LINE__);


	if (csr) {
		iounmap(csr);
	}

	/* Erase the temporary base address */
	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, 0);

 Exit2:

	return (rc);

}

static int
bhpc_command(uint8_t slot, int command)
{
	int rc = 0;

	switch (command) {

	case BHPC_DISABLE_OMM_SERR:

		rc = disable_omm_serr(slot);
		break;

	case BHPC_SHOW_SFM_LINKS:

#ifndef CONFIG_SILKWORM_DCX
		rc = pciehp_show_sfm_data(slot);
#else
		printk(KERN_ERR "%s: Unsupported request BHPC_SHOW_SFM_LINKS\n",
			   __FUNCTION__);
#endif
		break;

	default:

		printk(KERN_ERR "%s: Unrecognized command %d\n", __FUNCTION__, command);
	}

	return (rc);

}

/*
 * Function for promoting PEX8606 PCIe switch NTP to USP. Also
 * works for PEX8632 and others. Will not work for PEX8664.
 *
 * Algorithm:
 *  1. Determine the current NTP port number from bits [25:24] and
 *	   current USP port number from bits [11:8] of Debug Control
 *     register (offset 0x1DC from BAR0 This register is accessible
 *     only via memory-mapped I/O.
 *  2. Set bit 15 of the Debug Control register to enable SW to change
 *     the USP/NTP port mapping.
 *  3. Program Debug Control register bits [11:8] for the new USP and
 *     bits [25:24] for the new NTP.
 *  4. Disable seeprom load on hot reset. Otherwise, the switch
 *     NTP/USP go back to the power-on default states.
 *  5. Clear bit 15 of the Debug Control register.
 *  6. Reset the bus hierarchy via the Bridge Control register of the
 *     parent downstream bridge on the MM. This is a config access.
 *
 * Input(s):
 *   csr - virtual base address of the PEX CSR space.
 *   *bus - Bus the blade bridge primary side connects to
 *   *devfn - Defines the PCI device number of the blade bridge primary side.
 *   *pdev - Parent downstream port on the MM
 */
static int
promote_port_pex8606(unsigned long __iomem *csr, u32 bar0, struct pci_bus *bus,
					 unsigned int devfn, struct pci_dev *pdev)
{

	u32 *dbg_ctrl_address, dbg_ctrl_data, port_mask;
	u16 ctrl;
	int rc = 0, port_num;

	/*
	 * Disable unused port to prevent it from getting the
	 * device hung in hot reset state.
	 */

	dbg_ctrl_address = (u32 *)(csr + (0x234/ sizeof(int)));
	dbg_ctrl_data = ioread32(dbg_ctrl_address);

	if ((dbg_ctrl_data & 0x10) == 0) {

		dbg_ctrl_data |= 0x10;

		iowrite32(dbg_ctrl_data, dbg_ctrl_address);
		mdelay(1);
		dbg_ctrl_data = ioread32(dbg_ctrl_address);

		printk(KERN_ERR "%s: Disable unused port for %s 0x%8.8x\n",
			   __FUNCTION__, pci_name(pdev), dbg_ctrl_data);
	}
	/*
	 * Set the appropriate bits for PEX8606 Switch
	 */

	dbg_ctrl_address = (u32 *)(csr + (PEX8606_DEBUG_CTRL/ sizeof(int)));
	dbg_ctrl_data = ioread32(dbg_ctrl_address);

	DBGPRINT( "%s: Debug Ctrl Before: 0x%8.8x  Addr: 0x%p csr 0x%p addr 0x%p\n",
		   __FUNCTION__, dbg_ctrl_data, dbg_ctrl_address, csr);


	if (!(dbg_ctrl_data & PEX8606_NT_MODE_ENABLED)) {
		printk("%s %d: NT MODE IS NOT ENABLED dbg_ctrl 0x%x\n",
			   __FUNCTION__, __LINE__, dbg_ctrl_data);
		rc = -1;
		goto Exit;
	}

	/*
	 * Bits [11:8] define USP, bits [25:24] define NTP
	 * Since we know we are talking to the NTP, then the
	 * current value of bits [25:24] tells us the port
	 * number.
	 */
	port_num = (dbg_ctrl_data >> PEX8606_NTP_SHIFT) & 0xF;
	dbg_ctrl_data &= ~PEX8606_PORT_CONFIG_MASK;
	DBGPRINT("%s: Current NTP is port %d\n", __FUNCTION__, port_num);

	switch(port_num) {

	case BHPC_SFM_MM0_PORT: /* NTP is currently port 0, USP is port 4 */

		port_mask = PEX8606_PORT0_USP_PORT4_NTP; /* Make port 0 the USP */
		break;

	case BHPC_SFM_MM1_PORT: /* NTP is currently port 4, USP is port 0 */

		/* Make port 4 USP, port 0 NTP */
		port_mask = PEX8606_PORT4_USP_PORT0_NTP; /* Make port 4 the USP */
		break;

	default: /* NTP port number is invalid */
		printk(KERN_ERR "%s: NTP is incorrectly programmed to port %d (0x%x)\n",
			   __FUNCTION__, port_num, dbg_ctrl_data);
		rc = -1;
		goto Exit;
	}


	DBGPRINT("%s %d: dbg_ctrl_data 0x%x address 0x%p\n",
		   __FUNCTION__, __LINE__, dbg_ctrl_data, dbg_ctrl_address);

	/* Enable SW control */
	dbg_ctrl_data |= PEX8606_SW_CONFIG_CTRL;
	iowrite32(dbg_ctrl_data, dbg_ctrl_address);

	/* Change the port config */
	dbg_ctrl_data |= (port_mask | PEX8606_DISABLE_SEEPROM_LOAD);
	iowrite32(dbg_ctrl_data, dbg_ctrl_address);

	/* Restore BAR0 */
	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, bar0);


	/* Lock it down to prevent accidental reconfig */
	dbg_ctrl_data &= ~PEX8606_SW_CONFIG_CTRL;
	iowrite32(dbg_ctrl_data, dbg_ctrl_address);

	DBGPRINT("%s: Debug Ctrl After 0x%8.8x  Addr: 0x%p\n",
		   __FUNCTION__, dbg_ctrl_data, dbg_ctrl_address);

	mdelay(2);


	/*
	 * Issue a hot-reset to the switch. Needs to be done
	 * from the parent MM downstream port.
	 *
	 */
#if 0
	printk(KERN_ERR "%s %d: Doing bus reset from device %s\n", __FUNCTION__, __LINE__,
		   pci_name(pdev));
#endif

	pci_read_config_word(pdev, PCI_BRIDGE_CONTROL, &ctrl);

	ctrl |= (PCI_BRIDGE_CTL_BUS_RESET);
	pci_write_config_word(pdev, PCI_BRIDGE_CONTROL, ctrl);

	// msleep(2); /* This delay causes PCT when trying to access C3

	ctrl &= ~PCI_BRIDGE_CTL_BUS_RESET;
	pci_write_config_word(pdev, PCI_BRIDGE_CONTROL, ctrl);

	msleep(20);/* give it some time to propagate downstream. */
	// msleep(200);

 Exit:

	return (rc);

}
/*
 * bhpc_promote_port()
 *
 * Description:
 *  This function is used to change the mode of the PCIe port on a blade
 *   from a non-transparent port to an upstream port.
 *
 *  Changing the USP/NTP port configuration requires accessing internal
 *  CSRs of the blade PCIe switch via its BAR0. But the switch has not yet
 *  been scanned, so the switch BAR0 does not yet have a base address. A
 *  temporary base address needs to be provided to the switch BAR0.
 *
 *  1. Determine the memory window provided by the parent MM downstream
 *     port. The base of this window will be used as the temporary base
 *     address of the blade PCIe switch BAR0. This address is guaranteed
 *     not to be in use at this point.
 *
 *  2. Program the blade PCIe switch BAR0 and enable the memory decoder.
 *  3. Calculate the resource value for this BAR value. It is not always
 *     the same as the BAR value itself.
 *  4. Obtain a virtual address for the BAR.
 *  5. Call the appropriate port promotion routine based on the
 *     blade PCIe switch ID. Default to pex8606 if the ID is not known.
 *  6. Release the virtual address.
 *  7. Clear the blade PCIe switch BAR0 temporary base address.
 *  
 * When Called:
 *   during blade initialize and during failover PCI scan
 *
 *
 * Inputs:
 *   p_slot - the slot object of the MM PCIe switch downstream port
 *            that connects to the target blade
 *
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */
static int
bhpc_promote_port(struct slot *p_slot)
{
	int rc = 0;
	unsigned int devfn, device;
	struct pci_bus *bus;
	struct pci_dev *dsp_dev;
	u16 cmd;
	u32 regval, bar0;
	struct pci_controller *hose;
	resource_size_t paddr = 0;
	unsigned long __iomem *csr = NULL;

	dsp_dev = p_slot->ctrl->pcie->port;
	bus = dsp_dev->subordinate;
	device = p_slot->ctrl->slot_cap >> 19;
	devfn = PCI_DEVFN(0, 0);

	DBGPRINT("%s: Entry for device %d on bus %d subord bus %d devfn %d\n",
		   __FUNCTION__,
		   device,
		   p_slot->ctrl->pcie->port->bus->number,
		   p_slot->ctrl->pcie->port->subordinate->number, devfn);


	/*
	 * Determine temporary BAR0 address from the window assigned to
	 * the MM's DSP.
	 */

	pci_read_config_dword(dsp_dev, PCI_MEMORY_BASE, &regval);
	bar0 = regval << 16;
	DBGPRINT("%s %d: parent bridge %s Mem rsrc 0x%x 0x%x\n", __FUNCTION__,
		   __LINE__, pci_name(dsp_dev), regval, bar0);

	/*
	 * Write address to BAR0 of the blade's USP/NTP
	 */

	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, bar0);
	pci_bus_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &regval);

	/* Enable device memory decoders */

	pci_bus_read_config_word(bus, devfn, PCI_COMMAND, &cmd);
	cmd |= PCI_COMMAND_MEMORY;
	pci_bus_write_config_word(bus, devfn, PCI_COMMAND, cmd);

	/* Calculate resource from BAR0 value and hose pci offset */
	hose = pci_bus_to_host(bus);
	paddr = bar0 + hose->pci_mem_offset;


	DBGPRINT("%s %d: BAR0 0x%x dsp cmd 0x%x pci_offset 0x%llx paddr 0x%llx\n",
		   __FUNCTION__, __LINE__, regval, cmd, hose->pci_mem_offset, (u64)paddr);

	/* ioremap() the resource */

	csr = ioremap_nocache(paddr, PEX8606_TOTAL_CSR_SIZE);
	if (!csr) {
		printk(KERN_ERR "%s: Cannot temporarily ioremap blade bridge\n",
			   __FUNCTION__);

		rc = -1;
		goto Exit;
	}


	pci_bus_read_config_dword(bus, devfn, PCI_VENDOR_ID, &regval);
	regval = (regval >> 16) & 0xFFFF;
	switch (regval) {

	case PEX8606_DEVICE_ID:

		rc = promote_port_pex8606(csr, bar0, bus, devfn, dsp_dev);
		break;

	case PEX8664_DEVICE_ID:

		rc = promote_port_pex8664(csr, bar0, bus, devfn, dsp_dev);
		break;

	case PEX8713_DEVICE_ID:
	case PEX8713_DEVICE_NTP0_ID:
	case PEX8713_DEVICE_NTP1_ID:

		rc = promote_port_pex8713(csr, bar0, bus, devfn, dsp_dev);
		break;

	case 0xffff:

		printk(KERN_ERR "%s: Invalid device Vendor ID 0x%x\n",
			   __FUNCTION__, regval);
		rc = -1;
		break;

	default:
		printk(KERN_ERR "%s: Unrecognized PCIe switch Vendor ID 0x%x "
			   "Default to 8606\n",
			   __FUNCTION__, regval);
		/*
		 * If we don't know what it is, try the 8606
		 * algorithm.
		 */
		rc = promote_port_pex8606(csr, bar0, bus, devfn, dsp_dev);

	}

 Exit:

	if (csr) {
		iounmap(csr);
	}

	/* Erase the temporary base address */
	pci_bus_write_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, 0);

	return (rc);
}
/*
 *  refcount_fixup()
 *
 *    Adjust the reference count of a specified device to what it
 *    was when the PCI scan was completed. This is to prevent a
 *    memory leak or kernel panic if a user of the device did
 *    something to alter the reference count.
 *
 * Inputs:
 *   pdev - pci_dev object of the target device
 *   expected_refcount - what the refernce count should be.
 *
 */
static void
refcount_fixup(struct pci_dev *pdev, int expected_refcount)
{
	int actual_refcount, delta, i;

	if (pdev) {
		actual_refcount = pci_dev_refcount(pdev);

		if (expected_refcount > actual_refcount) {
			delta = expected_refcount - actual_refcount;
			pr_debug("%s: Increment actual refcount by %d\n",
				   __FUNCTION__, delta);
			for (i = 0; i < delta; i++) {
				pci_dev_get(pdev);
			}
		} else {
			delta = actual_refcount - expected_refcount;
			pr_debug("%s: Decrement actual refcount by %d\n",
				   __FUNCTION__, delta);

			for (i = 0; i < delta; i++) {
				pci_dev_put(pdev);
			}
		}

		pr_debug("%s: After fixup actual refcount %d expected refcount %d\n",
			   __FUNCTION__, pci_dev_refcount(pdev), expected_refcount);

	} else {
		printk(KERN_ERR "%s: No pci_dev device. Cannot fixup refcount\n",
			   __FUNCTION__);
	}

}
/*
 * bhpc_blade_remove()
 *
 * Description:
 *  This function is used to clean up any PCI objects that were
 *  created when a blade was scanned
 *
 *
 * When Called:
 *   Whenever a blade is being un-instantiated.
 *
 *
 * Inputs:
 *   slot - defines the chassis logical slot
 *
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */
static int
bhpc_blade_remove(uint8_t slot)
{
	struct controller *ctrl;
	struct slot *p_slot;
	int rc = 0;
	struct pci_dev *pdev = NULL, *tdev = NULL;

	if (!bhpc_initialized) {
		printk(KERN_ERR "%s: Module not initialized\n", __FUNCTION__);
		rc = -1;
		goto Exit;
	}

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -1;
		goto Exit;
	}

	p_slot = ctrl->slot;
	if (!p_slot) {
		printk(KERN_ERR "%s: No slot object for slot %d\n", __FUNCTION__, slot);
		rc = -1;
		goto Exit;
	}

	/*
	 * Check reference count of the blade bridge. If it's not what it was when
	 * the PCI scan was completed, then correct it. Otherwise, a memory leak or
	 * kernel panic will occur.
	 *
	 * Switched to using pci_get_slot() because it reliably returns NULL
	 * if the object has been deleted. But it also changes the reference
	 * count, so it must be restored before trying to validate it.
	 */

	tdev = pci_get_slot(ctrl->pcie->port->subordinate, PCI_DEVFN(0, 0));
	if (tdev) {
		int refcount;

		pci_dev_put(tdev); /* Restore the refcount before validating it */

		bhpc_get_reference_count(slot, &refcount);
		printk(KERN_DEBUG "%s: SFM USP dev %p %s REFCOUNT is %d expected %d\n",
			  __FUNCTION__, tdev, pci_name(tdev),
			  pci_dev_refcount(tdev), refcount);

		if (pci_dev_refcount(tdev) != refcount) {
			printk(KERN_ERR "%s: WARNING - UNEXPECTED REFERENCE COUNT %d FOR DEV %s "
				   "EXPECTED %d\n", __FUNCTION__, pci_dev_refcount(tdev),
				   pci_name(tdev), refcount);
			refcount_fixup(tdev, refcount);
		}

	} else {
		/*
		 * This is not necessarily an error. It just means
		 * the PCIe objects have already been removed, and the
		 * caller may not be able to know that - so just exit.
		 */
		printk(KERN_ERR "%s: No PCIe objects to remove\n", __FUNCTION__);
		goto Exit;
	}

	DBGPRINT("%s %d: Calling pciehp_unconfigure_device for slot %d\n",
		   __FUNCTION__, __LINE__, slot);

	pdev = p_slot->ctrl->pcie->port;

	if (pdev) {
		int pos;
		u32 regval32 = 0;

		/* Ensure the slot status register is cleared */
		pos = pci_find_capability(ctrl->pcie->port, PCI_CAP_ID_EXP);
		if (pos) {
			pci_read_config_dword(ctrl->pcie->port,
								  pos + PCI_EXP_SLTCTL, &regval32);
			pci_write_config_dword(ctrl->pcie->port,
								   pos + PCI_EXP_SLTCTL, regval32);
		}

		pr_debug("%s %d: MM DSP  %s refcount %d\n",
				 __FUNCTION__, __LINE__, pci_name(pdev),
				 pci_dev_refcount(pdev));
	}

	rc = pciehp_unconfigure_device(p_slot);
	if (rc) {
		printk(KERN_ERR "%s: pciehp unconfigure device failed %d\n",
			   __FUNCTION__, slot);
	}

	pr_debug("%s %d: MM DSP refcount %d\n",
		   __FUNCTION__, __LINE__, pci_dev_refcount(pdev));

 Exit:

	return (rc);
}
/*
 * bhpc_blade_insert()
 *
 * Description:
 *  This function is the entry point to perform a PCI configuration scan
 *  of a blade. 
 *
 *   1. Determine the type of switch port we are connected to
 *      on the blade.
 *   2. If the port is not the upstream port, then "promote" it
 *      to the upstream port.
 *   3. Call the Linux HPC entry point for performing a configuration
 *      scan of a slot.
 *   4. Pass back a pointer to the list of PCI devices found by
 *      the scan.
 *
 * When Called:
 *   During blade initialize and during failover PCI scan
 *
 *
 * Inputs:
 *   slot - specifies the chassis logical slot to be scanned.
 *
 * Outputs:
 *   dev - used to pass back a pointer to the upstream port. This is
 *         the "blade bridge"
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

static int
bhpc_blade_insert(uint8_t slot, struct pci_dev **dev)
{
	int rc = 0, port_type = 0, refcount;
	struct controller *ctrl;
	struct slot *p_slot;
	struct pci_dev *tdev = NULL;

	pr_debug("%s: Entry Slot %d initialized %d\n",
		   __FUNCTION__, slot, bhpc_initialized);

	if (!bhpc_initialized) {
		printk(KERN_ERR "%s: Module not initialized\n", __FUNCTION__);
		rc = -1;
		goto Exit2;
	}

	if (!dev) {
		printk(KERN_ERR "%s: invalid pci device input parameter for slot %d\n",
			   __FUNCTION__, slot);
		rc = -1;
		goto Exit2;
	}

	ctrl = get_slot_controller(slot);
	if (!ctrl) {
		printk(KERN_ERR "%s: No controller object for slot %d\n",
			   __FUNCTION__, slot);
		rc = -1;
		goto Exit2;
	}

	/*
	 * There is only ever one slot with PCIe unless we are dealing
	 * with an upstream port - which we are not.
	 */
	p_slot = ctrl->slot;
	if (!p_slot) {
		printk(KERN_ERR "%s: No slot object for slot %d\n", __FUNCTION__, slot);
		rc = -1;
		goto Exit2;
	}


	/*
	 * Disable error reporting at the DSP. Do this before touching
	 * any config registers to avoid getting interrupts due to
	 * link problems. These don't cause machine checks and we
	 * have other ways to handle these types of problems.
	 */
	if (interrupts_enabled) {
		bhpc_disable_scan_error_reports(p_slot->ctrl->pcie->port);
	}

	/*
	 * See what state links are in to begin with
	 */
#ifndef CONFIG_SILKWORM_DCX
	pciehp_show_sfm_data(slot);
#endif

	rc = bhpc_get_port_type(p_slot, &port_type);
	if (rc) {
		printk(KERN_ERR "%s: Cannot determine the upstream port type "
			   "for slot %d\n", __FUNCTION__, slot);
		goto Exit;
	}


	DBGPRINT("%s %d: Check if Blade port type %d for slot %d Needs "
		   "promotion\n",__FUNCTION__, __LINE__, port_type, slot);

	/*
	 * Port type of 0 means its an endpoint. Only way for that
	 * to happen is is this is the NTP.
	 */
	if (port_type == 0) {
		printk("%s %d: Promote NTP to USP\n", __FUNCTION__, __LINE__);

		rc = bhpc_promote_port(p_slot);
		if (rc) {
			printk(KERN_ERR "%s: Cannot promote port for slot %d to upstream "
				   "port\n", __FUNCTION__, slot);
			goto Exit;
		}

#ifndef CONFIG_SILKWORM_DCX
		pciehp_show_sfm_data(slot);
#endif

	} else if (port_type ==  PCI_EXP_TYPE_UPSTREAM) {
		DBGPRINT("%s %d: USP found\n", __FUNCTION__, __LINE__);
	} else {
		printk(KERN_ERR "%s %d: Invalid port type %d found\n",
			   __FUNCTION__, __LINE__, port_type);
		rc = -ENODEV;
		goto Exit;
	}
	/* Tell Linux hot-plug service to scan the blade slot */

	rc = pciehp_configure_device(p_slot);
	if (rc) {
		printk(KERN_ERR "%s: Cannot configure blade in slot %d\n",
				   __FUNCTION__, slot);
		goto Exit;
	}

	/* Find the blade bridge to pass back to the caller */
	*dev = pci_get_slot(ctrl->pcie->port->subordinate, PCI_DEVFN(0, 0));
	if (!*dev) {
		printk(KERN_ERR "%s Cannot find blade bridge for slot %d "
			   "parent port %s\n", __FUNCTION__, slot,
			   pci_name(ctrl->pcie->port)); 
		rc = -1;
	} else {
		int pos;
		u32 regval32 = 0;

		/* Ensure the slot status register is cleared */
		pos = pci_find_capability(ctrl->pcie->port, PCI_CAP_ID_EXP);
		if (pos) {
			pci_read_config_dword(ctrl->pcie->port, pos + PCI_EXP_SLTCTL, &regval32);
			pci_write_config_dword(ctrl->pcie->port, pos + PCI_EXP_SLTCTL, regval32);
		}

		printk(KERN_INFO "%s Returning blade bridge %p %s\n", __FUNCTION__,
			   *dev, pci_name(*dev));

		save_device_state(*dev);

		/*
		 * The call to pci_get_slot() incremented the refcount. So
		 * decrement it here to get it back where it was.
		 */
		pci_dev_put(*dev);

		if (interrupts_enabled) {

			/*
			 * Need to enable error reporting in the devices
			 * that have been found.
			 */
			tdev = bhpc_get_blade_bridge(slot);
			if (tdev) {
				int domain;

				printk(KERN_DEBUG "%s: Before setup blade bridge %p %s rct %d\n",
					   __FUNCTION__,
					   tdev, pci_name(tdev), pci_dev_refcount(tdev));

				domain = pci_domain_nr(tdev->bus);
				bhpc_setup_configs(tdev->bus->self, domain, 1);

				printk(KERN_DEBUG "%s: After setup blade bridge %p %s rct %d\n",
					   __FUNCTION__, tdev, pci_name(tdev),
					   pci_dev_refcount(tdev));

			} else {
				printk(KERN_ERR "%s: No tdev\n", __FUNCTION__);
			}
		}

		/*
		 * Get and save the refrence count of the blade bridge. When
		 * the blade is removed, we will need to make sure the reference
		 * count is the same. Othewise, we can get a memory leak or a
		 * kernel panic
		 */
		refcount = pci_dev_refcount(*dev);
		pr_debug("\n%s %d: SFM USP %s REFCOUNT %d\n\n",
			   __FUNCTION__, __LINE__,
			   pci_name(*dev), refcount);

		bhpc_save_reference_count(slot, refcount);

	}
 Exit:

	if (interrupts_enabled) {
		bhpc_enable_scan_error_reports(*dev);
	}

 Exit2:

	return (rc);
}
/*
 * show_dsp_link()
 *
 * Description:
 *   This function over-rides the probe() method of the Linux hot plug
 *	 service driver, and is used to identify PCIe ports that provide hot plug
 *	 controller services.
 *
 * When Called:
 *   Whenever a PCIe port device is found.
 *
 * Inputs:
 *   None
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */
static int
show_dsp_link(struct pci_dev *pdev)
{
	int rc = 0, pos;
	u32 link_ctrl = ~0, vid = 0, pcie_caps = ~0, slot_ctrl = ~0;

	if (pdev) {
		pci_read_config_dword(pdev, PCI_VENDOR_ID, &vid);
		pos = pci_find_capability(pdev, PCI_CAP_ID_EXP);
		if (pos) {
			pci_read_config_dword(pdev, pos, &pcie_caps);
			pci_read_config_dword(pdev, pos + PCI_EXP_LNKCTL, &link_ctrl);
			pci_read_config_dword(pdev, pos + PCI_EXP_SLTCTL, &slot_ctrl);
		} 
		printk(KERN_ERR "%s: probe dev %s VID 0x%8.8x lnk ctrl 0x%8.8x slt ctrl 0x%8.8x caps 0x%8.8x\n",
			   __FUNCTION__, pci_name(pdev),vid, link_ctrl, slot_ctrl, pcie_caps);

		if (vid == ~0) {
			rc = -1;
		}

	} else {
		printk(KERN_ERR "%s: Called with NULL for input dev\n",
			   __FUNCTION__);
		rc = -1;
	}

	return (rc);

}
static int
bhpc_hpc_probe(struct pcie_device *dev)
{
	int rc = 0,rc2 = 0, port = -1;
	struct controller *ctrl;
	struct hotplug_slot_ops *ops_ptr;

	/*
	 * Let the original Linux HPC probe() method do its work, which
	 * includes creating a new slot controller object. Then we extract
	 * and save the new slot controller object for later use.
	 */
	if (hpc_service_ops.probe) {

		rc = show_dsp_link(dev->port);
		if (rc) {
			goto Exit;
		}

		rc = hpc_service_ops.probe(dev);
		if (rc) {
			printk(KERN_ERR "%s: pcie hotplug service driver probe failed\n",
				   __FUNCTION__);
			goto Exit;
		}

		ctrl = get_service_data(dev);

		ops_ptr = ctrl->slot->hotplug_slot->ops;

		if (!ops_ptr) {
			printk(KERN_ERR "%s: No hotplug slot ops for %s\n", __FUNCTION__,
				   pci_name(dev->port));
			rc = -1;
			goto Exit;
		}

		memset(&saved_hpc_ops, 0, sizeof(struct hotplug_slot_ops));

		if (ops_ptr->get_power_status) {

			dbg("  Overriding get_power_status Existing: 0x%p New: 0x%p\n",
				   ops_ptr->get_power_status, &bhpc_hpc_ops.get_power_status);

			saved_hpc_ops.get_power_status = ops_ptr->get_power_status;
			ops_ptr->get_power_status = bhpc_hpc_ops.get_power_status;
		}
#if 0
		if (hpc_ops->query_power_fault) {

			dbg("  Overriding query_power_fault Existing: 0x%p New: 0x%p\n",
				   ops_ptr->query_power_fault, hpc_ops->query_power_fault);

			saved_hpc_ops.query_power_fault = ops_ptr->query_power_fault;
			ops_ptr->query_power_fault = hpc_ops->query_power_fault;
		}
#endif
		if (ops_ptr->get_adapter_status) {

			dbg("  Overriding get_adapter_status Existing: 0x%p New: 0x%p\n",
				   ops_ptr->get_adapter_status, &bhpc_hpc_ops.get_adapter_status);

			saved_hpc_ops.get_adapter_status = ops_ptr->get_adapter_status;
			ops_ptr->get_adapter_status = bhpc_hpc_ops.get_adapter_status;
		}

		/* Only store devices that are in the MM/CP PCIe switch */
		if(dev->port->bus->number == 2) {
			int domain;
			port = PCI_SLOT(dev->port->devfn);
			domain =  pci_domain_nr(dev->port->bus);
			rc2 = store_slot_controller(ctrl, domain, port);
			if (rc2) {
				DBGPRINT(KERN_ERR "%s: store controller object failed\n",
						 __FUNCTION__);
				goto Exit;
			}
		}
	} else {
		printk(KERN_ERR "%s: NO PCIE HPC PROBE\n", __FUNCTION__);
		rc = -1;
	}

 Exit:
	return (rc);
}
/*
 * bhpc_hpc_remove()
 *
 * Description:
 *   This function will be called when the Linux HPC module is unloaded.
 *   When this occurs, we must not use the methods received from the
 *   Linux HPC anymore.
 *
 *   NOTE: In theory, it should not be possible to remove Linux HPC
 *         module while Brocade HPC module is loaded.
 *
 * Inputs:
 *   None
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

int bhpc_hpc_remove(void)
{
	int rc = 0;

	if (bhpc_initialized) {
		bhpc_plat_remove();
	}

	return (rc);
}

/*
 * bhpc_initialize()
 *
 * Description:
 *   This function will perform the required initialization of the Brocade
 *   hot plug service driver.
 *
 * When Called:
 *   During NOS startup, immediately after Active/Standby MM arbitration
 *   has been completed.
 *
 * Inputs:
 *   None
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

static int bhpc_initialize(void)
{
	int rc;

	/*
	 * Provide over-rides for the original pcie_port_service_driver
	 * methods. Upon return, hpc_service_ops will contain pointers to
	 * the over-ridden methods. This will be used if they need to be
	 * restored later.
	 */
	memcpy(&hpc_service_ops, &bhpc_service_ops,
		   sizeof (struct pcie_port_service_driver));

	DBGPRINT("%s %d: Call pcied_service_init with psd_ops 0x%p\n",
		   __FUNCTION__, __LINE__, &hpc_service_ops);

	rc = pcied_service_init(&hpc_service_ops);
	if (rc) {
		printk(KERN_ERR "%s: pcied service init failed\n", __FUNCTION__);
		goto Exit;
	}

	DBGPRINT("%s %d: Call pci_register_driver\n", __FUNCTION__, __LINE__);

	rc = pci_register_driver(&bhpc_aer_driver);
	if (rc) {
		/* XXX Do we need to restore the over-rides here ??? */
		printk(KERN_ERR "%s: pci register driver failed %d\n", __FUNCTION__, rc);
		goto Exit;
	}

	if (chassis_ptr->enable_irqs) {
		printk(KERN_ERR "%s: Init PCIe IRQ handler\n", __FUNCTION__);
		bhpc_irq_init(chassis_ptr->error_fn, chassis_ptr->dsp_map);
		interrupts_enabled = 1;
		silkworm_machine_check_control(1, 0);
	} else {
		silkworm_machine_check_control(0, 0);
		interrupts_enabled = 0;
		printk(KERN_DEBUG "%s: NOT using PCIe IRQ handler\n", __FUNCTION__);
	}

	bhpc_initialized = 1;

 Exit:

	return (rc);

}
/*
 * bhpc_plat_remove()
 *
 * Description:
 *   This function will called by the driver user when it wishes
 *   to stop using the services of the BHPC driver. It's the
 *   equivalent of close() for a device driver. If bhpc_driver_register()
 *   has been called, then this function must be called before
 *   unloading the module.
 *
 *   When this occurs, we must discard the callbacks received from
 *   platform module and restore the over-ridden Linux HPC methods.
 *
 * Inputs:
 *   None
 *
 * Outputs:
 *   None
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

static int bhpc_plat_remove(void)
{
	int rc = 0;

	DBGPRINT("%s: Entry init %d\n", __FUNCTION__, bhpc_initialized);
	if (bhpc_initialized) {
		printk("  Unregistering\n");
		cleanup_dsps();
		pci_unregister_driver(&bhpc_aer_driver);
		pcied_service_unregister();

		if (chassis_ptr->enable_irqs) {
			bhpc_irq_cleanup();
			interrupts_enabled = 0;
			silkworm_machine_check_control(0, 0);
		}

		bhpc_initialized = 0;
	}

	return (rc);
}


/*
 *  bhpc_driver_register()
 *
 * Description:
 *    This function is used to provide the BHPC a chassis logical slot
 *    map and get back the operations or methods supported by the BHPC
 *    driver. The logical slot map defines the connections between the
 *    chassis slots and MM PCIe switch.
 *
 * When Called:
 *    During COS startup, immediately after Active/Standby MM arbitration
 *    has been completed.
 *
 * Inputs:
 *   platform - Contains chassis slot map, error/log callbacks, and 
 *              hpc_ops callbacks.
 *
 * Outputs:
 *   svc_ops - Passes back the operations supported by the BHPC service driver
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

bhpc_svc_ops *
bhpc_driver_register(bhpc_platform *platform)
{

	if (platform == NULL) {
		printk(KERN_ERR "%s: Null platform pointer\n", __FUNCTION__);
		return (NULL);
	}
	memcpy(&chassis, platform, sizeof (bhpc_platform));
	memcpy(&slot_map_table, platform->slot_map, sizeof(bhpc_slot_map_table));
	chassis.slot_map = &slot_map_table;

	if (platform->plat_ops) {
		plat_ops = chassis_ptr->plat_ops;
	} else {
		printk(KERN_ERR "%s: Plat_ops ptr is NULL\n", __FUNCTION__);
		return (NULL);
	}
	
	bhpc_initialized = 1;

	return (&svc_ops);

}
EXPORT_SYMBOL(bhpc_driver_register);

/*
 * Create /proc file system entry to dump slot map table
 *
 *
 */

static struct proc_dir_entry *pciehp_bhpc;
static struct proc_dir_entry *slot_table;

static void
proc_init(void)
{

	pciehp_bhpc = proc_mkdir("pciehp_bhpc", NULL);
	if (pciehp_bhpc == NULL) {
		printk(KERN_ERR "%s: Could not create proc dir pciehp_bhpc\n",
			   __FUNCTION__);
		return;
	}

	slot_table = create_proc_entry("slot_table", 0, pciehp_bhpc);
	if (!slot_table) {
		printk(KERN_ERR "%s: Could not create proc file"
			   "pciehp_bhpc/slot_table \n", __FUNCTION__);

		if (pciehp_bhpc) {
			remove_proc_entry("pciehp_bhpc", NULL);
		}

		return;
	}

	slot_table->write_proc = &dump_slot_table_map;

}

/*
 *
 */
static void
proc_cleanup(void)
{

	if (slot_table) {
		remove_proc_entry("slot_table", pciehp_bhpc);
	}

	if (pciehp_bhpc) {
		remove_proc_entry("pciehp_bhpc", NULL);
	}
}
/*
 *  bhpc_module_init()
 *
 * Description:
 *   This is the module init function. There's nothing to do at this
 *   time except indicate the module is alive.
 *
 * When Called:
 *    During Linux initialization
 *
 * Inputs:
 *   N/A
 *
 * Outputs:
 *   N/A
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

int __init bhpc_module_init(void)
{

	printk("Silkworm PCIe Hotplug Subsystem loaded\n");
	memset(&chassis, 0, sizeof (bhpc_platform));
	memset(&slot_map_table, 0, sizeof (bhpc_slot_map_table));
	plat_ops = chassis_ptr->plat_ops;

	proc_init();

	return (0);
}
/*
 *  bhpc_module_exit()
 *
 * Description:
 *   This is the module unload function. If the remove() method
 *   needs to be called but hasn't been, then do it now. Othewise,
 *   just announce the unload occured.
 *
 * When Called:
 *    When the module is unloaded
 *
 * Inputs:
 *   N/A
 *
 * Outputs:
 *   N/A
 *
 * Returns:
 *   0 - Success
 *  <0 - Failure
 */

void __exit bhpc_module_exit(void)
{

	if (bhpc_initialized) {
		bhpc_plat_remove();
		bhpc_initialized = 0;
	}
	proc_cleanup();

	printk("Silkworm PCIe Hotplug Subsystem unloaded\n");
}

module_init(bhpc_module_init);
module_exit(bhpc_module_exit);
