/*
 *
 *    Copyright (c) 1996-2006 Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 *    Description:
 *      M41T11 RTC device driver.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>
#include <linux/i2c.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <asm/time.h>
#include <asm/m41t11.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#if defined(CONFIG_SILKWORM_RTC_FIXUP)
#include <brocade/cpld.h>
#include <brocade/silkworm.h>
#include <brocade/platform.h>
#include <linux/delay.h>
#endif

#define I2C_DRIVERID_M41T11     48     /* real-time-clock               */
#define I2C_M41T11_ADDRESS      0x68     

#if defined(CONFIG_SILKWORM) && defined(CONFIG_40x)
#include <brocade/cpld.h>
/* These macros are hooks to perform the I2C bus multiplexing
 */
#define SELECT_I2C()		\
		{ int ret = 0; \
			if ((ret = cpld_set_i2c_mux_local()) != 0) \
				return ret; \
		}
#define RELEASE_I2C()		cpld_release_i2c_mux(0)
#else
#define SELECT_I2C()		{}
#define RELEASE_I2C()		{}
#endif

/* Build time exclusion */
#if defined(CONFIG_SILKWORM_RTC_FIXUP)
#include <brocade/silkworm_rtc_fixup.h>
#define RTC_SELECT()        \
        { int ret = 0; \
            if ((ret = silkwork_set_rtc_access()) != 0) \
                return ret; \
        }
#define RTC_RELEASE()  silkwork_reset_rtc_access()
#define RTC_ON_IIC1()    silkworm_get_iic_bus_rtc()
#else
#define RTC_SELECT()       {}
#define RTC_RELEASE()      {}
#endif

/* Dallas DS1307 clock/calendar register numbers */

#define RTC_SECONDS 0
#define RTC_MINUTES 1
#define RTC_HOURS 2
#define RTC_WEEKDAY 3
#define RTC_DAY_OF_MONTH 4
#define RTC_MONTH 5
#define RTC_YEAR 6
#define RTC_CONTROL 7

#define	M41T11_SECOND_OFF	0x0
#define	M41T11_MINUTE_OFF	0x1
#define	M41T11_HOUR_OFF		0x2
#define	M41T11_DAY_OFF		0x3
#define	M41T11_DATE_OFF		0x4
#define	M41T11_MONTH_OFF	0x5
#define	M41T11_YEAR_OFF		0x6

#define	M41T11_SECOND_MASK	0x7F
#define	M41T11_MINUTE_MASK	0x7F
#define	M41T11_HOUR_MASK	0x3F
#define	M41T11_DAY_MASK		0x07
#define	M41T11_DATE_MASK	0x3F
#define	M41T11_MONTH_MASK	0x1F
#define	M41T11_YEAR_MASK	0xFF

#define RTC_VERSION		"1.1"
#define MAX_I2C_BUFFER_SIZE M41T11_MAX_RTC_REGISTERS + 2

/*
 * See M41T11 Data sheet for format of the clock.
 */
#define hextodec(x)	((((x) & 0xF0) >> 4) * 10 + ((x) & 0x0F))
#define dectohex(x)	((((x) / 10) << 4) + ((x) % 10))

#ifndef elemof
#define elemof(x)	(sizeof(x) / sizeof((x)[0]))
#endif


static struct i2c_driver rtc_i2c_driver;
static struct i2c_client client_template;
static struct i2c_client *m41t11_client;

static void rtc_i2c_init_client(struct i2c_client *pclient);
static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);
static int i2c_get_rtc_time(struct rtc_time *rtc_tm);
static int i2c_set_rtc_time(struct rtc_time *rtc_tm);
static int rtc_read_proc(char *page, char **start, off_t off,
		int count, int *eof, void *data);
static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos);

static struct file_operations rtc_fops = {
	owner:	THIS_MODULE,
	read:	rtc_read,
	ioctl:	rtc_ioctl,
};

static struct miscdevice rtc_dev=
{
	RTC_MINOR,
	"rtc0",
	&rtc_fops
};

struct m41t11_data {
	u8 features;
	struct rtc_device *rtc;
};

int i2c_get_rtc_registers(int address, u_char *data, int qty);
int i2c_set_rtc_registers(int address, u_char *data, int qty);

static void rtc_i2c_init_client(struct i2c_client *pclient)
{
	u_char reg;

	/* DS1307 Datasheet says the oscillator may not be enabled on the first
	 * powerup.  Enable it if necessary.
	 */
	if((i2c_get_rtc_registers(RTC_SECONDS, &reg, 1) == 0) && (reg & 0x80))
	{
		reg &= ~0x80;
		i2c_set_rtc_registers(RTC_SECONDS, &reg, 1);
	}

	/* Enable 4KHz output if not already enabled. 
	 */
	if((i2c_get_rtc_registers(RTC_CONTROL, &reg, 1) == 0) && 
	   (reg & 0x93) != 0x91)
	{
		reg &= ~0x93;
		reg |= 0x91;
		i2c_set_rtc_registers(RTC_CONTROL, &reg, 1);
	}
}

/* We get here after scanning the I2C device at one of the MT41T11 addresses.
 */
static int rtc_i2c_detect_rtc(struct i2c_adapter *adap, int addr, int kind)
{
	char c = 0;
	int found = 0;
	struct i2c_client *pclient;

	if (!(pclient = (struct i2c_client *)kmalloc(sizeof(struct i2c_client),
					GFP_KERNEL)))
	{
		printk(KERN_ERR "m41t11: Driver failed memory allocation\n");
		return -ENOMEM;
	}
	memcpy(pclient, &client_template, sizeof(struct i2c_client));
	pclient->addr = addr;
	pclient->adapter = adap;

	if(i2c_master_send(pclient, &c, 1) == 1)
	{
		if(i2c_master_recv(pclient, &c, 1) == 1)
		{
			printk(KERN_ERR "m41t11: I2C Real-Time-Clock detected on iic%d addr 0x%x\n", 
				i2c_adapter_id(adap), pclient->addr);
			found = 1;
		}
	}

	if(found)
	{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
		if((err = i2c_attach_client(pclient)))
		{
			printk(KERN_ERR "m41t11: failed to attach client\n");
			kfree(pclient);
			return err;
		}
#endif
		m41t11_client = pclient;	 /* hack! */
	}
	else
	{
		kfree(pclient);
		return -ENODEV;
	}
	
	return 0;
}

static int remove_client(struct i2c_client *client)
{
	struct m41t11_data *clientdata = i2c_get_clientdata(client);
	struct rtc_device *rtc = clientdata->rtc;

	if (rtc)
		rtc_device_unregister(rtc);
	kfree(clientdata);

	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
static int i2c_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	return -1;
}
#endif

static int m41t11_get_rtc_time(struct device *dev, struct rtc_time *tm)
{
	return i2c_get_rtc_time(tm);
}

static int m41t11_set_rtc_time(struct device *dev, struct rtc_time *tm)
{
	return i2c_set_rtc_time(tm);
}

static int m41t11_rtc_proc(struct device *dev, struct seq_file *seq)
{
	int i, ret;
	u_char tmp[MAX_I2C_BUFFER_SIZE];

	ret = i2c_get_rtc_registers(0, tmp, MAX_I2C_BUFFER_SIZE);
	if (ret < 0) {
		seq_printf(seq, "M41T11 driver: i2c_get_rtc_registers() failed\n");
		return (ret);
	}

	seq_printf(seq, "M41T11 driver\n");
	for (i = 0; i < M41T11_MAX_RTC_REGISTERS; i++) {
		seq_printf(seq, "Register: %d  Data: 0x%2.2x\n", i, tmp[i]);
	}

	return 0;
}

static struct rtc_class_ops m41t11_rtc_ops = {
	.read_time = m41t11_get_rtc_time,
	.set_time = m41t11_set_rtc_time,
	.proc = m41t11_rtc_proc,
};

/*
 * XXX: temporarily disable the check of the IIC index as this
 *      field now is private in i2c-ibm_iic.h
 */

static int attach_adapter(struct i2c_adapter *adap)
{
	int ret = -1, status = 0;
	struct rtc_device *rtc = NULL;

	/*
	 * When probing, we do arbitration and mux selection
	 * only on systems where there is no private I2C bus.
	 * That's limited to 405-based systems, so far.
	 *
	 * Also, do NOT probe on any bus except 0. There should be
	 * no reason to look for devices on the system I2C bus
	 * this early in startup.
	 */

	if (adap && adap->nr == 0) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
		ret = i2c_probe(adap, &addr_data, rtc_i2c_detect_rtc);
#else
		ret = rtc_i2c_detect_rtc(adap, I2C_M41T11_ADDRESS, 0); 
		if(ret) {
			printk(KERN_ERR "%s: RTC:: Probe  bus IIC-%d:"
				" device not found\n", __FUNCTION__, adap->nr);
		} else {
			printk(KERN_DEBUG "%s: RTC:: Probe  bus IIC-%d OK\n",
				__FUNCTION__, adap->nr);
		}
#endif
	} 
#if defined(CONFIG_SILKWORM_RTC_FIXUP)
	else if ((adap) && (adap->nr == 1) && (RTC_ON_IIC1())) {
		printk("m41t11: Called to probe for bus IIC-%d\n", adap->nr);

		RTC_SELECT();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
		ret = i2c_probe(adap, &addr_data, rtc_i2c_detect_rtc);
#else
		ret = rtc_i2c_detect_rtc(adap, I2C_M41T11_ADDRESS, 0);
#endif
		if(ret) {
			printk(KERN_ERR "%s: RTC:: Probe  bus IIC-%d:"
				" device not found\n", __FUNCTION__, adap->nr);
		} else {
			printk(KERN_DEBUG "%s: RTC:: Probe  bus IIC-%d OK\n",
				__FUNCTION__, adap->nr);
		}

		RTC_RELEASE();
	}
#endif /* CONFIG_SILKWORM_RTC_FIXUP */

	if (ret == 0) { /* Device found */
		printk(KERN_DEBUG "%s: RTC:: misc_register() bus IIC-%d\n",
				__FUNCTION__, adap->nr);
		if((status = misc_register(&rtc_dev)))
		{	
			printk(KERN_ERR "mt41t11: can't register device.\n");
			return status;
		}

		rtc = rtc_device_register(adap->name, &adap->dev,
				  &m41t11_rtc_ops, THIS_MODULE);
		if (IS_ERR(rtc)) {
			ret = PTR_ERR(rtc);
			rtc = NULL;
		}
	}

	return ret;
}

/*
 * ssize_t rtc_read()
 *
 * Description:
 *   This function normally implements a blocking call
 *   to wait for an interrupt from the RTC indicating
 *   the second counter has rolled over.
 *
 *   The RTC used in Brocade Linux-based products
 *   does not implement such an interrupt, so we
 *   just immediately return.
 *
 * Input(s):
 *   file  - file descriptor
 *   buf   - users buffer to return read data
 *   count - size of buf in bytes
 *   ppos  -
 *
 * Output(s):
 *   0L, in the users buffer
 *
 * Returns:
 *   -EFAULT if an error
 *   Number of bytes (4) copied to the users buffer if OK.
 *
 */
static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos)
{
	unsigned long data;
	ssize_t retval;

	/*
	 * We don't really support interrupts, but we need
	 * to make the user think we do to prevent busy-waiting
	 * on I2C bus reads of the RTC. That's much too slow.
	 */

	data = 0;
	retval = put_user(data, (unsigned long *)buf);

	if (!retval)
	  retval = sizeof (unsigned long);

	return (retval);

}

/*
 * u_char convert_hr()
 *
 * Description:
 *   Converts the hour value from 12 hour format
 *   to 24 hour format, if needed. This is for the
 *   DS1305 RTC device, and should not be used with
 *   the M41T11 RTC device.
 *
 *   If bit 6 is a 1, the value in hr is 24 hour format.
 *   If bit 6 is a 0, the value in hr is 12 hour format,
 *   and bit 5 = 1 indicates PM.
 *
 * Input(s):
 *   hr  - BCD hour value from the RTC
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   The appropriate conversion of the
 *   value in hr.
 *
 */
u_char
convert_hr(u_char hr)
{
	u_char	ret = 0;

	/*
	 * If bit 6 is 0, the value is either already
	 * in 24 hour format, or the RTC does not have
	 * this mode. In either case, return the existing
	 * value
	 */

	if (!(hr & 0x40))
		return (hr);
	else {
		if (hr & 0x020) {
			ret = (hr & 0x10) * 10 + (hr & 0xf) + 12;
			return (ret);
		} else {
			ret = (hr & 0x10) * 10 + (hr & 0xf);
			return (ret);
		}
	}
}

/*
 * int i2c_get_rtc_registers()
 *
 * Description:
 *   Read contents of a specified  RTC register
 *   via the I2C device driver.
 *
 *   Before the I2C driver is called, exclusive
 *   access to the I2C mux must be granted.
 *   Before returning, access to the I2C address
 *   mux MUST be relenquished.
 *
 * Input(s):
 *   address - Register address to read
 *   data    - Pointer to hold register data
 *
 * Output(s):
 *   Contents of the specified RTC register
 *
 * Returns:
 *  -1 for fail
 *   0 for pass
 *
 */
int
i2c_get_rtc_registers(int address, u_char *data, int qty)
{
	int    ret;
	struct i2c_client	*client = m41t11_client;
	struct i2c_msg msgs[2];
	u_char addr = address;
	int i;

	if (!data || !client)
	  return (-1);

	/* first build the transmit message */
	msgs[0].addr 	= client->addr;
	msgs[0].len 	= 1;
	msgs[0].flags	= 0;
	msgs[0].buf	= &addr;

	/* next build the receive message */
	msgs[1].addr	= client->addr;
	msgs[1].len	= qty;
	msgs[1].flags	= I2C_M_RD;
	msgs[1].buf	= data;

	SELECT_I2C();
	RTC_SELECT();

	for (i = 0; i < 3; i++) {
		ret = i2c_transfer(client->adapter, msgs, elemof(msgs));
		if (ret >= 0)
			break;
		printk(KERN_INFO "RTC: i2c_transfer() NAK. Retrying %d, ret = %d\n", i+1, ret);
		msleep(350);
	}

	RTC_RELEASE();
	RELEASE_I2C();

	if (ret < 0) {
		printk(KERN_ERR "RTC: i2c_transfer() failed with error code %d.\n", ret);
		return (-1);
	}



#ifdef DEBUG
	printk("RTC register read: %d  Data: 0x%x\n", address, data[0]);
#endif

	return (0);
}


/*
 * int i2c_set_rtc_registers
 *
 * Description:
 *   Program RTC registers via the I2C device
 *   driver.
 *
 *   Before the I2C driver is called, exclusive
 *   access to the I2C address mux must be granted.
 *   Before returning, access to the I2C address
 *   mux MUST be relenquished.
 *
 * Input(s):
 *   address  - Register address to write
 *   *data    - Pointer to data to be written
 *   qty      - Number of bytes of data to be written
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *  -1 for fail
 *   0 for pass
 *
 */
int
i2c_set_rtc_registers(int address, u_char *data, int qty)
{
	int	ret, i;
	u_char	buf[2];
	struct i2c_client	*client = m41t11_client;
	int retry = 3;

	if (!data || !client)
	  return (-1);

#ifdef DEBUG
	printk("RTC register Address: %d  Data: 0x%x\n", address, *data);
#endif

	/* Write clock entries one by one */
	for (i = 0; i < qty; i++) {
		buf[0] = address + i; /* Set to address of each register */
		buf[1] = data[i];
		SELECT_I2C();
		RTC_SELECT();
		ret = i2c_master_send(client, buf, sizeof (buf));
		RTC_RELEASE()
		RELEASE_I2C();
		if (ret <= 0) {
			retry--;
			printk(KERN_INFO "RTC: i2c_master_send() NAK while "
				   "writing RTC registers. return = %d, i =%d retry = %d\n", ret, i, 3-retry);
			i--;
			if (retry <= 0) {
				printk(KERN_ERR "RTC: i2c_master_send() failed after 3 attempts\n");
				return (-1);
				break;
			}
			msleep(350);
		}
	}

	return (0);
}

/*
 * int i2c_get_rtc_time()
 *
 * Description:
 *   Read contents of the I2C RTC via the I2C device
 *   driver.
 *
 *   Before the I2C driver is called, exclusive
 *   access to the I2C mux must be granted.
 *   Before returning, access to the I2C address
 *   mux MUST be relenquished.
 *
 * Input(s):
 *   rtc_time - Pointer to struct to return the
 *              RTC register contents
 * Output(s):
 *   Contents of the RTC registers
 *
 * Returns:
 *  -1 for fail
 *   0 for pass
 *
 */
static
int
i2c_get_rtc_time(struct rtc_time *rtc_time)
{
	int	ret;
	u_char	tmp[MAX_I2C_BUFFER_SIZE];

	ret = i2c_get_rtc_registers(0, tmp, 7);
	if (ret < 0)
	  return (ret);

	rtc_time->tm_sec  = hextodec(tmp[M41T11_SECOND_OFF]
				     & M41T11_SECOND_MASK);
	rtc_time->tm_min  = hextodec(tmp[M41T11_MINUTE_OFF]
				     & M41T11_MINUTE_MASK);
	rtc_time->tm_hour = hextodec(tmp[M41T11_HOUR_OFF]
				     & M41T11_HOUR_MASK);
	rtc_time->tm_wday = hextodec(tmp[M41T11_DAY_OFF]
				     & M41T11_DAY_MASK) - 1;
	rtc_time->tm_mday = hextodec(tmp[M41T11_DATE_OFF]
				     & M41T11_DATE_MASK);
	rtc_time->tm_mon  = hextodec(tmp[M41T11_MONTH_OFF]
				     & M41T11_MONTH_MASK) - 1;
	rtc_time->tm_year = hextodec(tmp[M41T11_YEAR_OFF]
				     & M41T11_YEAR_MASK);

	/*
	 * Make some adjustments to match what struct
	 * rtc_time expects
	 */

	rtc_time->tm_year += 100; /* Change from 20th century to 21st */

#ifdef DEBUG
	printk("Reading from i2c_get_rtc_time()\n");
	printk("sec = %d\n", rtc_time->tm_sec);
	printk("min = %d\n", rtc_time->tm_min);
	printk("hour = %d (0x%x)\n", rtc_time->tm_hour, tmp[2]);
	printk("mday = %d\n", rtc_time->tm_mday);
	printk("wday = %d\n", rtc_time->tm_wday);
	printk("mon = %d\n", rtc_time->tm_mon);
	printk("year = %d\n", rtc_time->tm_year);
#endif
	return (0);
}

/*
 * int i2c_set_rtc_time()
 *
 * Description:
 *   Program the I2C RTC via the I2C device
 *   driver.
 *
 *   Before the I2C driver is called, exclusive
 *   access to the I2C address mux must be granted.
 *   Before returning, access to the I2C address
 *   mux MUST be relenquished.
 *
 * Input(s):
 *   rtc_time - Pointer to struct containing the
 *              "broken-down" time to program to the
 *              RTC device
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *  -1 for fail
 *   0 for pass
 *
 */
int
i2c_set_rtc_time(struct rtc_time *rtc_time)
{
	int	ret;
	u_char	tmp[MAX_I2C_BUFFER_SIZE];

#ifdef DEBUG
	printk("sec = %d\n", rtc_time->tm_sec);
	printk("min = %d\n", rtc_time->tm_min);
	printk("hour = %d\n", rtc_time->tm_hour);
	printk("wday = %d\n", rtc_time->tm_wday);
	printk("day = %d\n", rtc_time->tm_mday);
	printk("mon = %d\n", rtc_time->tm_mon);
	printk("year = %d\n", rtc_time->tm_year);
#endif
	tmp[M41T11_SECOND_OFF] = dectohex((u_char)rtc_time->tm_sec);
	tmp[M41T11_MINUTE_OFF] = dectohex((u_char)rtc_time->tm_min);
	tmp[M41T11_HOUR_OFF]   = dectohex((u_char)rtc_time->tm_hour);
	tmp[M41T11_DAY_OFF]    = dectohex(((u_char)rtc_time->tm_wday + 1));
	tmp[M41T11_DATE_OFF]   = dectohex((u_char)rtc_time->tm_mday);
	tmp[M41T11_MONTH_OFF]  = dectohex(((u_char)rtc_time->tm_mon + 1));
	if (rtc_time->tm_year > 100)
	  tmp[M41T11_YEAR_OFF] = dectohex((u_char)(rtc_time->tm_year - 100));

	ret = i2c_set_rtc_registers(0, tmp, 7);

	return (ret);
}

/*
 * int rtc_ioctl()
 *
 * Description:
 *   Implements the IOCTL interface for the driver.
 *   IOCTLs supported are:
 *    RTC_UIE_OFF  - Enable  RTC update interrupts
 *    RTC_UIE_ON   - Disable RTC update interrupts
 *    RTC_RD_TIME  - Read the RTC time
 *    RTC_SET_TIME - Set the RTC time
 *    RTC_FT_ON    - Enable the RTC FT bit
 *    RTC_FT_OFF   - Disable the RTC FT bit
 *    RTC_FT_RD    - Get the status of the RTC FT bit
 *
 * The FT bit controls the 512Hz clock signal for the PCI
 * deadman timer. If the bit is a 1, the clock is turned on.
 * If the bit is a 0, the clock is turned off.
 *
 * Input(s):
 *   inode  - Pointer to the inode descriptor for the device
 *   file   - Pointer to an open file descriptor for the device
 *   cmd    - The IOCTL command
 *   arg    - User-space data
 *
 * Output(s):
 *   For RTC_RD_TIME, the contents of the RTC registers.
 *   For RTC_FT_RD, 0 = bit not set, 1 = bit is set
 *
 * Returns:
 *    0 - Success
 *   -EFAULT - Fail
 *
 */

static int
rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
	struct rtc_time wtime;
	u_char data;
	struct i2c_client *client = m41t11_client;

	if(client == NULL)
		return -EINVAL;

	switch (cmd) {

	case RTC_UIE_OFF:
	  {
	    /*
	     * We don't really have update interrupts, but
	     * we don't want users to do busy-waits for the
	     * time to change. That causes many reads via
	     * the I2C bus, and we need to minimize that
	     * activity on Ulysses
	     */

	    return (0);
	  }

	case RTC_UIE_ON:
	  {
	    /*
	     * We don't really have update interrupts, but
	     * we don't want users to do busy-waits for the
	     * time to change. That causes many reads via
	     * the I2C bus, and we need to minimize that
	     * activity on Ulysses
	     */
	    return (0);
	  }

	case RTC_RD_TIME:	/* Read the time/date from RTC	*/
	  {
	    if (i2c_get_rtc_time(&wtime) != 0)
	      return (-EFAULT);

	    break;
	  }

	case RTC_SET_TIME:	/* Set the RTC */
	  {
	    struct rtc_time rtc_tm;

	    if (!capable(CAP_SYS_TIME))
	      return (-EACCES);

	    if (copy_from_user(&rtc_tm, (struct rtc_time*)arg,
			       sizeof (struct rtc_time)))
	      return (-EFAULT);

	    if (i2c_set_rtc_time(&rtc_tm) != 0)
	      return (-EFAULT);

	    return (0);
	  }

	case RTC_FT_RD :

	  if (i2c_get_rtc_registers(M41T11_CMD_OFFSET, &data, 1) != 0)
	    return (-EFAULT);

	  data =  (data & M41T11_FT_BIT)? 1 : 0;

	  return (copy_to_user((void *)arg, &data,
			       sizeof (data)) ? -EFAULT : 0);


	case RTC_FT_ON :

	  if (i2c_get_rtc_registers(M41T11_CMD_OFFSET, &data, 1) != 0)
	    return (-EFAULT);

	  data |= M41T11_FT_BIT;

	  if (i2c_set_rtc_registers(M41T11_CMD_OFFSET, &data, 1) != 0)
	    return (-EFAULT);

	  return (0);

	case RTC_FT_OFF:

	  if (i2c_get_rtc_registers(M41T11_CMD_OFFSET, &data, 1) != 0)
	    return (-EFAULT);

	  data &= ~(M41T11_FT_BIT);

	  if (i2c_set_rtc_registers(M41T11_CMD_OFFSET, &data, 1) != 0)
	    return (-EFAULT);

	  return (0);

	case RTC_REG_SET:
	  {
	    long addr_data;
	    int addr;

	    if (copy_from_user(&addr_data, (long *)arg,
			       sizeof (long)))
	      return (-EFAULT);

	    addr = (addr_data >> 16) & 0xFF;
	    data = addr_data & 0xff;

	    if (i2c_set_rtc_registers(addr, &data, 1) != 0)
	      return (-EFAULT);

	    return (0);

	  }

	case RTC_REG_GET:
	  {

	    long addr_data;
	    int addr;

	    if (copy_from_user(&addr_data, (long *)arg,
			       sizeof (long)))
	      return (-EFAULT);

	    addr = (addr_data >> 16) & 0xFF;

	    if (i2c_get_rtc_registers(addr, &data, 1) != 0)
	      return (-EFAULT);

	    addr_data |= data;

	    return (copy_to_user((void *)arg, &addr_data,
				 sizeof (addr_data)) ? -EFAULT : 0);


	  }


	default:
	  return (-EINVAL);
	}

	return (copy_to_user((void *)arg, &wtime,
			     sizeof (wtime)) ? -EFAULT : 0);
}

static struct i2c_driver rtc_i2c_driver =
{
	.driver = {
		.name = "M41T11 and DS1307 RTC",
	},
	.attach_adapter = attach_adapter,
	.remove = remove_client,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,34)
	.id	= I2C_DRIVERID_M41T11,
	.command = i2c_command
#endif
};

static struct i2c_client client_template = 
{
	name:		"M41T11 I2C",
	id:		I2C_DRIVERID_M41T11,
	flags:		0,
	addr:		0,
	adapter:	NULL,
	driver:		&rtc_i2c_driver
};

/*
 * int __init rtc_init()
 *
 * Description:
 *   Initialzation function for the m41t11 device driver
 *
 * Input(s):
 *   N/A
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   0 = Success
 *   NonZero = Fail
 *
 */
static int __init
rtc_init(void)
{
	int status;
	int i;

	m41t11_client = NULL;
	printk(KERN_INFO "M41T11 Real-time-clock Driver v" RTC_VERSION "\n");

	for (i = 0; i < 3; i++) {
		if ((status = i2c_add_driver(&rtc_i2c_driver))) {
			printk(KERN_ERR "mt41t11: I2C driver registration failed. Retrying %d\n", i+1);
			msleep(350);
		} else
			break;
	}

	if(m41t11_client)
		rtc_i2c_init_client(m41t11_client);

	return (status);
}

/*
 * void __exit rtc_exit()
 *
 * Description
 *   Function to remove the device driver. Removes
 *   the proc file system entry and un-registers
 *   the file ops.
 *
 * Input(s):
 *   N/A
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   N/A
 *
 */
static void __exit
rtc_exit(void)
{
	misc_deregister(&rtc_dev);

	i2c_del_driver(&rtc_i2c_driver);
}

module_init(rtc_init);
module_exit(rtc_exit);

EXPORT_SYMBOL(i2c_get_rtc_registers);
EXPORT_SYMBOL(i2c_set_rtc_registers);

MODULE_DESCRIPTION("M41T11 and DS1307 RTC");
