/*
 * Watchdog timer for PowerPC Book-E systems
 *
 * Author: Matthew McClintock
 * Maintainer: Kumar Gala <galak@kernel.crashing.org>
 *
 * Copyright 2005, 2008 Freescale Semiconductor Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/smp.h>
#include <linux/miscdevice.h>
#include <linux/notifier.h>
#include <linux/watchdog.h>
#include <linux/uaccess.h>

#include <asm/reg_booke.h>
#include <asm/system.h>
#ifdef CONFIG_SILKWORM
#include <asm/machdep.h>
#include <brocade/platform.h>
#include <brocade/swbd.h>
#include <brocade/cpld.h>
extern sw_bd silkworm;
#endif

/* If the kernel parameter wdt=1, the watchdog will be enabled at boot.
 * Also, the wdt_period sets the watchdog timer period timeout.
 * For E500 cpus the wdt_period sets which bit changing from 0->1 will
 * trigger a watchog timeout. This watchdog timeout will occur 3 times, the
 * first time nothing will happen, the second time a watchdog exception will
 * occur, and the final time the board will reset.
 */

#ifdef	CONFIG_FSL_BOOKE
#ifdef CONFIG_SILKWORM
#ifdef CONFIG_PPC_FSL_BOOK3E 
#define WDT_PERIOD_DEFAULT 38	/* Temporary safe value for T1022 - Steve Lee */
#else
#define WDT_PERIOD_DEFAULT 23	/* timeout 83ms so around 16ms hit panic - ming */
#endif
#else
#define WDT_PERIOD_DEFAULT 63	/* Ex. wdt_period=28 bus=333Mhz,reset=~40sec */
#endif
#else
#define WDT_PERIOD_DEFAULT 3	/* Refer to the PPC40x and PPC4xx manuals */
#endif				/* for timing information */

u32 booke_wdt_enabled;
u32 booke_wdt_period = WDT_PERIOD_DEFAULT;

#ifdef	CONFIG_FSL_BOOKE
// check AN2804 "Watchdog Timer for e500"
#define WDTP(x)		((((63+1-x)&0x3)<<30)|(((63+1-x)&0x3c)<<15))
#define WDTP_MASK	(WDTP(0))
#else
#define WDTP(x)		(TCR_WP(x))
#define WDTP_MASK	(TCR_WP_MASK)
#endif

#define DEBUG_FPGA_WDT 0 /* for debugging system FPGA WDT */

static DEFINE_SPINLOCK(booke_wdt_lock);

static void __booke_wdt_ping(void *data)
{
	mtspr(SPRN_TSR, TSR_ENW|TSR_WIS);
}

static void booke_wdt_ping(void)
{
	on_each_cpu(__booke_wdt_ping, NULL, 0);
}

#ifdef CONFIG_SILKWORM
/*
 * BROCADE implementation notes:
 *
 * We always set the h/w watchdog at a known 83.5 millisec value.
 * Which allows us top catch interrupts locked overflow in 168 millisec.
 * To ensure that user space got a chance to run once in 30 sec we implement
 * A softwatchdog which is based on jiffies or HZ. A global count is decremented
 * every tick and the h/w is tickled every HZ. If the user space has not sent
 * a single keep-alive in last sw_wdt_period then siftware watchdog triggers.
 */

#ifdef CONFIG_WATCHDOG_INTERRUPT
#ifdef  DEBUG_WDT
#define REFRESHES_PER_SEC	(1)	/* refreshed every 2^29 clocks */
#else 
#define REFRESHES_PER_SEC	(3)	/* refreshed every 2^25 clocks */
#endif /* DEBUG_WDT */
#else
#define REFRESHES_PER_SEC	(HZ)	/* refreshed each timer tick */
#endif	/* CONFIG_WATCHDOG_INTERRUPT */

int sw_wdt_enable;
unsigned long sw_wdt_period;
atomic_t sw_wdt_heartbeat_count[8];
int sw_wdt_default = 1;

/* The user space wtd daemon passes value in micorsec */
#define SW_WDT_DEFAULT_PERIOD	120	/* system default 2 minutes */
#define	MAXONEHOUR		3600UL	/* Max timeout period 60 minutes */
#define MICROSECBASE		1000000

static inline void
sw_update_timer(void)
{
	int i;
	for (i = 0; i < 8; i++)
		atomic_set(&sw_wdt_heartbeat_count[i], sw_wdt_period);
}

#ifdef CONFIG_SILKWORM
/*
 * mm_disable_wdt - display the wdt count and disable the boot wdt in MM
 *
 * The value of FPGA version register is viewed as a set of
 * eight 4b nibbles, ABCDEFGH - [31:28,27:24,23:20,19:16,15:12,11:8,7:4,3:0]
 * The watchdog timer is disabled by writing the value: FBEHGACD,
 * followed by DAFGEBHC to the key sequence register
 */
static void
mm_disable_wdt(void)
{
    unsigned char regval8 = 0;
    unsigned int fpga_ver = 0;
    unsigned int regval = 0;

    /* print the wdt counter value - Mercury is the default */
    switch (silkworm.board) {
	case SWBD127:
	    cpld_read_register32(STARLIFTER4040_BOOT_WDT_CTR, &regval);
	    break;
	case SWBD131:
	case SWBD137:
	case SWBD138:
	case SWBD151:
	case SWBD153:
	case SWBD154:
	case SWBD155:        
		cpld_read_register32(CASTOR_BOOT_WDT_CTR, &regval);
		break;
	case SWBD141:
		cpld_read_register(DCX_BOOT_WDT_CTR, &regval8);
		regval = regval8;
		/* Remove once DCX FPGA working */
		printk (KERN_CRIT "Yoda u-boot wdt counter: %d.  FPGA under construction, exiting...\n", regval8);
		return;
		break;
	case SWBD148:
		cpld_read_register(SKYBOLT_BOOT_WDT_CTR, &regval8);
		regval = regval8;
		/* Remove once Skybolt FPGA working */
		printk (KERN_CRIT "Skybolt u-boot wdt counter: %d.  FPGA under construction, exiting...\n", regval8);
		return;
		break;
	case SWBD161:
	case SWBD162:
		cpld_read_register(WEDGE_BOOT_WDT_CTR, &regval8);
		regval = regval8;
		break;
	case SWBD156:
	case SWBD157:
	case SWBD158:
	case SWBD167: /* Shenlong */	
		cpld_read_register32(CARBON_CPLD_BOOT_WDT_CTR, &regval);
		break;
	default:
	    cpld_read_register32(MM_BOOT_WDT_CTR, &regval);
	    break;
    }
    printk (KERN_CRIT "Uboot wdt counter value: %u\n", regval);

    /* Disable the wdt */
    switch (silkworm.board) {
	case SWBD148:
	    /* TR000527235: fix for skybolt */
	    /*
	     * Offset       :   03 02 01 00
	     * FPGA version :   AB CD EF GH
	     * 
	     * For skybolt, FPGA version will be used differently to get
	     * the seed value (ABCDEFGH)
	     *
	     * 00: GH ( 1 Byte )
	     * 01: F ( 1 nibble )
	     * 02: E ( 1 nibble )
	     * 03: D ( 1 nibble )
	     * The seed values should be 0x000DEFGH.
	     */
	    cpld_read_register(MM_FPGA_VERSION, &regval8);
	    fpga_ver = regval8;
	    cpld_read_register(MM_FPGA_VERSION + 1, &regval8);
	    fpga_ver |= ((regval8 & 0xf) << 8);
	    cpld_read_register(MM_FPGA_VERSION + 2, &regval8);
	    fpga_ver |= ((regval8 & 0xf) << 12);
	    cpld_read_register(MM_FPGA_VERSION + 3, &regval8);
	    fpga_ver |= ((regval8 & 0xf) << 16);
	    break;
	case SWBD161:
	case SWBD162:
	    cpld_read_register(MM_FPGA_VERSION, &regval8);
	    fpga_ver = regval8;
	    cpld_read_register(MM_FPGA_VERSION + 1, &regval8);
	    fpga_ver |= ((regval8 & 0xff) << 8);
	    cpld_read_register(MM_FPGA_VERSION + 2, &regval8);
	    fpga_ver |= ((regval8 & 0x07) << 16);
	    cpld_read_register(MM_FPGA_VERSION + 3, &regval8);
	    fpga_ver |= ((regval8 & 0x0) << 24);
	    break;

	default:
	    cpld_read_register32(MM_FPGA_VERSION, &fpga_ver);
	    break;
    }

#if DEBUG_FPGA_WDT
    printk("FPGA WDT Revision Word = 0x%08x\n", fpga_ver);
#endif
 
    /* change the FPGA revision word to little endian */
    switch(silkworm.board) {
        case SWBD156:
        case SWBD157:
        case SWBD158:
	case SWBD167:/* Shenlong */	
            fpga_ver = __cpu_to_le32( fpga_ver );
#if DEBUG_FPGA_WDT
            printk("Changed to little endian of FPGA WDT Revision Word = 0x%08x\n", fpga_ver);
#endif
            break;
    }
    /* rearrange 4b nibbles of fpga_ver to make FBEHGACD */
    regval = ((fpga_ver & 0x00000f00) << 20) |
              ((fpga_ver & 0x0f000000) <<  0) |
              ((fpga_ver & 0x0000f000) <<  8) |
              ((fpga_ver & 0x0000000f) << 16) |
              ((fpga_ver & 0x000000f0) <<  8) |
              ((fpga_ver & 0xf0000000) >> 20) |
              ((fpga_ver & 0x00f00000) >> 16) |
              ((fpga_ver & 0x000f0000) >> 16) ;

#if DEBUG_FPGA_WDT
    printk("FPGA WDT disarm MSB 4 byte key = 0x%08x\n", regval);
#endif
    
    /* Mercury is the default */
    switch (silkworm.board) {
	case SWBD127:
	    cpld_write_register32(STARLIFTER4040_BOOT_WDT_CTRL, regval);
	    break;
	case SWBD131:
	case SWBD137:
	case SWBD138:
	case SWBD151:
	case SWBD153:
	case SWBD154:
	case SWBD155:        
		cpld_write_register32(CASTOR_BOOT_WDT_CTRL, regval);
		break;
	case SWBD148:
		regval8 = ((regval >> 24) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL, regval8);

                regval8 = ((regval >> 16) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 1, regval8);

                regval8 = ((regval >> 8) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 2, regval8);

                regval8 = (regval & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 3, regval8);
                break;
	case SWBD161:
	case SWBD162:
		regval8 = ((regval >> 24) & 0xff);
		cpld_write_register(WEDGE_BOOT_WDT_CTRL, regval8);

		regval8 = ((regval >> 16) & 0xff);
		cpld_write_register(WEDGE_BOOT_WDT_CTRL + 1, regval8);

		regval8 = ((regval >> 8) & 0xff);
		cpld_write_register(WEDGE_BOOT_WDT_CTRL + 2, regval8);

		regval8 = (regval & 0xff);
		cpld_write_register(WEDGE_BOOT_WDT_CTRL + 3, regval8);
		break;

	case SWBD156:
	case SWBD157:
	case SWBD158:
	case SWBD167:/* Shenlong */	
		cpld_write_register32(CARBON_CPLD_BOOT_WDT_CTRL_MSB, regval);
		break;
	default:
	    cpld_write_register32(MM_BOOT_WDT_CTRL, regval);
	    break;
    }
    
    /* rearrange 4b nibbles of fpga_ver to make DAFGEBHC */
    regval = ((fpga_ver & 0x000f0000) << 12) |
              ((fpga_ver & 0xf0000000) >>  4) |
              ((fpga_ver & 0x00000f00) << 12) |
              ((fpga_ver & 0x000000f0) << 12) |
              ((fpga_ver & 0x0000f000) <<  0) |
              ((fpga_ver & 0x0f000000) >> 16) |
              ((fpga_ver & 0x0000000f) <<  4) |
              ((fpga_ver & 0x00f00000) >> 20) ;
    
#if DEBUG_FPGA_WDT
    printk("FPGA WDT disarm LSB 4 byte key = 0x%08x\n", regval);
#endif

    /* Mercury is the default */
    switch (silkworm.board) {
	case SWBD127:
	    cpld_write_register32(STARLIFTER4040_BOOT_WDT_CTRL, regval);
	    break;
	case SWBD131:
	case SWBD137:
	case SWBD138:
	case SWBD151:
	case SWBD153:
	case SWBD154:
	case SWBD155:        
	    cpld_write_register32(CASTOR_BOOT_WDT_CTRL, regval);
	    break;
	case SWBD148:
		regval8 = ((regval >> 24) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 4, regval8);

                regval8 = ((regval >> 16) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 5, regval8);

                regval8 = ((regval >> 8) & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 6, regval8);

                regval8 = (regval & 0xff);
                cpld_write_register(SKYBOLT_BOOT_WDT_CTRL + 7, regval8);
                break;

	case SWBD161:
	case SWBD162:
				regval8 = ((regval >> 24) & 0xff);
                cpld_write_register(WEDGE_BOOT_WDT_CTRL + 4, regval8);

                regval8 = ((regval >> 16) & 0xff);
                cpld_write_register(WEDGE_BOOT_WDT_CTRL + 5, regval8);

                regval8 = ((regval >> 8) & 0xff);
                cpld_write_register(WEDGE_BOOT_WDT_CTRL + 6, regval8);

                regval8 = (regval & 0xff);
                cpld_write_register(WEDGE_BOOT_WDT_CTRL + 7, regval8);
                break;

	case SWBD156:
	case SWBD157:
	case SWBD158:
	case SWBD167:/* Shenlong */	
		cpld_write_register32(CARBON_CPLD_BOOT_WDT_CTRL_LSB, regval);
		break;
	default:
	    cpld_write_register32(MM_BOOT_WDT_CTRL, regval);
	    break;
    }
}

/* 
 * lc48x10g_disable_wdt - display the wdt count and disable the boot wdt in LC
 *
 * The value of FPGA version register is viewed as a set of
 * eight 4b nibbles, ABCDEFGH - [31:28,27:24,23:20,19:16,15:12,11:8,7:4,3:0]
 * The watchdog timer is disabled by writing the value: FBEHGACD,
 * followed by DAFGEBHC to the key sequence register
 */
static void
lc48x10g_disable_wdt(void)
{
    unsigned int fpga_ver = 0;
    unsigned int regval = 0;

    /* print the wdt counter value */
    cpld_read_register32(LC48X10G_BOOT_WDT_CTR, &regval);
    printk (KERN_CRIT "Uboot wdt count value: %u\n", regval);

    /* Disable the wdt */
    cpld_read_register32(LC48X10G_FPGA_VERSION, &fpga_ver);
    
    /* rearrange 4b nibbles of fpga_ver to make FBEHGACD */
    regval = ((fpga_ver & 0x00000f00) << 20) |
              ((fpga_ver & 0x0f000000) <<  0) |
              ((fpga_ver & 0x0000f000) <<  8) |
              ((fpga_ver & 0x0000000f) << 16) |
              ((fpga_ver & 0x000000f0) <<  8) |
              ((fpga_ver & 0xf0000000) >> 20) |
              ((fpga_ver & 0x00f00000) >> 16) |
              ((fpga_ver & 0x000f0000) >> 16) ;
    
    cpld_write_register32(LC48X10G_BOOT_WDT_CTRL, regval);
    
    /* rearrange 4b nibbles of fpga_ver to make DAFGEBHC */
    regval = ((fpga_ver & 0x000f0000) << 12) |
              ((fpga_ver & 0xf0000000) >>  4) |
              ((fpga_ver & 0x00000f00) << 12) |
              ((fpga_ver & 0x000000f0) << 12) |
              ((fpga_ver & 0x0000f000) <<  0) |
              ((fpga_ver & 0x0f000000) >> 16) |
              ((fpga_ver & 0x0000000f) <<  4) |
              ((fpga_ver & 0x00f00000) >> 20) ;
    
    cpld_write_register32(LC48X10G_BOOT_WDT_CTRL, regval);
}
#endif

/* 
 * ppc_booke_sw_wdt_heartbeat: detects intr locked.
 *
 * This is invoked from timer IRQ. The system is allowed to be locked 
 * for few millisec (?) otherwise the watchdog will kickin.
 * Otherwise if system is too busy, yet not locked the soft-watchdog 
 * timer is trigrred.
 */
void ppc_booke_sw_wdt_heartbeat(void)
{
	if (sw_wdt_default) {
		/* default used until wdt inits */
		atomic_set(&sw_wdt_heartbeat_count[smp_processor_id()], sw_wdt_period);
		sw_wdt_default = 0;
	}	

	if ((atomic_read(&sw_wdt_heartbeat_count[smp_processor_id()]) > 0) || ( !sw_wdt_enable )) {
		if (atomic_read(&sw_wdt_heartbeat_count[smp_processor_id()]) > 0)
			atomic_dec(&sw_wdt_heartbeat_count[smp_processor_id()]);
			__booke_wdt_ping(NULL);
        } else {
		printk("Watchdog Timer Timed out, system reset!");
		atomic_set(&sw_wdt_heartbeat_count[smp_processor_id()], sw_wdt_period);
		dump_stack();
		BUG();
	}
}

// only refresh wdt on local core
void kick_watchdog(void)
{
	__booke_wdt_ping(NULL);
}

void shut_watchdog(void)
{
	sw_wdt_enable = 0;
	printk (KERN_INFO "PowerPC Book-E Watchdog Shutdown soft timer\n");
}

EXPORT_SYMBOL(shut_watchdog);
EXPORT_SYMBOL(kick_watchdog);
#endif /* CONFIG_SILKWORM */

static void __booke_wdt_enable(void *data)
{
	u32 val;

	/* clear status before enabling watchdog */
	__booke_wdt_ping(NULL);
	val = mfspr(SPRN_TCR);
	val &= ~WDTP_MASK;
	val |= (TCR_WIE|TCR_WRC(WRC_CHIP)|WDTP(booke_wdt_period));

	mtspr(SPRN_TCR, val);
}

static ssize_t booke_wdt_write(struct file *file, const char __user *buf,
				size_t count, loff_t *ppos)
{
	booke_wdt_ping();
	return count;
}

static struct watchdog_info ident = {
	.options = WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
	.identity = "PowerPC Book-E Watchdog",
};

static long booke_wdt_ioctl(struct file *file,
				unsigned int cmd, unsigned long arg)
{
	u32 tmp = 0;
	u32 __user *p = (u32 __user *)arg;
#ifdef CONFIG_SILKWORM
	unsigned long period;
	int state;
#endif

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		if (copy_to_user((void __user*)arg, &ident, sizeof(struct watchdog_info)))
			return -EFAULT;
	case WDIOC_GETSTATUS:
		return put_user(ident.options, p);
	case WDIOC_GETBOOTSTATUS:
		/* XXX: something is clearing TSR */
		tmp = mfspr(SPRN_TSR) & TSR_WRS(3);
		/* returns 1 if last reset was caused by the WDT */
		return (tmp ? 1 : 0);
	case WDIOC_SETOPTIONS:
#ifdef CONFIG_SILKWORM
		if(copy_from_user(&state, (int*) arg, sizeof(int)))
			return -EFAULT;
		if (state & WDIOS_DISABLECARD) {
			sw_wdt_enable = 0;
			printk(KERN_NOTICE "Soft watchdog timer is disabled\n");
			break;
		}
		if (state & WDIOS_ENABLECARD) {
			sw_update_timer();
			sw_wdt_enable = 1;
			booke_wdt_ping();
			printk(KERN_NOTICE "Soft watchdog timer is enabled\n");
			break;
		}
		return 0;
#else
		if (get_user(tmp, p))
			return -EINVAL;
		if (tmp == WDIOS_ENABLECARD) {
			booke_wdt_ping();
			break;
		} else
			return -EINVAL;
		return 0;
#endif
	case WDIOC_KEEPALIVE:
#ifdef CONFIG_SILKWORM
		sw_update_timer();
#else
		booke_wdt_ping();
#endif
		return 0;
	case WDIOC_SETTIMEOUT:
		if (get_user(booke_wdt_period, p))
			return -EFAULT;
		mtspr(SPRN_TCR, (mfspr(SPRN_TCR) & ~WDTP_MASK) |
						WDTP(booke_wdt_period));
		return 0;
	case WDIOC_GETTIMEOUT:
		return put_user(booke_wdt_period, p);

#ifdef CONFIG_SILKWORM
	case WDIOC_GETPERIOD:
		/* return watchdog period (units = microseconds) */
		period = sw_wdt_period / REFRESHES_PER_SEC * MICROSECBASE;
		if (copy_to_user((unsigned long *)arg, &period, 
		    sizeof(period))) {
			return -EFAULT;
		}
		break;
	case WDIOC_SETPERIOD:
		/*
		** set watchdog period (units = microseconds)
		** value of zero means maximum
		**
		** Don't set a watchdog period to a value less than
		** the requested value (period will be rounded up to
		** next available interval the watchdog supports).
		**
		** The software watchdog will expire at some point in
		** the range of (rounded up period) ..
		** (rounded up period + 1 jiffie).  If interrupts are
		** disabled so that the software watchdog is unable to
		** reset the system, then the hardware watchdog will
		** eventually reset the system.
		*/
		if (copy_from_user(&period, (unsigned long *)arg,
		    sizeof(period))) {
			return -EFAULT;
		}
		if (period == 0)
			period = MAXONEHOUR * MICROSECBASE;        
		sw_wdt_period = period / MICROSECBASE * REFRESHES_PER_SEC;
		sw_update_timer();
		printk(KERN_NOTICE"PowerPC Book-E Watchdog: SW New period set to %lu millisec\n", sw_wdt_period);
		break;
#endif
	default:
		return -ENOTTY;
	}

	return 0;
}

static int booke_wdt_open(struct inode *inode, struct file *file)
{
	spin_lock(&booke_wdt_lock);
#ifdef CONFIG_SILKWORM
	if (sw_wdt_period == 0)
		sw_wdt_period = SW_WDT_DEFAULT_PERIOD * REFRESHES_PER_SEC;
	sw_update_timer();
	sw_wdt_enable = 1;
#endif
	if (booke_wdt_enabled == 0) {
		booke_wdt_enabled = 1;
		on_each_cpu(__booke_wdt_enable, NULL, 0);
		printk(KERN_INFO
		      "PowerPC Book-E Watchdog Timer Enabled (wdt_period=%d)\n",
				booke_wdt_period);
	}
	spin_unlock(&booke_wdt_lock);

	return nonseekable_open(inode, file);
}

static const struct file_operations booke_wdt_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.write = booke_wdt_write,
	.unlocked_ioctl = booke_wdt_ioctl,
	.open = booke_wdt_open,
};

static struct miscdevice booke_wdt_miscdev = {
	.minor = WATCHDOG_MINOR,
	.name = "watchdog",
	.fops = &booke_wdt_fops,
};

static void __exit booke_wdt_exit(void)
{
	misc_deregister(&booke_wdt_miscdev);
}

static int __init booke_wdt_init(void)
{
	int ret = 0;

	printk(KERN_INFO "PowerPC Book-E Watchdog Timer Loaded\n");
	ident.firmware_version = cur_cpu_spec->pvr_value;

	ret = misc_register(&booke_wdt_miscdev);
	if (ret) {
		printk(KERN_CRIT "Cannot register miscdev on minor=%d: %d\n",
				WATCHDOG_MINOR, ret);
		return ret;
	}

	spin_lock(&booke_wdt_lock);
	if (booke_wdt_enabled == 1) {

#ifdef CONFIG_SILKWORM
		if (sw_wdt_period == 0)
			sw_wdt_period = SW_WDT_DEFAULT_PERIOD * REFRESHES_PER_SEC;
		sw_update_timer();
		printk("already enabled\n");
#endif
		printk(KERN_INFO
		      "PowerPC Book-E Watchdog Timer Enabled (wdt_period=%d)\n",
				booke_wdt_period);
		on_each_cpu(__booke_wdt_enable, NULL, 0);
	}
	spin_unlock(&booke_wdt_lock);

#ifdef CONFIG_SILKWORM
	/* display and disable the uboot wdt count */
	switch (silkworm.board) {
		case SWBD127:
		case SWBD131:
		case SWBD137:
		case SWBD138:
		case SWBD141:
		case SWBD148:
		case SWBD161:
		case SWBD162:
		case SWBD151:
		case SWBD1000:
		case SWBD1001:
		case SWBD1002:
		case SWBD153:
		case SWBD154:
		case SWBD155:            
		case SWBD156: /* Redbird */
		case SWBD157: /* Graphite */
		case SWBD158: /* Carbon */
		case SWBD167: /* Shenlong */		
			mm_disable_wdt();
			break;
		case SWBD1004:
		case SWBD1005:
		case SWBD1006:
		case SWBD1007:
		case SWBD1008:
		case SWBD1009:
		case SWBD1010:
		case SWBD1011:
			lc48x10g_disable_wdt();
			break;
		case SWBD112:
		case SWBD123:
			/* Platform FPGA does not support WDT */
			break;
		default:
			printk (KERN_CRIT "Unable to fetch uboot wdt count\n");
			break;
	}
#endif
	return ret;
}
device_initcall(booke_wdt_init);
