/*
 * Freescale hypervisor ioctl interface
 *
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 *
 * This file is used by the Freescale hypervisor management driver.  It can
 * also be included by applications that need to communicate with the driver
 * via the ioctl interface.
 */

#ifndef FSL_HYPERVISOR_H
#define FSL_HYPERVISOR_H

#include <linux/types.h>

#ifdef CONFIG_SILKWORM
#include <linux/of.h>
#endif

/**
 * Freescale hypervisor ioctl parameter
 */
union fsl_hv_ioctl_param {

	/**
	 * @ret: Return value.
	 *
	 * This is always the first word of any structure.
	 */
	__u32 ret;

	/**
	 * struct fsl_hv_ioctl_restart: restart a partition
	 * @ret: return error code from the hypervisor
	 * @partition: the ID of the partition to restart, or -1 for the
	 *             calling partition
	 *
	 * Used by FSL_HV_IOCTL_PARTITION_RESTART
	 */
	struct fsl_hv_ioctl_restart {
		__u32 ret;
		__u32 partition;
	} restart;

	/**
	 * struct fsl_hv_ioctl_status: get a partition's status
	 * @ret: return error code from the hypervisor
	 * @partition: the ID of the partition to query, or -1 for the
	 *             calling partition
	 * @status: The returned status of the partition
	 *
	 * Used by FSL_HV_IOCTL_PARTITION_GET_STATUS
	 *
	 * Values of 'status':
	 *    0 = Stopped
	 *    1 = Running
	 *    2 = Starting
	 *    3 = Stopping
	 */
	struct fsl_hv_ioctl_status {
		__u32 ret;
		__u32 partition;
		__u32 status;
	} status;

	/**
	 * struct fsl_hv_ioctl_start: start a partition
	 * @ret: return error code from the hypervisor
	 * @partition: the ID of the partition to control
	 * @entry_point: The offset within the guest IMA to start execution
	 * @load: If non-zero, reload the partition's images before starting
	 *
	 * Used by FSL_HV_IOCTL_PARTITION_START
	 */
	struct fsl_hv_ioctl_start {
		__u32 ret;
		__u32 partition;
		__u32 entry_point;
		__u32 load;
	} start;

	/**
	 * struct fsl_hv_ioctl_stop: stop a partition
	 * @ret: return error code from the hypervisor
	 * @partition: the ID of the partition to stop, or -1 for the calling
	 *             partition
	 *
	 * Used by FSL_HV_IOCTL_PARTITION_STOP
	 */
	struct fsl_hv_ioctl_stop {
		__u32 ret;
		__u32 partition;
	} stop;

	/**
	 * struct fsl_hv_ioctl_memcpy: copy memory between partitions
	 * @ret: return error code from the hypervisor
	 * @source: the partition ID of the source partition, or -1 for this
	 *          partition
	 * @target: the partition ID of the target partition, or -1 for this
	 *          partition
	 * @local_addr: user-space virtual address of a buffer in the local
	 *              partition
	 * @remote_addr: guest physical address of a buffer in the
	 *           remote partition
	 * @count: the number of bytes to copy.  Both the local and remote
	 *         buffers must be at least 'count' bytes long
	 *
	 * Used by FSL_HV_IOCTL_MEMCPY
	 *
	 * The 'local' partition is the partition that calls this ioctl.  The
	 * 'remote' partition is a different partition.  The data is copied from
	 * the 'source' paritition' to the 'target' partition.
	 *
	 * The buffer in the remote partition must be guest physically
	 * contiguous.
	 *
	 * This ioctl does not support copying memory between two remote
	 * partitions or within the same partition, so either 'source' or
	 * 'target' (but not both) must be -1.  In other words, either
	 *
	 *      source == local and target == remote
	 * or
	 *      source == remote and target == local
	 */
	struct fsl_hv_ioctl_memcpy {
		__u32 ret;
		__u32 source;
		__u32 target;
		__u64 local_vaddr;
		__u64 remote_paddr;
		__u64 count;
	} memcpy;

	/**
	 * struct fsl_hv_ioctl_doorbell: ring a doorbell
	 * @ret: return error code from the hypervisor
	 * @doorbell: the handle of the doorbell to ring doorbell
	 *
	 * Used by FSL_HV_IOCTL_DOORBELL
	 */
	struct fsl_hv_ioctl_doorbell {
		__u32 ret;
		__u32 doorbell;
	} doorbell;

	/**
	 * struct fsl_hv_ioctl_prop: get/set a device tree property
	 * @ret: return error code from the hypervisor
	 * @handle: handle of partition whose tree to access
	 * @path: virtual address of path name of node to access
	 * @propname: virtual address of name of property to access
	 * @propval: virtual address of property data buffer
	 * @proplen: Size of property data buffer
	 *
	 * Used by FSL_HV_IOCTL_DOORBELL
	 */
	struct fsl_hv_ioctl_prop {
		__u32 ret;
		__u32 handle;
		__u64 path;
		__u64 propname;
		__u64 propval;
		__u32 proplen;
	} prop;


	/**
	 * struct fsl_hv_ioctl_sys_reset: reset the system
	 * @ret: return error code from the hypervisor if call fails
	 *
	 */
	struct fsl_hv_ioctl_reset {
		__u32 ret;
	} reset;


};

/*
 * ioctl commands.
 */
enum {
	FSL_HV_IOCTL_PARTITION_RESTART = 1, /* Boot another partition */
	FSL_HV_IOCTL_PARTITION_GET_STATUS = 2, /* Boot another partition */
	FSL_HV_IOCTL_PARTITION_START = 3, /* Boot another partition */
	FSL_HV_IOCTL_PARTITION_STOP = 4, /* Stop this or another partition */
	FSL_HV_IOCTL_MEMCPY = 5, /* Copy data from one partition to another */
	FSL_HV_IOCTL_DOORBELL = 6, /* Ring a doorbell */

	/* Get a property from another guest's device tree */
	FSL_HV_IOCTL_GETPROP = 7,

	/* Set a property in another guest's device tree */
	FSL_HV_IOCTL_SETPROP = 8,
	FSL_HV_IOCTL_SYS_RESET = 9,
#ifdef CONFIG_SILKWORM
	FSL_HV_IOCTL_PARTITION_RESTART_UNC = 10,
#endif	
};

#ifdef __KERNEL__

/**
 * has_fsl_hypervisor - return TRUE if we're running under FSL hypervisor
 *
 * This function checks to see if we're running under the Freescale
 * hypervisor, and returns zero if we're not, or non-zero if we are.
 *
 * First, it checks if MSR[GS]==1, which means we're running under some
 * hypervisor.  Then it checks if there is a hypervisor node in the device
 * tree.  Currently, that means there needs to be a node in the root called
 * "hypervisor" and which has a property named "fsl,hv-version".
 */

#ifdef CONFIG_SILKWORM
static inline
int has_hypervisor(void)
{
	if (of_find_node_by_path("/hypervisor"))
		return 1;

	return 0;
}

void fsl_hv_restart_unc(void);

#ifndef CONFIG_FSL_HYPERVISOR
static inline
int has_fsl_hypervisor(void)
{
	if (of_find_node_by_path("/hypervisor"))
		return 1;

	return 0;
}
#else
int has_fsl_hypervisor(void);
#endif 

#else 
int has_fsl_hypervisor(void);
#endif



/**
 * fsl_hv_restart - restart the current partition
 */
void fsl_hv_restart(char *cmd);

/**
 * fsl_hv_halt - halt the current partition
 */
void fsl_hv_halt(void);

/**
 * fsl_hv_claim -- claim a device
 */
#ifdef CONFIG_SILKWORM
    #ifdef CONFIG_FSL_HYPERVISOR
        int fsl_hv_claim(struct device_node *node);
    #else // CONFIG_FSL_HYPERVISOR
        static inline int fsl_hv_claim(struct device_node *node) {
            printk("Possible Bug: calling fsl_hv_claim without CONFIG_FSL_HYPERVISOR.\n");
            return (0);
        }
    #endif // CONFIG_FSL_HYPERVISOR
#else // CONFIG_SILKWORM
    int fsl_hv_claim(struct device_node *node);
#endif // CONFIG_SILKWORM


/**
 * fsl_hv_partition_stop_dma -- perform deferred dma-stop on (stopped) partition
 */
int fsl_hv_partition_stop_dma(struct device_node *node);

/**
 * fsl_hv_partition_reboot -- perform a reboot on a partition
 */
int fsl_hv_partition_reboot(struct device_node *node);

/**
 * fsl_hv_is_first_boot -- determines if the current partition has been rebooted
 *
 * Returns non-zero if the current partition has not been rebooted before
 */
int fsl_hv_is_first_boot(void);

#ifdef CONFIG_FSL_HV_MANAGER

/**
 * fsl_hv_failover_register -- register a callback for failover events
 *
 * This function is called by device drivers to register their callback
 * functions for fail-over events.
 */
int fsl_hv_failover_register(struct notifier_block *nb);

/**
 * fsl_hv_failover_unregister -- unregister a callback for failover events
 */
int fsl_hv_failover_unregister(struct notifier_block *nb);

#endif /* CONFIG_FSL_HV_MANAGER */

#ifdef CONFIG_FSL_HYPERVISOR_FAILOVER
#define HV_FAILOVER_KICKOFF			5
#define HV_FAILOVER_PRE_DMA_STOP	4
#define HV_FAILOVER_DMA_STOP		3
#define HV_FAILOVER_POST_DMA_STOP	2
#define HV_FAILOVER_REBOOT			0
#endif

#endif /* __KERNEL__ */

#endif
