/*
 *	linux/mm/msync.c
 *
 * Copyright (C) 1994-1999  Linus Torvalds
 */

/*
 * The msync() system call.
 */
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/file.h>
#include <linux/syscalls.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/timer.h>
#include <asm/timex.h>

#ifdef CONFIG_SILKWORM
/*
 * Copyright (C) 2011-2012  Brocade Communications Systems, Inc.
 * All rights reserved.
 *
 *	Enahnced msync() syscall to get a list of modified pages within
 *	specified range for Transparent HA framework.
 */
#include <linux/slab.h>
#include <linux/pagemap.h>
#include <linux/hugetlb.h>
#include <linux/highmem.h>

#include <asm/pgtable.h>
#include <asm/tlbflush.h>

/*
 * Structure to cachle first few TLB mappings to be invalidated and space
 * to hold modified bitmap for typical memory range at page granularity.
 * Note that the bitmap is variable length, default size covers 4MB range.
 */
#define	MMOD_FLUSH_TLB_MAX		32			/* Max individual TLB flush */
#define	MMOD_BITMAP_SIZE		128			/* Bitmap size in ulong words */

typedef struct {
		struct vm_area_struct	*vma;
		unsigned long			addr;
} mmod_flush_tlb_t;

typedef struct {
	int				flush_tlb_cnt;
	mmod_flush_tlb_t	flush_tlb[MMOD_FLUSH_TLB_MAX];
	unsigned long	bitmap[MMOD_BITMAP_SIZE / sizeof (unsigned long)];
} mmod_info_t;

/*
 * MMOD TLB flush mode and threshold
 */
static uint	mmod_flush_tlb_mode = 2;
static uint	mmod_flush_tlb_threshold = MMOD_FLUSH_TLB_MAX;
static uint	mmod_kmalloc_cnt;
static int	mmod_debug = 0;

#define	MMOD_STATS				1			/* set it for statistics */

#ifdef MMOD_STATS
static uint	mmod_test_clear_cnt;
static uint	mmod_clear_flush_cnt;
static uint	mmod_cnt;
static uint	mmod_test_clear_time;
static uint	mmod_clear_flush_time;
static uint	mmod_time;
static uint	mmod_test_clear_time_high;
static uint	mmod_clear_flush_time_high;
static uint	mmod_time_high;
static uint	mmod_sysctl_inited;
#endif

/* The following should be part of include/asm-generic/pgtable.h file */
#ifndef __HAVE_ARCH_PTEP_TEST_AND_CLEAR_DIRTY
#define ptep_test_and_clear_dirty(__vma, __address, __ptep)	\
({								  							\
	pte_t __pte = *__ptep;		  							\
	int r = 1;					  							\
	if (!pte_dirty(__pte))		  							\
		r = 0;					  							\
	else													\
		set_pte_at((__vma)->vm_mm, (__address), (__ptep),   \
			   pte_mkclean(__pte));							\
	r;							  							\
})
#endif

#ifndef __HAVE_ARCH_PTEP_CLEAR_DIRTY_FLUSH
#ifdef MMOD_STATS
#define ptep_clear_flush_dirty(__vma, __address, __ptep)			\
({								  									\
	int __dirty;													\
	uint	t1;														\
	t1 = get_tbl();													\
	__dirty = ptep_test_and_clear_dirty(__vma, __address, __ptep);	\
	t1 = get_tbl() - t1;											\
	mmod_test_clear_cnt++;											\
	if ((mmod_test_clear_time += t1) < t1)							\
		mmod_test_clear_time_high++;								\
	if (__dirty && mmod_flush_tlb_mode == 0)	{					\
		t1 = get_tbl();												\
		flush_tlb_page(__vma, __address);		   					\
		t1 = get_tbl() - t1;										\
		mmod_clear_flush_cnt++;										\
		if ((mmod_clear_flush_time += t1) < t1) 					\
			mmod_clear_flush_time_high++;							\
	}																\
	__dirty;														\
})
#else	/* !MMOD_STATS */
#define ptep_clear_flush_dirty(__vma, __address, __ptep)			\
({								  									\
	int __dirty;													\
	__dirty = ptep_test_and_clear_dirty(__vma, __address, __ptep);	\
	if (__dirty && mmod_flush_tlb_mode == 0)	{					\
		flush_tlb_page(__vma, __address);		   					\
	}																\
	__dirty;														\
})
#endif	/* !MMOD_STATS */
#endif

static unsigned long mmod_pte_range(struct vm_area_struct *vma, pmd_t *pmd,
							unsigned long addr, unsigned long end,
							mmod_info_t *mmodinfop, unsigned long base)
{
	pte_t *pte;
	unsigned long ret = 0;
	unsigned long  start = addr;				/* For debug only */
	int bitnum = (addr - base) >> PAGE_SHIFT;

	pte = pte_offset_map(pmd, addr);
	do {
		unsigned long pfn;
		struct page *page;

		if (!pte_present(*pte))
			continue;

		if (!pte_maybe_dirty(*pte))
			continue;

		pfn = pte_pfn(*pte);
		if (!pfn_valid(pfn))
			continue;
		page = pfn_to_page(pfn);
		if (PageReserved(page))
			continue;

		if (ptep_clear_flush_dirty(vma, addr, pte)) {
			set_page_dirty(page);
			ret++;
			if (mmod_flush_tlb_mode == 2) {
				if (mmodinfop->flush_tlb_cnt < mmod_flush_tlb_threshold) {
					mmod_flush_tlb_t *flushp;

					flushp = &mmodinfop->flush_tlb[mmodinfop->flush_tlb_cnt++];
					flushp->vma = vma;
					flushp->addr = addr;
				}
			}
			mmodinfop->bitmap[bitnum/BITS_PER_LONG] |=
					1UL << (bitnum & (BITS_PER_LONG - 1));
			if (mmod_debug && pte_dirty(*pte))
				printk(KERN_ERR "mmod: pte dirty even after clear for %lx\n",
					addr);
		}
	} while (pte++, addr += PAGE_SIZE, bitnum++, addr != end);
	pte_unmap(pte - 1);

	if (mmod_debug > 2)
		printk(KERN_ERR
			"mmod_pte_range pmd: %p addr: 0x%lx end: 0x%lx ret: %ld\n",
			pmd, start, end, ret);

	return ret;
}

static inline unsigned long mmod_pmd_range(struct vm_area_struct *vma,
		pud_t *pud, unsigned long addr, unsigned long end,
		mmod_info_t *mmodinfop, unsigned long base)
{
	pmd_t *pmd;
	unsigned long next;
	unsigned long ret = 0;

	pmd = pmd_offset(pud, addr);
	do {
		next = pmd_addr_end(addr, end);
		if (pmd_none_or_clear_bad(pmd))
			continue;
		ret += mmod_pte_range(vma, pmd, addr, next, mmodinfop, base);
	} while (pmd++, addr = next, addr != end);

	return ret;
}

static inline unsigned long mmod_pud_range(struct vm_area_struct *vma,
		pgd_t *pgd, unsigned long addr, unsigned long end,
		mmod_info_t *mmodinfop, unsigned long base)
{
	pud_t *pud;
	unsigned long next;
	unsigned long ret = 0;

	pud = pud_offset(pgd, addr);
	do {
		next = pud_addr_end(addr, end);
		if (pud_none_or_clear_bad(pud))
			continue;
		ret += mmod_pmd_range(vma, pud, addr, next, mmodinfop, base);
	} while (pud++, addr = next, addr != end);
	return ret;
}

static unsigned long mmod_page_range(struct vm_area_struct *vma,
							unsigned long addr, unsigned long end,
							mmod_info_t *mmodinfop, unsigned long base)
{
	struct mm_struct *mm = vma->vm_mm;
	pgd_t *pgd;
	unsigned long next;
	unsigned long ret = 0;

	/* For hugepages we can't go walking the page table normally,
	 * but that's ok, hugetlbfs is memory based, so we don't need
	 * to do anything more on an msync() */
	if (is_vm_hugetlb_page(vma))
		return 0;

	BUG_ON(addr >= end);
	pgd = pgd_offset(mm, addr);
	flush_cache_range(vma, addr, end);
	spin_lock(&mm->page_table_lock);
	do {
		next = pgd_addr_end(addr, end);
		if (pgd_none_or_clear_bad(pgd))
			continue;
		ret += mmod_pud_range(vma, pgd, addr, next, mmodinfop, base);
	} while (pgd++, addr = next, addr != end);
	spin_unlock(&mm->page_table_lock);

	return ret;
}

static struct ctl_table mmod_kern_table[] = {
	{
		.procname	= "mmod_kmalloc_cnt",
		.data		= &mmod_kmalloc_cnt,
		.maxlen		= sizeof(mmod_kmalloc_cnt),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_flush_tlb_threshold",
		.data		= &mmod_flush_tlb_threshold,
		.maxlen		= sizeof(mmod_flush_tlb_threshold),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_flush_tlb_mode",
		.data		= &mmod_flush_tlb_mode,
		.maxlen		= sizeof(mmod_flush_tlb_mode),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_debug",
		.data		= &mmod_debug,
		.maxlen		= sizeof(mmod_debug),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
#ifdef MMOD_STATS
	{
		.procname	= "mmod_test_clear_cnt",
		.data		= &mmod_test_clear_cnt,
		.maxlen		= sizeof(mmod_test_clear_cnt),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_test_clear_time_high",
		.data		= &mmod_test_clear_time_high,
		.maxlen		= sizeof(mmod_test_clear_time_high),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_test_clear_time",
		.data		= &mmod_test_clear_time,
		.maxlen		= sizeof(mmod_test_clear_time),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_clear_flush_cnt",
		.data		= &mmod_clear_flush_cnt,
		.maxlen		= sizeof(mmod_clear_flush_cnt),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_clear_flush_time_high",
		.data		= &mmod_clear_flush_time_high,
		.maxlen		= sizeof(mmod_clear_flush_time_high),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_clear_flush_time",
		.data		= &mmod_clear_flush_time,
		.maxlen		= sizeof(mmod_clear_flush_time),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_cnt",
		.data		= &mmod_cnt,
		.maxlen		= sizeof(mmod_cnt),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_time_high",
		.data		= &mmod_time_high,
		.maxlen		= sizeof(mmod_time_high),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname	= "mmod_time",
		.data		= &mmod_time,
		.maxlen		= sizeof(mmod_time),
		.mode		= 0644,
		.proc_handler = proc_dointvec,
	},
#endif /* MMOD_STATS */
	{}
};

static struct ctl_table mmod_root_table[] = {
	{
		.procname	= "kernel",
		.mode		= 0555,
		.child		= mmod_kern_table,
	},
	{}
};

int __init mmod_sysctl_init(void)
{
	mmod_sysctl_inited++;
	printk(KERN_ERR "mmod_sysctl_inited: %d\n", mmod_sysctl_inited);
	register_sysctl_table(mmod_root_table);
	return 0;
}

__initcall(mmod_sysctl_init);


/*
 * Check if any pages in the specified range have been modified since
 * last time and update the modify bit accordingly.
 */
/* asmlinkage */ long sys_mmod(unsigned long start, size_t len,
	unsigned long __user * vec)
{
	unsigned long end;
	struct vm_area_struct *vma;
	int error = -ENOMEM;
	int	offset = start & (PAGE_SIZE -1);
	int	dirty_pages = 0;
	int	done = 0;
	int	bitmapsz = 0;
	unsigned long base;
	mmod_info_t	mmodinfo;
	mmod_info_t	*mmodinfop = &mmodinfo;
#ifdef MMOD_STATS
	uint t1;
#endif

	if (mmod_sysctl_inited == 0)
		mmod_sysctl_init();

	if (vec == (unsigned long *)-1) {
		/* Print statistics only. */
		printk(KERN_ERR "%18s calls: 0x%8x CumTime: %8x %8x  CurTB: %8x\n",
			"sys_mmod", mmod_cnt, mmod_time_high, mmod_time, (uint)get_tbl());
		printk(KERN_ERR "%18s calls: 0x%8x CumTime: %8x %8x\n",
			"mmod_test_clear", mmod_test_clear_cnt, mmod_test_clear_time_high,
			 mmod_test_clear_time);
		printk(KERN_ERR "%18s calls: 0x%8x CumTime: %8x %8x\n",
			"mmod_clear_flush", mmod_clear_flush_cnt,
			mmod_clear_flush_time_high, mmod_clear_flush_time);
		return (0);
	}

#ifdef MMOD_STATS
	t1 = get_tbl();
#endif
	if (offset) {
		start -= offset;
		len += offset;
	}
	base = start & ((~(BITS_PER_LONG-1)) << PAGE_SHIFT);
	len = (len + ~PAGE_MASK) & PAGE_MASK;
	end = start + len;
	if (end < start)
		goto out;
	error = 0;
	if (end == start)
		goto out;

	bitmapsz = BITS_TO_LONGS((end - base) >> PAGE_SHIFT) * sizeof(long);
	if (bitmapsz > MMOD_BITMAP_SIZE) {
		if ((mmodinfop = (mmod_info_t *) kmalloc(
				sizeof (mmodinfo) + bitmapsz - MMOD_BITMAP_SIZE, 0)) == NULL) {
			error = -EAGAIN;
			goto out;
		}
		mmod_kmalloc_cnt++;
	}
	mmodinfop->flush_tlb_cnt = 0;
	memset(mmodinfop->bitmap, 0, bitmapsz);

	if (mmod_debug)
		printk(KERN_ERR "sys_mmod: start: %lx len: %u vec: %p base: %lx "
			"end: %lx veclen:%d\n",
			start, len, vec, base, end, bitmapsz);

	/*
	 * If the interval [start,end) covers some unmapped address ranges,
	 * just ignore them, but return -ENOMEM at the end.
	 */
	down_read(&current->mm->mmap_sem);
	if (vec != NULL && !access_ok(VERIFY_WRITE, vec,
						 (bitmapsz + PAGE_SIZE -1) >> PAGE_SHIFT)) {
		error = -EFAULT;
		goto out_unlock;
	}
	vma = find_vma(current->mm, start);
	if (!vma) {
		error = -ENOMEM;
		goto out_unlock;
	}
	do {
		if (mmod_debug > 1)
			printk(KERN_ERR "vma: %p (range %lx - %lx) Cur range %lx - %lx\n",
			vma, vma->vm_start, vma->vm_end, start, end);

		/* Here start < vma->vm_end. */
		if (start < vma->vm_start) {
			start = vma->vm_start;
		}

		/* Here vma->vm_start <= start < vma->vm_end. */
		if (start >= end) {
			done = 1;
		} else if (end <= vma->vm_end) {
			if (start < end) {
				int pagecnt = mmod_page_range(vma, start, end,
								mmodinfop, base);

#ifdef MMOD_STATS
				if (pagecnt > 0 && mmod_flush_tlb_mode == 1) {
					uint t1 = get_tbl();

					flush_tlb_mm(vma->vm_mm);
					t1 = get_tbl() - t1;
					mmod_clear_flush_cnt++;
					if ((mmod_clear_flush_time += t1) < t1)
						mmod_clear_flush_time_high++;
				}
#endif
				dirty_pages += pagecnt;
			}
			done = 1;
		} else {
			/* Here vma->vm_start <= start < vma->vm_end < end. */
			int pagecnt = mmod_page_range(vma, start, vma->vm_end,
								mmodinfop, base);

#ifdef MMOD_STATS
			if (pagecnt > 0 && mmod_flush_tlb_mode == 1) {
				uint t1 = get_tbl();

				flush_tlb_mm(vma->vm_mm);
				t1 = get_tbl() - t1;
				mmod_clear_flush_cnt++;
				if ((mmod_clear_flush_time += t1) < t1)
					mmod_clear_flush_time_high++;
			}
#endif
			dirty_pages += pagecnt;
		}

		if (error)
				goto out_unlock;
		vma = vma->vm_next;
	} while (vma && !done);

	if (dirty_pages > 0 && mmod_flush_tlb_mode == 2) {
#ifdef MMOD_STATS
		uint t1 = get_tbl();
#endif

		if (dirty_pages > mmod_flush_tlb_threshold) {
			flush_tlb_mm(current->mm);
			mmod_clear_flush_cnt++;
		} else {
			int		i;
			mmod_flush_tlb_t *flushp = &mmodinfop->flush_tlb[0];

			for (i = mmodinfop->flush_tlb_cnt; --i >= 0; flushp++)
				flush_tlb_page(flushp->vma, flushp->addr);
			mmod_clear_flush_cnt += mmodinfop->flush_tlb_cnt;
		}
#ifdef MMOD_STATS
		t1 = get_tbl() - t1;
		if ((mmod_clear_flush_time += t1) < t1)
			mmod_clear_flush_time_high++;
#endif
	}

out_unlock:
	up_read(&current->mm->mmap_sem);

out:
	if (error == 0 && mmodinfop != NULL) {
		if (vec != NULL && copy_to_user(vec, mmodinfop->bitmap, bitmapsz)) {
			error = -EFAULT;
		}
	}

	if (mmodinfop != NULL && mmodinfop != &mmodinfo) {
		kfree(mmodinfop);
	}

	if (error == 0) {
		error = dirty_pages;
	}
	mmod_cnt++;

#ifdef MMOD_STATS
	t1 = get_tbl() - t1;
	if ((mmod_time += t1) < t1)
		mmod_time_high++;
#endif

	if (mmod_debug)
		printk(KERN_ERR
			"sys_mmod: done error:%d (start: %lx end: %lx dirty_pages: %d)\n",
			error, start, end, dirty_pages);
	return error;
}

#endif /* CONFIG_SILKWORM */

/*
 * MS_SYNC syncs the entire file - including mappings.
 *
 * MS_ASYNC does not start I/O (it used to, up to 2.5.67).
 * Nor does it marks the relevant pages dirty (it used to up to 2.6.17).
 * Now it doesn't do anything, since dirty pages are properly tracked.
 *
 * The application may now run fsync() to
 * write out the dirty pages and wait on the writeout and check the result.
 * Or the application may run fadvise(FADV_DONTNEED) against the fd to start
 * async writeout immediately.
 * So by _not_ starting I/O in MS_ASYNC we provide complete flexibility to
 * applications.
 */
#ifdef CONFIG_SILKWORM
SYSCALL_DEFINE4(msync, unsigned long, start, size_t, len, int, flags,
	unsigned long __user *, vec)
#else
SYSCALL_DEFINE3(msync, unsigned long, start, size_t, len, int, flags)
#endif
{
	unsigned long end;
	struct mm_struct *mm = current->mm;
	struct vm_area_struct *vma;
	int unmapped_error = 0;
	int error = -EINVAL;

#ifdef CONFIG_SILKWORM
	/*
	 *
	 * This syscall has been enhanced to support a new lag, MS_MMOD_PAGES,
	 * which takes an additional "vec" bit array argument to return a list
	 * of modified pages since last such call.
	 */
	if (flags & MS_MMOD_PAGES)
		return sys_mmod(start, len, vec);
#endif

	if (flags & ~(MS_ASYNC | MS_INVALIDATE | MS_SYNC))
		goto out;
	if (start & ~PAGE_MASK)
		goto out;
	if ((flags & MS_ASYNC) && (flags & MS_SYNC))
		goto out;
	error = -ENOMEM;
	len = (len + ~PAGE_MASK) & PAGE_MASK;
	end = start + len;
	if (end < start)
		goto out;
	error = 0;
	if (end == start)
		goto out;
	/*
	 * If the interval [start,end) covers some unmapped address ranges,
	 * just ignore them, but return -ENOMEM at the end.
	 */
	down_read(&mm->mmap_sem);
	vma = find_vma(mm, start);
	for (;;) {
		struct file *file;

		/* Still start < end. */
		error = -ENOMEM;
		if (!vma)
			goto out_unlock;
		/* Here start < vma->vm_end. */
		if (start < vma->vm_start) {
			start = vma->vm_start;
			if (start >= end)
				goto out_unlock;
			unmapped_error = -ENOMEM;
		}
		/* Here vma->vm_start <= start < vma->vm_end. */
		if ((flags & MS_INVALIDATE) &&
				(vma->vm_flags & VM_LOCKED)) {
			error = -EBUSY;
			goto out_unlock;
		}
		file = vma->vm_file;
		start = vma->vm_end;
		if ((flags & MS_SYNC) && file &&
				(vma->vm_flags & VM_SHARED)) {
			get_file(file);
			up_read(&mm->mmap_sem);
			error = vfs_fsync(file, file->f_path.dentry, 0);
			fput(file);
			if (error || start >= end)
				goto out;
			down_read(&mm->mmap_sem);
			vma = find_vma(mm, start);
		} else {
			if (start >= end) {
				error = 0;
				goto out_unlock;
			}
			vma = vma->vm_next;
		}
	}
out_unlock:
	up_read(&mm->mmap_sem);
out:
	return error ? : unmapped_error;
}
