/*
 *    Copyright (c) 2012-2014 by Brocade Communications Systems, Inc.
 *    All rights reserved.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>   /* printk() */
#include <linux/init.h>
#include <linux/slab.h>	    /* kmalloc() */
#include <linux/fs.h>	    /* everything... */
#include <linux/errno.h>    /* error codes */
#include <linux/types.h>    /* size_t */
#include <linux/delay.h>	/* mdelay */
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/interrupt.h>
#include <linux/hardirq.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <asm-generic/irq_regs.h>
#include <sysdev/fsl_soc.h>
#include <brocade/silkworm_bhpc.h>

#include <asm/prom.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/tlb.h>
#include <asm/delay.h>
#include <asm/bitops.h>
#include <asm/time.h>
#include <asm/uaccess.h>

#ifdef DEBUG
#define DBGPRINT(x...) printk(x)
#else
#define DBGPRINT(x...)
#endif

MODULE_LICENSE("GPL");

extern raw_spinlock_t regs_lock;
extern struct pt_regs* regs_table[];

extern void silkworm_machine_check_control(int enable, int threshold);

struct pci_dev* lookup_mm_dsp(int domain, u32 bus, u32 dev);

static void show_bridge_bus_window(struct pci_dev *dev);
static void show_bridge_mem_window(struct pci_dev *dev);
static void clear_pcie_errors_bridges(struct pci_dev *dev, const char *name,
									  bhpc_dev_error *deverr, int direct);
static void collect_clear_pcie_errors_pci_bridges(struct pci_dev *dev,
												  const char *name, 
												  bhpc_dev_error *deverr);
static void collect_clear_pci_errors_ntb(struct pci_dev *dev, const char *name,
										 bhpc_dev_error *deverr);

static void (*bhpc_error_notify)(bhpc_error_data *ddata);

/* Domain-specific info needed for multiple domain support */

#define BHPC_MAX_PCI_DOMAIN 4

typedef struct _domain_tbl {
	int domain;		/* PCI controller domain */
	int err_irq;	/* Virtual IRQ for this domain */
	volatile void __iomem *pex_base_addr;	/* Controller registers */
	struct pci_dev *rc_dev;	/* Root complex for this domain */

} domain_tbl_t;

/* Read-only after init */
static domain_tbl_t domain_tbl[BHPC_MAX_PCI_DOMAIN];

/*
 * Proc file system controls
 */
static int irqs_enabled = 0;
static int verbose = 0;


#define DISPLAY_ALL			0

#define PCIE_OFFSET(x)		(pex_base_addr + (x))

#define PCIE_ERR_DTCT_REG	PCIE_OFFSET(0xE00)
#define PCIE_ERR_ENBL_REG	PCIE_OFFSET(0xE08)
#define PCI_ERR_ATTR_REG	PCI_OFFSET(0xE0C)
#define PCIE_ERR_DSBL_REG	PCIE_OFFSET(0xE10)

#define PCIE_ERR_CAP_STAT	PCIE_OFFSET(0xE20)
#define PCIE_ERR_CAP_R0		PCIE_OFFSET(0xE28)
#define PCIE_ERR_CAP_R1		PCIE_OFFSET(0xE2C)
#define PCIE_ERR_CAP_R2		PCIE_OFFSET(0xE30)
#define PCIE_ERR_CAP_R3		PCIE_OFFSET(0xE34)

#define MEM_READ_CMD		0x6000
#define CFG_READ_CMD		0xA000





static dsp_dev_map *devices = NULL;
static bhpc_dsp_map_table *default_dsp_table = NULL;

#if defined(CONFIG_SILKWORM_MM) || defined(CONFIG_SILKWORM_PB)

static bhpc_dsp_map_table  dsp_table = {
	.base_bus = 2,
	.max_entries = 10,
	.dsp_map = {

		{0, 2, 1, 12, 21, 0, 0, 0, NULL},
		{0, 2, 2, 22, 31, 0, 0, 3, NULL},
		{0, 2, 5, 32, 41, 0, 0, 0, NULL},
		{0, 2, 7, 42, 51, 0, 0, 0, NULL},
		{0, 2, 9, 52, 61, 0, 0, 0, NULL},
		{0, 2, 10, 62, 71, 0, 0, 4, NULL},
		{0, 2, 12, 72, 81, 0, 0, 5, NULL},
		{0, 2, 14, 82, 91, 0, 0, 0, NULL},

		{0, 1, 0, 0, 0, 0, 0, 0, NULL},
		{0, 0, 0, 0, 0, 0, 0, 0, NULL},
	}
};

#else
static bhpc_dsp_map_table  dsp_table = {
	.base_bus = 2,
	.max_entries = 14,
	.dsp_map = {
		{0, 2, 1, 3, 3, 0, 0, 0, NULL},
		{0, 2, 2, 4, 4, 0, 0, 0, NULL},
		{0, 2, 3, 5, 5, 0, 0, 0, NULL},
		{0, 2, 4, 6, 6, 0, 0, 0, NULL},
		{0, 2, 5, 7, 7, 0, 0, 0, NULL},
		{0, 2, 6, 8, 8, 0, 0, 0, NULL},
		{0, 2, 7, 9, 9, 0, 0, 0, NULL},
		{0, 2, 8, 10, 10, 0, 0, 0, NULL},
		{0, 2, 9, 11, 11, 0, 0, 0, NULL},
		{0, 2, 10, 12, 12, 0, 0, 0, NULL},
		{0, 2, 11, 13, 13, 0, 0, 0, NULL},
		{0, 2, 12, 14, 14, 0, 0, 0, NULL},
		{0, 1, 0, 0, 0, 0, 0, 0, NULL},
		{0, 0, 0, 0, 0, 0, 0, 0, NULL},
	}
};

#endif

static int num_devices_entries = 0;

static u32 config[30];

static char *uncorr_errs[] = {
	"N/A",
	"N/A",
	"N/A",
	"N/A",
	"DLPE", /* Bit 4 */
	"SDE",  /* Bit 5 */
	"N/A",
	"N/A",
	"N/A",
	"N/A",
	"N/A",
	"N/A",
	"PTLP", /* Bit 12 */
	"FCPE", /* Bit 13 */
	"CTO",  /* Bit 14 */
	"CA",   /* Bit 15 */
	"UC",   /* Bit 16 */
	"RXO",  /* Bit 17 */
	"MTLP", /* Bit 18 */
	"ECRCE", /* Bit 19 */
	"URE",  /* Bit 20 */
	"ACSVE", /* Bit 21 */
};

char *bhpc_errdr_message[] = {
	"ERRDR",
	"PCT",
	"PAT",
	"PCAC",
	"LDDE",
	"PNM"
};
EXPORT_SYMBOL(bhpc_errdr_message);

char *bhpc_pme_message[] = {
	"PME",
	"PTO",
	"ENL23",
	"EXL23",
	"HRD",
	"LDD",
	"LUD"
};
EXPORT_SYMBOL(bhpc_pme_message);

char *bhpc_rc_message[] = {
	"UNKNOWN",
	"PAERR",
	"STA",
	"RTA",
	"RMA",
	"SSERR",
	"DPAERR",
	"SDE",
	"CTO",
};
EXPORT_SYMBOL(bhpc_rc_message);


static int show_pex_error_regs(volatile void __iomem *vaddr, int domain, int verbose);
static void show_configs_rc(struct pci_dev *rdev, char *name, int verbose);
static int get_port_type(struct pci_dev *dev);
static int decode_bus_err_src(u32 src_ids, bhpc_error_data *ddata);
static int decode_mem_err_src (u32 memaddr, bhpc_error_data *ddata);
static int decode_pci_config1_err_src(u32 src_ids, bhpc_error_data *ddata);
static int decode_pci_config0_err_src(u32 ntb_cfg_addr, u32 bus_cfg_addr,
									  bhpc_error_data *ddata);

/**************************************************/
/*********** START OF DISPLAY ROUTINES ************/
/**************************************************/

/*
 * For display purposes only
 *
 * This finds and displays the MM PCIe switch port hierarchy
 * for a specified SFM device. The MM DSP and USP are all that's
 * shown.
 *
 * These are the devices that will need to have error status
 * cleared.
 */
static void
show_hierarchy(int domain, u32 bus, u32 dev, u32 func)
{
	struct pci_dev *pdev;

	printk("%s: domain %d bus %d dev %d\n", __FUNCTION__,
		   domain, bus, dev);

	pdev = lookup_mm_dsp(domain, bus, dev);
	if (pdev) {

		printk("Source device hierarchy %s ", pci_name(pdev));
		if ((pdev->bus)&&(pdev->bus->self)) {
			pdev = pdev->bus->self;
			printk("==> %s ", pci_name(pdev));

			if ((pdev->bus)&&(pdev->bus->self)) {
				pdev = pdev->bus->self;
				printk("==> %s", pci_name(pdev));
			}
		}
		printk("\n");
	}
}
/*
 * For display purposes only
 *
 * Decode config access packet.
 *
 */
static void display_decoded_config(u32 type, u32 fmt, u32 cap_r0,
								u32 cap_r1, u32 cap_r2, u32 cap_r3)
{


	switch(type) {

	case 4: // Type 0

		switch (fmt) {

		case 0x00:
			printk("%s: Config type 0 Read packet cap_r0 0x%8.8x cap_r1 0x%8.8x "
				   "cap_r2 0x%8.8x cap_r3 0x%8.8x\n", __FUNCTION__,
				   cap_r0, cap_r1, cap_r2, cap_r3);
			break;

		case 0x02:
			printk("%s: Config type 0 Write packet cap_r0 0x%8.8x cap_r1 0x%8.8x "
				   "cap_r2 0x%8.8x cap_r3 0x%8.8x\n", __FUNCTION__,
				   cap_r0, cap_r1, cap_r2, cap_r3);

			break;

		default:

			printk("%s: Unrecognized Type 0 config packet type %d fmt %d "
				   "cap_r0 0x%8.8x cap_r1 0x%8.8x cap_r2 0x%8.8x "
				   "cap_r3 0x%8.8x\n", __FUNCTION__, type, fmt,
				   cap_r0, cap_r1, cap_r2, cap_r3);
		}

		break;

	case 5:  // Type 1

		switch(fmt) {

		case 0x00:
			printk("%s: Config type 1 Read packet cap_r0 0x%8.8x cap_r1 0x%8.8x "
				   "cap_r2 0x%8.8x cap_r3 0x%8.8x\n", __FUNCTION__,
				   cap_r0, cap_r1, cap_r2, cap_r3);

			break;

		case 0x02:
			printk("%s: Config type 1 Write packet cap_r0 0x%8.8x cap_r1 0x%8.8x "
				   "cap_r2 0x%8.8x cap_r3 0x%8.8x\n", __FUNCTION__,
				   cap_r0, cap_r1, cap_r2, cap_r3);

			break;

		default:

			printk("%s: Unrecognized Type 1 config packet type %d fmt %d "
				   "cap_r0 0x%8.8x cap_r1 0x%8.8x cap_r2 0x%8.8x "
				   "cap_r3 0x%8.8x\n", __FUNCTION__, type, fmt,
				   cap_r0, cap_r1, cap_r2, cap_r3);
		}
		break;

	default:
		;
	}
}
/*
 * For display purposes only
 *
 * Display ID of error source ID in the
 * Error Source ID register. Also display the
 * device hierarchy from the source device to the RC.
 */
static void
display_src_ids(int domain, u32 src_ids)
{
	u32 sbus = 0, sfunc = 0, sdev = 0;

	if (src_ids & 0x0000FFFF) {

		/* decode the correctable source ID */

		sbus = (src_ids >> 8) & 0xFF;	/* Source bus number: bits 8-15 */
		sfunc = src_ids & 0x07;			/* Source Function: Bits 0-2 */
		sdev = (src_ids >> 3) & 0x1F;	/* Source Device: Bits 3-7 */
		printk("  Correctable Error Source "
			   "device 0x%4.4x:%2.2x.%1.1x  ", sbus, sdev, sfunc);
	}

	if (src_ids & 0xFFFF0000) {

		/* Decode the uncorrectable source ID */

		sbus = (src_ids >> 24) & 0xFF;		/* Source bus number: bits 24-31 */
		sfunc = (src_ids >> 16) & 0x07;		/* Source Function: Bits 16-18 */
		sdev = (src_ids >> 19) & 0x1F;		/* Source Device: Bits 19-23 */
		printk("  Uncorrectable Error Source "
			   "device 0x%4.4x:%2.2x.%1.1x", sbus, sdev, sfunc);

	}
	printk("\n");

	show_hierarchy(domain, sbus, sdev, sfunc);

}
/*
 * For display purposes only
 *
 * Decode a PCIe memory access packet and display the access type,
 * the Requester ID, and the device hierarcy of the requester
 */
static void display_decoded_pcie_memory(int domain, u32 type, u32 fmt,
										u32 cap_r0,	u32 cap_r1, u32 cap_r2,
										u32 cap_r3)
{
	u32 rbus, rfunc, rdev;

	rbus = (cap_r1) & 0xFF;			/* Requester bus number: bits 24-31 */
	rfunc = (cap_r1 >> 8) & 0x07;	/* Requester Function: Bits 21-23 */
	rdev = (cap_r1 >> 11) & 0x1F;	/* requester Device: Bits 16-20 */

	switch (fmt) {

	case 0x0: /* Mem Read w/o lock */
	case 0x1: /* Mem Read with lock */

		printk("%s: Mem Read Req packet from bus %d dev %d "
			   "func %d\n", __FUNCTION__, rbus, rdev, rfunc);
		break;

	case 0x2:
	case 0x3:
		printk("%s: Mem Write Req packet from bus %d dev %d "
			   "func %d\n", __FUNCTION__, rbus, rdev, rfunc);
		break;

	default:
		printk("%s Unknown Mem packet fmt 0x%x bus %d dev %d "
			   "func %d\n", __FUNCTION__, fmt, rbus, rdev, rfunc);
	}

	show_hierarchy(domain, rbus, rdev, rfunc);

}
/*
 * For display purposes only
 *
 * Decodes a PCIE completion packet and displays the
 * completer ID, the requester ID, and the completion
 * status
 */

static void display_decoded_pcie_completion(int domain, u32 type, u32 fmt, u32 cap_r0,
								u32 cap_r1, u32 cap_r2, u32 cap_r3)
{

	char *cs, *fs;
	u32 cbus, cfunc, cdev, status, rbus, rfunc, rdev;

	cbus = (cap_r1) & 0xFF;			/* Completer bus number: bits 24-31 */
	cfunc = (cap_r1 >> 8) & 0x07;	/* Completer Function: Bits 21-23 */
	cdev = (cap_r1 >> 11) & 0x1F;	/* Completer Device: Bits 16-20 */
	status = (cap_r1 >> 21) & 0x07; /* Completion status: Bits 8-10 */

	rbus = (cap_r2) & 0xFF;			/* Requester bus number: bits 24-31 */
	rfunc = (cap_r2 >> 8) & 0x07;	/* Requester Function: Bits 21-23 */
	rdev = (cap_r2 >> 11) & 0x1F;	/* Requester Device: Bits 16-20 */

	// status = (cap_r2 >> 21) & 0x03; /* Completion status: Bits 8-10 */

	switch (status) {
	case 0:
		cs = "SC";	/* Successful Completion */
		break;
	case 1:
		cs = "UR";	/* Unsupported Request */
		break;
	case 2:
		cs = "CRS";	/* Configuration Retry Request */
		break;
	case 4:
		cs = "CA";	/* Completer abort */
		break;
	default:
		cs = "Unknown";
	}

	switch (fmt) {
	case 0:
		fs = "No Data";
		break;
	case 1:
		fs = "With Data";
		break;
	default:
		fs = "Unknown";
	}

	printk("%s: Completion packet [%s] fmt [%s] from bus %d dev %d "
		   "func %d\n", __FUNCTION__, cs, fs, cbus, cdev, cfunc);

	printk("%s: Requester bus %d dev %d func %d\n",
		   __FUNCTION__, rbus, rdev, rfunc);

	show_hierarchy(domain, cbus, cdev, cfunc);
}
/*
 * For display purposes only
 *
 * Decodes the error capture status when transaction
 * originator is the PEX_CONFIG_ADDR/PEX_CONFIG_DATA registers.
 */
static void display_decoded_pex_config_errors(u32 cap_status, u32 cap_r0,
											  u32 cap_r1, u32 cap_r2, u32 cap_r3)
{
	u32 type, fmt;

	/*
	 * Outbound case (error is due to timeout or config_addr/config_data access.
	 */
	type = (cap_r0 >> 9) & 0x1F;		/* Packet type: bits 18-22 */
	fmt = (cap_r0 >> 14) & 0x03;	/* Packet format: bits 16-17 */

	switch (type) {

	case 0x4: /* Config type 0 */

				switch (fmt) {
				case 0:
					printk("  Config Type0 Read Request\n");
					break;

				case 2:
					printk("  Config Type0 Write request\n");
					break;

				default:
					printk("  Undecoded Config type 0 request type %d fmt %d cap_r0 0x%x\n",
						   type, fmt, cap_r0);
				}
				break;

	case 0x5: /* Config type 1 */
				switch (fmt) {
				case 0:
					printk("  Config Type1 Read Request\n");
					break;

				case 2:
					printk("  Config Type1 Write request\n");
					break;

				default:
					printk("  Undecoded Config type 1 request type %d fmt %d cap_r0 0x%x\n",
						   type, fmt, cap_r0);
				}

				break;

	default:
		printk("  Undecoded Config Outbound packet type 0x%x "
			   "fmt 0x%x\n", type, fmt);
	}


}
/*
 * For display purposes only
 *
 * Decodes the capture status register to determine what
 * type of packet was captured, then calls the appropriate
 * routine to decode that packet.
 *
 * Not all types are supported.
 */
static void display_decoded_pex_errors(int domain, u32 err_dr, u32 cap_status, u32 cap_r0,
								u32 cap_r1, u32 cap_r2, u32 cap_r3)
{
	int pex_config, gsid, inbound;
	u32 type, fmt;

	/* Make sure it's valid status */
	if ((cap_status & 1) == 0) {
		goto Exit;
	}

	pex_config = (cap_status & 0x80000000);

	if (pex_config) {
		display_decoded_pex_config_errors(cap_status, cap_r0, cap_r1,
										  cap_r2, cap_r3);
	} else {

		gsid = cap_status >> 1;

		/* Only looking at PCIe 0 controller for now */
		inbound = (gsid == 0x02);

		/*
		 * Inbound transactions that result in PNM, PCAC, CDNSC or
		 * CRSNC set will result in GSID == 0x02
		 */

		if (inbound) {

			/* For inbound, cap_r1 has the requester ID */

			type = (cap_r0 & 0x1F);		/* Packet type: bits 27-31 */
			fmt = (cap_r0 >> 5) & 0x03;	/* Packet format: bits 25-26 */

			switch (type) {

			case 0x0A:  /* Completion w/o lock packet */
			case 0x0B: /* Completion with lock packet */
				display_decoded_pcie_completion(domain, type, fmt, cap_r0, cap_r1,
												cap_r2, cap_r3);

				break;


			case 0x00000: /* Mem Read/Write */
			case 0x00001: /* Mem Read Lock */

				display_decoded_pcie_memory(domain, type, fmt, cap_r0, cap_r1,
											cap_r2, cap_r3);

				break;

			case 0x4: /* Config type 0 */
			case 0x5: /* Config type 1 */
				display_decoded_config(type, fmt,  cap_r0, cap_r1, cap_r2,
									   cap_r3);

				break;

			default:

				printk("%s Undecoded inbound packet type 0x%x "
					   "fmt 0x%x\n", __FUNCTION__, type, fmt);
			}

		} else {

			/*
			 * Outbounds have capture status only for timeouts or
			 * config accesses
			 */
			printk("%s %d: Undecoded output packet type\n",
				   __FUNCTION__, __LINE__);

		}
	}

 Exit:
	return;

}
/*
 * For display purposes only
 *
 * Decode & display the contents of the PCIe controller pme message detect register
 */
static void
display_pme_events(u32 detect)
{

	if (detect) {
		printk("%s: PME_DR 0x%8.8x ", __FUNCTION__, detect);

		if (detect & 0x8000) {
			printk(" [PTO]");
		}
		if (detect & 0x2000) {
			printk(" [ENL23]");
		}
		if (detect & 0x1000) {
			printk(" [EXL23]");
		}
		if (detect & 0x0400) {
			printk(" [HRD]");
		}
		if (detect & 0x0200) {
			printk(" [LDD]");
		}
		if (detect & 0x0080) {
			printk(" [LUD]");
		}
		printk("\n");

	}

}
/*
 *
 * Create ID strings for the bits in the PCIe controller
 * Error Detect Register.
 */
static void
decode_err_detect(bhpc_error_data *ddata)
{

	bhpc_pcie_controller *pex;

	if (ddata) {

		pex = &ddata->pex;
		/*
		 * Start with default ERR DR message and over-ride with a more
		 * specific message for the most likely to occur.
		 */
		ddata->error_idx = BHPC_ERRDR_IDX;
		printk("%s: ERR_DR 0x%8.8x ", __FUNCTION__, pex->detect);

		if (pex->detect & 0x80000000) {
			printk(" [ME]");
		}
		if (pex->detect & 0x00800000) {
			ddata->error_idx = BHPC_ERRDR_PCT_IDX;
			printk(" [PCT]");
		}
		if (pex->detect & 0x00400000) {
			ddata->error_idx = BHPC_ERRDR_PAT_IDX;
			printk(" [PAT]");
		}
		if (pex->detect & 0x00200000) {
			ddata->error_idx = BHPC_ERRDR_PCAC_IDX;
			printk(" [PCAC]");
		}
		if (pex->detect & 0x00100000) {
			ddata->error_idx = BHPC_ERRDR_PNM_IDX;
			printk(" [PNM]");
		}
		if (pex->detect & 0x00080000) {
			printk(" [CDNSC]");
		}
		if (pex->detect & 0x00040000) {
			printk(" [CRSNC]");
		}
		if (pex->detect & 0x00020000) {
			printk(" [ICCA]");
		}
		if (pex->detect & 0x00010000) {
			printk(" [IACA]");
		}
		if (pex->detect & 0x00008000) {
			printk(" [CRST]");
		}
		if (pex->detect & 0x00004000) {
			printk(" [MIS]");
		}
		if (pex->detect & 0x00002000) {
			printk(" [IOIS]");
		}
		if (pex->detect & 0x00001000) {
			printk(" [CIS]");
		}
		if (pex->detect & 0x00000800) {
			printk(" [CIEP]");
		}
		if (pex->detect & 0x00000400) {
			printk(" [IOIEP]");
		}
		if (pex->detect & 0x00000200) {
			printk(" [OAC]");
		}
		if (pex->detect & 0x00000100) {
			printk(" [IOIA]");
		}
		if (pex->detect & 0x00000080) {
			printk(" [IMBA]");
		}
		if (pex->detect & 0x00000040) {
			printk(" [IIOBA]");
		}
		if (pex->detect & 0x00000020) {
			ddata->error_idx = BHPC_ERRDR_LDDE_IDX;
			printk(" [LDDE]");
		}
		printk("\n");
	}
}
/*
 * For display purposes only
 *
 * Displays the contents of the PCIe controller error detect
 * registers.
 */
static void
display_pex_errors(int domain, u32 err_dr, u32 cap_status, u32 cap_r0,
								u32 cap_r1, u32 cap_r2, u32 cap_r3)
{

	if (err_dr) {
		printk("%s: ERR_DR 0x%8.8x\n", __FUNCTION__, err_dr);
		display_decoded_pex_errors(domain, err_dr, cap_status, cap_r0, cap_r1, cap_r2, cap_r3);
	}

}
/*
 * For display purposes only (so far)
 *
 * Display the header log contents. Their meaning varies depending
 * on what happened, so just display the values for now.
 */
static void
display_decoded_header_log(const char *name, bhpc_dev_error *deverr)
{
	u32 *log;

	if (deverr) {

		log = &deverr->header_log[0];

		printk("%s: %s Log0 0x%8.8x log1 0x%8.8x log2 0x%8.8x "
			   "log3 0x%8.8x\n", __FUNCTION__, name, 
			   log[0], log[1], log[2], log[3]);
	}
}

/*
 * Create ID strings for Device Status
 * register
 */
static void
print_devstatus_errors(u32 bitmap)
{

	if (bitmap & 0x00080000) {
		printk(" [UnsuppReq]");
	}
	if (bitmap & 0x00040000) {
		printk(" [FatalErrDet]");
	}
	if (bitmap & 0x00020000) {
		printk(" [NonFatalErrDet]");
	}
	if (bitmap & 0x00010000) {
		printk(" [CorrErrDet]");
	}
	printk("\n");
}

/*
 * Create ID strings for Root Error Status
 * Register
 */
static void
print_pci_root_errors(u32 bitmap)
{

	if (bitmap & 0x00000001) {
		printk(" [ECR]");
	}
	if (bitmap & 0x00000002) {
		printk(" [MECR]");
	}
	if (bitmap & 0x00000004) {
		printk(" [EFNR]");
	}
	if (bitmap & 0x00000008) {
		printk(" [MEFNR]");
	}
	if (bitmap & 0x00000010) {
		printk(" [FUF]");
	}
	if (bitmap & 0x00000020) {
		printk(" [NFEMR]");
	}
	if (bitmap & 0x00000040) {
		printk(" [FEMR]");
	}

	printk("\n");
}

/*
 * Create ID strings for Slot Status
 * Register. Only the upper 16 bits matter
 */
static void
print_pci_slot_errors(u32 bitmap)
{

	bitmap = bitmap >> 16;

	if (bitmap & 0x00000001) {
		printk(" [ABP]");
	}
	if (bitmap & 0x00000002) {
		printk(" [PFD]");
	}
	if (bitmap & 0x00000004) {
		printk(" [MRLSC]");
	}
	if (bitmap & 0x00000008) {
		printk(" [PDC]");
	}
	if (bitmap & 0x00000010) {
		printk(" [CC]");
	}

	printk("\n");
}
/*
 * Create ID string for error bits in Primary and
 * Secondary status registers
 */
static void
print_pci_errors(u32 bitmap)
{

	if (bitmap & 0x80000000) {
		printk(" [DPA]");
	}
	if (bitmap & 0x40000000) {
		printk(" [R/SSE]");
	}
	if (bitmap & 0x20000000) {
		printk(" [RMA]");
	}
	if (bitmap & 0x10000000) {
		printk(" [RTA]");
	}
	if (bitmap & 0x08000000) {
		printk(" [STA]");
	}
	if (bitmap & 0x01000000) {
		printk(" [MDPE]");
	}
	printk("\n");
}
/*
 * Create ID string for error bits in the Correctable
 * Error Status register
 */
static void
print_pci_correctable(u32 bitmap)
{
	if (bitmap & 0x00002000) { // Bit 13
		printk(" [ANFE]");
	}
	if (bitmap & 0x00001000) { // Bit 12
		printk(" [RTTO]");
	}
	if (bitmap & 0x00000100) { // Bit 8
		printk(" [RNR]");
	}
	if (bitmap & 0x00000080) { // Bit 7
		printk(" [BDLLP]");
	}
	if (bitmap & 0x00000040) { // Bit 6
		printk(" [BTLP]");
	}
	if (bitmap & 0x00000001) { // Bit 0
		printk(" [RXE]");
	}
	printk("\n");
}

/*
 * Create ID string for error bits in the Uncorrectable
 * Error Status register
 */
static void
print_pci_uncorrectable(u32 bitmap)
{

	if (bitmap & 0x00200000) { // Bit 21
		printk(" [ACSVE]");
	}

	if (bitmap & 0x00100000) { // Bit 20
		printk(" [URE]");
	}

	if (bitmap & 0x00080000) { // Bit 19
		printk(" [ECRCE]");
	}
	if (bitmap & 0x00040000) { // Bit 18
		printk(" [MTLP]");
	}
	if (bitmap & 0x00020000) { // Bit 17
		printk(" [RXO]");
	}
	if (bitmap & 0x00010000) { // Bit 16
		printk(" [UC]");
	}
	if (bitmap & 0x00008000) { // Bit 15
		printk(" [CA]");
	}
	if (bitmap & 0x00004000) { // Bit 14
		printk(" [CTO]");
	}
	if (bitmap & 0x00002000) { // Bit 13
		printk(" [FCPE]");
	}
	if (bitmap & 0x00001000) { // Bit 12
		printk(" [PTLP]");
	}
	if (bitmap & 0x00000020) { // Bit 5 Surprise Down Error
		printk(" [SDE]");
	}
	if (bitmap & 0x00000010) { // Bit 4
		printk(" [DLPE]");
	}
	printk("\n");
}
/*
 * Create ID string for error bits in the PCIE-PCI
 * bridge secondary Uncorrectable Error Status register
 */
static void
print_pci_sec_uncorrectable(u32 bitmap)
{

	if (bitmap & 0x00002000) { // Bit 13 Internal Bridge Error
		printk(" [IBERR]");
	}
	if (bitmap & 0x00001000) { // Bit 12 PERR Assertion detected
		printk(" [SERR]");
	}
	if (bitmap & 0x00000800) { // Bit 11 PERR Assertion detected
		printk(" [PERR]");
	}
	if (bitmap & 0x00000400) { // Bit 10 Delayed Transaction Discard Timer Expired
		printk(" [DTDTE]");
	}
	if (bitmap & 0x00000200) { // Bit 9 Ucorrectable Address Error
		printk(" [UCADDRE]");
	}
	if (bitmap & 0x00000100) { // Bit 8 Ucorrectable Attribute Error
		printk(" [UCATTRE]");
	}
	if (bitmap & 0x00000080) { // Bit 7 Ucorrectable Split Completion Error
		printk(" [UCSCE]");
	}
	if (bitmap & 0x00000040) { // Bit 6 Ucorrectable Split Completion Error
		printk(" [UCSCE]");
	}
	if (bitmap & 0x00000020) { // Bit 5 Unexpected Split Completion Error
		printk(" [UNSCE]");
	}
	if (bitmap & 0x00000008) { // Bit 3 Received Master Abort
		printk(" [RMA]");
	}
	if (bitmap & 0x00000004) { // Bit 2 Received Target Abort
		printk(" [RTA]");
	}
	if (bitmap & 0x00000002) { // Bit 1 Master Abort on Split Completion
		printk(" [MAOSC]");
	}
	if (bitmap & 0x00000001) { // Bit 0 Target Abort on Split Completion
		printk(" [TAOSC]");
	}
	printk("\n");
}

/**************************************************/
/********* END OF DISPLAY ROUTINES ****************/
/**************************************************/
/*
 * Find chassis slot number that contains a given device based on 
 * the domain, bus and device numbers.
 */
static int
get_slot_number(int domain, u32 bus, u32 dev, bhpc_error_data *ddata)
{
	u32 i = 0;
	int rc = -1;
	struct pci_dev *pdev = NULL;

	if (!ddata) {
		printk(KERN_ERR "%s:Invalid input pointer\n", __FUNCTION__);
		return (rc);
	}

#if defined (CONFIG_SILKWORM_MM) 
	for (i = 0; i < num_devices_entries; i++) {
		if ((devices[i].domain == domain)&&
			(devices[i].bus == bus)&&
			(devices[i].dev == dev)) {

			/* exact match means the given device is a DSP */

			ddata->slot = devices[i].slot;
			rc = 0;
			break;

		} else if ((devices[i].domain == domain)&&
				   (devices[i].first_bus <= bus)&&
				   (devices[i].last_bus >= bus)) {

			/*
			 * Found the DSP that parents the given device.
			 * If this DSP has subranges, then search them
			 * to find the exact slot.
			 */
			pdev = devices[i].pdev;
			if (devices[i].bus_subranges.number > 0) {
				dsp_subranges *ptr = NULL;
				int x;

				ptr = &devices[i].bus_subranges;
				for (x = 0; (x < ptr->number)&&
						 (x < BHPC_MAX_DSP_SUBRANGES); x++) {

					if ((bus >= ptr->subranges[x].base)&&
						(bus <= ptr->subranges[x].limit)) {
						if (verbose) {
							printk(KERN_INFO "%s x %d: Found subrange "
								   "Dev %s Base 0x%8.8x Limit 0x%8.8x "
								   "slot %d\n", __FUNCTION__, x, pci_name(pdev),
								   ptr->subranges[x].base,
								   ptr->subranges[x].limit,
								   ptr->subranges[x].slot);
						}
						ddata->slot = ptr->subranges[x].slot;
					}
				}
			} else {
				/*
				 * This means the DSP supports a range of busses,
				 * but has no subrange table. This should not happen
				 * but if it does, then just go with the DSP slot
				 */
				ddata->slot = devices[i].slot;
			}

			rc = 0;
			break;
		}
	}
#else /* For LCs, slot is always 0 */

	ddata->slot = 0;
	rc = 0;

#endif

	return (rc);
}

/*
 * void store_dev()
 *
 * Store RC and PCIe switch port information for use by
 * PCIe error interrupt handler.
 */
static void
store_dev(int domain, u32 bus, u32 dev, u32 type, struct pci_dev *pdev)
{
	u32 i, regval;

	for (i = 0; i < num_devices_entries; i++) {
		if ((devices[i].domain == domain)&&
			(devices[i].bus == bus)&&
			(devices[i].dev == dev)) {

			devices[i].pdev = pdev;

			if (type != PCI_TYPE_NTB) {
				/*
				 * Store the memory window provided by each bus device.
				 * Ignore NTBs as their window is same as their parent
				 * device.
				 */
				pci_read_config_dword(pdev, PCI_MEMORY_BASE, &regval);
				devices[i].mem_base = ((regval &0xFFFF) << 16);
				devices[i].mem_limit = regval | 0xFFFFF;
			}
			if (verbose) {
				printk(KERN_ERR "%s: i %d domain %d bus %d dev %d "
					   "mem_base 0x%8.8x mem_limit 0x%8.8x name %s\n",
					   __FUNCTION__, i, domain, bus, dev, devices[i].mem_base,
					   devices[i].mem_limit, pci_name(pdev));
			}

			break;
		}
	}

	if (i == num_devices_entries) {
		printk(KERN_INFO "%s: Could not store pci dev for domain  %d "
			   "bus %d dev %d\n",
			   __FUNCTION__, domain, bus, dev);
	}
}

/*
 * lookup_mm_dsp()
 *
 * For a given PCI device (domain:bus:dev), determine the MM PCIe switch
 * downstream port that connects to the blade containing that device.
 *
 *    Find the DSP port whose bus range contains the bus of the target
 *    PCI device
 */
struct pci_dev*
lookup_mm_dsp(int domain, u32 bus, u32 dev)
{
	int i;
	struct pci_dev *pdev = NULL;

	if (verbose) {
		printk("%s: Find MM dsp for domain %d bus %d dev %d\n",
			   __FUNCTION__, domain, bus, dev);
	}

	if (bus > default_dsp_table->base_bus) {

		/* The bus is on a blade */
		for (i = 0; i < num_devices_entries; i++) {
			if ((devices[i].domain == domain)&&
				(devices[i].first_bus <= bus)&&
				(devices[i].last_bus >= bus)) {
				bus = devices[i].bus;
				dev = devices[i].dev;
				break;
			}
		}
	}

	/* The bus is on the MM */
	for (i = 0; i < num_devices_entries; i++) {
		if ((devices[i].domain == domain)&&
			(devices[i].bus == bus)&&
			(devices[i].dev == dev)) {

			pdev = devices[i].pdev;
			break;
		}
	}

	if (!pdev) {
		if (verbose) {
		printk(KERN_ERR "%s %d: Could not find pci dev for domain  %d bus %d dev %d "
			   "Table size %d\n",
				   __FUNCTION__, i, domain, bus, dev, num_devices_entries);
		}
	} else if (verbose) {
		printk("%s: MM DSP found: %s\n", __FUNCTION__, pci_name(pdev));
	}

	return (pdev);
}
/*
 * int decode_header_log()
 *
 * Decode the ID of an error source device from the captured
 * AER header log. See PCIe specs for format of the header log.
 *
 */
static int
decode_header_log(bhpc_dev_error *deverr, bhpc_error_data *ddata)
{

	u32 type, fmt, src_ids = 0;
	int src_found = 0;

	type = ((deverr->header_log[0] >> 24) & 0x1F);
	fmt = ((deverr->header_log[0] >> 29) & 0x03);

	switch (type) {

	case 0x0A:  /* Completion w/o lock packet */
	case 0x0B: /* Completion with lock packet */

		src_ids = ((deverr->header_log[1] >> 16) & 0xFFFF);
		src_found = decode_bus_err_src(src_ids, ddata);
		break;


	case 0x00000: /* mem read/write */
	case 0x00001: /* mem read lock */

		src_ids = ((deverr->header_log[1] >> 16) & 0xFFFF);
		if (src_ids & 0xFFF8) {

			/* For inbound mem access, we want the Requester ID */
			src_found = decode_bus_err_src(src_ids, ddata);
		} else {
			/* For outbound mem access, we want the mem address */
			src_ids = deverr->header_log[2];
			src_found = decode_mem_err_src(src_ids, ddata);
		}
		break;

	case 0x4: /* config type 0 */
	case 0x5: /* config type 1 */

		/* for configs, we want the target device, not requester */
		src_ids = ((deverr->header_log[2] >> 16) & 0xFFFF);
		src_found = decode_bus_err_src(src_ids, ddata);
		break;

	default:

		if (type & 0x10) { /* Message Requests 0x10xxx */
			src_ids = ((deverr->header_log[1] >> 16) & 0xFFFF);
			src_found = decode_bus_err_src(src_ids, ddata);
		} else {
			printk("%s Undecoded packet type 0x%x "
				   "fmt 0x%x\n", __FUNCTION__, type, fmt);
		}
	}

	return (src_found);
}

/*
 * Decode a memory address and determine what slot it belongs to
 * by comparing the address with the previously-stored memory window
 * provided by the PCIe switch DSPs
 *
 * The memory window of a DSP that connects to PCI slots is divided into
 * a subrange per slot, so the subrange table (if present) also has to
 * searched.
 *
 * Inputs:
 *   memaddr - Memory address
 *   ddata   - Contains all information captured/decoded so far
 *
 * Returns:
 *  1 - Found the source device that owns the memory address
 *  0 - Could not determine the device
 */
static int
decode_mem_err_src (u32 memaddr, bhpc_error_data *ddata)
{
	struct pci_dev *dev = NULL;
	int i, src_found = 0, dev_domain;

	for (i = 0; i < num_devices_entries; i++) {

		dev = devices[i].pdev;
		if (dev) {

			dev_domain =  pci_domain_nr(dev->bus);
			if (ddata->domain != dev_domain) {
				continue;
			}
			if ((memaddr >= devices[i].mem_base)&&
				(memaddr <= devices[i].mem_limit)) {

				if (verbose) {
					printk(KERN_ERR "%s i %d: Found Dev %s Base 0x%8.8x "
						   "Limit 0x%8.8x memaddr 0x%x\n",
						   __FUNCTION__, i, pci_name(dev), devices[i].mem_base,
						   devices[i].mem_limit, memaddr);
				}
				/*
				 * If this DSP has subranges, then search them
				 * to find the exact slot.
				 */
				if (devices[i].mem_subranges.number > 0) {
					dsp_subranges *ptr = NULL;
					int x;

					ptr = &devices[i].mem_subranges;

					for (x = 0; (x < ptr->number)&&
							 ( x < BHPC_MAX_DSP_SUBRANGES); x++) {
						if ((memaddr >= ptr->subranges[x].base)&&
							(memaddr <= ptr->subranges[x].limit)) {

							if (verbose) {
								printk(KERN_ERR "%s x %d: Found subrange "
									   "Dev %s Base 0x%8.8x Limit 0x%8.8x "
									   "slot %d\n", __FUNCTION__, x,
									   pci_name(dev),
									   ptr->subranges[x].base,
									   ptr->subranges[x].limit,
									   ptr->subranges[x].slot);
							}
							ddata->slot = ptr->subranges[x].slot;
						}
					}
				} else {
					ddata->slot = devices[i].slot;
				}

				ddata->mm_dsp_dev = dev;
				ddata->mm_parent_dev.dev = PCI_SLOT(dev->devfn);
				ddata->mm_parent_dev.bus = dev->bus->number;

				ddata->src_dev.dev = 0;
				ddata->src_dev.bus = devices[i].first_bus;
				ddata->src_dev.func = 0;
				src_found = 1;
				break;

			} /* if ((memaddr >= devices[i].mem_base) ... */

		} /* if (dev) */

	} /* for ... */

	return (src_found);
}
/*
 * Decodes an encoded Source ID and then finds the MM DSP that
 * connects to the blade on which the Source device resides.
 *
 * The bus:dev:func ID of the Source device and the MM DSP device
 * is passed back in the ddata structure.
 *
 * Returns:
 *    0 - Source device not found
 *    1 - Source device found
 */
static int
decode_bus_err_src(u32 src_ids, bhpc_error_data *ddata)
{

	u32 sbus, sfunc, sdev;
	struct pci_dev *pdev;
	int src_found = 0;

	sbus = (src_ids >> 8) & 0xFF;	/* Source bus number: bits 8-15 */
	sfunc = src_ids & 0x07;			/* Source Function: Bits 0-2 */
	sdev = (src_ids >> 3) & 0x1F;	/* Source Device: Bits 3-7 */

	if (verbose) {
		printk(KERN_ERR "%s:  Error Source "
			   "device 0x%4.4x:%2.2x.%1.1x\n", __FUNCTION__, sbus, sdev, sfunc);
	}

	/* Now find the MM DSP in the hierarchy of this device */
	if ((sbus == 0)&&(sdev == 0)) {
		pdev = domain_tbl[ddata->domain].rc_dev;
	} else {
		pdev = lookup_mm_dsp(ddata->domain, sbus, sdev);
	}
	if (ddata) {
		if (pdev) {
			int rc = 0;

			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;

			rc = get_slot_number(ddata->domain, sbus, sdev, ddata);
			if ((!rc)&&(verbose)) {
				printk(KERN_ERR "%s Found slot %d for MM Parent device "
					   "bus %d dev %d\n", __FUNCTION__, ddata->slot,
					   ddata->mm_parent_dev.bus,
					   ddata->mm_parent_dev.dev);
			}			
		}

		ddata->src_dev.dev = sdev;
		ddata->src_dev.bus = sbus;
		ddata->src_dev.func = sfunc;
		src_found = 1;
	}

	return (src_found);
}
/*
 * Clear error status in Non-Transparent Bridge devices.
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
clear_pci_errors_ntb(struct pci_dev *dev, const char *name,
							bhpc_dev_error *deverr, int direct)
{
	u32 regval = 0;

	if (dev) {
		pci_read_config_dword(dev, PCI_COMMAND, &regval); /* Primary status */
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Primary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, PCI_COMMAND, regval);

		}

		/* Secondary status */
		pci_read_config_dword(dev, PCI_NTB_SECONDARY_STATUS, &regval);
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, PCI_NTB_SECONDARY_STATUS, regval);

		}
		pci_read_config_dword(dev, PCI_NTB_CHIP_STATUS, &regval); /* Chip status */
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Chip status error 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
			}
			pci_write_config_dword(dev, PCI_NTB_CHIP_STATUS, regval);

		}

	}

}
/*
 * Clear error status in endpoint devices.
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
clear_pcie_errors_endpoints(struct pci_dev *dev, const char *name,
							bhpc_dev_error *deverr, int direct)
{
	u32 regval = 0;
	uint pos;

	if (dev) {
		if (direct) {
			pci_read_config_dword(dev, PCI_COMMAND, &regval); /* Primary status */
		} else {
			regval = deverr->primary_status;
		}
		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: %s  Primary status error 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, PCI_COMMAND, regval);

		}
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {
			/* Device status */
			if (direct) {
				pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &regval);
			} else {
				regval = deverr->device_status;
			}
			if (regval & 0x000F0000) {
				if (verbose) {
					printk("%s: %s Device status error 0x%8.8x detected: ",
						   __FUNCTION__, name, regval);
					print_devstatus_errors(regval);
				}
				pci_write_config_dword(dev, pos + PCI_EXP_DEVCTL, regval);
			}
		}
	}
}
/*
 * Clear error status in the RC
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
clear_pcie_errors_rc(struct pci_dev *rdev, const char *name,
					 bhpc_dev_error *deverr, int direct)
{
	int pos;
	u32 regval;

	if (rdev) {

		if (direct) {
			pci_read_config_dword(rdev, PCI_COMMAND, &regval);
		} else {
			regval = deverr->primary_status;
		}

		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: RC Primary status error 0x%8.8x detected: ",
					   __FUNCTION__, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(rdev, PCI_COMMAND, regval);

		}

		/* Secondary status is in upper 16 bits (0x1E) */
		if (direct) {
			pci_read_config_dword(rdev, PCI_IO_BASE, &regval);
		} else {
			regval = deverr->secondary_status;
		}

		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: RC Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(rdev, PCI_IO_BASE, regval);

		}

		pos = pci_find_capability(rdev, PCI_CAP_ID_EXP);
		if (pos) {

			if (direct) {
				pci_read_config_dword(rdev, pos + PCI_EXP_DEVCTL, &regval);
			} else {
				regval = deverr->device_status;
			}

			/* Device status */
			if (regval & 0x000F0000) {
				if (verbose) {
					printk("%s: %s Device status error 0x%8.8x detected: ",
						   __FUNCTION__, name, regval);
					print_devstatus_errors(regval);
				}
				pci_write_config_dword(rdev, pos + PCI_EXP_DEVCTL, regval);
			}
		}

		pos = pci_find_ext_capability(rdev, PCI_EXT_CAP_ID_ERR);
		if (pos) {
			/* Uncorr. Err Status */
			if (direct) {
				pci_read_config_dword(rdev, pos + PCI_ERR_UNCOR_STATUS, &regval);
			} else {
				regval = deverr->uncorrectable_status;
			}
			if (regval) {
				if (verbose) {
					printk("%s: RC Uncorr. Err Status error 0x%8.8x "
						   "detected: ", __FUNCTION__, regval);
					print_pci_uncorrectable(regval);
				}

				/* W1C all bits */
				pci_write_config_dword(rdev, pos + PCI_ERR_UNCOR_STATUS, regval);
			}

			/* Corr. Err Status */
			if (direct) {
				pci_read_config_dword(rdev, pos + PCI_ERR_COR_STATUS, &regval);
			} else {
				regval = deverr->correctable_status;
			}
			if (regval) {
				if (verbose) {
					printk("%s: RC Corr. Err Status 0x%8.8x ",
						   __FUNCTION__, regval);
					print_pci_correctable(regval);
				}
				/* W1C all bits */
				pci_write_config_dword(rdev, pos + PCI_ERR_COR_STATUS, regval);
			}

			/* Root Err status */
			if (direct) {
				pci_read_config_dword(rdev, pos + PCI_ERR_ROOT_STATUS, &regval);
			} else {
				regval = deverr->root_err_status;
			}

			if (regval) {
				if (verbose) {
					printk("%s: RC Root Err Status 0x%8.8x ",
						   __FUNCTION__, regval);
					print_pci_root_errors(regval);
				}
				pci_write_config_dword(rdev, pos + PCI_ERR_ROOT_STATUS, regval);
			}

		}
	}
}
/*
 * Clear error status in PCIe-PCI bridges.
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
clear_pcie_errors_pci_bridges(struct pci_dev *dev, const char *name,
						  bhpc_dev_error *deverr, int direct)
{
	u32 regval = 0;
	uint pos = 0;

	/* First, clear the status on the PCIe side of the bridge */
	clear_pcie_errors_bridges(dev, name, deverr, direct);

	/* Now clear the PCI side of the bridge */

	if (dev) {

		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
		if (pos) {

			if (direct) {
				pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_STATUS,
									  &regval);
			} else {
				regval = deverr->sec_uncorrectable_status;
			}

			if (regval) {
				if (verbose) {
					printk(KERN_ERR "%s: %s Sec Uncorr. Error status 0x%8.8x "
						   "detected: ", __FUNCTION__, name, regval);
					print_pci_sec_uncorrectable(regval);
				}
				pci_write_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_STATUS,
									   regval);
			}
		}
	}
}
	
/*
 * Clear error status in bridges. This means upstream or downstream
 * switch ports.
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
clear_pcie_errors_bridges(struct pci_dev *dev, const char *name,
						  bhpc_dev_error *deverr, int direct)
{
	u32 regval = 0;
	uint pos = 0;

	if (dev) {

		if (direct) {
			pci_read_config_dword(dev, 0x04, &regval); /* Primary status */
		} else {
			regval = deverr->primary_status;
		}
		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: %s Primary status error 0x%8.8x detected: ",
					   __FUNCTION__, name,  regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, 0x04, regval);
		}

		if (direct) {
			pci_read_config_dword(dev, 0x1C, &regval); /* Secondary status */
		} else {
			regval = deverr->secondary_status;
		}
		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: %s Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			 /* Secondary status */
			pci_write_config_dword(dev, 0x1C, regval);
		}

		if (direct) {
			pci_read_config_dword(dev, 0x70, &regval); /* Device status */
		} else {
			regval = deverr->device_status;
		}
		if (regval & 0x000F0000) {
			if (verbose) {
				printk("%s: %s Device status error 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
				print_devstatus_errors(regval);
			}
			pci_write_config_dword(dev, 0x70, regval);

		}
		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);

		/* Uncorr. Err Status */

		if (direct) {
			pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, &regval);
		} else {
			regval = deverr->uncorrectable_status;
		}

		if (regval) {
			if (verbose) {
				printk("%s: %s Uncorr. Error status 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
				print_pci_uncorrectable(regval);
			}
			pci_write_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, regval);
		}

		/* Correctable Error Status */
		if (direct) {
			pci_read_config_dword(dev, pos + PCI_ERR_COR_STATUS, &regval);
		} else {
			regval = deverr->correctable_status;
		}

		if (regval) {
			if (verbose) {
				printk("%s: %s Corr. Error status 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
				print_pci_correctable(regval);
			}
			pci_write_config_dword(dev, pos + PCI_ERR_COR_STATUS, regval);

		}
	}

}

/*
 * collect_clear_pcie_errors_bridges()
 *
 *  Read and save the contents of the error control/status
 *  registers of a PCI bridge device. Then clear the status
 *  registers.
 *
 * Inputs:
 *   dev - PCI bridge device
 *   name - String to display when in verbose mode
 *   deverr - table to hold the error status/control registers
 *
 */
static void
collect_clear_pcie_errors_bridges(struct pci_dev *dev, const char *name, 
								  bhpc_dev_error *deverr)
{
	u32 regval = 0;
	uint pos = 0;
	int domain;
	struct pci_dev *rc_dev = NULL;

	domain =  pci_domain_nr(dev->bus);
	rc_dev = domain_tbl[domain].rc_dev;

	if (verbose) {
		if (rc_dev) {
			u32 regval2 = 0, regval3 = 0;

			/* show parent device state */
			pci_read_config_dword(rc_dev, PCI_COMMAND, &regval);
			pci_read_config_dword(rc_dev, PCI_MEMORY_BASE, &regval2);
			pci_read_config_dword(rc_dev, 0x404, &regval3);
			printk(KERN_ERR "%s: RC %s cmd 0x%8.8x Mem Limit/Base 0x%8.8x "
				   "LTSSM 0x%8.8x\n",
				   __FUNCTION__, pci_name(rc_dev), regval, regval2, regval3);
		}
	}
	if ((dev)&&(deverr)) {

		deverr->dev = dev;
		pci_read_config_dword(dev, PCI_COMMAND, &regval); /* Primary status */
		deverr->primary_status = regval;
		if (regval == ~0) {


			/*
			 * In case of a bad read, try to display enough info to
			 * determine why.
			 */
			if (dev->bus->self) {
				u32 pregval = 0, linkctl = 0;
				int pos = pci_pcie_cap(dev->bus->self);

				pci_read_config_dword(dev->bus->self, PCI_COMMAND, &pregval);
				pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL, &linkctl);

				printk(KERN_ERR "%s: %s PriStat bad read 0x%8.8x "
					   "parent %s PriStat 0x%8.8x link 0x%8.8x\n",
					   __FUNCTION__, name,  regval,
					   pci_name(dev->bus->self), pregval, linkctl);
			} else {
				printk(KERN_ERR "%s: %s Primary status bad read 0x%8.8x "
					   "detected\n", __FUNCTION__, name,  regval);
			}
			goto Out;
		}
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s Primary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name,  regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, 0x04, regval);

		}
		/* Secondary status is in upper 16 bits (0x1E) */
		pci_read_config_dword(dev, PCI_IO_BASE, &regval);
	    deverr->secondary_status = regval;
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			 /* Secondary status */
			pci_write_config_dword(dev, PCI_IO_BASE, regval);

		}

		/* Memory base and limit registers */
		pci_read_config_dword(dev, PCI_MEMORY_BASE, &regval);
	    deverr->mem_window = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s Mem Limit/Base 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {

			/* Bridge Control */
			pci_read_config_dword(dev, PCI_INTERRUPT_LINE, &regval);
			deverr->bridge_control = regval;

			/* Device status */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &regval);
			deverr->device_status = regval;
			if (regval & 0x000F0000) {
				if (verbose) {
					printk(KERN_ERR "%s: %s Device status error 0x%8.8x "
						   "detected: ", __FUNCTION__, name, regval);
					print_devstatus_errors(regval);
				}
				pci_write_config_dword(dev, pos + PCI_EXP_DEVCTL, regval);

			}

			/* Link Control */
			pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL, &regval);
			deverr->link_control = regval;
			if (verbose) {
				printk(KERN_ERR "%s: %s Link Ctrl/Status 0x%8.8x\n",
					   __FUNCTION__, name, regval);
			}
			pci_write_config_dword(dev, pos + PCI_EXP_LNKCTL, regval);

			/* Slot control/status */
			pci_read_config_dword(dev, pos + PCI_EXP_SLTCTL, &regval);
			deverr->slot_control = regval;
			if (verbose) {
				printk(KERN_ERR "%s: %s Slot Ctrl 0x%8.8x\n",
					   __FUNCTION__, name, regval);
			}
			pci_write_config_dword(dev, pos + PCI_EXP_SLTCTL, regval);
		}


		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);

		/*
		 * AER Capabilites and Control Read these before we start
		 * clearing things
		 */
		pci_read_config_dword(dev, pos + PCI_ERR_CAP, &regval);
		deverr->aer_cap = regval;

		/* Header log. */
		pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG, &deverr->header_log[0]);
		pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 4, &deverr->header_log[1]);
		pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 8, &deverr->header_log[2]);
		pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 12, &deverr->header_log[3]);

		/* Uncorr. Err Status */

		pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, &regval);
		deverr->uncorrectable_status = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s Uncorr. Error status 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_uncorrectable(regval);
			}
			pci_write_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, regval);
		}

		/* Uncorrectable Error Mask */
		pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_MASK, &regval);
		deverr->uncorrectable_mask = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s Uncorr. Error mask 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}

		/* Uncorrectable Error Severity */
		pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_SEVER, &regval);
		deverr->uncorrectable_sev = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s Uncorr. Error Sev 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}

		/* Correctable Error Status */
		pci_read_config_dword(dev, pos + PCI_ERR_COR_STATUS, &regval);
		deverr->correctable_status = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s Corr. Error status 0x%8.8x detected: ",
					   __FUNCTION__, name, regval);
				print_pci_correctable(regval);
			}
			pci_write_config_dword(dev, pos + PCI_ERR_COR_STATUS, regval);

		}

		/* Correctable Error Mask */
		pci_read_config_dword(dev, pos + PCI_ERR_COR_MASK, &regval);
		deverr->correctable_mask = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s Corr. Error mask 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}


		if (verbose) {
			printk(KERN_ERR "%s: %s AER Caps/Ctrl 0x%8.8x\n",
				   __FUNCTION__, name, deverr->aer_cap);

			printk(KERN_ERR "%s: LOG0 0x%8.8x LOG4 0x%8.8x LOG8 0x%8.8x "
				   "LOG12 0x%8.8x\n", __FUNCTION__,
				   deverr->header_log[0], deverr->header_log[1],
				   deverr->header_log[2], deverr->header_log[3]);
		}
	}

 Out:

	return;

}



/*
 * Collect and clear the PCIe switch port registers.
 *
 *
 * Inputs:
 *  ddata - Contains cached error/status/control registers
 *
 */
static void
collect_clear_pcie_switch_errors(bhpc_error_data *ddata)
{

	int i, offset = BHPC_PCIE_PCI_IDX+1, dev_domain;
	struct pci_dev *pdev = NULL;
	bhpc_dev_error *deverr = NULL;

	for (i = 0; i < num_devices_entries; i++) {
		if (devices[i].bus != 0) {

			pdev = devices[i].pdev;
			dev_domain =  pci_domain_nr(pdev->bus);
			if (ddata->domain != dev_domain) {
				continue;
			}
			if ((pdev)&&((offset + i) < MAX_DEV_ERROR)) {

				deverr = &ddata->dev_err[offset + i];
				deverr->dev_type = get_port_type(pdev);
				deverr->valid = 1;

				switch(deverr->dev_type) {

				case PCI_EXP_TYPE_PCI_BRIDGE:
					collect_clear_pcie_errors_pci_bridges(pdev, pci_name(pdev), deverr);
					break;

				case PCI_EXP_TYPE_UPSTREAM:
				case PCI_EXP_TYPE_DOWNSTREAM:
					collect_clear_pcie_errors_bridges(pdev, pci_name(pdev), deverr);
					break;

				case PCI_TYPE_NTB:
					collect_clear_pci_errors_ntb(pdev, pci_name(pdev), deverr);
					break;

				default:
					printk(KERN_ERR "%s %s Unexpected type %d\n",
						   __FUNCTION__, pci_name(pdev), deverr->dev_type);
					deverr->valid = 0;
				}
			}
		}
	}
}


/*
 * collect_clear_pcie_errors_pci_bridges()
 *
 *  Read and save the contents of the error control/status
 *  registers of a PCIe-PCI bridge device. Then clear the status
 *  registers.
 *
 * Inputs:
 *   dev - PCIE-PCI bridge device
 *   name - String to display when in verbose mode
 *   deverr - table to hold the error status/control registers
 *
 */
static void
collect_clear_pcie_errors_pci_bridges(struct pci_dev *dev, const char *name, 
									  bhpc_dev_error *deverr)
{
	u32 regval = 0;
	uint pos = 0;

	collect_clear_pcie_errors_bridges(dev, name, deverr);

	pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
	if (pos) {
		/*
		 * AER Capabilites and Control Read these before we start
		 * clearing things
		 */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_CAP, &regval);
		deverr->sec_aer_cap = regval;

		/* Header log. */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG, &deverr->sec_header_log[0]);
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 4, &deverr->sec_header_log[1]);
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 8, &deverr->sec_header_log[2]);
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 12, &deverr->sec_header_log[3]);

		/* Uncorr. Err Status */

		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_STATUS, &regval);
		deverr->sec_uncorrectable_status = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s SEC Uncorr. Error status 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_sec_uncorrectable(regval);
			}
			pci_write_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_STATUS, regval);
		}
		/* Uncorrectable Error Mask */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_MASK, &regval);
		deverr->sec_uncorrectable_mask = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s SEC Uncorr. Error mask 0x%8.8x\n",
					   __FUNCTION__, name, regval);
			}

		}

		/* Uncorrectable Error Severity */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_SEVER, &regval);
		deverr->sec_uncorrectable_sev = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s SEC Uncorr. Error Sev 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}

		if (verbose) {
			printk(KERN_ERR "%s: %s SEC AER Caps/Ctrl 0x%8.8x\n",
				   __FUNCTION__, name, deverr->sec_aer_cap);

			printk(KERN_ERR "%s: SEC LOG0 0x%8.8x LOG4 0x%8.8x LOG8 0x%8.8x "
				   "LOG12 0x%8.8x\n", __FUNCTION__,
				   deverr->sec_header_log[0], deverr->sec_header_log[1],
				   deverr->sec_header_log[2], deverr->sec_header_log[3]);
		}
	} /* if (pos) ... */
}
/*
 * Clear error status in Non-Transparent Bridge devices.
 *
 * Inputs:
 *  dev    - pci_dev object of the device to be cleared
 *  name   - String to display when in verbose mode
 *  deverr - Contains cached error/status/control registers for this device
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values.
 *
 */
static void
collect_clear_pci_errors_ntb(struct pci_dev *dev, const char *name,
							bhpc_dev_error *deverr)
{
	u32 regval = 0;

	if ((dev)&&(deverr)) {
		deverr->dev = dev;
		pci_read_config_dword(dev, PCI_COMMAND, &regval); /* Primary status */
		deverr->primary_status = regval;
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Primary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, PCI_COMMAND, regval);

		}

		pci_read_config_dword(dev, PCI_NTB_SECONDARY_STATUS, &regval); /* Secondary status */
	    deverr->secondary_status = regval;
		if (regval & 0xF9000000) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, name, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(dev, PCI_NTB_SECONDARY_STATUS, regval);

		}
		pci_read_config_dword(dev, PCI_NTB_CHIP_STATUS, &regval); /* Chip status */
		deverr->device_status = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Chip status error 0x%8.8x detected\n",
					   __FUNCTION__, name, regval);
			}
			pci_write_config_dword(dev, PCI_NTB_CHIP_STATUS, regval);

		}

		pci_read_config_dword(dev, PCI_NTB_CHIP_CONTROL, &regval); /* Chip control */
		deverr->bridge_control = regval;
		if (regval) {
			if (verbose) {
				printk(KERN_ERR "%s: %s  Chip control 0x%8.8x\n",
					   __FUNCTION__, name, regval);
			}
		}

		/*
		 * Read/save the cfg address register. This can be useful if
		 * problem was caused by config read of a device downstream of the NTB
		 */
		pci_read_config_dword(dev, PCI_NTB_CFG_ADDR, &regval);
		deverr->header_log[0] = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s: CfgAddr 0x%8.8x\n",
					   __FUNCTION__, name, regval);
		}
	}

}

/*
 * Decode the PEX_CONFIG_ADDR register to determine what bus, device,
 * and offset was being accessed.
 */
static int
find_config_addr_target(bhpc_error_data *ddata)
{
	char *sptr = NULL;
	u32 bus, dev, offset, upper, lower;
	u32 type, fmt, config_addr;
	int src_found = 0;
	bhpc_dev_error *deverr = NULL;
	struct pci_dev *pdev;

	/*
	 * Packet type: bits 18-22
	 * Packet format: bits 16-17
	 */
	type = (ddata->pex.capture_data[0] >> 9) & 0x1F;
	fmt = (ddata->pex.capture_data[0] >> 14) & 0x03;

	/*
	 * Target bus number: bits 8-15
	 * Target dev number: bits 16-19
	 */
	config_addr = ddata->pex.config_addr;
	bus = (config_addr >> 16) & 0xFF;
	dev = (config_addr >> 11) & 0x1F;
	lower = (config_addr >> 2) & 0x3F;
	upper = (config_addr >> 18) & 0x3C0; /* Shift bits 4-7 down to 22-25 */

	/*
	 * Lower 6 bits of offset are in bits 24-27. 
	 * Upper 4 bits of offset are in bits 4-7
	 */
	offset = (((config_addr >> 2) & 0x3F) |
			  ((config_addr >> 18) & 0x3C0));

	if ((bus == 0)&&(dev == 0)) {
		pdev = domain_tbl[ddata->domain].rc_dev; /* Target was the RC */
	} else {
		pdev = lookup_mm_dsp(ddata->domain, bus, dev);
	}

	if (ddata) {
		if (pdev) {
			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;
		}
		ddata->src_dev.dev = dev;
		ddata->src_dev.bus = bus;
		src_found = 1;
	}

	switch (type) {

		case 0x4: /* Config type 0 */

			switch (fmt) {
			case 0:
				sptr = "  Config Type0 Read Request";
				break;

			case 2:
				sptr = "  Config Type0 Write request";
				break;

			default:
				sptr = "  Undecoded Config type 0 request";
			}

			break;

		case 0x5: /* Config type 1 */
			switch (fmt) {
			case 0:
				sptr = "  Config Type1 Read Request";
				break;

			case 2:
				sptr = "  Config Type1 Write request";
				break;

			default:
				sptr = "  Undecoded Config type 1 request";
			}

			break;

		default:
			sptr = "  Undecoded Config Outbound packet";
	}

	deverr = &ddata->dev_err[BHPC_RC_IDX];

	printk(KERN_ERR "%s %d: %s type %d fmt %d cap_r0 0x%x config_addr 0x%x "
		   "bus %d dev %d offset 0x%x upper 0x%x lower 0x%x link "
		   "status/ctrl 0x%8.8x\n", __FUNCTION__, __LINE__, sptr, type,
		   fmt, ddata->pex.capture_data[0], config_addr, bus, dev,
		   offset, upper, lower, deverr->link_control);

	return (src_found);
}
/*
 * PME events always originate from the RC, so it is the
 * source device. Certain other events detected by the RC
 * are also attributed to the RC as well. 
 */
static int
find_rc_pme_event_source(bhpc_error_data *ddata)
{
	struct pci_dev *pdev = NULL;
	u32 cdev = 0, cbus = 0, cfunc = 0;
	int src_found = 0;

	/* This is always the RC - bus 0, dev 0 */

	if (ddata) {
		pdev = lookup_mm_dsp(ddata->domain, cbus, cdev);
		if (pdev) {
			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;
			if (verbose) {
				printk(KERN_ERR "%s %d: MM DSP %s dev %d bus %d\n",
					   __FUNCTION__, __LINE__, pci_name(pdev),
				   ddata->mm_parent_dev.dev, ddata->mm_parent_dev.bus);
			}
		}

		ddata->src_dev.dev = cdev;
		ddata->src_dev.bus = cbus;
		ddata->src_dev.func = cfunc;
		src_found = 1;
	}

	return (src_found);
}
/*
 * pcie_completion_src()
 *
 * Decode the contents of the PCIe controller capture status
 * registers to determine whe requester device and completer
 * device. The capture status registers contain the contents
 * of a PCIe completion packet.
 *
 * The requester device should always be the RC, and the
 * completer device will be considered to be the source
 * of an error/event.
 *
 * Then find the MM PCIe switch DSP that connects to the
 * completer device.
 *
 * Inputs:
 *   type - Type of packet that was used to populate the
 *          capture status registers.
 *   fmt  - Format of the packet that was used to populate
 *         the capture status registers.
 *   ddata - Contains all the error/control/status registers
 *           that have been collected. Also used to pass back
 *           the error/event source device and it's MM DSP.
 *
 * Returns:
 *   1 = Source device was found
 *   0 = Source device was not found
 *
 */
static int
pcie_completion_src(u32 type, u32 fmt, bhpc_error_data *ddata)

{
	char *cs, *fs;
	u32 cbus, cfunc, cdev, status, rbus, rfunc, rdev;
	struct pci_dev *pdev = NULL;
	int src_found = 0;

	/* Completer bus number: bits 24-31 */
	cbus = (ddata->pex.capture_data[1]) & 0xFF;

	/* Completer Function: Bits 21-23 */
	cfunc = (ddata->pex.capture_data[1] >> 8) & 0x07;

	/* Completer Device: Bits 16-20 */
	cdev = (ddata->pex.capture_data[1] >> 11) & 0x1F;

	/* Completion status: Bits 8-10 */
	status = (ddata->pex.capture_data[1] >> 21) & 0x07;

	/* Requester bus number: bits 24-31 */
	rbus = (ddata->pex.capture_data[2]) & 0xFF;

	/* Requester Function: Bits 21-23 */
	rfunc = (ddata->pex.capture_data[2] >> 8) & 0x07;

	/* Requester Device: Bits 16-20 */
	rdev = (ddata->pex.capture_data[2] >> 11) & 0x1F;

	switch (status) {
	case 0:
		cs = "SC";	/* Successful Completion */
		break;
	case 1:
		cs = "UR";	/* Unsupported Request */
		break;
	case 2:
		cs = "CRS";	/* Configuration Retry Request */
		break;
	case 4:
		cs = "CA";	/* Completer abort */
		break;
	default:
		cs = "Unknown";
	}

	switch (fmt) {
	case 0:
		fs = "No Data";
		break;
	case 1:
		fs = "With Data";
		break;
	default:
		fs = "Unknown";
	}

	if (verbose) {
		printk("%s: Completion packet [%s] fmt [%s] from bus %d dev %d "
			   "func %d\n", __FUNCTION__, cs, fs, cbus, cdev, cfunc);

		printk("%s: Requester bus %d dev %d func %d\n",
			   __FUNCTION__, rbus, rdev, rfunc);
	}

	pdev = lookup_mm_dsp(ddata->domain, cbus, cdev);

	if (ddata) {
		if (pdev) {
			int rc = 0;

			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;

			rc = get_slot_number(ddata->domain, cbus, cdev, ddata);

			if (verbose) {
				printk("%s %d: MM DSP %s dev %d bus %d slot %d\n",
					   __FUNCTION__, __LINE__, pci_name(pdev),
					   ddata->mm_parent_dev.dev, ddata->mm_parent_dev.bus,
					   ddata->slot);
			}
		}

		ddata->src_dev.dev = cdev;
		ddata->src_dev.bus = cbus;
		ddata->src_dev.func = cfunc;
		src_found = 1;
	}

	return (src_found);
}
/*
 * find_pex_err_dr_source()
 *
 * Decodes the PCIe controller registers to determine the
 * type of event/error that occured, the source device that
 * generated the event, and the MM DSP that connects to the
 * blade containing that device.
 *
 * The PCIe controller registers contain selected portions of
 * the PCIe packet that was used to report the event. There are
 * several types of packets, but we currently only look at
 * completion packets. Events/errors reported by other packet
 * types are decoded from the RC registers.
 *
 * Inputs:
 *   ddata - Contains all the error control/status registers
 *           that were collected. Will be used to pass back
 *           the source device and it's MM DSP parent, if found.
 *
 * Returns:
 *   0 = Source device could not be decoded from the register data
 *   1 = Source device was decoded from the register data.
 */
static int
find_pex_err_dr_source(bhpc_error_data *ddata)
{
	int config, gsid, inbound, err_src_found = 0;
	u32 type = 0, fmt = 0;


	/* Make sure it's valid status */
	if ((ddata->pex.capture_status & 1) == 0) {
		goto Exit;
	}

	config = (ddata->pex.capture_status & 0x80000000);
	gsid = ddata->pex.capture_status >> 1;

	/* Only looking at PCIe 0 controller for now */
	inbound = (gsid == 0x02);

	/*
	 * Inbound transactions that result in PNM, PCAC, CDNSC or
	 * CRSNC set will result in GSID == 0x02
	 */

	if (inbound) {

		/*
		 * For inbound, capture_data[0] has the packet type and
		 * capture_data[1] has the requester ID
		 */

		/* Packet type: bits 27-31 */
		type = (ddata->pex.capture_data[0] & 0x1F);

		/* Packet format: bits 25-26 */
		fmt = (ddata->pex.capture_data[0] >> 5) & 0x03;

		switch (type) {

		case 0x0A:  /* Completion w/o lock packet */
		case 0x0B: /* Completion with lock packet */

			err_src_found = pcie_completion_src(type, fmt, ddata);

			break;


		case 0x00000: /* Mem Read/Write */
		case 0x00001: /* Mem Read Lock */

#if 0	/* Future enhancement */
			err_src_found = pcie_memory_src(type, fmt, ddata);
#endif
			break;

		case 0x4: /* Config type 0 */
		case 0x5: /* Config type 1 */

#if 0	/* Future enhancement */
			err_src_found = config_src(type, fmt,  ddata);

#endif
			break;

		default:

			printk(KERN_ALERT "%s Undecoded inbound packet type 0x%x "
				   "fmt 0x%x ERR_CAP_STATUS 0x%8.8x ERR_CAP_R0 0x%8.8x "
				   "ERR_CAP_R1 0x%8.8x ERR_CAP_R2 0x%8.8x "
				   "ERR_CAP_R3 0x%8.8x\n",
				   __FUNCTION__, type, fmt, ddata->pex.capture_status,
				   ddata->pex.capture_data[0], ddata->pex.capture_data[1],
				   ddata->pex.capture_data[2], ddata->pex.capture_data[2]);
		}

	} else { /* It's an outbound transaction */
		bhpc_dev_error *deverr = &ddata->dev_err[BHPC_RC_IDX];

		if (config) {

			if (ddata->pex.detect & 0x00800000) {

				/*
				 * It's a completion timeout during a config access.
				 * The RC is unable to capture any info for this, so
				 * the only chance to determine the device being
				 * accessed is from the contents of the PEX_CONFIG_ADDR
				 * register. This assumes we were able to capture that
				 * register before something else could change it.
				 */

				err_src_found = find_config_addr_target(ddata);

			} else if (ddata->pex.detect & 0x00200000) {
				/*
				 * It's a target abort during a config access.
				 * The RC will capture info for this, so nothing to
				 * do here
				 */

			} else {

				printk(KERN_ALERT "%s %d: Undecoded config output packet type "
					   "ERR_DET 0x%8.8x "
					   "ERR_CAP_STATUS 0x%8.8x ERR_CAP_R0 0x%8.8x "
					   "ERR_CAP_R1 0x%8.8x ERR_CAP_R2 0x%8.8x "
					   "ERR_CAP_R3 0x%8.8x \n",
					   __FUNCTION__, __LINE__,
					   ddata->pex.detect, ddata->pex.capture_status,
					   ddata->pex.capture_data[0], ddata->pex.capture_data[1],
					   ddata->pex.capture_data[2], ddata->pex.capture_data[2]);

				printk(KERN_ALERT "%s %d: LOG0 0x%8.8x LOG4 0x%8.8x "
					   "LOG8 0x%8.8x LOG12 0x%8.8x type 0x%x fmt 0x%x "
					   "Config Addr 0x%8.8x\n",
					   __FUNCTION__, __LINE__,
					   deverr->header_log[0], deverr->header_log[1],
					   deverr->header_log[2], deverr->header_log[2],
					   type, fmt, ddata->pex.config_addr);
			}

		} else if (ddata->pex.detect & 0x00000020) {

			/*
			 * It's a root complex link down data error. The link
			 * between the RC and the PCIe switch USP has gone
			 * down.
			 */
			find_rc_pme_event_source(ddata);

		} else {

			printk(KERN_ALERT "%s %d: Undecoded output packet type "
				   "ERR_DET 0x%8.8x "
				   "ERR_CAP_STATUS 0x%8.8x ERR_CAP_R0 0x%8.8x "
				   "ERR_CAP_R1 0x%8.8x ERR_CAP_R2 0x%8.8x "
				   "ERR_CAP_R3 0x%8.8x \n",
				   __FUNCTION__, __LINE__,
				   ddata->pex.detect, ddata->pex.capture_status,
				   ddata->pex.capture_data[0], ddata->pex.capture_data[1],
				   ddata->pex.capture_data[2], ddata->pex.capture_data[2]);

			printk(KERN_ALERT "%s %d: LOG0 0x%8.8x LOG4 0x%8.8x "
				   "LOG8 0x%8.8x LOG12 0x%8.8x type 0x%x fmt 0x%x "
				   "Config Addr 0x%8.8x\n",
				   __FUNCTION__, __LINE__,
				   deverr->header_log[0], deverr->header_log[1],
				   deverr->header_log[2], deverr->header_log[2],
				   type, fmt, ddata->pex.config_addr);
		}
	}

 Exit:
	if ((ddata)&&(verbose)) {
		printk("%s: Source bus %d dev %d MM Parent "
			   "device bus %d dev %d\n",
			   __FUNCTION__, ddata->src_dev.bus, ddata->src_dev.dev,
			   ddata->mm_parent_dev.bus, ddata->mm_parent_dev.dev);
	}

	return (err_src_found);

}
/*
 * collect_clear_rc_errors()
 *
 * Read, store, and optionally clear the RC error control/status
 * registers. All registers are in config space.
 *
 * Inputs:
 *   rdev - The RC pci_dev object
 *   ddata - where the collected register data will be stored
 *   clear - True = clear the status after reading. False = leave
 *           status as-is after reading.
 *
 * Returns::
 *   N/A
 */
static void
collect_clear_rc_errors(struct pci_dev *rdev, bhpc_error_data *ddata, int clear)
{
	int pos;
	u32 regval;
	bhpc_dev_error *deverr;
	char *name = "RC";

	if ((rdev)&&(ddata)) {

		deverr = &ddata->dev_err[BHPC_RC_IDX];
		deverr->dev_type = PCI_EXP_TYPE_ROOT_PORT;
		deverr->valid = 1;
		deverr->dev = rdev;

		pci_read_config_dword(rdev, PCI_COMMAND, &regval);
		deverr->primary_status = regval;
		if ((regval & 0xF9000000)||(verbose)) {
			if (verbose) {
				printk(KERN_ERR "%s: RC Primary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, regval);
				print_pci_errors(regval);
			}
			if (clear) {
				pci_write_config_dword(rdev, PCI_COMMAND, regval);
			}
		}

		/* Secondary status is in upper 16 bits (0x1E) */

		pci_read_config_dword(rdev, PCI_IO_BASE, &regval);
		deverr->secondary_status = regval;
		if ((regval & 0xF9000000)||(verbose)) {
			if (verbose) {
				printk(KERN_ERR "%s: RC Secondary status error 0x%8.8x "
					   "detected: ", __FUNCTION__, regval);
				print_pci_errors(deverr->secondary_status);
			}
			if (clear) {
				pci_write_config_dword(rdev, PCI_IO_BASE, regval);
			}
		}

		/* Memory base and limit registers */
		pci_read_config_dword(rdev, PCI_MEMORY_BASE, &regval);
	    deverr->mem_window = regval;
		if (verbose) {
			printk(KERN_ERR "%s: %s Mem Limit/Base 0x%8.8x\n",
				   __FUNCTION__, name, regval);
		}

		/* Bridge Control */
		pci_read_config_dword(rdev, PCI_INTERRUPT_LINE, &regval);
		deverr->bridge_control = regval;

		pos = pci_find_capability(rdev, PCI_CAP_ID_EXP);
		if (pos) {
			pci_read_config_dword(rdev, pos + PCI_EXP_DEVCTL, &regval);
			deverr->device_status = regval;

			/* Device control/status */
			if ((regval & 0x000F0000)||(verbose)) {
				if (verbose) {
					printk(KERN_ERR "%s: %s Device status error 0x%8.8x "
							"detected: ", __FUNCTION__, name, regval);
					print_devstatus_errors(regval);
				}
				if (clear) {
					pci_write_config_dword(rdev, pos + PCI_EXP_DEVCTL, regval);
				}
			}

			/* Slot control/status */
			pci_read_config_dword(rdev, pos + PCI_EXP_SLTCTL, &regval);
			deverr->slot_control = regval;
			if (verbose) {
				printk(KERN_ERR "%s: %s Slot Ctrl 0x%8.8x  ",
					   __FUNCTION__, name, regval);
			}
			if (clear) {
				pci_write_config_dword(rdev, pos + PCI_EXP_SLTCTL, regval);
			}

			/* Link control/status. Nothing clearable */
			pci_read_config_dword(rdev, pos + PCI_EXP_LNKCTL, &regval);
			deverr->link_control = regval;
			if (verbose) {
				printk(KERN_ERR "Link Ctrl 0x%8.8x  ", regval);
			}

			/* Root control */
			pci_read_config_dword(rdev, pos + PCI_EXP_RTCTL, &regval);
			deverr->root_control = regval;
			if (verbose) {
				printk(KERN_ERR "Root Ctrl 0x%8.8x  ", regval);
			}

			/* Root status */
			pci_read_config_dword(rdev, pos + PCI_EXP_RTSTA, &regval);
			deverr->root_status = regval;
			if (verbose) {
				printk(KERN_ERR "Root status 0x%8.8x\n", regval);
			}


			if (clear) {
				pci_write_config_dword(rdev, pos + PCI_EXP_RTSTA, regval);
			}

		}

		pos = pci_find_ext_capability(rdev, PCI_EXT_CAP_ID_ERR);
		if (pos) {

			/*
			 * AER Capabilites and Control Read these before we start
			 * clearing things
			 */
			pci_read_config_dword(rdev, pos + PCI_ERR_CAP, &regval);
			deverr->aer_cap = regval;

			if ((deverr->aer_cap & 0x1F) != 0) {
				printk(KERN_ERR "%s: ******* FIRST ERROR PTR SET 0x%8.8x\n",
					   __FUNCTION__, deverr->aer_cap);
			}
			/* Header log */
			pci_read_config_dword(rdev, pos + PCI_ERR_HEADER_LOG,
								  &deverr->header_log[0]);
			pci_read_config_dword(rdev, pos + PCI_ERR_HEADER_LOG + 4,
								  &deverr->header_log[1]);
			pci_read_config_dword(rdev, pos + PCI_ERR_HEADER_LOG + 8,
								  &deverr->header_log[2]);
			pci_read_config_dword(rdev, pos + PCI_ERR_HEADER_LOG + 12,
								  &deverr->header_log[3]);

			/* Uncorr. Err Status */
			pci_read_config_dword(rdev, pos + PCI_ERR_UNCOR_STATUS, &regval);
			deverr->uncorrectable_status = regval;
			if (regval || verbose) {
				if (verbose) {
					printk(KERN_ERR "%s: RC Uncorr. Err Status error 0x%8.8x "
						   "detected: ", __FUNCTION__, regval);
					print_pci_uncorrectable(regval);
				}

			}

			pci_read_config_dword(rdev, pos + PCI_ERR_UNCOR_MASK, &regval);
			deverr->uncorrectable_mask = regval;

			/* Uncorrectable Error Severity */
			pci_read_config_dword(rdev, pos + PCI_ERR_UNCOR_SEVER, &regval);
			deverr->uncorrectable_sev = regval;

			/* Corr. Err Status */
			pci_read_config_dword(rdev, pos + PCI_ERR_COR_STATUS, &regval);
			deverr->correctable_status = regval;
			if (regval || verbose) {
				if (verbose) {
					printk(KERN_ERR "%s: RC Corr. Err Status 0x%8.8x ",
						   __FUNCTION__, regval);
					print_pci_correctable(regval);
				}

			}

			pci_read_config_dword(rdev, pos + PCI_ERR_COR_MASK, &regval);
			deverr->correctable_mask = regval;

			/* Root Err status */
			pci_read_config_dword(rdev, pos + PCI_ERR_ROOT_STATUS, &regval);
			deverr->root_err_status = regval;

			if (verbose) {
				printk(KERN_ERR "%s: RC Root Err Status 0x%8.8x ",
					   __FUNCTION__, regval);
				print_pci_root_errors(regval);
			}

			/* Link training and status state machine */
			pci_read_config_dword(rdev, 0x404, &regval);
			deverr->rc_ltssm = regval;

			if (verbose) {
				printk(KERN_ERR "%s: Link Ctrl 0x%8.8x LTSSM 0x%x\n",
					   __FUNCTION__, deverr->link_control, regval);
			}

			pci_read_config_dword(rdev, 0x5A0, &regval);
			if (verbose) {
				printk(KERN_ERR "%s: SecStatusMask 0x%8.8x\n",
					   __FUNCTION__, regval);
			}
			pci_read_config_dword(rdev, pos + PCI_ERR_ROOT_COR_SRC, &regval);
			deverr->corr_err_src = (u16)(regval & 0x0000FFFF);
			deverr->uncorr_err_src = (u16)((regval >> 16) & 0x0000FFFF);

			if (verbose) {
				printk(KERN_ERR "%s: RC Errsrc 0x%8.8x corrSrc 0x%8.8x "
					   "uncorrSrc 0x%8.8x\n", __FUNCTION__, regval,
					   deverr->corr_err_src, deverr->uncorr_err_src);
			}

			if (clear) {
				/*
				 * W1C all bits. Do this after capturing the entire
				 * register group for cohenency. Some of the registers
				 * can begin latching new status once a bit is cleared
				 */

				pci_write_config_dword(rdev, pos + PCI_ERR_UNCOR_STATUS,
									   deverr->uncorrectable_status);

				pci_write_config_dword(rdev, pos + PCI_ERR_COR_STATUS,
									   deverr->correctable_status);

				pci_write_config_dword(rdev, pos + PCI_ERR_ROOT_STATUS,
									   deverr->root_err_status);


			}
			if (verbose) {
				printk(KERN_ERR "%s: AER CAPS and FEP 0x%8.8x first err %s\n",
					   __FUNCTION__, deverr->aer_cap,
					   uncorr_errs[(deverr->aer_cap & 0x1f)]);

				printk(KERN_ERR "%s: LOG0 0x%8.8x LOG4 0x%8.8x LOG8 0x%8.8x "
					   "LOG12 0x%8.8x\n", __FUNCTION__,
					   deverr->header_log[0], deverr->header_log[1],
					   deverr->header_log[2], deverr->header_log[3]);
			}

			if (verbose) {
				show_bridge_bus_window(rdev);
				show_bridge_mem_window(rdev);
			}
		}
	}
}
/*
 * collect_clear_pex_errors()
 *
 * Read, store, and optionally clear all error control/status registers
 * in the PCIe controller. All registers are memory-mapped.
 *
 * Inputs:
 *   vaddr - The ioremap'd base address of the PCIe controller registers
 *   ddata - where the collected register data will be stored
 *   clear - True = clear the status after reading. False = leave
 *           status as-is after reading.
 *
 * Returns::
 *   N/A
 */
static void
collect_clear_pex_errors(volatile void __iomem *vaddr,
						 bhpc_error_data *ddata, int clear)
{
	volatile void __iomem *regaddr;


	if (ddata) {

		regaddr = vaddr + 0xe00;	/* ERR_DR */
		ddata->pex.detect = readl_be(regaddr);
		if (clear) {
			writel_be(ddata->pex.detect, regaddr);
		}

		regaddr = vaddr + 0xe08;  /* ERR_EN */
		ddata->pex.enable = readl_be(regaddr);

		regaddr = vaddr + 0xe10;  /* ERR_DIS */
		ddata->pex.disable = readl_be(regaddr);

		regaddr = vaddr + 0xe20;  /* ERR_CAP_STAT */
		ddata->pex.capture_status = readl_be(regaddr);
		if (clear) {
			writel_be(ddata->pex.capture_status, regaddr);
			writel_be(0, regaddr);
		}

		regaddr = vaddr + 0xe28;  /* ERR_CAP_R0 */
		ddata->pex.capture_data[0] = readl_be(regaddr);

		if (clear) {
			writel_be(0, regaddr);
		}

		regaddr = vaddr + 0xe2c;  /* ERR_CAP_R1 */
		ddata->pex.capture_data[1] = readl_be(regaddr);

		if (clear) {
			writel_be(0, regaddr);
		}

		regaddr = vaddr + 0xe30; /* ERR_CAP_R2 */
		ddata->pex.capture_data[2] =  readl_be(regaddr);

		if (clear) {
			writel_be(0, regaddr);
		}

		regaddr = vaddr + 0xe34; /* ERR_CAP_R3 */
		ddata->pex.capture_data[3] =  readl_be(regaddr);

		if (clear) {
			writel_be(0, regaddr);
		}

		ddata->pex.valid = 1;

#ifndef CONFIG_SILKWORM_3041
		/*
		 * Collect the PME data. Even though PME is disabled, it
		 * can still generate interrupts.
		 */
		regaddr = vaddr + 0x014; // PME_CONFIG
		ddata->pme.config = readl_be(regaddr);

		regaddr = vaddr + 0x020;  // PME_DR
		ddata->pme.detect = readl_be(regaddr);
		if (clear) {
			writel_be(ddata->pme.detect, regaddr);
		}

		regaddr = vaddr + 0x024;  // PME_DISR
		ddata->pme.disable = readl_be(regaddr);

		regaddr = vaddr + 0x028;  // PME_IER
		ddata->pme.enable = readl_be(regaddr);

		regaddr = vaddr + 0xe2c;  // PME_PMCR
		ddata->pme.command = readl_be(regaddr);

		ddata->pme.valid = 1;

		if ((ddata->pex.detect)||(verbose)) {

			if (verbose) {
				printk(KERN_ERR "%s %d: ERR_DR 0x%8.8x ERR_EN 0x%8.8x "
				   "ERR_DISR 0x%8.8x ERR_CAP_STATUS 0x%8.8x "
				   "Config addr 0x%8.8x\n",
				   __FUNCTION__, __LINE__,
				   ddata->pex.detect, ddata->pex.enable, ddata->pex.disable,
				   ddata->pex.capture_status, ddata->pex.config_addr);


				printk(KERN_ERR "%s %d: ERR_CAP_R0 0x%8.8x ERR_CAP_R1 0x%8.8x "
				   "ERR_CAP_R2 0x%8.8x ERR_CAP_R3 0x%8.8x\n",
				   __FUNCTION__, __LINE__,
				   ddata->pex.capture_data[0], ddata->pex.capture_data[1],
				   ddata->pex.capture_data[2], ddata->pex.capture_data[3]);
			}
			if (ddata->pex.detect) {
				display_pex_errors(ddata->domain, ddata->pex.detect, ddata->pex.capture_status,
								ddata->pex.capture_data[0],
								ddata->pex.capture_data[1],
								ddata->pex.capture_data[2],
								ddata->pex.capture_data[3]);
			}
		}

		if ((ddata->pme.detect)||(verbose)) {

			if (verbose) {
				printk(KERN_ERR "%s %d: PME_CONFIG 0x%8.8x PME_DR 0x%8.8x "
					   "PME_IER 0x%8.8x PME_DISR 0x%8.8x PEX_PMCR 0x%8.8x\n",
					   __FUNCTION__,__LINE__,
					   ddata->pme.config, ddata->pme.detect, ddata->pme.enable,
					   ddata->pme.disable, ddata->pme.command);
			}

			if (ddata->pme.detect) {
				display_pme_events(ddata->pme.detect);
			}
		}
#endif
	}

}
/*
 * Decode & display the contents of the PCIe controller pme message detect register
 */
static void
decode_pme(bhpc_error_data *ddata)
{
	bhpc_pcie_pme *pme = NULL;

	if (ddata) {

		pme = &ddata->pme;

		ddata->error_idx = BHPC_PME_IDX;

		printk("%s: PME_DR 0x%8.8x ", __FUNCTION__, pme->detect);

		if (pme->detect & 0x8000) {
			ddata->error_idx = BHPC_PME_PTO_IDX;
			printk(" [PTO]");
		}
		if (pme->detect & 0x2000) {
			ddata->error_idx = BHPC_PME_ENL23_IDX;
			printk(" [ENL23]");
		}
		if (pme->detect & 0x1000) {
			ddata->error_idx = BHPC_PME_EXL23_IDX;
			printk(" [EXL23]");
		}
		if (pme->detect & 0x0400) {
			ddata->error_idx = BHPC_PME_HRD_IDX;
			printk(" [HRD]");
		}
		if (pme->detect & 0x0200) {
			ddata->error_idx = BHPC_PME_LDD_IDX;
			printk(" [LDD]");
		}
		if (pme->detect & 0x0080) {
			ddata->error_idx = BHPC_PME_LUD_IDX;
			printk(" [LUD]");
		}
		printk("\n");
	}
}
/*
 * process_err_detect()
 *
 * Anaylize and decode the PCIe controller error control/status
 * registers to determine what event/error occured, the source
 * device that generated the event, and the MM DSP port connected
 * to the blade that contains the source device.
 *
 * If the error detect register detected an event/error, then
 * set error_decode to True.
 *
 * If the source device is found, then src_decode is set to
 * PCIE_ERROR_ERR_DETECT.
 *
 * This should always be the first place we look to decode an
 * event/error
 *
 * Inputs:
 *   ddata - Contains all the register data to be analized.
 *           Also used to pass back the ID of source device and parent
 *           MM DSP
 *
 * Returns:
 *  N/A
 *
 */

static void
process_err_detect(bhpc_error_data *ddata)
{
	int rc = 0;

	if (ddata) {
		if (ddata->pex.detect != 0) {

			decode_err_detect(ddata);
			ddata->error_detected = 1;

			rc = find_pex_err_dr_source(ddata);
			if (rc) {
				ddata->src_decode = PCIE_ERROR_ERR_DETECT;
			}
		}

		if ((!ddata->error_detected)&&(ddata->pme.detect != 0)) {
			//			display_pme_events(ddata->pme.detect);
			decode_pme(ddata);
			ddata->error_detected = 1;

			rc = find_rc_pme_event_source(ddata);
			if (rc) {
				ddata->src_decode = PCIE_ERROR_PME_DETECT;
			}
		}
	}
}
/*
 * decode_rc_legacy_err()
 *
 * Analyze and decode the RC primary and secondary error status
 * to determine what type of legacy PCI event occured (PAERR,
 * SERR, RMA, etc).
 *
 * If these registers are the only thing that recorded an
 * event/error, then we don't really know the source of the
 * event. So we credit it to the RC itself.
 *
 * If an event/error was detected, then set error_decode to True.
 *
 * This should always be the last place we look to decode an
 * event/error, since it provides the least amount of information
 *
 * Inputs:
 *   rdev - The RC pci_dev object
 *   ddata - Contains all the register data to be analized.
 *           Also used to pass back the ID of source device and parent
 *           MM DSP
 *
 * Returns:
 *  0 = No legacy error was reported
 *  1 = a legacy error was reported.
 *
 */

static void
decode_rc_legacy_err(struct pci_dev *rdev, bhpc_error_data *ddata)
{
	bhpc_dev_error *deverr = NULL;

	if ((ddata)&&(ddata->src_decode != PCIE_ERROR_PME_DETECT)&&
		(ddata->src_decode != PCIE_ERROR_ERR_DETECT)) {

		deverr = &ddata->dev_err[BHPC_RC_IDX];

		if ((deverr->primary_status & 0xF9000000)||
			(deverr->secondary_status & 0xF9000000)) {
			ddata->error_detected = 1;

			if (ddata->src_decode == PCIE_ERROR_NONE) {
				ddata->src_decode = PCIE_ERROR_LEGACY_ERR;
			}

			/*
			 * If no source device has been found, then the
			 * RC is the only device reporting an error
			 */
			if (!ddata->mm_dsp_dev) {
				int src_found;

				src_found = decode_bus_err_src(deverr->uncorr_err_src, ddata);
				if (!src_found) {
					ddata->mm_dsp_dev = rdev;
					ddata->mm_parent_dev.dev = PCI_SLOT(rdev->devfn);
					ddata->mm_parent_dev.bus = rdev->bus->number;
					ddata->src_dev.dev = 0;
					ddata->src_dev.bus = 0;
					ddata->src_dev.func = 0;
				}

			}
			if ((deverr->primary_status & 0x80000000)||
				(deverr->secondary_status & 0x80000000)) {

				ddata->error_idx = BHPC_SPAERR_IDX;

			} else if ((deverr->primary_status & 0x01000000)||
					   (deverr->secondary_status & 0x01000000)){

				ddata->error_idx = BHPC_MPAERR_IDX;

			} else if ((deverr->primary_status & 0x10000000)||
					  (deverr->secondary_status & 0x10000000)) {

				ddata->error_idx= BHPC_RTA_IDX;

			} else if ((deverr->primary_status & 0x08000000)||
					   (deverr->secondary_status & 0x08000000)){

				ddata->error_idx = BHPC_STA_IDX;

			} else if ((deverr->primary_status & 0x20000000)||
					   (deverr->secondary_status & 0x20000000)) {

				ddata->error_idx = BHPC_RMA_IDX;

			} else if ((deverr->primary_status & 0x40000000)||
					   (deverr->secondary_status & 0x40000000)) {

				/* SERR can sometimes be further decoded */
				if (deverr->root_status & BHPC_ERROR_NON_COR) {
					bhpc_dev_error *dsp_deverr;

					dsp_deverr = &ddata->dev_err[BHPC_MM_DSP_IDX];

					if ((dsp_deverr)&&
						(dsp_deverr->uncorrectable_status & 0x00000020)) {
						ddata->error_idx = BHPC_SDE_IDX;
					}else if (deverr->uncorrectable_status & 0x00004000) {
						ddata->error_idx = BHPC_CTO_IDX;
					}
				} else {
					ddata->error_idx = BHPC_SSERR_IDX;
				}
			}
		}
	}
}

/*
 * Decode error source based on captured PCI (not PCIe) bus state.
 * Used with secondary side registers of PCIe-PCI bridges.
 *
 * This function decodes Type 1 config addresses.
 */
static int
decode_pci_config1_err_src(u32 src_ids, bhpc_error_data *ddata)
{

	u32 sbus = 0, sdev;
	struct pci_dev *pdev = NULL, *dev = NULL;
	int src_found = 0, i;


	sdev = (src_ids >> 11) & 0x1F;
	sbus = (src_ids >> 16) & 0xFFFF;

	for (i = 0; i < num_devices_entries; i++) {

		if ((devices[i].domain == ddata->domain)&&
			(devices[i].first_bus <=sbus)&&
			(devices[i].last_bus >= sbus)) {

			/*
			 * If this bus device has subranges, then search them
			 * to find the exact slot.
			 */
			if (devices[i].bus_subranges.number > 0) {
				dsp_subranges *ptr = NULL;
				int x;

				ptr = &devices[i].bus_subranges;
				dev = devices[i].pdev;

				for (x = 0; (x < ptr->number)&&
						 ( x < BHPC_MAX_DSP_SUBRANGES); x++) {
					if ((sbus >= ptr->subranges[x].base)&&
						(sbus <= ptr->subranges[x].limit)) {
						if (verbose) {
							printk(KERN_ERR "%s x %d: Found subrange "
								   "Dev %s Base 0x%8.8x Limit 0x%8.8x "
								   "slot %d\n", __FUNCTION__, x, pci_name(dev),
								   ptr->subranges[x].base,
								   ptr->subranges[x].limit,
								   ptr->subranges[x].slot);
						}
						ddata->slot = ptr->subranges[x].slot;
					}
				}
			} else {
				ddata->slot = devices[i].slot;
			}
			src_found = 1;
			break;
		}
	}

	if (verbose) {
		printk(KERN_ERR "%s:  Error Source "
			   "device 0x%4.4x:%2.2x\n", __FUNCTION__, sbus, sdev);
	}

	if ((sbus == 0)&&(sdev == 0)) {
		pdev = domain_tbl[ddata->domain].rc_dev;
	} else {
		pdev = lookup_mm_dsp(ddata->domain, sbus, sdev);
	}

	if (ddata) {
		if (pdev) {
			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;
		}
		ddata->src_dev.dev = sdev;
		ddata->src_dev.bus = sbus;
	}

	return (src_found);
}
/*
 * Decode error source based on captured PCI (not PCIe) bus state.
 * Used with secondary side registers of PCIe-PCI bridges.
 *
 * This function decodes Type 0 config addresses. 
 *
 * This means the NTB was trying to access a blade bridge. To determine
 * the chassis slot, we need to do a table lookup based on:
 *   - The target device that was being accessed. This is encoded in
 *     the ntb_cfg_addr value.
 *   - The bus on which the target device is attached. For this, we need to
 *     know which NTB was involved, which is determined by contents of the
 *     bus_cfg_addr and a lookup table.
 *
 * Inputs:
 *   ntb_cfg_addr - Contents of the NTB configuration address register
 *   bus_cfg_addr - Contents of the PCI bus AD bits as captured by the
 *                  PCIe-PCI bridge when the error occured.
 *
 *                  For both values, bits [31:11] represent idsels [20:0].
 *                  Each idsel is wired directly to the idsel pin of a single.
 *
 */
static int
decode_pci_config0_err_src(u32 ntb_cfg_addr, u32 bus_cfg_addr, bhpc_error_data *ddata)
{

	u32 sbus = 0, sdev = 0, idsels = 0, ntb;
	struct pci_dev *pdev = NULL, *dev = NULL;
	int src_found = 0, i;


	/*
	 * This will identify the NTB that was being used to indirectly access
	 * the target device. From that, we can determine the bus on which the
	 * target device resides. This is needed to help identify the chassis
	 * slot number of the target device.
	 */
	idsels = (bus_cfg_addr >> 11) & 0x1FFFFF;
	for (i = 0; i < num_devices_entries; i++) {

		if ((devices[i].domain == ddata->domain)&&
			(devices[i].idsel == idsels)) {
				sbus = devices[i].first_bus;
				ntb = devices[i].dev;
				if (verbose) {
					printk(KERN_ERR "%s: Found NTB %d sbus %d idsel 0x%x "
						   "bus_cfg_addr 0x%8.8x ntb_cfg_addr 0x%8.8x\n",
						   __FUNCTION__, ntb, sbus, idsels, bus_cfg_addr,
						   ntb_cfg_addr);
				}
				break;
		}
	}

	/*
	 * This will identify the device being indirectly accessed via the NTB.
	 * The device, along with the bus determined above are used to lookup
	 * the chassis slot number of the target device.
	 */

	idsels = (ntb_cfg_addr >> 11) & 0x1FFFFF;
	for (i = 0; i < num_devices_entries; i++) {

		if ((devices[i].domain == ddata->domain)&&
			(devices[i].first_bus <=sbus)&&
			(devices[i].last_bus >= sbus)) {

			/*
			 * If this bus device has subranges, then search them
			 * to find the exact slot.
			 */
			if (devices[i].bus_subranges.number > 0) {
				dsp_subranges *ptr = NULL;
				int x;

				ptr = &devices[i].bus_subranges;
				dev = devices[i].pdev;

				for (x = 0; (x < ptr->number)&&
						 ( x < BHPC_MAX_DSP_SUBRANGES); x++) {

					if (ptr->subranges[x].idsel == idsels) {
						if (verbose) {
							printk(KERN_ERR "%s x %d: Found subrange "
								   "Dev %s Base 0x%8.8x Limit 0x%8.8x "
								   "slot %d\n", __FUNCTION__, x, pci_name(dev),
								   ptr->subranges[x].base,
								   ptr->subranges[x].limit,
								   ptr->subranges[x].slot);
						}
						ddata->slot = ptr->subranges[x].slot;
						sdev = ptr->subranges[x].devid;
						break;
					}
				}
			} else {
				ddata->slot = devices[i].slot;
			}
			src_found = 1;
			break;
		}
	}

	if (verbose) {
		printk(KERN_ERR "%s: Error Source "
			   "device 0x%4.4x:%2.2x slot %d\n",
			   __FUNCTION__, sbus, sdev, ddata->slot);
	}

	if ((sbus == 0)&&(sdev == 0)) {
		pdev = domain_tbl[ddata->domain].rc_dev;
	} else {
		pdev = lookup_mm_dsp(ddata->domain, sbus, sdev);
	}

	if (ddata) {
		if (pdev) {
			ddata->mm_dsp_dev = pdev;
			ddata->mm_parent_dev.dev = PCI_SLOT(pdev->devfn);
			ddata->mm_parent_dev.bus = pdev->bus->number;
		}
		ddata->src_dev.dev = sdev;
		ddata->src_dev.bus = sbus;
	}

	return (src_found);
}

/*
 * Decode the contents of a PCIe-PCI bridge secondary side header log.
 * This log contains the exact state of the secondary-side PCI bus at the time
 * an error is detected by the bridge.
 *
 * Upon return, the device being accessed when the error occured will be
 * decoded along with the chassis slot that contains that device.
 *
 * Inputs:
 *   deverr - The device status/error registers from the PCIe-PCI bridge
 *   ntpdeverr - The device status/error registers from the NTB on the
 *               secondary side of the bridge
 *   ddata - Contains saved error status/control registers of all bus
 *           devices as well as decoded information.
 */
static int
decode_sec_header_log(bhpc_dev_error *deverr, bhpc_dev_error *ntpdeverr,
					  bhpc_error_data *ddata)
{

	int src_found = 0;
	u32 cbe = 0, ntb_cfg_addr = 0, bus_cfg_addr = 0;


	cbe = (deverr->sec_header_log[1] >> 4 ) & 0xF;
	bus_cfg_addr = deverr->sec_header_log[2];

	printk(KERN_INFO "%s: cbe 0x%x bus_cfg_addr 0x%8.8x\n",
		   __FUNCTION__, cbe, bus_cfg_addr);

	switch (cbe) {

	case 0x06: /* Mem read */
	case 0x07: /* Mem write */

		src_found = decode_mem_err_src(bus_cfg_addr, ddata);
		break;

	case 0x0A: /* Config Read */
	case 0x0B: /* Config Write */

		if ((bus_cfg_addr & 0x1) == 1) {

			/* It's a type 1 address - decode the address value as-is */
			src_found = decode_pci_config1_err_src(bus_cfg_addr, ddata);

		} else {
			/*
			 * If it's a type 0 address, then use the cfg addr
			 * register read from the child NTB. This value
			 * can also be either type 1 or type 0
			 */
			ntb_cfg_addr = ntpdeverr->header_log[0];
			if (ntb_cfg_addr & 0x1) {
				src_found = decode_pci_config1_err_src(ntb_cfg_addr, ddata);
			} else {
				src_found = decode_pci_config0_err_src(ntb_cfg_addr,
													   bus_cfg_addr, ddata);
			}
		}
		break;

	default:
		printk(KERN_ERR "%s: Unexpected PCI command 0x%x\n",
			   __FUNCTION__, cbe);
	}

	return (src_found);
}
static void
print_pcie_pci_bridge_err(bhpc_dev_error *deverr)
{

	printk(KERN_ERR "%s: DETECTED PCIE-PCI BRIDGE DEV %s\n",
		   __FUNCTION__, pci_name(deverr->dev));

	printk("%s: Primary status error 0x%8.8x detected: ",
				   __FUNCTION__, deverr->primary_status);
	print_pci_errors(deverr->primary_status);

	printk("%s: Secondary status error 0x%8.8x "
		   "detected: ", __FUNCTION__, deverr->secondary_status);

	print_pci_errors(deverr->secondary_status);

	printk("%s: Mem Limit/Base 0x%8.8x\n",
		   __FUNCTION__, deverr->mem_window);

	printk("%s: Link Ctrl/Status 0x%8.8x\n",
		   __FUNCTION__, deverr->link_control);

	printk("%s: Uncorr. Error status 0x%8.8x detected: ",
		   __FUNCTION__, deverr->uncorrectable_status);

	print_pci_uncorrectable(deverr->uncorrectable_status);

	printk("%s: Corr. Error status 0x%8.8x detected: ",
		   __FUNCTION__, deverr->correctable_status);

	print_pci_correctable(deverr->correctable_status);

	printk("%s: AER Caps/Ctrl 0x%8.8x\n",
		   __FUNCTION__, deverr->aer_cap);

	printk("%s: LOG0 0x%8.8x LOG4 0x%8.8x LOG8 0x%8.8x LOG12 0x%8.8x\n",
		   __FUNCTION__, deverr->header_log[0], deverr->header_log[1],
		   deverr->header_log[2], deverr->header_log[3]);

	printk(KERN_ERR "%s: Sec Uncorr. Error status 0x%8.8x detected: ",
		   __FUNCTION__, deverr->sec_uncorrectable_status);

	print_pci_sec_uncorrectable(deverr->sec_uncorrectable_status);

	printk(KERN_ERR "%s: SEC Uncorr. Error mask 0x%8.8x\n",
		   __FUNCTION__, deverr->sec_uncorrectable_mask);

	printk(KERN_ERR "%s: SEC AER Caps/Ctrl 0x%8.8x\n",
		   __FUNCTION__, deverr->sec_aer_cap);

	printk(KERN_ERR "%s: SEC LOG0 0x%8.8x LOG4 0x%8.8x LOG8 0x%8.8x LOG12 0x%8.8x\n",
		   __FUNCTION__, deverr->sec_header_log[0], deverr->sec_header_log[1],
		   deverr->sec_header_log[2], deverr->sec_header_log[3]);

}

/*
 * Examine status registers of PCIe switch ports to
 * see if any errors were detected/reported.
 *
 * Inputs:
 *   ddata - Contains the collected status registers of the switch ports
 *
 * Outputs:
 *   N/A
 */
static void
process_pcie_switch_err(bhpc_error_data *ddata)
{
	int i, offset = BHPC_PCIE_PCI_IDX+1, found;
	bhpc_dev_error *deverr = NULL;

	for (i = offset; (i < MAX_DEV_ERROR); i++) {

		deverr = &ddata->dev_err[i];
		if ((deverr)&&(deverr->valid == 1)) {

			/*
			 * There could be more than one error, but we decode
			 * only the first one we find. We still log all of them
			 * just in case.
			 *
			 * Only care about the ones that are unmasked.
			 */
			if ((deverr->uncorrectable_status &
				 (~deverr->uncorrectable_mask))!= 0) {
				ddata->error_detected = 1;

				found = decode_header_log(deverr, ddata);
				if (found) {
					ddata->src_decode = PCIE_ERROR_UNCORR_ERR;
					break;
				}
			} else if (deverr->dev_type == PCI_EXP_TYPE_PCI_BRIDGE) {

				if (deverr->sec_uncorrectable_status != 0) {
					ddata->error_detected = 1;

					if (verbose) {
						print_pcie_pci_bridge_err(deverr);
					}
					found = decode_sec_header_log(deverr,
												  &ddata->dev_err[i+1], ddata);
					if (found) {
						break;
					}
				}

			} /* else if (deverr->dev_type == PCI_EXP_TYPE_PCI_BRIDGE) */

		} /* if ((deverr) ...) */

	} /* for () */

}
/*
 * process_rc_err()
 *
 * Anaylize and decode the RC error/status registers to
 * determine what event/error occured, the source device that
 * generated the event, and the MM DSP port connected
 * to the blade that contains the source device.
 *
 * The root error status register will indicate what events
 * were detected, and the Source ID registers indicate the
 * source device.
 *
 * 1. Read Root error status
 * 2. If bit 0 and/or 1 is set, source is in Correctable source ID register
 * 3. If bit 2 and/or 6 is set, then source is in Non-correctable source
 *    ID register
 *
 * If the error detect register detected an event/error, then
 * set error_decode to True.
 *
 * If the source device is found, then src_decode is set to
 * PCIE_ERROR_ERR_DETECT.
 *
 * This should always be the second place we look to decode an
 * event/error after the PCIe controller registers. Even if the
 * error has been detected and decoded via the controller error
 * detect registers, we still want to see the RC error status
 * registers
 *
 * Inputs:
 *   rdev - pci_dev object of the RC
 *   ddata - Contains all the register data to be analized.
 *           Also used to pass back the ID of source device and parent
 *           MM DSP
 *
 * Returns:
 *  N/A
 *
 */

static void
process_rc_err(struct pci_dev *rdev, bhpc_error_data *ddata)
{

	bhpc_dev_error *deverr;
	int src_found = 0;

	deverr = &ddata->dev_err[BHPC_RC_IDX];

	if (deverr->root_err_status & BHPC_ERROR_NON_COR) {

		ddata->error_detected = 1;

		if (verbose) {
			printk("%s %d: UN Correctable errors 0x%8.8x: ",
				   __FUNCTION__, __LINE__, deverr->uncorrectable_status);

			print_pci_uncorrectable(deverr->uncorrectable_status);
		}
		if (!ddata->mm_dsp_dev) {
			if (deverr->uncorr_err_src) {
				src_found = decode_bus_err_src(deverr->uncorr_err_src, ddata);
				if (src_found) {
					ddata->src_decode = PCIE_ERROR_RC_UNCORR_ERR_SRC;
				}
			} else if (deverr->uncorrectable_status) {
				src_found = decode_header_log(deverr, ddata);
				if (src_found) {
					ddata->src_decode = PCIE_ERROR_RC_UNCORR_ERR_HLOG;
				}
			}
		}
	}

	if (deverr->root_err_status & BHPC_ERROR_COR) {
		ddata->error_detected = 1;
		if (verbose) {
			printk("%s %d: Correctable errors 0x%8.8x ",
				   __FUNCTION__, __LINE__, deverr->correctable_status);
			print_pci_correctable(deverr->correctable_status);
		}
		if (!ddata->mm_dsp_dev) {
			src_found = decode_bus_err_src(deverr->corr_err_src, ddata);
			if (src_found) {
				ddata->src_decode = PCIE_ERROR_RC_CORR_ERR_SRC;
			}
		}
	}

	if (verbose) {
		printk(KERN_ERR "%s: src_decode %d err idx %d\n", __FUNCTION__,
			   ddata->src_decode, ddata->error_idx);
		display_decoded_header_log(pci_name(rdev), deverr);
	}

}


/*
 *  process_pcie_errors()
 *
 *  Collect and clear the error control/status registers from the PCIe
 *  controller and the RC. Then analyze the registers to determine
 *  what happened, the source device that generated the event,
 *  and the MM DSP that connects to the blade containing the source
 *  device.
 *
 *  1. Collect/clear the registers.
 *  2. Determine the type of error.
 *     1. Check ERR_DR
 *     2. Check RC Correctable error status
 *     3. Check RC UnCorrectable error status
 *     4. Check RC Secondary status
 *  3. Determine the source device of the error based on error type.
 *  4. Progress up the MM device hierarchy cleaning up the status.
 *  5. Look up and save the logical slot number of the blade containing
 *     the source device.
 *
 * Returns:
 *  0 = No error was detected
 *  1 = An error was detected
 */
static void
process_pcie_errors(volatile void __iomem *vaddr, bhpc_error_data *ddata)
{

	/*
	 * Error source is determined 1 of 2 ways:
	 * 1. Via the PCI Express Error Detect register
	 *    and associated registers.
	 * 2. Root Complex Error source ID registers
	 */


	/* 1. First, collect the registers and clear them */

	collect_clear_pex_errors(vaddr, ddata, 1);
	collect_clear_rc_errors(domain_tbl[ddata->domain].rc_dev, ddata, 1);
	collect_clear_pcie_switch_errors(ddata);

	ddata->src_decode = PCIE_ERROR_NONE;

	/*
	 * 2-5. Analyze the registers to determine what happened and where,
	 *      and determine the chassis slot, if possible
	 *
	 * The order matters, so don't change the sequence.
	 */
	process_err_detect(ddata);
	process_rc_err(domain_tbl[ddata->domain].rc_dev, ddata);
	process_pcie_switch_err(ddata);
	decode_rc_legacy_err(domain_tbl[ddata->domain].rc_dev, ddata);

	if ((!ddata->error_detected)&&(verbose)) {
		printk(KERN_ERR "%s %d: No errors found\n", __FUNCTION__, __LINE__);
		show_configs_rc(domain_tbl[ddata->domain].rc_dev, "RC", 0);
	}
}

/*
 * clear_pcie_errors()
 *
 * Walks the entire PCI dev hierarchy from the RC on down
 * clearing the error status registers of each device. The PCIe
 * controller memory-mapped error detect status registers are
 * also cleared
 *
 * This is called during start-up init or when re-enable IRQs is requested
 * (via /proc file system) to clean up any "stale" errors.
 *
 *
 * Inputs:
 *  ddata - Contains all the saved error status/control registers
 *  direct - True = clear registers based on current register contents
 *           False = clear registers based on cached values in ddata.
 *
 */
static void
clear_pcie_errors(bhpc_error_data *ddata, int direct)
{
	u32 regval = 0, type;
	int dev_domain;
	struct pci_dev *dev = NULL;
	void volatile __iomem *regaddr = NULL;
	bhpc_dev_error *deverr = NULL;
	volatile void __iomem *pex_base_addr = NULL;

	for_each_pci_dev(dev) {

		dev_domain = pci_domain_nr(dev->bus);
		if (dev_domain != ddata->domain) {
			continue;
		}

		type = get_port_type(dev);
		switch (type) {

		case PCI_EXP_TYPE_ENDPOINT:
			deverr = &ddata->dev_err[BHPC_ENDPOINT_IDX];
			deverr->dev_type = PCI_EXP_TYPE_ENDPOINT;
			clear_pcie_errors_endpoints(dev, "ENDPOINT", deverr, direct);
			break;

		case PCI_EXP_TYPE_ROOT_PORT:
			deverr = &ddata->dev_err[BHPC_RC_IDX];
			deverr->dev_type = PCI_EXP_TYPE_ROOT_PORT;
			clear_pcie_errors_rc(dev, "RC", deverr, direct);
			break;

		case PCI_EXP_TYPE_UPSTREAM:

			deverr = &ddata->dev_err[BHPC_MM_USP_IDX];
			deverr->dev_type = PCI_EXP_TYPE_UPSTREAM;

			if (dev->bus->number == 1) {
				clear_pcie_errors_bridges(dev, "MM USP", deverr, direct);
			} else {
				clear_pcie_errors_bridges(dev, "SFM USP", deverr, direct);
			}
			break;

		case PCI_EXP_TYPE_DOWNSTREAM:

			deverr = &ddata->dev_err[BHPC_MM_DSP_IDX];
			deverr->dev_type = PCI_EXP_TYPE_DOWNSTREAM;

			if (dev->bus->number == 2) {
				clear_pcie_errors_bridges(dev, "MM DSP", deverr, direct);
			} else {
				clear_pcie_errors_bridges(dev, "SFM DSP", deverr, direct);
			}
			break;

		case PCI_EXP_TYPE_PCI_BRIDGE:
			deverr = &ddata->dev_err[BHPC_PCIE_PCI_IDX];
			deverr->dev_type = PCI_EXP_TYPE_PCI_BRIDGE;
			clear_pcie_errors_pci_bridges(dev, "PCIE-PCI", deverr, direct);

			break;

		case PCI_TYPE_NTB:
			clear_pci_errors_ntb(dev, "PCI NTB", NULL, direct);
			break;

		default:
			printk(KERN_ERR "%s Unrecognized device type %d %s\n",
				   __FUNCTION__, type, pci_name(dev));
		}

	}

	pex_base_addr = domain_tbl[ddata->domain].pex_base_addr;

	regaddr = pex_base_addr + 0xe00; /* ERR_DR */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);

	regaddr = pex_base_addr + 0xe20; /* ERR_CAP_STAT */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);
	writel_be(0, regaddr);

	regaddr = pex_base_addr + 0xe28; /* ERR_CAP_STAT0 */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);
	writel_be(0, regaddr);

	regaddr = pex_base_addr + 0xe2c; /* ERR_CAP_STAT1 */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);
	writel_be(0, regaddr);

	regaddr = pex_base_addr + 0xe30; /* ERR_CAP_STAT2 */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);
	writel_be(0, regaddr);

	regaddr = pex_base_addr + 0xe34; /* ERR_CAP_STAT3 */
	regval = readl_be(regaddr);
	writel_be(regval, regaddr);
	writel_be(0, regaddr);

}
/*
 * setup_pcie()
 *
 * Enable or Disable PCIe event/error detection via the PCIe controller
 * error detect registers.
 *
 * Inputs;
 *   vaddr - ioremap'd base address of the PCIe controller error detect registers
 *   enable - True = Enable error detect. False = disable error detect
 *
 */
static void
setup_pcie(volatile void __iomem *vaddr, int enable)
{
	volatile void __iomem *regaddr;
	u32 rege08, rege10;

	if (enable) {

		/* Do this to enable PCAC interrupts */

		regaddr = vaddr + 0xe08;
		rege08 = readl_be(regaddr);

		rege08 |= 0x00200000;
		writel_be(rege08, regaddr);

		regaddr = vaddr + 0xe10;
		rege10 = readl_be(regaddr);

		rege10 &= ~(0x00200000);
		writel_be(rege10, regaddr);

		regaddr = vaddr + 0xe08;
		rege08 = readl_be(regaddr);
		regaddr = vaddr + 0xe10;
		rege10 = readl_be(regaddr);

		if (verbose) {
			printk("%s %d: After enable PCAC ERR_EN 0x%8.8x "
				   "ERR_DISR 0x%8.8x \n", __FUNCTION__, __LINE__,
				   rege08, rege10);
		}

	} else {

		/* Do this to disable PCAC interrupts. */
		regaddr = vaddr + 0xe08;
		rege08 = readl_be(regaddr);

		rege08 &= ~(0x00200000);
		writel_be(rege08, regaddr);

		regaddr = vaddr + 0xe10;
		rege10 = readl_be(regaddr);

		rege10 |= 0x00200000;
		writel_be(rege10, regaddr);

		regaddr = vaddr + 0xe08;
		rege08 = readl_be(regaddr);
		regaddr = vaddr + 0xe10;
		rege10 = readl_be(regaddr);

		if (verbose) {
			printk("%s %d: After disable PCAC ERR_EN 0x%8.8x "
				   "ERR_DISR 0x%8.8x \n",
				   __FUNCTION__, __LINE__, rege08, rege10);
		}
	}

}
/*
 * bhpc_disable_scan_error_reports()
 *
 * When PCIe error reporting is enabled, master aborts occuring
 * during PCI config scan generate PCIe error interrupts. Since
 * master aborts during config scans are normal (and there are
 * a large number of them), we need to disable the error
 * reporting to prevent error interrupts from being generated.
 *
 * The only way to do this is apparently via the RMA bit in the
 * vendor-specific Secondary Status Interrupt Mask config
 * register in the RC at offset 0x5A0.
 *
 * There does not appear to be anything that can be done on the
 * MM DSP to prevent a master abort from being reported to the RC.
 *
 * A side-affect is that no master aborts occuring anywhere in
 * the system for any reason will be detected while this bit is
 * masked. So it needs to be masked as briefly as possible.
 *
 * Inputs:
 *   dev - pci_dev object of the MM DSP connected to the
 *         blade that is going to be scanned.
 *
 */
void
bhpc_disable_scan_error_reports(struct pci_dev *dev)
{
	u_int devfn;
	struct pci_dev *rdev;
	u32 regval = 0;
	int domain = 0;
	volatile void __iomem *pex_base_addr = NULL;

	pr_debug("%s: Disable scan error reporting for dev %s\n",
		   __FUNCTION__, pci_name(dev));

	/* Need the root port for this device's domain */
	domain = pci_domain_nr(dev->bus);
	devfn = PCI_DEVFN(0, 0);
	rdev = pci_get_domain_bus_and_slot(domain, 0, devfn);
	if (rdev) {

		/* This is all that's needed to prevent RMA IRQs during scan */
		pci_read_config_dword(rdev, 0x5A0, &regval);
		/*
		 * With Yoda CPs, we still get some PCIe errors during
		 * scan even with RMA masked. So mask everything for now.
		 * Try going back to this when the new Yoda's arrive:
		 * regval |= 0x08;
		 */
		regval |= 0x0F;
		pci_write_config_dword(rdev, 0x5A0, regval);
		pr_debug("  SecStatusMask 0x%8.8x %s\n", regval, pci_name(rdev));

		/* Also mask completion timeout in the RC */
		pex_base_addr = domain_tbl[domain].pex_base_addr;
		if (pex_base_addr) {
			volatile void __iomem *regaddr;
			u32 rege08, rege10;

			/* Disable completion timeout interrupt */
			regaddr = pex_base_addr + 0xe08;
			rege08 = readl_be(regaddr);
			rege08 &= ~(0x00800000);
			writel_be(rege08, regaddr);

			/* Disable completion timeout detection */
			regaddr = pex_base_addr + 0xe10;
			rege10 = readl_be(regaddr);
			rege10 |= 0x00800000;
			writel_be(rege10, regaddr);
		}
	}

}
/*
 * bhpc_enable_scan_error_reports()
 *
 * Re-enable master abort error reporting via the RC.
 *
 * To enable IRQs for master aborts after scan, must:
 *   1. Clear any residual errors from RC Primary and Secondary status
 *   2. Re-enable RMA reporting via 0x5A0 on RC
 *
 * Inputs:
 *   dev - pci_dev object of the MM DSP connected to the
 *         blade that is going to be scanned.
 *
 */
void
bhpc_enable_scan_error_reports(struct pci_dev *dev)
{

	u_int devfn;
	struct pci_dev *rdev;
	u32 regval = 0;
	int domain = 0;
	volatile void __iomem *pex_base_addr = NULL;

	pr_debug("%s: Enable scan error reporting for dev %s\n",
		   __FUNCTION__, pci_name(dev));

	domain = pci_domain_nr(dev->bus);
	devfn = PCI_DEVFN(0, 0);
	rdev = pci_get_domain_bus_and_slot(domain, 0, devfn);
	if (rdev) {

		/* Clear any residual errors before enabling error reporting */
		pci_read_config_dword(rdev, PCI_COMMAND, &regval); /* Primary status */
		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: RC Primary status error 0x%8.8x detected: ",
					   __FUNCTION__, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(rdev, PCI_COMMAND, regval);
		}
		/* Secondary status is in upper 16 bits (0x1E) */
		pci_read_config_dword(rdev, PCI_IO_BASE, &regval);
		if (regval & 0xF9000000) {
			if (verbose) {
				printk("%s: RC Secondary status error 0x%8.8x detected: ",
					   __FUNCTION__, regval);
				print_pci_errors(regval);
			}
			pci_write_config_dword(rdev, PCI_IO_BASE, regval);
		}

		/* Unmask secondary side master abort reporting */
		pci_read_config_dword(rdev, 0x5A0, &regval);
		/*
		 * With Yoda CPs, we still get some PCIe errors during
		 * scan even with RMA masked. So mask everything for now.
		 * Try going back to this when the new Yoda's arrive:
		 * regval &= ~0x08;
		 */
		regval &= ~0x0F;
		pci_write_config_dword(rdev, 0x5A0, regval);
		if (verbose) {
			printk("  SecStatusMask 0x%8.8x (Dev %s)\n", regval, pci_name(rdev));
		}

		/* Also renable completion timeout in the RC */
		pex_base_addr = domain_tbl[domain].pex_base_addr;
		if (pex_base_addr) {
			volatile void __iomem *regaddr;
			u32 rege08, rege00, rege10;

			regaddr = pex_base_addr + 0xe00;
			rege00 = readl_be(regaddr);
			if (rege00 & 0x008000000) {
				printk(KERN_ERR "%s: PCT 0x%8.8x was detected: ",
					   __FUNCTION__, rege00);
				
				writel_be(rege00, regaddr);
			}

			/* Enable completion timeout IRQ */
			regaddr = pex_base_addr + 0xe08;			
			rege08 = readl_be(regaddr);
			rege08 |= 0x00800000;
			writel_be(rege08, regaddr);

			/* Enable completion timeout detection */
			regaddr = pex_base_addr + 0xe10;			
			rege10 = readl_be(regaddr);
			rege10 &= ~(0x00800000);
			writel_be(rege10, regaddr);
		}

	}
}

/*
 * setup_configs_pci_ntb()
 *
 *  Setup PCI NTB devices to report PCI errors.
 *
 *
 * Inputs:
 *   dev - pci dev object of the endpoint
 *   name - String to display when in verbose mode.
 *
 */
static void
setup_configs_pci_ntb(struct pci_dev *dev, char *name)
{
	u32 regval32 = 0;

	if (dev) {
		if (verbose) {
			printk(KERN_ERR "%s: SETUP %s REGISTERS %s\n",
				   __FUNCTION__, name, pci_name(dev));
		}
		pci_read_config_dword(dev, PCI_NTB_CHIP_CONTROL, &regval32);
		regval32 |= 0x81;
		pci_write_config_dword(dev, PCI_NTB_CHIP_CONTROL, regval32);
	}
}
/*
 * setup_configs_endpoints()
 *
 *  Setup endpoint device to report PCIe errors.
 *
 *   Primary Command register:
 *     Enable SERR, PAERR in Primary Command register
 *
 *   Device Control register:
 *     Enable Non-fatal, fatal, and URE reporting in
 *
 *
 * Inputs:
 *   dev - pci dev object of the endpoint
 *   name - String to display when in verbose mode.
 *
 */
static void
setup_configs_endpoints(struct pci_dev *dev, char *name)
{
	u32 regval = 0;
	int pos;

	if (dev) {
		if (verbose) {
			printk(KERN_ERR "%s: SETUP %s REGISTERS %s\n",
				   __FUNCTION__, name, pci_name(dev));
		}
		  /* Primary Command/Status */
		pci_read_config_dword(dev, PCI_COMMAND, &regval);
		regval |= 0x0146; /* Enable SERR, PAERR, Bus Master, Mem decode */
		pci_write_config_dword(dev, PCI_COMMAND, regval);

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {
			/* Device Control/status */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &regval);
			regval |= 0xf; /* Report non-fatal, fatal, and unsupported request */
			pci_write_config_dword(dev, pos + PCI_EXP_DEVCTL, regval);
		}
	}
}
/*
 * setup_configs_ports()
 *
 *  Setup switch port devices to report PCIe errors.
 *
 *   Primary Command register:
 *     Enable SERR, PAERR in Primary Command register
 *
 *   Device Control register:
 *     Enable Non-fatal, fatal, and URE reporting in
 *
 *   Bridge Control:
 *     Enable SERR and PAERR
 *
 *   Correctable Err mask:
 *     Unmask ANFE
 *
 * Inputs:
 *   dev - pci dev object of the switch port
 *   name - String to display when in verbose mode.
 *
 */
static void
setup_configs_ports(struct pci_dev *dev, char *name)
{
	u32 regval = 0, pos;
	u16 regval16 = 0;

	if ((dev)&&(dev->subordinate)) {

		/* Primary Command/Status */
		pci_read_config_dword(dev, PCI_COMMAND, &regval);
		regval |= 0x0140; /* Enable SERR and PAERR */
		pci_write_config_dword(dev, PCI_COMMAND, regval);

		/* Secondary status, W1C the status bits */
		pci_read_config_word(dev, PCI_SEC_STATUS, &regval16);
		pci_write_config_word(dev, PCI_SEC_STATUS, regval16);

		/* Bridge Control */
		pci_read_config_word(dev, PCI_BRIDGE_CONTROL, &regval16);
		regval16 |= (PCI_BRIDGE_CTL_PARITY | PCI_BRIDGE_CTL_SERR);
		pci_write_config_word(dev, PCI_BRIDGE_CONTROL, regval16);

		dev->subordinate->bridge_ctl = regval16;
		pci_read_config_word(dev, PCI_BRIDGE_CONTROL, &regval16);

		if (verbose) {
			printk(KERN_ERR "%s: BRIDGE CONTROL 0x%4.4x (0x%4.4x) for dev %s\n",
				   __FUNCTION__, regval16, dev->subordinate->bridge_ctl,
				   pci_name(dev));
		}

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);

		 /* Device Control/status */
		pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &regval);
		regval |= 0xf; /* Report non-fatal, fatal, and unsupported request */
		pci_write_config_dword(dev, pos + PCI_EXP_DEVCTL, regval);


		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);

		/*
		 * Uncorr err mask - Mask Surprise down. FOS/NOS
		 * hot plug code handles this.
		 */
		pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_MASK, &regval);
		regval |= 0x20;
		pci_write_config_dword(dev, pos + PCI_ERR_UNCOR_MASK, regval);

		/* Corr err mask - Unmask ANFE */
		pci_read_config_dword(dev, pos + PCI_ERR_COR_MASK, &regval);
		regval &= ~0x00002000;
		pci_write_config_dword(dev, pos + PCI_ERR_COR_MASK, regval);
	} else {
		printk(KERN_ERR "%s: Invalid device\n", __FUNCTION__);
	}

}

/*
 * setup_configs_pci_bridges()
 *
 *  Setup PCIE-PCI bridge devices to report PCIe errors.
 *
 *   Primary Command register:
 *     Enable SERR, PAERR in Primary Command register
 *
 *   Device Control register:
 *     Enable Non-fatal, fatal, and URE reporting in
 *
 *   Bridge Control:
 *     Enable SERR and PAERR
 *
 *   Correctable Err mask:
 *     Unmask ANFE
 *
 *   Secondary Bridge Control:
 *     Enable master abort reporting.
 *
 *   Secondary UCORR mask:
 *     Unmask everything
 *
 * Inputs:
 *   dev - pci dev object of the bridge
 *   name - String to display when in verbose mode.
 *
 */
static void
setup_configs_pci_bridges(struct pci_dev *dev, char *name)
{
	u32 regval = 0, pos;
	u16 regval16 = 0;

	/* First, setup the primary side */
	setup_configs_ports(dev, name);

	if (verbose) {
		printk(KERN_ERR "%s: Dev %s\n", __FUNCTION__, pci_name(dev));
	}

	/* Bridge Control */
	pci_read_config_word(dev, PCI_BRIDGE_CONTROL, &regval16);
	regval16 |= PCI_BRIDGE_CTL_MASTER_ABORT; /* Enable Master abort reporting */
	pci_write_config_word(dev, PCI_BRIDGE_CONTROL, regval16);

	pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
	if (pos) {

		regval = 0;
		pci_write_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_MASK, regval);

	}
}
/*
 * setup_configs_rc()
 *
 *  Setup Root Complex to report PCIe errors.
 *
 *   Primary Command register:
 *     Enable SERR, PAERR in Primary Command register
 *
 *   Device Control register:
 *     Enable Non-fatal, fatal, and URE reporting in
 *
 *   Bridge Control:
 *     Enable SERR and PAERR
 *
 *   Root Control:
 *     Enable system error on fatal and non-fatal errors
 *
 *   Root Error Command:
 *     Enable Fatal and non-fatal error reporting
 *
 *   Secondary Status Interrupt Mask Register:
 *     Unmask all bits
 *
 * Inputs:
 *   dev - pci dev object of the RC
 *   enable_legacy - True = enable legacy PCI error reporting
 *   name - String to display when in verbose mode.
 *
 */
static void
setup_configs_rc(struct pci_dev *dev, int enable_legacy, char *name)
{
	u32 regval = 0, pos;
	u16 regval16;

	if (dev) {

		if ((verbose)&&(dev->bus)) {
			struct pci_controller *hose;

			hose = pci_bus_to_host(dev->bus);
			if (hose) {
				printk(KERN_ERR "%s %d Hose mem resources offset 0x%llx "
					   "start 0x%llx end 0x%llx\n",
						 __FUNCTION__, __LINE__, hose->pci_mem_offset,
						 hose->mem_resources[0].start,
						 hose->mem_resources[0].end);
			}
		}

		/* Primary Command/Status */
		pci_read_config_word(dev, PCI_COMMAND, &regval16);
		/*
		 * Omit SERR for now due to spurious errors seen in
		 * some boards. Restore this if/when those are understood.
		 * regval |= (PCI_COMMAND_PARITY | PCI_COMMAND_SERR); // Enable SERR and PAERR
		 */
		regval16 |= PCI_COMMAND_PARITY;	 /* Enable PAERR */
		pci_write_config_word(dev, PCI_COMMAND, regval16);

		/* Secondary status is in upper 16 bits (0x1E) */
		pci_read_config_word(dev, PCI_SEC_STATUS, &regval16);
		pci_write_config_word(dev, PCI_SEC_STATUS, regval16); /* W1C the status bits */

		/* Bridge Control is in Upper word 0x3E */
		pci_read_config_word(dev, PCI_BRIDGE_CONTROL, &regval16);
		regval16 |= (PCI_BRIDGE_CTL_PARITY | PCI_BRIDGE_CTL_SERR);
		pci_write_config_word(dev, PCI_BRIDGE_CONTROL, regval16);

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {

			/* Device Control/status */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &regval);

			/* Report non-fatal, fatal, and unsupported request */
			regval |= (PCI_EXP_DEVCTL_NFERE | PCI_EXP_DEVCTL_FERE | PCI_EXP_DEVCTL_URRE);
			pci_write_config_dword(dev, pos + PCI_EXP_DEVCTL, regval);

			pci_read_config_dword(dev,  pos + PCI_EXP_RTCTL, &regval); /* Root Control */

			/* System error on non-fatal and fatal errors */
			regval |= (PCI_EXP_RTCTL_SENFEE | PCI_EXP_RTCTL_SEFEE);
			pci_write_config_dword(dev, pos + PCI_EXP_RTCTL, regval);
		}

		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
		if (pos) {

			/* Uncorr. Err Status */
			pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, &regval);
			/* W1C all bits */
			pci_write_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, regval);

			/* Uncorr. Err Mask */
			pci_write_config_dword(dev, pos + PCI_ERR_UNCOR_MASK, 0);


			/* Root Err Cmd */
			pci_read_config_dword(dev, pos + PCI_ERR_ROOT_COMMAND, &regval);

			/* Report non-fatal and fatal errors */
			regval |= (PCI_ERR_ROOT_CMD_FATAL_EN | PCI_ERR_ROOT_CMD_NONFATAL_EN);
			pci_write_config_dword(dev, pos + PCI_ERR_ROOT_COMMAND, regval);

			/* Root Err status */
			pci_read_config_dword(dev, pos + PCI_ERR_ROOT_STATUS, &regval);
			/* W1C all bits */
			pci_write_config_dword(dev, pos + PCI_ERR_ROOT_STATUS, regval);

		}

		if (enable_legacy) {
			/*
			 * Secondary status mask. Enable all for legacy PCI err reporting
			 * This is device specific register for FSL p40x0 CPU
			 */

			pci_write_config_dword(dev, 0x5A0, 0);

		} else {

			/* Do this to disable legacy PCI error reporting */
			pci_write_config_dword(dev, 0x5A0, 0xffffffff);
		}



	}
}
/*
 * bhpc_setup_configs()
 *
 *   Walk down through a device hierarchy and enable error
 *   reporting for each device
 *
 * Inputs:
 *   dev - pci device object for the device at the top of
 *         the hierarchy. If NULL, start with the RC
 *   enable_legacy - enable legacy PCI error reporting
 *
 * NOTE: This function assumes only one PCIe controller.
 */
void
bhpc_setup_configs(struct pci_dev *dev, int domain, int enable_legacy)
{
	u32 type;
	struct pci_dev *tdev = NULL;
	int dev_domain;


	if (dev != NULL) {
		tdev = dev;
		if (verbose) {
			printk("%s: Called for dev %s\n", __FUNCTION__,
				   pci_name(tdev));
		}
	}

	for_each_pci_dev(dev) {

		dev_domain = pci_domain_nr(dev->bus);
		if (domain != dev_domain) {
			continue;
		}
		type = get_port_type(dev);
		switch (type) {

		case PCI_EXP_TYPE_ENDPOINT:
			setup_configs_endpoints(dev, "ENDPOINT");
			break;

		case PCI_EXP_TYPE_ROOT_PORT:
			setup_configs_rc(dev, enable_legacy, "RC");
			break;

		case PCI_EXP_TYPE_UPSTREAM:
			if (dev->bus->number == 1) {
				setup_configs_ports(dev, "MM USP");
			} else {
				setup_configs_ports(dev, "SFM USP");
			}
			break;

		case PCI_EXP_TYPE_DOWNSTREAM:
			if (dev->bus->number == 2) {
				setup_configs_ports(dev, "MM DSP");
			} else {
				setup_configs_ports(dev, "SFM DSP");
			}
			break;

		case PCI_EXP_TYPE_PCI_BRIDGE:
			setup_configs_pci_bridges(dev, "PCIE-PCI");
			break;

		case PCI_TYPE_NTB:
			setup_configs_pci_ntb(dev, "PCI-NTB");
			break;

		default:
			printk(KERN_INFO "%s Unrecognized device type %d %s\n",
				   __FUNCTION__, type, pci_name(dev));
		}

	}

	/*
	 * Need to restore the refcount of the dev that
	 * was passed in, if there was one.
	 */
	if (tdev) {
		pci_dev_get(tdev);
	}
}
/*
 * show_bridge_mem_window()
 *
 *   Display the memory resources provided by a bridge
 *
 * Inputs:
 *   dev - pci_dev object of the bridge.
 *
 */
static void
show_bridge_mem_window(struct pci_dev *dev)
{
	u16 config[2];

	pci_read_config_word(dev, PCI_MEMORY_BASE, &config[0]);
	pci_read_config_word(dev, PCI_MEMORY_LIMIT, &config[1]);

	printk("  PCIBusMemRange: 0x%x - 0x%x",
		   (int)(config[0] << 16), (int)(config[1] << 16));

	printk("  PhysMemRange: 0x%llx - 0x%llx\n",
		   dev->resource[8].start, dev->resource[8].end);

}
/*
 * show_bridge_bus_window()
 *
 *   Display the bus resources provided by a bridge
 *
 * Inputs:
 *   dev - pci_dev object of the bridge.
 *
 */
static void
show_bridge_bus_window(struct pci_dev *dev)
{
	u8 config[3];

	pci_read_config_byte(dev, PCI_PRIMARY_BUS, &config[0]);
	pci_read_config_byte(dev, PCI_SECONDARY_BUS, &config[1]);
	pci_read_config_byte(dev, PCI_SUBORDINATE_BUS, &config[2]);

	printk("  Bus Range 0x%x to 0x%x\n", config[1], config[2]);

}
/*
 * show_configs_rc()
 *
 *   Display the error control/status registers of the RC
 *
 * intputs:
 *   rdev - pci_dev object of the RC
 *   name - String to display
 *   verbose - True = Display all the registers. False = Display
 *             only the registers that currently contain error status.
 *
 */
static void
show_configs_rc(struct pci_dev *rdev, char *name, int verbose)
{
	u32 errors_detected = 0;

	if (rdev) {

		memset(&config, 0, sizeof(config));
		printk("%s: %s REGISTERS %s\n",
			   __FUNCTION__, name, pci_name(rdev));
		pci_read_config_dword(rdev, PCI_COMMAND, &config[0]);  /* Status */
		pci_read_config_dword(rdev, 0x1C, &config[1]); /* Secondary status */
		pci_read_config_dword(rdev, 0x3C, &config[2]); /* Bridge Control */

		/* Device Control/status */
		pci_read_config_dword(rdev, 0x54, &config[3]);
		pci_read_config_dword(rdev, 0x68, &config[4]); /* Root Control */
		pci_read_config_dword(rdev, 0x74, &config[5]); /* Dev Ctrl 2 */
		pci_read_config_dword(rdev, 0x104, &config[6]); /* Uncorr. Err Status */
		pci_read_config_dword(rdev, 0x108, &config[7]); /* Uncorr. Err Mask */
		pci_read_config_dword(rdev, 0x10C, &config[8]); /* Uncorr. Err Sev */

		pci_read_config_dword(rdev, 0x11C, &config[9]); /* PCIe Header Log0 */
		pci_read_config_dword(rdev, 0x120, &config[10]); /* PCIe Header Log4 */
		pci_read_config_dword(rdev, 0x124, &config[11]); /* PCIe Header Log8 */
		pci_read_config_dword(rdev, 0x128, &config[12]); /* PCIe Header Log12 */

		pci_read_config_dword(rdev, 0x12C, &config[13]); /* Root Err Cmd */
		pci_read_config_dword(rdev, 0x130, &config[14]); /* Root Err status */
		pci_read_config_dword(rdev, 0x134, &config[15]); /* ErrSrcID */

		/* Secondary status mask */
		pci_read_config_dword(rdev, 0x5A0, &config[16]);

		pci_read_config_dword(rdev, 0x110, &config[17]); // Corr. Err Status
		pci_read_config_dword(rdev, 0x114, &config[18]); // Corr. Err Mask
		pci_read_config_dword(rdev, 0x118, &config[19]); // AECC

		pci_read_config_dword(rdev, 0x5C, &config[20]); // Link Status/Control
		pci_read_config_dword(rdev, 0x64, &config[21]); // Slot Status/Control

		pci_read_config_dword(rdev, 0x404, &config[22]); // RC LTSSM State

		errors_detected = ((config[0] & 0xF9000000)||(config[1] & 0xF9000000)||
						   config[3] & 0x000F0000||
						   config[6]||config[14]||config[15]||config[9]||
						   config[10]||config[11]||config[12]||config[17]);

		if (errors_detected || (verbose)) {

			printk("  PriStatus 0x%8.8x SecStatus 0x%8.8x "
				   "BridgeCtrl 0x%8.8x DevCtrl 0x%8.8x LinkCtrl 0x%8.8x SlotCtrl 0x%8.8x\n",
				   config[0], config[1], config[2], config[3], config[20], config[21]);

			printk("  RootCtrl 0x%8.8x DevCtrl2 0x%8.8x "
				   "UncorrErrStatus 0x%8.8x UncorrErrMask 0x%8.8x LTSSM 0x%x\n",
				   config[4], config[5], config[6], config[7], config[22]);

			printk("  UncorrErrSev 0x%8.8x RootErrCmd 0x%8.8x "
				   "RootErrStatus 0x%8.8x SecStatusMask 0x%8.8x\n",
				   config[8], config[13], config[14], config[16]);

			printk("  CorrErrStatus 0x%8.8x CorrErrMask 0x%8.8x "
				   "AE Capabilities & Ctrl 0x%8.8x\n",
				   config[17], config[18], config[19]);

			printk("  ErrSrcID 0x%8.8x  UncorrErrSrcID 0x%4.4x "
				   "CorrErrSrcId 0x%4.4x\n", config[15],
				   ((config[15] >> 16) & 0x0000FFFF),
				   ((config[15] & 0x0000FFFF)));

			printk("  Log0 0x%8.8x log4 0x%8.8x log8 0x%8.8x "
				   "log12 0x%8.8x\n", config[9], config[10], config[11],
				   config[12]);

			if (config[0] & 0xF9000000) {
				printk("  RC Primary status error 0x%8.8x "
					   "detected: ", config[0]);
				print_pci_errors(config[0]);
			}

			if (config[1] & 0xF9000000) {
				printk("  RC Secondary status error 0x%8.8x "
					   "detected: ", config[1]);
				print_pci_errors(config[1]);
			}


			if (config[3] & 0x000F0000) {
				printk("  %s Device status error 0x%8.8x detected: ",
					   name, config[3]);
				print_devstatus_errors(config[3]);
			}

			if (config[17]) {
				printk("  RC Correctable error 0x%8.8x detected: ",
					   config[17]);
				print_pci_correctable(config[17]);
			}

			if (config[6]) {
				printk("  RC Uncorrectable error 0x%8.8x detected: ",
					   config[6]);
				print_pci_uncorrectable(config[6]);
			}
			if (config[14]) {
				printk("  RC Root error 0x%8.8x detected: ",
					   config[14]);
				print_pci_root_errors(config[14]);
			}
			if (config[21]) {
				printk("  RC Slot status 0x%8.8x detected: ",
					   config[21]);
				print_pci_slot_errors(config[21]);
			}

			if (config[15]) {
				int domain;

				domain =  pci_domain_nr(rdev->bus);
				display_src_ids(domain, config[15]);
			}

			show_bridge_mem_window(rdev);
			show_bridge_bus_window(rdev);
		}
	}
}
/*
 * show_configs_bridges()
 *
 *   Display the error control/status registers of a bridge device
 *
 * intputs:
 *   rdev - pci_dev object of the bridge device
 *   name - String to display
 *   verbose - True = Display all the registers. False = Display
 *             only the registers that currently contain error status.
 *
 */
static void
show_configs_bridges(struct pci_dev *dev, char *name, int verbose)
{
	uint pos, errors_detected = 0;

	if (dev) {
		memset(&config, 0, sizeof(config));
		pci_read_config_dword(dev, PCI_COMMAND, &config[0]);  /* Status */
		pci_read_config_dword(dev, 0x1C, &config[1]); /* Secondary status */
		pci_read_config_dword(dev, 0x3C, &config[2]); /* Bridge Control */

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {
			/* Device Control/status */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &config[3]);

			/* Device Control/status 2 */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL2, &config[4]);

			/* Link Control/status */
			pci_read_config_dword(dev, pos + PCI_EXP_LNKCTL, &config[20]);

		}

		pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
		if (pos) {
			/* Uncorr. Err Status */
			pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS, &config[6]);
			/* Uncorr. Err Mask */
			pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_MASK, &config[7]);
			/* Uncorr. Err Sev */
			pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_SEVER, &config[8]);

			/* Correctable Error Status */
			pci_read_config_dword(dev, pos + PCI_ERR_COR_STATUS, &config[9]);
			/* Correctable Err Mask */
			pci_read_config_dword(dev, pos + PCI_ERR_COR_MASK, &config[10]);

			/* Advanced Error Capabilites and Ctrl */
			pci_read_config_dword(dev, pos + PCI_ERR_CAP, &config[11]);


			/* PCIe Header Log0 */
			pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG, &config[12]);
			/* PCIe Header Log4 */
			pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 4, &config[13]);
			/* PCIe Header Log8 */
			pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 8, &config[14]);
			/* PCIe Header Log12 */
			pci_read_config_dword(dev, pos + PCI_ERR_HEADER_LOG + 12, &config[15]);
		}

		errors_detected = ((config[0] & 0xF9000000)||(config[1] & 0xF9000000)||
						   (config[3] & 0x000F0000));

		if (errors_detected || verbose) {

			printk("%s: %s REGISTERS %s ",
				   __FUNCTION__, name, pci_name(dev));

			show_bridge_bus_window(dev);

			printk("  PriStatus 0x%8.8x SecStatus 0x%8.8x "
				   "BridgeCtrl 0x%8.8x DevCtrl 0x%8.8x DevCtrl2 0x%8.8x LinkStat/Ctrl 0x%8.8x\n",
				   config[0], config[1], config[2], config[3], config[4], config[20]);

			printk("  UncorrErrStatus 0x%8.8x UncorrErrMask 0x%8.8x "
				   "UncorrErrSev 0x%8.8x\n",
				   config[6], config[7], config[8]);

			printk("  CorrErrStatus 0x%8.8x CorrErrMask 0x%8.8x "
				   "AE Capabilities & Ctrl 0x%8.8x\n",
				   config[9], config[10], config[11]);

			printk("  Log0 0x%8.8x log4 0x%8.8x log8 0x%8.8x "
				   "log12 0x%8.8x\n",
				   config[12], config[13], config[14], config[15]);


			if (config[0] & 0xF9000000) {
				printk("  %s Primary status error 0x%8.8x "
					   "detected: ", name, config[0]);
				print_pci_errors(config[0]);
			}

			if (config[1] & 0xF9000000) {
				printk("  %s Secondary status error 0x%8.8x "
					   "detected: ", name, config[1]);
				print_pci_errors(config[1]);
			}

			if (config[3] & 0x000F0000) {
				printk("  %s Device status error 0x%8.8x detected: ",
					   name, config[3]);
				print_devstatus_errors(config[3]);
			}

			if (config[6]) {
				printk("  %s Uncorrectable error 0x%8.8x detected: ",
					   name, config[6]);
				print_pci_uncorrectable(config[6]);
			}

			if (config[9]) {
				printk("  %s Correctable error 0x%8.8x detected: ",
					   name, config[9]);
				print_pci_correctable(config[9]);
			}


		}

		show_bridge_mem_window(dev);
		show_bridge_bus_window(dev);

	}
}

/*
 * show_configs_pci_bridges()
 *
 *   Display the error control/status registers of a PCIE-PCI bridge device
 *
 * intputs:
 *   dev - pci_dev object of the bridge device
 *   name - String to display
 *   verbose - True = Display all the registers. False = Display
 *             only the registers that currently contain error status.
 *
 */
static void
show_configs_pci_bridges(struct pci_dev *dev, char *name, int verbose)
{
	uint pos;

	/* First, show the primary side registers */

	show_configs_bridges(dev, name, verbose);

	pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
	if (pos) {
		/* Uncorr. Err Status */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_STATUS, &config[6]);
		/* Uncorr. Err Mask */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_MASK, &config[7]);
		/* Uncorr. Err Sev */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_UNCOR_SEVER, &config[8]);

		/* Advanced Error Capabilites and Ctrl */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_CAP, &config[11]);

		/* PCIe Header Log0 */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG, &config[12]);
		/* PCIe Header Log4 */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 4, &config[13]);
		/* PCIe Header Log8 */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 8, &config[14]);
		/* PCIe Header Log12 */
		pci_read_config_dword(dev, pos + PCI_SEC_ERR_HEADER_LOG + 12, &config[15]);
	}

	if ((config[6])||(verbose)) {
		printk(KERN_ERR "  %s Uncorrectable error 0x%8.8x detected: ",
			   name, config[6]);
		print_pci_sec_uncorrectable(config[6]);

		printk(KERN_ERR "  Log0 0x%8.8x log4 0x%8.8x log8 0x%8.8x "
			   "log12 0x%8.8x\n",
			   config[12], config[13], config[14], config[15]);
	}
}
/*
 * show_configs_endpoints()
 *
 *   Display the error control/status registers of an endpoint device
 *
 * intputs:
 *   rdev - pci_dev object of an endpoint device
 *   name - String to display
 *   verbose - True = Display all the registers. False = Display
 *             only the registers that currently contain error status.
 *
 */
static void
show_configs_endpoints(struct pci_dev *dev, char *name, int verbose)
{
	uint pos, errors_detected = 0;
	bhpc_dev_error deverr;

	if (dev) {

		pci_read_config_dword(dev, PCI_COMMAND, &config[0]);  /* Status */

		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {
			/* Device Control/Status */
			pci_read_config_dword(dev, pos + PCI_EXP_DEVCTL, &config[3]);
		} else {
			config[3] = 0;
		}

		errors_detected = ((config[0] & 0xF9000000)||(config[3] & 0x000F0000));

		if (errors_detected || (verbose)) {

			printk("%s: %s REGISTERS %s\n", __FUNCTION__,
				   name, pci_name(dev));
			printk("  PriStatus 0x%8.8x DevCtrl 0x%8.8x\n",
				   config[0], config[3]);

			if (config[0] & 0xF9000000) {
				printk("  %s Primary status error 0x%8.8x "
					   "detected: ", name, config[0]);
				print_pci_errors(config[0]);
			}

			if (config[3] & 0x000F0000) {
				printk("  %s Device status error 0x%8.8x detected: ",
					   name, config[3]);
				print_devstatus_errors(config[3]);
			}


			pos = pci_find_ext_capability(dev, PCI_EXT_CAP_ID_ERR);
			if (pos) {

				/* Uncorr. Err Status */
				pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_STATUS,
									  &config[0]);
				/* Uncorr. Err Mask */
				pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_MASK,
									  &config[1]);
				/* Uncorr. Err Sev */
				pci_read_config_dword(dev, pos + PCI_ERR_UNCOR_SEVER,
									  &config[2]);

				/* Correctable Error Status */
				pci_read_config_dword(dev, pos + PCI_ERR_COR_STATUS,
									  &config[3]);
				/* Correctable Err Mask */
				pci_read_config_dword(dev, pos + PCI_ERR_COR_MASK,
									  &config[4]);

				/* Advanced Error Capabilites and Ctrl */
				pci_read_config_dword(dev, pos + PCI_ERR_CAP,
									  &config[5]);

				printk("  UncorrErrStatus 0x%8.8x UncorrErrMask 0x%8.8x "
					   "UncorrErrSev 0x%8.8x\n",
					   config[0], config[1], config[2]);

				printk("  CorrErrStatus 0x%8.8x CorrErrMask 0x%8.8x "
					   "AE Capabilities & Ctrl 0x%8.8x\n",
					   config[3], config[4], config[5]);

				display_decoded_header_log(pci_name(dev), &deverr);
			}
		}
	}
}
/*
 * get_port_type()
 *
 *   Determine the type of PCIe port
 *
 * Inputs:
 *    dev - pci_dev object of the port device
 *
 * Returns:
 *   PCI_EXP_TYPE_ENDPOINT
 *   PCI_EXP_TYPE_ROOT_PORT
 *   PCI_EXP_TYPE_UPSTREAM
 *   PCI_EXP_TYPE_DOWNSTREAM
 *   PCI_EXP_TYPE_PCI_BRIDGE
 *   PCI_TYPE_NTB
 */
static int
get_port_type(struct pci_dev *dev)
{
	int pos, type = -1;
	u16 reg16;

	if (dev) {
		pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
		if (pos) {
			pci_read_config_word(dev, pos + PCI_EXP_FLAGS, &reg16);
			type = (reg16 & PCI_EXP_FLAGS_TYPE) >> 4;
		} else {
			u16 class = 0;

			/* See if it's a PCI NTB */

			pci_read_config_word(dev, PCI_CLASS_DEVICE, &class);

			if (class == 0x680) {
				type = PCI_TYPE_NTB;
			}

		}

	}
	return (type);
}
/*
 * show_device_configs()
 *
 *   Display the PCIe error status/control registers for a
 *   specified device.
 *
 * Inputs:
 *   dev - pci_dev object of the specified device
 *   verbose - True = show all the registers. False = show only
 *             the registers that currently contain error status
 *
 */
static void 
show_device_configs(struct pci_dev *dev, int verbose)
{
	u32 type;

	if (!dev) {
		printk(KERN_ERR "%s: Error NULL input ptr\n", __FUNCTION__);
		return;
	}

	type = get_port_type(dev);
	switch (type) {

	case PCI_EXP_TYPE_ENDPOINT:
		show_configs_endpoints(dev, "ENDPOINT", verbose);
		break;

	case PCI_EXP_TYPE_ROOT_PORT:
		show_configs_rc(dev, "RC", verbose);
		break;

	case PCI_EXP_TYPE_UPSTREAM:
		if (dev->bus->number == 1) {
			show_configs_bridges(dev, "MM USP", verbose);
		} else {
			show_configs_bridges(dev, "SFM USP", verbose);
		}
		break;

	case PCI_EXP_TYPE_DOWNSTREAM:
		if (dev->bus->number == 2) {
			show_configs_bridges(dev, "MM DSP", verbose);
		} else {
			show_configs_bridges(dev, "SFM DSP", verbose);
		}
		break;

	case PCI_EXP_TYPE_PCI_BRIDGE:
		show_configs_pci_bridges(dev, "PCIE-PCI", verbose);
		break;

	default:
		printk("%s Unrecognized device type %d %s\n",
			   __FUNCTION__, type, pci_name(dev));
	}
}
/*
 * show_configs()
 *
 *   Display the PCIe error status/control registers for the
 *   entire device hierarchy.
 *
 * Inputs:
 *   verbose - True = show all the registers. False = show only
 *             the registers that currently contain error status
 *
 */
static void
show_configs(int verbose)
{
	u32 devfn;
	int domain = 0;
	struct pci_dev *dev = NULL;

	devfn = PCI_DEVFN(0, 0);
	dev = pci_get_domain_bus_and_slot(domain, 0, devfn);

	if (dev) {
		show_configs_rc(dev, "RC", verbose);
	}
	for_each_pci_dev(dev) {

		show_device_configs(dev, verbose);
	}
}

/*
 * show_pex_error_regs()
 *
 *    Display the PCIe controller error detect registers. Also
 *    display error control/status config registers of the entire
 *    device hierarchy if not in interrupt context.
 *
 *    If in interrupt context, then just display RC error
 *    control/status config registers.
 *
 * Inputs:
 *    vaddr - ioremap'd base address of the PCIe controller
 *            error detect registers
 *    domain  - Defines the PCI domain (PCI controller) to use.
 *    verbose - True = display all registers. False = display
 *              only registers with current error status
 *
 */
static int
show_pex_error_regs(volatile void __iomem *vaddr, int domain, int verbose)
{
	u32 reg00c, rege00, rege08, rege10, rege20, rege28,
		rege2c, rege30, rege34;
	u32 reg014, reg020, reg024, reg028, reg02c;
	volatile void __iomem *regaddr;
	uint errors = 0;


	/* Completion timeout value */

	regaddr = vaddr + 0x00C;
	reg00c = readl_be(regaddr);

	/* PEX config */

	regaddr = vaddr + 0x014;
	reg014 = readl_be(regaddr);

	/* PME registers */

	regaddr = vaddr + 0x020;
	reg020 = readl_be(regaddr);

	regaddr = vaddr + 0x024;
	reg024 = readl_be(regaddr);

	regaddr = vaddr + 0x028;
	reg028 = readl_be(regaddr);

	regaddr = vaddr + 0x02c;
	reg02c = readl_be(regaddr);

	/* Err Detect registers */

	regaddr = vaddr + 0xe00;
	rege00 = readl_be(regaddr);

	regaddr = vaddr + 0xe08;
	rege08 = readl_be(regaddr);

	regaddr = vaddr + 0xe10;
	rege10 = readl_be(regaddr);

	regaddr = vaddr + 0xe20;
	rege20 = readl_be(regaddr);

	regaddr = vaddr + 0xe28;
	rege28 = readl_be(regaddr);

	regaddr = vaddr + 0xe2c;
	rege2c = readl_be(regaddr);

	regaddr = vaddr + 0xe30;
	rege30 = readl_be(regaddr);

	regaddr = vaddr + 0xe34;
	rege34 = readl_be(regaddr);

	if (rege00) {
		printk("%s: PCIe error(s) detected. PEX_ERR_DR: 0x%8.8x\n",
			   __FUNCTION__, rege00);
		errors = 1;
	} else if (verbose) {
		printk("%s: PCIe error(s) NOT detected. "
			   "PEX_ERR_DR: 0x%8.8x\n", __FUNCTION__, rege00);
	}

	if ((rege00)||(verbose)) {
		printk("%s %d: ERR_DR 0x%8.8x ERR_EN 0x%8.8x "
			   "ERR_DISR 0x%8.8x ERR_CAP_STATUS 0x%8.8x "
			   "Compl Timeout 0x%8.8x\n",
			   __FUNCTION__, __LINE__,
			   rege00, rege08, rege10, rege20, reg00c);

		printk("%s %d: ERR_CAP_R0 0x%8.8x ERR_CAP_R1 0x%8.8x "
			   "ERR_CAP_R2 0x%8.8x ERR_CAP_R3 0x%8.8x\n",
			   __FUNCTION__, __LINE__,
			   rege28, rege2c, rege30, rege34);
		if (verbose) {
			display_pex_errors(domain, rege00, rege20, rege28, rege2c, rege30, rege34);
		}
	}

	if (reg020) {
		printk("%s: PCIe PME event(s) detected. PME_DETECT: 0x%8.8x\n",
			   __FUNCTION__, reg020);
		errors = 1;
	} else if (verbose) {
		printk("%s: PCIe PME event(s) NOT detected. "
			   "PME_DR: 0x%8.8x\n", __FUNCTION__, reg020);
	}

	if ((reg020)||(verbose)) {
		printk("%s: PEX_CONFIG 0x%8.8x PME_DR 0x%8.8x PME_IER 0x%8.8x "
			   "PME_DISR 0x%8.8x PEX_PMCR 0x%8.8x\n", __FUNCTION__,
			   reg014, reg020, reg028, reg024, reg02c);
	}

	if (in_interrupt()) {
		if (domain_tbl[domain].rc_dev) {
			show_configs_rc(domain_tbl[domain].rc_dev, "RC", verbose);
		} else {
			printk("%s: Cannot display PCIe config registers in interrupt context\n",
			   __FUNCTION__);
		}
	} else {
		show_configs(verbose);
	}

	return (errors);
}

/*
 * pcie_event_isr()
 *
 *   PCIe error interrupt handler.
 *
 */
static irqreturn_t pcie_event_isr(int irq, void *ptr)
{
	struct pt_regs *regs = NULL;
	u32 core, msr, pex_config_addr;
	int slot = 0;
	irqreturn_t rc = IRQ_HANDLED;
	unsigned long mcsr = mfspr(SPRN_MCSR);
	bhpc_error_data err_ddata;
	static int count = 0;
	domain_tbl_t *domain_ptr = (domain_tbl_t *)ptr;

	/* Capture this register, hopefully before something else can change it. */
	pex_config_addr = readl_be(domain_ptr->pex_base_addr);

	msr = mfmsr();
	core = mfspr(SPRN_PIR) & 0x3fff;
	regs = get_irq_regs();
	count++;

	if (domain_ptr) {
		printk(KERN_ERR "%s: ***** virq %d core %d MCSR 0x%8.8lx MSR 0x%8.8x "
			   "count %d ptr 0x%p err_irq %d domain %d\n", __FUNCTION__, irq, core,
			   mcsr, msr, count, domain_ptr, domain_ptr->err_irq, domain_ptr->domain);
	} else {

		printk(KERN_ERR "%s: ###### domain_ptr is NULL 0x%p\n", __FUNCTION__,
			   domain_ptr);
	}

	if (regs) {
		printk(KERN_INFO "%s: ***** virq %d core %d MCSR 0x%8.8lx MSR 0x%8.8x "
			   "NIP 0x%8.8lx count %d\n", __FUNCTION__, irq, core,
			   mcsr, msr, regs->nip, count);
	} else {
		printk(KERN_ERR "\n%s: ###### virq %d core %d MCSR 0x%lx MSR 0x%8.8x count %d\n",
			   __FUNCTION__, irq, core, mcsr, msr, count);
	}

	if (verbose) {
		show_pex_error_regs(domain_ptr->pex_base_addr, domain_ptr->domain, verbose);
	}

	memset(&err_ddata, 0, sizeof(bhpc_error_data));
	err_ddata.message_source = BHPC_INTERRUPT_SOURCE;
	err_ddata.err_irq = domain_ptr->err_irq;
	err_ddata.pex.config_addr = pex_config_addr;
	err_ddata.domain = domain_ptr->domain;

	/*
	 * upon return, we know everything there is to know
	 * about the error and which device caused it.
	 */
	(void) process_pcie_errors(domain_ptr->pex_base_addr, &err_ddata);
	if (err_ddata.slot < 0) {
		/*
		 * XXXX: We will always get 1 spurious IRQ when the handler is registered.
		 *       But we need to track these and disable IRQ if we exceed
		 *       a threshold within a certain time window.
		 */
		printk("%s: No error source device found %d\n", __FUNCTION__, slot);
		show_configs_rc(domain_ptr->rc_dev, "RC", DISPLAY_ALL);
	}

	/*
	 * See if we need to increment the NIP. We have to do this if
	 * the PCIe error occured on this core and the machine check
	 * didn't run.
	 */
	if ((mcsr & 0x0000A000) == 0x00002000) {
		if (verbose) {
			printk("\n%s: Before increment NIP 0x%lx core %d\n",
				   __FUNCTION__, regs->nip, core);
		}

		regs->nip += 4;

		if (verbose) {
			printk("\n%s: After increment NIP 0x%lx\n",
				   __FUNCTION__, regs->nip);
		}
	}

	if (mcsr) {
		mtspr(SPRN_MCSR, mcsr);
	}

	if (bhpc_error_notify) {
		bhpc_error_notify(&err_ddata);
	} else {
		printk(KERN_ERR "%s: NO ERROR CALLBACK REGISTERED\n", __FUNCTION__);
	}

	return (rc);
}
/*
 * find_pex_base()
 *
 *    Find and ioremap the base address of the PCIe controller memory-mapped
 *    registers from the device and fill in the domain table entry.
 *
 *  node - pointer to device tree node for the PCIe controller.
 *  domain_ptr - pointer to domain table entry for the PCIe controller.
 *
 */

void
find_pex_base(struct device_node *node, domain_tbl_t *domain_ptr)
{
	struct resource res;

	if ((node)&&(domain_ptr)) {
		of_address_to_resource(node, 0, &res);

		if (verbose) {
			printk(KERN_ERR "Found %s property address name %s start 0x%8.8llx "
				   "end 0x%8.8llx flags 0x%8.8lx\n", node->full_name,
				   res.name, res.start, res.end, res.flags);
		}

		domain_ptr->pex_base_addr = ioremap(res.start,
								(unsigned long)(res.end - res.start));

		if (verbose) {
			printk(KERN_ERR "%s pex_base_addr 0x%p\n",
				   __FUNCTION__, domain_ptr->pex_base_addr);
		}
	}
}
/*
 * pcie_init_error_init()
 *
 *   Initialization routine for the PCIe error interrupt handler
 *
 * Inputs:
 *    node - device tree node for the PCIe controller
 *    domain - specifies the domain by which the PCIe controller
 *             will be identified.
 */

int
pcie_init_error_init(struct device_node *node, int domain)
{
	int rc = 0;
	u32 core, msr;
	bhpc_error_data ddata;
	domain_tbl_t *domain_ptr = NULL;

	if (domain >= BHPC_MAX_PCI_DOMAIN) {
		printk(KERN_ERR "%s: Invalid PCIe domain %d. Should be 0-%d\n",
			   __FUNCTION__, domain, (BHPC_MAX_PCI_DOMAIN - 1));
		rc = -1;
		goto Exit;
	}

	domain_ptr = &domain_tbl[domain];
	domain_ptr->domain = domain;
	domain_ptr->err_irq = of_irq_to_resource(node, 0, NULL);

	core = mfspr(SPRN_PIR) & 0x3fff;

	printk(KERN_INFO "%s core %d: PCIe error irq mapped to %d\n",
		   __FUNCTION__, core, domain_ptr->err_irq);

	if (domain_ptr->err_irq < 0) {
		rc = -1;
		goto Exit;
	}

	find_pex_base(node, domain_ptr);
	if (domain_ptr->pex_base_addr) {
		memset(&ddata, 0, sizeof(bhpc_error_data));
		ddata.message_source = BHPC_INIT_SOURCE;
		clear_pcie_errors(&ddata, 1); // Clear any existing status
	} else {
		printk(KERN_ERR "%s: Could not map pex registers\n", __FUNCTION__);
		rc = -1;
		goto Exit;
	}

	bhpc_setup_configs(NULL, domain_ptr->domain, 1);
	setup_pcie(domain_ptr->pex_base_addr, 1);
	if ((domain_ptr->pex_base_addr)&&(verbose)) {
		show_pex_error_regs(domain_ptr->pex_base_addr,
							domain_ptr->domain, DISPLAY_ALL);
	}

	if (verbose) {
		printk("%s: Calling request_irq for IRQ %d\n", __FUNCTION__,
			   domain_ptr->err_irq);
	}

	silkworm_machine_check_control(1, 0);

	rc = request_irq(domain_ptr->err_irq, pcie_event_isr, IRQF_SHARED,
					 "pcie-err", domain_ptr);
	if (rc) {
		printk(KERN_ERR "%s: Call to request_irq failed %d\n",
			   __FUNCTION__, rc);
		goto Exit;
	}

	irqs_enabled = 1;

	if (verbose) {
		msr = mfmsr();
		printk("%s Original MSR 0x%8.8x\n", __FUNCTION__, msr);
	}

Exit:

	return (rc);
}
/****************************************************************/
/************************ /PROC FILESYSTEM CODE *****************/
/****************************************************************/
/*
 * write_pcie_irq_regdump()
 *
 *  Dump registers on request. Just dumps to stdout.
 *
 */
static int
write_pcie_irq_regdump(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];
	int regdump = 0, domain;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(ptr, buffer, count)) {
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return (-EFAULT);
	}

	if (!sscanf(ptr, "%d", &regdump)) {
		return (-EFAULT);
	};


	if (regdump > 0) {
		volatile void __iomem *pex_base_addr;
		for (domain = 0; domain < BHPC_MAX_PCI_DOMAIN; domain++) {

			pex_base_addr = domain_tbl[domain].pex_base_addr;
			if (pex_base_addr) {
				printk("\n%s %d: DUMP PCIE ERROR REGISTERS BEGIN\n",
					   __FUNCTION__, __LINE__);
				show_pex_error_regs(pex_base_addr, domain, verbose);
				printk("\n%s: DUMP PCIE ERROR REGISTERS COMPLETE\n", __FUNCTION__);
			} else {
				printk("%s: Cannot dump error regs. No pex_base_addr\n",
					   __FUNCTION__);
			}
		}
	}
	return (count);
}
/*
 * write_pcie_irq_table()
 *
 *  Dump DSP table on request. Just dumps to stdout.
 *
 */
static int
write_pcie_irq_table(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];
	int regdump = 0;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(ptr, buffer, count)) {
		printk(KERN_ERR "%s: copy_from_user failed.\n", __FUNCTION__);
		return (-EFAULT);
	}

	if (!sscanf(ptr, "%d", &regdump)) {
		return (-EFAULT);
	};


	if (regdump > 0) {
		int i;
		dsp_dev_map *ptr = NULL;
		subrange *sr_ptr = NULL;

		printk(KERN_ERR "Max Entries %d Base bus %d\n",
			   default_dsp_table->max_entries,
			   default_dsp_table->base_bus);

		for (i = 0; i < num_devices_entries; i++) {
			ptr = &devices[i];
			if (ptr) {
				printk(KERN_ERR "Domain %d Bus %d dev %d virtual slot %d idsel 0x%8.8x\n",
					   ptr->domain, ptr->bus, ptr->dev, ptr->slot, ptr->idsel);
				printk(KERN_ERR "  First bus %d Last bus %d\n", ptr->first_bus, ptr->last_bus);
				if (ptr->bus_subranges.number > 0) {
					int x;
					for (x = 0; x < ptr->bus_subranges.number; x++) {
						sr_ptr = &ptr->bus_subranges.subranges[x];
						printk(KERN_ERR "    Bus subrange %d slot %d devid %d idesl 0x%8.8x "
							   "base %d limit %d\n", x, sr_ptr->slot, sr_ptr->devid, sr_ptr->idsel,
							   sr_ptr->base, sr_ptr->limit);
					}
				}
				printk(KERN_ERR "  MemBase 0x%8.8x MemLimit 0x%8.8x\n",
					   ptr->mem_base, ptr->mem_limit);
				if (ptr->mem_subranges.number > 0) {
					int x;
					for (x = 0; x < ptr->mem_subranges.number; x++) {
						sr_ptr = &ptr->mem_subranges.subranges[x];
						printk(KERN_ERR "    Mem subrange %d slot %d devid %d idesl 0x%8.8x "
							   "base 0x%8.8x limit 0x%8.8x\n", x, sr_ptr->slot, sr_ptr->devid,
							   sr_ptr->idsel, sr_ptr->base, sr_ptr->limit);
					}
				}
			}				   

		}
	}
	return (count);
}
/*
 * write_pcie_verbose()
 *
 *   Enable/Disable verbose mode
 *
 */
static int
write_pcie_verbose(struct file *file, const char *buffer,
			     unsigned long count, void *data)
{
	char ptr[PAGE_SIZE];

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(ptr, buffer, count)) {
		printk("%s: copy_from_user failed.\n", __FUNCTION__);
		return (-EFAULT);
	}

	if (!sscanf(ptr, "%d", &verbose)) {
		return (-EFAULT);
	};

	return (count);
}
/*
 * read_pcie_verbose()
 *
 *  Display state of verbose mode flag. 1 = enabled, 0 = disabled
 *
 */
static int
read_pcie_verbose(char *page, char **start, off_t off,
								   int count, int *eof, void *private)
{
	int len;
	char *out = page;

	len = snprintf(out, count, "%d\n", verbose);

	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) {
			return (0);
		}
	} else {
		len = count;
	}
	*start = page + off;

	return (len);
}

/*
 * Create various /proc file system entries:
 *
 *  Dump error-related registers
 *  Enable/Disable verbose mode
 *
 */

static struct proc_dir_entry *pcie_irq;
static struct proc_dir_entry *pcie_irq_regdump;
static struct proc_dir_entry *pcie_irq_verbose;
static struct proc_dir_entry *pcie_irq_table;

void
proc_init(void)
{

	pcie_irq = proc_mkdir("pcie_irq", NULL);
	if (pcie_irq == NULL) {
		printk(KERN_ERR "%s: Could not create proc dir pcie_irq\n", __FUNCTION__);
		return;
	}

	pcie_irq_regdump = create_proc_entry("regdump", 0, pcie_irq);
	if (!pcie_irq_regdump) {
		printk(KERN_ERR "%s: Could not create proc file pcie_irq/regdump\n",
			   __FUNCTION__);
		return;
	}

	pcie_irq_verbose = create_proc_entry("verbose", 0, pcie_irq);
	if (!pcie_irq_verbose) {
		printk(KERN_ERR "%s: Could not create proc file pcie_irq/verbose\n",
			   __FUNCTION__);
		return;
	}

	pcie_irq_table = create_proc_entry("table", 0, pcie_irq);
	if (!pcie_irq_table) {
		printk(KERN_ERR "%s: Could not create proc file pcie_irq/table\n",
			   __FUNCTION__);
		return;
	}

	pcie_irq_regdump->write_proc = &write_pcie_irq_regdump;

	pcie_irq_verbose->read_proc = &read_pcie_verbose;
	pcie_irq_verbose->write_proc = &write_pcie_verbose;

	pcie_irq_table->write_proc = &write_pcie_irq_table;
}

/*
 *
 */
void
proc_cleanup(void)
{

	if (pcie_irq_verbose) {
		remove_proc_entry("verbose", pcie_irq);
	}

	if (pcie_irq_regdump) {
		remove_proc_entry("regdump", pcie_irq);
	}


	if (pcie_irq) {
		remove_proc_entry("pcie_irq", NULL);
	}
}
/*
 * bhpc_irq_init()
 *
 *  Module init function.
 *
 */
int
bhpc_irq_init(void (*error_fn)(bhpc_error_data *ddata), bhpc_dsp_map_table *dsp_ptr)
{
	struct device_node *dn;
	u_int  type;
	int domain = 0, i = 0;
	struct pci_dev *dev = NULL;

	if (error_fn) {
		bhpc_error_notify = error_fn;
	} else {
		printk(KERN_ERR "%s: INVALID ERROR CALLBACK PTR\n", __FUNCTION__);
	}
	/*
	 * Find all the pci_dev objects for the MM PCIe switch ports and the RC.
	 * Cache them so we can use them when an IRQ occurs. Linux does not
	 * allow searching pci_dev lists in interrupt context.
	 */

	if (dsp_ptr) {
		devices = dsp_ptr->dsp_map;
		num_devices_entries = dsp_ptr->max_entries;
		default_dsp_table = dsp_ptr;
		printk(KERN_ERR "%s: Over-ride default device map. Size %d\n",
			   __FUNCTION__, num_devices_entries);
	} else {
		devices = &dsp_table.dsp_map[0];
		num_devices_entries = dsp_table.max_entries;
		default_dsp_table = &dsp_table;
		printk(KERN_ERR "%s: Using default device map. Size %d\n",
			   __FUNCTION__, num_devices_entries);
	}

	/* Make sure all domain table entries are defaulted to empty */
	for (i = 0; i < BHPC_MAX_PCI_DOMAIN; i++) {
		domain_tbl[i].pex_base_addr = NULL;
		domain_tbl[i].err_irq = -1;
		domain_tbl[i].rc_dev = NULL;
		domain_tbl[i].domain = -1;
	}

	/*
	 * Parse device tree to see which PCIe domains support error handling and
	 * fill in their domain table entries and init the HW in that domain.
	 */
	i = 0;
	for_each_compatible_node(dn, NULL, "fsl,p4080-pcie") {
		if (of_get_property(dn, "pciehp-bhpc-irq", NULL) != NULL) { 

			if (i >= BHPC_MAX_PCI_DOMAIN) {
				printk(KERN_ERR "%s: Too many PCIe domains defined %d\n",
					   __FUNCTION__, i);
				break;
			}
			for_each_pci_dev(dev) {

				domain = pci_domain_nr(dev->bus);

				if (domain != i) {
					printk(KERN_ERR "%s: %s: IGNORE DEVICES IN DOMAIN %d\n",
						   __FUNCTION__, pci_name(dev), domain);
					continue;
				}

				type = get_port_type(dev);
				switch (type) {

				case PCI_EXP_TYPE_ENDPOINT: /* Don't care about endpoints */
					break;

				case PCI_EXP_TYPE_ROOT_PORT:
					domain_tbl[domain].rc_dev = dev;
					/* Intentional fall-thru */

				case PCI_EXP_TYPE_UPSTREAM:
				case PCI_EXP_TYPE_DOWNSTREAM:
				case PCI_EXP_TYPE_PCI_BRIDGE:
				case PCI_TYPE_NTB:
					store_dev(domain, dev->bus->number, PCI_SLOT(dev->devfn),
							  type, dev);
					break;

				default:
					printk(KERN_ERR "%s Unrecognized device type %d %s\n",
						   __FUNCTION__, type, pci_name(dev));
				}

			} /* for_each_pci_dev() */

			if (!pcie_init_error_init(dn, i)) {
				printk(KERN_ERR "%s: Pcie domain %d err interrupt "
					   "handler Installed\n", __FUNCTION__, i);
			} else {
				printk(KERN_ERR "%s: Pcie domain %d err interrupt handler "
					   "NOT installed\n", __FUNCTION__, i);
			}

		} else { /* of_get_property() */
			printk(KERN_ERR "%s: PCIe domain %d does not support error hander\n",
				   __FUNCTION__, i);
		}

		i++;

	} /* for_each_compatible_node() */

	proc_init();

	return (0);
}
EXPORT_SYMBOL(bhpc_irq_init);

void
bhpc_irq_cleanup(void)
{
	int i;
	domain_tbl_t *domain_ptr = NULL;

	printk("%s: pcie_err Unloaded\n", __FUNCTION__);

	for (i = 0; i < BHPC_MAX_PCI_DOMAIN; i++) {
		domain_ptr = &domain_tbl[i];

		if (domain_ptr->err_irq >= 0) {
			free_irq(domain_ptr->err_irq, domain_ptr);
			printk(KERN_ERR "%s: pcie_err Unloading\n", __FUNCTION__);
		}
		if (domain_ptr->pex_base_addr) {
			iounmap(domain_ptr->pex_base_addr);
		}
	}
	silkworm_machine_check_control(0, 0);
	proc_cleanup();
}
